#!/usr/bin/perl -w

# ArrayPipe - a microarray data processing pipeline

# developed by the Brinkman Laboratory, Simon Fraser University,
# Burnaby, BC, Canada for the Genome Canada Pathogenomics project 
# (http://www.pathogenomics.ca/)
# June 2003 - October 2004

my $license = "

#    ArrayPipe - a microarray data processing pipeline
#    Copyright (C) 2004 Karsten Hokamp and Fiona Brinkman,  
#    all rights reserved.
#
#   This program is free software; you can redistribute it and/or
#   modify it under the terms of the Perl Artistic License
#   as published by Larry Wall
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#   Artistic License for more details.
#
#   You should have received a copy of the Artistic License with this program,
#   in the file named 'LICENCE.txt'. If not, We'll be glad to provide
#   one.
#
#   For more information on ArrayPipe
#   see http://www.pathogenomics.ca/arraypipe
#

";

my $revision_number = '$Revision: 1.4 $'; # '  $Date: 2004/10/13 17:16:53 $
$revision_number =~ s/[\$Revision\: | \$]//g;

my $help_message = "

#  HELP MESSAGE FOR ARRAYPIPE ($revision_number)
#  
#  USAGE: arraypipe.pl [--help|--copy|--version|--listing [action]]
#         or
#         arraypipe.pl --BATCH [OPTIONS]
#  
#  OPTIONS:
#  
#    GENERAL:
#      --help
#      --copy
#      --version
#      --listing          = list all functions and filters
#      --listing <string> = list all variables for a certain action module
#
#    PROCESSING:
#      --config <string>  = location of configuration file
#      --action <string>  = one or more (comma separated) actions to carry out
#      --profile <string> = name of action profile to carry out
#      --file <string>    = one or more (comma separated) files, files of files, 
#                           or directory names containing these for data input
#      --list <string>    = one or more (comma separated) lists, files of lists, 
#                           or directory names containing these for data input
#      --out <string>     = name of output directory

";

# the aim of this tool is to set up a pipeline with functions 
# (or modules) for each sub-task
# this will allow flexible arrangement of transformation and 
# visualization steps
# the sub-tasks are divided into actions (visualization of data) 
# and filters (transformation of data)

# the following main data storage is being used:
# %data{file}{__DATA__}{label}{header}{mrow}{mcol}{row}{col} = value
# (the __DATA__ key was introduced so that other keys could be added
#  in the future)
#
# the data hash is organized by files
# for each file there will be data section (__DATA__), generated 
# from parsing the input file (see sub read_data)
# additionally, a storage location is provided for each format in 
# which the data has been written to disk (which might be required 
# for modules that are called via system calls)
#
# the data section contains values for each specified column for each 
# label for each spot
# the spots are organized in meta rows, meta columns, rows and columns 
# (see variables below for definition of chip- and subgrid dimensions)
#
# the ArrayVision software organizes the output in a number of columns 
# whose headers are duplicated for each label (Ag,Au), e.g.:
#
#line 1: |      |   Ag   |   Ag     |   Ag    |   Au   |   Au     |
#        |------+--------+----------+---------+--------+----------+
#line 2: |Labels|MTM Dens|MAD-Levels|SD-Levels|MTM Dens|MAD-Levels|
#
# a special label '__NO_LABEL__' is used for columns that are not 
# label-specific, such as 'Spot labels'
#
# each module needs to specify the columns on which it needs to carry out 
# its function
# these should be returned when the according function is called with 
# '__columns__' e.g.: @{$columns{$module}} = &function('__columns__');
#
# for each specified action/filter item the required columns are 
# collected and passed on to the &read_data function to acquire all 
# necessary data upfront; 
# this brings the reading of the file down to a single pass and also 
# reduces the memory space for holding unnecessary data
#
# all data is stored in the %data hash, which is passed on to the 
# function by reference
# the function cycles through each file (first level keys) and can 
# retrieve the required data for each spot from the label/header keys
# if the module can't read the data directly, it can be stored in 
# a temporary location in a specified format, where it will be 
# available for subsequent functions
#
# the return string from each module called in action mode
# should be in HTML format that can be written to a file
#
# filters carry out their transformations directly on the values 
# in the %data hash
#
# to make this program independent of the specific headers used in 
# the arrayvision output it would be desirable to design a mapping 
# mechanism that matches a set of predefined terms to the 
# according header, e.g.:
# FG MEDIAN - sMTMDens
# BG MEDIAN - Bkgd
# FLAG - Flag
# etc.
# but we can leave this for a later exercise...
# Oct 21st, 2003: DONE!

# ADDING A FUNCTION:
#
# - locate the hash %action_items below
#
# - add a key-value pair (keyword - short description) to the 
#   action_items, e.g.: 
#   'action_R_histogram', 'Plot R histogram'
#   (starting the action items with 'action_' and the filter items
#   with 'filter_' is just a convention to recognize them easier)
#
# - add a key-value pair (same keyword - long description) to
#   the %item_description hash below, e.g.:
#   'action_R_histogram','Plot histogram of foreground signal through R'
#
# - add a short subroutine with the keyword as the name, e.g.:
#   sub action_R_histogram {
#     # each subroutine gets four arguments:
#     my $num = shift;    # the number in the action chain
#                         # (useful for action specific parameters)
#     my $file = shift;   # the name of the file containing
#                         # the microarray data
#     my $label = shift;  # the label to work on (Ag or Au)
#     my $action = shift; # the name of the action 
#                         # (same as subroutine name)
#  
#      # return list with the headers required for this function
#      # when called with special argument
#      if ($num eq '__columns__') {
#  	return 'MTM Dens - Levels';
#      }
#  
#      # put all the main calculations into a separate subroutine
#      # the return value should be HTML code that will be directly
#      # included into the results page
#      print &R_histogram_plotting($num,$file,$label,'MTM Dens - Levels',$action);
#   }
#
# - add key to function hash, e.g.:
#   $function{'action_R_histogram'} = \&action_R_histogram;
#
# - define rest of your function somewhere below
# - done!


# TO DO LIST:
# - if a data column is chosen for both foreground and background, the MA plots don't work
# - allow compressed archives as example files
# - modify get_array_specs, so that it doesn't report grids that are not used (when debugging)
# - applying flag other than 'list' doesn't seem to work in filter_by_value
# - make a Refine that works per action step and not per file
# - skipped files don't seem to be subtracted from time estimation
# - add an 'auto' option to filter by value, which is 'all' when only one treatment group is there and separately when we have multiple treatment groups (another distinction should be 'all within treatment group' in case we haven't merged all replicates yet)
# - show settings in the order they were chosen in action setup
# - when loading an action setup, make sure the according settings will be shown
# - reduce report of waiting time by half for data that consists of two files per array
# - if no flags were selected in saved file, don't select any when reading in file
# - add a box for the sorting to ignore sign
# - parameter for log_transfer in draw_chip
# - multiple successive merging on the same type results in data loss (one module seems to delete everything before doing the merging, check for in- and out-header being the same, although, since a new file is created it should work there! Must be another problem...)
# - zip files with the same name as one of the data files cause problems!
# - use whiskers to determine threshold for chip visualization 
# - histogram of ratios
# - BG correction of partial subset will be ignored at normalization?
# - carry out interscatter plot only within a group (or technical replicates) and optionally for intensities, between ratios by default
# - delete unused spots after merging (or at least give the option to do so)
# - box plot with ratios doesn't seem to work properly
# - changing group for user apache not permitted?
# - create a variable for the .htaccess file that will be put into public dirs.
# - allow copying and pasting of settings between different spreadsheets (via cookies?)
# - omit message of partial results when refreshing for spreadsheet output
# - imagene files starting with the label identifier (Cy3,etc) will not be recognized properly, i.e. array identifier should be first!
# - put label onto y-axis of box-plots
# - filter by value: if no values passed through, don't save list or up number of flag
# - can't seem to download http://www.pseudomonas.com/download/UnionTableU_10-Jun-2004.tab for annotation of Pseudomonas slides
# - use different first page if one spreadsheet exists already (fixed caching for the moment - test with IExplorer)
# - fix transpath
# - store uploaded data in subdirctory of project and make it available just like example data!
# - when changing output names, do a replacement of all previous directories to the new directory!
# - give proper warning when 'Do it!' is pressed without an action selected
# - check why refine doesn't work for t-tests
# - when listing results from different treatments, do not include control when sorting alphabetically
# - when loading a file into spreadsheet - add a button to clear all settings
# - allow saving of data set (in private accounts)
# - run spreadsheet from clients and not from koch
# - after loess, the graphic looks as if it is displaying background values
# - add more documentation to chip visualization options
# - figure out the best way to present the names of both Imagene files in the output
# - if a module produces many output sections (one per file) include a next button
# - for (de)selection of columns, show two boxes, one with selected, one with deselected columns
# - make assignment of list flags flexible (wherever 'list1' is assigned)
# - do the refine only per action step, not per file
# - make button to save created spread-sheet under a name
# - add red stars to data file fields that need to be filled in
# - link hugo names to GeneLynx: http://human.genelynx.org/cgi-bin/a?page=home
# - allow an array design to be specified during data load, which will be applied to all matching files
# - signal scatter plot by subgrid doesn't work for Jin et al files (auto, MROW)
# - list the total amount of spots included in histogram
# - plus sign in filename causes problems
# - when picking a different channel (e.g. for foreground signal) in a two file format such as ImaGene, then this is forced upon the second file as well
# - when selecting gene and gene hits (filtering for gene hits = 1) and then downloading the list, it doesn't contain gene hits
# - web-based spreadsheet: when filtering and then changing the number of rows shown, the filter doesn't work anymore
# - add a module that allows to start working on previous data level (e.g. go back to FG_COR)
# - bug in signal scatter plot: load a file with 4 subgrids and choose to print 3 per plot - last one is not printed and border is too high
# - add option to translate certain flag columns from values to words
# - avoid multiple upload of annotation file (checkbox to force overwrite)
# - let the user select which part of the results should be written to separate pages...
# - make use of precompiled annotation file or FuncDB optional
# - warning comes up when saving settings
# - make 'Retrieve Data' a default in the action list
# - make &die that reports back to the web (e.g. in &list_files)
# - allow user to specify flag for each list
# - make it possible to get intersection from all - x files that have common genes
# - change order of merged file
# - make generated lists available for inclusion (similar as example files)
# - generate comment in first line of list file (which can be used as explanation for point above)
# - if merged file is based on logg-ed values, do also provide the unlogged values
# - add GO plotter to HTML output
# - make ratio plot green and red (or yellow and blue)
# - avoid creating directory with help file when producing login page (no submit)
# - when renaming directory make sure all file paths are changed as well (except for example files)
# - when duplicate neighbours are listed, give them the same colour in spreadsheet
# - add function for showing flags in chip with white or black background
# - apply rotation suggested by Dana
# - make output available as compressed tar-file
# - die if array layout wasn't specified
# - text field to insert info about array
# - if reporter list specifier is out of array limit, this will go unnoticed
# - if files are unloaded and loaded, they keep the array layout of the previous one!
# - make settings savings public or private
# - create a file repository for registered users where all the uploads go to
# - if no data from one file upload was retrieved, the whole process dies
# - make an overall slide overview function (integrated into draw_image)
# - allow splitting of output file into action sections
# - add a section for general settings (e.g. image size) in the Settings section
# - list execution times for each module
# - add a checkbox beside each action item to temporarily disable step
# - in list comparison of multiple files list flags for each single file as well
# - make list of markers flexible (all spots that are labeled more than MROW*MCOL times)
# - add 'all' to skip flags in HTML output
# - time calculation overestimates for two-file data set (ImaGene)
# - refine time calculation (partial files, less input lines, etc. , e.g. base it on number of spots)
# - in URL FLAG_file is added repeatedly
# - avoid break of lines in column specifications
# - warn if grid doesn't match number of lines
# - return error if grid settings don't make sense or no channels specified!
# - allow filtering on multiple headers at the same time (difficulty: combine with AND or OR?)
# - color selection in Z-score sometimes doesn't work
# - printing of ratios after MEV LOESS doesn't work
# - if output skips flags, annotate only those that will get through
# - make tar'ed archive available for download
# - add hr between output of multiple files
# - update index file with size and date of latest results in sub finalize
# - Load files and Do it doesn't work with lists
# - when merging have strict option (drop gene if one value is missing)
# - for merging have strict settings to skip all points that have at least one flag set
# - for sim chip: get colour gradient only from those that aren't flagged (or maybe not)
# - only redo steps if parameter changed which affect data
# - print physical measurement limits into MA plot
# - send e-mail when finished (to be activated in temporary reload page)
# - skip empty fields doesn't work well in output module
# - if ($ch2) in box plot might cause problems (only $val1 is set, tests for both vals will fail)
# - add histogram of dup_diff's
# - tries to open tmp and example directory before config is read in
# - have a special exit_warnings routine
# - add button to kill processes
# - wrap image header around if wider than image
# - beside the required data header, add a desired data header
# - read in file format with multiple sets returned by arraypipe
# - when unloading a file, make sure it is not ticked anymore in the selection list

use strict;
use Cwd;
use GD;
#use CGI qw/:standard -nosticky/;
use CGI qw/:standard/;
use CGI::Carp 'fatalsToBrowser';
use File::Basename;
use File::Path;
use File::Copy;
use Getopt::Long;

use constant PI => 3.1415926536;
use constant SIGNIFICANT => 5; # number of significant digits to be returned

#############################
### VARIABLE DECLARATIONS ###
#############################

# avoid STDOUT buffering
$| = 1;

my $FAILURE = 1;
my $SUCCESS = 0;

# grab command line arguments first to avoid them being used by $cgi object
# (since we have our own routine to deal with that)
my @CMD_ARG = @ARGV;

my $batch_processing = '';
my $modules_listing = 0;
if (@CMD_ARG) {
    # -BATCH is the trigger for command line only processing
    # if this or a help switch is specified, read in command line arguments
    if (grep /^-{0,2}BATCH/, @CMD_ARG
	or
	grep /^-{1,2}h/, @CMD_ARG
	or
	grep /^-{1,2}c/, @CMD_ARG
	or
	grep /^-{1,2}v/, @CMD_ARG
	or
	grep /^-{1,2}l/, @CMD_ARG
	or
	grep /^-{1,2}\?/, @CMD_ARG) {
	&process_command_line;
    }    
}
@ARGV = ();

my $cgi = new CGI;

if (defined $cgi->param('submit')
    and
    $cgi->param('submit') eq 'Kill Processes!') {
    &kill_processes; # EXPERIMENTAL!
}

my $FPMI_specific = 0;  # some features are tailor-made for the FPMI project
                        # (see www.pathogenomics.ca)
my %section_view = ();
$section_view{'SETTINGS'} = 'show all';
#$section_view{'SETTINGS'} = 'Show used settings!';
$section_view{'SETTINGS'} = 'show used';
my $from_dir = '';
my $to_dir = '';
my $clean_up_files = 1;
my $individual_results_pages = 0;
my $webstart = 1;
my $second_call = '';
my $first_call = 1;
my $new_results_file = '';

if (@CMD_ARG
    or
    $batch_processing) {
    &set_cgi_params;
    $webstart = 0;
}

my $progname = basename $0;
my $tmp_dir = '';  # set this in the configuration file
my $tmp_dir_html = '';  # set this in the configuration file
my $results_file = '';
my $results_file_html = '';
my $save_settings_dir = '';  # set this in the configuration file
my $remote_user = '';  # set this in the configuration file
my $remote_config_file = '';  # set this in the configuration file
my $public_username = 'public';  # overwrite this in the configuration file
my $home_link = "http://www.arraypipe.ca";  # overwrite this in the configuration file
my $cgi_link = "http://www.arraypipe.ca/cgi-bin/start";  # overwrite this in the configuration file
my $user_instructions_link = "http://www.arraypipe.ca/instructions.html";  # overwrite this in the configuration file
my $project_logo = '';  # set this in the configuration file
my $operator_email = 'please_define@config.file'; # overwrite this in the configuration file
my $tutorial_link = "http://koch.pathogenomics.ca/arraypipe/tutorial";  # overwrite this in the configuration file
my $verbose = 0;
my $lynx_location = '';  # set this in the configuration file
my $allow_web_file_load = '';  # set this in the configuration file
my $check_example_files = 0;  # overwrite this in the configuration file
my $help_file_URL = '';

# the main data hash
my %data = ();
my %swissprot = ();
my %file_specs = ();
my %list_specs = ();
my $web_link_annotation_file = '';
my %first_data_file = ();
my %first_data_file_name = ();
my %second_data_file = ();
my %list_flag_match = ();
my %annotation = ();
my @annotation_headers_all = ();
my %annotation_done = ();
my %annotation_file_read = ();
my %annotation_select_probelynx = ();
my %annotation_select_funcdb = ();
my %annotation_hostdb_id = ();
my %annotation_links = (
			'HUGO accession', 'http://bioinfo.weizmann.ac.il/cards-bin/cardsearch.pl?search=__REPLACE__',
			'gene symbol', 'http://bioinfo.weizmann.ac.il/cards-bin/cardsearch.pl?search=__REPLACE__',
			'GO', 'http://www.ensembl.org/Homo_sapiens/goview?depth=2&query=__REPLACE__',
			'go description', 'http://www.ensembl.org/Homo_sapiens/goview?depth=2&query=__REPLACE__',
			'go term', 'http://www.ensembl.org/Homo_sapiens/goview?acc=__REPLACE__',
#			'TransPath', 'http://www.brinkman.mbb.sfu.ca/cgi-bin/biobase/transpath/4.4/bin/graph.cgi?out=text%2Fhtml&dstdown1=2&style=HMAP&mode=1&id=__REPLACE__',
			'BioCarta', 'http://www.biocarta.com/pathfiles/h___REPLACE__.asp',
			'InterPro ID', 'http://www.ebi.ac.uk/interpro/IEntry?ac=__REPLACE__',
			'interpro interpro id', 'http://www.ebi.ac.uk/interpro/IEntry?ac=__REPLACE__',
			'LocusLink accession', 'http://www.ncbi.nlm.nih.gov/LocusLink/LocRpt.cgi?l=__REPLACE__',
			'EMBL', 'http://www.ebi.ac.uk/cgi-bin/emblfetch?__REPLACE__',
			'EMBL accession', 'http://www.ebi.ac.uk/cgi-bin/emblfetch?__REPLACE__',
			'emb', 'http://www.ebi.ac.uk/cgi-bin/emblfetch?__REPLACE__',
#			'protein', 'http://www.ncbi.nlm.nih.gov:80/entrez/query.fcgi?cmd=Search&db=Protein&term=__REPLACE__[accn]&doptcmdl=GenPept',
			'protein', 'http://www.ncbi.nlm.nih.gov/entrez/viewer.fcgi?val=__REPLACE__',
			'gb', 'http://www.ncbi.nlm.nih.gov/entrez/viewer.fcgi?val=__REPLACE__',
			'dbj', 'http://www.ncbi.nlm.nih.gov/entrez/viewer.fcgi?val=__REPLACE__',
			'GP', 'http://www.ncbi.nlm.nih.gov/entrez/viewer.fcgi?db=protein&val=__REPLACE__',
			'OMIM accession', 'http://www.ncbi.nlm.nih.gov/entrez/dispomim.cgi?id=__REPLACE__',
			'RefSeq accession', 'http://srs.sanger.ac.uk/srsbin/cgi-bin/wgetz?-e+[REFSEQ-ID:__REPLACE__]',
			'SPTREMBL accession', 'http://srs6.ebi.ac.uk/srsbin/cgi-bin/wgetz?-e+[libs%3d{SWALL_SP_REMTREMBL}-id:__REPLACE__*]|[libs-acc:__REPLACE__*]',
			'SWISSPROT accession', 'http://srs6.ebi.ac.uk/srsbin/cgi-bin/wgetz?-e+[libs%3d{SWALL_SP_REMTREMBL}-id:__REPLACE__*]|[libs-acc:__REPLACE__*]',
			'RefSeq', 'http://srs.sanger.ac.uk/srsbin/cgi-bin/wgetz?-e+[REFSEQ-ID:__REPLACE__]',
			'SPTREMBL', 'http://srs6.ebi.ac.uk/srsbin/cgi-bin/wgetz?-e+[libs%3d{SWALL_SP_REMTREMBL}-id:__REPLACE__*]|[libs-acc:__REPLACE__*]',
			'SWISSPROT', 'http://srs6.ebi.ac.uk/srsbin/cgi-bin/wgetz?-e+[libs%3d{SWALL_SP_REMTREMBL}-id:__REPLACE__*]|[libs-acc:__REPLACE__*]',
			'SP', 'http://srs6.ebi.ac.uk/srsbin/cgi-bin/wgetz?-e+[libs%3d{SWALL_SP_REMTREMBL}-id:__REPLACE__*]|[libs-acc:__REPLACE__*]',
			'PDB accession', 'http://www.rcsb.org/pdb/cgi/explore.cgi?pdbId=__REPLACE__',
			'TRANSCRIPT', 'http://www.ensembl.org/Homo_sapiens/transview?transcript=__REPLACE__',
			'ORTHOLOGUE', 'http://www.ensembl.org/Homo_sapiens/transview?transcript=__REPLACE__',
			'PIR', 'http://pir.georgetown.edu/cgi-bin/nbrfget?uid=__REPLACE__',
			'pdb', 'http://www.rcsb.org/pdb/cgi/explore.cgi?pdbId=__REPLACE__',
			'TC', 'http://www.tigr.org/docs/tigr-scripts/tgi/tc_report.pl?tc=__REPLACE__&species=Cattle',
			'putative ortholog hit identifier', 'http://www.tigr.org/docs/tigr-scripts/tgi/tc_report.pl?tc=__REPLACE__&species=Human',
			'putative ortholog TOG', 'http://www.tigr.org/tigr-scripts/tgi/ego/ego_report.pl?ego=__REPLACE__',
			'DIP', 'http://dip.doe-mbi.ucla.edu/dip/DIPview.cgi?PK=__REPLACE__',
			'BIND', 'http://bind.ca/cgi-bin/dataget?rectype=28&submit=Go&get=search&id=__REPLACE__',
			'TransFac Factor', 'http://www.brinkman.mbb.sfu.ca/cgi-bin/biobase/transfac/7.4/bin/getTFProf.cgi?__REPLACE__',
			'TransFac Gene', 'http://www.brinkman.mbb.sfu.ca/cgi-bin/biobase/transfac/7.4/bin/getTFProf.cgi?__REPLACE__',
			'gene','http://www.ensembl.org/Homo_sapiens/geneview?gene=__REPLACE__',
			'THC ENSEMBL TRANSCRIPT','http://www.ensembl.org/Homo_sapiens/transview?transcript=__REPLACE__',
			'THC ENSEMBL GENE','http://www.ensembl.org/Homo_sapiens/geneview?gene=__REPLACE__',
			'putative ortholog Ensembl transcript','http://www.ensembl.org/Homo_sapiens/transview?transcript=__REPLACE__',
			'putative ortholog Ensembl gene','http://www.ensembl.org/Homo_sapiens/geneview?gene=__REPLACE__',
			'transcript','http://www.ensembl.org/Homo_sapiens/transview?transcript=__REPLACE__',
#			'proteinfamily family id','http://www.ensembl.org/Homo_sapiens/familyview?family=__REPLACE__',
#			'proteinfamily family description','http://www.ensembl.org/Homo_sapiens/textview?type=Family&q=__REPLACE__',
			
);

my %annotation_links_alternatives = (
'domain accession number' => {
    '^PF' => 'http://www.sanger.ac.uk/cgi-bin/Pfam/getacc?__REPLACE__',
#    '^PS' => 'http://www.isrec.isb-sib.ch/cgi-bin/get_pstprf?__REPLACE__',
    '^PS' => 'http://us.expasy.org/cgi-bin/prosite-search-ac?__REPLACE__',
    '^PR' => 'http://blocks.fhcrc.org/blocks-bin/getblock.sh?__REPLACE__'
    },
'Ensembl accession' => {
    '^ENSG' => 'http://www.ensembl.org/Homo_sapiens/geneview?gene=__REPLACE__',
    '^ENST' => 'http://www.ensembl.org/Homo_sapiens/transview?transcript=__REPLACE__',
    '^ENSP' => 'http://www.ensembl.org/Homo_sapiens/protview?peptide=__REPLACE__'
    },
'TransPath' => {
    '\(MO' => 'http://www.brinkman.mbb.sfu.ca/cgi-bin/biobase/transpath/4.4/bin/graph.cgi?maxnodes=500&isoff=indirect%7Cinteraction%7Cexpression&dpt=1&dstup2=6&dstup3=10&isoffk=EF&isofft=reaction&rng=SPEC&format=HTML&db=TRANSPATH&hideb=on&out=text%2Fhtml&dstdown2=6&dstdown3=10&id=__REPLACE__&mode=1&dstup1=2&dstdown1=2&endwith=LIGANDS&dens=med&style=HMAP&ishitk=NA&ishitt=molecule%7Cgene%7Creaction&maxnodes=500&shrink=300',
    '\(G0' => 'http://www.brinkman.mbb.sfu.ca/cgi-bin/biobase/transpath/4.4/bin/get.cgi?__REPLACE__',
    }
);
my %annotation_remove = (
			 'putative ortholog TOG', [ 'TOG' ],
			 'DIP', [ 'N' ],
			 'TransFac Factor', [ '.*\(', '\)' ],
			 'TransFac Gene', [ '.*\(', '\)' ],
			 'TransPath', [ '.*\(', '\)' ],
#			 'domain accession number', [ ';' ],
			 #'protein', '\.\d+$',
			 );
my %annotation_item_match = (
			     'HUGO accession','gene symbol',
			     'Ensembl','gene',
			     'Ensembl accession','gene',
			     'EMBL','EMBL accession',
			     'RefSeq','RefSeq accession',
			     'SPTREMBL','SPTREMBL accession',
			     'SWISSPROT','SWISSPROT accession',
			     );
my %annotation_skip = (
		       'transcript', 'medium probability intergenic',
		       'gene', 'medium probability intergenic',
		       );
my %data_column_available = ();
my $last_column_used = '';
my $FG_SIGNAL_TERM = 'FG';
my $FG_COR_SIGNAL_TERM = 'FG_CRD';
my $FOLD_CHANGE_TERM = 'fold change';
my $RATIO_TERM = 'ratio';
my $BG_SIGNAL_TERM = 'BG';
my $SPOT_ID_TERM = 'SPOT ID';
my $FLAG_CHANNEL_TERM = 'FL';
my $SPOT_QUALITY_TERM = 'QCscore';
my $WEIGHTS_TERM = 'WEIGHTS';
my $VALID_ENTRIES_TERM = 'VALLID ENTRIES';
my $SPOT_QUALITY_THRESHOLD = 0;
my $RATIO_DIFF_TERM = 'RATIO_DIFF';
my $MERGED_RATIOS_TERM = 'merged ratios';
my $MERGED_FILE_NAME = 'merge_';
my $COMMON_GENE_NAME = 'Gene Name';
my $COMMON_GENE_NAME_SOURCE = 'Gene Name Source';
my $COMMON_GENE_DESCRIPTION = 'Gene Description';
my $COMMON_GENE_DESCRIPTION_SOURCE = 'Gene Description Source';
my $IMAGE_WIDTH = 570;
my $IMAGE_HEIGHT = 830;
my $SCALE_IMAGE = 1;
my @known_channel_labels = ('Ag', 'Au', 'Cy3', 'Cy5','CH1', 'CH2');

# all the interesting data columns (listed without channel):
my @output = ('Probe ID', 
	      $SPOT_ID_TERM, 
	      $FG_SIGNAL_TERM, 
	      'DUP_DIFF', 
	      $FG_SIGNAL_TERM.'_L', 
	      $FG_SIGNAL_TERM.'_M', 
	      $FG_SIGNAL_TERM.'_M_SD', 
	      $FG_SIGNAL_TERM.'_M_orig', 
	      $FG_SIGNAL_TERM.'_M_list', 
	      $FG_SIGNAL_TERM.'_M_i', 
	      $FG_SIGNAL_TERM.'_L_M', 
	      $FG_SIGNAL_TERM.'_L_M_SD', 
	      $FG_SIGNAL_TERM.'_L_M_list', 
	      $FG_SIGNAL_TERM.'_L_M_i', 
	      $FG_SIGNAL_TERM.'_M_L', 
	      $FG_SIGNAL_TERM.'_M_L_SD', 
	      $FG_SIGNAL_TERM.'_M_L_list', 
	      $FG_SIGNAL_TERM.'_M_L_i', 
	      # background:
	      $BG_SIGNAL_TERM.'_M', 
	      $BG_SIGNAL_TERM, 
	      'BG_CALC_L', 
	      'BG_CALC', 
	      # background corrected foreground:
	      $FG_COR_SIGNAL_TERM, 
	      $FG_COR_SIGNAL_TERM.'_L', 
	      $FG_COR_SIGNAL_TERM.'_M', 
	      $FG_COR_SIGNAL_TERM.'_M_SD', 
	      $FG_COR_SIGNAL_TERM.'_M_orig', 
	      $FG_COR_SIGNAL_TERM.'_M_list', 
	      $FG_COR_SIGNAL_TERM.'_M_i', 
	      $FG_COR_SIGNAL_TERM.'_L_M', 
	      $FG_COR_SIGNAL_TERM.'_L_M_SD', 
	      $FG_COR_SIGNAL_TERM.'_L_M_list', 
	      $FG_COR_SIGNAL_TERM.'_L_M_i', 
	      $FG_COR_SIGNAL_TERM.'_M_L', 
	      $FG_COR_SIGNAL_TERM.'_M_L_SD', 
	      $FG_COR_SIGNAL_TERM.'_M_L_list', 
	      $FG_COR_SIGNAL_TERM.'_M_L_i', 
	      # normalized values:
	      'Norm', 
	      'Norm_M', 
	      'Norm_M_SD',
	      'Norm_M_list', 
	      'Norm_M_i', 
	      'Norm_M_R', 
	      'Norm_L', 
	      'Norm_L_M', 
	      'Norm_L_M_SD', 
	      'Norm_L_M_list', 
	      'Norm_L_M_i', 
	      'Norm_L_M_R', 
	      'Norm_M_L', 
	      'Norm_M_L_SD', 
	      'Norm_M_L_list', 
	      'Norm_M_L_i', 
	      'Norm_M_L_R', 
	      'Dehub', 
	      'Dehub_M_SD', 
	      'Dehub_M_list', 
	      'Dehub_M', 
	      'Dehub_L', 
	      'Dehub_L_M', 
	      'Dehub_L_M_SD', 
	      'Dehub_L_M_list', 
	      'Dehub_L_M_i', 
	      'Dehub_L_M_R', 
	      'Dehub_M_L', 
	      'Dehub_M_L_SD', 
	      'Dehub_M_L_list', 
	      'Dehub_M_L_i', 
	      'Dehub_M_L_R', 
	      'Norm_I', 
	      'Norm_L_I', 
	      'Dehub_L_I', 
	      $FG_COR_SIGNAL_TERM.'_I', 
	      $FG_SIGNAL_TERM.'_I', 
	      'Norm_R', 
	      'Norm_R_M', 
	      'Norm_R_M_SD', 
	      'Norm_R_M_list', 
	      'Norm_R_M_i', 
	      'Norm_L_R', 
	      'Norm_L_R_M', 
	      'Norm_L_R_M_SD', 
	      'Norm_L_R_M_list', 
	      'Norm_L_R_M_i', 
	      'Dehub_L_R', 
	      'Dehub_L_R_M', 
	      'Dehub_L_R_M_SD', 
	      'Dehub_L_R_M_list', 
	      'Dehub_L_R_M_i', 
	      $MERGED_RATIOS_TERM.'_L', 
	      $MERGED_RATIOS_TERM.'_L_SD', 
	      $MERGED_RATIOS_TERM.'_L_i', 
	      $MERGED_RATIOS_TERM.'_L_list', 
	      $MERGED_RATIOS_TERM, 
	      $MERGED_RATIOS_TERM.'_SD', 
	      $MERGED_RATIOS_TERM.'_i', 
	      $MERGED_RATIOS_TERM.'_list', 
	      $FOLD_CHANGE_TERM, 
	      $RATIO_TERM, 
	      'Z-score', 
	      'p-value', 
	      'p-permute', 
	      'p-student', 
	      'p-wilcox', 
	      'p-welch', 
	      'p-paired', 
	      'p-anova', 
	      'F-value',
	      $RATIO_DIFF_TERM, 
	      $FG_COR_SIGNAL_TERM.'_R', 
	      $FG_COR_SIGNAL_TERM.'_R_M', 
	      $FG_COR_SIGNAL_TERM.'_R_M_SD', 
	      $FG_COR_SIGNAL_TERM.'_R_M_list', 
	      $FG_COR_SIGNAL_TERM.'_R_M_i', 
	      $FG_SIGNAL_TERM.'_L_R_M', 
	      $FG_SIGNAL_TERM.'_L_R_M_SD', 
	      $FG_SIGNAL_TERM.'_L_R_M_list', 
	      $FG_SIGNAL_TERM.'_M_L_R', 
	      $FG_SIGNAL_TERM.'_L_M_R', 
	      $FG_SIGNAL_TERM.'_L_R_M_i', 
	      $FG_SIGNAL_TERM.'_L_R', 
	      $FG_SIGNAL_TERM.'_L_R_SD', 
	      $FG_SIGNAL_TERM.'_L_R_list', 
	      $FG_SIGNAL_TERM.'_L_R_i', 
	      $FG_SIGNAL_TERM.'_L_I', 
	      $FG_COR_SIGNAL_TERM.'_L_R', 
	      $FG_COR_SIGNAL_TERM.'_L_R_SD', 
	      $FG_COR_SIGNAL_TERM.'_L_R_list', 
	      $FG_COR_SIGNAL_TERM.'_L_R_i', 
	      $FG_COR_SIGNAL_TERM.'_L_I', 
	      $FG_SIGNAL_TERM.'_R_L', 
	      $FG_SIGNAL_TERM.'_R', 
	      $FG_SIGNAL_TERM.'_R_M', 
	      $FG_SIGNAL_TERM.'_M_R', 
	      $FG_SIGNAL_TERM.'_R_M_SD', 
	      $FG_SIGNAL_TERM.'_R_M_list', 
	      $FG_SIGNAL_TERM.'_R_M_list', 
	      $FG_SIGNAL_TERM.'_R_M_i', 
	      $FLAG_CHANNEL_TERM,
	      $SPOT_QUALITY_TERM,
	      $WEIGHTS_TERM,
	      $VALID_ENTRIES_TERM
	      );

# list available action items (mostly plotting), shortname and full name
# add your own action items here:
my %action_items = ('action_01_bg_plot','Chip Visualization (BG)',
		    'action_02_fg_plot','Chip Visualization (FG)',
		    'action_04_ratio_plot','Chip Visualization (ratios)',
		    'action_08_norm_plot','Chip Visualization (norm)',
#		    'action_10_R_histogram', 'Plot R histogram',
#		    'action_11_R_ArrayPlot', 'Plot of mean intensities of sub-grids',
		    'action_12_R_quintile_histogram', 'Histogram',
		    'action_13_signal_scatter_plot', 'Signal scatterplot', 
		    'action_14_signal_boxplot', 'Signal box plot', 
		    'action_15_signal_boxplot_printtip', 'Signal box plot (print tip)', 
		    'action_17_rep_signal_scatter_plot', 'Inter-slide signal scatterplot', 
		    'action_18_ratio_signal_scatter_plot', 'Inter-slide ratio scatterplot', 
		    'action_20_MA_plot', 'MA-plot',
		    'action_30_compare_lists', 'Compare Lists',
		    'action_35_correlation_coefficient', 'Correlation Coefficient',
		    'action_40_annotate', 'Annotate data',
		    'action_50_output', 'Retrieve data',
		    'filter_01_dup_flagging', 'Flag flawed duplicates',
		    'filter_01b_rep_flawed_flagging', 'Flag flawed replicate spots',
		    'filter_02_fg_min_bg','Foreground - Background',
		    'filter_03_fg_min_percentile_fg', 'By-subgrid BG correction',
		    'filter_03b_loess_bg_correction', 'Loess BG correction',
		    'filter_04_flag_markers', 'Flag markers',
		    'filter_05_set_cutoffs', 'Set cutoffs (data shift)',
		    'filter_06a_limma_loess_printtip', 'limma loess (subgrid)',
		    'filter_06b_limma_loess_global', 'limma loess (global)',
		    'filter_07_vsn_normalization', 'Normalize with VSN',
#		    'filter_08_marrayNorm_normalization', 'Normalize with marrayNorm',
		    'filter_08a_marrayNorm_median', 'marrayNorm: median',
		    'filter_08b_marrayNorm_loess', 'marrayNorm: loess',
		    'filter_08c_marrayNorm_twoD', 'marrayNorm: twoD',
		    'filter_08d_marrayNorm_printTipLoess', 'marrayNorm: printTipLoess',
		    'filter_08e_marrayNorm_globalMAD', 'marrayNorm: globalMAD',
		    'filter_08f_marrayNorm_printTipMAD', 'marrayNorm: printTipMAD',
		    'filter_08g_marrayNorm_scalePrintTipMAD', 'marrayNorm: scalePrintTipMAD',
		    'filter_10_merge_dups', 'Merge duplicate spots',
		    'filter_11_merge_tech_reps', 'Merge technical replicates',
		    'filter_12_merge_all_reps', 'Merge replicate arrays',
		    'filter_20_by_value', 'Filter by value',
		    'filter_31_local_int_dep_z', 'Local intensity dependent Z-scores',
		    'filter_35_student_ttest', 'Student t-test (within group)',
		    'filter_35a_welch_ttest', 'Welch t-test (within group)',
		    'filter_35b_wilcox_ttest', 'Wilcox  rank sum test (within group)',
		    'filter_35c_paired_ttest', 'Paired t-test (within group)',
		    'filter_33_permutation_p_value', 'Permutation p-value (within group)',
		    'filter_38_student_ttest', 'Student t-test (between two groups)',
		    'filter_38a_welch_ttest', 'Welch t-test (between two groups)',
		    'filter_38b_wilcox_ttest', 'Wilcox rank sum test (between two groups)',
		    'filter_40_anova', 'ANOVA',
		    'filter_04g_flag_change', 'Flag change',
		    );

my %module_category = (
		       '0 often used' => [
					     'filter_04_flag_markers',
					     'action_02_fg_plot',
					     'action_13_signal_scatter_plot',
					     'action_20_MA_plot',
					     'action_14_signal_boxplot',
					     'filter_03_fg_min_percentile_fg', 
					     'filter_05_set_cutoffs',
					     'filter_06a_limma_loess_printtip',
					     'filter_20_by_value',
					     'action_50_output',
					     ],					     
		       '1 flagging and filtering' => [
						      'filter_04_flag_markers',
						      'filter_04g_flag_change',
						      'filter_01_dup_flagging',
						      'filter_01b_rep_flawed_flagging',
						      'filter_05_set_cutoffs',
						      'filter_20_by_value',
						      ],
		       '2 visualization' => [
					     'action_01_bg_plot',
					     'action_02_fg_plot',
					     'action_04_ratio_plot',
					     'action_08_norm_plot',
#					     'action_10_R_histogram',
					     'action_12_R_quintile_histogram',
					     'action_13_signal_scatter_plot',
					     'action_14_signal_boxplot',
					     'action_15_signal_boxplot_printtip',
					     'action_17_rep_signal_scatter_plot',
					     'action_18_ratio_signal_scatter_plot',
					     'action_20_MA_plot',
					     ],
		       '3 background correction' => [
						     'filter_02_fg_min_bg',
						     'filter_03_fg_min_percentile_fg',
						     'filter_03b_loess_bg_correction',
						     ],
		       '4 normalization' => [
					     'filter_06a_limma_loess_printtip',
					     'filter_06b_limma_loess_global',
					     'filter_07_vsn_normalization',
#					     'filter_08_marrayNorm_normalization',
					     'filter_08a_marrayNorm_median',
					     'filter_08b_marrayNorm_loess',
					     'filter_08c_marrayNorm_twoD',
					     'filter_08d_marrayNorm_printTipLoess',
					     'filter_08g_marrayNorm_scalePrintTipMAD',
					     'filter_08e_marrayNorm_globalMAD',
					     'filter_08f_marrayNorm_printTipMAD',
						     ],
		       '5 replicate handling' => [
						  'action_30_compare_lists',
						  'action_35_correlation_coefficient',
						  'filter_10_merge_dups',
						  'filter_11_merge_tech_reps',
						  'filter_12_merge_all_reps',
						     ],
		       '6 differential expression' => [
						       'filter_31_local_int_dep_z',
						       'filter_35_student_ttest',
						       'filter_35c_paired_ttest',
						       'filter_35a_welch_ttest',
						       'filter_35b_wilcox_ttest', 
						       'filter_33_permutation_p_value', 
						       'filter_38_student_ttest',
						       'filter_38a_welch_ttest',
						       'filter_38b_wilcox_ttest', 
						       'filter_40_anova', 
						     ],
		       '7 miscellaneous' => [
					     'action_40_annotate',
					     'action_50_output',
						     ],
#		       '8 experimental' => [
#					    'action_11_R_ArrayPlot',
#					    'action_25_pathway_plot',
#					    ],
		       );


# the approximate number of seconds your action might run (optional)
my %action_wait = ('action_01_bg_plot', 6,
		    'action_02_fg_plot', 6,
		    'action_08_norm_plot', 6,
		    'action_04_ratio_plot', 6,
		    'action_10_R_histogram', 4,
		    'action_11_R_ArrayPlot', 4,
		    'action_12_R_quintile_histogram', 3,
		    'action_13_signal_scatter_plot', 14, 
		    'action_14_signal_boxplot', 4, 
		    'action_15_signal_boxplot_printtip', 5, 
		    'action_17_rep_signal_scatter_plot', 4, 
		    'action_18_ratio_signal_scatter_plot', 4, 
		    'action_20_MA_plot', 10,
		    'action_30_compare_lists', 6,
		    'action_35_correlation_coefficient', 10,
		    'action_40_annotate', 20,
		    'action_50_output', 30,
		    'filter_01_dup_flagging', 3,
		    'filter_01b_rep_flawed_flagging', 5,
		    'filter_02_fg_min_bg', 3,
		    'filter_03_fg_min_percentile_fg', 3,
		    'filter_03b_loess_bg_correction', 240,
		    'filter_04_flag_markers', 3,
		    'filter_04g_flag_change', 3,
		    'filter_05_set_cutoffs', 2,
		   'filter_06a_limma_loess_printtip', 10,
		   'filter_06b_limma_loess_global', 10,
		    'filter_07_vsn_normalization', 20,
#		    'filter_08_marrayNorm_normalization', 40,
		   'filter_08a_marrayNorm_median', 30,
		   'filter_08b_marrayNorm_loess', 120,
		   'filter_08c_marrayNorm_twoD', 40,
		   'filter_08d_marrayNorm_printTipLoess', 40,
		   'filter_08g_marrayNorm_scalePrintTipMAD', 60,
		   'filter_08e_marrayNorm_globalMAD', 40,
		   'filter_08f_marrayNorm_printTipMAD', 40,
		    'filter_10_merge_dups', 3,
		    'filter_11_merge_tech_reps', 3,
		    'filter_12_merge_all_reps', 3,
		    'filter_20_by_value', 4,
		    'action_25_pathway_plot', 15,
		    'filter_31_local_int_dep_z', 20,
		    'filter_35_student_ttest', 180,
		    'filter_35c_paired_ttest', 180,
		    'filter_35a_welch_ttest', 180,
		    'filter_35b_wilcox_ttest', 180,
		    'filter_33_permutation_p_value', 20,
		    'filter_38_student_ttest', 180,
		    'filter_38a_welch_ttest', 180,
		    'filter_38b_wilcox_ttest', 180,
		    'filter_40_anova', 60,
		    );


# description of each item, 
# each item listed above must have one
my %item_description = ('action_01_bg_plot','Plot a chip graphic to visualize the background intensities of an array',
			'action_02_fg_plot','Plot a chip graphic to visualize the foreground intensities of an array. The default settings (no borders specified on a relative scale (\'rel\')) will plot a gradient of the middle 50 percent of the data, with the intensities of the lower quartile in black and the upper quartile in white. This reduction of complexity is useful for detecting spatial bias on the chip.',
			'action_08_norm_plot','Plot a chip graphic to visualize the foreground intensities after normalization',
			'action_04_ratio_plot','Plot a chip graphic to visualize the ratios (ch2/ch1); normalized values will be taken automatically, if available',
			'action_10_R_histogram','Plot histogram of foreground signal through R.',
			'action_11_R_ArrayPlot', 'Color representation of sub-grid intensities, experimental (implemented by David Goode, Finlay lab, UBC, BC, Canada, based on R-code by Dana Aeschliman, Statistics, also UBC).',
			'action_12_R_quintile_histogram', 'Plot histogram of intensity values, with quintile lines added and log-transferred if requested (implemented by David Goode, Finlay lab, UBC, BC, Canada, based on R-code by Dana Aeschliman, Statistics, also UBC, expanded by Karsten Hokamp, Brinkman Lab, SFU). Optionally sets minimum value to \'1\', to avoid log-transferring of negative values.',
			'action_13_signal_scatter_plot', 'Plots foreground signals from both channels against each other', 
			'action_14_signal_boxplot', 'Show box plots of each channel for each file beside each other (useful to do before and after normalization). This was derived from a tool used by Marc-Etienne Rousseau at Inimex Pharmaceuticals Inc., BC, Canada.', 
			'action_15_signal_boxplot_printtip', 'Draws box plots for each subgrid beside each other. Channels can be plotted individually, beside each other or combined. Useful to do before and after normalization.', 
			'action_17_rep_signal_scatter_plot', 'This module creates signal scatter plots for duplicate channels in replicate or dye-flip experiments (use the boxes in the data section to label channels accordingly).', 
			'action_18_ratio_signal_scatter_plot', 'This module creates scatter plots of the ratios between two slides.', 
			'action_20_MA_plot', 'Create MA-plot, i.e., plot 0.5*log(Ch1*Ch2) versus log(Ch2/Ch1)',
			'action_25_pathway_plot', 'Overlay expression data onto BioCarta pathways. This module uses code provided by Fiona Roche.',
			'action_30_compare_lists', 'Compare subsets of data identified by flags and compute the number of spots shared between them.',
			'action_35_correlation_coefficient', 'Calculate the correlation coefficient as a measurement of how close replicate slides are to each other.',
			'action_40_annotate', 'Annotate spots with external biological data. You can upload an output file from <A HREF="http://www.pathogenomics.ca/probelynx">ProbeLynx</A>, which matches your probe sequences to genes (Ensembl) or tentative consensus sequences (TCs, TIGR) and combines further annotation from the respective databases.Additionally, annotation info from FuncDB (an extensive data resource assembled by Fiona Roche and Michael Acab, Brinkman lab) can be integrated. This is mapped to a probe either through the Ensembl transcript that it maps to or through a putative human orthologue (in case of organisms other than human). Please be aware of the quality information regarding the matches between probes and genes/TCs and between TCs and orthologues, as well as the number of transcripts mapped to each probe. This is contained in columns such as <B>hit_identity</B>, <B>hit_coverage</B>, <B>hit_flag</B>, etc. ',
			'action_50_output', 'Write data into a file and make it available for download',
			'filter_01_dup_flagging', 'Flag duplicated spots (within a slide) that differ by a user specified cutoff. The unit used can be either the absolute difference, the fold difference or their standard deviations. If a difference larger than the cutoff is found in at least one channel, the spot and its duplicate neighbour is flagged with the specified flag. This module was implemented with the help of Byron Kuo, Inimex Pharmaceuticals Inc., BC, Canada.',
			'filter_01b_rep_flawed_flagging', 'Flag replicated spots (between multiple slides) that differ by a user specified cutoff. The unit used can be either the absolute difference, the fold difference or their standard deviations. If a difference larger than the cutoff is found in at least one x percent of the replicates, either all of the spots or only those that vary are flagged with the specified flag.',
			'filter_02_fg_min_bg','Background correction through simply subtracting background from foreground. Spots that end up with values <= 0 will be flagged as \'floor\'.',
			'filter_03_fg_min_percentile_fg','Background correction using the median of the lower percentile (0.1 by default) of the foreground values per subgrid as a background estimation, which is then subtracted from the intensity signal within that subgrid (or only flag those that are at background level or below). This module is base on R-code developed by Jenny Bryan, Statistics, UBC, BC, Canada. The output shows a chip visualization with the calculated background value for each subgrid.',
			'filter_03b_loess_bg_correction','Localized background correction using Loess parameters. This module is base on R-code provided by Dana Aeschliman, Statistics, UBC, BC, Canada.',
			'filter_04_flag_markers', 'Flag every spot that is either empty or a marker, so that it can be filtered out lateron',
			'filter_04g_flag_change', 'Change one or more flags into another',
			'filter_05_set_cutoffs', 'Specify lower and upper ceilings for the data and either delete or adjust data points outside the thresholds',
			'filter_06a_limma_loess_printtip', 'Intensity dependent normlization using <A HREF="http://www.bioconductor.org">BioConductor\'s</A> <A HREF="http://www.bioconductor.org/repository/release1.3/package/html/limma.html">limma</A>\'s <B>loess</B> function, applied per printtip', 
			'filter_06b_limma_loess_global', 'Intensity dependent normlization using <A HREF="http://www.bioconductor.org">BioConductor\'s</A> <A HREF="http://www.bioconductor.org/repository/release1.3/package/html/limma.html">limma</A>\'s <B>loess</B> function, applied to the whole array', 
			'filter_07_vsn_normalization', 'Normalize signal intensities with Huber\'s VSN function (integrated into R). For global comparison, "the arrays have to be reasonably similar, i.e. from one production batch" (Huber).',
#			'filter_08_marrayNorm_normalization', 'Normalize signal intensities with <A HREF="http://www.bioconductor.org">BioConductor\'s</A> <A HREF="http://www.bioconductor.org/repository/release1.3/package/html/marrayNorm.html">marrayNorm</A> package.<BR>The following methods are available:<UL><LI><B>median</B>, for global median location normalization<LI><B>loess</B> for global intensity or A-dependent location normalization using the loess function<LI><B>twoD</B>, for 2D spatial location normalization using the loess function<LI><B>printTipLoess</B>, for within-print-tip-group intensity dependent location normalization using the loess function<LI><B>scalePrintTipMAD</B>, for within-print-tip-group intensity dependent location normalization followed by within-print-tip-group scale normalization using the median absolute deviation<LI><B>globalMAD</B>, for global scale normalization using the median absolute deviation<LI><B>printTipMAD</B>, for within-print-tip-group scale normalization using the median absolute deviation.',
			'filter_08a_marrayNorm_median', 'Normalize signal intensities with <A HREF="http://www.bioconductor.org">BioConductor\'s</A> <A HREF="http://www.bioconductor.org/repository/release1.3/package/html/marrayNorm.html">marrayNorm</A> package using the <B>median</B> function for global median location normalization.',
			'filter_08b_marrayNorm_loess', 'Normalize signal intensities with <A HREF="http://www.bioconductor.org">BioConductor\'s</A> <A HREF="http://www.bioconductor.org/repository/release1.3/package/html/marrayNorm.html">marrayNorm</A> package using the <B>loess</B> function for global intensity or A-dependent location normalization.', 
			'filter_08c_marrayNorm_twoD', 'Normalize signal intensities with <A HREF="http://www.bioconductor.org">BioConductor\'s</A> <A HREF="http://www.bioconductor.org/repository/release1.3/package/html/marrayNorm.html">marrayNorm</A> package using the <B>loess</B> function, for 2D spatial location normalization',
			'filter_08d_marrayNorm_printTipLoess', 'Normalize signal intensities with <A HREF="http://www.bioconductor.org">BioConductor\'s</A> <A HREF="http://www.bioconductor.org/repository/release1.3/package/html/marrayNorm.html">marrayNorm</A> package using the <B>loess</B> function for within-print-tip-group intensity dependent location normalization',
			'filter_08g_marrayNorm_scalePrintTipMAD', 'Normalize signal intensities with <A HREF="http://www.bioconductor.org">BioConductor\'s</A> <A HREF="http://www.bioconductor.org/repository/release1.3/package/html/marrayNorm.html">marrayNorm</A> package using the <B>loess</B> function for within-print-tip-group intensity dependent location normalization followed by within-print-tip-group scale normalization using the median absolute deviation (<B>MAD</B>). This is the same as using first printTipLoess and then printTipMAD.',
			'filter_08e_marrayNorm_globalMAD', 'Normalize signal intensities with <A HREF="http://www.bioconductor.org">BioConductor\'s</A> <A HREF="http://www.bioconductor.org/repository/release1.3/package/html/marrayNorm.html">marrayNorm</A> package using the median absolute deviation (<B>MAD</B>) for global scale normalization.',
			'filter_08f_marrayNorm_printTipMAD', 'Normalize signal intensities with <A HREF="http://www.bioconductor.org">BioConductor\'s</A> <A HREF="http://www.bioconductor.org/repository/release1.3/package/html/marrayNorm.html">marrayNorm</A> package using the median absolute deviation (<B>MAD</B>) for within-print-tip-group scale normalization.',
			'filter_10_merge_dups', 'Merge multiple values of the same entity (probe/gene) into a single value.',
			'filter_11_merge_tech_reps', 'Merge values from multiple technical replicate arrays per biological sample into a single data set (using the specified method).',
			'filter_12_merge_all_reps', 'Merge values from multiple replicate arrays from each group into a single data set (using the specified method).',
			'filter_20_by_value', "Filter spots/genes according to restrictions on a specific data column (the latest, highest level data type by default); data that will not pass through the filter is marked with the specified flag. To reverse this, i.e. only flag those spots that get through the filter, insert a name into the \'Make list\'-field to generate a list that is used for flagging. Choose \'auto\' if the module is called multiple times.<BR>The \'auto\' keyword in the filter settings is translated as follows:<BR>
<TABLE BORDER=2>
<TR><TH>field</TH><TH>data</TH><TH>value</TH></TR>
<TR><TD ROWSPAN=6>Operator,<BR>Threshold</TD><TD>Z-score</TD><TD>\&gt; 1.96</TD></TR>
<TR><TD>p-value</TD><TD>\&lt; 0.05</TD></TR>
<TR><TD>ratio (log'ed)</TD><TD>\&gt; 1</TD></TR>
<TR><TD>ratio (raw)</TD><TD>\&gt; 2</TD></TR>
<TR><TD>$FG_COR_SIGNAL_TERM</TD><TD>\&gt; -50</TD></TR>
<TR><TD>rest</TD><TD>\&gt; 1000</TD></TR>
<TR><TD ROWSPAN=6>Operator2,<BR>Threshold2</TD><TD>Z-score</TD><TD>\&gt; 1.96</TD></TR>
<TR><TD>p-value</TD><TD>\&lt; 0.05</TD></TR>
<TR><TD>ratio (log'ed)</TD><TD>\&lt; -1</TD></TR>
<TR><TD>ratio (raw)</TD><TD>\&lt; 0.5</TD></TR>
<TR><TD>$FG_COR_SIGNAL_TERM</TD><TD>\&gt; -50</TD></TR>
<TR><TD>rest</TD><TD>\&gt; 1000</TD></TR>
<TR><TD ROWSPAN=4>Labels</TD><TD>Z-score</TD><TD>combined</TD></TR>
<TR><TD>Ratio</TD><TD>combined</TD></TR>
<TR><TD>p-value</TD><TD>combined or<BR>CH1 and CH2</TD></TR>
<TR><TD>rest</TD><TD>CH1 and CH2</TD></TR>
</TABLE>",
			'filter_31_local_int_dep_z', 'A sliding window across log transferred intensities calculates local standard deviations of the log transferred ratios and assigns Z-scores to define significantly differiantially expressed genes (based on <A HREF="http://www.ncbi.nlm.nih.gov/entrez/query.fcgi?cmd=Retrieve&db=PubMed&list_uids=12429061&dopt=Abstract">Yang <I>et al.</I></A>, Genome Biology 2002, 3(11):research0062.1 0062.12)',
			'filter_35_student_ttest', 'The Student t-test will produce p-values for the processed data. This module is applicable for processing data from one experimental group, either the two foreground intensities (two-sample) or the ratios (one-sample). You need to specify in the loaded data section which channel is the control and which is the treatment before running this module. Replicate slides should carry the same channel assignment.<BR>The R-code was provided by Marc-Etienne Rousseau from Inimex Pharmaceuticals Inc., BC, Canada.',
			'filter_35c_paired_ttest', 'The paired t-test will produce p-values for the processed data. This module is applicable only for processing data from one experimental group using the two foreground intensities. You need to specify in the loaded data section which channel is the control and which is the treatment before running this module. Replicate slides should carry the same channel assignment.',
			'filter_35a_welch_ttest', 'The Welch t-test will produce p-values for the processed data. This module is applicable for processing data from one experimental group, using two foreground intensities (two-sample). When working on the ratios (one-sample) it will act as a Student t-test. You need to specify in the loaded data section which channel is the control and which is the treatment before running this module. Replicate slides should carry the same channel assignment.<BR>The R-code was provided by Marc-Etienne Rousseau from Inimex Pharmaceuticals Inc., BC, Canada.<BR>Reference for the method: Welch, B.L. (1938). The significance of the difference between two means when the population variances are unequal. Biometrika, 29, 350-362.',
			'filter_35b_wilcox_ttest', 'The Wilcox t-test will produce p-values for the processed data. This module is applicable for processing data from one experimental group, either the two foreground intensities (two-sample = Wilcoxon rank sum test) or the ratios (one-sample = Wilcoxon signed rank test). You need to specify in the loaded data section which channel is the control and which is the treatment before running this module. Replicate slides should carry the same channel assignment.<BR>The R-code was provided by Marc-Etienne Rousseau from Inimex Pharmaceuticals Inc., BC, Canada.',
			'filter_33_permutation_p_value', 'Shuffle data and compute p-values as the percentage of simulated means which are greater than the actual observed means relative to the number of simulated samples. This module uses a C++ program developed by Byron Kuo, Inimex Pharmaceuticals Inc., BC, Canada.',
			'filter_38_student_ttest', 'The Student t-test will produce p-values for the processed data. This module is applicable for processing data (ratios) from two experimental groups. You need to specify in the loaded data section which channel is the control and which is the treatment before running this module. Replicate slides should carry the same channel assignment.',

			'filter_38a_welch_ttest', 'The Welch t-test will produce p-values for the processed data. This module is applicable for processing data (ratios) from two experimental groups. You need to specify in the loaded data section which channel is the control and which is the treatment before running this module. Replicate slides should carry the same channel assignment.<BR>The R-code was provided by Marc-Etienne Rousseau from Inimex Pharmaceuticals Inc., BC, Canada.<BR>Reference for the method: Welch, B.L. (1938). The significance of the difference between two means when the population variances are unequal. Biometrika, 29, 350-362.',
			'filter_38b_wilcox_ttest', 'The Wilcox t-test will produce p-values for the processed data. This module is applicable for processing data (ratios) from two experimental groups. You need to specify in the loaded data section which channel is the control and which is the treatment before running this module. Replicate slides should carry the same channel assignment.<BR>The R-code was provided by Marc-Etienne Rousseau from Inimex Pharmaceuticals Inc., BC, Canada.',
			'filter_40_anova', 'Carry out Analysis of Variance, to test if the null hypothesis that the effects of all treatments are the same is true. You need at least two groupings of your input files (done by the channel annotation). The ratios for the spots in each group are taken as input. An F-value and a p-value are returned and assigned to each spot from each each input file that was used.',
			);

# dataset each item is working on 
# (in order of preference)
my %item_data_column = (
			'TIGR MEV', 
			[ 'Dehub_M', 
			  'Dehub', 
			  'Norm_M', 
			  'Norm_M_L', 
			  'Norm_L_M', 
			  'Norm', 
			  'Norm_L', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_M_L', 
			  $FG_COR_SIGNAL_TERM.'_L_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M_L', 
			  $FG_SIGNAL_TERM.'_L_M', 
			  $FG_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L', 
			  ],
			
			'action_02_fg_plot', 
			[ 'Dehub_M', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  $FG_COR_SIGNAL_TERM.'_L_M', 
			  $FG_COR_SIGNAL_TERM.'_M_L', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_L_M', 
			  $FG_SIGNAL_TERM.'_M_L', 
			  $FG_SIGNAL_TERM.'_M', 
			  'Dehub', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM 
			  ],

			'action_04_ratio_plot', 
			[ 'Dehub_M', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  'Dehub', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ],
			
			'action_10_R_histogram', 
			[ 'Dehub_L', 
			  'Norm_L', 
			  'Norm', 
			  'Dehub', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ],

			'action_11_R_ArrayPlot', 
			[ $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ],

			'action_12_R_quintile_histogram', 
			[ $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ],
			
			'action_13_signal_scatter_plot', 
			[ 'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_M', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  'Dehub_L', 
			  'Dehub', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ], 
			
			'action_14_signal_boxplot', 
			[ 'Dehub_L_R_M', 
			  'Norm_L_R_M', 
			  'Norm_R_M', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  'Dehub_L_R', 
			  'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_M', 
			  'Dehub_L', 
			  'Dehub', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ], 

			'action_15_signal_boxplot_printtip', 
			[ 'Dehub_L_R_M', 
			  'Norm_L_R_M', 
			  'Norm_R_M', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  'Norm_L_R', 
			  'Dehub_L_R', 
			  'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_M', 
			  'Dehub_L', 
			  'Dehub', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_L_R', 
			  $FG_SIGNAL_TERM.'_L_R', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM 
			  ], 

			'action_17_rep_signal_scatter_plot', 
			[ 'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_M', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  'Dehub_L', 
			  'Dehub', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ], 

			'action_18_ratio_signal_scatter_plot', 
			[ 'Dehub_L_M_R', 
			  'Dehub_M_L_R', 
			  'Norm_M_R', 
			  $FG_COR_SIGNAL_TERM.'_M_R', 
			  $FG_SIGNAL_TERM.'_M_R', 
			  'Dehub_L_R_M', 
			  'Norm_L_R_M', 
			  'Norm_R_M', 
			  $FG_COR_SIGNAL_TERM.'_R_M', 
			  $FG_SIGNAL_TERM.'_R_M', 
			  'Dehub_L_R', 
			  'Norm_L_R', 
			  'Norm_R', 
			  $FG_COR_SIGNAL_TERM.'_R', 
			  $FG_SIGNAL_TERM.'_R' 
			  ],
			
			'action_20_MA_plot', 
			[ 'Dehub_M', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  'Dehub', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ],
			
			'action_25_pathway_plot', 
			[ 'Dehub_L_M_R', 
			  'Dehub_M_L_R', 
			  'Norm_M_R', 
			  $FG_COR_SIGNAL_TERM.'_M_R', 
			  $FG_SIGNAL_TERM.'_M_R', 
			  'Dehub_L_R_M', 
			  'Norm_L_R_M', 
			  'Norm_R_M', 
			  $FG_COR_SIGNAL_TERM.'_R_M', 
			  $FG_SIGNAL_TERM.'_R_M', 
			  'Dehub_L_R', 
			  'Norm_L_R', 
			  'Norm_R', 
			  $FG_COR_SIGNAL_TERM.'_R', 
			  $FG_SIGNAL_TERM.'_R', 
			  'Dehub_M', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  'Dehub', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ],
			
			'action_35_correlation_coefficient', 
			[ 'Norm_L_R', 
			  'Dehub_L_R', 
			  'Dehub_L', 
			  'Dehub', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_L_R', 
			  $FG_SIGNAL_TERM.'_L_R', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM 
			  ],
			
			'action_50_output', 
			[ 'Dehub_L_R_M', 
			  'Dehub_L_M_R', 
			  'Dehub_M_L_R', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Dehub_L_R', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  'Dehub_L', 
			  'Dehub', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ],
			
			'filter_01_dup_flagging', 
			[ 
			  'Dehub_L_M_R', 
			  'Dehub_M_L_R', 
			  'Dehub_L_R_M', 
			  'Norm_L_M_R', 
			  'Norm_M_L_R', 
			  'Norm_L_R_M', 
			  'Dehub_R_M', 
			  'Dehub_M_R', 
			  'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_M', 
			  'Norm_R_M', 
			  'Norm_M_R', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Dehub_L_R', 
			  'Dehub_R_L', 
			  'Dehub_R', 
			  'Dehub_L', 
			  'Dehub', 
			  'Norm_L_R', 
			  'Norm_R_L', 
			  'Norm_R', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_L_M_R', 
			  $FG_COR_SIGNAL_TERM.'_M_L_R', 
			  $FG_COR_SIGNAL_TERM.'_L_R_M', 
			  $FG_COR_SIGNAL_TERM.'_R_M', 
			  $FG_COR_SIGNAL_TERM.'_M_R', 
			  $FG_COR_SIGNAL_TERM.'_L_M', 
			  $FG_COR_SIGNAL_TERM.'_M_L', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L_R', 
			  $FG_COR_SIGNAL_TERM.'_R_L', 
			  $FG_COR_SIGNAL_TERM.'_R', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L_M_R', 
			  $FG_SIGNAL_TERM.'_M_L_R', 
			  $FG_SIGNAL_TERM.'_L_R_M', 
			  $FG_SIGNAL_TERM.'_R_M', 
			  $FG_SIGNAL_TERM.'_M_R', 
			  $FG_SIGNAL_TERM.'_L_M', 
			  $FG_SIGNAL_TERM.'_M_L', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_L_R', 
			  $FG_SIGNAL_TERM.'_R_L', 
			  $FG_SIGNAL_TERM.'_R', 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM, 
			  ],
			
			'filter_01b_rep_flawed_flagging', 
			[ 
			  'Dehub_L_M_R', 
			  'Dehub_M_L_R', 
			  'Dehub_L_R_M', 
			  'Norm_L_M_R', 
			  'Norm_M_L_R', 
			  'Norm_L_R_M', 
			  'Dehub_R_M', 
			  'Dehub_M_R', 
			  'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_M', 
			  'Norm_R_M', 
			  'Norm_M_R', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Dehub_L_R', 
			  'Dehub_R_L', 
			  'Dehub_R', 
			  'Dehub_L', 
			  'Dehub', 
			  'Norm_L_R', 
			  'Norm_R_L', 
			  'Norm_R', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_L_M_R', 
			  $FG_COR_SIGNAL_TERM.'_M_L_R', 
			  $FG_COR_SIGNAL_TERM.'_L_R_M', 
			  $FG_COR_SIGNAL_TERM.'_R_M', 
			  $FG_COR_SIGNAL_TERM.'_M_R', 
			  $FG_COR_SIGNAL_TERM.'_L_M', 
			  $FG_COR_SIGNAL_TERM.'_M_L', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L_R', 
			  $FG_COR_SIGNAL_TERM.'_R_L', 
			  $FG_COR_SIGNAL_TERM.'_R', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L_M_R', 
			  $FG_SIGNAL_TERM.'_M_L_R', 
			  $FG_SIGNAL_TERM.'_L_R_M', 
			  $FG_SIGNAL_TERM.'_R_M', 
			  $FG_SIGNAL_TERM.'_M_R', 
			  $FG_SIGNAL_TERM.'_L_M', 
			  $FG_SIGNAL_TERM.'_M_L', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_L_R', 
			  $FG_SIGNAL_TERM.'_R_L', 
			  $FG_SIGNAL_TERM.'_R', 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM, 
			  ],

			'filter_02_fg_min_bg', 
			[ $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_03_fg_min_percentile_fg', 
			[ $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_03b_loess_bg_correction', 
			[ $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_03_fg_min_percentile_fg_draw', 
			[ 'BG_CALC' ],

			'filter_04_flag_markers', 
			[ $SPOT_ID_TERM ],

			'filter_04g_flag_change', 
			[ $FLAG_CHANNEL_TERM ],

			'filter_05_set_cutoffs', 
			[ 'Norm_M', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  'Dehub', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_06a_limma_loess_printtip', 
			[ 'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ], 

			'filter_06b_limma_loess_global', 
			[ 'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ], 
			
			'filter_07_vsn_normalization', 
			[ 'Norm', 
			  'Norm_L', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ], 

			'filter_08a_marrayNorm_median', 
			[ 'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ], 

			'filter_08b_marrayNorm_loess', 
			[ 'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ], 

			'filter_08c_marrayNorm_twoD', 
			[ 'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ], 

			'filter_08d_marrayNorm_printTipLoess', 
			[ 'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ], 

			'filter_08g_marrayNorm_scalePrintTipMAD', 
			[ 'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ], 

			'filter_08e_marrayNorm_globalMAD', 
			[ 'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ], 

			'filter_08f_marrayNorm_printTipMAD', 
			[ 'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ], 

			'filter_10_merge_dups', 
			[ 'Dehub_L_R', 
			  'Dehub_L', 
			  'Dehub_R', 
			  'Dehub', 
			  'Norm_L_R', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_L_R', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM.'_L_R', 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM.'_R', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_R', 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_12_merge_all_reps', 
			[ 'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_M', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  $FG_COR_SIGNAL_TERM.'_L_M', 
			  $FG_COR_SIGNAL_TERM.'_M_L', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_L_M', 
			  $FG_SIGNAL_TERM.'_M_L', 
			  $FG_SIGNAL_TERM.'_M', 
			  'Dehub_L', 
			  'Dehub', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_11_merge_tech_reps', 
			[ 'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_M', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  $FG_COR_SIGNAL_TERM.'_L_M', 
			  $FG_COR_SIGNAL_TERM.'_M_L', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_L_M', 
			  $FG_SIGNAL_TERM.'_M_L', 
			  $FG_SIGNAL_TERM.'_M', 
			  'Dehub_L', 
			  'Dehub', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_20_by_value', 
			[ 'p-value', 
			  'p-anova', 
			  'p-paired',
			  'p-wilcox',
			  'p-welch',
			  'p-student',
			  'p-permute',
			  'Z-score', 
			  $MERGED_RATIOS_TERM.'_L', 
			  $MERGED_RATIOS_TERM, 
			  'Dehub_L_R_M', 
			  'Dehub_L_M_R', 
			  'Dehub_M_L_R', 
			  'Norm_L_R_M', 
			  'Norm_R_M', 
			  $FG_COR_SIGNAL_TERM.'_R_M', 
			  $FG_SIGNAL_TERM.'_R_M', 
			  'Dehub_L_R', 
			  'Norm_L_R', 
			  'Norm_R', 
			  $FG_COR_SIGNAL_TERM.'_L_R', 
			  $FG_SIGNAL_TERM.'_L_R', 
			  $FG_COR_SIGNAL_TERM.'_R', 
			  $FG_SIGNAL_TERM.'_R', 
			  'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  $FG_COR_SIGNAL_TERM.'_L_I', 
			  $FG_SIGNAL_TERM.'_L_I', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_M', 
			  'Dehub_L_R', 
			  'Norm_L', 
			  'Norm', 
			  'Dehub_L', 
			  'Dehub', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_31_local_int_dep_z', 
			[ 'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_L', 
			  'Dehub_M', 
			  'Dehub', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L_M', 
			  $FG_SIGNAL_TERM.'_M_L', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_35_student_ttest', 
			[ 'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_L', 
			  'Dehub_M', 
			  'Dehub', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L_M', 
			  $FG_SIGNAL_TERM.'_M_L', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_35c_paired_ttest', 
			[ 'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_L', 
			  'Dehub_M', 
			  'Dehub', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L_M', 
			  $FG_SIGNAL_TERM.'_M_L', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_35a_welch_ttest', 
			[ 'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_L', 
			  'Dehub_M', 
			  'Dehub', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L_M', 
			  $FG_SIGNAL_TERM.'_M_L', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_35b_wilcox_ttest', 
			[ 'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_L', 
			  'Dehub_M', 
			  'Dehub', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L_M', 
			  $FG_SIGNAL_TERM.'_M_L', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_33_permutation_p_value', 
			[ 'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_L', 
			  'Dehub_M', 
			  'Dehub', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L_M', 
			  $FG_SIGNAL_TERM.'_M_L', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_38_student_ttest', 
			[ 'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_L', 
			  'Dehub_M', 
			  'Dehub', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L_M', 
			  $FG_SIGNAL_TERM.'_M_L', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_38a_welch_ttest', 
			[ 'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_L', 
			  'Dehub_M', 
			  'Dehub', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L_M', 
			  $FG_SIGNAL_TERM.'_M_L', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_38b_wilcox_ttest', 
			[ 'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_L', 
			  'Dehub_M', 
			  'Dehub', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_L_M', 
			  $FG_SIGNAL_TERM.'_M_L', 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM.'_L', 
			  $FG_SIGNAL_TERM 
			  ],

			'filter_40_anova', 
			[ 'Dehub_L_M', 
			  'Dehub_M_L', 
			  'Dehub_L', 
			  'Dehub_M', 
			  'Dehub', 
			  'Norm_L_M', 
			  'Norm_M_L', 
			  'Norm_M', 
			  'Norm_L', 
			  'Norm', 
			  $FG_COR_SIGNAL_TERM.'_M', 
			  $FG_COR_SIGNAL_TERM.'_L', 
			  $FG_COR_SIGNAL_TERM, 
			  $FG_SIGNAL_TERM.'_M', 
			  $FG_SIGNAL_TERM 
			  ],
			);

my %function = (); # this hash keeps a list of pointers to the functions

my $refine_parameters = '';


#######################################
#### Add your own subroutine here: ####
#######################################

# (to keep this part concise, it would be best to 
# define a frame subroutine for each action/filter
# that calls other subroutines which can be defined 
# lateron)


##########################
### ACTION SUBROUTINES ###
##########################

# define action subroutine

# example for R stuff
sub action_10_R_histogram {
    # each subroutine gets four arguments:
    my $num = shift;    # the number in the action chain
                        # (useful for action specific parameters)
    my $action = shift; # the name of the action 
                        # (same as subroutine name)

    # return list with the headers required for this function
    # when called with special argument
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM;
    }

    # carry out action on each file ...
    foreach my $file (sort keys %data) {
	
	# ... and each label
	foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	    
	    # skip columns that are not label specific ('spot labels')
	    next if ($label eq '__NO_LABEL__');
	    
	    my ($image, $info);
	    if ($refine_parameters) {
		$info = 'Parameter Refinement!<BR>';
	    } else {

		# put all the main calculations into a separate subroutine
		# the return value should be HTML code that will be directly
		# included into the results page
		$image =  &R_histogram_plotting($num,$file,$label,$action);
	    }

	    # format output (include image and choice for parameters)
	    print &format_output( 
				  {
				      'prefix' => 'hist_',
				      'num' => $num,
				      'file' => $file,
				      'label' => $label,
				      'action' => $action,
				      'image' => $image,
				      'info' => $info
				      }
				  );
	}
    }
}
$function{'action_10_R_histogram'} = \&action_10_R_histogram;


sub action_30_compare_lists {
    # each subroutine gets four arguments:
    my $num = shift;    # the number in the action chain
                        # (useful for action specific parameters)
    my $action = shift; # the name of the action 
                        # (same as subroutine name)

    # return list with the headers required for this function
    # when called with special argument
    if ($num eq '__columns__') {
	return $FLAG_CHANNEL_TERM;
    }

    &compare_lists_setup($num,$action);
}
$function{'action_30_compare_lists'} = \&action_30_compare_lists;


sub action_35_correlation_coefficient {
    # each subroutine gets four arguments:
    my $num = shift;    # the number in the action chain
                        # (useful for action specific parameters)
    my $action = shift; # the name of the action 
                        # (same as subroutine name)

    # return list with the headers required for this function
    # when called with special argument
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM,$FLAG_CHANNEL_TERM;
    }

    &correlation_coefficient_setup($num,$action);
}
$function{'action_35_correlation_coefficient'} = \&action_35_correlation_coefficient;


sub action_40_annotate {
    # each subroutine gets four arguments:
    my $num = shift;    # the number in the action chain
                        # (useful for action specific parameters)
    my $action = shift; # the name of the action 
                        # (same as subroutine name)

    # return list with the headers required for this function
    # when called with special argument
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM;
    }

    &annotation_setup($num,$action);
}
$function{'action_40_annotate'} = \&action_40_annotate;



sub action_50_output {
    # each subroutine gets four arguments:
    my $num = shift;    # the number in the action chain
                        # (useful for action specific parameters)
    my $action = shift; # the name of the action 
                        # (same as subroutine name)

    # return list with the headers required for this function
    # when called with special argument
    if ($num eq '__columns__') {
	return @output;
    }

    &output_file_setup($num,$action);
}
$function{'action_50_output'} = \&action_50_output;



sub action_11_R_ArrayPlot {
    # each subroutine gets two arguments:
    my $num = shift;    # the number in the action chain
                        # (useful for action specific parameters)
    my $action = shift; # the name of the action 
                        # (same as subroutine name)

    # return list with the headers required for this function
    # when called with special argument
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM;
    }

    # carry out action on each file ...
    foreach my $file (sort keys %data) {
	
	# ... and each label
	foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	    
	    # skip columns that are not label specific ('spot labels')
	    next if ($label eq '__NO_LABEL__');

	    # put all the main calculations into a separate subroutine
	    # the return value should be HTML code that will be directly
	    # included into the results page
	    print &R_ArrayPlot($num,$file,$label,$action);
	}
    }
}
$function{'action_11_R_ArrayPlot'} = \&action_11_R_ArrayPlot;

sub action_12_R_quintile_histogram {
    # each subroutine gets two arguments:
    my $num = shift;    # the number in the action chain
                        # (useful for action specific parameters)
    my $action = shift; # the name of the action 
                        # (same as subroutine name)

    # return list with the headers required for this function
    # when called with special argument
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM;
    }

    # carry out action on each file ...
    foreach my $file (sort keys %data) {
	
	# ... and each label
#	foreach my $label (sort keys %{$data{$file}{'__DATA__'}}, 'ratio') {
	foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	    
	    # skip columns that are not label specific ('spot labels')
	    if ($label eq '__NO_LABEL__') {
		next;
	    }

	    my ($image, $info);

	    if ($refine_parameters) {
		$info = 'Parameter Refinement!<BR>';
	    } else {

		# put all the main calculations into a separate subroutine
		# the return value should be HTML code that will be directly
		# included into the results page
		$image = &R_quintile_histogram($num,$file,$label,$action);
	    }

	    # format output (include image and choice for parameters)
	    print &format_output(
				 {
				     'prefix' => 'histo_',
				     'num' => $num,
				     'file' => $file,
				     'label' => $label,
				     'action' => $action,
				     'image' => $image,
				     'info' => $info
				     }
				 );
	}
    }
}
$function{'action_12_R_quintile_histogram'} = \&action_12_R_quintile_histogram;

sub action_13_signal_scatter_plot {
    my $num = shift;    # the number in the action chain
                        # (useful for action specific parameters)
    my $action = shift; # the name of the action 
                        # (same as subroutine name)

    # return list with the headers required for this function
    # when called with special argument
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &signal_plotting_setup($num,$action);
}
$function{'action_13_signal_scatter_plot'} = \&action_13_signal_scatter_plot;


sub action_14_signal_boxplot {
    my $num = shift;    # the number in the action chain
                        # (useful for action specific parameters)
    my $action = shift; # the name of the action 
                        # (same as subroutine name)

    # return list with the headers required for this function
    # when called with special argument
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    # put all the main calculations into a separate subroutine
    # the return value should be HTML code that will be directly
    # included into the results page
    my ($image,$info) = &signal_boxplotting($num,$action);
    
#    print &format_output('boxplot_',$num,'','__ALL__',$action,$image,$info);
    print &format_output( 
			  {
			      'prefix' => 'boxplot_',
			      'num' => $num,
			      'label' => '__ALL__',
			      'action' => $action,
			      'image' => $image,
			      'info' => $info,
			      'outtext' => "All files, all labels, data column: <B>$FG_SIGNAL_TERM</B>"
			      }
			  );

}
$function{'action_14_signal_boxplot'} = \&action_14_signal_boxplot;

sub action_15_signal_boxplot_printtip {
    my $num = shift;    # the number in the action chain
                        # (useful for action specific parameters)
    my $action = shift; # the name of the action 
                        # (same as subroutine name)

    # return list with the headers required for this function
    # when called with special argument
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &boxplot_pt_setup($num,$action);
}
$function{'action_15_signal_boxplot_printtip'} = \&action_15_signal_boxplot_printtip;

sub action_17_rep_signal_scatter_plot {
    my $num = shift;    # the number in the action chain
                        # (useful for action specific parameters)
    my $action = shift; # the name of the action 
                        # (same as subroutine name)

    # return list with the headers required for this function
    # when called with special argument
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    # put all the main calculations into a separate subroutine
    # the return value should be HTML code that will be directly
    # included into the results page
    &rep_signal_plotting_setup($num,$action);
}
$function{'action_17_rep_signal_scatter_plot'} = \&action_17_rep_signal_scatter_plot;

sub action_18_ratio_signal_scatter_plot {
    my $num = shift;    # the number in the action chain
                        # (useful for action specific parameters)
    my $action = shift; # the name of the action 
                        # (same as subroutine name)

    # return list with the headers required for this function
    # when called with special argument
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    # put all the main calculations into a separate subroutine
    # the return value should be HTML code that will be directly
    # included into the results page
    &ratio_signal_plotting_setup($num,$action);
}
$function{'action_18_ratio_signal_scatter_plot'} = \&action_18_ratio_signal_scatter_plot;

sub action_20_MA_plot {
    my $num = shift;    # the number in the action chain
                        # (useful for action specific parameters)
    my $action = shift; # the name of the action 
                        # (same as subroutine name)

    # return list with the headers required for this function
    # when called with special argument
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    # carry out action on each file ...
    foreach my $file (sort keys %data) {
	
	# put all the main calculations into a separate subroutine
	# the return value should be HTML code that will be directly
	# included into the results page
	my ($label,$image,$info) = &MA_plotting($num,$file,$action);

	print &format_output( 
			      {
				  'prefix' => 'ma_plot_',
				  'num' => $num,
				  'file' => $file,
				  'label' => '__ALL__',
				  'action' => $action,
				  'image' => $image,
				  'info' => $info
				  }
			      );
    }
}
$function{'action_20_MA_plot'} = \&action_20_MA_plot;


sub action_25_pathway_plot {
    my $num = shift;    # the number in the action chain
                        # (useful for action specific parameters)
    my $action = shift; # the name of the action 
                        # (same as subroutine name)

    # return list with the headers required for this function
    # when called with special argument
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &pathway_plotting($num,$action);
}
$function{'action_25_pathway_plot'} = \&action_25_pathway_plot;

sub action_01_bg_plot {
    my $num = shift;
    my $action = shift;    

    # return list with required columns
    if ($num eq '__columns__') {
	return $BG_SIGNAL_TERM,$FLAG_CHANNEL_TERM;
    }

    # carry out action on each file ...
    foreach my $file (sort keys %data) {
	
	# ... and each label
	foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	    
	    # skip columns that are not label specific ('spot labels')
	    next if ($label eq '__NO_LABEL__');
	    
	    # draw chip with values from background column
	    my $header = $BG_SIGNAL_TERM;
	    foreach my $test ('BG_CALC', $BG_SIGNAL_TERM.'_M') {
		if (defined $data_column_available{$file}{$label}{$test}) {
		    $header = $test;
		}
	    }
	    $last_column_used = $header;
	    my ($image,$info,$min_val,$max_val);

	    if ($refine_parameters) {
		$info = 'Parameter Refinement!<BR>';
	    } else {
		($image,$info,$min_val,$max_val) = &draw_chip( # 'bg_',$num,$file,$label,$header,$action
							      {
								  'prefix' => 'bg_',
								  'num' => $num,
								  'file' => $file,
								  'label' => $label,
								  'header' => $header,
								  'action' => $action
								  }
							      );
	    }
	    
	    if (defined $image
		and
		$image =~ /Error\:/i) {
		print "<BR><B>$image<B><BR>\n";
		next;
	    }

	    # format output (include image and choice for parameters)
	    print &format_output( 
				  {
				      'prefix' => 'bg_',
				      'num' => $num,
				      'file' => $file,
				      'label' => $label,
				      'action' => $action,
				      'image' => $image,
				      'info' => $info,
				      'min' => $min_val,
				      'max' => $max_val,
#				      'image_link' => "<A HREF=\"http://localhost/cgi-bin/arraypipe.pl?position\" TARGET=\"big_pic\">"
				      }
				  );
	}
	
	print "<BR>\n";
    }
}
 
# add reference handle to function hash:
$function{'action_01_bg_plot'} = \&action_01_bg_plot;


sub action_02_fg_plot {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM,$FLAG_CHANNEL_TERM;
    }

    my $file_counter = scalar keys %data;

    # carry out action on each file ...
    foreach my $file (sort keys %data) {
	
	# ... and each label
	foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	    
	    # skip columns that are not label specific ('spot labels')
	    next if ($label eq '__NO_LABEL__');
	    
	    # draw chip with values from (corrected/merged) foreground column
	    my $header = $FG_SIGNAL_TERM;
	    foreach my $column (@{$item_data_column{$action}}) {
		if (defined $data_column_available{$file}{$label}{$column}) {
		    $header = $column;
		    last;
		}
	    }
	    $last_column_used = $header;
	    my ($image,$info,$min_val,$max_val);

	    if ($refine_parameters) {
		$info = 'Parameter Refinement!<BR>';
		$image = '';
	    } else {
		($image,$info,$min_val,$max_val) = &draw_chip( # 'fg_',$num,$file,$label,$header,$action
							       {
								   'prefix' => 'fg_',
								   'num' => $num,
								   'file' => $file,
								   'label' => $label,
								   'header' => $header,
								   'action' => $action
								   }
							       );	    
		if ($image =~ /Error\:/i) {
		    print "<BR><B>$image<B><BR>\n";
		    next;
		}
	    }

	    # format output (include image and choice for parameters)
	    print &format_output( 
				  {
				      'prefix' => 'fg_',
				      'num' => $num,
				      'file' => $file,
				      'label' => $label,
				      'action' => $action,
				      'image' => $image,
				      'info' => $info,
				      'min' => $min_val,
				      'max' => $max_val
				      }
				  );
	}
	print "<BR>\n";
	$file_counter--;
	if ($file_counter) {
	    print "<HR><BR>\n";
	}
    }
}
$function{'action_02_fg_plot'} = \&action_02_fg_plot;


sub action_08_norm_plot {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM,$FLAG_CHANNEL_TERM;
    }

    # carry out action on each file ...
    foreach my $file (sort keys %data) {
	
	# ... and each label
	foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	    
	    # skip columns that are not label specific ('spot labels')
	    next if ($label eq '__NO_LABEL__');
	    
	    my ($image,$info,$min_val,$max_val);

	    if ($refine_parameters) {
		$info = 'Parameter Refinement!<BR>';
		$image = '';
	    } else {
		unless (defined $data{$file}{'__DATA__'}{$label}{'Norm'}
			or
			defined $data{$file}{'__DATA__'}{$label}{'Norm_L'}) {
		    print "<B>No normalized data available for ".(basename $file).", $label, please apply filter first!</B><BR>\n";
		    next;
		}
		
		# draw chip with values from background column
		my $header = '';
		foreach my $test ('Norm_L_M', 'Norm_M_L', 'Norm_M', 'Norm_L', 'Norm') {
		    if (defined $data_column_available{$file}{$label}{$test}) {
			$header = $test;
			last;
		    }
		}
		$last_column_used = $header;
		($image,$info,$min_val,$max_val) = &draw_chip( # 'norm_',$num,$file,$label,$header,$action
							       {
								   'prefix' => 'norm_',
								   'num' => $num,
								   'file' => $file,
								   'label' => $label,
								   'header' => $header,
								   'action' => $action
								  }
							       );
		
		if ($image =~ /Error\:/i) {
		    print "<BR><B>$image<B><BR>\n";
		    next;
		}
	    }

	    # format output (include image and choice for parameters)
	    print &format_output( 
				  {
				      'prefix' => 'norm_',
				      'num' => $num,
				      'file' => $file,
				      'label' => $label,
				      'action' => $action,
				      'image' => $image,
				      'info' => $info,
				      'min' => $min_val,
				      'max' => $max_val
				      }
				  );
	}
	print "<BR>\n";
    }
}
$function{'action_08_norm_plot'} = \&action_08_norm_plot;

sub action_04_ratio_plot {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM,$FLAG_CHANNEL_TERM;
    }

    # carry out action on each file ...
    foreach my $file (sort keys %data) {
	
	my ($header,$info_ratio,$image,$info,$min_val,$max_val);

	if ($refine_parameters) {
	    $info = 'Parameter Refinement!<BR>';
	    $image = '';
	} else {
	    # calculate ratios either from (bg-corrected) foreground signals
	    # or from normalized values
	    ($header,$info_ratio) = &ratio_calc($num,$file,$action,'__NO_LABEL__');
	    
	    # draw chip with values from ratio column
	    ($image,$info,$min_val,$max_val) = &draw_chip( # 'ratio_',$num,$file,'__NO_LABEL__',$header,$action
							   {
							       'prefix' => 'ratio_',
							       'num' => $num,
							       'file' => $file,
							       'label' => '__NO_LABEL__',
							       'header' => $header,
							       'action' => $action
							       }
							   );
	    
	    if ($image =~ /Error\:/i) {
		print "<BR><B>$image<B><BR>\n";
		next;
	    }
	}

	# format output (include image and choice for parameters)
	print &format_output( 
			      {
				  'prefix' => 'ratio_',
				  'num' => $num,
				  'file' => $file,
				  'label' => '__NO_LABEL__',
				  'action' => $action,
				  'image' => $image,
				  'info' => $info_ratio.$info,
				  'min' => $min_val,
				  'max' => $max_val
				  }
			      );

	print "<BR>\n";
    }
}
$function{'action_04_ratio_plot'} = \&action_04_ratio_plot;


##########################
### FILTER SUBROUTINES ###
##########################

sub filter_02_fg_min_bg {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $BG_SIGNAL_TERM;
    }

    &simple_background_correction_setup($num,$action);
}
$function{'filter_02_fg_min_bg'} = \&filter_02_fg_min_bg;


# background correction ala Jenny Bryan:
# take the median of the lower x % of the forground values for each subgrid
# and subtract from the foreground
sub filter_03_fg_min_percentile_fg {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM;
    }

    &bg_correction_percentile_setup($num,$action);

}
$function{'filter_03_fg_min_percentile_fg'} = \&filter_03_fg_min_percentile_fg;


# background correction with Loess parameters:
# predict loess for a certain neighbourhood and subtract it from the foreground
sub filter_03b_loess_bg_correction {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM;
    }

    &bg_correction_loess_setup($num,$action);

}
$function{'filter_03b_loess_bg_correction'} = \&filter_03b_loess_bg_correction;


sub filter_04_flag_markers {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM;
    }

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $header = ''; 
    foreach my $column (@{$item_data_column{$action}}) {
	my $found = 1;
	foreach my $file (keys %data) {
	    unless ($data_column_available{$file}{'__NO_LABEL__'}{$column}) {
		$found = 0;
	    }
	}
	if ($found) {
	    $header = $column;
	    last;
	}
    }

    # check for specific settings:
    $header = &get_user_settings($header,'markers_data_column',$force_global_opt,$num);

    unless ($header) {
	return "Error: No data column defined to work on for $action!\n";
    }

    # carry out action on each file ...
    foreach my $file (sort keys %data) {
		    
	my ($info);
	if ($refine_parameters) {
	    $info = 'Parameter Refinement!<BR>';
	} else {
	    
	    $info = &flag_markers($num,$file,$action,$header,$FLAG_CHANNEL_TERM);
	}
	print &format_output( # info'markers_',$num,$file,'__ALL__',$action,'',$info
			     {
				 'prefix' => 'markers_',
				 'num' => $num,
				 'file' => $file,
				 'label' => '__ALL__',
				 'action' => $action,
				 'info' => $info
				 }
			     );
    }
}
$function{'filter_04_flag_markers'} = \&filter_04_flag_markers;

# read in upper and lower threshold
# either adjust data points outside the range
# or delete them
sub filter_05_set_cutoffs {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FG_COR_SIGNAL_TERM;
    }

    &set_cutoffs_setup($num,$action);
}
$function{'filter_05_set_cutoffs'} = \&filter_05_set_cutoffs;


# read in upper and lower threshold
# either adjust data points outside the range
# or delete them
sub filter_04g_flag_change {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM, $FG_COR_SIGNAL_TERM;
    }

    &flag_change_setup($num,$action);
}
$function{'filter_04g_flag_change'} = \&filter_04g_flag_change;


# flagging of duplicate spots:
# divide the higher value from a pair of neighbouring duplicates
# by the lower value and flag the ones that are above a user
# specified threshold
# (modified from Byron Kuo's script 'flaggingChannels.pl', 
# Inimex Pharmaceuticals Inc., June 25th, 2003).
sub filter_01_dup_flagging {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM, $FG_SIGNAL_TERM;
    }

    # carry out action on each file ...
    foreach my $file (sort keys %data) {
	
	my $out = &dup_spot_flagging($num,$file,$action,$SPOT_ID_TERM,$FLAG_CHANNEL_TERM);
	print &format_output( # 'flag_dups_',$num,$file,'__ALL__',$action,'',$out
			     {
				 'prefix' => 'flag_dups_',
				 'num' => $num,
				 'file' => $file,
				 'label' => '__ALL__',
				 'action' => $action,
				 'info' => $out
				 }
			     );
    }
}
$function{'filter_01_dup_flagging'} = \&filter_01_dup_flagging;


# flagging of flawed replicate spots:
sub filter_01b_rep_flawed_flagging {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM, $FG_SIGNAL_TERM;
    }

    &rep_flawed_flagging_setup($num,$action);
}
$function{'filter_01b_rep_flawed_flagging'} = \&filter_01b_rep_flawed_flagging;


# normalization with marrayNorm's loess
sub filter_06a_limma_loess_printtip {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM, $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &limma_loess_setup($num,$action,'printtip');
    
}
$function{'filter_06a_limma_loess_printtip'} = \&filter_06a_limma_loess_printtip;



# normalization with marrayNorm's loess
sub filter_06b_limma_loess_global {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM, $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &limma_loess_setup($num,$action,'global');
    
}
$function{'filter_06b_limma_loess_global'} = \&filter_06b_limma_loess_global;


# normalization with vsn
# (with help from Marc Rousseau's scripts (Inimex Pharmaceuticals Inc., BC, Canada, June 25th, 2003)
sub filter_07_vsn_normalization {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM, $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &vsn_normalization_setup($num,$action);
    
}
$function{'filter_07_vsn_normalization'} = \&filter_07_vsn_normalization;

# normalization with marrayNorm
#sub filter_08_marrayNorm_normalization {
#    my $num = shift;
#    my $action = shift;

#    # return list with required columns
#    if ($num eq '__columns__') {
#	return $SPOT_ID_TERM, $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
#    }

#    &marrayNorm_normalization_setup($num,$action);
    
#}
#$function{'filter_08_marrayNorm_normalization'} = \&filter_08_marrayNorm_normalization;


# normalization with marrayNorm's median
sub filter_08a_marrayNorm_median {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM, $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &marrayNorm_normalization_setup($num,$action,'median');
    
}
$function{'filter_08a_marrayNorm_median'} = \&filter_08a_marrayNorm_median;


# normalization with marrayNorm's loess
sub filter_08b_marrayNorm_loess {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM, $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &marrayNorm_normalization_setup($num,$action,'loess');
    
}
$function{'filter_08b_marrayNorm_loess'} = \&filter_08b_marrayNorm_loess;


# normalization with marrayNorm's twoD
sub filter_08c_marrayNorm_twoD {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM, $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &marrayNorm_normalization_setup($num,$action,'twoD');
    
}
$function{'filter_08c_marrayNorm_twoD'} = \&filter_08c_marrayNorm_twoD;

# normalization with marrayNorm's printTipLoess
sub filter_08d_marrayNorm_printTipLoess {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM, $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &marrayNorm_normalization_setup($num,$action,'printTipLoess');
    
}
$function{'filter_08d_marrayNorm_printTipLoess'} = \&filter_08d_marrayNorm_printTipLoess;


# normalization with marrayNorm's globalMAD
sub filter_08e_marrayNorm_globalMAD {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM, $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &marrayNorm_normalization_setup($num,$action,'globalMAD');
    
}
$function{'filter_08e_marrayNorm_globalMAD'} = \&filter_08e_marrayNorm_globalMAD;


# normalization with marrayNorm's printTipMAD
sub filter_08f_marrayNorm_printTipMAD {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM, $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &marrayNorm_normalization_setup($num,$action,'printTipMAD');
    
}
$function{'filter_08f_marrayNorm_printTipMAD'} = \&filter_08f_marrayNorm_printTipMAD;


# normalization with marrayNorm's scalePrintTipMAD
sub filter_08g_marrayNorm_scalePrintTipMAD {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM, $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &marrayNorm_normalization_setup($num,$action,'scalePrintTipMAD');
    
}
$function{'filter_08g_marrayNorm_scalePrintTipMAD'} = \&filter_08g_marrayNorm_scalePrintTipMAD;


# merge duplicate data points together
sub filter_10_merge_dups {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $SPOT_ID_TERM, $FG_SIGNAL_TERM, $BG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }


    # carry out action on each file ...
    foreach my $file (sort keys %data) {
	
	my $out = &merge_dups($num,$action,$file,$SPOT_ID_TERM,$FLAG_CHANNEL_TERM);
    }        
}
$function{'filter_10_merge_dups'} = \&filter_10_merge_dups;


# merging of replicates into one data set
sub filter_12_merge_all_reps {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &merge_replicates_setup($num,$action,'all');
    
}
$function{'filter_12_merge_all_reps'} = \&filter_12_merge_all_reps;

# merging of replicates into one data set
sub filter_11_merge_tech_reps {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &merge_replicates_setup($num,$action,'tech');
    
}
$function{'filter_11_merge_tech_reps'} = \&filter_11_merge_tech_reps;


# filter data points according to thresholds imposed on their values
sub filter_20_by_value {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &filter_by_value_setup($num,$action);
}
$function{'filter_20_by_value'} = \&filter_20_by_value;


# selecting differentially expressed genes
# (using local window and SD)
sub filter_31_local_int_dep_z {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &local_int_dep_z_setup($num,$action);
    
}
$function{'filter_31_local_int_dep_z'} = \&filter_31_local_int_dep_z;


# selecting differentially expressed genes
# (using Welch t-test)
sub filter_35_student_ttest {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &ttest_one_group_setup($num,$action,'student');
    
}
$function{'filter_35_student_ttest'} = \&filter_35_student_ttest;


# selecting differentially expressed genes
# (using paired t-test)
sub filter_35c_paired_ttest {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &ttest_one_group_setup($num,$action,'paired');
    
}
$function{'filter_35c_paired_ttest'} = \&filter_35c_paired_ttest;

# selecting differentially expressed genes
# (using Welch t-test)
sub filter_35a_welch_ttest {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &ttest_one_group_setup($num,$action,'welch');
    
}
$function{'filter_35a_welch_ttest'} = \&filter_35a_welch_ttest;

# selecting differentially expressed genes
# (using Wilcox t-test)
sub filter_35b_wilcox_ttest {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &ttest_one_group_setup($num,$action,'wilcox');
#    &wilcox1_ttest_setup($num,$action);
    
}
$function{'filter_35b_wilcox_ttest'} = \&filter_35b_wilcox_ttest;

# selecting differentially expressed genes
# (using Permutation for calculation of p-value)
sub filter_33_permutation_p_value {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &permutation_p_value_setup($num,$action);
    
}
$function{'filter_33_permutation_p_value'} = \&filter_33_permutation_p_value;

 
# selecting differentially expressed genes
# (using Welch t-test)
sub filter_38_student_ttest {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &ttest_between_groups_setup($num,$action,'student');
    
}
$function{'filter_38_student_ttest'} = \&filter_38_student_ttest;

 
# selecting differentially expressed genes
# (using Welch t-test)
sub filter_38a_welch_ttest {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &ttest_between_groups_setup($num,$action,'welch');
    
}
$function{'filter_38a_welch_ttest'} = \&filter_38a_welch_ttest;

# selecting differentially expressed genes
# (using Wilcox t-test)
sub filter_38b_wilcox_ttest {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &ttest_between_groups_setup($num,$action,'wilcox');
#    &wilcox1_ttest_setup($num,$action);
    
}
$function{'filter_38b_wilcox_ttest'} = \&filter_38b_wilcox_ttest;

# selecting differentially expressed genes
# (using Permutation for calculation of p-value)
sub filter_40_anova {
    my $num = shift;
    my $action = shift;

    # return list with required columns
    if ($num eq '__columns__') {
	return $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM;
    }

    &anova_setup($num,$action);
    
}
$function{'filter_40_anova'} = \&filter_40_anova;


#####################################
### variables for HTML interface: ###
#####################################

# FLAG settings:
#my @flag_names = (  'no flag','automatic','markers', 'dup-flaw','floor', 'ceiling', 'warning', 'no_ratio','error',  'undefined','low_quality','user1', 'list1',   'list2','list3','list4',  'list5', 'list6', 'list7', 'list8', 'list9', 'list10');
my @flag_names = (  'no flag','automatic','markers', 'dup-flaw','floor', 'ceiling','warning', 'no_ratio','error',  'undefined','low_quality', 'user1', 'list1');
my @flag_colours = ('black',  'cyan',     'purple',  'magenta',  'maroon','darkgreen',   'olive',   'yellow',  'red',     'pink',   'thistle','coral', 'orange',  'blue', 'magenta', 'green','yellow', 'cyan', 'purple', 'red', 'darkorange', 'olive');
my @pref_flag_colours = ('cyan','magenta','orange','green','blue','yellow','maroon','firebrick','darkcyan','darkmagenta','darkgreen','brown','turquoise','springgreen','pink','darkorange','coral','thistle','red','sienna');
my %pref_flag2col = ();
my %flag_values = ();
my $flag_num_counter = -1;
my %flag_values_num = ();
my %used_flags = ();
my $highest_flag_num = 0;
# calculate a number for each flag
# these are products of 2, which allows to add them
# and take them apart again
foreach my $flag (@flag_names) {
    my $num = $flag_num_counter + 1;
    if ($flag_num_counter >=0) {
	$num = 2**$flag_num_counter;
    }
    $flag_num_counter++;
    $flag_values{$flag} = $num;
    $flag_values_num{$num} = $flag;
    $highest_flag_num = $num;
}

my @all_flags = ();
my @all_flags_num = ();
foreach (sort { $a <=> $b } keys %flag_values_num) {
    next if ($_ == 0);
    push @all_flags, $flag_values_num{$_};
    
    # sort @all_flags_num in descending order
    unshift @all_flags_num, $_;
}

# assign some important flags with colours already
$pref_flag2col{$flag_values{'no flag'}} = 'black';
$pref_flag2col{'black'} = $flag_values{'no flag'};
$pref_flag2col{$flag_values{'automatic'}} = 'cyan';
$pref_flag2col{'cyan'} = $flag_values{'automatic'};
$pref_flag2col{$flag_values{'markers'}} = 'magenta';
$pref_flag2col{'magenta'} = $flag_values{'markers'};

# we need a few more colours here for combination of flags
my %score2col = (0,'black',
		 1,'blue',
		 1.96,'red',
		 2,'red',
		 3,'orange',
		 4,'green',
		 5,'cyan',
		 6,'magenta',
		 7,'pink',
		 8,'yellow',
		 );

# variables for each item (hash with array as value)
# each item listed above should have one or more
my %item_var = (
		'action_01_bg_plot', [ 'bg_square_size',
				       'bg_grid',
				       'bg_skip_flags',
				       'bg_orientation',
				       'colour_spread_header',
				       'bg_min',
				       'bg_max',
				       'bg_border_setting' ],
		
		'action_02_fg_plot', [ 'fg_square_size',
				       'fg_grid',
				       'fg_skip_flags',
				       'fg_orientation',
				       'colour_spread_header',
				       'fg_min',
				       'fg_max',
				       'fg_border_setting' ],

		'action_08_norm_plot', [ 'norm_square_size',
					 'norm_grid',
					 'norm_skip_flags',
					 'norm_orientation',
					 'colour_spread_header',
					 'norm_min',
					 'norm_max',
					 'norm_border_setting' ],

		'action_04_ratio_plot', [ 'ratio_square_size',
					  'ratio_grid',
					  'ratio_skip_flags',
					  'ratio_orientation',
					  'ratio_ch1',
					  'log_trans_ratios',
					  'colour_spread_header',
					  'ratio_min',
					  'ratio_max',
					  'ratio_border_setting' ],

		'action_12_R_quintile_histogram', [ 'histo_take_min_val', 
						    'histo_take_log', 
						    'histo_add_quintile', 
						    'histo_skip_flags' ],
		'action_13_signal_scatter_plot', [ 'signal_x', 
						   'signal_log', 
						   'signal_layout', 
						   'signal_grids_per_plot', 
						   'signal_coloured', 
						   'signal_skip_flags' ],
		
		'action_14_signal_boxplot', [ 'boxplot_labels', 
					      'boxplot_log', 
					      'boxplot_skip_flags' ],
		
		'action_15_signal_boxplot_printtip', [ 'boxplot_pt_files', 
						       'boxplot_pt_labels', 
						       'boxplot_pt_colour', 
						       'boxplot_pt_log', 
						       'boxplot_pt_add_info', 
						       'boxplot_pt_skip_flags' ],
		
		'action_17_rep_signal_scatter_plot', [ 'rep_signal_log', 
						       'rep_signal_skip_flags' ],
		
		'action_18_ratio_signal_scatter_plot', [ 'ratio_signal_files', 
							 'ratio_signal_skip_flags' ],
		
		'action_20_MA_plot', [ 'ma_plot_ch1', 
				       'ma_skip_flags' ],
		
		'action_25_pathway_plot', [ 'pathway_files', 
					    'pathway_labels', 
					    'pathway_boxplot', 
					    'pathway_connectors', 
					    'pathway_skip_flags'],
		
		'action_30_compare_lists', [ 'compare_lists_files', 
					     'compare_lists_make_new', 
					     'compare_lists_flags' ],
		
		'action_35_correlation_coefficient', [ 'correlation_coefficient_files', 
						       'correlation_coefficient_method', 
						       'correlation_coefficient_ratio', 
						       'correlation_coefficient_log', 
						       'correlation_coefficient_rounding', 
						       'correlation_coefficient_skip_flags' ],
		
		'action_40_annotate', [ 'annotate_files', 
					'annotate_web_file', 
					'annotate_items_probelynx', 
					'annotate_items_funcdb' ],
		
		'action_50_output', [ 'output_columns', 
				      'output_skip_flags', 
				      'output_sample_lines', 
				      'output_rounding' ],
		
		'filter_02_fg_min_bg', [ 'fg_min_bg_files', 
					 'fg_min_bg_skip_flags', 
					 'fg_min_bg_operator' ],
		
		'filter_03_fg_min_percentile_fg', [ 'fg_min_percentile_files', 
						    'fg_min_percentile_skip_flags',
						    'percentile', 
						    'percentile_action' ],
		
		'filter_03b_loess_bg_correction', [ 'bg_loess_files', 
						    'bg_loess_skip_flags',
						    'bg_loess_span' ],
		
		'filter_04_flag_markers', [ 'marker_flag', 
					    'markers' ],
		
		'filter_04g_flag_change', [ 'flag_change_files', 
					    'flag_change_in', 
					    'flag_change_out' ],
		
		'filter_05_set_cutoffs', [ 'cutoff_files', 
					   'cutoff_allowed_min', 
					   'cutoff_min', 
					   'cutoff_min_action', 
					   'cutoff_max', 
					   'cutoff_max_action', 
					   'cutoff_skip_flags'],
		
		'filter_01_dup_flagging', [ 'dup_cutoff', 
					    'dup_unit', 
					    'dup_flag', 
					    'dup_skip_flags' ],
		
		'filter_01b_rep_flawed_flagging', [ 
						    'rep_flawed_files', 
						    'rep_flawed_cutoff', 
						    'rep_flawed_unit', 
#						    'rep_flawed_percent', 
						    'rep_flawed_flag', 
#						    'rep_flawed_extend', 
						    'rep_flawed_flag_single', 
						    'rep_flawed_within_slide', 
						    'rep_flawed_label', 
						    'rep_flawed_skip_flags' 
						    ],

		'filter_06a_limma_loess_printtip', [ 'limma_loess_printtip_files', 
						     'limma_loess_printtip_span', 
						     'limma_loess_printtip_bin', 
						     'limma_loess_printtip_iter', 
						     'limma_loess_printtip_skip_flags' ],
		
		'filter_06b_limma_loess_global', [ 'limma_loess_global_files', 
						   'limma_loess_global_span', 
						   'limma_loess_global_bin', 
						   'limma_loess_global_iter', 
						   'limma_loess_global_skip_flags' ],
		
		'filter_07_vsn_normalization', [ 'vsn_iter', 
						 'vsn_dehuberize', 
						 'vsn_skip_flags',
						 'vsn_global' ],
		
		'filter_08a_marrayNorm_median', [ 'marrayNormmedian_files', 
						  'marrayNormmedian_skip_flags' ],
		
		'filter_08b_marrayNorm_loess', [ 'marrayNormloess_files', 
						 'marrayNormloess_span', 
						 'marrayNormloess_skip_flags' ],
		
		'filter_08c_marrayNorm_twoD', [ 'marrayNormtwoD_files', 
						'marrayNormtwoD_span', 
						'marrayNormtwoD_skip_flags' ],
		
		'filter_08d_marrayNorm_printTipLoess', [ 'marrayNormprintTipLoess_files', 
							 'marrayNormprintTipLoess_span', 
							 'marrayNormprintTipLoess_skip_flags' ], 
		
		'filter_08g_marrayNorm_scalePrintTipMAD', [ 'marrayNormscalePrintTipMAD_files', 
							    'marrayNormscalePrintTipMAD_span', 
							    'marrayNormscalePrintTipMAD_skip_flags' ], 
		
		'filter_08e_marrayNorm_globalMAD', [ 'marrayNormglobalMAD_files', 
						     'marrayNormglobalMAD_global', 
						     'marrayNormglobalMAD_skip_flags' ], 
		
		'filter_08f_marrayNorm_printTipMAD', [ 'marrayNormprintTipMAD_files', 
						       'marrayNormprintTipMAD_global', 
						       'marrayNormprintTipMAD_skip_flags' ], 
		
		'filter_10_merge_dups', [ 'merge_dups_target', 
					  'merge_skip_flags' ],
		
		'filter_11_merge_tech_reps', [ 'merge_tech_reps_skip_flags', 
					       'merge_tech_reps_dups', 
#					       'merge_tech_weight', 
					       'merge_tech_reps_continuation', 
					       'merge_tech_reps_method' ],
		
		'filter_12_merge_all_reps', [ 'merge_all_reps_skip_flags', 
					      'merge_all_reps_dups', 
#					      'merge_all_weight', 
					      'merge_all_reps_continuation', 
					      'merge_all_reps_method' ],
		
		'filter_20_by_value', [ 'by_value_files', 
					'by_value_strict_skip', 
					'by_value_labels', 
					'by_value_max_field', 
					'by_value_ignore_sign1', 
					'by_value_max_op', 
					'by_value_max', 
					'by_value_connector', 
					'by_value_files2', 
					'by_value_labels2', 
					'by_value_min_field', 
					'by_value_ignore_sign2', 
					'by_value_min_op', 
					'by_value_min', 
					'by_value_flag', 
					'by_value_skip_flags', 
					'by_value_make_list', 
					'by_value_copy_list', 
					'by_value_sticky_list' ],
		
		'filter_31_local_int_dep_z', [ 'local_int_dep_z_files', 
					       'local_int_dep_z_window', 
					       'local_int_dep_z_orient', 
					       'local_int_dep_z_ch1', 
					       'local_int_dep_z_skip_flags' ],
		
		'filter_35_student_ttest', [ 'student_ttest_1_skip_flags', 
					     'student_ttest_1_dups', 
					     'student_ttest_1_input', 
					     'student_ttest_1_outheader', 
					     'student_ttest_1_outchannel' ],
		
		'filter_35c_paired_ttest', [ 'paired_ttest_1_skip_flags', 
					     'paired_ttest_1_dups', 
					     'paired_ttest_1_outheader', 
					     'paired_ttest_1_outchannel' ],
		
		'filter_35a_welch_ttest', [ 'welch_ttest_1_skip_flags', 
					    'welch_ttest_1_dups', 
					    'welch_ttest_1_input', 
					    'welch_ttest_1_outheader', 
					    'welch_ttest_1_outchannel' ],
		
		'filter_35b_wilcox_ttest', [ 'wilcox_ttest_1_skip_flags', 
					     'wilcox_ttest_1_dups', 
					     'wilcox_ttest_1_input', 
					     'wilcox_ttest_1_outheader', 
					     'wilcox_ttest_1_outchannel' ],
		
		'filter_33_permutation_p_value', [ 'permutation_p_value_skip_flags', 
						   'permutation_p_value_dups', 
						   'permutation_p_value_iterations', 
						   'permutation_p_value_outheader', 
						   'permutation_p_value_outchannel' ],
		
		'filter_38_student_ttest', [ 'student_ttest_2_skip_flags', 
					     'student_ttest_2_dups', 
					     'student_ttest_2_outheader', 
					     'student_ttest_2_outchannel' ],
		
		'filter_38a_welch_ttest', [ 'welch_ttest_2_skip_flags', 
					    'welch_ttest_2_dups', 
					    'welch_ttest_2_outheader', 
					    'welch_ttest_2_outchannel' ],
		
		'filter_38b_wilcox_ttest', [ 'wilcox_ttest_2_skip_flags', 
					     'wilcox_ttest_2_dups', 
					     'wilcox_ttest_2_outheader', 
					     'wilcox_ttest_2_outchannel' ],
		
		'filter_40_anova', [ 'anova_skip_flags', 
				     'anova_dups', 
				     'anova_outheader', 
				     'anova_outchannel' ],
		);

# definitions of variables, 
# each variable listed above must be defined!

# type of variable:
my %var_def_type = (
		    'histo_take_min_val', 'text',
		    'bg_loess_span', 'text',
		    'permutation_p_value_iterations', 'text',
		    'local_int_dep_z_files', 'list',
		    'local_int_dep_z_window', 'text', 
		    'local_int_dep_z_orient', 'radio', 
		    'local_int_dep_z_ch1', 'select', 
		    'local_int_dep_z_skip_flags', 'list',
		    'student_ttest_1_skip_flags', 'list',
		    'student_ttest_1_dups', 'radio',
		    'student_ttest_1_input', 'select',
		    'student_ttest_1_outheader', 'select',
		    'student_ttest_1_outchannel', 'select',
		    'paired_ttest_1_skip_flags', 'list',
		    'paired_ttest_1_dups', 'radio',
		    'paired_ttest_1_outheader', 'select',
		    'paired_ttest_1_outchannel', 'select',
		    'welch_ttest_1_skip_flags', 'list',
		    'welch_ttest_1_dups', 'radio',
		    'welch_ttest_1_input', 'select',
		    'welch_ttest_1_outheader', 'select',
		    'welch_ttest_1_outchannel', 'select',
		    'welch_ttest_2_skip_flags', 'list',
		    'welch_ttest_2_dups', 'radio',
		    'welch_ttest_2_outheader', 'list',
		    'welch_ttest_2_outchannel', 'select',
		    'student_ttest_2_skip_flags', 'list',
		    'student_ttest_2_dups', 'radio',
		    'student_ttest_2_outheader', 'list',
		    'student_ttest_2_outchannel', 'select',
		    'correlation_coefficient_skip_flags', 'list',
		    'correlation_coefficient_files', 'list',
		    'correlation_coefficient_method', 'radio',
		    'correlation_coefficient_ratio', 'radio',
		    'correlation_coefficient_log', 'radio',
		    'correlation_coefficient_rounding', 'select',
		    'merge_tech_reps_skip_flags', 'list',
		    'merge_tech_reps_dups', 'radio',
		    'merge_tech_weight', 'radio',
		    'merge_tech_reps_method', 'radio',
		    'merge_tech_reps_continuation', 'radio',
		    'merge_all_reps_skip_flags', 'list',
		    'merge_all_reps_dups', 'radio',
		    'merge_all_weight', 'radio',
		    'merge_all_reps_method', 'radio',
		    'merge_all_reps_continuation', 'radio',
		    'wilcox_ttest_1_skip_flags', 'list',
		    'wilcox_ttest_1_dups', 'radio',
		    'wilcox_ttest_1_input', 'select',
		    'wilcox_ttest_1_outheader', 'select',
		    'wilcox_ttest_1_outchannel', 'select',
		    'wilcox_ttest_2_skip_flags', 'list',
		    'wilcox_ttest_2_dups', 'radio',
		    'wilcox_ttest_2_outheader', 'list',
		    'wilcox_ttest_2_outchannel', 'select',
		    'permutation_p_value_skip_flags', 'list',
		    'permutation_p_value_dups', 'radio',
		    'permutation_p_value_outheader', 'select',
		    'permutation_p_value_outchannel', 'select',
		    'anova_skip_flags', 'list',
		    'anova_dups', 'radio',
		    'anova_outheader', 'select',
		    'anova_outchannel', 'select',
		    'boxplot_skip_flags', 'list',
		    'boxplot_labels', 'list',
		    'by_value_strict_skip', 'radio',
		    'by_value_ignore_sign1', 'radio',
		    'by_value_ignore_sign2', 'radio',
		    'boxplot_log', 'radio',
		    'boxplot_pt_skip_flags', 'list',
		    'boxplot_pt_files', 'list',
		    'boxplot_pt_labels', 'list',
		    'boxplot_pt_colour', 'radio',
		    'boxplot_pt_log', 'radio',
		    'boxplot_pt_add_info', 'radio',
		    'flag_change_in', 'list',
		    'flag_change_out', 'select',
		    'flag_change_files', 'list',
		    'marker_flag', 'list',
		    'markers', 'list',
		    'signal_x', 'select',
		    'signal_coloured', 'radio',
		    'signal_layout', 'select',
		    'signal_grids_per_plot', 'select',
		    'histo_add_quintile', 'radio',
		    'histo_take_log', 'radio',
		    'signal_log', 'radio',
		    'rep_signal_log', 'radio',
		    'histo_skip_flags', 'list',
		    'signal_skip_flags', 'list',
		    'rep_signal_skip_flags', 'list',
		    'ratio_signal_skip_flags', 'list',
		    'ratio_signal_files', 'list',
		    'pathway_files', 'list',
		    'compare_lists_files', 'list',
		    'annotate_files', 'list',
		    'annotate_web_file', 'text',
		    'annotate_items_probelynx', 'list',
		    'annotate_items_funcdb', 'list',
		    'output_files', 'select',
		    'pathway_labels', 'list',
		    'pathway_skip_flags', 'list',
		    'pathway_boxplot', 'check_single',
		    'pathway_connectors', 'check_single',
		    'output_skip_empty', 'check_single',
		    'output_skip_flags', 'list',
		    'output_columns', 'list',
		    'output_format', 'list',
		    'output_rounding', 'select',
		    'compare_lists_flags', 'list',
		    'compare_lists_make_new', 'radio',
		    'ma_skip_flags', 'list',
		    'output_sample_lines', 'text',
		    'by_value_make_list', 'text',
		    'by_value_copy_list', 'radio',
		    'by_value_sticky_list', 'check_single',
		    'by_value_max_op', 'select',
		    'by_value_max', 'text',
		    'by_value_max_field', 'select',
		    'by_value_min_field', 'select',
		    'by_value_connector', 'select',
		    'by_value_min_op', 'select',
		    'by_value_min', 'text',
		    'by_value_skip_flags', 'list',
		    'by_value_flag', 'select',
		    'by_value_files', 'list',
		    'by_value_files2', 'select',
		    'by_value_labels', 'select',
		    'by_value_labels2', 'select',
		    'colour_spread_header', 'header',
		    'bg_square_size','text',
		    'bg_grid', 'radio',
		    'bg_skip_flags', 'list',
		    'fg_min_bg_files', 'list',
		    'fg_min_percentile_files', 'list',
		    'bg_loess_files', 'list',
		    'fg_min_bg_skip_flags', 'list',
		    'fg_min_percentile_skip_flags', 'list',
		    'bg_loess_skip_flags', 'list',
		    'bg_orientation', 'radio',
		    'bg_min', 'text',
		    'bg_max', 'text',
		    'bg_border_setting', 'radio',
		    'fg_square_size','text',
		    'fg_grid', 'radio',
		    'fg_skip_flags', 'list',
		    'fg_orientation', 'radio',
		    'fg_min', 'text',
		    'fg_max', 'text',
		    'fg_border_setting', 'radio',
		    'norm_square_size','text',
		    'norm_grid', 'radio',
		    'norm_skip_flags', 'list',
		    'norm_orientation', 'radio',
		    'norm_min', 'text',
		    'norm_max', 'text',
		    'norm_border_setting', 'radio',
		    'ratio_square_size','text',
		    'ratio_grid', 'radio',
		    'ratio_skip_flags', 'list',
		    'ratio_orientation', 'radio',
		    'ratio_min', 'text',
		    'ratio_max', 'text',
		    'ratio_border_setting', 'radio',
		    'ratio_ch1', 'select',
		    'ma_plot_ch1', 'select',
		    'log_trans_ratios', 'radio',
		    'fg_min_bg_operator', 'select',
		    'percentile', 'text',
		    'percentile_action', 'radio',
		    'dup_cutoff', 'text',
		    'dup_flag', 'select',
		    'dup_unit', 'select',
		    'rep_flawed_files', 'list',
		    'rep_flawed_cutoff', 'text',
		    'rep_flawed_unit', 'select',
		    'rep_flawed_percent', 'text',
		    'rep_flawed_flag', 'select',
		    'rep_flawed_extend', 'radio',
		    'rep_flawed_within_slide', 'radio',
		    'rep_flawed_flag_single', 'radio',
		    'rep_flawed_label', 'list',
		    'rep_flawed_skip_flags', 'list',
		    'limma_loess_printtip_files', 'list',
		    'limma_loess_printtip_span', 'text',
		    'limma_loess_printtip_bin', 'text',
		    'limma_loess_printtip_iter', 'text',
		    'limma_loess_printtip_skip_flags', 'list',
		    'limma_loess_global_files', 'list',
		    'limma_loess_global_span', 'text',
		    'limma_loess_global_bin', 'text',
		    'limma_loess_global_iter', 'text',
		    'limma_loess_global_skip_flags', 'list',
		    'vsn_skip_flags', 'list',
		    'vsn_iter', 'text',
		    'vsn_global', 'radio',
		    'vsn_dehuberize', 'radio',
		    'marrayNormmedian_files', 'list',
		    'marrayNormtwoD_files', 'list',
		    'marrayNormprintTipLoess_files', 'list',
		    'marrayNormscalePrintTipMAD_files', 'list',
		    'marrayNormglobalMAD_files', 'list',
		    'marrayNormprintTipMAD_files', 'list',
		    'marrayNormloess_files', 'list',
		    'marrayNormmedian_skip_flags', 'list',
		    'marrayNormtwoD_skip_flags', 'list',
		    'marrayNormprintTipLoess_skip_flags', 'list',
		    'marrayNormscalePrintTipMAD_skip_flags', 'list',
		    'marrayNormglobalMAD_skip_flags', 'list',
		    'marrayNormprintTipMAD_skip_flags', 'list',
		    'marrayNormloess_skip_flags', 'list',
#		    'marrayNorm_skip_flags', 'list',
		    'marrayNormtwoD_span', 'text',
		    'marrayNormprintTipLoess_span', 'text',
		    'marrayNormscalePrintTipMAD_span', 'text',
		    'marrayNormloess_span', 'text',
#		    'marrayNorm_span', 'text',
#		    'marrayNorm_global', 'radio',
		    'marrayNormglobalMAD_global', 'radio',
		    'marrayNormprintTipMAD_global', 'radio',
#		    'marrayNorm_method', 'select',
		    'perm_skip_flags', 'text',
		    'cutoff_files', 'list',
		    'cutoff_skip_flags', 'list',
		    'cutoff_allowed_min', 'text',
		    'cutoff_min', 'text',
		    'cutoff_min_action', 'radio',
		    'cutoff_max', 'text',
		    'cutoff_max_action', 'radio',
		    'merge_dups_target', 'select',
		    'merge_skip_flags', 'list',
		    'dup_skip_flags', 'list',
		    );

# title of variable:
my %var_def_text = (
		    'fg_min_bg_operator', 'Operator: ',
		    'local_int_dep_z_files', 'Files: ',
		    'local_int_dep_z_window', 'Window size: ', 
		    'local_int_dep_z_orient', 'Add orientation: ', 
		    'local_int_dep_z_ch1', 'Numerator: ', 
		    'local_int_dep_z_skip_flags','Skip flags: ' ,
		    'flag_change_in', 'From: ',
		    'flag_change_out', 'To: ',
		    'flag_change_files', 'Files: ',
		    'marker_flag', 'Flag: ',
		    'markers', 'Markers: ',
		    'signal_x', 'x-axis: ',
		    'signal_coloured', 'colour subgrids: ',
		    'signal_layout', 'Graphics layout: ',
		    'signal_grids_per_plot', 'subgrids per plot: ',
		    'student_ttest_1_skip_flags','Skip flags: ' ,
		    'student_ttest_1_input', 'Input: ',
		    'student_ttest_1_outheader', 'Output field: ',
		    'student_ttest_1_outchannel', 'Assign to: ',
		    'student_ttest_1_dups', 'Use duplicate neighbours as replicates? ',
		    'paired_ttest_1_skip_flags','Skip flags: ' ,
		    'paired_ttest_1_outheader', 'Output field: ',
		    'paired_ttest_1_outchannel', 'Assign to: ',
		    'paired_ttest_1_dups', 'Use duplicate neighbours as replicates? ',
		    'welch_ttest_1_skip_flags','Skip flags: ' ,
		    'welch_ttest_1_input', 'Input: ',
		    'welch_ttest_1_outheader', 'Output field: ',
		    'welch_ttest_1_outchannel', 'Assign to: ',
		    'welch_ttest_1_dups', 'Use duplicate neighbours as replicates? ',
		    'welch_ttest_2_skip_flags','Skip flags: ' ,
		    'welch_ttest_2_outheader', 'Output field: ',
		    'welch_ttest_2_outchannel', 'Assign to: ',
		    'welch_ttest_2_dups', 'Use duplicate neighbours as replicates? ',
		    'student_ttest_2_skip_flags','Skip flags: ' ,
		    'student_ttest_2_outheader', 'Output field: ',
		    'student_ttest_2_outchannel', 'Assign to: ',
		    'student_ttest_2_dups', 'Use duplicate neighbours as replicates? ',
		    'correlation_coefficient_skip_flags', 'Skip flags: ',
		    'correlation_coefficient_files', 'Files: ',
		    'correlation_coefficient_method', 'Method: ',
		    'correlation_coefficient_ratio', 'Data: ',
		    'correlation_coefficient_log', 'Take log: ',
		    'correlation_coefficient_rounding', 'Round to: ',
		    'merge_tech_reps_skip_flags','Skip flags: ' ,
		    'merge_tech_reps_dups', 'Use duplicate neighbours as replicates? ',
		    'merge_tech_weight', 'Apply weights from previous merging?',
		    'merge_tech_reps_method', 'Specify method for averaging: ',
		    'merge_tech_reps_continuation', 'Apply following modules to merged file only? ',
		    'merge_all_reps_skip_flags','Skip flags: ' ,
		    'merge_all_reps_dups', 'Use duplicate neighbours as replicates? ',
		    'merge_all_weight', 'Apply weights from previous merging?',
		    'merge_all_reps_method', 'Specify method for averaging: ',
		    'merge_all_reps_continuation', 'Apply following modules to merged file only? ',
		    'wilcox_ttest_1_skip_flags','Skip flags: ' ,
		    'wilcox_ttest_1_input', 'Input:',
		    'wilcox_ttest_1_outheader', 'Output field: ',
		    'wilcox_ttest_1_outchannel', 'Assign to: ',
		    'wilcox_ttest_1_dups', 'Use duplicate neighbours as replicates? ',
		    'wilcox_ttest_2_skip_flags','Skip flags: ' ,
		    'wilcox_ttest_2_outheader', 'Output field: ',
		    'wilcox_ttest_2_outchannel', 'Assign to: ',
		    'wilcox_ttest_2_dups', 'Use duplicate neighbours as replicates? ',
		    'permutation_p_value_skip_flags','Skip flags: ' ,
		    'permutation_p_value_outheader', 'Output field: ',
		    'permutation_p_value_outchannel', 'Assign to: ',
		    'permutation_p_value_dups', 'Use duplicate neighbours as replicates? ',
		    'anova_skip_flags','Skip flags: ' ,
		    'anova_outheader', 'Output field: ',
		    'anova_outchannel', 'Assign to: ',
		    'anova_dups', 'Use duplicate neighbours as replicates? ',
		    'by_value_strict_skip', 'Strict: ',
		    'by_value_ignore_sign1', 'Ignore sign: ',
		    'by_value_ignore_sign2', 'Ignore sign: ',
		    'boxplot_log', 'Take log',
		    'boxplot_labels', 'Labels: ',
		    'boxplot_pt_skip_flags', 'Skip flags: ',
		    'boxplot_pt_files', 'Files: ',
		    'boxplot_pt_labels', 'Labels: ',
		    'boxplot_pt_colour', 'Use colours?',
		    'boxplot_pt_log', 'Take log?',
		    'boxplot_pt_add_info', 'List spot info?',
		    'histo_add_quintile', 'Add quintile lines: ',
		    'histo_take_log', 'Take log: ',
		    'signal_log', 'Take log: ',
		    'rep_signal_log', 'Take log',
		    'boxplot_skip_flags', 'Skip flags: ',
		    'histo_skip_flags', 'Skip flags: ',
		    'signal_skip_flags', 'Skip flags: ',
		    'rep_signal_skip_flags', 'Skip flags: ',
		    'ratio_signal_skip_flags', 'Skip flags: ',
		    'ratio_signal_files', 'Files: ',
		    'compare_lists_files', 'Files: ',
		    'annotate_files', 'Data files: ',
		    'annotate_web_file', 'Annotation file: ',
		    'annotate_items_probelynx', 'ProbeLynx Annotation: ',
		    'annotate_items_funcdb', 'FuncDB Annotation: ',
		    'output_files', 'Output: ',
		    'pathway_files', 'Files: ',
		    'pathway_labels', 'Labels: ',
		    'pathway_skip_flags', 'Skip flags: ',
		    'pathway_boxplot', 'Boxplot: ',
		    'pathway_connectors', 'Connectors: ',
		    'output_skip_empty', 'Skip empty rows',
		    'output_skip_flags', 'Skip flags: ',
		    'output_columns', 'Columns: ',
		    'output_format', 'Format: ',
		    'output_rounding', 'Round to',
		    'compare_lists_flags', 'Use flags: ',
		    'compare_lists_make_new', 'Create new list: ',
		    'ma_skip_flags', 'Skip flags: ',
		    'output_sample_lines', 'Show',
		    'by_value_make_list', 'Make list:',
		    'by_value_copy_list', 'Copy flags:',
		    'by_value_sticky_list', 'Sticky list:',
		    'by_value_max_field', 'Data column: ',
		    'by_value_min_field', 'Data column: ',
		    'by_value_max_op', 'Operator: ',
		    'by_value_max', 'Threshold: ',
		    'by_value_connector', '',
		    'by_value_min_op', '2nd operator: ',
		    'by_value_min', '2nd threshold: ',
		    'by_value_flag', '<BR>Flag: ',
		    'by_value_files', 'Files: ',
		    'by_value_labels', 'Labels: ',
		    'by_value_files2', 'File 2: ',
		    'by_value_labels2', 'Label 2: ',
		    'by_value_skip_flags', 'Skip flags',
		    'colour_spread_header', '<BR>Border Settings: ',
		    'bg_square_size','Spotsize: ',
		    'bg_colour_spread', 'Colour Spread: ',
		    'bg_grid', 'Grid',
		    'bg_skip_flags', 'Ignore flags: ',
		    'fg_min_bg_files', 'Files: ',
		    'fg_min_percentile_files', 'Files: ',
		    'bg_loess_files', 'Files: ',
		    'fg_min_bg_skip_flags', 'Skip flags: ',
		    'fg_min_percentile_skip_flags', 'Skip flags: ',
		    'bg_loess_skip_flags', 'Skip flags: ',
		    'bg_orientation', 'Orientation: ',
		    'bg_min', 'Minimum: ',
		    'bg_max', 'Maximum: ',
		    'bg_border_setting', '',
		    'fg_square_size','Spotsize: ',
		    'fg_colour_spread', 'Colour Spread: ',
		    'fg_grid', 'Grid',
		    'fg_skip_flags', 'Ignore flags: ',
		    'fg_orientation', 'Orientation: ',
		    'fg_min', 'Minimum: ',
		    'fg_max', 'Maximum: ',
		    'fg_border_setting', '',
		    'norm_square_size','Spotsize: ',
		    'norm_colour_spread', 'Colour Spread: ',
		    'norm_grid', 'Grid',
		    'norm_skip_flags', 'Ignore flags: ',
		    'norm_orientation', 'Orientation: ',
		    'norm_min', 'Minimum: ',
		    'norm_max', 'Maximum: ',
		    'norm_border_setting', '',
		    'ratio_square_size','Spotsize: ',
		    'ratio_colour_spread', 'Colour Spread: ',
		    'ratio_grid', 'Grid',
		    'ratio_skip_flags', 'Ignore flags: ',
		    'ratio_orientation', 'Orientation: ',
		    'ratio_ch1', 'Numerator: ',
		    'ratio_min', 'Minimum: ',
		    'ratio_max', 'Maximum: ',
		    'ratio_border_setting', '',
		    'log_trans_ratios', 'Log-transfer: ',
		    'percentile', 'Percentile: ',
		    'percentile_action', 'Action',
		    'dup_cutoff', 'Cutoff: ',
		    'dup_flag', 'Flag: ',
		    'dup_unit', 'Unit: ',
		    'rep_flawed_files', 'Files: ',
		    'rep_flawed_percent', 'Trigger: ',
		    'rep_flawed_extend', 'Extend: ',
		    'rep_flawed_within_slide', 'Use dups: ',
		    'rep_flawed_flag_single', 'Flag single: ',
		    'rep_flawed_cutoff', 'Cutoff: ',
		    'rep_flawed_flag', 'Flag: ',
		    'rep_flawed_unit', 'Unit: ',
		    'rep_flawed_label', 'Channel: ',
		    'rep_flawed_skip_flags', 'Skip flags: ',
		    'limma_loess_printtip_files', 'Files: ',
		    'limma_loess_printtip_span', 'Span: ',
		    'limma_loess_printtip_bin', 'Bin: ',
		    'limma_loess_printtip_iter', 'Iterations: ',
		    'limma_loess_printtip_skip_flags', 'Skip Flags: ',
		    'limma_loess_global_files', 'Files: ',
		    'limma_loess_global_span', 'Span: ',
		    'limma_loess_global_bin', 'Bin: ',
		    'limma_loess_global_iter', 'Iterations: ',
		    'limma_loess_global_skip_flags', 'Skip Flags: ',
		    'vsn_skip_flags', 'Skip Flags: ',
		    'vsn_global', 'Multiple Files: ',
		    'vsn_dehuberize', 'De-Huberize: ',
		    'vsn_iter', 'Iterations: ',
		    'marrayNormmedian_files', 'Files: ',
		    'marrayNormtwoD_files', 'Files: ',		    
		    'marrayNormprintTipLoess_files', 'Files: ',		    
		    'marrayNormscalePrintTipMAD_files', 'Files: ',		    
		    'marrayNormglobalMAD_files', 'Files: ',		    
		    'marrayNormprintTipMAD_files', 'Files: ',		    
		    'marrayNormloess_files', 'Files: ',		    
		    'marrayNormmedian_skip_flags', 'Skip Flags: ',
		    'marrayNormtwoD_skip_flags', 'Skip Flags: ',		    
		    'marrayNormprintTipLoess_skip_flags', 'Skip Flags: ',		    
		    'marrayNormscalePrintTipMAD_skip_flags', 'Skip Flags: ',		    
		    'marrayNormglobalMAD_skip_flags', 'Skip Flags: ',		    
		    'marrayNormprintTipMAD_skip_flags', 'Skip Flags: ',		    
		    'marrayNormloess_skip_flags', 'Skip Flags: ',		    
#		    'marrayNorm_skip_flags', 'Skip Flags: ',
#		    'marrayNorm_global', 'Multiple Files: ',
		    'marrayNormglobalMAD_global', 'Multiple Groups: ',
		    'marrayNormprintTipMAD_global', 'Multiple Groups: ',
#		    'marrayNorm_method', 'Method: ',
		    'marrayNormtwoD_span', 'Span: ',
		    'marrayNormprintTipLoess_span', 'Span: ',
		    'marrayNormscalePrintTipMAD_span', 'Span: ',
		    'marrayNormloess_span', 'Span: ',
#		    'marrayNorm_span', 'Span: ',
		    'perm_skip_flags', 'Skip Flags: ',
		    'ma_plot_ch1', 'Numerator: ',
		    'cutoff_files', 'Files: ',
		    'cutoff_skip_flags', 'Skip Flags: ',
		    'cutoff_allowed_min', 'Allowed minimum: ',
		    'cutoff_min', 'Floor: ',
		    'cutoff_min_action', '',
		    'cutoff_max', 'Ceiling: ',
		    'histo_take_min_val', 'Minimum: ',
		    'bg_loess_span', 'Span: ',
		    'permutation_p_value_iterations', 'Iterations: ',
		    'cutoff_max_action', '',
		    'merge_dups_target', 'Target: ',
		    'merge_skip_flags', 'Skip flags: ',
		    'dup_skip_flags', 'Skip flags: ',
		    );

# default size for text boxes or scrolling lists
my %var_def_size = (
		    'by_value_files', 4,
		    'annotate_files', 4,
		    'annotate_items_probelynx', 10,
		    'annotate_items_funcdb', 10,
		    'annotate_web_file', 30,
		    'local_int_dep_z_files', 3,
		    'local_int_dep_z_window', 5, 
		    'local_int_dep_z_ch1', 2, 
		    'local_int_dep_z_skip_flags', 3,
		    'correlation_coefficient_skip_flags', 3,
		    'correlation_coefficient_files', 3,
		    'merge_tech_reps_skip_flags', 3,
		    'merge_all_reps_skip_flags', 3,
		    'student_ttest_1_skip_flags', 3,
		    'paired_ttest_1_skip_flags', 3,
		    'welch_ttest_1_skip_flags', 3,
		    'wilcox_ttest_1_skip_flags', 3,
		    'welch_ttest_2_skip_flags', 3,
		    'wilcox_ttest_2_skip_flags', 3,
		    'student_ttest_2_skip_flags', 3,
		    'permutation_p_value_skip_flags', 3,
		    'anova_skip_flags', 3,
		    'bg_skip_flags', 3,
		    'fg_min_bg_files', 3,
		    'fg_min_percentile_files', 3,
		    'bg_loess_files', 3,
		    'fg_min_bg_skip_flags', 3,
		    'fg_min_percentile_skip_flags', 3,
		    'bg_loess_skip_flags', 3,
		    'fg_skip_flags', 3,
		    'norm_skip_flags', 3,
		    'ratio_skip_flags', 3,
		    'limma_loess_printtip_files', 4,
		    'limma_loess_printtip_span', 3,
		    'limma_loess_printtip_bin', 3,
		    'limma_loess_printtip_iter', 3,
		    'limma_loess_printtip_skip_flags', 3,
		    'limma_loess_global_files', 4,
		    'limma_loess_global_span', 3,
		    'limma_loess_global_bin', 3,
		    'limma_loess_global_iter', 3,
		    'limma_loess_global_skip_flags', 3,
		    'vsn_iter', 3,
		    'marrayNormtwoD_span', 3,
		    'marrayNormprintTipLoess_span', 3,
		    'marrayNormscalePrintTipMAD_span', 3,
		    'marrayNormloess_span', 3,
#		    'marrayNorm_span', 3,
		    'flag_change_in', 3,
		    'flag_change_files', 3,
		    'marker_flag', 3,
		    'markers', 5,
		    'boxplot_pt_skip_flags', 3,
		    'boxplot_pt_files', 3,
		    'boxplot_pt_labels', 2,
		    'boxplot_skip_flags', 3,
		    'histo_skip_flags', 3,
		    'signal_skip_flags', 3,
		    'rep_signal_skip_flags', 3,
		    'ratio_signal_skip_flags', 3,
		    'ratio_signal_files', 3,
		    'pathway_skip_flags', 3,
		    'output_sample_lines', 5,
		    'by_value_make_list', 12,
		    'by_value_max', 6,
		    'by_value_min', 6,
		    'by_value_skip_flags', 3,
		    'output_skip_flags', 3,
		    'output_columns', 3,
		    'output_format', 2,
		    'compare_lists_flags', 3,
		    'ma_skip_flags', 3,
		    'by_value_labels', 2,
		    'boxplot_labels', 2,
		    'bg_square_size',2,
		    'bg_min', 8,
		    'bg_max', 8,
		    'fg_square_size',2,
		    'fg_min', 8,
		    'fg_max', 8,
		    'norm_square_size',2,
		    'norm_min', 8,
		    'norm_max', 8,
		    'ratio_square_size',2,
		    'ratio_min', 8,
		    'ratio_max', 8,
		    'percentile',4,
		    'dup_cutoff',4,
		    'dup_flag',1,
		    'dup_unit',1,
		    'rep_flawed_cutoff',4,
		    'rep_flawed_flag',1,
		    'rep_flawed_unit',1,
		    'rep_flawed_label', 3,
		    'rep_flawed_skip_flags', 3,
		    'rep_flawed_files',3,
		    'rep_flawed_percent',3,
		    'vsn_skip_flags', 3,
		    'marrayNormmedian_files', 3,
		    'marrayNormtwoD_files', 3,
		    'marrayNormprintTipLoess_files', 3,
		    'marrayNormscalePrintTipMAD_files', 3,
		    'marrayNormglobalMAD_files', 3,
		    'marrayNormprintTipMAD_files', 3,
		    'marrayNormloess_files', 3,
		    'marrayNormmedian_skip_flags', 3,
		    'marrayNormtwoD_skip_flags', 3,
		    'marrayNormprintTipLoess_skip_flags', 3,
		    'marrayNormscalePrintTipMAD_skip_flags', 3,
		    'marrayNormglobalMAD_skip_flags', 3,
		    'marrayNormprintTipMAD_skip_flags', 3,
		    'marrayNormloess_skip_flags', 3,
#		    'marrayNorm_skip_flags', 3,
		    'perm_skip_flags', 2,
		    'cutoff_files', 3,
		    'cutoff_skip_flags', 3,
		    'cutoff_allowed_min', 8,
		    'cutoff_min', 8,
		    'cutoff_max', 8,
		    'histo_take_min_val', 3,
		    'bg_loess_span', 3,
		    'permutation_p_value_iterations', 3,
		    'merge_skip_flags', 3,
		    'dup_skip_flags', 3,
		    );

# values for multiple choices
my %var_def_values = (
		      'fg_min_bg_operator', [ '-', '/' ],
		      'local_int_dep_z_files', '_X__all_X__@work_files',
		      'local_int_dep_z_ch1', [ 'CH1', 'CH2' ], 
		      'local_int_dep_z_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'local_int_dep_z_orient', [ 'yes', 'no' ], 
		      'flag_change_in', [ 'all lists', 'all lists but last', @all_flags ],
		      'flag_change_out', [ 'none', @all_flags ],
		      'flag_change_files', '_X__all_X__@work_files',
		      'marker_flag', [ @all_flags ],
		      'markers', [ 'all', 'B/W','no label','GFP Marker', 'AC005609', 'U17997', 'Buffer', 'DNA', '3XSSC Buffer Control', 'blank', 'N/A', 'EMPTY', 'NORF', 'BLANK', 'undefined', '3X SSC', 'GFP', 'Empty', 'Empty Spot', 'Arabidopsis Negative Control', 'ABBINDING', 'Actin', 'GAPDH', 'HPRT', 'Cot1', 'ENV', 'MSG', 'POLYA', 'Rubisco', 'TAX', 'abbinding', 'Gapdh', 'Hprt', 'PolyA', 'Msg', 'Tax', ':Data not found:::::Arabidopsis', ':Data not found:::::683068', ':Data not found:::::3XSSC', ':Data not found:::::683190', ':Data not found:::::No clone in this well', '"NONE"', 'Randomized negative control', '3XSSC Buffer Control', 'undefined', 'Hypothetical protein' ],
		      'signal_x', [ 'CH1', 'CH2' ],
		      'signal_coloured', [ 'yes', 'no' ],
		      'signal_layout', [ 'auto', '1x1', '2x1', '2x2', '2x3', '2x4', '3x1', '3x2', '3x3', '3x4', '3x5', '4x2', '4x3', '4x4', '5x3', '5x4', '6x3', '6x4' ],
		      'signal_grids_per_plot', [ 'all', 'MROW', 'MCOL', (1..8) ],
		      'student_ttest_1_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'student_ttest_1_input', [ 'ratios', 'intensities' ],
		      'student_ttest_1_outheader', [ 'p-value', 'p-student', 'both' ],
		      'student_ttest_1_outchannel', [ 'file', 'channel' ],
		      'student_ttest_1_dups', [ 'yes', 'no' ],
		      'paired_ttest_1_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'paired_ttest_1_outheader', [ 'p-value', 'p-paired', 'both' ],
		      'paired_ttest_1_outchannel', [ 'file', 'channel' ],
		      'paired_ttest_1_dups', [ 'yes', 'no' ],
		      'welch_ttest_1_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'welch_ttest_1_input', [ 'ratios', 'intensities' ],
		      'welch_ttest_1_outheader', [ 'p-value', 'p-welch', 'both' ],
		      'welch_ttest_1_outchannel', [ 'file', 'channel' ],
		      'welch_ttest_1_dups', [ 'yes', 'no' ],
		      'welch_ttest_2_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'welch_ttest_2_outheader', [ 'p-value', 'p-welch', 'label p-value', 'label p-welch' ],
		      'welch_ttest_2_outchannel', [ 'file', 'channel' ],
		      'welch_ttest_2_dups', [ 'yes', 'no' ],
		      'student_ttest_2_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'student_ttest_2_outheader', [ 'p-value', 'p-student', 'label p-value', 'label p-student' ],
		      'student_ttest_2_outchannel', [ 'file', 'channel' ],
		      'student_ttest_2_dups', [ 'yes', 'no' ],
		      'correlation_coefficient_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'correlation_coefficient_files', '_X__between technical replicates_X__within treatment group_X__all_X__@work_files',
		      'correlation_coefficient_method', [ 'Spearman', 'Pearson', 'both' ],
		      'correlation_coefficient_ratio', [ 'ratios', 'intensities', 'both'  ],
		      'correlation_coefficient_log', [ 'yes', 'no' ],
		      'correlation_coefficient_rounding', [ 14,13,12,11,10,9,8,7,6,5,4,3,2,1 ],
		      'merge_tech_reps_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'merge_tech_reps_dups', [ 'yes', 'no' ],
		      'merge_tech_weight', [ 'yes', 'no' ],
		      'merge_tech_reps_method', [ 'mean', 'median', 'weighted mean' ],
		      'merge_tech_reps_continuation', [ 'yes', 'no' ],
		      'merge_all_reps_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'merge_all_reps_dups', [ 'yes', 'no' ],
		      'merge_all_weight', [ 'yes', 'no' ],
		      'merge_all_reps_method', [ 'mean', 'median', 'weighted mean' ],
		      'merge_all_reps_continuation', [ 'yes', 'no' ],
		      'wilcox_ttest_1_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'wilcox_ttest_1_input', [ 'ratios', 'intensities' ],
		      'wilcox_ttest_1_outheader', [ 'p-value', 'p-wilcox', 'both' ],
		      'wilcox_ttest_1_outchannel', [ 'file', 'channel' ],
		      'wilcox_ttest_1_dups', [ 'yes', 'no' ],
		      'wilcox_ttest_2_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'wilcox_ttest_2_outheader', [ 'p-value', 'p-wilcox', 'label p-value', 'label p-wilcox' ],
		      'wilcox_ttest_2_outchannel', [ 'file', 'channel' ],
		      'wilcox_ttest_2_dups', [ 'yes', 'no' ],
		      'permutation_p_value_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'permutation_p_value_outheader', [ 'p-value', 'p-permute', 'both' ],
		      'permutation_p_value_outchannel', [ 'file', 'channel' ],
		      'permutation_p_value_dups', [ 'yes', 'no' ],
		      'anova_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'anova_outheader', [ 'p-value', 'p-anova', 'both' ],
		      'anova_outchannel', [ 'file', 'channel' ],
		      'anova_dups', [ 'yes', 'no' ],
		      'boxplot_pt_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'boxplot_pt_files', '_X__all_X__@work_files',
		      'boxplot_pt_labels', [ 'auto', 'ratios', 'CH1', 'CH2', 'separate', 'together', 'combined' ],
		      'boxplot_pt_colour', [ 'yes', 'no' ],
		      'boxplot_pt_log', [ 'yes', 'no' ],
		      'boxplot_pt_add_info', [ 'yes', 'no' ],
		      'by_value_strict_skip', [ 'yes', 'no' ],
		      'by_value_ignore_sign1', [ 'yes', 'no' ],
		      'by_value_ignore_sign2', [ 'yes', 'no' ],
		      'boxplot_log', [ 'yes', 'no' ],
		      'histo_add_quintile', [ 'yes', 'no' ],
		      'histo_take_log', [ 'yes', 'no' ],
		      'signal_log', [ 'yes', 'no' ],
		      'rep_signal_log', [ 'yes', 'no' ],
		      'boxplot_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'histo_skip_flags', [ 'none', 'all', 'all but lists', 'unflagged', @all_flags ],
		      'signal_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'rep_signal_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'ratio_signal_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'ratio_signal_files', '_X__auto_X__all_X__merged between groups_X__all between groups_X__merged within groups_X__all within groups_X__between technical replicates_X__@work_files',
		      'pathway_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'compare_lists_files', '_X__separately_X__all_X__@work_files',
		      'annotate_files', '_X__all_X__@work_files',
		      'annotate_items_probelynx', [ 'all', 'TC', 'TC hits', 'number of transcript hits','gene hits','hit identity', 'hit coverage', 'hit gaps', 'hit flag','hit start','hit end', 'chr name','chr strand','exon coverage','status','TC description', 'SP', 'PIR', 'GP', 'emb', 'gb', 'dbj', 'protein', 'pdb', 'transcript','gene','gene symbol','gene description','EMBL accession','RefSeq accession','SPTREMBL accession','SWISSPROT gene name','SWISSPROT function','SWISSPROT protein name','go term','go description','go ec', 'interpro interpro id','interpro description','proteinfamily family description','feature prediction method','feature start','feature end','domain prediction method','domain accession number','domain start','domain end','domain description', 'putative ortholog TOG', 'putative ortholog organism', 'putative ortholog hit identifier', 'putative ortholog description', 'putative ortholog Ensembl gene', 'putative ortholog Ensembl transcript', 'putative ortholog Ensembl cover', 'putative ortholog Ensembl ident', 'putative ortholog Ensembl gaps', 'putative ortholog Ensembl status'],
		      'annotate_items_funcdb', [ 'all', 'EMBL accession', 'Ensembl accession','gene description', 'HUGO accession', 'LocusLink accession', 'OMIM accession', 'PDB accession', 'RefSeq accession', 'SWISSPROT accession', 'SWISSPROT gene name','SWISSPROT function','SWISSPROT protein name','SPTREMBL accession', 'interpro interpro id','interpro description','proteinfamily family description','feature prediction method','feature start','feature end','domain prediction method','domain accession number','domain start','domain end','domain description', 'BIND', 'BIND interaction partners', 'BioCarta', 'DIP', 'DIP interaction partners', 'TransFac Factor Num', 'TransFac Gene Num', 'TransFac Factor', 'TransFac Gene', 'TransPath', ],
		      'output_files', [ 'separately', 'combined' ],
		      'pathway_files', '_X__separately_X__all_X__@work_files',
		      'pathway_labels', [ 'auto', 'ratios', 'CH1', 'CH2' ],
		      'by_value_max_op', [ 'auto', '>', '>=', '=', '<=', '<' ],
		      'by_value_max_field', [ 'auto', @output ],
		      'by_value_min_field', [ 'auto', @output ],
		      'limma_loess_printtip_files', '_X__all_X__@work_files', 
		      'limma_loess_printtip_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'limma_loess_global_files', '_X__all_X__@work_files', 
		      'limma_loess_global_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'vsn_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'marrayNormmedian_files', '_X__all_X__@work_files',
		      'marrayNormtwoD_files', '_X__all_X__@work_files', 
		      'marrayNormprintTipLoess_files', '_X__all_X__@work_files', 
		      'marrayNormscalePrintTipMAD_files', '_X__all_X__@work_files', 
		      'marrayNormglobalMAD_files', '_X__all_X__@work_files', 
		      'marrayNormprintTipMAD_files', '_X__all_X__@work_files', 
		      'marrayNormloess_files', '_X__all_X__@work_files', 
		      'marrayNormmedian_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'marrayNormtwoD_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'marrayNormprintTipLoess_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'marrayNormscalePrintTipMAD_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'marrayNormglobalMAD_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'marrayNormprintTipMAD_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'marrayNormloess_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
#		      'marrayNorm_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'by_value_min_op', [ 'auto', '<', '>=', '=', '<=', '>' ],
		      'by_value_connector', [ '--', 'AND', 'OR', 'BUTNOT' ],
		      'by_value_flag', [ 'auto', @all_flags ],
		      'by_value_files', '_X__separately_X__all_X__@work_files',
		      'boxplot_labels', [ 'auto', 'ratios', 'CH1', 'CH2', 'combined' ],
		      'by_value_labels', [ 'auto', 'ratios', 'log-ratios', 'CH1', 'CH2', 'CH1 and CH2', 'CH1 or CH2' ],
		      'by_value_files2', '_X__separately_X__@work_files',
		      'by_value_labels2', [ 'auto', 'ratios', 'log-ratios', 'CH1', 'CH2', 'CH1 and CH2', 'CH1 or CH2' ],
		      'by_value_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'by_value_copy_list', [ 'yes', 'no' ],
		      'output_skip_flags', [ 'none', 'no flag', 'all', 'all but lists', @all_flags ],
		      'output_columns', [ 'all', 'i', 'MROW', 'MCOL', 'ROW', 'COL', @output, 'Description' ],
		      'output_format', [ 'default', 'TIGR MEV', 'BASE' ],
		      'output_rounding', [ 14,13,12,11,10,9,8,7,6,5,4,3,2,1,0 ],
		      'compare_lists_flags', [ 'none', 'all', 'all lists', @all_flags ],
		      'compare_lists_make_new', [ 'yes', 'no' ],
		      'ma_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'bg_colour_spread', [ 'all', 'conc' ],
		      'bg_grid', [ 'on', 'off' ],
		      'bg_skip_flags',  [ 'none', 'all', 'all but lists', @all_flags ],
		      'fg_min_bg_files',  '_X__all_X__@work_files', 
		      'fg_min_percentile_files',  '_X__all_X__@work_files', 
		      'bg_loess_files',  '_X__all_X__@work_files', 
		      'fg_min_bg_skip_flags',  [ 'none', 'all', 'all but lists', 'all but lists and floor', @all_flags ],
		      'fg_min_percentile_skip_flags',  [ 'none', 'all', 'all but lists', 'all but lists and floor', @all_flags ],
		      'bg_loess_skip_flags',  [ 'none', 'all', 'all but lists', 'all but lists and floor', @all_flags ],
		      'bg_orientation', [ 'portrait', 'landscape' ],
		      'bg_border_setting', [ 'abs', 'rel' ],
		      'fg_colour_spread', [ 'all', 'conc' ],
		      'fg_grid', [ 'on', 'off' ],
		      'fg_skip_flags',  [ 'none', 'all', 'all but lists', @all_flags ],
		      'fg_orientation', [ 'portrait', 'landscape' ],
		      'fg_border_setting', [ 'abs', 'rel' ],
		      'norm_colour_spread', [ 'all', 'conc' ],
		      'norm_grid', [ 'on', 'off' ],
		      'norm_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'norm_orientation', [ 'portrait', 'landscape' ],
		      'norm_border_setting', [ 'abs', 'rel' ],
		      'ratio_colour_spread', [ 'all', 'conc' ],
		      'ratio_grid', [ 'on', 'off' ],
		      'ratio_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'ratio_orientation', [ 'portrait', 'landscape' ],
		      'ratio_ch1', [ 'CH1', 'CH2' ],
		      'ratio_border_setting', [ 'abs', 'rel' ],
		      'log_trans_ratios', [ 'yes', 'no' ],
		      'percentile_action', [ 'flag', 'adjust', 'both' ],
		      'vsn_global', [ 'global', 'local' ],
		      'vsn_dehuberize', [ 'yes', 'no' ],
#		      'marrayNorm_global', [ 'yes', 'no' ],
		      'marrayNormglobalMAD_global', [ 'yes', 'no' ],
		      'marrayNormprintTipMAD_global', [ 'yes', 'no' ],
#		      'marrayNorm_method', [ 'none', 'median', 'loess', 'twoD', 'printTipLoess', 'scalePrintTipMAD', 'globalMAD', 'printTipMAD' ],
		      'ma_plot_ch1', [ 'CH1', 'CH2' ],
		      'cutoff_files', '_X__separately_X__all_X__@work_files',
		      'cutoff_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'cutoff_min_action', [ 'individual adjust', 'global shift', 'flag', 'do nothing' ],
		      'cutoff_max_action', [ 'individual adjust', 'flag', 'do nothing' ],
		      'merge_dups_target', [ 'whole slide', 'direct neighbours' ],
		      'merge_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'dup_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'dup_flag', [ @all_flags ],
		      'dup_unit', [ 'StdDev abs diff', 'abs diff', 'StdDev fold diff', 'fold diff' ],
		      'rep_flawed_label', [ 'all', 'ratio', 'CH1', 'CH2' ],
		      'rep_flawed_skip_flags', [ 'none', 'all', 'all but lists', @all_flags ],
		      'rep_flawed_flag', [ @all_flags, 'list' ],
		      'rep_flawed_unit', [ 'StdDev abs diff', 'abs diff', 'StdDev rel diff', 'rel diff' ],
		      'rep_flawed_extend', [ 'all', 'flawed' ],
		      'rep_flawed_within_slide', [ 'yes', 'no' ],
		      'rep_flawed_flag_single', [ 'yes', 'no' ],
		      'rep_flawed_files', '_X__between technical replicates_X__within treatment group_X__all_X__@work_files',
		      );

# labels for values
my %var_def_labels = (
		      'local_int_dep_z_files', '_X__%work_files_base',
		      'flag_change_files', '_X__%work_files_base',
		      'cutoff_files', '_X__%work_files_base',
		      'by_value_files', '_X__%work_files_base',
		      'by_value_files2', '_X__%work_files_base',
		      'compare_lists_files', '_X__%work_files_base',
		      'correlation_coefficient_files', '_X__%work_files_base',
		      'rep_flawed_files', '_X__%work_files_base',
		      'annotate_files', '_X__%work_files_base',
#		      'output_files', '_X__%work_files_base',
		      'pathway_files', '_X__%work_files_base',
		      'limma_loess_printtip_files', '_X__%work_files_base', 
		      'limma_loess_global_files', '_X__%work_files_base', 
		      'marrayNormmedian_files', '_X__%work_files_base',
		      'marrayNormtwoD_files', '_X__%work_files_base', 
		      'marrayNormprintTipLoess_files', '_X__%work_files_base', 
		      'marrayNormscalePrintTipMAD_files', '_X__%work_files_base', 
		      'marrayNormglobalMAD_files', '_X__%work_files_base', 
		      'marrayNormprintTipMAD_files', '_X__%work_files_base', 
		      'marrayNormloess_files', '_X__%work_files_base', 
		      'boxplot_pt_files', '_X__%work_files_base',
		      'fg_min_bg_files', '_X__%work_files_base',
		      'fg_min_percentile_files', '_X__%work_files_base',
		      'bg_loess_files', '_X__%work_files_base',
		      'ratio_signal_files', '_X__auto_X__all_X__merged between groups_X__all between groups_X__merged within groups_X__between technical replicates_X__all within groups_X__%work_files_base',
		      );

# default value
my %var_def_default = (
		       'local_int_dep_z_files', 'all',
		       'local_int_dep_z_window', 500, 
		       'local_int_dep_z_orient', 'yes', 
		       'fg_min_bg_operator', '-',
		       'local_int_dep_z_ch1', 'CH2', 
		       'local_int_dep_z_skip_flags', 'all but lists',
		       'flag_change_in', 'all lists but last',
		       'flag_change_out', 'user1',
		       'flag_change_files', 'all',
		       'marker_flag', 'markers',
		       'limma_loess_printtip_files', 'all',
		       'limma_loess_printtip_span', '',
		       'limma_loess_printtip_bin', '',
		       'limma_loess_printtip_iter', '',
		       'limma_loess_printtip_skip_flags', 'all but lists',
		       'limma_loess_global_files', 'all',
		       'limma_loess_global_span', '',
		       'limma_loess_global_bin', '',
		       'limma_loess_global_iter', '',
		       'limma_loess_global_skip_flags', 'all but lists',
		       'vsn_iter', 10,
		       'marrayNormtwoD_span', '',
		       'marrayNormprintTipLoess_span', '',
		       'marrayNormscalePrintTipMAD_span', '',
		       'marrayNormloess_span', '',
#		       'marrayNorm_span', '',
		       'markers', [ 'all' ],
		       'signal_x', 'CH1',
		       'signal_coloured', 'no',
		       'signal_layout', 'auto',
		       'signal_grids_per_plot', 'all',
		       'student_ttest_1_skip_flags', 'all but lists',
		       'student_ttest_1_input', 'ratios',
		       'student_ttest_1_outheader', 'p-student',
		       'student_ttest_1_outchannel', 'file',
		       'student_ttest_1_dups', 'no',
		       'paired_ttest_1_skip_flags', 'all but lists',
		       'paired_ttest_1_input', 'ratios',
		       'paired_ttest_1_outheader', 'p-paired',
		       'paired_ttest_1_outchannel', 'file',
		       'paired_ttest_1_dups', 'no',
		       'welch_ttest_1_skip_flags', 'all but lists',
		       'welch_ttest_1_input', 'intensities',
		       'welch_ttest_1_outheader', 'p-welch',
		       'welch_ttest_1_outchannel', 'file',
		       'welch_ttest_1_dups', 'no',
		       'welch_ttest_2_skip_flags', 'all but lists',
		       'welch_ttest_2_outheader', [ 'p-value', 'label p-welch' ],
		       'welch_ttest_2_outchannel', 'file',
		       'welch_ttest_2_dups', 'no',
		       'student_ttest_2_skip_flags', 'all but lists',
		       'student_ttest_2_outheader', [ 'p-value', 'label p-student' ],
		       'student_ttest_2_outchannel', 'file',
		       'student_ttest_2_dups', 'no',
		       'correlation_coefficient_skip_flags', 'all but lists',
		       'correlation_coefficient_files', 'between technical replicates',
		       'correlation_coefficient_method', 'both',
		       'correlation_coefficient_ratio', 'both',
		       'correlation_coefficient_log', 'yes',
		       'correlation_coefficient_rounding', 3,
		       'merge_tech_reps_skip_flags', 'all but lists',
		       'merge_tech_reps_dups', 'yes',
		       'merge_tech_weight', 'yes',
		       'merge_tech_reps_method', 'weighted mean',
		       'merge_tech_reps_continuation', 'yes',
		       'merge_all_reps_skip_flags', 'all but lists',
		       'merge_all_reps_dups', 'no',
		       'merge_all_weight', 'yes',
		       'merge_all_reps_method', 'weighted mean',
		       'merge_all_reps_continuation', 'no',
		       'wilcox_ttest_1_skip_flags', 'all but lists',
		       'wilcox_ttest_1_input', 'ratios',
		       'wilcox_ttest_1_outheader', 'p-wilcox',
		       'wilcox_ttest_1_outchannel', 'file',
		       'wilcox_ttest_1_dups', 'no',
		       'wilcox_ttest_2_skip_flags', 'all but lists',
		       'wilcox_ttest_2_outheader', [ 'p-value', 'label p-wilcox' ],
		       'wilcox_ttest_2_outchannel', 'file',
		       'wilcox_ttest_2_dups', 'no',
		       'permutation_p_value_skip_flags', 'all but lists',
		       'permutation_p_value_outheader', 'p-permute',
		       'permutation_p_value_outchannel', 'file',
		       'permutation_p_value_dups', 'no',
		       'anova_skip_flags', 'all but lists',
		       'anova_outheader', 'p-anova',
		       'anova_outchannel', 'file',
		       'anova_dups', 'no',
		       'boxplot_pt_skip_flags', 'all but lists',
		       'boxplot_pt_files', 'all',
		       'boxplot_pt_labels', 'ratios',
		       'boxplot_pt_colour', 'yes',
		       'boxplot_pt_log', 'yes',
		       'boxplot_pt_add_info', 'no',
		       'by_value_strict_skip', 'yes',
		       'by_value_ignore_sign1', 'yes',
		       'by_value_ignore_sign2', 'yes',
		       'by_value_sticky_list', 0,
		       'boxplot_log', 'yes',
		       'histo_add_quintile', 'yes',
		       'histo_take_log', 'yes',
		       'signal_log', 'yes',
		       'rep_signal_log', 'yes',
		       'boxplot_skip_flags', [ 'all but lists' ],
		       'histo_skip_flags', [ 'none' ],
		       'signal_skip_flags', [ 'none' ],
		       'rep_signal_skip_flags', [ 'none' ],
		       'ratio_signal_skip_flags', [ 'none' ],
#		       'ratio_signal_files', 'merged between groups',
		       'ratio_signal_files', 'auto',
		       'compare_lists_files', 'all',
		       'annotate_files', 'all',
		       'annotate_items_probelynx', 'all',
		       'annotate_items_funcdb', 'all',
		       'output_files', 'separately',
		       'pathway_files', 'all',
		       'pathway_labels', 'auto',
		       'pathway_skip_flags', [ 'all but lists' ],
		       'pathway_boxplot', 0,
		       'pathway_connectors', 1,
		       'output_sample_lines', 0,
		       'output_skip_empty', 0,
		       'by_value_max_op', 'auto', 
		       'by_value_max_field', 'auto',
		       'by_value_min_field', 'auto',
		       'by_value_make_list', 'auto',
		       'by_value_copy_list', 'yes',
		       'by_value_max', 'auto',
		       'by_value_min_op', 'auto',
		       'by_value_min', 'auto',
		       'by_value_connector', '--',
		       'by_value_flag', 'auto',
#		       'by_value_files', '_X__@work_files',
		       'by_value_files', 'all',
		       'by_value_labels', 'auto',
		       'boxplot_labels', 'ratios',
		       'by_value_files2', 'separately',
		       'by_value_labels2', 'auto',
		       'by_value_skip_flags', [ 'all but lists' ],
		       'output_skip_flags', [ 'none' ],
		       'output_columns', [ 'all' ],
		       'output_format', [ 'default' ],
		       'output_rounding', 5,
		       'compare_lists_flags', [ 'all lists' ],
		       'compare_lists_make_new', 'yes',
		       'ma_skip_flags', [ 'none' ],
		       'bg_square_size', 2,
		       'bg_colour_spread', 'all',
		       'bg_grid', 'on',
		       'bg_skip_flags', [ 'none' ],
		       'fg_min_bg_files', 'all',
		       'fg_min_percentile_files', 'all',
		       'bg_loess_files', 'all',
		       'fg_min_bg_skip_flags', [ 'all but lists and floor' ],
		       'fg_min_percentile_skip_flags', [ 'all but lists and floor' ],
		       'bg_loess_skip_flags', [ 'all but lists and floor' ],
		       'bg_orientation', 'landscape',
		       'bg_min', '2',
		       'bg_max', '98',
		       'bg_border_setting', 'rel',
		       'fg_square_size','2',
		       'fg_colour_spread', 'all',
		       'fg_grid', 'on',
		       'fg_skip_flags', [ 'none' ],
		       'fg_orientation', 'landscape',
		       'fg_min', '',
		       'fg_max', '',
		       'fg_border_setting', 'rel',
		       'norm_square_size','2',
		       'norm_colour_spread', 'all',
		       'norm_grid', 'on',
		       'norm_skip_flags', [ 'none' ],
		       'norm_orientation', 'landscape',
		       'norm_min', '',
		       'norm_max', '',
		       'norm_border_setting', 'rel',
		       'ratio_square_size','2',
		       'ratio_colour_spread', 'conc',
		       'ratio_grid', 'off',
		       'ratio_skip_flags', [ 'all but lists' ],
		       'ratio_orientation', 'landscape',
		       'ratio_ch1', 'CH2',
		       'ratio_min', '95',
		       'ratio_max', '100',
		       'ratio_border_setting', 'rel',
		       'log_trans_ratios', 'yes',
		       'percentile_action', 'both',
		       'percentile', 0.1,
		       'dup_cutoff', 2,
		       'dup_flag', 'dup-flaw',
		       'dup_unit', 'StdDev abs diff',
		       'rep_flawed_cutoff', 2,
		       'rep_flawed_flag', 'dup-flaw',
		       'rep_flawed_unit', 'StdDev rel diff',
		       'rep_flawed_label', 'ratio',
		       'rep_flawed_skip_flags', 'all but lists',
		       'rep_flawed_files', 'between technical replicates',
		       'rep_flawed_percent', 50,
		       'rep_flawed_extend', 'flawed',
		       'rep_flawed_within_slide', 'yes',
		       'rep_flawed_flag_single', 'yes',
		       'vsn_skip_flags', 'all but lists',
		       'marrayNormmedian_files', 'all',
		       'marrayNormtwoD_files', 'all',
		       'marrayNormprintTipLoess_files', 'all',
		       'marrayNormscalePrintTipMAD_files', 'all',
		       'marrayNormglobalMAD_files', 'all',
		       'marrayNormprintTipMAD_files', 'all',
		       'marrayNormloess_files', 'all',
		       'marrayNormmedian_skip_flags', 'all but lists',
		       'marrayNormtwoD_skip_flags', 'all but lists',
		       'marrayNormprintTipLoess_skip_flags', 'all but lists',
		       'marrayNormscalePrintTipMAD_skip_flags', 'all but lists',
		       'marrayNormglobalMAD_skip_flags', 'all but lists',
		       'marrayNormprintTipMAD_skip_flags', 'all but lists',
		       'marrayNormloess_skip_flags', 'all but lists',
#		       'marrayNorm_skip_flags', 'all but lists',
		       'perm_skip_flags', 0,
		       'vsn_global', 'global',
		       'vsn_dehuberize', 'no',
#		       'marrayNorm_global', 'local',
		       'marrayNormglobalMAD_global', 'yes',
		       'marrayNormprintTipMAD_global', 'yes',
#		       'marrayNorm_method', 'printTipLoess',
		       'ma_plot_ch1', 'CH2',
		       'cutoff_files', 'separately',
		       'cutoff_skip_flags', 'none',
		       'cutoff_allowed_min', -50,
		       'cutoff_min', 1,
		       'cutoff_min_action', 'global shift',
		       'cutoff_max', 65534,
		       'histo_take_min_val', '',
		       'bg_loess_span', 0.01,
		       'permutation_p_value_iterations', 1000,
		       'cutoff_max_action', 'flag',
		       'merge_dups_target', 'whole slide',
		       'merge_skip_flags', 'all but lists',
		       'dup_skip_flags', 'all but lists',
		       );


# text to be added after variable
my %var_def_add = (
		       'flag_change_in', ' (select one or more flags to be changed)',
		       'flag_change_out', ' (select the new flag to be used instead)',
		       'boxplot_pt_add_info', '<BR>(prints a table with number of found and missing spots per subgrid)',
		   'marrayNormglobalMAD_global', '<BR>Explanation from marrayNorm: If \'yes\', the MAD of each group is divided by the geometric
          mean of the MADs across groups (cf. Yang et al. (2002)). This
          allows observations to retain their original units.',
		   'marrayNormprintTipMAD_global', '<BR>Explanation from marrayNorm: If \'yes\', the MAD of each group is divided by the geometric
          mean of the MADs across groups (cf. Yang et al. (2002)). This
          allows observations to retain their original units.',
		   'fg_min_bg_operator', 'Choose \'/\' to use division instead of subtraction.',
		   'annotate_web_file', '<BR>(specify a URL to a file on the web, such as the ProbeLynx output)',
		   'local_int_dep_z_orient', '<BR>(when \'yes\' adds \'-\' to score if ratio is lower than median)', 
		   'compare_lists_make_new', '<BR>(from intersection of <B>all</B> used flags)',
		   'histo_take_min_val', '<BR>(enter a number to shift data set so that lowest value lies there)',
		   'bg_loess_span', '(percentage of spots to be used for smoothing, 0&lt;x&lt;=1)',
		   'by_value_make_list', '<BR>(alpha-numeric only, no spaces, use <B>auto</B> if applied multiple times)',
		   'by_value_copy_list', '<BR>(applies flag to all other files if activated)',
		   'by_value_sticky_list', '(carry on list to next steps)',
		   'local_int_dep_z_ch1', '<BR>(for calculating ratio; denominator set automatically)', 
		   'signal_x', '<BR>(y-axis set automatically)',
		   'signal_coloured', '<BR>(set to \'yes\' to plot each subgrid in one of 8 different colours)',
		   'signal_layout', '<BR>(arrangement of plots per image, i.e. rows x colums, necessary when multiple plots are drawn)',
		   'signal_grids_per_plot', '<BR>(specify how many subgrids should be combined into one plot; MROW or MCOL will be replaced with the number of meta rows or meta columns, respectively)',
		   'by_value_strict_skip', '<BR>(for multiple files: if set to \'yes\' skip entry unless all files comply)',
		   'by_value_ignore_sign1', '<BR>(work only on absolute value if set to \'yes\')',
		   'by_value_ignore_sign2', '<BR>(work only on absolute value if set to \'yes\')',
		   'boxplot_log', '(unless log\'ed already)',
		   'merge_tech_reps_dups', '<BR>(automatically skips single spots if \'yes\')',
		   'merge_tech_weight', '<BR>(only for averaging by mean)',
		   'merge_tech_reps_method', '<BR>(weighted mean takes spread of previously merged replicates into account but defaults to median if no weights are available)',
#		   'merge_tech_reps_continuation', '<BR>(excludes all files used for the merging from most of the subsequent functions, except data retrieval if set to \'yes\')',
		   'merge_tech_reps_continuation', '<BR>(excludes all individual files used for the merging from subsequent functions if set to \'yes\')',
		   'merge_all_reps_dups', '<BR>(automatically skips single spots if \'yes\')',
		   'merge_all_weight', '<BR>(only for averaging by mean)',
		   'merge_all_reps_method', '<BR>(weighted mean takes spread of previously merged replicates into account but defaults to median if no weights are available)',
		   'merge_all_reps_continuation', '<BR>(excludes all individual files used for the merging from subsequent functions if set to \'yes\')',
		   'student_ttest_1_dups', '<BR>(automatically skips single spots if \'yes\')',
		   'paired_ttest_1_dups', '<BR>(automatically skips single spots if \'yes\')',
		   'welch_ttest_1_dups', '<BR>(automatically skips single spots if \'yes\')',
		   'wilcox_ttest_1_dups', '<BR>(automatically skips single spots if \'yes\')',
		   'permutation_p_value_dups', '<BR>(automatically skips single spots if \'yes\')',
		   'welch_ttest_2_dups', '<BR>(automatically skips single spots if \'yes\')',
		   'wilcox_ttest_2_dups', '<BR>(automatically skips single spots if \'yes\')',
		   'student_ttest_2_dups', '<BR>(automatically skips single spots if \'yes\')',
		   'anova_dups', '<BR>(automatically skips single spots if \'yes\')',
#		   'histo_add_quintile', '(unless),
		   'histo_take_log', '(unless log\'ed already)',
		   'signal_log', '(unless log\'ed already)',
		   'rep_signal_log', '(unless log\'ed already)',
		   'output_sample_lines', 'sample lines',
		   'output_rounding', 'digits',
		   'correlation_coefficient_rounding', 'digits',
		   'output_skip_empty', '(of last data set)',
		   'fg_square_size','(in pixels)',
		   'bg_square_size','(in pixels)',
		   'norm_square_size','(in pixels)',
		   'ratio_square_size','(in pixels)',
		   'ratio_ch1','<BR>(for calculating ratio; denominator set automatically)',
		   'log_trans_ratios', '(raw data only)',
		   'percentile', '(0&lt;x&lt;1)',
		   'dup_cutoff', '<BR>The maximum allowed value (in the unit chosen below) for each duplicated pair.',
		   'dup_unit', '<BR>The cutoff can be applied to the fold difference (large/small), the absolute difference (large - small) or the standard deviations for each duplicated pair.',
		   'dup_flag', '<BR>The flag that will be applied to spots with differences above the cutoff.',
		   'dup_skip_flags', '<BR>Spots flagged as specified will be excluded from the calculation.',
#		   'rep_flawed_label', '<BR>Spots flagged as specified will be excluded from the calculation.',
		   'rep_flawed_skip_flags', '<BR>Spots flagged as specified will be excluded from the calculation.',
		   'rep_flawed_cutoff', '<BR>The maximum allowed value (in the unit chosen below) for each duplicated pair (expected values for relative difference: 1, for absolute difference: 0).',
		   'rep_flawed_unit', '<BR>The cutoff can be applied to the relative difference (large/small), the absolute difference (large - small) or the standard deviations for each set of replicates.',
		   'rep_flawed_flag', '<BR>The flag that will be applied to spots with differences above the cutoff.',
#		   'rep_flawed_files', '<BR>.',
		   'rep_flawed_percent', '%<BR>Trigger action if at least that many percent of replicates are flawed.',
		   'rep_flawed_extend', '<BR>Apply flagging to all replicates or only the flawed ones.',
		   'rep_flawed_within_slide', '<BR>Include values from duplicate spots within each slide.',
		   'rep_flawed_flag_single', '<BR>Specify if you would like to flag sets with only one value, for wich differences can not be calculated .',
		   'ma_plot_ch1', '<BR>(for calculating ratio; denominator set automatically)',
		   'marrayNormtwoD_span', '(0&lt;x&lt;1, defaults to 0.4)',
		   'marrayNormprintTipLoess_span', '(0&lt;x&lt;1, defaults to 0.4)',
		   'marrayNormscalePrintTipMAD_span', '(0&lt;x&lt;1, defaults to 0.4)',
		   'marrayNormloess_span', '(0&lt;x&lt;1, defaults to 0.4)',
#		   'marrayNorm_span', '(0&lt;x&lt;1, defaults to 0.4)',
		   'fg_min_bg_skip_flags', "<BR>Spots with the selected flag(s) will not get a value for '$FG_COR_SIGNAL_TERM'.",
		   'fg_min_percentile_skip_flags', "<BR>Spots with the selected flag(s) will be excluded from the background estimation and will not get a value for '$FG_COR_SIGNAL_TERM'.",
		   'bg_loess_skip_flags', "<BR>Spots with the selected flag(s) will be excluded from the background estimation and will not get a value for '$FG_COR_SIGNAL_TERM'.",
		   'cutoff_allowed_min', '<br>Everything below this value will be ignored and flagged.',
		   'cutoff_min', '<br>Everything below this value will be treated according to the following option:',
		   'cutoff_max', '<br>Everything above this value will be treated according to the following option:',
		   'output_files', '<br>Provide data separate for each file or combined as one large output.',
		   'limma_loess_printtip_span', 'empty box results in limma defaults, currently 0.3',
		   'limma_loess_printtip_bin', 'empty box results in limma defaults, currently 0.01',
		   'limma_loess_printtip_iter', 'empty box results in limma defaults, currently 4',
		   'limma_loess_global_span', 'empty box results in limma defaults, currently 0.3',
		   'limma_loess_global_bin', 'empty box results in limma defaults, currently 0.01',
		   'limma_loess_global_iter', 'empty box results in limma defaults, currently 4',
		   );

my $counter = 0;

my $spot_limit = '';
my $spot_limit_start = '';
my $subgrid_start = '';
my $subgrid_end = '';
my $spotid_filter = '';

my $image_stem = 'chip_plot';

my $LOAD_LIMIT = 0.9;

# variables for web interface:
# style generously provided by Indra Konnerth (http://www.web-site-bau.de)
my $css_style = 'A:link  { color: navy; font-weight: bold; text-decoration: none }
A:visited  { color: #5533FF; font-weight: bold; text-decoration: none }
A:hover  { color: black; text-decoration: underline overline; }
A:active  { color: navy; text-decoration: none }
td { font-size: 12px; font-family: Arial, Helvetica, Geneva, Swiss, SunSans-Regular }
th { font-size: 12px; font-family: Arial, Helvetica, Geneva, Swiss, SunSans-Regular }
div  { font-size: 12px; font-family: Arial, Helvetica, Geneva, Swiss, SunSans-Regular }
.small  { font-size: 10px; line-height: 11px }
h4 { text-indent: 10px }
';

my $table_element_size_limit = 38;
my $table_element_size_excess = 'hide';
my $max_files = $cgi->param('max_files') || 4;   # for file upload fields
my $max_actions = $cgi->param('max_actions') || 10; # for chain of actions
my $action_alternatives = $cgi->param('action_alternatives') || 2;
# other variables
my $waiting_time = 0;
my $waiting_time_remain = 0;
my $default_wait = 5;
my $file_num = 0;
my %control_labels = ();
my $occurence_threshold = 0;
my %param_added = ();
my %data_done = ();
my %file_input = ();
my %probe_ids = ();
my %probe_names = ();
my %orthologues = ();
my %probe_reporter_match = ();
my %labels = ();
my $map_out = '';
my $plot = '';
my @files = ();
my @reporter_lists = ();
my %list_match = ();
my %added_parameters = ();
my $check_files = '';  # check file format before carrying out action
                       # useful for command line, because otherwise files
                       # are only checked during loading
my %saved_settings_action = ();
my %saved_settings_var = ();
my $saved_settings_html = '';
my %work_files = ();
my %work_files_skip = (); # files that are to be skipped from all modules following the setting 
                          # (e.g. in merge replicates), except for data retrieval
my %work_lists = ();
my %annotation_files = ();
my @annotation_files = ();
my %unload_files = ();
my %skip_files = ();
my %wrong_format_files = ();
my @request_columns = ();
my $action_step = 0;
my $action_step_num = 0;
my $action_step_ext = '';
my $show_parameters = 0;

my $example_dir = '';  # set this in the configuration file
my $example_dir_html = '';  # set this in the configuration file
my %example_files = ();
my $log_file = '';  # set this in the configuration file

my $time1 = time;
my ($time2,$time_diff);

my $PID = $$;

my %channel_swaps = ();

my $R_executable = '/usr/local/bin/R'; # overwrite this in the configuration file
my $allow_debug = 0;
my $transpath_user = '';
my $transpath_URL = '';
my $use_cluster_nodes = '';
my $common_user_group = 'users';

my $data_db_setup = '';
my $data_db_user = '';
my $data_db_passwd = '';
my $data_db_host = '';
my $data_db_type = '';
my %data_db = ();

my $main_width = 500;
my $javascript = '';
my $pointer_image = '';  # set this in the configuration file
my $popup_box_width = 220;
my $help_javascript = "

// script to determine mouse position 
// (taken from http://www.breakingpar.com/bkp/home.nsf/Doc?OpenNavigator&U=87256B14007C5C6A87256B4B0005BFA6)
// Set Netscape up to run the \"captureMousePosition\" function whenever
// the mouse is moved. For Internet Explorer and Netscape 6, you can capture
// the movement a little easier.
if (document.layers) { // Netscape
    document.captureEvents(Event.MOUSEMOVE);
    document.onmousemove = captureMousePosition;
} else if (document.all) { // Internet Explorer
    document.onmousemove = captureMousePosition;
} else if (document.getElementById) { // Netcsape 6
    document.onmousemove = captureMousePosition;
}
// Global variables
xMousePos = 0; // Horizontal position of the mouse on the screen
yMousePos = 0; // Vertical position of the mouse on the screen
xMousePosMax = 0; // Width of the page
yMousePosMax = 0; // Height of the page

function captureMousePosition(e) {
    if (document.layers) {
        // When the page scrolls in Netscape, the event's mouse position
        // reflects the absolute position on the screen. innerHight/Width
        // is the position from the top/left of the screen that the user is
        // looking at. pageX/YOffset is the amount that the user has
        // scrolled into the page. So the values will be in relation to
        // each other as the total offsets into the page, no matter if
        // the user has scrolled or not.
        xMousePos = e.pageX;
        yMousePos = e.pageY;
        xMousePosMax = window.innerWidth+window.pageXOffset;
        yMousePosMax = window.innerHeight+window.pageYOffset;
    } else if (document.all) {
        // When the page scrolls in IE, the event's mouse position
        // reflects the position from the top/left of the screen the
        // user is looking at. scrollLeft/Top is the amount the user
        // has scrolled into the page. clientWidth/Height is the height/
        // width of the current page the user is looking at. So, to be
        // consistent with Netscape (above), add the scroll offsets to
        // both so we end up with an absolute value on the page, no
        // matter if the user has scrolled or not.
        xMousePos = window.event.x+document.body.scrollLeft;
        yMousePos = window.event.y+document.body.scrollTop;
        xMousePosMax = document.body.clientWidth+document.body.scrollLeft;
        yMousePosMax = document.body.clientHeight+document.body.scrollTop;
    } else if (document.getElementById) {
        // Netscape 6 behaves the same as Netscape 4 in this regard
        xMousePos = e.pageX;
        yMousePos = e.pageY;
        xMousePosMax = window.innerWidth+window.pageXOffset;
        yMousePosMax = window.innerHeight+window.pageYOffset;
    }
}

// Function for popping up help boxes
// (taken from http://codepunk.hardwar.org.uk/bjs19.htm)
// (see also sub help)
    function menu (whichMenu,whatState){
	var xNewPos = xMousePos;
	var width = $popup_box_width;
	if (xNewPos + width > xMousePosMax) {
	    xNewPos = xMousePosMax - width - 15;
	}
	var yNewPos = yMousePos;
//	if (yNewPos + width > yMousePosMax) {
//	    yNewPos = yMousePosMax - height - 15;
//	}
	if (document.getElementById) {
            document.getElementById(whichMenu).style.top = yNewPos + \"px\";
            document.getElementById(whichMenu).style.left = xNewPos + \"px\";
            document.getElementById(whichMenu).style.visibility = whatState;
        } else {
            document[whichMenu].top = yNewPos;
            document[whichMenu].left = xNewPos;
            document[whichMenu].visibility = whatState;
	}
    }
";
 
 
# variables for user authentication
my $username = '';
my $password = '';
my $password_check = '';

my $help_file = '';
my $help_file_link = '';
my %pipeline_help_title = ('refresh', 'Automatic Page Loading',
                           'popup_boxes', 'Popup Boxes',
                           'login', 'ArrayPipe Login',
		           'username', 'ArrayPipe Username',
		           'data_input', 'Data Input',
		           'data_input_modify', 'Data Input Modification',
		           'input_files', 'Input Files',
		           'web_files', 'Web Files',
		           'input_lists', 'Input Lists',
		           'file_compression', 'File Compression',
		           'upload_fields', 'Upload Fields',
		           'action_chain', 'Action Chain',
		           'action_settings', 'Action Settings',
		           'parameter_settings', 'Parameter Settings',
		           'data_auto_detect', 'Data Auto Detection',
		           'public_login', 'Public Entry',
		           'authentication', 'Authentication',
		           'array_layout', 'Array Layout',
		           'column_headers', 'Column Headers',
		           'flag_column', 'Flag Column',
		           'add_column', 'Additional Columns',
		           'lines_column', 'File Info',
		           'matches_column', 'Gene List Match',
		           'swap_channels', 'Swap Channels',
		           'tmev_output', 'TIGR MEV Output',
		           'tmev_output_ok', 'TIGR MEV Output',
		           'bio_rep', 'Biological Replicate',
		           'tech_rep', 'Technical Replicate',
		           'skip_file', 'Skip File',
		           'file_handle', 'File Alias',
		           'unload_box', 'Unload Files',
		           'flag_box', 'List Flag',
		           'ch_assign', 'Channel Assignment (Label and Annotation)',
		           'load_files_and_do_it', '\'Load Files and Do it\' - Button',
		           'example_files', 'Example Files',
		           'previous_steps', 'Previous Steps',
		           'unload_files_and_do_it', '\'Unload Files and Do it\' - Button',
		           'reset_button', 'Reset Button',
		           'clear_button', 'Clear Button',
		           'refine_button', 'Refine Button',
		           'do_it_button', 'Do It Button',
		           'save_settings_button', 'Save Settings Button',
		           'change_output_name', 'Output Name',
		           'line_numbers', 'Data Rows',
		           'chunk_size', 'Table Lines',
		           'header_repeat', 'Header Repeat',
		           'hide_columns', 'Hide Columns',
		           'show_columns', 'Show Columns',
		           'add_links', 'Add Links',
		           'reduce_table_element_size', 'Table element size limit',
		           'skip_flags', 'Skip Flags',
		           'rounding', 'Rounding',
		           'show_lines', 'Listing',
		           'table_font', 'Table Font',
		           'filter_operators', 'Operators',
		           'barplot_range', 'Range',
		           'barplot_breaks', 'Breaks',
		           'data_retrieval', 'Data Retrieval',
		           'header_probe_id', 'Probe ID',
		           'header_spot_id', 'SPOT ID',
		           'header_hash', '# (hash)',
		           'header_i', 'i',
		           'header_mrow', 'MROW',
		           'header_mcol', 'MCOL',
		           'header_row', 'ROW',
		           'header_col', 'COL',
		           'header_fg', 'FG',
		           'header_bg', 'BG',
		           'header_fg_l', 'FG_L',
		           'header_bg_l', 'BG_L',
		           'header_fg_l_r', 'FG_L_R',
		           'header_fg_l_i', 'FG_L_I',
		           'header_fl', 'FL',
		           'header_dup_diff', 'DUP_DIFF',
		           'header_fg_crd', 'FG_CRD',
		           'header_norm', 'Norm',
		           'header_norm_l', 'Norm_L',
		           'header_norm_l_i', 'Norm_L_I',
		           'header_norm_l_r', 'Norm_L_R',
		           'header_zscore', 'Z-score',
		           'table_window_position', 'Window Navigation',
		           'table_window_start', 'Window Start',
		           'data_file_organization', 'Data File Organization',
		           'ORTHOLOGUE', 'ORTHOLOGUE',
		           'TC', 'TC',
		           'TC description', 'TC Description',
		           'SP', 'SP',
		           'PIR', 'PIR',
		           'GP', 'GP',
		           'emb', 'emb',
		           'gb', 'gb',
		           'dbj', 'dbj',
		           'protein', 'protein',
		           'pdb', 'pdb',
		           'go term', 'go term',
		           'go description', 'go description',
		           'go ec', 'go ec',
		           'putative ortholog TOG', 'putative ortholog TOG',
		           'putative ortholog organism', 'putative ortholog organism',
		           'putative ortholog hit identifier', 'putative ortholog hit identifier',
		           'putative ortholog description', 'putative ortholog description',
		           'putative ortholog Ensembl gene', 'putative ortholog Ensembl gene',
		           'putative ortholog Ensembl transcript', 'putative ortholog Ensembl transcript',
		           'putative ortholog cover', 'putative ortholog cover',
		           'putative ortholog ident', 'putative ortholog ident',
		           'putative ortholog gaps', 'putative ortholog gaps',
		           'putative ortholog status', 'putative ortholog status',
		           'TC hits', 'TC hits',
		           'hit identity', 'hit identity',
		           'hit coverage', 'hit coverage',
		           'hit gaps', 'hit gaps',
		           'hit flag', 'hit flag',
		           'THC ENSEMBL GENE', 'THC ENSEMBL GENE',
		           'THC ENSEMBL TRANSCRIPT', 'THC ENSEMBL TRANSCRIPT',
		           'THC ENSEMBL', 'THC ENSEMBL',
		           'THC ENSEMBL COVERAGE', 'THC ENSEMBL COVERAGE',
		           'THC ENSEMBL IDENTITY', 'THC ENSEMBL IDENTITY',
		           'THC ENSEMBL GAPS', 'THC ENSEMBL GAPS',
		           'THC ENSEMBL STATUS', 'THC ENSEMBL STATUS',
		           'THC ENSEMBL HITS', 'THC ENSEMBL HITS',
		           'reporter_numbers', 'Reporters (probes)',
		           'reporter_missing', 'Missing',
		           'TransFac Factor Num', 'TransFac Factor Num',
		           'TransFac Gene Num', 'TransFac Gene Num',
		           'TransFac Factor', 'TransFac Factor',
		           'TransFac Gene', 'TransFac Gene',
		           'TransPath', 'TransPath',
		           'number of transcript hits', 'number of transcript hits',
		           'gene hits', 'gene hits',
		           'hit start', 'hit start',
		           'hit end', 'hit end',
		           'chr name', 'chr name',
		           'chr strand', 'chr strand',
		           'status', 'status',
		           'exon coverage', 'exon coverage',
		           'transcript', 'transcript',
		           'gene', 'gene',
		           'gene symbol', 'gene symbol',
		           'gene description', 'gene description',
		           'domain prediction method', 'domain prediction method',
		           'domain accession number', 'domain accession number',
		           'domain start', 'domain start',
		           'domain end', 'domain end',
		           'domain description', 'domain description',
		           'EMBL accession', 'EMBL accession',
		           'feature prediction method', 'feature prediction method',
		           'feature start', 'feature start',
		           'feature end', 'feature end',
		           'RefSeq accession', 'RefSeq accession',
		           'SPTREMBL accession', 'SPTREMBL accession',
		           'interpro interpro id', 'interpro interpro id',
		           'interpro description', 'interpro description',
		           'SWISSPROT accession', 'SWISSPROT accession',
		           'SWISSPROT function', 'SWISSPROT function',
		           'SWISSPROT gene name', 'SWISSPROT gene name',
		           'SWISSPROT protein name', 'SWISSPROT protein name',
		           'proteinfamily family id', 'proteinfamily family id',
		           'proteinfamily family description', 'proteinfamily family description',
		           'BIND', 'BIND',
		           'BIND interaction partners', 'BIND interaction partners',
		           'BioCarta', 'BioCarta',
		           'DIP', 'DIP',
		           'DIP interaction partners', 'DIP interaction partners',
		           'Ensembl accession', 'Ensembl accession',
		           'LocusLink accession', 'LocusLink accession',
		           'OMIM accession', 'OMIM accession',
		           'PDB accession', 'PDB accession',
		           'HUGO accession', 'HUGO accession',
		           'simple_view', 'Simple Spreadsheet View',
		           'spreadsheet_files', 'Input files for Spreadsheet',
		           'spreadsheet_content', 'Data categories for Spreadsheet',
		           'spreadsheet_hidden', 'Hidden data categores for Spreadsheet',
		           '', '',
		          );
my %pipeline_help_text = ('refresh', 
#                          'The browser will reload the results page automatically in intervals that are approximately equivalent to the execution time of each processing step. This mechanism helps overcome time-out problems for long processes and enables viewing of partial results as soon as they become available.',
                          'ArrayPipe tries to estimate how long a process will take and tells the browser to load the results page after that time. You can click on the results link earlier to see parts of the output. This mechanism helps overcome time-out problems for long processes and enables viewing of partial results as soon as they become available. Please note that for technical reasons it is not always possible to give a precise estimate.<P>WARNING: Due to technical reasons you could get a page error when clicking on the link too soon. In that case please reload or use the back button of your browser and try again.',
                          'popup_boxes', 
                          'Wherever you see this sign you can get a popup box with additional information when you hover with your mouse over it. It will dissappear again when you move your mouse away. This feature needs <B>Javascript</B> to be activated.<BR>You can also <B>click on the link</B>, which will open up a new window with the help text. This should work for all browsers.',
                          'login', 
			  "To protect your data from others, login with your username and password. Choose something that's easy to remember and not so easy to guess, but <B>please</B> don't use important passwords because they might be visible to the operators of the program. Any data uploaded and generated after login will only be accessible with the correct username and password (<B>Note:</B> these are case-sensitive!).",
			  'username', 
			  "An <B>empty</B> username or the username <B>$public_username</B> does not require a password. Simply click the 'Enter' or 'Login' button. Data stored this way can be accessed by (and shared with) anyone within the project.",
		           'data_input', 
                           'There are currently two ways to submit data to the pipeline:<OL><LI>Upload files directly from your computer.</LI><LI>Pick one or more of the provided examples (if available).</LI></OL>',
		           'data_input_modify', 'If you decide to remove one of the files you have worked on (for example, if it is flawed or skews the results) or if you wish to add more files to your current selection you can do so. There is also the option to upload gene lists to overlay them onto the data.',
		           'input_files', 
                           'You can upload output files from ImaGene, ScanAlyze, GenePix or ArrayVision. Simple tab-delimited files with a one-line column header should work in most cases (one array data set per file!).',
		           'web_files', 
                           'You can specify a URL to download files straight from the web. Ideally, these should be in ImaGene, ScanAlyze, GenePix or ArrayVision format, but simple tab-delimited files with a one-line column header should work in most cases (one array data set per file!).',
		           'input_lists', 
                           'You can upload files that specify a selection of spots via grid coordinates or label names. There are several possibilities to specify reporters:<UL><LI>write down the spot or probe ids (one item per line)</LI><LI>indicate the position with the coordinates MROW,MCO,ROW, and COL (separated by TABs, one entry per line)</LI><LI>use lists generated by ArrayPipe</UL>',
		           'file_compression', 'To reduce the upload time, you can compress one or more files into a <B>ZIP archive</B> or compress and bundle them with <B>gzip</B> and <B>tar</B> on Linux.',
		           'upload_fields', 'There is currently no limit for the number of files to be processed. If you wish to upload more files at one time than boxes available, please increase the field number and click the \'Refresh!\'-button.',
		           'action_chain', 'In the table below you can choose the type of actions and the order in which they are applied to your data. It starts with the top left and ends at the bottom right. Empty fields are skipped. Leave space (= empty fields) between your actions to be able to add more steps later.',
		           'action_settings', 'The actions that were applied in the last step are listed at the bottom of the page. You can modify them by adding or deleting steps, or changing their order.',
		           'parameter_settings', 'The parameter section contains a listing for each action with explanatory information and mechanisms to change the parameters.',
		           'data_auto_detect', 'For known file types, such as ArrayVision or ImaGene, the program knows which columns contain the important data, such as foreground intensities and flags. For other types of files it is only guessing and you might need to specify the right columns.<BR>For the array layout, the program tries to guess fitting parameters from the number of data rows. Please check that they are correct.',
		           'public_login', 'For this entry mode you don\'t need a username or password. Your data will be stored in public space.<BR>Simply click on \'Enter\' to continue.',
		           'authentication', 'Right after the login you will be asked again by the web-server for your username and password.',
		           'array_layout', 'Normally, arrays are divided into subgrids that are specified as <b>meta rows x meta columns</b>, and each subgrid contains <b>rows x columns</b> spots. Please specify the layout below (MCOL = meta columns, etc.). This is important if you like to see correct chip visualizations.',
		           'column_headers', 'The program needs to know which column headers are used for spot labels (<B>Spot ID</B>), foreground intensities (<B>FG</B>), and background intensities (<B>BG</B>). The intensities should be reported individually for each channel (<B>CH1, CH2</B>).<BR>The Spot ID can be set to \'<I>undefined</I>\' in which case only the spot positions are used, assigning them in the order they are read in, from top left corner to the bottom right of the slide.',
		           'flag_column', 'Specify the column header(s) that contain flagging information (either assigned automatically by the image quantification software, or added manually). To ignore all flags, set to \'<I>undefined</I>\'.',
		           'add_column', 'Specify which additional columns from the input file you would like to carry over into the ArrayPipe output file.<BR>(This has no effect unless you use the \'Retrieve Data\' function.)',
		           'lines_column', 'This column contains general info about each file. It reports the format (if recognized), the size (in Bytes) and the number of lines found in the input file (data + header + comments).',
		           'matches_column', 'This column shows the the number of times that the items of each gene list match within each data file.',
		           'bio_rep', 'To be able to merge technical replicates from the same biological sample, assign the same identifier (letters or numbers) to them.',
		           'tech_rep', 'This is currently optional - all files that have the same treatment group labeling and the same biological replicate labeling will be treated as technical replicates.',
		           'swap_channels', 'Normally, the first column of intensities is assigned to channel 1 (<B>CH1</B>)and the next to channel 2 (<B>CH2</B>). Tick this box in case you would like to switch this around. This will affect the ratios, which are normally calculated CH2/CH1',
		           'tmev_output', 'TIGR\'s Multiple Experiment Viewer requires intensities for both channels as input. Annotation information can be added separately. ArrayPipe searches the spreadsheet output for suitable data and bundles them into a downloadable compressed archive. In this case, however, <FONT COLOR=RED>no channel intensities were found!</FONT> Please load the appropriate columns (\'... FG\', \'... Norm_L_M\') or, if necessary, go back and expand the data selection for the spreadsheet view, e.g. select <b><i>everything</i></B> to be shown and click the \'Extended Spreadsheet View\' button.',
		           'tmev_output_ok', 'ArrayPipe automatically selects the channel intensities that it deems most suitable for MEV input. If you are not happy with the choice, just remove (= hide) the columns from your spreadsheet output.',

		           'skip_file', 'Tick boxes for files that you would like to temporarily exclude from the analysis.',
		           'file_handle', 'This has no function at the moment. Eventually it will allow you to specify a short name for your input file(s), which should be of particular interest in cases where you have two files for one data set, as is the case with ImaGene output.',
		           'flag_box', 'Specify the flag you would like to assign to the spots matching this list.',
		           'unload_box', 'To unload one or more of your working files, tick this box. It will come into effect next time you press any of the \'(Un)Load Files!\' or \'(Un)Load Files and Do it!\' buttons.',
		           'ch_assign', 'You can choose which label to give each channel (this is also another way of specifying a dye swap).<BR>Some modules might need to know if a channel carries the signal of a control or of a treatment. This can be assigned in this column with the letters \'<B>C</B>\' and \'<B>T</B>\'.<BR>To distinguish replicates from different controls or treatments, add numbers to the letters, e.g. <B>T1</B> and <B>T2</B> would be regarded as two different treatments, but <B>T3</B> and <B>T3</B> would refer to treatment 3 and its replicate. ',
		           'load_files_and_do_it', 'For a quick way to repeat a process on additional files, specify the data to load and click this button.',
		           'example_files', 'Your pipeline operator might have provided some example files for you, in which case they will be listed below. Tick the boxes of those files you would like to work on and click the \'Load Files!\'-button.',
		           'previous_steps', 'This section provides links to previously performed action steps. By default the submission button that was clicked and the type of actions carried out (if any) are assigned as descriptive labels. You can modify the text and assign your own description for any of the previous and the current step.',
		           'unload_files_and_do_it', 'For a quick way to repeat a process on fewer files, tick the \'unload\' box for the files you would like to drop and click this button.',
		           'reset_button', 'Click this button to reset the page to its original state.',
		           'clear_button', 'Click this button to clear the action menu.',
		           'refine_button', 'Click this button to refine the parameters for each action step (<B>Experimental!</B>).',
		           'do_it_button', 'Click this button to start processing the data.',
		           'save_settings_button', 'Click this button to save your settings (parameters and action chain) under the name specified in the text box to the right (add a description if you like).',
		           'change_output_name', 'By default the program stores each project under a directory consisting of a number (e.g. \'<I>_22981</I>\'. This text field allows you to make this name more meaningful (e.g. \'<I>batch 1 normalization</I>\'). This will be activated the next time the \'Do It\' button is clicked. (Blanks will be replaced by underscores.)',
		           'line_numbers', 'Specify how many data rows to be displayed in the table.',
		           'chunk_size', 'The output can be divided into chunks with a specified number of lines to make handling of otherwise large tables easier.',
		           'header_repeat', 'Normally, a header is printed at the beginning and end of each chunk. For more frequent headers, specify a number which forces printing of a header line after that many lines of output.',
		           'hide_columns', 'Mark the headers of the columns that you would like to hide in the output. Multiple selections can be made by holding down the CTRL-key while clicking on a list item.<BR>If group headers are listed, then you can hide multiple headers at once, i.e. those matching the string plus wild-card (*) combination that is shown.',
		           'show_columns', 'Mark the headers of the columns that you would like to show in the output. Multiple selections can be made by holding down the CTRL-key while clicking on a list item.',
		           'add_links', 'If your data contains annotation columns, these can be enriched with links to external sources. Otherwise this selection box is empty.',
		           'reduce_table_element_size', 'To avoid overly large table elements, you can specify a maximum limit for the length of the allowed text per element (specify \'0\' for no limit). Additionally, you can choose to remove the excessive text or hide it in a pop-up box.',
		           'skip_flags', '!! to be generated within the program to get the proper definition of flags ',
		           'rounding', 'To reduce the number of digits after the decimal point, select how many positions to round to. This applies to all cells in the table output that contain numbers only.',
		           'show_lines', 'Specify how many entries you would like to have listed (in addition to the barplot/histogram). Use \'<B>0</B>\' for no listing and \'<B>all</B>\' for everything (this could get BIG).',
		           'table_font', 'Change this value to increase or decrease the font of the text in the output table.',
		           'filter_operators', 'The operators \'<B>&gt;</B>\', \'<B>&gt;=</B>\', \'<B>&lt;</B>\', and \'<B>&lt;=</B>\' only work for numbers.<BR>The operator \'<B>=</B>\' can be used for both numbers and text, whereas in the latter case only an exact match will trigger a filter action.<BR>You can use the \'<B>//</B>\' operator for regular expressions, i.e. \'<I>AC\d{1,4}&nbsp;(foo|bar)</I>\' would match the string \'AC\' followed by one to four digits, then a space and then either a \'foo\' or a \'bar\'. <BR>The \'<B>!//</B>\' operator is the reverse, i.e. the entry will be shown if the pattern does not match. This is needed only for the first filter field where it is not possible to negate the match with the \'<B>BUTNOT</B>\' connector.<BR>To work with absolute numbers only, tick the box \'<B>ignore sign</B>\'. This way you can filter, for example, for a fold-change of greater than 2 or smaller than -2 in one go.',
		           'barplot_range', 'You can choose to do a barplot of only the data that you see on this page (<I>visible</I>) or of all entries in the data file (<I>all</I>).',
		           'barplot_breaks', 'For histograms, specify how many columns to show in the plot (no effect in barplot!).',
		           'data_retrieval', 'You need to select the <B>\'Retrieve Data\'</B> function at the end of your action chain, otherwise ArrayPipe will not return the calculated values in tab-delimited format or for web-based spreadsheet operations.',
		           'header_probe_id', 'If a special annotation file (derived from ProbeLynx) exists, a probe id is assigned to the spot, otherwise the spot id is listed.',
		           'header_spot_id', 'The spot id retrieved from the original input file.',
		           'header_hash', 'Sequential numbering of the rows in the output table (optional).',
		           'header_i', 'Sequential numbering of the lines in the ArrayPipe output file.',
		           'header_mrow', 'Meta rows of the array layout, i.e. specifies how many rows with subgrids the array contains.',
		           'header_mcol', 'Meta columns of the array layout, i.e. specifies how many columns with subgrids the array contains.',
  		           'header_row', 'Row numbers per subgrid of the array layout, i.e. specifies how many rows with spots are contained in a subgrid within the array.',
  		           'header_col', 'Column numbers per subgrid of the array layout, i.e. specifies how many columns with spots are contained in a subgrid within the array.',
		           'header_fg', 'Foreground intensity',
		           'header_bg', 'Background intensity',
		           'header_fl', 'Flags',
		           'header_fg_l', 'Log-transferred foreground intensity.',
		           'header_bg_l', 'Log-transferred background intensity.',
		           'header_fg_l_r', 'Ratios of foreground intensity (log-transferred).',
		           'header_fg_l_i', 'Intensity product of foreground intensity (log-transferred).',
		           'header_dup_diff', 'Factor by which duplicate neighbouring spots or replicate spots (maximum only) differ.',
		           'header_fg_crd', 'The background corrected foreground intensities.',
		           'header_norm', 'Normalized foreground intensities.',
		           'header_norm_l', 'Normalized and log-transferred foreground intensities.',
		           'header_norm_l_i', 'Intensity product of normalized and log-transferred foreground intensities.',
		           'header_norm_l_r', 'Ratio of normalized and log-transferred foreground intensities.',
		           'header_zscore', 'The Z-score, calculated as number of standard deviations that a ratio is different to the median ratio within a neighbourhood of spots.',
		           'table_window_position', 'These controls allow you to quickly position or slide the window of lines shown on the screen.<BR>You can jump to the start (\'|<\') or end (\'>|\') of the file, or you can move the window forwards (\'>\') or backwards (\'<\') by the number of lines specified in the text field.',
		           'table_window_start', 'Specify the start of the window, i.e. which is the first data row to be shown on the screen.',
		           'data_file_organization', 'There are two main ways in which the data is organized in a file:<OL><LI><B>by-row</B>:<BR>The spots are listed from the left to the right across the whole array from top to bottom. <LI><B>by-subgrid</B>:<BR>The spots are listed from the left to the right across each subgrid from top to bottom.</OL> To make sure that the chip visualization conforms to the real layout you should specify the option that applies to your data file.',
		           'Human Orthologue', 'Ensembl identifier for putative human orthologue - this is used to add more annotation to non-human probes.',
		           'TC', 'The Tentative Consensus (TC) sequences of non-human EST assemblies, according to TIGR.',
		           'TC description', 'Description of the functional role of each Tentative Consensus (TC) sequences, according to TIGR. This might include identifiers for other databases such as SwissProt, PIR and EMBL, which are extracted and reported in separate columns.',
		           'SP', 'SwissProt identifiers, extracted from TC description.',
		           'PIR', 'PIR identifiers, extracted from TC description.',
		           'GP', 'GP identifiers (GenBank Protein?), extracted from TC description.',
		           'emb', 'EMBL identifiers, extracted from TC description.',
		           'gb', 'GenBank identifiers, extracted from TC description.',
		           'dbj', 'DDBJ identifiers, extracted from TC description.',
		           'protein', 'Protein identifiers, extracted from TC description.',
		           'pdb', 'PDB identifiers, extracted from TC description.',
		           'go term', 'Gene Ontology term assigned to probe through ProbeLynx.',
		           'go description', 'Description of Gene Ontology term assigned to probe through ProbeLynx.',
		           'go ec', 'EC number of Gene Ontology term assigned to probe through ProbeLynx.',
		           'putative ortholog TOG', 'Tentative Orthologous Group (TOG) at TIGR assigned to a probe through ProbeLynx. A TOG is a Tentative Orthologous Group and contains a list of TCs from different TIGR Gene Indices (i.e. different species) which are grouped together as putative orthologs. TOGs are identified by carrying out pairwise sequence similarity searches of all TCs from all TIGR Gene Indices against each other. Those TC sequences which show a reciprocal best hit across three or more species with a minimum of 75% identity over a length of 400bp or more are included in the TOG group (Quackenbush et al., 2001).',
		           'putative ortholog organism', 'The name of the organism the putative ortholog is derived from.',
		           'putative ortholog hit identifier', 'The identifier given to the putative ortholog assigned to a probe by ProbeLyns. In most cases this will be a Tentative Human Consensus (THC) identifier.',
		           'putative ortholog description', 'Description of the functional role of a putative ortholog to this probe as assigned by ProbeLynx.',
		           'TC hits', 'Number of Tentative Consensus (TC) sequences that match this probe, according to ProbeLynx.',
		           'hit identity', 'The percentage of identity for a hit between a probe and a gene or Tentative Consensus (TC) sequence according to ProbeLynx',
		           'hit coverage', 'Percentage of probe sequence that aligns with each gene or Tentative Consensus (TC) sequence, according to ProbeLynx.',
		           'hit gaps', 'Number of gaps in the alignment between probe and gene or Tentative Consensus (TC) sequence, according to ProbeLynx.',
		           'hit flag', 'Indicator of the quality of the match between a probe sequence and a gene or Tentative Consensus (TC) sequence, ranging from 1 (for best match) to 5. For more info see ProbeLynx documentation.',
		           'THC ENSEMBL', 'The human Ensembl gene that best matches the tentative human consensus from TIGR.',
		           'THC ENSEMBL GENE', 'The human Ensembl belonging to the transcript that best matches the tentative human consensus from TIGR.',
		           'THC ENSEMBL TRANSCRIPT', 'The human Ensembl transcript that best matches the tentative human consensus from TIGR.',
		           'THC ENSEMBL COVERAGE', 'Length of alignment in percent between the tentative human consensus from TIGR and the best matching human Ensembl gene. If the orthologue has been derived by a provided accession number, this is noted as \'acc\'.',
		           'THC ENSEMBL IDENTITY', 'Identity of alignment in percent between the tentative human consensus from TIGR and the best matching human Ensembl gene. If the orthologue has been derived by a provided accession number, this is noted as \'acc\'.',
		           'THC ENSEMBL STATUS', 'THC ENSEMBL STATUS',
		           'THC ENSEMBL GAPS', 'Gaps in alignment between the tentative human consensus from TIGR and the best matching human Ensembl gene. If the orthologue has been derived by a provided accession number, this is noted as \'acc\'.',
		           'THC ENSEMBL HITS', 'Number of human Ensembl genes that best match the tentative human consensus from TIGR.',
		           'putative ortholog Ensembl gene', 'The human Ensembl belonging to the transcript that best matches the tentative human consensus from TIGR.',
		           'putative ortholog Ensembl transcript', 'The human Ensembl transcript that best matches the tentative human consensus from TIGR.',
		           'putative ortholog cover', 'Length of alignment in percent between the tentative human consensus from TIGR and the best matching human Ensembl gene.',
		           'putative ortholog ident', 'Identity of alignment in percent between the tentative human consensus from TIGR and the best matching human Ensembl gene. If the orthologue has been derived by a provided accession number, this is noted as \'acc\'.',
		           'putative ortholog gaps', 'Gaps in alignment between the tentative human consensus from TIGR and the best matching human Ensembl gene. If the orthologue has been derived by a provided accession number, this is noted as \'acc\'.',
                           'putative ortholog status', 'One of \'ortholog\' (derived from TIGR\'s EGO db), \'homolog\' (derived from sequence similarity search) or \'unknown\' (no proper match in EGO or Ensembl sequences).',
		           'reporter_numbers', 'Number of reporters (or probes) for which annotation was found.',
		           'reporter_missing', 'Number of reporters (or probes) for which annotation is missing.',
		           'TransFac Factor Num', 'Number of transcription factors related to this probe, according to TransFac.',
		           'TransFac Gene Num', 'Number of transcription factor-regulated genes that this probe matches to, according to TransFac.',
		           'TransFac Factor', 'List of the transcription factors that either match directly to the probe or indirectly through the TF-regulated gene that is represented by the probe.',
		           'TransFac Gene', 'List of the TF-regulated genes that either match directly to the probe or indirectly through the transcription factors linked to the probe.',
		           'TransPath', 'TransPath pathways build from transcription factors and binding sites.',
		           'number of transcript hits', 'The number of unique transcripts a probe maps to, according to ProbeLynx.',
		           'gene hits', 'The number of unique genes a probe maps to, according to ProbeLynx.',
		           'hit start', 'The chromosomal coordinates in which the start of the aligned probe sequence maps to, according to ProbeLynx.',
		           'hit end', 'The chromosomal coordinates in which the end of the aligned probe sequence maps to, according to ProbeLynx.',
		           'chr name', 'The name of the chromosome that the probe maps to, according to ProbeLynx.',
		           'chr strand', 'The strand of the chromosome that the probe maps to, according to ProbeLynx.',
		           'status', 'The status for a probe match: exon, intron, or medium probability intergenic, according to ProbeLynx.',
		           'exon coverage', 'The percent of the alignment that covers an exon, according to ProbeLynx.',
		           'transcript', 'ID of the Ensembl transcript matching to a probe.',
		           'gene', 'ID of the Ensembl gene associated with the transcript.',
		           'gene symbol', 'Gene symbol (HUGO ID) for this gene, derived from Ensembl.',
		           'gene description', 'Description for this gene, derived from Ensembl.',
		           'domain prediction method', 'Method(s) used for predicting a domain, derived from Ensembl.',
		           'domain accession number', 'Domain accession number, derived from Ensembl.',
		           'domain start', 'Start coordinate of the domain, derived from Ensembl.',
		           'domain end', 'End coordinate of the domain, derived from Ensembl.',
		           'domain description', 'Description of the domain, derived from Ensembl.',
		           'EMBL accession', 'EMBL accession for this gene, derived from either TIGR or Ensembl.',
		           'feature prediction method', 'Method(s) used for predicting a feature, derived from Ensembl.',
		           'feature start', 'Start coordinate of the feature, derived from Ensembl.',
		           'feature end', 'End coordinate of the feature, derived from Ensembl.',
		           'RefSeq accession', 'RefSeq accession for this gene, derived from either TIGR or Ensembl.',
		           'SPTREMBL accession', 'SPTREMBL accession for this gene, derived from either TIGR or Ensembl.',
		           'interpro interpro id', 'Interpro ID, derived from Ensembl.',
		           'interpro description', 'Interpro description, derived from Ensembl.',
		           'SWISSPROT accession', 'SWISSPROT accession for this gene, derived from either TIGR or Ensembl.',
		           'SWISSPROT function', 'Functionional annotation according to SwissProt.',
		           'SWISSPROT gene name', 'Gene name according to SwissProt.',
		           'SWISSPROT protein name', 'Protein name according to SwissProt.',
		           'proteinfamily family id', 'Protein family id, derived from Ensembl.',
		           'proteinfamily family description', 'Protein family description, derived from Ensembl.',
		           'BIND', 'Molecular interaction associated with the gene matched to this probe, derived from the Biomolecular Interaction Network Database (BIND).',
		           'BIND interaction partners', 'Interaction partners, according to BIND',
		           'BioCarta', 'Pathways that the gene matched to this probe is associated with, according to BioCarta.',
		           'DIP', 'Protein-protein interactions that this probe is associated with, derived from the Database of Interacting Proteins (DIP).',
		           'DIP interaction partners', 'Interaction partners, according to DIP.',
		           'Ensembl accession', 'Ensembl accession for transcript, gene and protein.',
		           'LocusLink accession', 'LocusLink accession for this gene, derived from Ensembl.',
		           'OMIM accession', 'OMIM accession for this gene, derived from Ensembl.',
		           'PDB accession', 'PDB accession for this gene, derived from Ensembl.',
		           'HUGO accession', 'HUGO accession for this gene, derived from Ensembl.',
		           'simple_view', 'Clicking on the \'Simple Spreadsheet View\' button transfers some of the important parts of your output data into a spreadsheet. If the annotation function was selected (and succeeded) it starts with gene name and function, followed by ratios or fold changes. Other annotation fields or significance values (if available) are hidden and can be selected for viewing in the settings section below the spreadsheet. The files to be included can be selected in the box below.',
		           'spreadsheet_files', 'Select the file(s) which which shall be used as input source(s) for the spreadsheet. The selection affects both the Simple and the Extended Spreadsheet View.',
		           'spreadsheet_content', 'Select the data categories that you would like to show up in the spreadsheet.',
		           'spreadsheet_hidden', 'The categories selected here will be included in the spreadsheet but are hidden by default (unless marked in the previous box). They can be selected for viewing in the settings section below the spreadsheet.',
			  );
my $back_to_info = '<SMALL><A HREF="#INFO">[INFO]</A></SMALL>';
my $no_script_warning = '<NOSCRIPT><BR><B>Attention:</B> Your browser is either not <B>Javascript</B> capable or you disabled this feature. This is no problem for this site, except that you won\'t see the help popup boxes. Instead you will need to access the help page by clicking on the green <FONT COLOR="#008000">[info]</FONT> items.<BR><BR></NOSCRIPT>';

#######################################################
################### MAIN PROGRAM ######################
#######################################################

my $remote_address = CGI::remote_host() || 'undefined';
my $referral_address = CGI::referer() || 'undefined';
my %flag_done = ();
my %spread_done = ();

# read local settings from configuration files:
my @config_files = ('/etc/arraypipe.config','arraypipe.config');
if (defined $ENV{'ARRAYPIPE_CONFIG'}) {
    push @config_files, $ENV{'ARRAYPIPE_CONFIG'};
}
if (defined $ENV{'HOME'}) {
    push @config_files, "$ENV{'HOME'}/.arraypipe.config";
}

if (defined $cgi->param('remote_config_file')) {
    push @config_files, $cgi->param('remote_config_file');
}
	
# set variables from input parameters and configuration files
&set_variables;


if ($webstart) {
    # if the program start is from the web
    # and no username is specified
    # let's show the login screen
    if (!defined $cgi->param('pipeline_user')
	or
	$cgi->param('pipeline_user') !~ /\w/) {

	if (defined $cgi->param('submit')
	    and
	    (
	     $cgi->param('submit') eq 'Login'
	     or
	     $cgi->param('submit') eq 'Enter'
	     )
	    ) {
	    # unless we are coming from the login screen
	    # in which case an empty username was chosen
	    # and it needs to be set to the public one:
	    $cgi->param(-name=>'pipeline_user',
			-value=>$public_username);
	} elsif (defined $cgi->param('submit')
	    and
	    $cgi->param('submit') eq 'Change Password') {
	    die "No username specified!";
	} else {
	    # get username and password
	    &login;
	}
    } else {
	if (defined $cgi->param('submit')
	    and
	    $cgi->param('submit') eq 'Change Password') {	
	    my $message = &change_password;
	    if ($username eq $public_username) {
		die "Can't change password for public account ($public_username) - please go back and enter your own username!";
	    }
	    &login($username,$message);
	}	
    }
} else {
    # we have been started from the command line!
    if ($first_call) {
#	print STDERR "Command line start!\n";
#	exit;
    }
}

my %output_collect = ();

if (defined $cgi->param('submit')
    and
    $cgi->param('submit') eq 'Load Settings and Do it!') {
    $output_collect{'WARNINGS'} .= &load_settings('apply');
    $cgi->param(-name=>'submit',
		-values=>'Do it!');
}
    
# collect all actions, so we know if we have to refresh quickly
my @actions = ();
# cycle through each action
for my $i (0..100) {    
    # skip undefined actions
    next unless ($cgi->param("action$i"));   
    push @actions, $cgi->param("action$i");
}

unless (@actions) {
    if ($cgi->param('submit') =~ /Do it/) {
	unless ($modules_listing) {
	    die "No action specified!";
	}
    }
}

# start script again in background if 
# a possibly lengthy action was requested:

my ($cmd_line,$cmd_config_file) = &get_cgi_params;

if ($username ne $public_username
    and
    $username ne ''
    and 
    $cgi->param('submit') eq 'Login') {
#	warn "checking password";
    
    # password check is only necessary at login
    # actually, maybe it's not necessary at all
    # since we have the .htaccess file in place
    # WRONG - it is necessary to avoid overwriting of password without check!
#    warn "Checking login!";
    &check_login;
}
    
unless ($cmd_line =~ /\&pid=/) {
    # we have been started directly from the browser

#    unless (-e "$tmp_dir/job.step_$action_step") {
#	my $prog = '/home/khokamp/FPMI/ArrayPipe/node/'.(basename ($0));
#	open (JOB, ">$tmp_dir/job.step_$action_step")
#	    or die "Can't write to '$tmp_dir/job.step_$action_step': $!";
#	print JOB "$prog '$cmd_line' 2> $tmp_dir/job.step_$action_step.err";
#	close JOB;
#    }

    # determine if job needs to be run in the background
    # (and on a different cluster node)
    # which should be the case if the calculation takes longer than 60 seconds
    $waiting_time = 0;
    if (@actions
	 and
	 $cgi->param('submit') =~ /Do it/i
	) {
	$file_num = 0;
	foreach ($cgi->param('work_files')) {
	    $file_num++;
	}

	# add example and upload files if 'Load Files and Do it!' was selected
	# (presuming that they are all o.k.)
	if (defined $cgi->param('submit')
	    and
	    $cgi->param('submit') eq 'Load Files and Do it!') {
	    
	    for my $i (1..$max_files) {
		if ($cgi->param("file_name$i")) {
		    $file_num++;
		}
		if ($cgi->param("web_file_name$i")) {
		    $file_num++;
		}		
	    }
	    
	    foreach ($cgi->param('example_files')) {
		$file_num++;
	    }
	}

	# subtract number of files that are to be removed
	if (defined $cgi->param('submit')
	    and
	    $cgi->param('submit') eq 'Unload Files and Do it!') {
	    
	    foreach ($cgi->param('unload_files')) {
		$file_num--;
	    }
	    foreach ($cgi->param('skip_files')) {
		$file_num--;
	    }
	}	
	    

    	for (my $i = 0; $i <= $#actions; $i++) {
	    $waiting_time += (defined $action_wait{$actions[$i]} ? $action_wait{$actions[$i]} : $default_wait);
	}

	# multiply with number of files to work on
	$waiting_time *= ($file_num ? $file_num : 1 );
    } elsif (defined $cgi->param('submit')
	     and
	     $cgi->param('submit') =~ /Spreadsheet View/) {
	# determine waiting time by number and sizes of files to be combined
	$waiting_time = 0;
	# collect files
	my %output_available = ();
	foreach ($cgi->param("output_info")) {
	    my @tmp = split /\|/, $_;
	    my $out_file = shift @tmp;
	    foreach (@tmp) {
		$output_available{$_} = $out_file;
	    }
	}
	my @input_files = ();
	foreach ($cgi->param("spreadsheet_files")) {
	    if ($_ eq 'all') {
		@input_files = values %output_available;
		last;
	    } else {
		push @input_files, $output_available{$_};
	    }
	}
	my %selected_fields = ();
	if (defined $cgi->param("spreadsheet_content")) {
	    my @fields =  $cgi->param("spreadsheet_content");
	    foreach (@fields) {
		$selected_fields{$_}++;
	    }
	}
	if (defined $cgi->param("spreadsheet_hidden_content")) {
	    my @fields =  $cgi->param("spreadsheet_hidden_content");
	    foreach (@fields) {
		$selected_fields{$_}++;
#		$hidden_fields{$_}++;
	    }
	}
	my $size_factor = 0.4; # calculate roughly that many seconds for processing of each megabyte
	my $sort_factor = 2;
	if (defined $selected_fields{'everything'}) {
	    $sort_factor = 1;
	}
	foreach my $file (@input_files) {
	    my $size = (stat $file)[7];
	    $waiting_time += $sort_factor * $size_factor * $size / 1048576;
	}

	$results_file = "$tmp_dir/spreadsheet_output_$action_step.html";
	$results_file_html = "$tmp_dir_html/spreadsheet_output_$action_step.html";
    }

    # TODO: the following doesn't seem to work!
    if (%second_data_file) {
	$waiting_time = int ($waiting_time / 2);
    }

    warn "# KH: Expected waiting time is '$waiting_time'";
    
    $waiting_time_remain = $waiting_time;

#    open (OUT, ">/home/khokamp/cmd")
#        or die "$!";
#    print OUT "$cmd_line";
#    close OUT;
#    my $tmp_host = `hostname`;
#    print STDOUT $cgi -> header().&html_head("$progname - DUMP ($tmp_host)");
#    print STDOUT "<BR>$cmd_line<BR><BR>".$cgi->Dump();
#    exit;


    if ($waiting_time >= 3) {
	# some lengthy action required - run as background command
	# to avoid timeout
	$cmd_line .= '&spreadsheet_background='.$results_file;
	&add_config($cmd_config_file, 'spreadsheet_background='.$results_file);
	$cmd_line .= '&prev_page='.$referral_address;
	&add_config($cmd_config_file, 'prev_page='.$referral_address);
	$cmd_line .= '&pid='.$PID;
	&add_config($cmd_config_file, 'pid='.$PID);
	$cmd_line .= '&results_file='.$results_file unless ($cmd_line =~ /\&results_file=/);	
	&add_config($cmd_config_file, 'results_file='.$results_file) unless ($cmd_line =~ /\&results_file=/);
	$cmd_line =~ s/'/\\'/g;

	if (-e $results_file) {
	    unlink $results_file
		or die "Can't delete $results_file!\n";
	}

	warn "# time to wait: $waiting_time";
	my $job = "$0 '$cmd_line\&waiting_time=$waiting_time\&second_call=1' $cmd_config_file &";
	&add_config($cmd_config_file, "waiting_time=$waiting_time");
	&add_config($cmd_config_file, "second_call=1");

	my $node = '';
	if ($use_cluster_nodes) {
	    # get available nodes
	    my @nodes = &get_nodes;

	    if ($nodes[0] eq 'No node available!') {
		die "Currently no computing node available - please try again later!\n";
	    }

	    foreach (@nodes) {
		chomp;
		my @tmp = split;
		if ($tmp[0] =~ /^c(\d\d\d)$/i) {
		    $node = $tmp[0];
		    last;
		}
	    }

	    if ($node) {
		
                warn "# KH: node is '$node'";
	    	my $prog = '/home/khokamp/FPMI/ArrayPipe/node/'.(basename ($0));
		my $cmd_line2 = $cmd_line;
#		$cmd_line2 .= '&swap_dir=/www/koch.pathogenomics.ca/data/pipeline/&dir_swap=/home/khokamp/FPMI/pipeline';
		if ($webstart) {
		    $cmd_line2 .= '&executable='.$cgi -> url();
		    &add_config($cmd_config_file, 'executable='.$cgi -> url());
		}

		$cmd_line2 .= "\&waiting_time=$waiting_time\&second_call=1";
		&add_config($cmd_config_file, "waiting_time=$waiting_time\&second_call=1");

		# add setting of environmental variable for specific configuratoin file 
		# to be called from the nodes
		my $remote_config = '';
		if (defined $remote_config_file) {
		    #$remote_config = "export ARRAYPIPE_CONFIG='$remote_config_file'; ";
		    $cmd_line2 .= '&remote_config_file='.$remote_config_file;
		    &add_config($cmd_config_file, 'remote_config_file='.$remote_config_file);
		}		    

		# write job command to a file (makes the call much shorter
		# and allows to pipe STDERR into file)
		open (JOB, ">$tmp_dir/job.step_$action_step")
		    or die "Can't write to '$tmp_dir/job.step_$action_step': $!";
		$cmd_line2 =~ s/\'/_/g;
		print JOB "$prog '&test=test' $cmd_config_file";
		close JOB;

                $job = "rsh -n -l $remote_user $node sh '$tmp_dir/job.step_$action_step' > $tmp_dir/job.step_$action_step.err 2>\&1 &";

	    }
	}

        warn "# KH: job: '$job'";
	
	system "$job";
	
	# wait until background program has written refresh file:
        my $sleeper = 0;

#        warn "# KH: Found '$results_file' after $sleeper seconds!\n";

	# redirect browser to that file
	# changed to printing HTML code directly which will redirect to results file:
	my $direct_output = $cgi -> header().&html_head_refresh("$progname (step $action_step, in progress)","$waiting_time; URL=$results_file_html",'-1',$node);

	# The following should only be done for the spreadsheet view
	# otherwise the action_step number is increased:
	if (defined $cgi->param('submit')
	    and
	    $cgi->param('submit') =~ /Spreadsheet View/) {
	    if (open (RES, ">$results_file")) {
		print RES &html_head_refresh("$progname (step $action_step, in progress)","$waiting_time; URL=$results_file_html",'-1',$node);
		close RES;
		system "chgrp $common_user_group '$results_file' 2> /dev/null";
		system "chmod 666 '$results_file' 2> /dev/null";			
	    }
	}

#	$direct_output .= "<BR>Final results will become available at <A HREF=\"$results_file_html\">$results_file_html</A>\n<BR>(Please click on this link if your browser window does not refresh automatically or to see partial results as they come in.)\n</BODY></HTML>";
	print STDOUT $direct_output unless (defined $cgi->param('second_call'));
	close STDOUT;
#	print STDOUT $cgi->redirect("$results_file_html");
#	print STDOUT "\n<BR>\ncmd: $cmd_line\n";
	exit;
    }
} else {

    # either the waiting time is below the limit
    # or
    # we are in the second round (called from CGI-script):

    if (defined $cgi->param('waiting_time')) {
	my @tmp = $cgi->param('waiting_time');
	$waiting_time_remain = pop @tmp;
	$waiting_time = $waiting_time_remain;
#	warn "#KH: getting time: $waiting_time_remain";
    }
   

    if ($username ne $public_username) {
#	warn "Checking login!";
        #&check_login;
    }    

#    unless (-e "$tmp_dir/job.step_$action_step") {
#	my $prog = '/home/khokamp/FPMI/ArrayPipe/node/'.(basename ($0));
#	open (JOB, ">$tmp_dir/job.step_$action_step")
#	    or die "Can't write to '$tmp_dir/job.step_$action_step': $!";
#	print JOB "$prog '$cmd_line' 2> $tmp_dir/job.step_$action_step.err";
#	close JOB;
#    }

}
	
my $tmp_results_file = $results_file.'.tmp';

    my $show_command = 0;
#    $show_command = 1;
    if ($show_command) {
	warn "#KH: '$cmd_line'\n";
	open (TMP, ">/tmp/cmd")
            or die "$!";
	print TMP "perl -d /home/khokamp/FPMI/arraypipe/arraypipe.pl '$cmd_line'\n";
#	print TMP "perl -d /var/www/cgi-bin/arraypipe.pl '$cmd_line'\n";
	close TMP;
	exit;
    }

# LOG report
if ($log_file) {
    if (open (LOG, ">>$log_file")) {
	print LOG "\n";
	print LOG '-' x 60;
	print LOG "\n$progname $PID: Date: ".(localtime)."\n";
	print LOG "$progname $PID: Remote address: $remote_address\n";
	print LOG "$progname $PID: Referral address: $referral_address\n";
    } else {
	warn "Can't append to log '$log_file': $!";
    }
}

#####################
### START PROGRAM ###
#####################

if ($modules_listing) {
    # general module listing
    print "\n\nModules listing for ArrayPipe:\n";
    foreach (sort keys %action_items) {
	print "  $_ ($action_items{$_})\n";
    }
    
    exit;
}

my $results_section = 0;
# initialize each possible section
foreach ('WARNINGS','INFO','RESULTS','ACTION SETUP','SETTINGS','DATA FILES','INPUT','FINISH') {
    $output_collect{$_} = '';
}

my $submit = $cgi->param('submit') || '';

# first check for calls that diverge from the main pipeline
if ($submit =~ /Spreadsheet/
    or
    $submit eq 'Reprint Table!'
    or
    $submit eq 'Save List!'
    or
    $submit =~ /^Show/
    or
    $submit =~ /^Hide/
    or
    $submit =~ /[>|<]/
    or 
    (defined $cgi->param('program_part')
     and
     $cgi->param('program_part') eq 'html_table_output')
    ) {
    if ($submit eq 'Barplot!') {
	&plot_barplot('Barplot');
    } elsif ($submit eq 'Histogram!') {
	&plot_barplot('Histogram');
    } elsif ($submit eq 'Plot Pathways!') {
	&plot_pathways;
    } elsif ($submit eq 'TransPath') {
	&generate_transpath_output;
    } elsif ($submit eq 'TIGR MEV') {
	&generate_mev_output;
    } elsif ($submit eq 'Add Column!') {
	&add_spreadsheet_column;
	&generate_html_output;
    } else {
	&generate_html_output;
    }
    exit;
}

#warn "#KH: $cmd_line";
#exit;

# write header and select output channel
my $refresh = 1;
if (@actions
    and
    ($cgi->param('submit') =~ /Do it/i
     or
     $cgi->param('submit') =~ /Refine/i)
    ) {
    $refresh = (defined $action_wait{$actions[0]} ? $action_wait{$actions[0]} : $default_wait);
    # multiply with number of files to work on
    $file_num = 0;
    foreach ($cgi->param('work_files')) {
	$file_num++;
    }
    for my $i (1..$max_files) {
	if ($cgi->param("file_name$i")) {
	    $file_num++;
	}
    }
    $refresh *= ($file_num ? $file_num : 1 );
}

# CHANGED: Use waiting time instead:
#&initialize_output($refresh,'first');
#sleep 1;
&initialize_output($waiting_time,'first');

# a self-refreshing header is written to the results file
# and the default file handle points to the temporary results file
# (in append mode)
if ($cgi->param('submit') =~ /Do it/i
    or
    $cgi->param('submit') =~ /refine/i) {
    # we might need to refresh a couple of times
    # if processing was requested
    print STDOUT $cgi->redirect("$results_file_html");
}
# otherwise just wait until the new page is completed...

# print HTML output
my $url_executable = $cgi -> url();

unless ($webstart) {
    $url_executable = $cgi_link;
    if (defined $cgi->param('executable')
	and
	$cgi->param('executable')) {
	$url_executable = $cgi->param('executable');
    }
}


$output_collect{'START'} .= "\n".($cgi -> start_multipart_form({action => $url_executable, name => 'pipeline'}))."\n";

$output_collect{'START'} .= "<DIV>\n<H1>ArrayPipe";

if ($FPMI_specific) {
    if ($progname =~ /FPMI/i) {
	 $output_collect{'START'} .= " <FONT COLOR=RED>FPMI</FONT>";
    }
    unless ($progname =~ /latest/) {
	my $progname_latest = $cgi->url();
	$progname_latest =~ s/\.pl/_latest\.pl/; 
	if (-e basename $progname_latest) {	
	    if (-x basename $progname_latest) {
#		$output_collect{'START'} .= " (try <A HREF=\"$progname_latest\">latest version</A>)";
	    } else {
		warn "# WARNING: Can't execute $progname_latest!";
	    }
	}
	$output_collect{'START'} .= " (version $revision_number)";
    } else {
	$output_collect{'START'} .= " ($progname $revision_number) - work in progress!";
#	$output_collect{'START'} .= " <FONT COLOR=RED>FPMI</FONT>, $revision_number";
    }
} else {
    $output_collect{'START'} .= " (version $revision_number)";
}

$output_collect{'START'} .= "</H1>\n$no_script_warning\n";

my $new_files = 0;

# section for selecting/uploading files:
$output_collect{'INPUT'} .= &input_section;

# Carry out file checks and loads and collect output
# before printing the help message which goes to the top
# of the screen:

# save settings:
if ($submit eq 'Save Settings!') {
    $output_collect{'WARNINGS'} .= &save_settings;
}

my $settings_loaded = '';
if ($submit eq 'Load Settings!') {
    $settings_loaded = 1;
    $output_collect{'WARNINGS'} .= &load_settings;
}

# have to do this earlier, so we can calculate the length of operation!
#if ($submit eq 'Load Settings and Do it!') {
#    $settings_loaded = 1;
#    $output_collect{'WARNINGS'} .= &load_settings('apply');
#    $submit = 'Do it!';
#}

# load newly selected files
if ($submit eq 'Load Files!') {
    $output_collect{'WARNINGS'} .= &get_files;
}
if ($submit eq 'Load Files and Do it!') {
    $output_collect{'WARNINGS'} .= &get_files;
    $submit = 'Do it!';
}
    

# unload checked files
if ($submit eq 'Unload Files!') {
    $output_collect{'WARNINGS'} .= &unload_files;
}
if ($submit eq 'Unload Files and Do it!') {
    $output_collect{'WARNINGS'} .= &unload_files;
    $submit = 'Do it!';
}
    
# remove invalid files
if ($submit eq 'Remove files!') {
    # NOT DONE YET!
    $output_collect{'WARNINGS'} .= &remove_files;
}

# list files available for action
$output_collect{'DATA FILES'} .= &list_files;


# help section is the first thing to be printed
$output_collect{'INFO'} .= &help_section;

if (%wrong_format_files) {
    $output_collect{'WARNINGS'} .= "<H2>Warnings</H2><FONT COLOR=RED>\n";
    foreach my $file (sort keys %wrong_format_files) {
	$output_collect{'WARNINGS'} .= "$wrong_format_files{$file}\n<BR>\n";
    }
    $output_collect{'WARNINGS'} .= "</FONT><BR><BR>\n";
}
%wrong_format_files = ();			     


if ($submit eq 'Do it!') {
    my $load = &get_system_load;
#    warn "#KH: Load: $load";
    if ($load > $LOAD_LIMIT) {
	$submit = '';
	die "System load is currently too high ($load) - please try again later...\n";
#	print "<BR><BR><FONT COLOR=RED><BIG><B>System load is currently too high ($load) - please try again later...</B></BIG></FONT><BR>\n";
    }
}

#print "</DIV>\n";

# copy everything collected so far to the refreshed file
system "cat '$tmp_results_file' >> '$results_file'";

# this is the call that produces the web-page when we fork out to another node:
my $hostname = `hostname`;

if ($submit =~ /do it/i
    or
    $cgi->param('submit') =~ /Refine/i) {   
    print STDOUT $cgi->redirect("$results_file_html") unless (defined $cgi->param('second_call'));
}

##############
# Action part:
##############

if ($submit eq 'Refine!') {
    $submit = 'Do it!';
    $refine_parameters = 1;
    $show_parameters = 1;
}
    

if ($submit =~ /show/
    or
    $submit =~ /hide/) {

    # check if the previous file contained results
    
    my $prev_page = $referral_address;
    
    if (defined $cgi->param('prev_page')) {
	$prev_page = $cgi->param('prev_page');
    }

    my $prev_results = '';
    if (open (PREV, $prev_page)) {
	$prev_results = "reading from $prev_page";
    } else {
	$prev_page = "$tmp_dir/".(basename $prev_page);
	if (open (PREV, $prev_page)) {
	    $prev_results = "reading from $prev_page";
	} else {
    #	$prev_results = "Couldn't open previous page ($prev_page): $!\n";
	}
    }

    if ($prev_results) {
	$prev_results = '';
	while (<PREV>) {
	    last if (eof(PREF));
	    if (/<\!-- START of INFO section -->/) {
		$prev_results = $_;
		while (<PREV>) {
		    $prev_results .= $_;
		    if (/<!-- END of RESULTS section -->/) {
			last;
		    }
		    last if (eof(PREF));
		}
	    }
	}
	close PREV;
    }

    if ($prev_results) {
	print "\n$output_collect{'START'}";

	# print everything from start of INFO section

	$results_section = 1;
	$output_collect{'RESULTS'} = 1;
	print $prev_results;
    }

# end of treatment for hide/show section submit

} elsif ($submit eq 'Do it!'
    and
    @actions) {
    # ACTION REQUESTED

    # collect the columns required by each action item    
    my @required_columns = &get_required_columns;

    my $warning_messages = '';

    # read all necessary columns into hash
    # and take note on which columns are available
    my $valid_files = 0;
    foreach my $file_key (sort keys %work_files) {
	my @files = ($file_key);
	if (defined $second_data_file{$file_key}) {
	    push @files, $second_data_file{$file_key};
	}
	foreach my $file (@files) {
	    my $rc = &read_data($file,@required_columns);
	    unless (%{$data{$file}{'__DATA__'}}) {
		# second part of multi-part input, such as ImaGene:
		delete $data{$file};
	    }
	    if ($rc =~ /ok;?\s*(.+)/) {
		my $add_warning = $1;
		$add_warning =~ s/;/<BR>WARNING: /g;
		$warning_messages .= "<BR><B><FONT COLOR=RED>WARNING: $add_warning</FONT></B><BR>\n";
		$rc = 'ok';
	    }
	    unless ($rc eq 'ok') {
		$warning_messages .= "<BR><B><FONT COLOR=RED>WARNING: $rc</FONT></B><BR>\n";
	    } else {
		$valid_files++;
	    }
	}
    }

    # set flags for uploaded lists
    my $list_counter = 0;
    foreach my $file (sort keys %work_lists) {
	$list_counter++;
#	my $flag = (defined $flag_values{"list$list_counter"} ? $flag_values{"list$list_counter"} : $flag_values{'list1'}); 
	my $flag = 'list1';
	if (defined $list_specs{$file}{'FLAG'}) {
	    $flag = $list_specs{$file}{'FLAG'};
	}
	unless (defined $flag_values{$flag}) {
	    # create a new list
	    $highest_flag_num *= 2;
	    $flag_values{$flag} = $highest_flag_num;
	    $flag_values_num{$highest_flag_num} = $flag;
	    push @all_flags, $flag;
	    unshift @all_flags_num, $highest_flag_num;
	}
	# keep track of list flags:
	if ($flag =~ /^list\d+/) {
	    $used_flags{$flag}++;
	}

	my $flag_num = $flag_values{$flag};

	my $rc = &read_reporter_list($file,$flag_num);
	unless ($rc eq 'ok') {
	    $warning_messages .= "<BR><B><FONT COLOR=RED>WARNING: $rc</FONT></B><BR>\n";
	} else {
	    $valid_files++;
	    push @{$list_flag_match{'FLAG'}{$flag_values_num{$flag_num}}}, $file;
	    $list_flag_match{'FILE'}{$file} = $flag_values_num{$flag_num};
	}
    }

    if ($warning_messages) {
	$warning_messages .= "<BR>\n";
    }

#    $output_collect{'RESULTS'} .= "\n<BR><TABLE><TR><TD>\n<BR>\n<BR><TABLE WIDTH=$main_width cellpadding=\"6\"><TR BGCOLOR=lightgrey><TD><CENTER><BIG><B><A NAME=\"results_section\">RESULTS</A></B></BIG></CENTER></TD></TR></TABLE></TD></TR></TABLE>";
    print "\n$output_collect{'START'}";
    print "\n<!-- START of INFO section -->\n";
    print "\n<BR><table WIDTH=$main_width><tr><td>
<table WIDTH=$main_width cellpadding=\"6\"><tr BGCOLOR=lightgrey><td><center><big><b><a NAME=\"INFO\">INFO</a></b></big></center></td></tr></table>
<center><small>
<a HREF=\"#INPUT\">[INPUT]</a>
<a HREF=\"#DATA FILES\">[DATA FILES]</a>
<a HREF=\"#RESULTS\">[RESULTS]</a>
<a HREF=\"#ACTION SETUP\">[ACTION SETUP]</a>
<a HREF=\"#SETTINGS\">[SETTINGS]</a>
</small></center>
</td></tr></table>
<!-- END of INFO section -->\n";

    print "\n$output_collect{'INFO'}";
    print "\n<!-- START of RESULTS section -->\n";
    print "\n<BR><table WIDTH=$main_width><tr><td>
<table WIDTH=$main_width cellpadding=\"6\"><tr BGCOLOR=lightgrey><td><center><big><b><a NAME=\"RESULTS\">RESULTS</a></b></big></center></td></tr></table>
<center><small>
<a HREF=\"#INFO\">[INFO]</a>
<a HREF=\"#INPUT\">[INPUT]</a>
<a HREF=\"#DATA FILES\">[DATA FILES]</a>
<a HREF=\"#ACTION SETUP\">[ACTION SETUP]</a>
<a HREF=\"#SETTINGS\">[SETTINGS]</a>
</small></center>
</td></tr></table>\n";

# <TABLE><TR><TD>\n<BR>\n<BR><TABLE WIDTH=$main_width cellpadding=\"6\"><TR BGCOLOR=lightgrey><TD><CENTER><BIG><B><A NAME=\"results_section\">RESULTS</A></B></BIG></CENTER></TD></TR></TABLE></TD></TR></TABLE>";

    $results_section = 1;
    $output_collect{'RESULTS'} = 1;

#    $output_collect{'WARNINGS'} .= $warning_messages;
    print $warning_messages;
    
    unless (@actions) {
	$output_collect{'WARNINGS'} .= "<BR><B>Please select at least one action to be performed!</B>\n";
    }

    unless ($valid_files) {
	$output_collect{'WARNINGS'} .= "<BR><B>None of the input files seems readable or valid!</B>\n";
	@actions = ();
    } else {
	if (@actions > 1) {
#	    $output_collect{'RESULTS'} .= "Action sections:\n<BR><UL>\n";
	    print "Action sections:\n<BR><UL>\n";
	    for (my $i = 0; $i <= $#actions; $i++) {
		
		my $action = $actions[$i];
		
#		$output_collect{'RESULTS'} .= "<LI><A HREF=\"#action".($i+1)."\">Action ".($i+1).": "
		print "<LI><A HREF=\"#action".($i+1)."\">Action ".($i+1).": $action_items{$action}</A></LI>";	
	    }
#	    $output_collect{'RESULTS'} .= "</UL><BR>\n";
	    print "</UL><BR>\n";
	}
    }

    # cycle through each action
    $refresh = 4;
    for (my $i = 0; $i <= $#actions; $i++) {

	my $action = $actions[$i];

#	$output_collect{'RESULTS'} .= "<BR>
	print "<BR>
<TABLE><TR><TD>
  <TABLE><TR><TD BGCOLOR=LIGHTGREY>
    <H3>
      <A name=\"action".($i+1)."\">Action ".($i+1).": $action_items{$action}\&nbsp;</A>
    </H3>
  </TD></TR></TABLE>
</TD><TD>
  <TABLE><TR><TD><SMALL>"
    # $i+1 is the anchor for the current action
    . ($i > 0 ? "<A HREF=\"#action$i\">[previous action]</A>&nbsp;" : '')
    . ($i < $#actions ? "<A HREF=\"#action".($i+2)."\">[next action]</A><BR>" : '<BR>')
    . "<A HREF=\"#".($section_view{'SETTINGS'} eq 'hide all' ? 'SETTINGS' : $action)."\">[settings]</A>&nbsp;<A HREF=\"#INFO\">[info section]</A><BR>
  </SMALL></TD></TR></TABLE>
</TD></TR></TABLE>\n";

	unless ($function{$action}) {
	    warn "No action '$action' defined!";
#	    $output_collect{'RESULTS'} .= "<FONT COLOR=RED>No action '$action' defined!</FONT><BR>\n";
	    print "<FONT COLOR=RED>No action '$action' defined!</FONT><BR>\n";
	    next;
	}
		
	
#	$output_collect{'RESULTS'} .= "\n\n";
	print "\n\n";
	
	# call the sub routine named after the action
	my $action_time = time;
	eval { $function{$action}->($i,$action) }; 
	$action_time = time - $action_time;
	# warn if action didn't execute properly
	if ($@) {
	    warn "Function $action couldn't be executed: $@";
#	    $output_collect{'RESULTS'} .= "\n<BR><B>ERROR: <BR>\nFunction $action couldn't be executed: $@</B><BR>\n";
	    print "\n<BR><B>ERROR: <BR>\nFunction $action couldn't be executed: $@</B><BR>\n";
	}

#	$output_collect{'RESULTS'} .= "\n\n";
	print "\n\n";

	if ($i == $#actions) {
	    print "<BR>\n<!-- END of RESULTS section -->\n";
	}

	close RESULT;

#	if (defined $actions[$i+1]) {	    
#	    $refresh = (defined $action_wait{$actions[$i+1]} ? $action_wait{$actions[$i+1]} : $default_wait);
#	} else {
#	    $refresh = $default_wait;
#	}
#	# multiply with number of files to work on
#	$refresh *= scalar (keys %work_files);
#

	my $current_refresh = (defined $action_wait{$actions[$i]} ? $action_wait{$actions[$i]} : $default_wait);
	$current_refresh *= scalar (keys %work_files);

	$waiting_time_remain -= $current_refresh;
	if ($waiting_time_remain < $default_wait) {
	    $waiting_time_remain = $default_wait;
	}

	# CHANGED: refresh at new estimated end-time (waiting time - refresh times)
#	&initialize_output($refresh);
	warn "#KH: action time: $action_time, remain: $waiting_time_remain ($action)\n";
	&initialize_output($waiting_time_remain);
	
	# copy everything collected so far to the refreshed file
	system "cat '$tmp_results_file' >> '$results_file'";
    }
} # end of submit with @actions section


# provide interface for selection of action items
# if one or more files are uploaded
if (%work_files) {
    &list_action_items;
}

#print "<BR><BR><B>DUMP:</B><BR>".$cgi->Dump();

my $original_action_step = $action_step;
$action_step_num++;
$action_step = $action_step_num.$action_step_ext;
my $results_file_new = $results_file;
$results_file_new =~ s/_step_\d+(\.\d+)*\./_step_${action_step}\./;

# hidden values:
$output_collect{'FINISH'} .= "<input type=\"hidden\" name=\"waiting_time\" value=\"$waiting_time\">\n";
$output_collect{'FINISH'} .= "<input type=\"hidden\" name=\"previous_action_step\" value=\"$original_action_step\">\n";
$output_collect{'FINISH'} .= "<input type=\"hidden\" name=\"action_step\" value=\"$action_step\">\n";
$output_collect{'FINISH'} .= "<input type=\"hidden\" name=\"results_file\" value=\"$results_file_new\">\n";
$output_collect{'FINISH'} .= "<input type=\"hidden\" name=\"tmp_dir\" value=\"$tmp_dir\">\n";
$output_collect{'FINISH'} .= "<input type=\"hidden\" name=\"tmp_dir_html\" value=\"$tmp_dir_html\">\n";
$output_collect{'FINISH'} .= "<input type=\"hidden\" name=\"last_pid\" value=\"$$\">\n";
$output_collect{'FINISH'} .= "<input type=\"hidden\" name=\"pipeline_user\" value=\"$username\">\n";
foreach (keys %section_view) {
    $output_collect{'FINISH'} .= "<input type=\"hidden\" name=\"$_\" value=\"$section_view{$_}\">\n";
}
  
if (defined $annotation{$COMMON_GENE_NAME}) {
    $output_collect{'FINISH'} .=  "<input type=\"hidden\" name=\"spreadsheet_annotation\" value=\"1\">\n";
} else {
    $output_collect{'FINISH'} .=  "<input type=\"hidden\" name=\"spreadsheet_annotation\" value=\"0\">\n";
}
$output_collect{'FINISH'} .=  "<input type=\"hidden\" name=\"spreadsheet_annotation_headers\" value=\"".(join '|', @annotation_headers_all)."\">\n";
# store flags
foreach (%flag_values) {
    next unless (defined $flag_values{$_});
    $output_collect{'FINISH'} .= "<input type=\"hidden\" name=\"flags\" value=\"$_|$flag_values{$_}\">\n";
}
if (%work_files) {
    $output_collect{'FINISH'} .= "<BR>Last chance to start the action: <input type=\"submit\" name=\"submit\" value=\"Do it!\">\n";
    $output_collect{'FINISH'} .= "<BR>
<center><small>
<a HREF=\"#INPUT\">[INPUT]</a>
<a HREF=\"#DATA FILES\">[DATA FILES]</a>
<a HREF=\"#RESULTS\">[RESULTS]</a>
<a HREF=\"#ACTION SETUP\">[ACTION SETUP]</a>
<a HREF=\"#SETTINGS\">[SETTINGS]</a>
</small></center>
";
}

#$output_collect{'FINISH'} .= "<input type=\"hidden\" name=\"pipeline_password_check\" value=\"$password_check\">\n";

# create file for added output
#my $file_add = &check_for_added_data;
#if ($file_add) {
#    $output_collect{'FINISH'} .= "<input type=\"hidden\" name=\"add_data\" value=\"$file_add\">\n";
#}    


$output_collect{'FINISH'} .= $cgi -> end_form();

$output_collect{'FINISH'} .= &ending('start over', 'home', 'tutorial', 'show_processing_time');


# add description to warnings:
if ($output_collect{'WARNINGS'}) {
    $output_collect{'WARNINGS'} = "<B>The following warnings were issued by the program:</B><BR>\n".$output_collect{'WARNINGS'};
}

# write output in intelligent order
&write_output;

close LOG if ($log_file);

close RESULT;

&finalize_output;

unless ($submit =~ /do it/i) {
    print STDOUT $cgi->redirect("$results_file_html");
}

# write added output
#if ($file_add) {
#    &write_add_data($file_add);
#}

system "chmod 666 '$tmp_dir/*' 2> /dev/null";

unlink $tmp_results_file;

my $notify_address = 'karsten_hokamp@sfu.ca';
$notify_address = '';
if ($notify_address) {
    system "mail '$notify_address' -s 'ArrayPipe output ready at $results_file' < /dev/null 2> /dev/null";
}

if (-t STDIN) {
    select STDOUT;
    print "\nPlease point your browser to '$results_file_html'\n\n";
}

###################
### SUBROUTINES ###
###################


sub section_view {
    my $action = shift;
    my $direction = shift;

    $action =~ /(.*?) (.*)\!/;
    my $scope = lc $1;
    my $section = uc $2;

    my $tmp_dir = $cgi->param('tmp_dir');
    my $in_file = "$tmp_dir/".(basename $referral_address);
    my $out_file = "$in_file.tmp.html";

    my $new_section = '';

    $section_view{$section} = "$direction $scope";

    return;

    # the rest was a previous version where the content of the file was simply changed and returned
    # this has changed now, so this subroutine is only required to change the settings of %section_view

#    warn "# setting view: $direction $scope";
    if ($section eq 'SETTINGS') {
	$new_section .= &print_settings_section;
    }

    open (IN, $in_file)
	or die "Can't read file $in_file: $!";
    open (OUT, ">$out_file")
	or die "Can't write file $out_file: $!";
    system "chmod 666 $out_file 2> /dev/null";
    while (<IN>) {
	if (/<\!-- INFO START of $section section -->/) {
	    print OUT $new_section;
	    until (/<\!-- END of $section section -->/) {
		$_ = <IN>;
	    }
	    print OUT $_;
	} else {
	    print OUT $_;
	}
    }
    close IN;
    close OUT;

    # creating a new file might require a lot of changing of variables (action_step, etc.), 
    # so we just overwrite the existing one and send it out again:
    system "mv $out_file $in_file";
    warn "# sending back to $referral_address";
    print STDOUT $cgi->redirect($referral_address);

#    print STDOUT $cgi -> header().&html_head("$progname - Spreadsheet, ERROR");
#    print STDOUT "<BR><BR>action: $action, referral: $referral_address, scope: $scope, section: $section<BR><B>DUMP:</B><BR>".$cgi->Dump();
#    exit;
}

sub proper_header {
    my $header = shift;
    my $header_out = $header;
    # supress breaking of ' end' and ' start',
    # to make box wider and avoid deep staggering
    # of many numbers, which would increase row height
    $header_out =~ s/ end/\&nbsp;end/;
    $header_out =~ s/ start/\&nbsp;start/;
    $header_out =~ s/hit flag/hit\&nbsp;flag/;
#    $header_out =~ s/_/ /g;
    my $out = '';
    my $channel = '';
    my $text_overwrite = '';
    my $title_overwrite = '';

    if ($header eq 'MROW') {
	$out = &help(
		     {
			 'topic' => 'header_mrow',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header eq 'MCOL') {
	$out = &help(
		     {
			 'topic' => 'header_mcol',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header eq 'ROW') {
	$out = &help(
		     {
			 'topic' => 'header_row',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header eq 'COL') {
	$out = &help(
		     {
			 'topic' => 'header_col',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header eq 'Probe ID') {
	$out = &help(
		     {
			 'topic' => 'header_probe_id',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header eq 'SPOT ID') {
	$out = &help(
		     {
			 'topic' => 'header_spot_id',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header eq $COMMON_GENE_NAME) {
	$text_overwrite = "A gene name for this spot derived from the following annotation fields (whatever contains a value first): gene symbol, HUGO accession, SWISSPROT gene name, SWISSPROT accession, SPTREMBL accession, EMBL accession, gene, Ensembl accession. If no name was found 'unknown' is used.";
	$title_overwrite = $COMMON_GENE_NAME;
	$out = &help(
		     {
			 'topic' => 'header_gene_name',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header eq $COMMON_GENE_NAME_SOURCE) {
	$text_overwrite = "The database that provided the common gene name.";
	$title_overwrite = $COMMON_GENE_NAME_SOURCE;
	$out = &help(
		     {
			 'topic' => 'header_gene_name_type',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header eq $COMMON_GENE_DESCRIPTION) {
	$text_overwrite = "A gene description for this spot derived from the following annotation fields (whatever contains a value first): SWISSPROT function, SWISSPROT protein name, interpro description, gene description, proteinfamily family description,domain description, feature prediction method. If no description was found 'unknown' is used.";
	$title_overwrite = $COMMON_GENE_DESCRIPTION;
	$out = &help(
		     {
			 'topic' => 'header_gene_description',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header eq $COMMON_GENE_DESCRIPTION_SOURCE) {
	$text_overwrite = "The database that provided the common gene description.";
	$title_overwrite = $COMMON_GENE_DESCRIPTION_SOURCE;
	$out = &help(
		     {
			 'topic' => 'header_gene_description_type',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header eq 'i') {
	$out = &help(
		     {
			 'topic' => 'header_i',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header eq '#') {
	$out = &help(
		     {
			 'topic' => 'header_hash',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?$FG_SIGNAL_TERM$/) {
	$channel = $1;
	if ($channel) {
	    $text_overwrite = "Foreground intensity from $channel channel";
	    $title_overwrite = 'Foreground';
	}
	$out = &help(
		     {
			 'topic' => 'header_fg'.$channel,
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?)_i$/) {
	$text_overwrite = "Number of spots that were merged. Only the first of the spots merged together will get this value assigned.";
	$title_overwrite = "Merged Spots (i)";
	$out = &help(
		     {
			 'topic' => 'header_merged_intensity_numbers',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?${FG_SIGNAL_TERM}_M$/) {
	$text_overwrite = "Merged foreground intensities.";
	$title_overwrite = "${FG_SIGNAL_TERM}_M";
	$out = &help(
		     {
			 'topic' => 'header_merged_intensities',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?${MERGED_RATIOS_TERM}$/) {
	$text_overwrite = "Ratios from multiple files (and duplicated spots) merged into a single ratio.";
	$title_overwrite = "${FG_SIGNAL_TERM}";
	$out = &help(
		     {
			 'topic' => 'header_merged_ratios',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?${MERGED_RATIOS_TERM}_L$/) {
	$text_overwrite = "Log2-ratios from multiple files (and duplicated spots) merged into a single log2-ratio.";
	$title_overwrite = "${FG_SIGNAL_TERM}_L";
	$out = &help(
		     {
			 'topic' => 'header_merged_ratios_log',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?${FG_SIGNAL_TERM}_L_M$/) {
	$text_overwrite = "Merged log-transferred foreground intensities.";
	$title_overwrite = "${FG_SIGNAL_TERM}_M";
	$out = &help(
		     {
			 'topic' => 'header_merged_intensities',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?${FG_SIGNAL_TERM}_M_L_R$/) {
	$text_overwrite = "Log-transferred ratios of merged foreground intensities.";
	$title_overwrite = "${FG_SIGNAL_TERM}_M_L_R";
	$out = &help(
		     {
			 'topic' => 'header_fg_M_L_R',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?${FG_SIGNAL_TERM}_R$/) {
	$text_overwrite = "Ratios from foreground intensities";
	$title_overwrite = 'Ratios';
	$out = &help(
		     {
			 'topic' => 'header_ratios',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?${FG_SIGNAL_TERM}_R_L$/) {
	$text_overwrite = "Log-transferred ratios from foreground intensities";
	$title_overwrite = 'Log ratios';
	$out = &help(
		     {
			 'topic' => 'header_log_ratios',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?${FG_SIGNAL_TERM}_L$/) {
	$channel = $1;
	if ($channel) {
	    $text_overwrite = "Log-transferred (base 2) foreground intensities in channel $channel";
	    $title_overwrite = 'Log-values';
	}
	$out = &help(
		     {
			 'topic' => 'header_log'.$channel,
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?$BG_SIGNAL_TERM$/) {
	$channel = $1;
	if ($channel) {
	    $text_overwrite = "Background intensity from $channel channel";
	    $title_overwrite = 'Background';
	}
	$out = &help(
		     {
			 'topic' => 'header_bg'.$channel,
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?${FG_SIGNAL_TERM}_L$/) {
	$channel = $1;
	if ($channel) {
	    $text_overwrite = "Log-transferred foreground intensity from $channel channel";
	    $title_overwrite = 'Log\'ed Foreground';
	}
	$out = &help(
		     {
			 'topic' => 'header_fg_l'.$channel,
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?${BG_SIGNAL_TERM}_L$/) {
	$channel = $1;
	if ($channel) {
	    $text_overwrite = "Log-transferred background intensity from $channel channel";
	    $title_overwrite = 'Log\'ed Background';
	}
	$out = &help(
		     {
			 'topic' => 'header_bg_l'.$channel,
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?Norm_L_M$/) {
	$channel = $1;
	$text_overwrite = "Several foreground intensities have been normalized, log-transferred and merged into a single value.";
	$title_overwrite = 'Merged Log\'ed Normalized Foreground';
	$out = &help(
		     {
			 'topic' => 'header_norm_l_m',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?)_SD$/) {
	$channel = $1;
	$text_overwrite = "The standard deviation calculated from the values that have been merged into a single value.";
	$title_overwrite = 'Standard Deviation (SD)';
	$out = &help(
		     {
			 'topic' => 'header_norm_l_m_sd',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?)_list$/) {
	$channel = $1;
	$text_overwrite = "The individual values that have been merged into a single value.";
	$title_overwrite = 'Value List (list)';
	$out = &help(
		     {
			 'topic' => 'header_norm_l_m_list',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?)_orig$/) {
	$channel = $1;
	$text_overwrite = "The original foreground intensities from which the merged value has been derived. This gives an indication of the level of intensity.";
	$title_overwrite = 'Value List (orig)';
	$out = &help(
		     {
			 'topic' => 'fg_m_orig',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?Norm_L_M_R$/) {
	$channel = $1;
	$text_overwrite = "The ratios calculated from normalized, log-transferred and merged foreground intensities.";
	$title_overwrite = 'Ratios of Merged Log\'ed Normalized Foreground';
	$out = &help(
		     {
			 'topic' => 'header_norm_l_m_r',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?$FOLD_CHANGE_TERM$/) {
	$channel = $1;
	$text_overwrite = "Fold change, based on the highest level intensity values available (normalized over background corrected over foreground intensities and merged over log-transformed over natural values).<BR>A value of 1 is equal to no change. Everything greater than 1 means a positive change (up-regulation) and everything smaller than -1 means a negative change (down-regulation) from one condition to another (usually from CH1 to CH2).";
	$title_overwrite = 'fold change';
	$out = &help(
		     {
			 'topic' => 'header_fold_change',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?$RATIO_TERM$/) {
	$channel = $1;
	$text_overwrite = "Ratio, based on the highest level intensity values available (normalized over background corrected over foreground intensities and merged over log-transformed over natural values).<BR>A value of 1 is equal to no change. Everything greater than 1 means a positive change (up-regulation) and everything smaller than 1 means a negative change (down-regulation) from one condition to another (usually from CH1 to CH2).";
	$title_overwrite = 'ratio';
	$out = &help(
		     {
			 'topic' => 'header_ratio',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?p-student/) {
	$channel = $1;
	$text_overwrite = "p-value calculated with Student t-test.";
	$title_overwrite = 'p-student';
	$out = &help(
		     {
			 'topic' => 'header_p-student',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?p-paired/) {
	$channel = $1;
	$text_overwrite = "p-value calculated with paired t-test.";
	$title_overwrite = 'p-paired';
	$out = &help(
		     {
			 'topic' => 'header_p-paired',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?p-welch/) {
	$channel = $1;
	$text_overwrite = "p-value calculated with Welch's t-test.";
	$title_overwrite = 'p-welch';
	$out = &help(
		     {
			 'topic' => 'header_p-welch',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?p-wilcox/) {
	$channel = $1;
	$text_overwrite = "p-value calculated with Wilcoxon test.";
	$title_overwrite = 'p-wilcox';
	$out = &help(
		     {
			 'topic' => 'header_p-wilcox',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?p-value/) {
	$channel = $1;
	$text_overwrite = "p-value calculated by t-test or ANOVA.";
	$title_overwrite = 'p-value';
	$out = &help(
		     {
			 'topic' => 'header_p-value',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?F-value$/) {
	$channel = $1;
	$text_overwrite = "F-value calculated by ANOVA.";
	$title_overwrite = 'p-value';
	$out = &help(
		     {
			 'topic' => 'header_F-value',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?p-permute$/) {
	$channel = $1;
	$text_overwrite = "p-value calculated by permutation and paired t-test.";
	$title_overwrite = 'p-permute';
	$out = &help(
		     {
			 'topic' => 'header_p-perm',
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /${FG_SIGNAL_TERM}_L_R$/) {
	$out = &help(
		     {
			 'topic' => 'header_fg_l_r',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /${FG_SIGNAL_TERM}_L_I$/) {
	$out = &help(
		     {
			 'topic' => 'header_fg_l_i',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?$FLAG_CHANNEL_TERM$/) {
	$channel = $1;
	if ($channel) {
	    $text_overwrite = "Flags specific to $channel channel";
	    $title_overwrite = 'Flags';
	}
	$out = &help(
		     {
			 'topic' => 'header_fl'.$channel,
			 'label' => $header_out,
			 'text' => $text_overwrite,
			 'title' => $title_overwrite,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?DUP_DIFF$/) {
	$out = &help(
		     {
			 'topic' => 'header_dup_diff',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?FG_CRD$/) {
	$out = &help(
		     {
			 'topic' => 'header_fg_crd',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?Norm$/) {
	$out = &help(
		     {
			 'topic' => 'header_norm',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?Norm_L$/) {
	$out = &help(
		     {
			 'topic' => 'header_norm_l',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?Norm_L_R$/) {
	$out = &help(
		     {
			 'topic' => 'header_norm_l_r',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?Norm_L_I$/) {
	$out = &help(
		     {
			 'topic' => 'header_norm_l_i',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif ($header =~ /(.*?) ?Z-score$/) {
	$out = &help(
		     {
			 'topic' => 'header_zscore',
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif (grep /\b$header\b/, @{$var_def_values{'annotate_items_probelynx'}}
	     and
	     defined $pipeline_help_title{$header}) {
	$out = &help(
		     {
			 'topic' => $header,
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } elsif (grep /\b$header\b/, @{$var_def_values{'annotate_items_funcdb'}}
	     and
	     defined $pipeline_help_title{$header}) {
	$out = &help(
		     {
			 'topic' => $header,
			 'label' => $header_out,
			 'no_small' => 1
		     }
		     );
    } else {
	$out = $header_out;
    }

    return $out;
}

sub check_login {

    # make sure that we either have the plain password
    # or the encrypted:
    $password = $cgi->param('pipeline_password') || '';
    unless ($password =~ /\w/) {
	unless (defined $cgi->param('pipeline_password_check') 
		and
		$cgi->param('pipeline_password_check') =~ /\w/) {	
	    die "No password! Please go back and specify a password.";
	}
    }

    # check if user exists already
    # by finding a .htpasswd file in the user directory
    my $password_file = $tmp_dir;
    $password_file =~ s/(.*$username\/).*/$1/;
    $password_file .= '.htpasswd';
    if (-r $password_file) {
	# get previously defined password:
	open (PASS, $password_file)
	    or die "Can't open password file";
	my ($user, $passwd);
	while (<PASS>) {
	    chomp;
	    ($user,$passwd) = split /\:/, $_;
	    if ($user eq $username) {
		last;
	    }
	}
	close PASS;

	unless ($passwd) {
	    die "No previous password for $username found!";
	}	
	# make sure passwords match
	if (defined $cgi->param('pipeline_password_check') 
	    and
	    $cgi->param('pipeline_password_check') =~ /\w/) {
	    $password_check = $cgi->param('pipeline_password_check');
	} else {
	    my $salt = substr($passwd, 0, 2);
	    $password_check = crypt($password,$salt);
	}
	if ($password_check ne $passwd) {
	    die "Wrong password - please go back and try again!";
	}

	# create authentication file
	my $auth_file = "$tmp_dir/.htaccess";
	unless (-e $auth_file) {
	    open (ACC, ">$auth_file")
		or die "Can't write to access file ($auth_file): $!";
	    print ACC "<Files \"*\">
AuthType Basic
AuthName \"ArrayPipe Output for $username\"
require user $username
AuthUserFile $password_file
</Files>
";
	    close ACC;
#	system "chmod 400 '$auth_file'";
	}
    } else {
	# add username/password to password file
	open (PASS, ">>$password_file")
	    or die "Can't write to password file ($password_file): $!";
	flock(PASS, 2);
	
	my $salt = (localtime)[0];
	$salt .= $salt if (length($salt) == 1);
	$password_check = crypt($password,$salt);
	# try to avoid special characters in encrypted password
	$salt = 9;
	while ($password_check =~ /\W/) {
	    $salt++;
	    last if ($salt > 99);
	    $password_check = crypt($password,$salt);
	}
	print PASS "$username:$password_check\n";
	flock(PASS, 8); # free the lock;     
	close PASS;
#	system "chmod 400 '$password_file'";

	# write htaccess file
	my $auth_file = "$tmp_dir/.htaccess";
	open (ACC, ">$auth_file")
	    or die "Can't write to access file ($auth_file): $!";
	print ACC "<Files \"*\">
AuthType Basic
AuthName \"ArrayPipe Output for $username\"
require user $username
AuthUserFile $password_file
</Files>
";
	close ACC;
#	system "chmod 400 '$auth_file'";
    }
}


sub login {

    my $default_username = shift || $public_username;
    my $password_change = shift || '';
    
    my $title = "$progname - Login";
    $javascript .= "
	function setFocus()
	{
	    // set autofocus to pipeline_user
	    document.pipeline.pipeline_user.focus();
	    document.pipeline.pipeline_user.select();
	}
	
	function clearForm()
	{
	    document.pipeline.pipeline_user.defaultValue = '';
	    document.pipeline.pipeline_user.value = '';
	    setFocus();
	}
";
                     
    $javascript .= $help_javascript;
    print $cgi -> header();
    print $cgi->start_html(-title=>"$title",
			   -author=>"$operator_email",
			   -BGCOLOR=>'white',
			   -style=>{-code=>$css_style},                         
			   -script=>$javascript,
			   -onload=>"setFocus()"
			   );
    
    print "\n".($cgi -> start_multipart_form({action => $url_executable, name => 'pipeline'}))."\n";
    print "<DIV style=\"width:500px; padding:10px; font-family: Arial, Helvetica, Geneva, Swiss, SunSans-Regular\">\n<H1>Welcome to ArrayPipe ";
#    warn "# FPMI: $FPMI_specific, prog: $progname";
    if ($FPMI_specific
	and
	$progname =~ /FPMI/i) {
	print "<FONT COLOR=RED>FPMI</FONT> ";
    }
    print "(version $revision_number)</H1>\n$no_script_warning";
    print "<TABLE><TR><TD BGCOLOR=YELLOW><BIG><B>&nbsp;&nbsp;Info&nbsp;&nbsp;</BIG></B></TD></TR></TABLE><BR>\n";
    print "<table width=500><TR><TD>\nThis tool aims to help visualizing and analyzing your two-colour microarray slides.<BR>\nThroughout the process you will find pointers"
	. ($pointer_image ? ' ('.&pointer.')' : '')
	. " with guidance information and <B>popup boxes</B> (hidden behind \"".&help( 
										       {
											   'topic' => 'popup_boxes',
											   'URL' => $help_file_URL,
										       }
										       )
	."\") with additional explanations. Please note that this feature is based on Javascript and to avoid compatibility problems, the labelled text also acts as a hyperlink to a web page with the same help text.<br>\n";
    if ($tutorial_link) {
	print "<BR>A tutorial is provided at <A HREF=\"$tutorial_link\">$tutorial_link</A> to make you familiar with the basic and advanced usage of ArrayPipe.\n";
    }

    print "</TD></TR></TABLE>\n<BR><P><P><TABLE><TR><TD BGCOLOR=LIGHTBLUE><BIG><B>&nbsp;&nbsp;Login&nbsp;&nbsp;</BIG></B></TD></TR></TABLE><BR>\n";
    print &pointer
	."To protect your data, please enter a username and password and log in, otherwise just click the 'Enter' button."
	.&help( 
		{
		    'topic' => 'login',
		    'URL' => $help_file_URL,
		}
		)
	."\n"; 
    print "<TABLE>
<TR>
<TD>public access: </TD>
<TD><TABLE><TR><TD><TABLE CELLPADDING=2><TR><TD BGCOLOR=LIGHTBLUE><input name=\"submit\" value=\"Enter\" type=\"submit\"></TD></TR></TABLE></TD><TD>"
.&help( 
	{
	    'topic' => 'public_login',
	    'URL' => $help_file_URL,
	}
	)
    ."</TD></TR></TABLE></TD>
</TR>
<TR>
<TD COLSPAN=2><B>or</B> protect your data with a username/password combination:<BR><BR></TD>
</TR>
<TR>
<TD>username: </TD><TD><input name=\"pipeline_user\" type=\"text\" size=\"15\" maxlength=\"15\" value='$default_username'>"
    .&help( 
	    {
		'topic' => 'username',
		'URL' => $help_file_URL,
	    }
	    )
."</TD>
</TR>
<TR>
<TD>password: </TD><TD><input name=\"pipeline_password\" type=\"password\" size=\"12\" maxlength=\"12\"></TD>
</TR>
<TR>
<TD> </TD>
<TD><TABLE><TR><TD><TABLE CELLPADDING=2><TR><TD BGCOLOR=BLUE><input name=\"submit\" value=\"Login\" type=\"submit\"></TD></TR></TABLE></TD><TD>"
    .&help( 
	    {
		'topic' => 'authentication',
		'URL' => $help_file_URL,
	    }
	    )
    ."</TD></TR></TABLE></TD>
</TR>
</TABLE>
";

    print "<BR><P><P><TABLE><TR><TD BGCOLOR=LIGHTGREEN><BIG><B>&nbsp;&nbsp;Password Change&nbsp;&nbsp;</BIG></B></TD></TR></TABLE><BR>\n";
#    print "<BR><H2>Password Change</H2>\n";
    
    if ($password_change) {
	print "$password_change";
    } else {
	
	print &pointer."If you would like to change your password, <B>fill out the form above</B> and in the section below type your new password twice, then click the button.<BR><BR>\n"; 
	print "<TABLE>
<TR>
<TD>new password: </TD><TD><input name=\"pipeline_new_password\" type=\"password\" size=\"12\" maxlength=\"12\"></TD>
</TR>
<TR>
<TD>control: </TD><TD><input name=\"pipeline_new_password_control\" type=\"password\" size=\"12\" maxlength=\"12\"></TD>
</TR>
<TR>
<TD> </TD><TD><input name=\"submit\" value=\"Change Password\" type=\"submit\"></TD>
</TR>
</TABLE>
";
	
    }
    
    
    print &ending('start over', 'home', 'tutorial');

    # add comment in case program was started from the command line
    print "
<!-- 

This was the login page for the ArrayPipe web server!

In case you started this program from the command line, 
please use the '-help' switch for instructions!

-->
";

    exit;
}

    
sub change_password {

    # make sure that we either have the plain password
    # or the encrypted:
    $password = $cgi->param('pipeline_password') || '';
    unless ($password =~ /\w/) {
	unless (defined $cgi->param('pipeline_password_check') 
		and
		$cgi->param('pipeline_password_check') =~ /\w/) {	
	    die "No password! Please go back and specify your old password.";
	}
    }

    my $new_password = '';

    # check if user exists already
    # by finding a .htpasswd file in the user directory
    my $password_file = $tmp_dir;
    $password_file =~ s/(.*$username\/).*/$1/;
    $password_file .= '.htpasswd';
    if (-r $password_file) {

	# make sure the two new passwords match
	my $new_password1 = '';
	my $new_password2 = '';

	unless (defined $cgi->param('pipeline_new_password')) {
	    die "No new password specified!";
	} else {
	    $new_password1 = $cgi->param('pipeline_new_password');
	}

	unless (defined $cgi->param('pipeline_new_password_control')) {
	    die "No control of the new password specified!";
	} else {
	    $new_password2 = $cgi->param('pipeline_new_password_control');
	}

	unless ($new_password1 =~ /\w/) {
	    die "No new password specified!";
	}

	unless ($new_password2 =~ /\w/) {
	    die "No control of new password specified!";
	}

	unless ($new_password1 eq $new_password2) {
	    die "Mismatch between new password and control!";
	} else {
	    $new_password = $new_password1;
	}

	# get previously defined password:
	open (PASS, $password_file)
	    or die "Can't open password file";
	my ($user, $passwd);
	while (<PASS>) {
	    chomp;
	    ($user,$passwd) = split /\:/, $_;
	    if ($user eq $username) {
		last;
	    }
	}
	close PASS;

	unless ($passwd) {
	    die "No previous password for $username found!";
	}	
	# make sure passwords match
	if (defined $cgi->param('pipeline_password_check') 
	    and
	    $cgi->param('pipeline_password_check') =~ /\w/) {
	    $password_check = $cgi->param('pipeline_password_check');
	} else {
	    my $salt = substr($passwd, 0, 2);
	    $password_check = crypt($password,$salt);
	}
	if ($password_check ne $passwd) {
	    die "Wrong password - please go back and try again!";
	}

	# create authentication file
	my $auth_file = "$tmp_dir/.htaccess";
	unless (-e $auth_file) {
	    open (ACC, ">$auth_file")
		or die "Can't write to access file ($auth_file): $!";
	    print ACC "<Files \"*\">
AuthType Basic
AuthName \"ArrayPipe Output for $username\"
require user $username
AuthUserFile $password_file
</Files>
";
	    close ACC;
#	system "chmod 400 '$auth_file'";
	}
    } else {
	if ($username eq $public_username) {
	    die "Can't change password for public account ($public_username) - please go back and enter your own username!";
	}
	die "User '$username' doesn't exist!";
    }

    # read in all existing passwords
    open (IN, "$password_file")
	or die "Can't read password file ($password_file): $!";
    my %passwords = ();
    while (<IN>) {
	my ($user,$pass) = split /\:/, $_, 2;
	$passwords{$user} = $pass;
    }
    close IN;
    
    
    # change the password
    my $salt = (localtime)[0];
    $salt .= $salt if (length($salt) == 1);
    $password_check = crypt($new_password,$salt);
    # try to avoid special characters in encrypted password
    $salt = 9;
    while ($password_check =~ /\W/) {
	$salt++;
	last if ($salt > 99);
	$password_check = crypt($new_password,$salt);
    }
    $passwords{$username} = $password_check;

    # write new password file:
    open (PASS, ">$password_file")
	or die "Can't write to password file ($password_file): $!";
    flock(PASS, 2);
    
    foreach my $username (keys %passwords) {
	print PASS "$username:$passwords{$username}\n";
    }

    flock(PASS, 8); # free the lock;     
    close PASS;
#	system "chmod 400 '$password_file'";

    return "<FONT COLOR=\"green\"><B>Password for user '$username' changed!<BR>Please log in above with the new password.</B></FONT><BR>\n";

}

       
sub pointer {
    if ($pointer_image) {
	return "<IMG SRC=\"$pointer_image\" BORDER=0 ALT=\"Tip: \">\n";	
    } else {
	return '';
    }
}


sub help {

    my $topic = shift;
    
    my %args = ();
    if (ref($topic) eq 'HASH') {
	%args = %$topic;
	$topic = $args{'topic'};
    }
	
#    my $text = shift || '[more&nbsp;info]';
    my ($text, $width, $help_text_overwrite, $help_title_overwrite, $help_URL);
    if (defined $args{'label'}) {
	$text = $args{'label'};
    } elsif (@_) {
	$text = shift;
	unless ($text =~ /\w/) {
	    $text = '[info]';
	}
    } else {
	$text = '[info]';
    }

    if (defined $args{'width'}) {
	$width = $args{'width'};
    } elsif (@_) {
	$width = shift;
	unless ($width =~ /\d/) {
	    $width = $popup_box_width;
	}
    } else {
	$width = $popup_box_width;
    }

    if (defined $args{'text'}
	and
	$args{'text'}) {
	$help_text_overwrite = $args{'text'};
    } elsif (@_) {
	$help_text_overwrite = shift;
    } else {
	$help_text_overwrite = '';
    }

    if (defined $args{'title'}
	and
	$args{'title'}) {
	$help_title_overwrite = $args{'title'};
    } elsif (@_) {
	$help_title_overwrite = shift;
    } else {
	$help_title_overwrite = '';
    }

    if (defined $args{'URL'}) {
	$help_URL = $args{'URL'};
    } elsif (@_) {
	$help_URL = shift;
    } else {
	$help_URL = '';
    }

    my $pipeline_help_title = 'Sorry - no help available, yet!';
    my $pipeline_help_text = '';
    if (defined $pipeline_help_title{$topic}) {
	$pipeline_help_title = $pipeline_help_title{$topic};
    } else {
	warn "# $progname: no help title for $topic available!\n" unless ($help_title_overwrite);
    }

    if (defined $pipeline_help_text{$topic}) {
	$pipeline_help_text = $pipeline_help_text{$topic};
    } else {
	warn "# $progname: no help text for $topic available!\n" unless ($help_text_overwrite);
    }

    if ($help_text_overwrite) {
	$pipeline_help_text = $help_text_overwrite;
    }

    if ($help_title_overwrite) {
	$pipeline_help_title = $help_title_overwrite;
    }

    # (based on code from http://codepunk.hardwar.org.uk/bjs19.htm)

    my $link = '';
    if ($help_URL) {
	$link = $help_URL;
    } else {
	$link = $help_file_link;
    }

    my $small_class = 'class=small';
    if (defined $args{'no_small'}
	and
	$args{'no_small'}) {
	$small_class = '';
    }

    my $popup_background = '#ffff00';
    if (defined $args{'background'}
	and
	$args{'background'}) {
	$popup_background = $args{'background'};
    }

    my $no_link = 0;
    if (defined $args{'no_link'}
	and
	$args{'no_link'}) {
	$no_link = $args{'no_link'};
    }
    my $help = "<A HREF=\"$link#$topic\" TARGET=\"pipeline_help\" $small_class style=\"font-weight:bold;color:#008000;cursor:hand\" onMouseover=\"menu('$topic','visible')\" onMouseout=\"menu('$topic','hidden')\">
$text</A>";
    if ($no_link) {
	$help = "<A HREF=\"\" $small_class style=\"font-weight:bold;color:#008000;cursor:hand\" onMouseover=\"menu('$topic','visible')\" onMouseout=\"menu('$topic','hidden')\">
$text</A>";
}
$help .= "
<div id=\"$topic\" style=\"position: absolute;  top: 0;  left: 0;  width: ${width}px;  border: solid 3px #000000;  background: $popup_background;  padding: 3px;  z-index: 1;  visibility: hidden;\">";
unless ($pipeline_help_title eq '__none__') {
    $help .= "<CENTER><SMALL>[click on text to show help file]</SMALL></CENTER>\n"
    ."<P><B>$pipeline_help_title</B>
<P>";
}
$help .=
"$pipeline_help_text
</div>
";
#<BR>
#<BR>
#<a href=\"javascript://\" onMouseOver=\"menu('$topic','hidden')\" class=\"menulink\">CLOSE</a>
#</div>";
    return $help;

}


sub get_nodes {
    
    my @nodes = `ruptime -lr`;
    chomp @nodes;
    foreach (@nodes) {
#	warn "#: node '$_'";
	next if (/down/);
	my @tmp = split;
	next if ($tmp[0] =~ /00[1-3]/
		 or
		 $tmp[0] =~ /04\d/
		 );
	my $load = $tmp[6];
	chop $load;
	next if ($load > 1.1);
	my $time_now = time;
	my $node_file = "/www/koch.pathogenomics.ca/data/arraypipe/logs/$tmp[0]";
	if (-e $node_file) {
	    # make sure the last usage was more than a minute ago!
#	    my $time = (stat($node_file))[9];
	    my $time = `tail -1 $node_file`;
	    chomp $time;
	    my $diff = $time_now - $time;
	    warn "#KH: diff for node $tmp[0] is '$diff'";
	    if ($diff < 120) {
		next;
	    }
	}
#	system "touch $node_file";
	system "echo $time_now >> $node_file";
	my $time = localtime;
#	warn "#KH: touching $node_file at $time";
	return $tmp[0];
    }
    return 'No node available!';
}



sub add_spreadsheet_column {
    
#    print STDOUT $cgi -> header().&html_head("$progname - Spreadsheet, ERROR");
#    print STDOUT "<BR><BR><B>DUMP:</B><BR>".$cgi->Dump();
#    exit;

    my $show_command = 0;
#    $show_command = 1;
    if ($show_command) {
	warn "#KH: '$cmd_line'\n";
	open (TMP, ">/tmp/cmd");
	print TMP "perl -d /home/khokamp/FPMI/arraypipe/arraypipe.pl '$cmd_line'\n";
	close TMP;
	exit;
    }

    my $col1 = $cgi->param('add_column1') || '';
    my $col2 = $cgi->param('add_column2') || '';
    if ($col2 eq '-') {
	$col2 = '';
    }
    my $operator = $cgi->param('add_column_op') || '';
    my $label = $cgi->param('add_column_name') || '';
    
    unless ($col1 =~ /\w/) {
	die "Please specify column1!";
    }
    
    my %binary_op = ('+', 1, '-', 1, '/', 1, '*', 1);

    unless ($col2 =~ /\w/) {
	if (
	    defined $binary_op{$operator}
#	    $operator eq '+'
#	    or
#	    $operator eq '-'
#	    or
#	    $operator eq '/'
#	    or
#	    $operator eq '*'
	    ) {
	    die "Please specify column2!";
	}
    }
    
    if ($operator eq '') {
	die "Please specify operator!";
    }
    
    unless ($label =~ /\w/) {
	die "Please specify a column label!";
    }
    if ($label =~ /(\W+)/) {
	unless ($1 eq '_'
		or
		$1 eq ' '
		or
		$1 eq '-') {
	    die "Error: string contains non-alphanumeric symbol ('$1') - please go back and change the column label!";
	}
    }
    
    # get the file to be worked on:
    
    my $input_file = '';
    unless (defined $cgi->param('spreadsheet_input')) {
	die "No input file specified!";
    }
    $input_file = $cgi->param('spreadsheet_input');
    $input_file =~ s/(\.flagged|\.filtered|\.sorted)//g;

    my $data_dir = dirname($input_file);
    
#    warn " # reaading file '$input_file', calculating '$label' from '$col1' $operator '$col2'";

    # read file
    open (SHEET_DATA, $input_file)
        or die "Can't read input file '$input_file': $!";

    # open output file
    my $output_file = $input_file;
    my $delete_file = '';
    if ($output_file =~ /\.add(\d+)$/) {
	my $ending_num = $1 + 1;
#	warn "# changing $1 to $ending_num";
	$output_file =~ s/\d+$/$ending_num/;
	$delete_file = $input_file;
    } else {
	$output_file .= '.add1';
    }


#    warn "# new output_file: $output_file";

    open (OUT, ">$output_file") 
	or die "Can't write output ($output_file): $!";

    my @headers = ();
    my %headers = ();
	
    # get the header first:

    while (<SHEET_DATA>) {

	# remove comments (everything following a hash)
	my ($line,$rest) = split /\#/;

	# skip if nothing's on the line
	unless ($line =~ /\w/) {
	    print OUT $_;
	    next;
	}

	chomp $line;
	
	# first line with data -> use it for header
	my $header_counter = -1;
	foreach (split /\t/, $line) {
	    $header_counter++;
	    push @headers, $_;
	    $headers{$_} = $header_counter;
	}

	# add new column header:
	$line .= "\t$label\n";
	print OUT $line;

	unless (defined $headers{$col1}) {
	    die "ERROR: Column1 ($col1) not found in headers!";
	}

	if ($col2) {
	    unless (defined $headers{$col2}) {
		die "ERROR: Column2 ($col2) not found in headers!";
	    }
	}

	last;
    }

    # read in data
    
    my %op = (
	      '-' => \&subtract_op,
	      '+' => \&add_op,
	      '*' => \&multiply_op,
	      '/' => \&divide_op,
	      'log2' => \&log2_op,
	      'exp2' => \&exp2_op,
	      );
    
    my $counter = 0;
    while (<SHEET_DATA>) {
	
	$counter++;

	# remove newline without touching any trailing tabs
	s/\n$//;
	s/\r$//;

#	last if ($counter > 10);

	my $add = '';

	my @tmp = split /\t/, $_;

	my $pos1 = $headers{$col1};
	my $pos2 = '';
	if ($col2) {
	    $pos2 = $headers{$col2};
	}

	my $val1 = '';
	if ($tmp[$pos1]  =~ /\d/) {
	    $val1 = $tmp[$pos1];
	}
	my $val2 = '';
	if ($col2) {
	    if ($tmp[$pos2]  =~ /\d/) {
		$val2 = $tmp[$pos2];
	    }
	}
	
#	warn "# action $counter: $val1 $operator $val2";
	if (defined $binary_op{$operator}) {
	    if ($val1 eq ''
		or
		$val2 eq '') {
		$add = '';
	    } else {
		eval { $add = $op{$operator}->($val1, $val2) };
		if ($@) {
		    $add = '';
		}
	    }
	} else {
	    if ($val1 eq '') {
		$add = '';
	    } else {
		eval { $add = $op{$operator}->($val1) };
		if ($@) {
		    $add = '';
		}
	    }
	}
	
#	warn "# result: $add\n";

	$_ .= "\t$add\n";

	print OUT $_;
    }

    close SHEET_DATA;
    close OUT;

    system "chgrp $common_user_group '$output_file' 2> /dev/null";
    system "chmod 666 '$output_file' 2> /dev/null";	
    
    if ($delete_file) {
	unlink $delete_file
	    or warn "Can't delete $delete_file: $!";
    }

    # new column has been added

    # make sure now that it will be included in the spreadsheet output
	$cgi->param(-name=>'spreadsheet_input',
		   -value=>$output_file);
    
}

sub subtract_op {
    my ($var1,$var2) = @_;
    return $var1 - $var2;
}
 
sub add_op {
    my ($var1,$var2) = @_;
    return $var1 + $var2;
}
 
sub multiply_op {
    my ($var1,$var2) = @_;
    return $var1 * $var2;
}
 
sub divide_op {
    my ($var1,$var2) = @_;
    return $var1 / $var2;
}

sub log2_op {
    my ($var1) = @_;
    return &log_base(2,$var1);
}

sub exp2_op {
    my ($var1) = @_;
    return 2**$var1;
}
 


sub generate_html_output {

#    print STDOUT $cgi -> header().&html_head("$progname - Spreadsheet, ERROR");
#    print STDOUT "<BR><BR><B>DUMP:</B><BR>".$cgi->Dump();
#    exit;

    my $show_command = 0;
#    $show_command = 1;
    if ($show_command) {
	warn "#KH: '$cmd_line'\n";
	open (TMP, ">/tmp/cmd");
	print TMP "perl -d /home/khokamp/FPMI/arraypipe/arraypipe.pl '$cmd_line'\n";
	close TMP;
	exit;
    }
    
    
    my $output_file_title = '';
    my $lines = 0;
    my $new_lines = 0;
    my $file_counter = 1;
    my %hidden_fields = ();
    if ($submit =~ /Spreadsheet (\d+)/) {
	$file_counter = $1;
    } else {
	$file_counter = '';
    }
    my ($input_file, $data_dir);
    if (defined $submit
	and
	$submit eq 'Load File!') {
	if (defined $cgi->param('table_upload_file')
	    and
	    $cgi->param('table_upload_file') =~ /\w/) {
	    my $upload_file = &upload_1_file($cgi, $cgi->param('table_upload_file'), $tmp_dir,'');
	    # decompress if necessary
	    ($input_file) = &unpack_archive($upload_file);
	} elsif ($cgi->param('original_input_file') =~ /\w/) {
	    $input_file = $cgi->param('original_input_file');
	    if ($FPMI_specific) {
		if ($input_file =~ /$example_dir/) {
		    my $tmp_file = $input_file;		    
		    $input_file = $tmp_dir.'/'.(basename $input_file);
#		    warn "cp $tmp_file $input_file";
		    system "cp '$tmp_file' '$input_file'";
		}
	    }
	} else {
	    die "No file uploaded!";
	}
	$data_dir = $tmp_dir;
    } elsif ($cgi->param('spreadsheet_input') =~ /\w/)  {
	$input_file = $cgi->param('spreadsheet_input');
	$data_dir = dirname($input_file);
	warn "# using for input: $input_file";
    } else {
	# started directly from ArrayPipe output
	# compile a file if necessary
	
	# submit options are Simple or Extended Spreadsheet View

	# check which files are available from output:
	unless (defined $cgi->param('output_info')
		and
		$cgi->param('output_info') =~ /\w/) {
	    die "No files available!";
	}

	my %output_available = ();
	foreach ($cgi->param("output_info")) {
	    my @tmp = split /\|/, $_;
	    my $out_file = shift @tmp;

	    foreach (@tmp) {
		# don't include MEV files in output:
#		next if ($_ =~ /\.mev$/
#			 or
#			 $_ =~ /\.ann$/);
		
		$output_available{$_} = $out_file;
	    }
	}

	# which files are requested?
	unless (defined $cgi->param("spreadsheet_files")
		and
		$cgi->param("spreadsheet_files") =~ /\w/) {
	    die "No input specified!";
	}
	my @input_files = ();
	foreach ($cgi->param("spreadsheet_files")) {
	    if ($_ eq 'all') {
		@input_files = values %output_available;
		last;
	    } else {
		push @input_files, $output_available{$_};
	    }
	}
	# remove duplicates from @input_files
	my %input_tmp = ();
	foreach (@input_files) {
	    # don't include MEV files in output:
	    next if ($_ =~ /\.mev$/
		     or
		     $_ =~ /\.ann$/);
	    $input_tmp{$_}++;
	}
	@input_files = sort keys %input_tmp;

	unless (@input_files) {
	    die "No files selected!";
	}

	# compile output of split input, depending on what is provided and what is selected:
	my $input_type = '';
	if (defined $cgi->param("output_type")
	    and
	    $cgi->param("output_type") =~ /\w/) {
	    $input_type = $cgi->param("output_type");
	}

	# check which fields are to be skipped
	my %selected_fields = ();
	if (defined $cgi->param("spreadsheet_content")) {
	    my @fields =  $cgi->param("spreadsheet_content");
	    foreach (@fields) {
		$selected_fields{$_}++;
	    }
	}
	if (defined $cgi->param("spreadsheet_hidden_content")) {
	    my @fields =  $cgi->param("spreadsheet_hidden_content");
	    foreach (@fields) {
#		$selected_fields{$_}++;
		$hidden_fields{$_}++;
	    }
	}
	if ($submit =~ /Quick/i
	    or
	    $submit =~ /Simple/i) {
	    my $annotation_available = $cgi->param('spreadsheet_annotation');
	    %selected_fields = ();
	    if ($annotation_available) {
		$selected_fields{'gene info'}++;
		$hidden_fields{'probe and spot ids'}++;
	    } else {
		$selected_fields{'probe and spot ids'}++;
	    }
	    
	    $selected_fields{'significance values'}++;

	    %hidden_fields = ();
	    if ($FPMI_specific) {
		if ($remote_address =~ /^128.233/
		    or
		    $remote_address =~ /^142.165/) {
		    $selected_fields{'ratios'}++;
		    $hidden_fields{'fold changes'}++;		
		} else {
		    $selected_fields{'fold changes'}++;
		}
	    } else {
		$selected_fields{'fold changes'}++;
	    }
	    $hidden_fields{'ratios'}++;
	    $hidden_fields{'all annotations'}++;
	    $hidden_fields{'spot coordinates'}++;
	    $hidden_fields{'probe and spot ids'}++;
	}

	# the rest is the same for Simple or Extended Spreadsheet view:

	my $merge_field = 'position';
	$input_file = &combine_files(\%selected_fields, \%hidden_fields, $merge_field, @input_files);
	$data_dir = dirname($input_file);
	if (@input_files == 1) {
	    $output_file_title = basename $input_files[0];
	    if ($output_file_title =~ /(.*)_$progname/) {
		$output_file_title = $1;
	    } elsif ($output_file_title =~ /(.*)_step\d+/) {
		$output_file_title = $1;
	    } elsif ($output_file_title =~ /^$MERGED_FILE_NAME(.*?)_(.*?)_/) {
		$output_file_title = "$1 $2";
	    }
	    $output_file_title = " for $output_file_title";
	}
    }

#    warn "# KH: '$input_file', '$data_dir'";
#    exit;

    if (defined $cgi->param('output_file_title')) {
	$output_file_title = $cgi->param('output_file_title');
    }

    my $html_ending = "</CENTER>
<BR>
<HR>
<address>automatically generated at ".(localtime)." localtime
<BR>for questions or remarks e-mail <A HREF=\"MAILTO:$operator_email\">$operator_email</A>
</address>
</DIV>
</BODY>
</HTML>\n";

    # get flags
    foreach ($cgi->param('flags')) {
	my ($flag,$num) = split /\|/, $_;
	if ($num > $highest_flag_num) {
	    $highest_flag_num = $num;
	}
	$flag_values{$flag} = $num;
	$flag_values_num{$num} = $flag;
	# keep track of list flags:
	if ($flag =~ /^list\d+/) {
	    $used_flags{$flag}++;
	}
    }

    @all_flags = ();
    @all_flags_num = ();
    foreach (sort { $a <=> $b } keys %flag_values_num) {
	next if ($_ == 0);
	push @all_flags, $flag_values_num{$_};
	
	# sort @all_flags_num in descending order
	unshift @all_flags_num, $_;
    }

    my %skip_headers = ();
    my %skip_group_headers = ();
    foreach ($cgi->param('skip_headers')) {
	next if (/^-- /);
	if (/^group_header (.*)/) {
	    $skip_group_headers{$1}++;;
	} else {
	    $skip_headers{$_}++;
	}
    }
    
    # hidden fields have been updated by &combine_files
    # (to switch from column names to column positions)
    foreach (keys %hidden_fields) {
	$skip_headers{$_}++;
    }

    my %show_headers = ();
    my %show_group_headers = ();
    foreach ($cgi->param('show_headers')) {
	next if (/^-- /);
	if (/^group_header (.*)/) {
	    $show_group_headers{$1}++;;
	} else {
	    $show_headers{$_}++;
	}
    }

    # add all previously selected skip headers
    # unless they are now selected under show headers
    foreach ($cgi->param('skip_headers_all')) {
	next if (defined $show_headers{$_});
	if (/^group_header (.*)/) {
	    next;
	} else {
	    $skip_headers{$_}++;
	}
    }

    # setup input file
    my %skip_flags = ();
    my %prev_flags = ();
    foreach ($cgi->param('skip_flags')) {
	$skip_flags{$_} = 1;
    }
    foreach ($cgi->param('prev_flags')) {
	$prev_flags{$_} = 1;
    }

    # if there is a change of flags to be skipped, then we have to redo the input file
    foreach (keys %skip_flags) {
	$prev_flags{$_}--;
    }
    my $flag_change = '';
    foreach (keys %prev_flags) {
	if ($prev_flags{$_}) {
	    $flag_change = $_;
	    # check here if the change is from a flag to '-'
	    # in which case no flagging needs to be done!
	    if ($_ eq '-'
		and
		$skip_flags{$_} == 1) {
		last;
	    } 
	}
    }
    if ($flag_change eq '-') {
	# just use the original file as input,
	# don't do any flag-filtering
	$input_file =~ s/(\.flagged|\.filtered|\.sorted)//g;
	$flag_change = 0;
    }
    if ($flag_change) {
	# read in original file
	my $input_orig = $input_file;

	if ($input_orig =~ /(\.flagged|\.filtered|\.sorted)/) {
	    # if the file has been processed before,
	    # go back to the original file
	    $input_orig =~ s/(\.flagged|\.filtered|\.sorted)//g;
	}	    

	unless (-r $input_orig) {
	    print STDOUT $cgi->header(-target=>'full_html', -expires=>'-1s');
	    print STDOUT &html_head("$progname - Spreadsheet, ERROR");
	    print STDOUT "Can't read input file ($input_orig): $!\n";
	    print STDOUT $html_ending;
	    exit;
	}

	# and append .flagged suffix
	$input_file = $input_orig.'.flagged';

	open (IN, $input_orig);

	unless (open (OUT, ">$input_file")) {
	    print STDOUT $cgi->header(-target=>'full_html', -expires=>'-1s');
	    print STDOUT &html_head("$progname - Spreadsheet, ERROR");
	    print STDOUT "Can't write file ($input_file): $!\n";
	    print STDOUT $html_ending;
	    exit;
	}
	    
	my @flag_headers = ();

	while (<IN>) {
	    # remove comments (everything following a hash)
	    my ($line,$rest) = split /\#/; 	
	    
	    # skip if nothing's on the line
	    next unless ($line =~ /\w/);

	    # check for 'flag' header
	    unless ($line =~ /$FLAG_CHANNEL_TERM/i) {
		next;
	    }
	    
	    chomp $line;
	    
	    # first line with data -> use it for header
	    my $counter = -1;
	    foreach (split /\t/, $line) {
		$counter++;
		# match flag term at the end only
		# (some columns might start with 'FLAG...'
		# but are actually not flagged)
		if (/$FLAG_CHANNEL_TERM$/) {
		    # check if group header is selected for this
		    my @tmp = split;
		    if (defined $skip_group_headers{"$tmp[0]\*"}
			or
			defined $skip_group_headers{"\*$tmp[-1]"}) {
			$skip_headers{$counter}++;
		    } else {
			push @flag_headers, $counter;
		    }
		}
	    }
	    # write header to output file
	    print OUT "$line\n";	    
	    last;
	}

	unless (@flag_headers) {
	    print STDOUT $cgi->header(-target=>'full_html', -expires=>'-1s');
	    print STDOUT &html_head("$progname - Spreadsheet, ERROR");
	    print STDOUT "No flag columns found in '$_', '$input_orig': $!\n";
	    print STDOUT $html_ending;
	    exit;
	}
	    
#	warn "# KH: flag headers: ".(join ',', @flag_headers);
#	warn "# KH: flags: ".(join ',', keys %skip_flags);
	# go through each line and extract flags
	IN:while (my $line = <IN>) {	    
	    # skip if nothing's on the line
	    next unless ($line =~ /\w/);
	    
	    chomp $line;

	    my @tmp = split /\t/, $line;
	    foreach my $pos (@flag_headers) {
		# don't look at flags from columns that are to be skipped
		next if (defined $skip_headers{$pos});		
		if (defined $tmp[$pos]) {
		    my $flag = $tmp[$pos];
#		    warn "#KH: flag: '$flag'";
		    my @flags = &get_flags($flag,'add_zero');
		    foreach (@flags) {
			if (defined $skip_flags{$flag_values_num{$_}}) {
			    # skip line
			    next IN;
			}
		    }
		}
	    }
	    # if we got here then we can print out the line
	    print OUT "$line\n";
	}
	close OUT;
	close IN;
    }	    
	    
    # check if filtering has changed:
    my $filter_change = '';
    my $filter_change_lists = '';
    my $filter_change_fields = '';
    my %filter_fields = ();
    my @filter_fields = ();
    my @filter_fields_value = ();
    my @filter_fields_absolute = ();
    my @filter_fields_operator = ();
    my @filter_fields_connector = ();
    my $filter_fields_max_default = 5;
    my $max_list_keys = 3;
    my %filter_lists = ();
    my @filter_lists = ();
    my @filter_lists_field = ();
    my @filter_lists_column = ();
    my @filter_lists_match = ();
    my @filter_lists_connector = ();
    my $filter_lists_max_default = 3;
    my $max_filter_keys = 5;
    my $filter_fields_max = $filter_fields_max_default;
    if (defined $cgi->param('filter_fields_max')) {
	$filter_fields_max = $cgi->param('filter_fields_max');
    } 
    if ($filter_fields_max > $max_filter_keys) {
	$filter_fields_max = $max_filter_keys;
    }
    my $filter_lists_max = $filter_lists_max_default;
    if (defined $cgi->param('filter_lists_max')) {
	$filter_lists_max = $cgi->param('filter_lists_max');
    } 
    if ($filter_lists_max > $max_list_keys) {
	$filter_lists_max = $max_list_keys;
    }

    my $prev_filter_lists_max = 0;
    if (defined $cgi->param('prev_filter_lists_max')) {
	$prev_filter_lists_max = $cgi->param('prev_filter_lists_max');
    }
    # check for change in filter lists:
    FILTER:foreach (my $i = 1; $i <= $prev_filter_lists_max; $i++) {
	foreach my $suffix ('', '_column', '_match', '_field', '_connector') {
	    if (defined $cgi->param('prev_filter_list'.$suffix.$i)) {
		unless (defined $cgi->param('filter_list'.$suffix.$i)) {
		    $filter_change_lists = 1;
#			warn "KH: FILTER: $i,$suffix,$filter_change_lists";
		    last FILTER;
		} else {
		    if ($cgi->param('prev_filter_list'.$suffix.$i) ne $cgi->param('filter_list'.$suffix.$i)) {
			$filter_change_lists = 1;
			my $other_vals = 0;
			if ($i == 1
			    and
			    $suffix eq ''
			    and
			    $cgi->param('filter_list'.$i) eq '-') {
			    # filter_change_lists == '-' has special meaning, so we have to
			    # check if there are lists and values specified in the other sections
			    for (my $k = 2; $k <= $filter_lists_max; $k++) {
				if (defined $cgi->param('filter_list'.$k)
				    and
				    $cgi->param('filter_list'.$k) =~ /\w/
				    and
				    defined $cgi->param('filter_list_value'.$k)
				    and
				    $cgi->param('filter_list_value'.$k) =~ /\w/) {
				    $other_vals = 1;
				    last;
				}
			    }
			}
			$filter_change_lists = $cgi->param('filter_list'.$i) unless ($other_vals);

#			warn "KH: FILTER: $i,$suffix,$filter_change_lists";
			last FILTER;
		    }
		}
	    } else {
		if ($suffix eq ''
		    and
		    $cgi->param('filter_list'.$suffix.$i) eq '-') {
		    next FILTER;
		}
		$filter_change_lists = 1;
#			warn "KH: FILTER: $i,$suffix,$filter_change_lists";
		last FILTER;
	    }
	}
    }

    # if filter lists have been added
    # and if they contain values, we have to react to this change
    if ($filter_lists_max > $prev_filter_lists_max) {
	foreach (my $i = $prev_filter_lists_max; $i <= $filter_lists_max; $i++) {
	    if (defined $cgi->param('filter_list'.$i)
		and 
		$cgi->param('filter_list'.$i) =~ /\w/
		and
		$cgi->param('filter_list'.$i) ne '-'
		and
		defined $cgi->param('filter_list_value'.$i)
		and 
		$cgi->param('filter_list_value'.$i) =~ /\w/
		and
		defined $cgi->param('filter_list_operator'.$i)
		and 
		$cgi->param('filter_list_operator'.$i) =~ /\w/) {
		$filter_change_lists = $cgi->param('filter_list'.$i);
		last;
	    }
	}
    }
#			warn "KH: FILTER: $filter_change_lists";
		
	    
    my $used_filter_lists = 0;
    for (my $i = 1; $i <= $filter_lists_max; $i++) {	
	if (defined $cgi->param('filter_list'.$i)
	    and
	    $cgi->param('filter_list'.$i) =~ /\w/) {
	    my $upload_file = $cgi->param('filter_list'.$i);
	    if ($upload_file =~ /^__UPLOADED__(.*)/) {
		$upload_file = "$tmp_dir/$1";
	    } else {
		# this one has just been uploaded
		$upload_file = &upload_1_file($cgi, $cgi->param('filter_list'.$i), $tmp_dir,'');
	    }
	    # decompress if necessary
	    ($upload_file) = &unpack_archive($upload_file);	    
	    push @filter_lists, $upload_file;
	    $filter_lists{$i}{'file'} = $upload_file;
	    my $column = '';
	    if (defined $cgi->param('filter_list_column'.$i)
		and
		$cgi->param('filter_list'.$i) =~ /\w/) {
		$column = $cgi->param('filter_list_column'.$i);

		$filter_lists{$i}{'column'} = $column;
		$used_filter_lists++;
	    }
	    my $match = '';
	    if (defined $cgi->param('filter_list_match'.$i)
		and
		$cgi->param('filter_list'.$i) =~ /\w/) {
		$match = $cgi->param('filter_list_match'.$i);

		$filter_lists{$i}{'match'} = $match;
		$used_filter_lists++;
	    }
	    my $field = '';
	    if (defined $cgi->param('filter_list_field'.$i)
		and
		$cgi->param('filter_list'.$i) =~ /\w/) {
		$field = $cgi->param('filter_list_field'.$i);

		$filter_lists{$i}{'field'} = $field;
		$used_filter_lists++;
	    }
	    my $connector = '';
	    if (defined $cgi->param('filter_list_connector'.$i)
		and
		$cgi->param('filter_list'.$i) =~ /\w/) {
		$connector = $cgi->param('filter_list_connector'.$i);

		$filter_lists{$i}{'connector'} = $connector;
		$used_filter_lists++;
	    }
	    push @filter_lists_column, $column;
	    push @filter_lists_match, $match;
	    push @filter_lists_connector, $connector;
	    push @filter_lists_field, $field;
	}
    }
    if ($used_filter_lists >= $filter_lists_max) {
	$filter_lists_max = $used_filter_lists + 1;
    } else {
	if ($filter_lists_max > $filter_lists_max_default) {
	    my $diff = $filter_lists_max - $used_filter_lists;
	    if ($diff > 1) {
		$filter_lists_max = $used_filter_lists + 1;
	    }
	}
    }
    if ($filter_lists_max > $max_list_keys) {
	$filter_lists_max = $max_list_keys;
    }
    
    # check the filter fields:
    my $prev_filter_fields_max = 0;
    if (defined $cgi->param('prev_filter_fields_max')) {
	$prev_filter_fields_max = $cgi->param('prev_filter_fields_max');
    }
    FILTER:foreach (my $i = 1; $i <= $prev_filter_fields_max; $i++) {
	foreach my $suffix ('', '_value', '_absolute', '_operator', '_connector') {
	    if (defined $cgi->param('prev_filter_field'.$suffix.$i)) {
		unless (defined $cgi->param('filter_field'.$suffix.$i)) {
		    $filter_change_lists = 1;
		    last FILTER;
		} else {
		    if ($cgi->param('prev_filter_field'.$suffix.$i) ne $cgi->param('filter_field'.$suffix.$i)) {
			$filter_change_lists = 1;
			my $other_vals = 0;
			if ($i == 1
			    and
			    $suffix eq ''
			    and
			    $cgi->param('filter_field'.$i) eq '-') {
			    # filter_change_lists == '-' has special meaning, so we have to
			    # check if there are fields and values specified in the other sections
			    for (my $k = 2; $k <= $filter_fields_max; $k++) {
				if (defined $cgi->param('filter_field'.$k)
				    and
				    $cgi->param('filter_field'.$k) =~ /\w/
				    and
				    defined $cgi->param('filter_field_value'.$k)
				    and
				    $cgi->param('filter_field_value'.$k) =~ /\w/) {
				    $other_vals = 1;
				    last;
				}
			    }
			}
			$filter_change_lists = $cgi->param('filter_field'.$i) unless ($other_vals);

#			warn "KH: FILTER: $i,$suffix,$filter_change_lists";
			last FILTER;
		    }
		}
	    } else {
		if ($suffix eq ''
		    and
		    $cgi->param('filter_field'.$suffix.$i) eq '-') {
		    next FILTER;
		}
		$filter_change_lists = 1;
#			warn "KH: FILTER: $i,$suffix,$filter_change_lists";
		last FILTER;
	    }
	}
    }

    # if filter fields have been added
    # and if they contain values, we have to react to this change
    if ($filter_fields_max > $prev_filter_fields_max) {
	foreach (my $i = $prev_filter_fields_max; $i <= $filter_fields_max; $i++) {
	    if (defined $cgi->param('filter_field'.$i)
		and 
		$cgi->param('filter_field'.$i) =~ /\w/
		and
		$cgi->param('filter_field'.$i) ne '-'
		and
		defined $cgi->param('filter_field_value'.$i)
		and 
		$cgi->param('filter_field_value'.$i) =~ /\w/
		and
		defined $cgi->param('filter_field_operator'.$i)
		and 
		$cgi->param('filter_field_operator'.$i) =~ /\w/) {
		$filter_change_lists = $cgi->param('filter_field'.$i);
		last;
	    }
	}
    }
#			warn "KH: FILTER: $filter_change_lists";
		
	    
    my $used_filter_fields = 0;
    for (my $i = 1; $i <= $filter_fields_max; $i++) {	
	if (defined $cgi->param('filter_field'.$i)
	    and
	    $cgi->param('filter_field'.$i) =~ /\w/) {
	    push @filter_fields, $cgi->param('filter_field'.$i);
	    $filter_fields{$i}{'field'}{$cgi->param('filter_field'.$i)}++;
	    my $value = '';
	    if (defined $cgi->param('filter_field_value'.$i)
		and
		$cgi->param('filter_field'.$i) =~ /\w/) {
		$value = $cgi->param('filter_field_value'.$i);

		$filter_fields{$i}{'value'}{$value}++;
		$used_filter_fields++;
	    }
	    my $connector = '';
	    if (defined $cgi->param('filter_field_connector'.$i)
		and
		$cgi->param('filter_field'.$i) =~ /\w/) {
		$connector = $cgi->param('filter_field_connector'.$i);

		$filter_fields{$i}{'connector'}{$connector}++;
		$used_filter_fields++;
	    }
	    my $absolute = '';
	    if (defined $cgi->param('filter_field_absolute'.$i)
		and
		$cgi->param('filter_field'.$i) =~ /\w/) {
		$absolute = $cgi->param('filter_field_absolute'.$i);
		$filter_fields{$i}{'absolute'}{$absolute}++;
		$used_filter_fields++;
	    }
	    my $operator = '';
	    if (defined $cgi->param('filter_field_operator'.$i)
		and
		$cgi->param('filter_field'.$i) =~ /\w/) {
		$operator = $cgi->param('filter_field_operator'.$i);

		$filter_fields{$i}{'operator'}{$operator}++;
		$used_filter_fields++;
	    }
	    push @filter_fields_value, $value;
	    push @filter_fields_operator, $operator;
	    push @filter_fields_absolute, $absolute;
	    push @filter_fields_connector, $connector;
	}
    }
    if ($used_filter_fields >= $filter_fields_max) {
	$filter_fields_max = $used_filter_fields + 1;
    } else {
	if ($filter_fields_max > $filter_fields_max_default) {
	    my $diff = $filter_fields_max - $used_filter_fields;
	    if ($diff > 1) {
		$filter_fields_max = $used_filter_fields + 1;
	    }
	}
    }
    if ($filter_fields_max > $max_filter_keys) {
	$filter_fields_max = $max_filter_keys;
    }
    
    if ($filter_change_lists eq '-') {
	if ($filter_change_lists eq '_') {
	    $filter_change = '-';
	} else {
	    $filter_change = $filter_change_lists;
	}
    } else {
	$filter_change = $filter_change_lists;
    }


    if ($filter_change eq '-') {
	# just use the original (or flag-filtered) file as input,
	# don't do any filtering
	unless ($flag_change) {
	    $input_file =~ s/(\.flagged|\.filtered|\.sorted)//g;
	}
	$filter_change = 0;
    } else {
	if ($flag_change) {
	    # if flag change has occured
	    # we have to do filtering
	    # unless no filter is set
	    $filter_change = 1 unless ($flag_change);
	}
    }

#    warn "KH: FILTER: $filter_change,$flag_change";
    if ($filter_change) {
	# read in original file
	my $input_orig = $input_file;
	if ($input_orig =~ /(\.filtered|\.sorted)/) {
	    # if the file has been processed before,
	    # go back to the original file
	    $input_orig =~ s/(\.filtered|\.sorted)//g;
	}	    
	# and append .filtered suffix
	$input_file = $input_orig.'.filtered';

	unless (-r $input_orig) {
	    print STDOUT $cgi->header(-target=>'full_html', -expires=>'-1s');
	    print STDOUT &html_head("$progname - Spreadsheet, ERROR");
	    print STDOUT "Can't read input file ($input_orig): $!\n";
	    print STDOUT $html_ending;
	    exit;
	}
	open (IN, $input_orig);

	unless (open (OUT, ">$input_file")) {
	    print STDOUT $cgi->header(-target=>'full_html', -expires=>'-1s');
	    print STDOUT &html_head("$progname - Spreadsheet, ERROR");
	    print STDOUT "Can't write file ($input_file): $!\n";
	    print STDOUT $html_ending;
	    exit;
	}
	    
	my %header_positions = ();
	while (<IN>) {
	    # remove comments (everything following a hash)
	    my ($line,$rest) = split /\#/; 	
	    
	    # skip if nothing's on the line
	    next unless ($line =~ /\w/);

	    # check for 'flag' header
	    my @fields = ();
	    if (@filter_lists_field) {
		@fields = @filter_lists_field;
	    } else {
		@fields = @filter_fields;
	    }
	    unless ($line =~ /\Q$fields[0]/i) {
		next;
	    }
	    
	    chomp $line;
	    
	    # first line with data -> use it for header
	    foreach my $field (@filter_fields, @filter_lists_field) {
#		warn "# filter field: '$field'";
		my $counter = -1;
		foreach (split /\t/, $line) {
		    $counter++;
#		    warn "# field: '$_'";
		    if ($_ eq $field) {
			$header_positions{$field} = $counter;
			last;
		    }
		}
	    }
	    # write header to output file
	    print OUT "$line\n";	    
	    last;
	}

	unless (%header_positions) {
	    print STDOUT $cgi->header(-target=>'full_html', -expires=>'-1s');
	    print STDOUT &html_head("$progname - Spreadsheet, ERROR");
	    print STDOUT "No filter column '$filter_fields[0] / $filter_lists_field[0]' found in '$input_orig': $!\n";
	    print STDOUT $html_ending;
	    exit;
	}
	    
	# load data from lists	
	my %filter_list_items = ();
	foreach my $i (sort { $a <=> $b } keys %filter_lists) {	
	    my $file = $filter_lists{$i}{'file'};
	    unless (open (FILTER, $file)) {
		warn "Can't open file '$file': $!\n";
		next;
	    }
	    my $column = $filter_lists{$i}{'column'} - 1;
	    my $linearize_list = 1;
	    while (<FILTER>) {
		chomp;
		s/\cM//;
		my @tmp = split /\t/, $_;
		if (defined $tmp[$column]) {
		    my $item = $tmp[$column];
		    my @items = ($item);
		    if ($linearize_list) {
			@items = ();
			foreach (split /; /, $item) {
			    push @items, $_;
			}
		    }
		    foreach my $item (@items) {
			$filter_list_items{$i}{$item}++;
		    }
		}
	    }
	    close FILTER;
	}

	# go through each line and extract values
	while (my $line = <IN>) {	    
	    # skip if nothing's on the line
	    next unless ($line =~ /\w/);
	    
	    chomp $line;

	    my @tmp = split /\t/, $line;
	    my $prev_result = '';
	    my $result = '';
	    my $first_i = (sort { $a <=> $b } keys %filter_lists)[0];
	    my $last_i = '';
	    my $last_i_lists = '';
	    foreach my $i (sort { $a <=> $b } keys %filter_lists) {		
		my $header = $filter_lists{$i}{'field'};
		unless (defined $header_positions{$header}) {
		    warn "# No position for header '$header'!";
		    next;
		}
		my $pos = $header_positions{$header};
		my $value = $tmp[$pos];
		my $match = $filter_lists{$i}{'match'};
		$result = 0;
		if ($match eq 'matches exactly') {
		    if (defined $filter_list_items{$i}{$value}) {
			$result = 1;
		    } 
		} elsif ($match eq 'does not match') {
		    unless (defined $filter_list_items{$i}{$value}) {
			$result = 1;
		    }
		} else {
		    my $found_match = 0;
		    foreach my $item (keys %{$filter_list_items{$i}}) {
			if ($value =~ /\Q$item/) {
			    $found_match++;
			    if ($match eq 'is contained in') {
				$result = 1;
				last;
			    }
			}
		    }
		    if ($match eq 'is not contained in'
			and 
			$found_match == 0) {
			$result = 1;
		    }
		}
		if ($i == $first_i) {
		    $prev_result = $result;
		    $last_i = $i;
		    $last_i_lists = $i;
		    next;
		}
		
		my $connector = $filter_lists{$last_i}{'connector'};
		$last_i = $i;
		$last_i_lists = $i;
		if ($connector eq 'AND') {
		    $result = $prev_result && $result;
		} elsif ($connector eq 'OR') {
		    $result = $prev_result || $result;
		} else {
		    if ($prev_result
			and
			$result == 0) {
			$result = 1;
		    } else {
			$result = 0;
		    }
		}
		# if result is 0 after connecting two filters
		# get out of loop
		# WRONG - if the next operator is 'OR' then we could still continue!
#		last if ($result == 0);

		# keep copy of result for next round
		$prev_result = $result;
	    }


	    # now check the filter fields:

	    my $first_reverse = 0;
	    unless (%filter_lists) {
		# even though there are no lists reported, 
		# we might have a connector set to 'BUTNOT' to reverse the match:

		# doesn't work because without a list upload we don't get the connector values!

#		for (my $i = 1; $i <= $filter_lists_max; $i++) {	
#		    if (defined $filter_lists{$i}
#			and
#			defined $filter_lists{$i}{'connector'}
#			and
#			$filter_lists{$i}{'connector'} eq 'BUTNOT') {
#			$first_reverse = 1;
#			last;
#		    }
#		}
		$first_i = (sort { $a <=> $b } keys %filter_fields)[0];
		$last_i = '';
	    }

	    foreach my $i (sort { $a <=> $b } keys %filter_fields) {		
		my $header = (keys %{$filter_fields{$i}{'field'}})[0];
		unless (defined $header_positions{$header}) {
		    warn "# No position for header '$header'!";
		    next;
		}
		my $pos = $header_positions{$header};
		my $value = $tmp[$pos];
		my $absolute = (keys %{$filter_fields{$i}{'absolute'}})[0];
		my $operator = (keys %{$filter_fields{$i}{'operator'}})[0];
		my $comparison = (keys %{$filter_fields{$i}{'value'}})[0];
		$result = &check_data_simple($value,$absolute,$operator,$comparison);
#		warn "$result,$value,$operator,$comparison";
		# ( returns either 0 or 1)
		my $connector = '';
		if ($i == $first_i) {
		    unless (%filter_lists) {
			# even though there are no lists reported, 
			# we might have a connector set to 'BUTNOT' to reverse the match:
			if ($first_reverse) {
			    if ($result == 0) {
				$result = 1;
			    } else {
				$result = 0;
			    }
			}
			$prev_result = $result;
			$last_i = $i;
			next;
		    } else {
			# use connector from last uploaded list
			$connector = $filter_lists{$last_i_lists}{'connector'};
		    }
		} else {
		    $connector = (keys %{$filter_fields{$last_i}{'connector'}})[0];
		}
		$last_i = $i;
		if ($connector eq 'AND') {
		    $result = $prev_result && $result;
		} elsif ($connector eq 'OR') {
		    $result = $prev_result || $result;
		} else {
		    if ($prev_result
			and
			$result == 0) {
			$result = 1;
		    } else {
			$result = 0;
		    }
		}
		# if result is 0 after connecting two filters
		# get out of loop
		# WRONG - if the next operator is 'OR' then we could still continue!
#		last if ($result == 0);

		# keep copy of result for next round
		$prev_result = $result;
	    }
	    next unless ($result);

	    # if we got here then we can print out the line
	    print OUT "$line\n";
#	    warn "$line";
	}
	close OUT;
	close IN;
    }	    


# 	warn "last line";
#	    exit;
  # check if sorting has changed:
    my %sort_fields = ();
    my @sort_fields = ();
    my @sort_fields_direction = ();
    my $sort_fields_max_default = 5;
    my $max_sort_keys = 5;
    my $sort_fields_max = $sort_fields_max_default;
    if (defined $cgi->param('sort_fields_max')) {
	$sort_fields_max = $cgi->param('sort_fields_max');
    } 
    if ($sort_fields_max > $max_sort_keys) {
	$sort_fields_max = $max_sort_keys;
    }

    # if there is a change of sorting parameters, then we have to redo the input file
    my $sort_change = 0;
    my $prev_sort_fields_max = 0;
    if (defined $cgi->param('prev_sort_fields_max')) {
	$prev_sort_fields_max = $cgi->param('prev_sort_fields_max');
    }
    foreach (my $i = 1; $i <= $prev_sort_fields_max; $i++) {
	if (defined $cgi->param('prev_sort_field'.$i)) {
	    unless (defined $cgi->param('sort_field'.$i)) {
		$sort_change = 1;
		last;
	    } else {
		if ($cgi->param('prev_sort_field'.$i) ne $cgi->param('sort_field'.$i)) {
		    $sort_change = 1;
		    $sort_change = $cgi->param('sort_field'.$i) if ($i == 1);
		    last;
		}
	    }
	} else {
	    $sort_change = 1;
	    last;
	}
	# check for change of direction
	if (defined $cgi->param('prev_sort_field_direction'.$i)) {
	    unless (defined $cgi->param('sort_field_direction'.$i)) {
		$sort_change = 1;
		last;
	    } else {
		if ($cgi->param('prev_sort_field_direction'.$i) ne $cgi->param('sort_field_direction'.$i)) {
		    $sort_change = 1;
		    last;
		}
	    }
	} else {
	    $sort_change = 1;
	    last;
	}
    }
    if ($sort_fields_max > $prev_sort_fields_max) {
	foreach (my $i = $prev_sort_fields_max; $i <= $sort_fields_max; $i++) {
	    if (defined $cgi->param('sort_field'.$i)
		and 
		$cgi->param('sort_field'.$i) =~ /\w/) {
		$sort_change = 1;
		last;
	    }
	}
    }
		
	    
    my $used_sort_fields = 0;
    for (my $i = 1; $i <= $sort_fields_max; $i++) {	
	if (defined $cgi->param('sort_field'.$i)
	    and
	    $cgi->param('sort_field'.$i) =~ /\w/) {
	    push @sort_fields, $cgi->param('sort_field'.$i);
	    $sort_fields{$i}{'field'}{$cgi->param('sort_field'.$i)}++;
	    my $direction = 'ascending';
	    if ($cgi->param('sort_field_direction'.$i)) {
		$direction = $cgi->param('sort_field_direction'.$i);
	    }
	    $sort_fields{$i}{'direction'}{$direction}++;
	    push @sort_fields_direction, $direction;
	    $used_sort_fields++;
	} else {
	    # set default sort direction to descending:
	    $sort_fields{$i}{'direction'}{'descending'}++;
	}
    }
    if ($used_sort_fields >= $sort_fields_max) {
	$sort_fields_max = $used_sort_fields + 1;
    } else {
	if ($sort_fields_max > $sort_fields_max_default) {
	    my $diff = $sort_fields_max - $used_sort_fields;
	    if ($diff > 1) {
		$sort_fields_max = $used_sort_fields + 1;
	    }
	}
    }
    if ($sort_fields_max > $max_sort_keys) {
	$sort_fields_max = $max_sort_keys;
    }
    
    if ($sort_change eq '-') {
	# just use the original file as input,
	# don't do any sorting
	$input_file =~ s/\.sorted//g;
	$sort_change = 0;
    } else {
 	if ($flag_change
	    or
	    $filter_change) {
	    # we have to sort again
	    $sort_change = 1;
	}
    }


    if ($sort_change
	and
	$used_sort_fields) {
	# read in original file
	my $input_orig = $input_file;
	if ($input_orig =~ /\.sorted/) {
	    $input_orig =~ s/\.sorted//g;
	}
	unless (-r $input_orig) {
	    print STDOUT $cgi->header(-target=>'full_html', -expires=>'-1s');
	    print STDOUT &html_head("$progname - Spreadsheet, ERROR");
	    print STDOUT "Can't read input file ($input_orig): $!\n";
	    print STDOUT $html_ending;
	    exit;
	}
	open (IN, $input_orig);

	$input_file = $input_orig.'.sorted';

	unless (open (OUT, ">$input_file")) {
	    print STDOUT  $cgi->header(-target=>'full_html', -expires=>'-1s');
	    print STDOUT &html_head("$progname - Spreadsheet, ERROR");
	    print STDOUT "Can't write file ($input_file): $!\n";
	    print STDOUT $html_ending;
	    exit;
	}
	    
	my %headers = ();
	while (<IN>) {
	    # remove comments (everything following a hash)
	    my ($line,$rest) = split /\#/; 	
	    
	    # skip if nothing's on the line
	    next unless ($line =~ /\w/);

	    chomp $line;
	    
	    # first line with data -> use it for header
	    my $counter = -1;
	    foreach (split /\t/, $line) {
		$counter++;
		$headers{$_} = $counter;
	    }
	    # write header to output file
	    print OUT "$line\n";
	    last;
	}

	unless (%headers) {
	    print STDOUT $cgi->header(-target=>'full_html', -expires=>'-1s');
	    print STDOUT &html_head("$progname - Spreadsheet, ERROR");
	    print STDOUT "No headers found in '$input_orig': $!\n";
	    print STDOUT $html_ending;
	    exit;
	}
	    
	# go through each line and extract sort headers
	$counter = 0;
	my @lines = ();
	my %sort = ();
	my %sort_type = ();
	foreach (my $i = 0; $i <= $#sort_fields; $i++) {
	    $sort_type{$i} = 'num';
	}

	IN:while (my $line = <IN>) {	    
	    # skip if nothing's on the line
	    next unless ($line =~ /\w/);
	    
	    chomp $line;

	    push @lines, $line;

	    # get values for sort headers
	    my @tmp = split /\t/, $line;
	    my @vals = ();
	    foreach (my $i = 0; $i <= $#sort_fields; $i++) {
		my $sort_header = $sort_fields[$i];
		my $val = $tmp[$headers{$sort_header}];
		if ($val =~ /[a-z]/i) {
		    $sort_type{$i} = 'text' unless ($val =~ /\d+e\-\d/);
		}
		push @vals, $val;
	    }

	    for (my $i = 0; $i <= $max_sort_keys; $i++) {
		unless (defined $vals[$i]) {
		    $vals[$i] = 0;
		}
	    }

	    push @{$sort{$vals[0]}{$vals[1]}{$vals[2]}{$vals[3]}{$vals[4]}}, $counter;

	    $counter++;
	}
	close IN;

	# print the data lines in order
	my @val0 = ();
	if (defined $sort_type{0}
	    and $sort_type{0} eq 'num') {
	    @val0 = sort { $a <=> $b } keys %sort;
	} else {
	    @val0 = sort keys %sort;
	}
	if (defined $sort_fields_direction[0] 
	    and
	    $sort_fields_direction[0] eq 'descending') {
	    @val0 = reverse @val0;
	}
	my @val1 = ();
	foreach my $val0 (@val0) {
	    if (defined $sort_type{1}
		and $sort_type{1} eq 'num') {
		@val1 = sort { $a <=> $b } keys %{$sort{$val0}};
	    } else {
		@val1 = sort keys %{$sort{$val0}};
	    }
	    if (defined $sort_fields_direction[1] 
		and
		$sort_fields_direction[1] eq 'descending') {
		@val1 = reverse @val1;
	    }

	    my @val2 = ();
	    foreach my $val1 (@val1) {
		if (defined $sort_type{2}
		    and $sort_type{2} eq 'num') {
		    @val2 = sort { $a <=> $b } keys %{$sort{$val0}{$val1}};
		} else {
		    @val2 = sort keys %{$sort{$val0}{$val1}};
		}
		if (defined $sort_fields_direction[2] 
		    and
		    $sort_fields_direction[2] eq 'descending') {
		    @val2 = reverse @val2;
		}

		my @val3 = ();
		foreach my $val2 (@val2) {
		    if (defined $sort_type{3}
			and $sort_type{3} eq 'num') {
			@val3 = sort { $a <=> $b } keys %{$sort{$val0}{$val1}{$val2}};
		    } else {
			@val3 = sort keys %{$sort{$val0}{$val1}{$val2}};
		    }
		    if (defined $sort_fields_direction[3] 
			and
			$sort_fields_direction[3] eq 'descending') {
			@val3 = reverse @val3;
		    }
		    my @val4 = ();
		    foreach my $val3 (@val3) {
			if (defined $sort_type{4}
			    and $sort_type{4} eq 'num') {
			    @val4 = sort { $a <=> $b } keys %{$sort{$val0}{$val1}{$val2}{$val3}};
			} else {
			    @val4 = sort keys %{$sort{$val0}{$val1}{$val2}{$val3}};
			}
			if (defined $sort_fields_direction[4] 
			    and
			    $sort_fields_direction[4] eq 'descending') {
			    @val4 = reverse @val4;
			}
			foreach my $val4 (@val4) {
			    my @pos = sort { $a <=> $b } @{$sort{$val0}{$val1}{$val2}{$val3}{$val4}};
			    foreach my $pos (@pos) {
				print OUT "$lines[$pos]\n";
			    }
			}
		    }
		}
	    }
	}
	close OUT;
    }
    
    # font for table
    my $table_font_default = 12;
    my $table_font = $table_font_default;
    if (defined $cgi->param('table_font')) {
	$table_font = $cgi->param('table_font');
    }
    # change font
    $css_style =~ s/td \{ font-size\: (\d+)px/td \{ font-size\: ${table_font}px/s;
    $css_style =~ s/th \{ font-size\: (\d+)px/th \{ font-size\: ${table_font}px/s;

    my $step = 1;
    if (defined $cgi->param('step')) {
	$step = $cgi->param('step');
    }
    my $extension = 0;
    my $appendix = $step;
    while (-e "$input_file.$appendix.html") {
	$extension++;
	$appendix = "$step.$extension";
    }
    my $html_output = "$input_file.$appendix.html";	    

    my $file_output = $html_output;

    $html_output =~ s/$tmp_dir/$tmp_dir_html/;

    unless (open (HTML, ">$file_output")) {
	print STDOUT $cgi->header(-target=>'full_html', -expires=>'-1s');
	print STDOUT &html_head("$progname - Spreadsheet, ERROR");
	print STDOUT "Can't write html file ('$file_output'): $!\n";
	print STDOUT $html_ending;
	exit;
    }
    system "chgrp $common_user_group '$file_output' 2> /dev/null";
    system "chmod 666 '$file_output' 2> /dev/null";	

    select HTML;


#    print $cgi->header(-target=>'full_html', -expires=>'-1s');
    print &html_head("$progname - Web Spreadsheet, step $step");

    print "<DIV>\n";

    # start FORM
    my $url_executable = $cgi -> url();
    
    unless ($webstart) {
	$url_executable = $cgi_link;
	if (defined $cgi->param('executable')
	    and
	    $cgi->param('executable')) {
	    $url_executable = $cgi->param('executable');
	}
    }
    print "\n".($cgi -> start_multipart_form({action => $url_executable, name => 'full_html'}))."\n";

    print "<H1>ArrayPipe ";
    if ($FPMI_specific
	and
	$progname =~ /FPMI/i) {
	print "<FONT COLOR=RED>FPMI</FONT> ";
    }
    
    print "Table Generator</H1>\n<TABLE><TR><TD>".(&pointer)."</TD><TD>Carry out spreadsheet operations on the pipeline output through the Web.\n<BR>Use the forms <A HREF=\"#table_settings_section\">at the bottom</A> for filtering, sorting, extra tools and settings.<BR>Click on "
	. "<TABLE CELLPADDING=5><TR><TD BGCOLOR=RED>"
	. $cgi -> submit(-name=>'submit', 
			 -value=>'Reprint Table!')
	. '</TD></TR></TABLE>'
	. " after making any changes.</TD></TR></TABLE>\n";

    my $prev_page = '';
    if (defined $cgi->param('prev_page')) {
	$prev_page = $cgi->param('prev_page');
    } else {
	$prev_page = $referral_address;
    }
	    
    print "<BR><A HREF=\"$prev_page\">[back to ArrayPipe results page]</A>, <A HREF=\"$prev_page\" TARGET=\"_new\">[in a new window]</A><BR>\n";
    
    my $output_collect = '';

#    warn "KH: opening $input_file";
    unless (open (SHEET_DATA, $input_file)) {
	$output_collect .= "ERROR: Couldn't open file: $!\n$html_ending";
	exit;
    }

    my $header = '';
    my $header_no_tab = '';
    my @headers = ();
    my @bg_colours = ('WHITE', 'LIGHTGREY');

    my %header_group_start = ();
    my %header_group_end = ();

    # get the header first:

    while (<SHEET_DATA>) {

	# remove comments (everything following a hash)
	my ($line,$rest) = split /\#/; 	

	# skip if nothing's on the line
	next unless ($line =~ /\w/);

	chomp $line;

	# first line with data -> use it for header
	my $header_counter = -1;
	foreach (split /\t/, $line) {
	    $header_counter++;
	    push @headers, $_;
	
	    next if (defined $skip_headers{$header_counter});

	    # check if group header is selected for this
	    my @tmp = split;
	    if (defined $skip_group_headers{"$tmp[0]\*"}
		or
		defined $skip_group_headers{"\*$tmp[-1]"}) {
		$skip_headers{$header_counter}++;
		next;
	    }

	    # get first and last word to compile header groups
	    # from those headers that are shown
	    $header_group_start{$tmp[0]}++;
	    $header_group_end{$tmp[-1]}++;
	}
	last;
    }
	
    my @group_headers = ();
    foreach (sort keys %header_group_start) {
	if ($header_group_start{$_} > 1) {
	    push @group_headers, "$_*";
	}
    }
    foreach (sort keys %header_group_end) {
	if ($header_group_end{$_} > 1) {
	    push @group_headers, "*$_";
	}
    }

    # parameters for organization of pages
    my $warnings = '';
    my $line_counter = 0;
    my $line_print_counter = 0;
    my $last_header = 1;
    my $page_counter = 0;
    my $line_start = 1;
    my $line_numbers = 20;
    my $chunk_size = 50;
    my $header_repeat = '';

    if (defined $cgi->param('table_window_start')) {
	$line_start = $cgi->param('table_window_start');
    }
    if (defined $cgi->param('line numbers')) {
	$line_numbers = $cgi->param('line numbers');
    }

    # check if one of the navigation buttons was pressed:
    my $orig_input = $input_file;
    $orig_input =~ s/(\.flagged|\.filtered|\.sorted)//g;    
    $lines = `grep -c -v '^#' '$orig_input'`;
    chomp $lines;
    $lines--;
    unless ($orig_input eq $input_file) {
	$new_lines = `grep -c -v '^#' '$input_file'`; 
	chomp $new_lines;
	$new_lines--;
    } else {
	$new_lines = $lines;
    }

    my $shift = 0;
    if ($submit =~ /[>|<]/) {
	$shift = $cgi->param('window_slide') || $line_numbers;
	$line_numbers = $shift;	
    }

    if ($submit eq '|<') {
	$line_start = 1;
#	$line_end = $line_start + $shift - 1;
    } elsif ($submit eq '<') {
	$line_start -= $shift;
#	$line_end -= $shift;
    }

    if ($line_start < 1) {
	$warnings .= "Invalid line start ($line_start), adjusted to 1!<BR>";
	$line_start = 1;
    }

    if ($new_lines > 1
	and
	$line_start >= $new_lines) {
	$warnings .= "Not enough input lines for start at $line_start, adjusted to 1!<BR>";
	$line_start = 1;
    }

    if ($line_numbers =~ /\D/
	or
	$line_numbers < 0) {
	$warnings .= "Invalid number of data rows ($line_numbers), adjusted to 10!<BR>";
	$line_numbers = 10;
    }

    if ($submit eq '>|') {
	$line_start = $new_lines - $shift + 1;
#	$line_end = $new_lines;
    } elsif ($submit eq '>') {
	$line_start += $shift;
#	$line_end += $shift;
    }

    if (defined $cgi->param('chunk size')) {
	$chunk_size = $cgi->param('chunk size');
	if ($chunk_size < 1) {
	    $warnings .= "Invalid chunk size ($chunk_size), adjusted to 1!<BR>";
	    $chunk_size = 1;
	}
    }
    if (defined $cgi->param('header repeat')
	and
	$cgi->param('header repeat') ne '') {
	$header_repeat = $cgi->param('header repeat');
	if ($header_repeat < 1) {
	    $warnings .= "Invalid header repeat ($header_repeat), adjusted to 1!<BR>";
	    $header_repeat = 1;
	}
    }

    if ($warnings) {
	print "<H2>Warnings</H2><FONT COLOR=RED>$warnings</FONT><BR><BR>\n";
    }

    # preference menu
    $output_collect .= "\n<H2><A NAME=\"table_settings_section\">Settings</A></H2>\n".(&pointer)."Click on green labels or ".&help('popup_boxes')."for more information.<BR><BR>";	

    $output_collect .= "<TABLE>\n";

    # menu for hiding columns
    $output_collect .= "<TR><TD><B>hide columns:</B>".&help('hide_columns')."<BR><SELECT name=\"skip_headers\" multiple size=8><OPTION VALUE=->-</OPTION>\n";

    if (@group_headers) {
#	$output_collect .= "<OPTION VALUE=\"-- skip1 --\"></OPTION>\n";
	$output_collect .= "<OPTION VALUE=\"-- group headers --\">-- group headers --</OPTION>\n";
	foreach (@group_headers) {
	    $output_collect .= "<OPTION VALUE=\"group_header $_\">$_</OPTION>\n";
	}
	$output_collect .= "<OPTION VALUE=\"-- skip2 --\"></OPTION>\n";
	$output_collect .= "<OPTION VALUE=\"-- individual headers --\">-- individual headers --</OPTION>\n";
    }
    
    my $header_counter = 0;
    foreach (@headers) {	
	unless (defined $skip_headers{$header_counter}) {	    
	    $output_collect .= "<OPTION VALUE=$header_counter>$_</OPTION>\n";
	}
	$header_counter++;
    }
    $output_collect .= "</SELECT></TD><TD>&nbsp;</TD>\n";

    if (%skip_headers
	or
	%skip_group_headers) {
	$output_collect .= "<TD><B>show columns:</B>".&help('show_columns')."<BR><SELECT name=\"show_headers\" multiple size=8><OPTION VALUE=->-</OPTION>\n";
	
	if (%skip_group_headers) {
	    $output_collect .= "<OPTION VALUE=\"-- group headers --\">-- group headers --</OPTION>\n";
	    foreach (sort keys %skip_group_headers) {
		$output_collect .= "<OPTION VALUE=\"group_header $_\">$_</OPTION>\n";
	    }
	    $output_collect .= "<OPTION VALUE=\"-- skip2 --\"></OPTION>\n";
	    $output_collect .= "<OPTION VALUE=\"-- individual headers --\">-- individual headers --</OPTION>\n";
	}
	
	my $header_counter = 0;
	foreach (@headers) {	
	    if (defined $skip_headers{$header_counter}) {
		$output_collect .= "<OPTION VALUE=$header_counter>$_</OPTION>\n";
	    }
	    $header_counter++;
	}
	$output_collect .= "</SELECT></TD><TD>&nbsp;</TD>\n";
    }	

    # menu for adding links
    $output_collect .= "<TD><B>add links:</B>".&help('add_links')."<BR><SELECT name=\"add_links\" multiple size=8><OPTION VALUE=-2>none</OPTION>\n";
    my %add_links = ();
    foreach ($cgi->param('add_links')) {
	$add_links{$_}++;
    }
       

    $header_counter = -1;
    my $all_option = 0;
    unless (defined $cgi->param('add_links')) {
	# first call
	# preselect 'all'
	$add_links{-1}++;
    }

    foreach (@headers) {
	$header_counter++;
	# skip columns that were selected to be skipped
	next if (defined $skip_headers{$header_counter});
	# skip columns that don't have annotation:
	next unless (defined $annotation_links{$_}
		     or
		     defined $annotation_links_alternatives{$_});

	# here we have a specific annotation field that could have the link option enabled:
	if (defined $add_links{-1}) {
	    # everything's supposed to have links:
	    # $output_collect .= selected 'all' option
	     unless ($all_option) {
		 $output_collect .= "<OPTION VALUE=-1 SELECTED>all</OPTION>\n";
		 $all_option = 1;
	     }
	     $output_collect .= "<OPTION VALUE=$header_counter>$_</OPTION>\n";
	     $add_links{$header_counter}++;
	 } else {
	    # $output_collect .= 'all' option
	     unless ($all_option) {
		 $output_collect .= "<OPTION VALUE=-1>all</OPTION>\n";
		 $all_option = 1;
	     }
	     my $select = '';
	     if (defined $add_links{$header_counter}) {
		 $select = ' SELECTED';
		 $add_links{$header_counter}++;
	     }		
	     $output_collect .= "<OPTION VALUE=$header_counter$select>$_</OPTION>\n";
	 }
    }
       
    
    if (defined $add_links{-1}) {
	undef $add_links{-1};
    }

    $output_collect .= "</SELECT></TD><TD>&nbsp;</TD>\n";	

    # menu for skipping flags
    my $skip_flags_help_text = "To hide data from spots with certain flags (or no flag) from the output, select the flags to be skipped. Their numerical representation in the table output is as follows:<BR><TABLE>\n";
    foreach my $flag ('no flag', @all_flags) {	
	my $num = $flag_values{$flag};
	$skip_flags_help_text .= "<TR><TD>$flag:</TD><TD ALIGN=RIGHT>$num</TD></TR>\n";
    }
    $skip_flags_help_text .= "</TABLE>If a spot has multiple flags assigned, it will show the sum of their values.\n";

    $output_collect .= "<TD><B>skip flags:</B>".&help('skip_flags', '', '', $skip_flags_help_text)."<BR><SELECT name=\"skip_flags\" multiple size=8><OPTION>-</OPTION>\n";

    foreach my $flag ('no flag', @all_flags) {	
	my $select = '';
	if (defined $skip_flags{$flag}) {
	    $select = ' SELECTED';
	}
	$output_collect .= "<OPTION$select>$flag</OPTION>\n";
    }
    $output_collect .= "</SELECT></TD><TD>&nbsp;</TD>\n";

    # menu for cutting table elements down to size
    if (defined $cgi->param('table_element_size')) {
	$table_element_size_limit = $cgi->param('table_element_size');
    } 
    if (defined $cgi->param('table_element_excess')) {
	$table_element_size_excess = $cgi->param('table_element_excess');
    }

    $output_collect .= "<TD><B>reduce text size:</B>".&help('reduce_table_element_size')."<BR>
Limit: <INPUT TYPE=TEXT name=\"table_element_size\" SIZE=\"3\" VALUE=\"$table_element_size_limit\">
<BR>Excess: <SELECT name=\"table_element_excess\">\n";
    foreach ('remove', 'hide') {	
	my $select = '';
	if ($table_element_size_excess eq $_) {
	    $select = ' SELECTED';
	}
	$output_collect .= "<OPTION$select>$_</OPTION>\n";
    }
    $output_collect .= "</SELECT></TD><TD>&nbsp;</TD>\n";


    # menu for rounding
    my $table_rounding = '-';
    if (defined $cgi->param('table_rounding')) {
	$table_rounding = $cgi->param('table_rounding');
    }

    $output_collect .= "<TD><B>rounding:</B>".&help('rounding')."<BR><SELECT name=\"table_rounding\">\n";
    
    foreach my $num ('-', (0..15)) {	
	my $select = '';
	if ($table_rounding =~ /-/) {
	    if ($num =~ /-/) {
		$select = ' SELECTED';
	    }
	} elsif ($num =~ /\d/) {
	    if ($num == $table_rounding) {
		$select = ' SELECTED';
	    }
	}
	$output_collect .= "<OPTION$select>$num</OPTION>\n";
    }
    $output_collect .= "</SELECT></TD><TD>&nbsp;</TD>\n";

    # menu for table font
    $output_collect .= "<TD><B>table font:</B>".&help('table_font')."<BR><SELECT name=\"table_font\">\n";
    
    foreach my $font (6..20) {	
	my $select = '';
	if ($font == $table_font) {
	    $select = ' SELECTED';
	}
	$output_collect .= "<OPTION$select>$font</OPTION>\n";
    }
    $output_collect .= "</SELECT></TD>\n";

    $output_collect .= "</TR></TABLE>\n";


    $output_collect .= "<H2>Filtering</H2>\n".(&pointer)."Combine up to $filter_lists_max_default uploaded list(s)";
    if ($FPMI_specific) {
	$output_collect .= ", such as the <A HREF=\"http://www.pathogenomics.ca/internal/data/lists/\">ones provided for FPMI members</A> ";
    }
    $output_collect .= "and/or $filter_fields_max_default criteria via Boolean operators to filter out certain entries from the data.<BR><BR>";

    $output_collect .= "<TABLE><TR>\n";

    # menu for filtering

    # first the uploadable lists
    my $max_header_length = 0;
    $header_counter = -1;
    foreach (@headers) {
	$header_counter++;
	next if (defined $skip_headers{$header_counter});
	if (length($_) > $max_header_length) {
	    $max_header_length = length($_);
	}
    }
    if ($max_header_length > 20) {
	$max_header_length -= 12;
    }

    for (my $i = 1; $i <= $filter_lists_max; $i++) {
	$output_collect .= "<TD><TABLE>\n";
	$output_collect .= "<TR><TH>filter list $i</TH></TR>\n";
	if (defined $filter_lists{$i}{'file'}) {
	    # a file has been uploaded previously
	    $output_collect .= "<TR><TD><SELECT name=\"filter_list$i\">\n<OPTION>-</OPTION>\n<OPTION VALUE=\"__UPLOADED__".(basename $filter_lists{$i}{'file'})."\" SELECTED>".(basename $filter_lists{$i}{'file'})."</OPTION>\n</SELECT></TD></TR>\n";
	} else {
	    $output_collect .= "<TR><TD>".($cgi -> filefield({name => "filter_list$i", size => $max_header_length}))."</TD></TR>\n";
	}
	$output_collect .= "<TR><TD>column <SELECT name=\"filter_list_column$i\">\n";
	for my $j (1..20) {
	    my $select = '';
	    if (defined $filter_lists{$i}{'column'} 
		and
		$filter_lists{$i}{'column'} eq $j) {
		$select = ' SELECTED';
	    }
	    $output_collect .= "<OPTION$select>$j</OPTION>\n";
	}
	$output_collect .= "</SELECT> <SELECT name=\"filter_list_match$i\">\n";
	foreach my $k ('matches exactly', 'is contained in', 'does not match', 'is not contained in') {
	    my $select = '';
	    if (defined $filter_lists{$i}{'match'} 
		and
		$filter_lists{$i}{'match'} eq $k) {
		$select = ' SELECTED';
	    }
	    $output_collect .= "<OPTION$select>$k</OPTION>\n";
	}
	$output_collect .= "</SELECT></TD></TR>\n";
	$output_collect .= "<TR><TD><SELECT name=\"filter_list_field$i\">\n";
	my $header_counter = -1;
	foreach (@headers) {
	    $header_counter++;
	    next if (defined $skip_headers{$header_counter});
	    my $select = '';
	    if (defined $filter_lists{$i}{'field'} 
		and
		$filter_lists{$i}{'field'} eq $_) {
		$select = ' SELECTED';
	    }
	    $output_collect .= "<OPTION$select>$_</OPTION>\n";
	}
	$output_collect .= "</SELECT></TD></TR>\n";
	$output_collect .= "</TABLE></TD>\n";

#	if ($i == $filter_lists_max) {
#	    # print connector into the next row if we are in the last field
#	    $output_collect .= "</TR></TABLE>\n";   
#	    $output_collect .= "<BR>\n";
#	    $output_collect .= "<TABLE><TR>\n";
#	}

	$output_collect .= "<TD VALIGN=MIDDLE><TABLE>\n";
	my $default = (defined $cgi->param("filter_list_connector$i") ? $cgi->param("filter_list_connector$i") : '');
	$output_collect .= "<TR><TD><SELECT name=\"filter_list_connector$i\">\n";
	foreach ('AND', 'OR', 'BUTNOT') {
	    my $select = '';
	    if (defined $filter_lists{$i}{'connector'}
		and
		$filter_lists{$i}{'connector'} eq $_) {
		$select = ' SELECTED';
	    }
	    $output_collect .= "<OPTION$select>$_</OPTION>\n";
	}
	$output_collect .= "</SELECT>\n";
	$output_collect .= "</TD></TR>\n";
	$output_collect .= "</TABLE></TD>\n";
	
    }

    $output_collect .= "</TR></TABLE>\n";   

    $output_collect .= "<TABLE><TR>\n";

    # no the fields
    for (my $i = 1; $i <= $filter_fields_max; $i++) {
	$output_collect .= "<TD><TABLE>\n";
	$output_collect .= "<TR><TH>filter field $i</TH></TR>\n<TR><TD><SELECT name=\"filter_field$i\"><OPTION>-</OPTION>\n";
	my $header_counter = -1;
	foreach (@headers) {
	    $header_counter++;
	    next if (defined $skip_headers{$header_counter});
	    my $select = '';
	    if (defined $filter_fields{$i}{'field'}{$_}) {
		$select = ' SELECTED';
	    }
	    $output_collect .= "<OPTION$select>$_</OPTION>\n";
	}
	$output_collect .= "</SELECT></TD></TR>\n";
	my $default = (defined $cgi->param("filter_field_operator$i") ? $cgi->param("filter_field_operator$i") : '');
	$output_collect .= "<TR><TD><SELECT name=\"filter_field_operator$i\">\n";
	my @operators = ('>', '>=', '=', '<=', '<', '//');
	if ($i == 1) {
	    push @operators, '!//';
	}
	foreach (@operators) {
	    my $select = '';
	    if (defined $filter_fields{$i}{'operator'}{$_}) {
		$select = ' SELECTED';
	    }
	    $output_collect .= "<OPTION$select>$_</OPTION>\n";
	}
	$output_collect .= "</SELECT>\n";

	# absolute box
	my $checked = '';
	if (defined $filter_fields{$i}{'absolute'}{1}) {
	    $checked = ' CHECKED';
	}
	$output_collect .= " <input type=\"checkbox\" name=\"filter_field_absolute$i\" value=\"1\"$checked> ignore sign<BR>\n";

	# filter value:
	$default = (defined $cgi->param("filter_field_value$i") ? $cgi->param("filter_field_value$i") : '');
	$output_collect .= $cgi ->  textfield(-name=>"filter_field_value$i",
					      -override=>1,
					      -default => $default, 
					      -size => 20);
	$output_collect .= '<BR>'.&help('filter_operators')."</TD></TR>\n";
	$output_collect .= "</TABLE></TD>\n";

	last if ($i == $filter_fields_max);
	# print connector unless we are in the last field

	$output_collect .= "<TD VALIGN=MIDDLE><TABLE>\n";
	$default = (defined $cgi->param("filter_field_connector$i") ? $cgi->param("filter_field_connector$i") : '');
	$output_collect .= "<TR><TD><SELECT name=\"filter_field_connector$i\">\n";
	foreach ('AND', 'OR', 'BUTNOT') {
	    my $select = '';
	    if (defined $filter_fields{$i}{'connector'}{$_}) {
		$select = ' SELECTED';
	    }
	    $output_collect .= "<OPTION$select>$_</OPTION>\n";
	}
	$output_collect .= "</SELECT>\n";
	$output_collect .= "</TD></TR>\n";
	$output_collect .= "</TABLE></TD>\n";
	
    }

    $output_collect .= "</TR></TABLE>\n";   


    $output_collect .= "<H2>Sorting</H2>\n".(&pointer)."Specify up to 5 columns (in decreasing precedence) that will be used to sort your data.<BR><BR>";

    $output_collect .= "<TABLE><TR>\n";

    # menu for sorting

    for (my $i = 1; $i <= $sort_fields_max; $i++) {
	$output_collect .= "<TD><TABLE>\n";
	$output_collect .= "<TR><TH>sort field $i</TH></TR>\n<TR><TD><SELECT name=\"sort_field$i\"><OPTION>-</OPTION>\n";
	my $header_counter = -1;
	foreach (@headers) {
	    $header_counter++;
	    next if (defined $skip_headers{$header_counter});
	    my $select = '';
	    if (defined $sort_fields{$i}{'field'}{$_}) {
		$select = ' SELECTED';
	    }
	    $output_collect .= "<OPTION$select>$_</OPTION>\n";
	}
	$output_collect .= "</SELECT></TD></TR>\n";
#	$output_collect .= "<TR><TD><SELECT name=\"sort_field_direction$i\">\n";
#	foreach ('ascending', 'descending') {
#	    my $select = '';
#	    if (defined $sort_fields{$i}{'direction'}{$_}) {
#		$select = ' SELECTED';
#	    }
#	    $output_collect .= "<OPTION$select>$_</OPTION>\n";
#	}
#	$output_collect .= "</SELECT></TD></TR>\n";

	# use radio button instead of select list
	# because it's quicker to use:
	foreach ('ascending', 'descending') {
	    my $checked = '';
	    if (defined $sort_fields{$i}{'direction'}{$_}) {
		$checked = ' checked="checked"';
	    }	
	    $output_collect .= "<TR><TD><INPUT TYPE=RADIO name=\"sort_field_direction$i\" value=\"$_\"$checked>$_\n";
	}
	$output_collect .= "</TABLE></TD>\n";
	
    }

    $output_collect .= "</TR></TABLE>\n";


    # check for saved lists
    my %saved_lists = ();
    if (defined $cgi->param('saved_lists')) {
	foreach my $list ($cgi->param('saved_lists')) {
	    next unless (-r $list);
	    $saved_lists{$list}{'name'} = basename $list;
	    $saved_lists{$list}{'step'} = $cgi->param("$list.step") || '?';
	    $saved_lists{$list}{'lines'} = $cgi->param("$list.lines") || '?';
	}
    }
    if ($submit eq 'Save List!') {
	my $list = $cgi->param('spreadsheet_input');
	my $name = $cgi->param('list_name');
	unless ($name =~ /\w/) {
	    die "No name given to save list! Please specify a name under which to save the list.\n";
	}
	if ($name =~ /\W/) {
	    die "Invalid character in listname! Please specify a name with alpha-numerical characters only under which to save the list.\n";
	}
	my $new_name = "$data_dir/$name";
	if (-e $new_name) {
	    unlink $new_name or die "Can't remove previous list '$new_name': $!\n";
	}
	system "cp '$list' '$new_name'";
	unless (-e $new_name) {
	    die "Couldn't create new list '$new_name'";
	}
	my $lines = `cat '$new_name'| grep -v '^#' | wc`;
	chomp $lines;
	$lines--; # remove header from count
	$saved_lists{$new_name}{'name'} = $name;
	$saved_lists{$new_name}{'lines'} = $lines;
	$saved_lists{$new_name}{'step'} = $step;
    }
    if (%saved_lists) {
	$output_collect .= "<H2>Saved List(s)</H2>\n";
	
	$output_collect .= "<TABLE BORDER=2><TR><TH>#</TH><TH>name</TH><TH>entries</TH><TH>step</TH></TR>\n";
	my $counter = 0;
	foreach my $list (sort { $saved_lists{$a}{'step'} <=> $saved_lists{$b}{'step'} } keys %saved_lists) {
	    $counter++;
	    my $link = $list;
	    $link = basename ($link);
	    $output_collect .= "<TR><TD>$counter</TD><TD><A HREF=\"$link\" TARGET=\"_list\">$saved_lists{$list}{'name'}</A></TD><TD ALIGN=RIGHT>$saved_lists{$list}{'lines'}</TD><TD ALIGN=RIGHT>$saved_lists{$list}{'step'}</TD></TR>\n";
	}
	$output_collect .= "</TABLE>\n";
    }

    my $settings_name = 'list';
    # check how many we have already that start with this string
    
    if (defined $cgi->param('list_name')) {
	$settings_name = $cgi->param('list_name');
	$settings_name =~ s/\d+$//;
    }

    if (opendir(DIR, $data_dir)) {
	my @nums = ();
	while (defined(my $file = readdir(DIR))) {
	    if ($file =~ /$settings_name\s*(\d+)/) {
		push @nums, $1;
	    }
	}
	close DIR;
	if (@nums) {
	    @nums = sort { $a <=> $b } @nums;
	    $settings_name .= $nums[$#nums] + 1;
	} else {
	    $settings_name .= 1;
	}
    } else {
	$settings_name .= 1;
    }
    
    my $formatted_file = $orig_input.".list_$appendix.tab";    

    # hidden fields:
    # input file
    $output_collect .=  "<input type=\"hidden\" name=\"spreadsheet_input\" value=\"$input_file\">\n";
    # formatted file
    $output_collect .=  "<input type=\"hidden\" name=\"formatted_file\" value=\"$formatted_file\">\n";
    # previous page
    $output_collect .=  "<input type=\"hidden\" name=\"prev_page\" value=\"$prev_page\">\n";
    # step
    $step++;    
    $output_collect .=  "<input type=\"hidden\" name=\"step\" value=\"$step\">\n";
    # flags and values
    foreach (%flag_values) {
	next unless (defined $flag_values{$_});
	$output_collect .= "<input type=\"hidden\" name=\"flags\" value=\"$_|$flag_values{$_}\">\n";
    }
    # skip flags
    foreach (keys %skip_flags) {
	$output_collect .=  "<input type=\"hidden\" name=\"prev_flags\" value=\"$_\">\n";
    }
    # filter lists
    foreach my $i (sort { $a <=> $b } keys %filter_lists) {
	my $file = $filter_lists{$i}{'file'} || '';
	my $column = $filter_lists{$i}{'column'} || '';
	my $match = $filter_lists{$i}{'match'} || '';
	my $field = $filter_lists{$i}{'field'} || '';
	my $connector = $filter_lists{$i}{'connector'} || '';
	$output_collect .= "<input type=\"hidden\" name=\"prev_filter_list$i\" value=\"__UPLOADED__".(basename $file)."\">\n";
	$output_collect .= "<input type=\"hidden\" name=\"prev_filter_list_column$i\" value=\"$column\">\n";
	$output_collect .= "<input type=\"hidden\" name=\"prev_filter_list_match$i\" value=\"$match\">\n";
	$output_collect .= "<input type=\"hidden\" name=\"prev_filter_list_field$i\" value=\"$field\">\n";
	$output_collect .= "<input type=\"hidden\" name=\"prev_filter_list_connector$i\" value=\"$connector\">\n";
    }

    # number of lists
    $output_collect .= "<input type=\"hidden\" name=\"filter_lists_max\" value=\"$filter_lists_max\">\n";
    $output_collect .= "<input type=\"hidden\" name=\"prev_filter_lists_max\" value=\"$filter_lists_max\">\n";

    # filter fields
    foreach my $i (sort { $a <=> $b } keys %filter_fields) {
	my @values = keys %{$filter_fields{$i}{'value'}};
	my @absolutes = keys %{$filter_fields{$i}{'absolute'}};
	my @operators = keys %{$filter_fields{$i}{'operator'}};
	my @connectors = keys %{$filter_fields{$i}{'connector'}};
	foreach my $field (keys %{$filter_fields{$i}{'field'}}) {
	    $output_collect .= "<input type=\"hidden\" name=\"prev_filter_field$i\" value=\"$field\">\n";
	    my $value = shift @values || '';
	    $output_collect .= "<input type=\"hidden\" name=\"prev_filter_field_value$i\" value=\"$value\">\n";
	    my $absolute = shift @absolutes || '';
	    $output_collect .= "<input type=\"hidden\" name=\"prev_filter_field_absolute$i\" value=\"$absolute\">\n";
	    my $operator = shift @operators || '>';
	    $output_collect .= "<input type=\"hidden\" name=\"prev_filter_field_operator$i\" value=\"$operator\">\n";
	    
	    # skip the last connector
	    last if ($i == (scalar keys %filter_fields));
	    my $connector = shift @connectors || 'AND';
	    $output_collect .= "<input type=\"hidden\" name=\"prev_filter_field_connector$i\" value=\"$connector\">\n";
	}
    }
    # number of filters
    $output_collect .= "<input type=\"hidden\" name=\"filter_fields_max\" value=\"$filter_fields_max\">\n";
    $output_collect .= "<input type=\"hidden\" name=\"prev_filter_fields_max\" value=\"$filter_fields_max\">\n";
    # sort fields
    foreach my $i (keys %sort_fields) {
	my @directions = keys %{$sort_fields{$i}{'direction'}};
	foreach my $field (keys %{$sort_fields{$i}{'field'}}) {
	    $output_collect .= "<input type=\"hidden\" name=\"prev_sort_field$i\" value=\"$field\">\n";
	    my $direction = shift @directions || 'ascending';
	    $output_collect .= "<input type=\"hidden\" name=\"prev_sort_field_direction$i\" value=\"$direction\">\n";
	}
    }
    # number of sorts
    $output_collect .= "<input type=\"hidden\" name=\"sort_fields_max\" value=\"$sort_fields_max\">\n";
    $output_collect .= "<input type=\"hidden\" name=\"prev_sort_fields_max\" value=\"$sort_fields_max\">\n";
    
    # saved lists
    foreach (keys %saved_lists) {
	$output_collect .= "<input type=\"hidden\" name=\"saved_lists\" value=\"$_\">\n";
	$output_collect .= "<input type=\"hidden\" name=\"$_.step\" value=\"$saved_lists{$_}{'step'}\">\n";
	$output_collect .= "<input type=\"hidden\" name=\"$_.lines\" value=\"$saved_lists{$_}{'lines'}\">\n";
    }
    # other fields
    $output_collect .= "<input type=\"hidden\" name=\"action_step\" value=\"$action_step\">\n";
    $output_collect .= "<input type=\"hidden\" name=\"prev_page\" value=\"$referral_address\">\n";
    if ($output_file_title) {
	$output_collect .= "<input type=\"hidden\" name=\"output_file_title\" value=\"$output_file_title\">\n";
    }
#    $output_collect .= "<input type=\"hidden\" name=\"results_file\" value=\"$results_file_new\">\n";
    $output_collect .= "<input type=\"hidden\" name=\"tmp_dir\" value=\"$tmp_dir\">\n";
    $output_collect .= "<input type=\"hidden\" name=\"tmp_dir_html\" value=\"$tmp_dir_html\">\n";
    $output_collect .= "<input type=\"hidden\" name=\"last_pid\" value=\"$PID\">\n";
    $output_collect .= "<input type=\"hidden\" name=\"current_line_start\" value=\"$line_start\">\n";
    $output_collect .= "<input type=\"hidden\" name=\"current_line_numbers\" value=\"$line_numbers\">\n";
    $output_collect .= "<input type=\"hidden\" name=\"pipeline_user\" value=\"$username\">\n";
    $output_collect .= "<input type=\"hidden\" name=\"pipeline_password_check\" value=\"$password_check\">\n";

    foreach (keys %skip_headers) {
	$output_collect .= "<input type=\"hidden\" name=\"skip_headers_all\" value=\"$_\">\n";
    }
    foreach (keys %skip_group_headers) {
	$output_collect .= "<input type=\"hidden\" name=\"skip_headers_all\" value=\"group_header $_\">\n";
    }

    foreach my $file ($cgi->param('work_files')) {
	$output_collect .= "<input type=\"hidden\" name=\"work_files\" value=\"$file\">\n";
    }
	
    $output_collect .= "<BR>\n<H2>Action</H2>"
	. "<TABLE CELLPADDING=5><TR><TD BGCOLOR=RED>"
	. $cgi -> submit(-name=>'submit', 
			 -value=>'Reprint Table!')
	. '</TD></TR></TABLE>'
#	. "<TABLE CELLPADDING=5><TR><TD BGCOLOR=YELLOW>"
#	. $cgi -> submit(-name=>'submit', 
#			 -value=>'Save List!')
#	. '&nbsp;'
#	. 'specify list name: '.$cgi ->  textfield(-name => 'list_name', 
#						   -override=>1,
#						   -default => $settings_name, 
#						   -size => (length($settings_name)+5))
#	. '(alpha-numerics only, no whitespace)'
#	. '</TD></TR></TABLE>'
	. "\n<BR><BR>\n";
    
    # menu for advanced tools
 
    $output_collect .= "<H2>Extra Tools</H2>\n".(&pointer)."Each panel provides extra functionality - click the according buttons to run.<BR><BR>\n";

    # barplot plot of selected column
    
    my $column_selection = "<SELECT name=\"barplot_column\">\n<OPTION VALUE=\"-1\">-</OPTION>\n";
    $header_counter = -1;
    my %preselect = ();
    if (defined $cgi->param('barplot_column')) {
	foreach ($cgi->param('barplot_column')) {
	    $preselect{$_-1}++;
	}
    }

    my $remain_counter = 1;
    foreach (@headers) {
	$header_counter++;
	next if (defined $skip_headers{$header_counter});
	my $select = '';
	if (defined $preselect{$header_counter}) {
	    $select = ' SELECTED';
	}
	$column_selection .= "<OPTION VALUE=\"$remain_counter\"$select>$_</OPTION>\n";
	$remain_counter++;
    }
    $column_selection .= "</SELECT>\n";
	
    $output_collect .= "<TABLE CELLSPACING=10 BORDER=2 bgcolor=\"#FFFFCC\" bordercolor=\"#006699\"><TR valign=top><TD BGCOLOR=WHITE>\n";

    # menu for Barplot or Histogram
    $output_collect .= "<TABLE WIDTH=200><TR>
<TH COLSPAN=2>Create Barplot or Histogram</TH></TR>
<TR><TD COLSPAN=2>Plot a histogram (numbers only) or a barplot (numbers or text) from the selected column:</TD></TR>
<TR><TD COLSPAN=2>$column_selection</TD></TR>";
    $output_collect .= "<TR><TD>rounding:</TD><TD><SELECT name=\"barplot_rounding\">\n";
    
    my $bar_plot_rounding = '-';
    if (defined $cgi->param('barplot_rounding')) {
	$bar_plot_rounding = $cgi->param('barplot_rounding');
    }
    foreach my $num ('-', (-4..15)) {	
	my $select = '';
	if ($bar_plot_rounding =~ /\d/) {
	    if ($num =~ /\d/
		and
		$bar_plot_rounding == $num) {
		$select = ' SELECTED';
	    }
	} else {
	    if ($num =~ /^-$/) {
		$select = ' SELECTED';
	    }
	}
	$output_collect .= "<OPTION$select>$num</OPTION>\n";
    }
    $output_collect .= "</SELECT>".&help('rounding')."</TD></TR>\n";
    my $bar_plot_show_lines = 10;
    if (defined $cgi->param('barplot_show_lines')) {
	$bar_plot_show_lines = $cgi->param('barplot_show_lines');
    }
    $output_collect .= "<TR><TD>listing:</TD><TD><input type=\"text\" name=\"barplot_show_lines\" value=\"$bar_plot_show_lines\" size=\"4\">lines".&help('show_lines')."</TR></TD>\n";
    $output_collect .= "<TR><TD>range:</TD><TD>"
	. $cgi -> radio_group(-name=>'barplot_range',
				  -values=>[ ('all', 'visible') ],
				  -default=>'all',
				  )
	.&help('barplot_range')
	."</TR></TD>";

    my $bar_plot_breaks = 'auto';
    if (defined $cgi->param('barplot_breaks')) {
	$bar_plot_breaks = $cgi->param('barplot_breaks');
    }
    $output_collect .= "<TR><TD>breaks:</TD><TD><input type=\"text\" name=\"barplot_breaks\" value=\"$bar_plot_breaks\" size=\"4\">lines".&help('barplot_breaks')
	."</TD></TR>\n<TR ALIGN=CENTER><TD COLSPAN=2><BR>"
	. $cgi -> submit(-name=>'submit', 
			 -value=>'Barplot!')
	. $cgi -> submit(-name=>'submit', 
			 -value=>'Histogram!')
	."</TD></TR>\n</TABLE>\n";


    # Menu for Transpath:

    if ($FPMI_specific) {
	$column_selection = "<SELECT name=\"transpath_column\">\n";
	$header_counter = -1;
	%preselect = ();
	if (defined $cgi->param('transpath_column')) {
	    foreach ($cgi->param('transpath_column')) {
		$preselect{$_-1}++;
	    }
	}
	
	$remain_counter = 1;
	my $found_column = 0;
	foreach (@headers) {
	    $header_counter++;
	    next if (defined $skip_headers{$header_counter});
	    unless (/fold change/) {
		$remain_counter++;
		next;
	    }
	    my $select = '';
	    if (defined $preselect{$header_counter}) {
		$select = ' SELECTED';
	    }
	    $column_selection .= "<OPTION VALUE=\"$remain_counter\"$select>$_</OPTION>\n";
	    $found_column++;
	    $remain_counter++;
	}
	$column_selection .= "</SELECT>\n";
	
	unless ($found_column) {
	    # no header found
	    $column_selection = "No column with fold-changes found - please select one from the <i><A HREF=\"#table_settings_section\">show columns</A></i> box above!\n";
	}
	
	$output_collect .= "</TD><TD BGCOLOR=WHITE>\n<TABLE WIDTH=200><TR>
<TH>Link to TransPath</TH></TR>
<TR><TD>Sends fold-changes from the selected column to TransPath:</TD></TR>
<TR><TD>$column_selection</TD></TR>
";

	# make sure that some annotation is available:
	my %suitable_columns = ('SWISSPROT', -1,
				'EMBL', -1,
				'SPTREMBL',-1,
				'SP', -1,
				'dbj', -1,
				'emb', -1,
				'gb', -1,
				'SWISSPROT accession', -1,
				'EMBL accession', -1,
				'SPTREMBL accession',-1);
	my $found_annotation = 0;
	$header_counter = -1;
	foreach (@headers) {
	    $header_counter++;
	    next if (defined $skip_headers{$header_counter});
	    if (defined $suitable_columns{$_}) {
		$found_annotation++;
		last;
	    }
	}
	
	if ($found_annotation) {	    
	    $output_collect .= "<TR ALIGN=CENTER><TD><BR><input name=\"submit\" value=\"TransPath\" type=\"submit\"></TD></TR>
</TABLE>\n";
	} else {
	    $output_collect .= "<TR ALIGN=CENTER><TD><BR>No EMBL/GenBank/DDBJ/SWISSPROT column found - please make sure you chose the annotation module and select column(s) from the <i><A HREF=\"#table_settings_section\">show columns</A></i> box above!</TD></TR>
</TABLE>\n";
	}
    }
    
    # Menu for Pathway plotter
    $column_selection = "<SELECT name=\"pathway_plotter_id_column\">\n<OPTION VALUE=\"-1\">-</OPTION>\n";
    $header_counter = -1;
    %preselect = ();
    if (defined $cgi->param('pathway_plotter_id_column')) {
	foreach ($cgi->param('pathway_plotter_id_column')) {
	    $preselect{$_-1}++;
	}
    }

    $remain_counter = 1;
    my $selected = 0;
    foreach (@headers) {
	$header_counter++;
	next if (defined $skip_headers{$header_counter});
	my $select = '';
	if (%preselect) {
	    if (defined $preselect{$header_counter}) {
		$select = ' SELECTED';
		$selected = 1;
	    }
	} else {
	    if ($selected == 0
		and
		($_ eq 'Probe ID'
		 or
		 $_ eq $SPOT_ID_TERM)) {
		$selected = 1;
		$select = ' SELECTED';
	    }
	}
	$column_selection .= "<OPTION VALUE=\"$remain_counter\"$select>$_</OPTION>\n";
	$remain_counter++;
    }
    $column_selection .= "</SELECT>\n";

    $output_collect .= "</TD><TD BGCOLOR=WHITE>\n<TABLE WIDTH=200>
<TR>
<TH>Pathway Plotter</TH>
</TR>
<TR>
<TD>requires BioCarta output, at least two fold change or ratios columns and spot id
<P>
id column:
</TR>
<TR>
<TD>$column_selection</TD>
</TR>
";
    $column_selection = "<SELECT name=\"pathway_plotter_value_column\" multiple size=4>\n<OPTION VALUE=\"-1\">-</OPTION>\n";
    $header_counter = -1;
    %preselect = ();
    if (defined $cgi->param('pathway_plotter_value_column')) {
	foreach ($cgi->param('pathway_plotter_value_column')) {
	    $preselect{$_-1}++;
	}
    }

    $remain_counter = 1;
    foreach (@headers) {
	$header_counter++;
	next if (defined $skip_headers{$header_counter});
	my $select = '';
	if (%preselect) {
	    if (defined $preselect{$header_counter}) {
		$select = ' SELECTED';
	    }
	} else {
	    if (/_R/
		or
		/$MERGED_RATIOS_TERM/) {
		$select = ' SELECTED';
	    }
	}
	$column_selection .= "<OPTION VALUE=\"$remain_counter\"$select>$_</OPTION>\n";
	$remain_counter++;
    }
    $column_selection .= "</SELECT>\n";

$output_collect .= "
<TR>
<TD>value column(s):
</TR>
<TR>
<TD>$column_selection</TD>
</TR>
<TR ALIGN=CENTER>
<TD>
<input name=\"submit\" value=\"Plot Pathways!\" type=\"submit\">
</TD>
</TR>
";

    $output_collect .= "\n</TABLE>\n</TD><TD BGCOLOR=WHITE>\n<TABLE WIDTH=200>";
    $column_selection = "<SELECT name=\"add_column1\">\n";
    $header_counter = -1;
    %preselect = ();
    if (defined $cgi->param('add_column1')) {
        foreach ($cgi->param('add_column1')) {
            $preselect{$_}++;
        }
    }

    foreach (@headers) {
        $header_counter++;
        next if (defined $skip_headers{$header_counter});
        my $select = '';
        if (%preselect) {
            if (defined $preselect{$_}) {
                $select = ' SELECTED';
            }
	}
        $column_selection .= "<OPTION VALUE=\"$_\"$select>$_</OPTION>\n";
    }
    $column_selection .= "</SELECT>\n";

    $output_collect .= "
<TR>
<TH>Calculate a Column</TH>
</TR>
<TR>
<TD>Add a new column to the output based on the input and operator specified below:
<P>
column 1:
</TR>
<TR>
<TD>$column_selection</TD>
</TR>
";


    $column_selection = "<SELECT name=\"add_column_op\">\n";
    %preselect = ();
    if (defined $cgi->param('add_column_op')) {
        foreach ($cgi->param('add_column_op')) {
            $preselect{$_}++;
        }
    }

    my @operators = ('-', '+', '*', '/', 'log2', 'exp2');
    foreach (@operators) {
        my $select = '';
        if (%preselect) {
            if (defined $preselect{$_}) {
                $select = ' SELECTED';
            }
        }
        $column_selection .= "<OPTION VALUE=\"$_\"$select>$_</OPTION>\n";
    }
    $column_selection .= "</SELECT>\n";

    $output_collect .= "<TR>
<TD>
<P>
operator:&nbsp;
$column_selection</TD>
</TR>
";

    $column_selection = "<SELECT name=\"add_column2\">\n<OPTION VALUE='-'>-</OPTOIN>";
    $header_counter = -1;
    %preselect = ();
    if (defined $cgi->param('add_column2')) {
        foreach ($cgi->param('add_column2')) {
            $preselect{$_}++;
        }
    }

    foreach (@headers) {
        $header_counter++;
        next if (defined $skip_headers{$header_counter});
        my $select = '';
        if (%preselect) {
            if (defined $preselect{$_}) {
                $select = ' SELECTED';
            }
        }
        $column_selection .= "<OPTION VALUE=\"$_\"$select>$_</OPTION>\n";
    }
    $column_selection .= "</SELECT>\n";

    $output_collect .= "
<TR>
<TD>
<P>
column 2:
</TR>
<TR>
<TD>$column_selection</TD>
</TR>
";

    $output_collect .= "
<TR>
<TD>
<P>
column label:
</TR>
<TR>
<TD>
<input type=\"text\" size=25 name=\"add_column_name\">
</TD>
</TR>
";

    $output_collect .= "
<TR ALIGN=CENTER>
<TD>
<input name=\"submit\" value=\"Add Column!\" type=\"submit\">
</TD>
</TR>
</TABLE>
";

    $output_collect .= "</TD></TR></TABLE>\n";


    print "<H2><A NAME=\"output\">Output$output_file_title</A></H2>\n";
    if ($table_element_size_limit) {
	print "Table elements with text exceeding $table_element_size_limit characters are truncuated!";
	if ($table_element_size_excess eq 'hide') {
	    print "<BR>Click on <FONT COLOR=GREEN><B>'[...]'</B></FONT> to show the whole content (requires Javascript) or change <A HREF=\"#table_settings_section\">settings</a>.\n";
	}
	print "<BR>\n";
    } else {
	print "<SMALL><A HREF=\"#table_settings_section\">[SETTINGS]</A></SMALL><BR>";
    }
    print "<B>$new_lines</B> + 1 lines, \n";
    print &help(
		     {
			 'topic' => 'table_window_position',
			 'label' => 'window&nbsp;navigation:&nbsp;',
			 'no_small' => 1
		     }
		);
    print $cgi->submit(-name => 'submit',
		       -value => '|<');
    print $cgi->submit(-name => 'submit',
		       -value => '<');
    print $cgi->textfield(-name => 'window_slide',
			  -default => $line_numbers,
			  -override=>1,
			  -size => 4);
    print $cgi->submit(-name => 'submit',
		       -value => '>');
    print $cgi->submit(-name => 'submit',
		       -value => '>|');
    print '&nbsp;&nbsp;&nbsp;';
    print &help(
		     {
			 'topic' => 'table_window_start',
			 'label' => 'window&nbsp;start:&nbsp;',
			 'no_small' => 1
		     }
		)
	.$cgi ->  textfield(-name => 'table_window_start', 
			    -override=>1,
			    -default => $line_start, 
			    -size => 6)
	.&help(
		     {
			 'topic' => 'line_numbers',
			 'label' => '&nbsp;entries:&nbsp;',
			 'no_small' => 1
		     }
		)
	.$cgi ->  textfield(-name => 'line numbers', 
			    -override=>1,
			    -default => $line_numbers, 
			    -size => 4)
	.&help(
		     {
			 'topic' => 'chunk_size',
			 'label' => '&nbsp;&nbsp;lines:&nbsp;',
			 'no_small' => 1
		     }
		)
	.$cgi ->  textfield(-name => 'chunk size', 
			    -override=>1,
			    -default => $chunk_size, 
			    -size => 4)
	.&help(
		     {
			 'topic' => 'header_repeat',
			 'label' => '&nbsp;&nbsp;headers:&nbsp;',
			 'no_small' => 1
		     }
		)
	.$cgi ->  textfield(-name => 'header repeat', 
			    -override=>1,
			    -default => $header_repeat, 
			    -size => 4);
#	.' &nbsp;&nbsp;'
#	. "<TABLE CELLPADDING=5><TR><TD BGCOLOR=RED>"
#	. $cgi -> submit(-name=>'submit', 
#			 -value=>'Reprint Table!')
#	. '</TD></TR></TABLE>';
   print "\n<BR><BR>\n";


    my $line_end = $line_start + $line_numbers - 1;

    # assemble header
    $header_counter = -1;
    $header_no_tab = "<TR BGCOLOR=yellow><TH>".&proper_header('#')."</TH>";
#    my @headers_for_file = ('COLUMNS:#');
    my @headers_for_file = ('line');
    foreach (@headers) {
	$header_counter++;
	next if (defined $skip_headers{$header_counter});
	push @headers_for_file, $_;
	$header_no_tab .= "<TH>".&proper_header($_)."</TH>\n";;
    }
	
    open (OUT, ">$formatted_file")
	or die "Can't write to $formatted_file: $!";
    print OUT (join "\t", @headers_for_file)."\n";
    $header_no_tab .= "</TR>\n";
    $header = "<TABLE BORDER=1 CELLPADDING=3>$header_no_tab";
    print "$header";

    my $te_counter = 0;

    my $max_column_entries = @headers_for_file;
    while (my $line = <SHEET_DATA>) {
	# skip if nothing's on the line
	next unless ($line =~ /\w/);

	chomp $line;
	# data lines
	$line_counter++;

	# write output to file
	my @tmp = (split /\t/, $line);
	if ($line =~ /\t$/) {
	    push @tmp, '';
	}

	my $line_counter_cp = 0;
	if ($line_counter >= $line_start) {
	    $line_counter_cp = $line_counter - $line_start + 1;
	}

	$header_counter = -2;
	my @out_filtered = ();
	my $bg_colour = shift @bg_colours;
	push @bg_colours, $bg_colour;
#	my $out .= "<TR BGCOLOR=$bg_colour><TD>$line_counter</TD>\n";
	my $out .= "<TR BGCOLOR=$bg_colour>\n";
	foreach my $element ($line_counter_cp,@tmp) {
#	foreach my $element (@tmp) {
	    $header_counter++;
	    next if (defined $skip_headers{$header_counter});

	    if ($element =~ /\w/) {
		my @element = ($element);
		if ($element =~ /\;/) {
		    @element = split /; /, $element;
		}
		foreach (@element) {
		    if ($table_rounding =~ /\d/) {
			if (/\d+\.\d+e-\d+/) {
			    $_ = sprintf "%f", $_;
			}
			if (/\.(\d+)$/) {
			    if (length($1) > $table_rounding) {
				$_ = sprintf "%.${table_rounding}f", $_;
			    }
			}
#		    $element =~ s/\.0+$//;
		    }
		}
		$element = join '; ', @element;
	    }
	    # collect the values that will be written to the csv file:
	    push @out_filtered, $element;
	    
	    # do further formatting for the table output
	    # unless these lines are to be skipped
	    if ($line_counter_cp < 1) {
		next;
	    }
	    if ($line_end =~ /\d/
		and $line_counter > $line_end) {
		next;
	    }

	    my $cut = 0;
	    if ($table_element_size_limit > 0
		and
		length($element) > $table_element_size_limit) {
#		$cut = substr $element, $table_element_size_limit-5, -1;
		$cut = $element;
		$element = substr $element, 0, $table_element_size_limit-5;
	    }
	    # if we got here then this line is to be printed
	    if ($element eq ''
		or
		$element eq ' ') {
		$element = '&nbsp;';
	    } elsif (defined $add_links{$header_counter}) {
		my $out_field = '';
		foreach my $item (split /; /, $element) {
		    # remove any trailing semicolons
		    $item =~ s/;\s*$//;
		    my $url = '';
		    my $column = $headers[$header_counter];
		    if (defined $annotation_links{$column}) {
			$url = $annotation_links{$column};
		    } elsif (defined $annotation_links_alternatives{$column}) {
			foreach my $pattern (keys %{$annotation_links_alternatives{$column}}) {
			    if ($item =~ /$pattern/) {				
				$url = $annotation_links_alternatives{$column}{$pattern};
				last;
			    }
			}
		    }
		    my $item_orig = $item;
		    $item =~ s/ /\+/g;
		    if (defined $annotation_remove{$headers[$header_counter]}) {
			foreach my $pattern (@{$annotation_remove{$headers[$header_counter]}}) {
			    $item =~ s/$pattern//g;
			}
		    }
		    $url =~ s/__REPLACE__/$item/g;
		    if (defined $annotation_skip{$column}
			and
			$item_orig eq $annotation_skip{$column}) {
			$out_field .= $item_orig;
		    } else {
			$out_field .= "<A HREF=\"$url\" TARGET=\"external_links\">$item_orig</A>; ";
		    }
		}
		chop $out_field;
		chop $out_field;
		$element = $out_field;
	    }
	    my $align = ' ALIGN=RIGHT';
	    if ($element =~ /[a-z]/i) { # entry contains a letter
		$align = '';
	    }
	    if ($cut) {
		if ($table_element_size_excess eq 'hide') {
		    $te_counter++;
		    $element .= &help(
				      {
					  'topic' => "TE$te_counter",
					  'label' => '[...]',
					  'text' => $cut,
					  'title' => '__none__',
					  'no_small' => 1,
					  'background' => '#ffffff',
					  'no_link' => 1
					  }
				      );
		} else {
		    $element .= '...';
		}
	    } 

	    $out .= "<TD$align>$element</TD>\n";
	} # end of elements for this line

	$out .= "</TR>\n";
	print OUT (join "\t", (@out_filtered))."\n";
	

	if ($line_counter_cp < 1) {
	    next;
	}
	if ($line_end =~ /\d/
	    and $line_counter > $line_end) {
	    next;
	}
	
	$page_counter++;
	if ($chunk_size =~ /\d/
	    and
	    $page_counter > $chunk_size) {
	    $page_counter = 1;
	    # close table and start new one
	    print "</TABLE>\n$header";
	    $last_header = $line_print_counter;
	}

	if ($header_repeat =~ /\d/
	    and
	    $page_counter > 1
	    and
	    ($page_counter-1) % $header_repeat == 0) {
#	    $page_counter--;
	    # close table and start new one
	    print "$header_no_tab";
	    $last_header = $line_print_counter;
	}   

	# print the formatted output row
	print $out;
	$line_print_counter++;

    } # end of lines

    # print the header again in case we had more than 9 lines between the last one
    if ($line_print_counter - $last_header > 9) {
	print "$header_no_tab";
    }

    print "</TABLE>\n";
    close SHEET_DATA;

    close OUT;

    $output_collect .= "<H2>Downloads</H2>\n".(&pointer)."Click on the links to retrieve the results file form the pipeline, or the output generated by the table tool:<BR>";
    $output_collect .= "<TABLE>\n<TR><TD><A HREF=\"".(basename $orig_input)."\">pipeline output</TD><TD ALIGN=RIGHT>($lines + 1 lines, ".(&file_size($orig_input)).")</TD></TR>\n";
    $output_collect .= "<TR><TD><A HREF=\"".(basename $formatted_file)."\">new list</TD><TD ALIGN=RIGHT>($new_lines + 1 lines, ".(&file_size($formatted_file)).")</TD></TR>\n";

    # TIGR output:
    # check if we have intensities
    my $file_selection = "<B>TIGR MEV</B>: <FONT COLOR=RED>No intensity information found!</FONT>"
	. &help(
		{
		    'topic' => 'tmev_output',
#		    'label' => 'info'
		    }
		)
	;
    my %tmev_selection = ();
    foreach (@headers_for_file) {
	if (/(.*) ?(ag|au|cy3|cy5|ch1|ch2) (.*)/i) {
	    my $file = $1;
	    my $type = $3;
	    # skip flag and background channels
	    next if ($type eq $BG_SIGNAL_TERM
		     or
		     $type eq $FLAG_CHANNEL_TERM);
	    $tmev_selection{$file}{$type}++;
	}
    }
    my %types = ();
    foreach my $file (sort keys %tmev_selection) {
	my @types = sort keys %{$tmev_selection{$file}};
	foreach my $type (@types) {
	    # we need the info for both channels:
	    unless ($tmev_selection{$file}{$type} == 2) {
		next;
	    }
	    # count how many files provide this type
	    $types{$type}++;
	}
    }
    my $default_type = '';
#    if (defined $cgi->param('tmev_type')) {
#	$default_type = $cgi->param('tmev_type');
#    }
    if (%types) {

	# how many types?
	my $types_num = scalar keys %types;

	if ($types_num > 1) {

	    # make sure the selected one still exists:
	    if ($default_type) {
		unless (defined $types{$default_type}) {
		    $default_type = '';
		}
	    }

	    # unless a default type is selected, make an automatic choice:
	    if ($default_type eq '') {
		# find the highest number of files for any type:
		my $max_files = 0;
		foreach (keys %types) {
		    if ($types{$_} > $max_files) {
			$max_files = $types{$_};
		    }
		}
		# our preference list:
		my @prefs = ('Dehub_M_L', 
			     'Dehub_L_M', 
			     'Dehub_L', 
			     'Dehub_M', 
			     'Dehub', 
			     'Norm_M_L', 
			     'Norm_L_M', 
			     'Norm_L', 
			     'Norm_M', 
			     'Norm', 
			     $FG_COR_SIGNAL_TERM.'_M_L', 
			     $FG_COR_SIGNAL_TERM.'_L_M', 
			     $FG_COR_SIGNAL_TERM.'_M', 
			     $FG_COR_SIGNAL_TERM.'_L', 
			     $FG_COR_SIGNAL_TERM, 
			     $FG_SIGNAL_TERM.'_M_L', 
			     $FG_SIGNAL_TERM.'_L_M', 
			     $FG_SIGNAL_TERM.'_M', 
			     $FG_SIGNAL_TERM.'_L', 
			     $FG_SIGNAL_TERM
			     );
	    # go through all types for which the highest amount of files is available
		foreach my $pref_type (@prefs) {
		    if (defined $types{$pref_type}) {
			next unless ($types{$pref_type} == $max_files);
			$default_type = $pref_type;
			last;
		    }
		}
	    }
	    # TODO: make a selection box to choose which type to use
	} else {
	    $default_type = (keys %types)[0];
	}
	
	$file_selection = "<input name=\"submit\" value=\"TIGR MEV\" type=\"submit\"><BR>using $default_type for $types{$default_type} files\n"
	    . &help(
		    {
			'topic' => 'tmev_output_ok',
#			'label' => 'info'
			}
		    )
	    
	    . "<input type=\"hidden\" name=\"tmev_type\" value=\"$default_type\">\n";
    }

    $output_collect .= "<TR><TD COLSPAN=2><P><P>or generate output from the new list in a special format:</TD></TR>\n";
    $output_collect .= "<TR><TD COLSPAN=2>$file_selection</TD></TR>\n";
#    if ($FPMI_specific) {
#	$output_collect .="<TR><TD COLSPAN=2>Generate output from the new list in one of the following formats:</TD></TR>\n";
#	$output_collect .= "<TR><TD COLSPAN=2>\n";
#   #    $output_collect .= "<input name=\"submit\" value=\"TIGR MEV\" type=\"submit\">\n";
#	$output_collect .= "<input name=\"submit\" value=\"TransPath\" type=\"submit\">\n";    
#	$output_collect .= "</TD></TR>\n";
#    }

    $output_collect .= "</TABLE>\n";

    # add button to upload a file:
    $output_collect .= "<H2>Uploads</H2>\n".(&pointer)."You can use the table generator to look at the original data input files or to check out any other file in tab-delimited format.<BR><BR>";
    $output_collect .= "<BR>Upload a file: ".($cgi -> filefield({name => "table_upload_file", size => 40}))."<BR>\n";

    # make original input files available
    my %input_files = ();
    my @input_files = ();
    foreach my $file ($cgi->param('work_files')) {
	my $base = basename $file;
	$input_files{$file} = $base;
    }
    
    # list input files (if specified/found):
    if (%input_files) {
	$output_collect .= "or pick an input file:<BR>";
	@input_files = ();
#	my @tmp = sort { $input_files{$a} cmp $input_files{$b} } keys %input_files;
	my @tmp = sort keys %input_files;
	# filter out those that don't have the correct format
	foreach my $file (@tmp) {
	    push @input_files, $file;
	}
	$output_collect .= $cgi -> radio_group(-name=>'original_input_file',
				  -values=>[@input_files],
				  -linebreak=>'true',
				  -default=>'-',
				  -labels=>\%input_files)
	    . "<BR>\n";    
    }


    $output_collect .= "<input type=\"hidden\" name=\"program_part\" value=\"html_table_output\">\n";
    $output_collect .= '<BR>'.$cgi -> submit(-name=>'submit', 
				    -value=>'Load File!')
    . "<BR>\n";    


    $output_collect .= &ending('start over', 'home', 'tutorial');

    
    print $output_collect;

    close HTML;

    my $redirect = "<html>
<head>
<title>arraypipe redirect</title>
<meta http-equiv=\"refresh\" content=\"0; URL=$html_output\" />
</html>
";

    if (defined $cgi->param('spreadsheet_background')) {
	$results_file = $cgi->param('spreadsheet_background');
	# we were running in the background...
	unless (open (REDIRECT, ">$results_file")) {
	    die "Can't write to $results_file: $!";
	}
	print REDIRECT $redirect;
	close REDIRECT;
    } else {
	print STDOUT $cgi->redirect("$html_output");
    }

    exit;

}

sub combine_files {
    my $ref = shift;
    my %selected_fields = %$ref;
    $ref = shift;
    my %hidden_fields = %$ref;
    my %hidden_fields_num = ();
    my $merge_field = shift;
    my @files = @_;

#    warn "# input_files: @files";

    my %done_field = ();
    my %start_pos = ();
    my %values_pos = ();
    my %annotation_pos = ();
    my @start_pos = ();
    my @values_pos = ();
    my %values_pos_new = ();
    my @annotation_pos = ();
    my %common_done = ();
    my @annotation_fields = ();
    my $annotation_fields = $cgi->param('spreadsheet_annotation_headers') || '';
    @annotation_fields = split /\|/, $annotation_fields;

    my @merge_fields = ();
    if ($merge_field eq 'position') {
	@merge_fields = ('MROW','MCOL','ROW','COL');
    }

#    foreach (@merge_fields) {
#	push @start_pos, $_;
#	$start_pos{$_}++;
#	$done_field{$_}++;
#    }


    # determine which headers are being used:

    # START section:
    if (defined $selected_fields{'everything'}
	and
	@files == 1) {
	# everything is selected!
	
	# there is nothing to be done
	# except determining the positions of the hidden fields
	if (%hidden_fields) {
	    # read the header
	    my $file = $files[0];
	    unless (open (IN, $file)) {
		warn "Can't read '$file': $!";
		next;
	    }
	    
	    my @headers = &get_header(*IN);
	    my $column_counter = 0;
	    foreach my $field (@headers) {
		$hidden_fields_num{$column_counter}++ if (defined $hidden_fields{$field});
		$column_counter++;		    
	    }
	    %$ref = %hidden_fields_num;
	}
	
	warn "# returning $files[0]!";
	return $files[0];
    }         
	
    # decipher short cuts:
    
    # select the common start if everything was selected
    if (defined $selected_fields{'everything'}) {
	foreach ("$COMMON_GENE_NAME","$COMMON_GENE_NAME_SOURCE", "$COMMON_GENE_DESCRIPTION","$COMMON_GENE_DESCRIPTION_SOURCE", 'Probe ID', $SPOT_ID_TERM, 'i', 'MROW','MCOL','ROW','COL') {
	    push @start_pos, $_;
	    $done_field{$_}++;
	    $start_pos{$_}++;
	}
    }    
	
    # 'probe and spot ids':
    if (defined $selected_fields{'probe and spot ids'}) {
	foreach ('Probe ID', $SPOT_ID_TERM) {
	    next if ($done_field{$_});
	    push @start_pos, $_;
	    $done_field{$_}++;
	    $start_pos{$_}++;
	}
	# hide probe id:
	$hidden_fields{'Probe ID'}++;
    }

    # 'spot coordinates':
    if (defined $selected_fields{'spot coordinates'}) {
	foreach ('i', 'MROW','MCOL','ROW','COL') {
	    push @start_pos, $_;
	    $done_field{$_}++;
		$start_pos{$_}++;
	}
    }
    
    # 'gene info':
    if (defined $selected_fields{'gene info'}) {
	foreach ("$COMMON_GENE_NAME","$COMMON_GENE_NAME_SOURCE", "$COMMON_GENE_DESCRIPTION","$COMMON_GENE_DESCRIPTION_SOURCE") {
	    push @start_pos, $_;
	    $done_field{$_}++;
	    $start_pos{$_}++;
	}
	# hide the sources
	foreach ("$COMMON_GENE_NAME_SOURCE", "$COMMON_GENE_DESCRIPTION_SOURCE") {
	    $hidden_fields{$_}++;
	}
    }
    
    
    # ANNOTATION section:
    if (defined $selected_fields{'everything'}
	or
	defined $selected_fields{'all annotations'}) {
	foreach (@annotation_fields) {
	    my $field = $_;
	    next if (defined $done_field{$field});
	    push @annotation_pos, $field;
#	    warn "# adding '$field'";
	}
    }
    foreach (@annotation_pos) {
	$done_field{$_}++;
	$annotation_pos{$_}++;
    }

    # NUMBERS section
    # 'all values'
    if (defined $selected_fields{'everything'}
	or
	defined $selected_fields{'all values'}) {
	foreach my $field (@output) {
	    next if (defined $done_field{$field});
	    push @values_pos, $field;
	    $done_field{$field}++;
#	    $values_pos{$field}++;
	}
    }

    # fold changes:
    if (defined $selected_fields{'fold changes'}) {
	my $field = $FOLD_CHANGE_TERM;
	unless (defined $done_field{$field}) {
	    push @values_pos, $field;
	    $done_field{$field}++;
#	    $values_pos{$field}++;
	}
    }

    # ratios:
    if (defined $selected_fields{'ratios'}) {
	my $field = $RATIO_TERM;
	unless (defined $done_field{$field}) {
	    push @values_pos, $field;
	    $done_field{$field}++;
#	    $values_pos{$field}++;
	}
    }

    # p-values etc.:
    if (defined $selected_fields{'significance values'}) {
	foreach my $field (@output) {
	    next if (defined $done_field{$field});
	    if ($field =~ /p-/
		or
		$field =~ /F-/
		or
		$field =~ /Z-/) {
		push @values_pos, $field;
		$done_field{$field}++;
#		$values_pos{$field}++;
	    }
	}
    }

    # normalized values:
    if (defined $selected_fields{'normalized values'}) {
	foreach my $field (@output) {
	    next if (defined $done_field{$field});
	    if ($field =~ /norm\b/i
		or
		$field =~ /dehub\b/i) {
		push @values_pos, $field;
		$done_field{$field}++;
#		$values_pos{$field}++;
	    }
	}
    }

    # merged values:
    if (defined $selected_fields{'merged values'}) {
	foreach my $field (@output) {
	    next if (defined $done_field{$field});
	    if ($field =~ /_M\b/i
		or
		$field =~ /$MERGED_RATIOS_TERM\b/
		or
		$field =~ /DUP_DIFF/) {
		push @values_pos, $field;
		$done_field{$field}++;
#		$values_pos{$field}++;
	    }
	}
    }

    # background corrected values:
    if (defined $selected_fields{'bkgd corrected values'}) {
	foreach my $field (@output) {
	    next if (defined $done_field{$field});
	    if ($field =~ /$FG_COR_SIGNAL_TERM\b/i
		or
		$field =~ /BG_CALC\b/i) {
		push @values_pos, $field;
		$done_field{$field}++;
#		$values_pos{$field}++;
	    }
	}
    }

    # raw values:
    if (defined $selected_fields{'raw values'}) {
	foreach my $field (@output) {
	    next if (defined $done_field{$field});
	    if ($field =~ /^$FG_SIGNAL_TERM/i
		or
		$field =~ /^$BG_SIGNAL_TERM/i) {
		push @values_pos, $field;
		$done_field{$field}++;
#		$values_pos{$field}++;
	    }
	}
    }

    # flags:
    if (defined $selected_fields{'flags'}) {
	unless (defined $done_field{$FLAG_CHANNEL_TERM}) {
	    my $field = $FLAG_CHANNEL_TERM;
	    push @values_pos, $field;
	    $done_field{$field}++;
#	    $values_pos{$field}++;
	}
    }

    # same for the hidden fields:
    if (defined $hidden_fields{'everything'}) {
	foreach ('Probe ID', $SPOT_ID_TERM, 'i', 'MROW','MCOL','ROW','COL') {
	    next if ($done_field{$_});
	    push @start_pos, $_;
	    $done_field{$_}++;
	    $start_pos{$_}++;
	    $hidden_fields{$_}++;
	}
    }    

    # 'probe and spot ids':
    if (defined $hidden_fields{'probe and spot ids'}) {
	foreach ('Probe ID', $SPOT_ID_TERM) {
	    next if ($done_field{$_});
	    push @start_pos, $_;
	    $done_field{$_}++;
	    $start_pos{$_}++;
	    $hidden_fields{$_}++;
	}
    }

    # 'spot coordinates':
    if (defined $hidden_fields{'spot coordinates'}) {
	foreach ('i', 'MROW','MCOL','ROW','COL') {
	    next if ($done_field{$_});
	    push @start_pos, $_;
	    $done_field{$_}++;
	    $start_pos{$_}++;
	    $hidden_fields{$_}++;
	}
    }
    
    # 'gene info':
    if (defined $hidden_fields{'gene info'}) {
	foreach ("$COMMON_GENE_NAME","$COMMON_GENE_NAME_SOURCE", "$COMMON_GENE_DESCRIPTION","$COMMON_GENE_DESCRIPTION_SOURCE") {
	    next if ($done_field{$_});
	    push @start_pos, $_;
	    $done_field{$_}++;
	    $start_pos{$_}++;
	    $hidden_fields{$_}++;
	}
    }
    
    # ANNOTATION section:
    if (defined $hidden_fields{'everything'}
	or
	defined $hidden_fields{'all annotations'}) {
	foreach (@annotation_fields) {
	    my $field = $_;
	    next if (defined $done_field{$field});
	    push @annotation_pos, $field;
	    $hidden_fields{$_}++;
#	    warn "# hiding '$field'";
	}
    }
    foreach (@annotation_pos) {
	$done_field{$_}++;
	$annotation_pos{$_}++;
    }

    # NUMBERS section
    # 'all values'

    if (defined $hidden_fields{'everything'}
	or
	defined $hidden_fields{'all values'}) {
	foreach my $field (@output) {
	    next if (defined $done_field{$field});
	    push @values_pos, $field;
	    $done_field{$field}++;
#	    $values_pos{$field}++;
	    $hidden_fields{$field}++;
	}
    }

    # fold changes:
    if (defined $hidden_fields{'fold changes'}) {
	my $field = $FOLD_CHANGE_TERM;
	unless (defined $done_field{$field}) {
	    push @values_pos, $field;
	    $done_field{$field}++;
#	    $values_pos{$field}++;
	    $hidden_fields{$field}++;
	}
    }

    # ratios:
    if (defined $hidden_fields{'ratios'}) {
	my $field = $RATIO_TERM;
	unless (defined $done_field{$field}) {
	    push @values_pos, $field;
	    $done_field{$field}++;
#	    $values_pos{$field}++;
	    $hidden_fields{$field}++;
	}
    }

    # p-values etc.:
    if (defined $hidden_fields{'significance values'}) {
	foreach my $field (@output) {
	    next if (defined $done_field{$field});
	    if ($field =~ /p-/
		or
		$field =~ /F-/
		or
		$field =~ /Z-/) {
		push @values_pos, $field;
		$done_field{$field}++;
#		$values_pos{$field}++;
		$hidden_fields{$field}++;
	    }
	}
    }
    
    # normalized values:
    if (defined $hidden_fields{'normalized values'}) {
	foreach my $field (@output) {
	    next if (defined $done_field{$field});
	    if ($field =~ /norm\b/i
		or
		$field =~ /dehub\b/i) {
		push @values_pos, $field;
		$done_field{$field}++;
#		$values_pos{$field}++;
		$hidden_fields{$field}++;
	    }
	}
    }

    # merged values:
    if (defined $hidden_fields{'merged values'}) {
	foreach my $field (@output) {
	    next if (defined $done_field{$field});
	    if ($field =~ /_M\b/i
		or
		$field =~ /$MERGED_RATIOS_TERM\b/
		or
		$field =~ /DUP_DIFF/) {
		push @values_pos, $field;
		$done_field{$field}++;
#		$values_pos{$field}++;
		$hidden_fields{$field}++;
	    }
	}
    }

    # background corrected values:
    if (defined $hidden_fields{'bkgd corrected values'}) {
	foreach my $field (@output) {
	    next if (defined $done_field{$field});
	    if ($field =~ /$FG_COR_SIGNAL_TERM\b/i
		or
		$field =~ /BG_CALC\b/i) {
		push @values_pos, $field;
		$done_field{$field}++;
#		$values_pos{$field}++;
		$hidden_fields{$field}++;
	    }
	}
    }

    # raw values:
    if (defined $hidden_fields{'raw values'}) {
	foreach my $field (@output) {
	    next if (defined $done_field{$field});
	    if ($field =~ /^$FG_SIGNAL_TERM/i
		or
		$field =~ /^$BG_SIGNAL_TERM/i) {
		push @values_pos, $field;
		$done_field{$field}++;
#		$values_pos{$field}++;
		$hidden_fields{$field}++;
	    }
	}
    }

    # flags:
    if (defined $hidden_fields{'flags'}) {
	unless (defined $done_field{$FLAG_CHANNEL_TERM}) {
	    my $field = $FLAG_CHANNEL_TERM;
	    push @values_pos, $field;
	    $done_field{$field}++;
#	    $values_pos{$field}++;
	    $hidden_fields{$field}++;
	}
    }


    # read from file
    # if columns were selected that are common to all files, extract them from the first file:
    my $collect_common_info = 1;
    my @order = ();
    my %done_order = ();
    my %start_out = ();
    my %annotation_out = ();
    my %value_out = ();
#    @files = ('/www/koch.pathogenomics.ca/data/arraypipe/internal/results/FPMI/_26773/merge_C_T1_10_14.4.1.1.2.10._out.txt.tmp');
#    @files = ('/www/koch.pathogenomics.ca/data/arraypipe/internal/results/FPMI/_26773/file1', 
#	      '/www/koch.pathogenomics.ca/data/arraypipe/internal/results/FPMI/_26773/file2', 
#	      '/www/koch.pathogenomics.ca/data/arraypipe/internal/results/FPMI/_26773/file3'
#);

    my $value_field_number = 0;

    foreach my $file (sort @files) {
        unless (open (IN, $file)) {
	    warn "Can't read '$file': $!";
	    next;
	}

	my @headers = &get_header(*IN);

	my %header_positions = ();
	
	my $counter = 0;
	foreach my $field (@headers) {
	    $header_positions{$field} = $counter;
	    $counter++;	
	}

	# reduce header arrays
	my @start_pos_tmp = ();
	foreach my $field (@start_pos) {
	    if (defined $header_positions{$field}) {
		push @start_pos_tmp, $field;
	    }
	}

	my @annotation_pos_tmp = ();
	foreach my $field (@annotation_pos) {
	    if (defined $header_positions{$field}) {
		push @annotation_pos_tmp, $field;
	    }
	}

	my @values_pos_tmp = ();
	my %values_pos_tmp = ();
	foreach my $field (@values_pos) {
	    foreach my $header_field (sort keys %header_positions) {
		if ($header_field =~ /\b$field\b/) {
		    push @{$values_pos_tmp{$field}}, $header_field;
		    next;
		}
	    }
	    push @values_pos_tmp, $field if (defined $values_pos_tmp{$field});
	}

	$value_field_number = 0;
	foreach my $field (@values_pos_tmp) {
	    @{$values_pos_tmp{$field}} = sort @{$values_pos_tmp{$field}};
	    $value_field_number += scalar @{$values_pos_tmp{$field}};
	}

#	warn "# merge fields: @merge_fields";

	while (<IN>) {
	    chomp;
	    my @tmp = split /\t/, $_;

	    # combine merge fields into a key:
	    my @merge = ();
	    foreach my $field (@merge_fields) {
		my $pos = $header_positions{$field};
		unless (defined $tmp[$pos]) {
		    die "No field '$field' in $file!";
		}
		my $key = $tmp[$pos];
		# padd numerical key with zeros, so that we can sort them alphabetically
		unless ($key =~ /\D/) {
		    until (length($key) > 4) {
			$key = '0'.$key;
		    }
		}
		push @merge, $key;
	    }
	    my $merge_key = join '|', @merge;
	    unless ($done_order{$merge_key}) {
		push @order, $merge_key;
		$done_order{$merge_key}++;
		$collect_common_info = 1;
	    } else {
		$collect_common_info = 0;
	    }

#	    warn "# merge: '$merge_key'";

	    # collect common information (unless done already):
	    if ($collect_common_info) {

		# first the start fields
		foreach my $field (@start_pos_tmp) {
		    my $pos = $header_positions{$field};
		    unless (defined $tmp[$pos]) {
			next;
		    }
		    push @{$start_out{$merge_key}}, $tmp[$pos];
		    $start_pos{$field}++;
		}

		# then the annotation fields:
		foreach my $field (@annotation_pos_tmp) {
		    my $pos = $header_positions{$field};
		    unless (defined $tmp[$pos]) {
			next;
		    }
		    push @{$annotation_out{$merge_key}}, $tmp[$pos];
		    $annotation_pos{$field}++;
		}
	    }
	    foreach my $field (@values_pos_tmp) {
		foreach my $header_field (@{$values_pos_tmp{$field}}) {
		    my $pos = $header_positions{$header_field};
		    unless (defined $tmp[$pos]) {
#			next;
			$tmp[$pos] = '';
		    }
		    if (($header_field eq $field  # this is covered by the second pattern match
			 or                       # but might speed things up for direct matches
			 $header_field =~ /\b$field\b/)
			) {
			push @{$value_out{$merge_key}{$file}}, $tmp[$pos];
			unless (defined $values_pos{$file}{$header_field}) {
			    $values_pos{$file}{$header_field} = 1;
			}
			if ($values_pos{$file}{$header_field} == 1) {
			    push @{$values_pos_new{$file}}, $header_field;
			}
			$values_pos{$file}{$header_field}++;
		    }
		}
	    }
	    next;
	}
	close IN;
    }

    # write merged output to file:
    my $output_file = "$tmp_dir/output_$action_step";
    my $step_ext = '';
    if (-e $results_file) {
	$step_ext .= '.0';
    }
    while (-e $output_file.$step_ext) {
	$step_ext =~ /(\d+)$/;
	my $num = $1 + 1;
	$step_ext =~ s/\d+$/$num/;
    }
    $output_file .= "$step_ext.tab";
    unless (open OUT, ">$output_file") {
	die "Can't write to $output_file: $!";
    }
    system "chgrp $common_user_group '$output_file' 2> /dev/null";
    system "chmod 666 '$output_file' 2> /dev/null";	

    # write header:
    my @headers_out = ();
    my $column_counter = 0;
    foreach my $field (@start_pos) {
	if ($start_pos{$field} > 1) {
	    push @headers_out, $field;
	    $hidden_fields_num{$column_counter}++ if (defined $hidden_fields{$field});
	    $column_counter++;
	}
    }

    my %occurrence = ();
    foreach my $file (sort @files) {
	my $file_out = '';
	if (@files > 1) {
	    $file_out = basename $file;
	    if ($file_out =~ /(.*)_$progname/) {
		$file_out = $1;
	    } elsif ($file_out =~ /(.*)_step\d+/) {
		$file_out = $1;
	    } elsif ($file_out =~ /^$MERGED_FILE_NAME(.*?)_(.*?)_/) {
		$file_out = "$1 $2";
	    }
	    $file_out .= ' ';
	}
	foreach my $field (@{$values_pos_new{$file}}) {
#	    if ($values_pos{$field} > 1) {
	    my $out = "$file_out$field";
	    if ($field =~ / vs .*p-\w+/) {
		my $occurrence = '';
		if (defined $occurrence{$field}) {
		    $occurrence = " $occurrence{$field}";
		    $occurrence{$field}++;
#		    warn "# KH: occ($field): $occurrence{$field}";
		} else {
		    $occurrence{$field} = 1;
		}
		$out = "$field$occurrence";
	    }
		push @headers_out, $out;
		if (defined $hidden_fields{$field}) {
		    $hidden_fields_num{$column_counter}++;
		} else {
		    foreach my $hidden (keys %hidden_fields) {
			if ($field =~ /\b$hidden\b/) {
			    $hidden_fields_num{$column_counter}++;
			}
		    }
		}
		$column_counter++;
	    }
#	}
    }

    foreach my $field (@annotation_pos) {
	if ($annotation_pos{$field} > 1) {
	    push @headers_out, $field;
	    $hidden_fields_num{$column_counter}++ if (defined $hidden_fields{$field});
	    $column_counter++;
	}
    }
    
    for (my $i = 0; $i <= $#headers_out; $i++) {
	my $header = $headers_out[$i];
	if ($header =~ /^${MERGED_FILE_NAME}(\w.*?)_(\w.*?)__.*?\.tab( .*)/) {
	    my $cond1 = $1;
	    my $cond2 = $2;
	    my $rest = $3;
	    if ($cond1 eq 'C') {
		$header = $cond2.$rest;
	    } else {
		$header = "$cond1 vs $cond2$rest";
	    }
	    $headers_out[$i] = $header;
	} elsif ($header =~ /^${MERGED_FILE_NAME}(\w.*?)_(\w.*?)__.*?\.txt( .*)/) {
	    my $cond1 = $1;
	    my $cond2 = $2;
	    my $rest = $3;
	    if ($cond1 eq 'C') {
		$header = $cond2.$rest;
	    } else {
		$header = "$cond1 vs $cond2$rest";
	    }
	    $headers_out[$i] = $header;
	}

	# other header substitutions:
	if ($header =~ /p-anova/) {
	    $header = 'ANOVA p-value';
	    $headers_out[$i] = $header;
	} elsif ($header =~ /F-value/) {
	    $header = 'ANOVA F-value';
	    $headers_out[$i] = $header;
	} elsif ($header =~ /$RATIO_DIFF_TERM$/) {
	    $header = "$RATIO_DIFF_TERM";
	    $headers_out[$i] = $header;
	}
    }

    print OUT (join "\t", @headers_out)."\n";

    # write data:
    foreach my $merge_key (sort keys %done_order) {
	
	my @out = ();
	
	if (defined $start_out{$merge_key}) {
	    push @out, @{$start_out{$merge_key}};		       
	} else {
	    warn "# WARNING: No start information for $merge_key!";
	    next;
	}

	if (defined $value_out{$merge_key}) {
	    foreach my $file (sort @files) {
#	    foreach my $file (sort keys %{$value_out{$merge_key}}) {
		if (defined @{$value_out{$merge_key}{$file}}) {
		    push @out, @{$value_out{$merge_key}{$file}};
		} else {
		    foreach (1..$value_field_number) {
			push @out, '';
		    }
		}
	    }
	}

	if (defined $annotation_out{$merge_key}) {
	    push @out, @{$annotation_out{$merge_key}};
	}

	print OUT (join "\t", @out)."\n";
    }
    close OUT;

    %$ref = %hidden_fields_num;

    return $output_file;
}


sub get_header {
    my $fh = shift;
    my @header = ();
    while (<$fh>) {
	# remove comments (everything following a hash)
	my ($line,$rest) = split /\#/; 	
	
	# skip if nothing's on the line
	next unless ($line =~ /\w/);
	
	chomp $line;

	# we have reached the header line
	@header = split /\t/, $line;
	last;
    }

    return @header;
}


sub time_formated {
    my $time = shift;
    my $time_out = '';
                                                                                                                             
    if ($time >= 86400) {
        my $days = int($time / 86400);
        $time = $time % 86400;
        $time_out .= "$days day";
	if ($days > 1) {
	    $time_out .= 's';
	}
	$time_out .= ' ';
    }
    if ($time >= 3600) {
        my $hours = int($time / 3600);
        $time = $time % 3600;
        $time_out .= "$hours hour";
	if ($hours > 1) {
	    $time_out .= 's';
	}
	$time_out .= ' ';
    }
    if ($time > 60) {
        my $min = int($time / 60);
        $time_out .= "$min minute";
	if ($min > 1) {
	    $time_out .= 's';
	}
	$time_out .= ' ';
        $time = $time % 60;
    }
    if ($time > 0) {
        $time_out .=  "$time second";
	if ($time > 1) {
	    $time_out .= 's';
	}
	$time_out .= ' ';
    }
    chop $time_out;
    return $time_out;
}


sub file_size {
    my $file = shift;
    my $size = '?';

    unless (-e $file) {
	return $size;
    } else {
	$size = (stat($file))[7];
    }
    
    my @units = ('Bytes', 'KB', 'MB', 'GB');
    my $unit = shift @units;
    while ($size > 1024
	   and
	   @units) {
	$unit = shift @units;
	$size /= 1024;
    }
    $size = sprintf "%.1f", $size;

    return "$size\&nbsp;$unit";
}

    
sub check_for_added_data {
    my $file_add = "$tmp_dir/data_step_$action_step.${PID}.add";

    # check if data has been added
    unless (%work_files) {
	return '';
    }

    # make sure data has been added:
    my $new_data = 0;
    OUTER:foreach my $file (keys %data_done) {
	foreach my $label (keys %{$data_done{$file}}) {
	    foreach my $header (keys %{$data_done{$file}{$label}}) {
		if (defined $file_input{$file}{$label}{$header}
		    and $file_input{$file}{$label}{$header}) {
		    next;
		} else {
		    $new_data = 1;
		    last OUTER;
		}
	    }
	}
    }
    if ($new_data == 0) {
	return '';
    }
	

    # make sure we can write to file
    unless (open (ADD, ">$file_add")) {
	warn "# $progname: Can't write to '$file_add': $!";
	$file_add = '';
    } else {
	close ADD;
    }

    return $file_add;
}

sub header_bar {

    my $ref = shift;
    my %args = %$ref;

    my $out = '';
    my %labels_def = ();
    my %link_def = ('input',
		    'INFO',
		    'info',
		    'info_section'
		      );

    my $title = $args{'title'} || '';
    my $colour = $args{'colour'} || '';
    my $anchor = $args{'anchor'} || '';

    my @links = ();
    if (defined $args{'links'}) {
	@links = @{$args{'links'}};
    }			 

    my %labels = ();
    if (defined $args{'labels'}) {
	%labels = %{$args{'labels'}};
    }

    my $width = $args{'width'} || $main_width;

    my $links = '';
    if (@links) {
	$links = "<CENTER><small>\n";
	foreach my $link (@links) {	    
	    my $label = '['.$link.']';
	    if (defined $labels_def{$link}) {
		$label = $labels_def{$link};
	    }
	    if (defined $labels{$link}) {
		$label = $labels{$link};
	    }
	    
	    my $link_out = $link;
	    if (defined $link_def{$link}) {
		$link_out = $link_def{$link};
	    }
	    $links .= "<A HREF=\"#$link_out\">$label</A>\n";
	}
	$links .= "</SMALL></CENTER>\n";
    }
	    
	    
    $out .=  "\n<TABLE WIDTH=$width><TR><TD>\n<TABLE WIDTH=$width cellpadding=\"6\"><TR BGCOLOR=$colour><TD><CENTER><BIG><B><A NAME=\"$anchor\">$title</A></B></BIG></CENTER></TD></TR></TABLE>\n$links</TD></TR></TABLE>\n";
    
    return $out;
}


sub write_output {

    my %done = ();
    my %available = ();
    my @link_order = ('INFO','INPUT','DATA FILES','RESULTS','ACTION SETUP','SETTINGS');
    my %link_tag = ('INFO', 'info',
		    'INPUT', 'input',
		    'DATA FILES', 'file',
		    'RESULTS', 'results',
		    'ACTION SETUP', 'action',
		    'SETTINGS', 'parameters');

    # standard output order:
    my @default_order = ('WARNINGS','INFO','RESULTS','ACTION SETUP','SETTINGS','DATA FILES','INPUT','FINISH');

    # push list of loaded files to front
    # if newly loaded files are available:
    if ($new_files) {
	@default_order = ('WARNINGS','INFO','DATA FILES','RESULTS','ACTION SETUP','SETTINGS','INPUT','FINISH');
    }
    
    my @order = ();
    foreach (@default_order) {
	if ($output_collect{$_}) {
	    push @order, $_;
	    $available{$_}++;
	}
    }
    
    my @links = ();
    foreach (@link_order) {
	if ($available{$_}) {
	    push @links, $_;
	}
    }


    # print start of page
    if ($output_collect{'START'}) {
	print $output_collect{'START'} unless ($results_section);
    }

    foreach my $section (@order) {

	if ($results_section) {
	    next if ($section eq 'RESULTS'
		     or
		     $section eq 'INFO'
		     or
		     $section eq 'START');
	}

	next if ($done{$section});
	$done{$section}++;

	my @link_list = ();
	foreach (@links) {
	    push @link_list, $_ unless ($_ eq $section);
	}

	print "\n<!-- START of $section section -->\n";
	unless ($section eq 'FINISH') {
	    print &header_bar(
			      {
				  title => $section,
				  colour => 'lightgrey',
				  anchor => $section,
				  links => [ @link_list ]
				  }
			      );
	}
	print "\n<BR>".$output_collect{$section}."\n<BR>\n";;
	print "<!-- END of $section section -->\n";
    }
}


sub write_add_data {

    my $file_add = shift;

    unless (open (ADD, ">$file_add")) {
	warn "# $progname: Can't write to '$file_add': $!";
	return;
    }

    my $date = localtime;
    my $prefix = "# output from $progname, version $revision_number at $date\n";
    $prefix .= "# Headers: $FG_SIGNAL_TERM = foreground, $BG_SIGNAL_TERM = background, $FLAG_CHANNEL_TERM = flag, $FG_COR_SIGNAL_TERM = background-corrected foreground, '_L' = log-transferred, '_R' = ratio, '_I' = intensity product, '_M' = merged\n";
    $prefix .= "# Headers: $FG_SIGNAL_TERM = foreground, $BG_SIGNAL_TERM = background, $FLAG_CHANNEL_TERM = flag, $FG_COR_SIGNAL_TERM = background-corrected foreground, '_L' = log-transferred, '_R' = ratio, '_I' = intensity product, '_M' = merged\n";
    foreach my $file (sort @files) {
	foreach my $key (sort keys %{$file_specs{$file}}) {
	    my $val = $file_specs{$file}{$key};
#	    warn "# val '$val' = '".(ref($val))."'";
	    if (ref($val) eq 'ARRAY') {
		$val = join ' | ',  @{$file_specs{$file}{$key}};
	    }
	    $prefix .= "# $key = $val\n";
	}
	if ($channel_swaps{$file}) {
	    $prefix .= "# CHANNELS WERE SWAPPED = 1\n";
	}
    }
    $prefix .= "# DATA SECTION:\n";

    print ADD $prefix;

    my @files =  (sort keys %data);

    my $counter = 0;
    foreach my $file (sort @files) {
	print ADD "# FILE$counter\t'$file'";
	$counter++;
	my $counterl = 0;
	foreach my $label (sort {$b cmp $a} keys %{$data{$file}{'__DATA__'}}) {
	    print ADD "\tLABEL$counterl\t'$label'";
	    $counterl++;
	}
	print ADD "\n";
    }

    my @labels = ();
    foreach my $label (sort {$b cmp $a} keys %{$data{$files[0]}{'__DATA__'}}) {
	push @labels, $label;
    }

    my @header = ('MROW','MCOL','ROW','COL');
    
    my @data_header = ();

    my %valid_headers = ();
    
    foreach my $file (@files) {
	foreach my $header (@output) {
	    foreach my $label (@labels) {
		unless (defined $data{$file}{'__DATA__'}{$label}{$header}) {
		    next;
		}
		unless (defined $data_done{$file}{$label}{$header}) {
		    next;
		}

		my $column = "$label __ $header __ $file";
		push @data_header, $column;

		$valid_headers{$file}{$header}{$label}++;
	    }
	}
    }

    print ADD (join "\t", (@header, @data_header))."\n";
    
    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@files);

    foreach my $mrow (1..$MROW) {
	foreach my $mcol (1..$MCOL) {
	    foreach my $row (1..$ROW) {
		foreach my $col (1..$COL) {

		    my @out = ();
			
		    foreach my $file (@files) {
			foreach my $header (@output) {
			    foreach my $label (@labels) {
				
				next unless (defined $valid_headers{$file}{$header}{$label});

				my $val = '';
				if (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}) {
				    $val = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
				}
				push @out, $val;
			    }
			}
		    }
		    unless (scalar @out == scalar @data_header) {
			warn "# $progname: mismatch at line $mrow,$mcol,$row,$col: ".(scalar @out)." != ".(scalar @data_header)."!";
			@out = ();
		    }

		    if (@out) {
			print ADD "$mrow\t$mcol\t$row\t$col\t".(join "\t", @out)."\n";
		    }
		}
	    }
	}
    }

    close ADD;
}


sub get_array_specs {
    my @files = @_;
    my ($MROW,$MCOL,$ROW,$COL) = (0,0,0,0);

    foreach my $file (@files) {

	foreach my $key ('MROW', 'MCOL', 'ROW', 'COL') {

	    if (defined $cgi->param("${key}_$file")
		and
		$cgi->param("${key}_$file") =~ /\d/) {
		$file_specs{$file}{$key} = $cgi->param("${key}_$file");
	    }
	}

	if ($file_specs{$file}{'MROW'} > $MROW) {
	    $MROW = $file_specs{$file}{'MROW'};
	}
	if ($file_specs{$file}{'MCOL'} > $MCOL) {
	    $MCOL = $file_specs{$file}{'MCOL'};
	}
	if ($file_specs{$file}{'ROW'} > $ROW) {
	    $ROW = $file_specs{$file}{'ROW'};
	}
	if ($file_specs{$file}{'COL'} > $COL) {
	    $COL = $file_specs{$file}{'COL'};
	}

	unless ($MROW =~ /^\d+$/
		and
		$MROW > 0
		and
		$MCOL =~ /^\d+$/
		and
		$MCOL > 0
		and
		$ROW =~ /^\d+$/
		and
		$ROW > 0
		and
		$COL =~ /^\d+$/
		and
		$COL > 0) {

#	    die "Grid specification for $file not complete!";
# can't do this because it is called from &read_data, while the temporary refresh page is in place -> ergo we would never see a change
# have to find a mechanism to proliferate this to the results page
# or maybe warn at a much earlier stage
# for the moment I just set the layout manually
	    ($MROW,$MCOL,$ROW,$COL) = (1,1,100,100);
	}
    }

    return ($MROW,$MCOL,$ROW,$COL);
}


sub save_settings {

    # check if an action has been specified:
    my $action = 0;
    my $action_counter = 0;
    SAVE:for my $i (1..$max_actions) {
	for my $k (1..$action_alternatives) {
	    $action_counter++;
	    if ($cgi->param("action$action_counter")) {
		$action++;
		last SAVE;
	    }
	}
    }
    
    unless (-d $save_settings_dir) {
	system "mkdir '$save_settings_dir'; chmod 777 '$save_settings_dir' 2> /dev/null";
    }
    my $file = $cgi->param('save_settings_name');
    unless ($file) {
	return "<BR><BR><B><BIG><FONT COLOR=RED>WARNING:</FONT></BIG></B><BR>Please specify a name for these settings!</B>\n";
    }
    if (-e "$save_settings_dir/$file") {
	return "<BR><BR><B><BIG><FONT COLOR=RED>WARNING:</FONT></BIG></B><BR>Saved settings '$file' exist already, please go back and specify a different name!</B>\n";
    }
    unless (open (OUT, ">$save_settings_dir/$file")) {
	return "<BR><BR><B><BIG><FONT COLOR=RED>WARNING:</FONT></BIG></B><BR>Can't write to $save_settings_dir/$file: $!</B>";
    }

    # print info
    my $date = localtime;
    print OUT "# $progname version $revision_number\n# $file saved by $remote_address at $date\n";
    my $note = $cgi->param('save_settings_note');
    if ($note eq '<Add Description>') {
	$note = '';
    }
    print OUT "# NOTE: \n$note\n";
    
    # get all actions
    $action_counter = 0;
    for my $i (1..$max_actions) {
	for my $k (1..$action_alternatives) {
	    $action_counter++;
	    if ($cgi->param("action$action_counter")) {
		print OUT "# ACTION $action_counter ".$cgi->param("action$action_counter")."\n";
	    }
	}
    }
    
    # get all variables
    foreach my $var (keys %var_def_type) {
	if (defined $cgi->param($var)) {
	    foreach ($cgi->param($var)) {
		print OUT "# VARIABLE $var $_\n";
	    }
	}
    }

    # other important variables:
    print OUT "# VARIABLE max_actions $max_actions\n";
    print OUT "# VARIABLE action_alternatives $action_alternatives\n";

    close OUT;
	    
    system "chmod 666 '$save_settings_dir/$file' 2> /dev/null";

    return "<BR><BR><B>Settings saved as '$file'.</B>\n";
}
	    
	

sub load_settings {

    my $apply = shift || '';

    unless (-d $save_settings_dir) {
	system "mkdir '$save_settings_dir'; chmod 777 '$save_settings_dir' 2> /dev/null";
    }
    my $file = $cgi->param('load_settings');
    unless ($file) {
	return "<BR><BR><B><BIG><FONT COLOR=RED>WARNING:</FONT></BIG></B><BR>Please specify a name for settings!</B>\n";
    }
    unless (open (IN, "$save_settings_dir/$file")) {
	return "<BR><BR><B><BIG><FONT COLOR=RED>WARNING:</FONT></BIG></B><BR>Can't open $save_settings_dir/$file: $!</B>\n";
    }

    my %vars = ();

    while (<IN>) {
	chomp;
	if (/^\# ACTION (\d+) (.*)/) {
	    my $action_num = $1;
	    my $action = $2;
	    $saved_settings_action{$action_num} = $action;
	}
	if (/^\# VARIABLE (\w+) (.*)/) {
	    my $var = $1;
	    my $value = $2;
	    unless (defined $value) {
		$value = '';
	    }
	    push @{$saved_settings_var{$var}}, $value;
	}
    }
    close IN;

    if ($apply) {

	# set actions
	foreach my $action_num (sort keys %saved_settings_action) {
	    my $action = $saved_settings_action{$action_num};
		$cgi->param(-name => "action$action_num",
			    -values => $action);
#	    push @actions, $action;
	}

	# set variables
	foreach my $key (sort keys %saved_settings_var) {
	    my @list = @{$saved_settings_var{$key}};
	    $cgi->param(-name => $key,
			-values => [@list]);
	}
    }
	

    return "<BR><BR><B>Retrieved ".(scalar keys %saved_settings_action)." actions and ".( scalar keys %saved_settings_var)." parameter settings from '$file'.</B>\n" if ($verbose);
    return '';
}
	    
	

sub kill_processes {
    my $prev_pid = $cgi->param('last_pid');    
    
    my @pids = ();
    if (opendir(DIR, '/proc/')) {
	while (defined(my $proc = readdir(DIR))) {
	    next if ($proc =~ /\D/);
	    unless (open (PROC, "/proc/$proc/status")) {
		next;
	    }
	    my $ppid = '';
	    while (<PROC>) {
		if (/PPid:\s*(\d+)/) {
		    $ppid = $1;
		    last;
		}
	    }
	    close PROC;
#	    warn "got ppid $ppid!";
	    if ($ppid =~ /\d/
		and
		$ppid == $prev_pid) {
		push @pids, $proc;
		warn "found prog: $proc!\n";
	    }
	}
    } else {
	warn "#WARNING: $progname: Can't read /proc directory!\n";
    } 

    warn "Match with $prev_pid!\n";
    
    foreach my $pid (sort { $a <=> $b } @pids) {
	next if ($pid == $$);
	warn "# killing $pid!\n";
    }

    exit;
}


sub get_system_load {
    my @procs = `cat /proc/cpuinfo`;
    my $procs = 0;
    foreach (@procs) {
	if (/^processor\s+\:\s*(\d+)/) {
	    $procs = $1;
	}
    }
    $procs++;


    my $loadavg = `cat /proc/loadavg`;
    my @tmp = split /\s+/, $loadavg,4;
    pop @tmp;
    foreach (@tmp) {
	$_ /= $procs if (/\d/);
    }

    return $tmp[0];
}

sub adjust_to_cutoffs {
    my $action_num = shift;
    my $action = shift;
    my @files = @_;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $header = ''; 
    my @work_labels = ('CH1','CH2');
    foreach my $column (@{$item_data_column{$action}}) {
	my $found = 1;
	foreach my $file (@files) {
	    foreach my $label (@work_labels) {
		unless ($data_column_available{$file}{$label}{$column}) {
		    $found = 0;
		}
	    }
	}
	if ($found) {
	    $header = $column;
	    last;
	}
    }

    # check for specific settings:
    $header = &get_user_settings($header,'data_column',$force_global_opt,$action_num,$files[0]);

    unless ($header) {
	return "Error: No data column defined to work on for $action!\n";
    }
    
    $last_column_used = $header;

    foreach my $file (@files) {
	&read_data($file,$header);
    }

    my @flags = &get_user_settings($var_def_default{'cutoff_skip_flags'},'cutoff_skip_flags',$force_global_opt,$action_num,$files[0]);
    my %skip_flags = ();
    %skip_flags = &get_skip_flags(@flags);

    my $allowed_min = &get_user_settings($var_def_default{'cutoff_allowed_min'},'cutoff_allowed_min',$force_global_opt,$action_num,$files[0]);
    my $allowed_min_spots = 0;
    my $lower = &get_user_settings($var_def_default{'cutoff_min'},'cutoff_min',$force_global_opt,$action_num,$files[0]);
    my $upper = &get_user_settings($var_def_default{'cutoff_max'},'cutoff_max',$force_global_opt,$action_num,$files[0]);
    my $lower_action = &get_user_settings($var_def_default{'cutoff_min_action'},'cutoff_min_action',$force_global_opt,$action_num,$files[0]);
    my $upper_action = &get_user_settings($var_def_default{'cutoff_max_action'},'cutoff_max_action',$force_global_opt,$action_num,$files[0]);

    my $upper_changed = 0;
    my $lower_changed = 0;
    my %skipped = ();

    my ($mrow,$mcol,$row,$col,$data);

    my ($all_min,$all_max);

    my %skip_lines = ();
    my %skip_spots = ();
    my %values = ();

    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@files);

    if ($lower_action eq 'global shift') {
	# find overall minimum

	# collect all data and skip rows that are not defined or flagged

	foreach $mrow (1..$MROW) {
	    foreach $mcol (1..$MCOL) {
		foreach $row (1..$ROW) {
		    COL:foreach $col (1..$COL) {
			
			# skip rows that have entry without value
			foreach my $file (@files) {
			    foreach my $label (@work_labels) {
				unless (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}) {
				    $skip_lines{$mrow}{$mcol}{$row}{$col}++;
				    next COL;
				}
				my $data = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
				if ($data =~ /[^\d\.\-\+e]/) {
				    $skip_lines{$mrow}{$mcol}{$row}{$col}++;
				    next COL;
				}
			    }
			}			

			if (%skip_flags) {
			    foreach my $file (@files) {
				foreach my $label (@work_labels) {
				    
				    # skip flagged spots
				    if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}
					or
					defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
					my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
					 if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
					     push @flags, &get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
					 }
					foreach (@flags) {
					    if (defined $skip_flags{$_}) {
						$skip_lines{$mrow}{$mcol}{$row}{$col}++;
						next COL;
					    }
					}
				    }		
				}
			    }
			}
			
			foreach my $file (@files) {
			    foreach my $label (@work_labels) {
				my $data = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};	
				# skip spots below the allowed minimum
				if ($data < $allowed_min) {
				    $allowed_min_spots++;

				    # flag file unless flag is set already:
				    my $skip_flagging = 0;
				    if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}
					and
					$data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} >= $flag_values{'floor'}) {
					my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
					foreach (@flags) {
					    if ($_ == $flag_values{'floor'}) {
						$skip_flagging++;
						last;
					    }
					}
				    }
				    $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'floor'} unless ($skip_flagging);

				    $skip_spots{$mrow}{$mcol}{$row}{$col}{$file}{$label}++;
				    next;
				}
				push @{$values{"$mrow|$mcol|$row|$col"}}, $data;
			    }
			}
		    }
		  }
	    }
	}
	
	$all_min = 999999;
	my @values = ();

	foreach $mrow (1..$MROW) {
	    foreach $mcol (1..$MCOL) {
		foreach $row (1..$ROW) {
		    foreach $col (1..$COL) {
			next if (defined $skip_lines{$mrow}{$mcol}{$row}{$col});
			
			if (defined $values{"$mrow|$mcol|$row|$col"}) {
			    push @values, @{$values{"$mrow|$mcol|$row|$col"}};
			}
		    }
		}
	    }
	}
	
	$all_min = (sort { $a <=> $b } @values)[0];
    }

			    
			   
    # transform data
    my $unflagged_spots = 0;

    foreach $mrow (1..$MROW) {
	foreach $mcol (1..$MCOL) {
	    foreach $row (1..$ROW) {
	      COL:foreach $col (1..$COL) {
		  
		  # skip rows that have entry without value
		  foreach my $file (@files) {
		      foreach my $label (@work_labels) {
			  unless (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}) {
			      $skipped{'undefined'}++;
			      next COL;
			  }
			  my $data = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
			  if ($data =~ /[^\d\.\-\+e]/) {
			      $skipped{$data}++;
			      next COL;
			  }
		      }
		  }			
		  
		  if (%skip_flags) {
		      foreach my $file (@files) {
			  foreach my $label (@work_labels) {
			      
			      # skip flagged spots
			      if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}
				  or
				  defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
				  my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
				  if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
				      push @flags, &get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
				  }
				  foreach (@flags) {
				      if (defined $skip_flags{$_}) {
					  next COL;
				      }
				  }
			      }		
			  }
		      }
		  }

		  foreach my $file (@files) {
		      foreach my $label (@work_labels) {
			  
			  # adjust data
			  
			  next unless (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col});
			  my $data = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
			  next unless ($data =~ /\d/);
			  # spot is below allowed limit
			  next if (defined $skip_spots{$mrow}{$mcol}{$row}{$col}{$file}{$label});

			  if ($lower_action eq 'global shift'
			      and
			      $all_min < $lower) {
			      $lower_changed++;
			      $data = $data - $all_min + $lower;
			  }
			    
			  if ($data < $lower) {
			      $lower_changed++;
			      if ($lower_action eq 'flag') {
				  my $skip_flagging = 0;
				  # check if floor flag is set already
				  if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}
				      and
				      $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} >= $flag_values{'floor'}) {
				      my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
				      foreach (@flags) {
					  if ($_ == $flag_values{'floor'}) {
					      $skip_flagging++;
					      last;
					  }
				      }
				  }
				  $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'floor'} unless ($skip_flagging);
			      } else {
				  $data = $lower;
			      }
			  } elsif ($data > $upper) {
			      $upper_changed++;
			      if ($upper_action eq 'flag') {
				  $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'ceiling'};
			      } else {
				  $data = $upper;
			      }
			  }
			  
			  # check if floor flag is set:
			  if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}
			      and
			      $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} >= $flag_values{'floor'}) {
			      my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
			      my $floor_flagged = 0;
			      foreach (@flags) {
				  if ($_ == $flag_values{'floor'}) {
				      $floor_flagged = 1;
				      last;
				  }
			      }
			      if ($data >= $lower
				  and
				  $floor_flagged) {
				  # remove floor flag
				  $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} -= $flag_values{'floor'};
				  $unflagged_spots++;
			      }
			  }

			  $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col} = $data;
		      }
		  }
	      }
	    }
	}
    }

    my $out = '';

    if ($lower_changed) {
	$out .= "Applied $lower_action to $lower_changed values ".($lower_action eq 'global shift'? "by ".(abs($lower - $all_min)) : "below $lower")."<BR>";
    } 
    if ($unflagged_spots) {
	$out .= "Removed 'floor' flag from $unflagged_spots spots.<BR>";
    } 
    if ($upper_changed) {
	$out .= "Applied $upper_action to $upper_changed values above $upper<BR>";
    } 
    if (%skipped) {
	$out .= "The following data points were skipped:<BR><UL>\n";
	foreach (sort keys %skipped) {
	    $out .= "<LI>$_ -> $skipped{$_} entries</LI>\n";
	}
	if ($allowed_min_spots) {
	    $out .= "<LI>below $allowed_min -> $allowed_min_spots entries</LI>\n";
	}
	$out .= "</UL>\n";
    }

    $out = 'No data points where flagged or adjusted!<BR>' if ($out eq '');

    return $out;
	
}



sub output_file {
    my $action_num = shift;
    my $action = shift;
    my $output_format = shift;
    my $full_html_counter = shift;
    my $ref = shift;
    my @files = ();
    my @check_files = ();

    if ($output_format eq 'TIGR MEV') {
	push @files, shift @_;
	@check_files = @_;
    } else {
	@files = @_;
	@check_files = @files;
    }

    unless (@files) {
	push @files, (sort keys %data)[0];
	@check_files = @files;
    }

    my $extra_info = '';
	
    if (@files == 1) {
	if ($files[0] eq 'annotation') {
	    $extra_info .= "annotation";
	} else {
	    $extra_info .= (basename $files[0]);
	}
    }

    my $date = localtime;
    my $prefix = "# output from $progname, version $revision_number at $date\n";
    $prefix .= "# Headers: $FG_SIGNAL_TERM = foreground, $BG_SIGNAL_TERM = background, $FLAG_CHANNEL_TERM = flag, $FG_COR_SIGNAL_TERM = background-corrected foreground, '_L' = log-transferred, '_R' = ratio, '_I' = intensity product, '_M' = merged\n";
    foreach my $file (sort @files) {
	next unless (defined $file_specs{$file});
	foreach my $key (sort keys %{$file_specs{$file}}) {	    
	    my $val = $file_specs{$file}{$key};
#	    warn "# val '$val' = '".(ref($val))."'";
	    if (ref($val) eq 'ARRAY') {
		$val = join ' | ',  @{$file_specs{$file}{$key}};
	    }
	    $prefix .= "# $key = $val\n";
	}
	if ($channel_swaps{$file}) {
	    $prefix .= "# CHANNELS WERE SWAPPED = 1\n";
	}
    }

    $prefix .= "# DATA SECTION:\n";

    if ($refine_parameters) {
	return 'Parameter Refinement!<BR>';
    }
	
    my @output = ($ref);
    if (ref($ref) eq 'ARRAY') {    
	@output = @$ref;
    }

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $suffix = '.tab';
    if ($output_format eq 'TIGR MEV') {
	$suffix = '.mev';
	$prefix = "# Data from ArrayPipe stored in TIGR's MEV format:\n";
	if ($files[0] eq 'annotation') {
	    $prefix = "# Annotation data from ArrayPipe stored in TIGR's MEV format:\n";
	    $suffix = '.ann';
	}
    } elsif ($output_format eq 'BASE') {
	$suffix = '.bas';
	$prefix = '';
    }

    my $base_name = basename $files[0];
    my $out_file = "$tmp_dir/$base_name";
    my $out_file_html = $base_name;
    if ($base_name =~ /^($MERGED_FILE_NAME.*\d\.)txt/) { 
	$out_file = "$tmp_dir/$1".'_out.tab';
	$out_file_html = $1.'_out.tab';
    } else {
	$out_file .= "_step$action_step.$action_num";
	$out_file_html .= "_step$action_step.$action_num";    
	$out_file .= $suffix;
	$out_file_html .= $suffix;
    }
    
    if (-e $out_file) {
	unlink $out_file
	    or return "Can't remove '$out_file': $!\n";
    }
    
    # open output file
    open (OUT, ">$out_file")
	or return "<BR><B>Can't write to $out_file: $!</B><BR>\n";
    print OUT $prefix;


    # print annotation if requested
    if ($files[0] eq 'annotation') {
	$output_format = 'MEV_annotation';
    }

    # collect all available labels
    my @labels = ();

    foreach my $label (sort keys %{$data{$files[0]}{'__DATA__'}}) {
	push @labels, $label;
    }


    my @columns = &get_user_settings($var_def_default{'output_columns'},'output_columns',$force_global_opt,$action_num,$files[0]);
    if ($columns[0] eq 'all') {
#	@columns = ('Probe ID', 'i', 'MROW', 'MCOL', 'ROW', 'COL', @output, 'Description');
	@columns = ('Probe ID', 'i', 'MROW', 'MCOL', 'ROW', 'COL', @output);
    }

    my $fold_change_header = '';
    my %fold_change_header = ();
    if ((grep /^$FOLD_CHANGE_TERM$/, @columns)
	and
	$output_format !~ /MEV/) {
	# add fold-change column
	# check what the highest-level intensity values are
	my @data_columns = ($MERGED_RATIOS_TERM, $MERGED_RATIOS_TERM.'_L', 'Dehub_M_L_R','Dehub_M_L','Dehub_L_M_R','Dehub_L_M','Dehub_L_R_M', 'Dehub_L_R','Dehub_L','Dehub_M_R','Dehub_M','Dehub_R','Dehub','Norm_M_L_R','Norm_M_L','Norm_L_M_R','Norm_L_M','Norm_L_R_M','Norm_L_R','Norm_L','Norm_M_R','Norm_M','Norm_R_M','Norm_R','Norm',$FG_COR_SIGNAL_TERM.'_M_R',$FG_COR_SIGNAL_TERM.'_M',$FG_COR_SIGNAL_TERM.'_L_R',$FG_COR_SIGNAL_TERM.'_L',$FG_COR_SIGNAL_TERM.'_R',$FG_COR_SIGNAL_TERM,$FG_SIGNAL_TERM.'_M_L_R',$FG_SIGNAL_TERM.'_M_L',$FG_SIGNAL_TERM.'_L_M_R',$FG_SIGNAL_TERM.'_L_M',$FG_SIGNAL_TERM.'_M_R',$FG_SIGNAL_TERM.'_M',$FG_SIGNAL_TERM.'_L_R_M',$FG_SIGNAL_TERM.'_L_R',$FG_SIGNAL_TERM.'_R_L',$FG_SIGNAL_TERM.'_L',$FG_COR_SIGNAL_TERM.'_R_M',$FG_COR_SIGNAL_TERM.'_M_R',$FG_SIGNAL_TERM.'_R',$FG_SIGNAL_TERM);
#	my @data_columns = reverse @output;
	FILE:foreach my $file (@files) {
	    foreach (@data_columns) {
		if (
		    (/_R/
		     or
		     /$MERGED_RATIOS_TERM/)
		    and
		    defined $data_done{$file}{'__NO_LABEL__'}{$_}) {
		    $fold_change_header{$file} = $_;	
		    next FILE;
		} elsif (defined $data_done{$file}{'CH1'}{$_}
			 and
			 defined $data_done{$file}{'CH2'}{$_}) {
		    $fold_change_header{$file} = $_;
		    next FILE;
		}		
	    }
	    
	    unless ($fold_change_header{$file}) {
		# assign foreground signal:
		$fold_change_header{$file} = $FG_SIGNAL_TERM;
		# and make sure it's available for each file:
		foreach my $file (@files) {	    
		    &read_data($file,$FG_SIGNAL_TERM);
		}
	    }
	}
    }    

    # calculate fold change
    if (%fold_change_header) {	
	foreach my $file (@files) {
	    my $fold_change_header = $fold_change_header{$file};
	    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);
	    foreach my $mrow (1..$MROW) {
		foreach my $mcol (1..$MCOL) {
		    foreach my $row (1..$ROW) {
			foreach my $col (1..$COL) {
			    # initialize field:
			    my $fold_change = '';
			    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FOLD_CHANGE_TERM}{$mrow}{$mcol}{$row}{$col} = $fold_change;
			    my $val1 = '';
			    my $val2 = '';
			    if ($fold_change_header =~ /_R/
				or
				$fold_change_header =~ /$MERGED_RATIOS_TERM/) {
				if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$fold_change_header}{$mrow}{$mcol}{$row}{$col}) {
				    $fold_change = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$fold_change_header}{$mrow}{$mcol}{$row}{$col};
				}
				next unless ($fold_change =~ /\d/);
				# reverse log
				if ($fold_change_header =~ /_L/) {
				    $fold_change = 2**$fold_change;
				}

			    } else {
				if (defined $data{$file}{'__DATA__'}{'CH1'}{$fold_change_header}{$mrow}{$mcol}{$row}{$col}) {
				    $val1 = $data{$file}{'__DATA__'}{'CH1'}{$fold_change_header}{$mrow}{$mcol}{$row}{$col};
				}
				if (defined $data{$file}{'__DATA__'}{'CH2'}{$fold_change_header}{$mrow}{$mcol}{$row}{$col}) {
				    $val2 = $data{$file}{'__DATA__'}{'CH2'}{$fold_change_header}{$mrow}{$mcol}{$row}{$col};
				}
				next unless ($val1 =~ /\d/
					     and
					     $val2 =~ /\d/);
				if ($fold_change_header =~ /_L/) {
				    $fold_change = $val2 - $val1;
				    $fold_change = 2**$fold_change;
				} else {
				    next if ($val1 == 0);
				    $fold_change = $val2 / $val1;
				}
			    }

			    if ($fold_change < 1) {
				if ($fold_change == 0) {
				    next;
				} else {
				    $fold_change = -1 * 1/$fold_change;
				}
			    }
			    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FOLD_CHANGE_TERM}{$mrow}{$mcol}{$row}{$col} = $fold_change;
			}
		    }
		}
	    }
    
	    unless (defined $data_column_available{$file}{'__NO_LABEL__'}{$FOLD_CHANGE_TERM}) {
		$data_column_available{$file}{'__NO_LABEL__'}{$FOLD_CHANGE_TERM} = 1;
		push @{$data_column_available{$file}{'__NO_LABEL__'}{'__ORDER__'}}, $FOLD_CHANGE_TERM;
	    }
	    $data_done{$file}{'__NO_LABEL__'}{$FOLD_CHANGE_TERM}++;
	}
    }
    

    my $ratio_header = '';
    my %ratio_header = ();
    if ((grep /^$RATIO_TERM$/, @columns)
	and
	$output_format !~ /MEV/) {
	# add ratio column
	# check what the highest-level intensity values are
	my @data_columns = ($MERGED_RATIOS_TERM, $MERGED_RATIOS_TERM.'_L', 'Dehub_M_L_R','Dehub_M_L','Dehub_L_M_R','Dehub_L_M','Dehub_L_R_M', 'Dehub_L_R','Dehub_L','Dehub_M_R','Dehub_M','Dehub_R','Dehub','Norm_M_L_R','Norm_M_L','Norm_L_M_R','Norm_L_M','Norm_L_R_M','Norm_L_R','Norm_L','Norm_M_R','Norm_M','Norm_R_M','Norm_R','Norm',$FG_COR_SIGNAL_TERM.'_M_R',$FG_COR_SIGNAL_TERM.'_M',$FG_COR_SIGNAL_TERM.'_L_R',$FG_COR_SIGNAL_TERM.'_L',$FG_COR_SIGNAL_TERM.'_R',$FG_COR_SIGNAL_TERM,$FG_SIGNAL_TERM.'_M_L_R',$FG_SIGNAL_TERM.'_M_L',$FG_SIGNAL_TERM.'_L_M_R',$FG_SIGNAL_TERM.'_L_M',$FG_SIGNAL_TERM.'_M_R',$FG_SIGNAL_TERM.'_M',$FG_SIGNAL_TERM.'_L_R_M',$FG_SIGNAL_TERM.'_L_R',$FG_SIGNAL_TERM.'_R_L',$FG_SIGNAL_TERM.'_L',$FG_COR_SIGNAL_TERM.'_R_M',$FG_COR_SIGNAL_TERM.'_M_R',$FG_SIGNAL_TERM.'_R',$FG_SIGNAL_TERM);
#	my @data_columns = reverse @output;
	FILE:foreach my $file (@files) {
	    foreach (@data_columns) {
		if (
		    (/_R/
		     or
		     /$MERGED_RATIOS_TERM/)
		    and
		    defined $data_done{$file}{'__NO_LABEL__'}{$_}) {
		    $ratio_header{$file} = $_;	
		    next FILE;
		} elsif (defined $data_done{$file}{'CH1'}{$_}
			 and
			 defined $data_done{$file}{'CH2'}{$_}) {
		    $ratio_header{$file} = $_;
		    next FILE;
		}		
	    }
	    
	    unless ($ratio_header{$file}) {
		# assign foreground signal:
		$ratio_header{$file} = $FG_SIGNAL_TERM;
		# and make sure it's available for each file:
		foreach my $file (@files) {	    
		    &read_data($file,$FG_SIGNAL_TERM);
		}
	    }
	}
    }    

    # calculate ratio
    if (%ratio_header) {	
	foreach my $file (@files) {
	    my $ratio_header = $ratio_header{$file};
	    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);
	    foreach my $mrow (1..$MROW) {
		foreach my $mcol (1..$MCOL) {
		    foreach my $row (1..$ROW) {
			foreach my $col (1..$COL) {
			    # initialize field:
			    my $ratio = '';
			    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$RATIO_TERM}{$mrow}{$mcol}{$row}{$col} = $ratio;
			    my $val1 = '';
			    my $val2 = '';
			    if ($ratio_header =~ /_R/
				or
				$ratio_header =~ /$MERGED_RATIOS_TERM/) {
				if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$ratio_header}{$mrow}{$mcol}{$row}{$col}) {
				    $ratio = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$ratio_header}{$mrow}{$mcol}{$row}{$col};
				}
				next unless ($ratio =~ /\d/);
				# reverse log
				if ($ratio_header =~ /_L/) {
				    $ratio = 2**$ratio;
				}

			    } else {
				if (defined $data{$file}{'__DATA__'}{'CH1'}{$ratio_header}{$mrow}{$mcol}{$row}{$col}) {
				    $val1 = $data{$file}{'__DATA__'}{'CH1'}{$ratio_header}{$mrow}{$mcol}{$row}{$col};
				}
				if (defined $data{$file}{'__DATA__'}{'CH2'}{$ratio_header}{$mrow}{$mcol}{$row}{$col}) {
				    $val2 = $data{$file}{'__DATA__'}{'CH2'}{$ratio_header}{$mrow}{$mcol}{$row}{$col};
				}
				next unless ($val1 =~ /\d/
					     and
					     $val2 =~ /\d/);
				if ($ratio_header =~ /_L/) {
				    $ratio = $val2 - $val1;
				    $ratio = 2**$ratio;
				} else {
				    next if ($val1 == 0);
				    $ratio = $val2 / $val1;
				}
			    }

			    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$RATIO_TERM}{$mrow}{$mcol}{$row}{$col} = $ratio;
			}
		    }
		}
	    }
    
	    unless (defined $data_column_available{$file}{'__NO_LABEL__'}{$RATIO_TERM}) {
		$data_column_available{$file}{'__NO_LABEL__'}{$RATIO_TERM} = 1;
		push @{$data_column_available{$file}{'__NO_LABEL__'}{'__ORDER__'}}, $RATIO_TERM;
	    }
	    $data_done{$file}{'__NO_LABEL__'}{$RATIO_TERM}++;
	}
    }
    

    my %out_header = ();
    my %columns = ();
    foreach (@columns) {
	$columns{$_}++;
    }

    # create headers for each header/label combination
    foreach my $header (@output) {
	my $found = 0;
	foreach my $label (@labels) {
	    my $label_out = $label.' ';
	    if ($label eq '__NO_LABEL__') {
		$label_out = '';
	    }
#	    if (defined %{$data{$files[0]}{'__DATA__'}{$label}{$header}} ) {
	    my $id_header = 0;
	    foreach my $file (@files) {
		if ($data_column_available{$file}{$label}{$header} 
		    and
		    $data_done{$file}{$label}{$header}
		    ) {
		    $found++;
		    if ($header eq $SPOT_ID_TERM) {
			next if ($id_header > 0);
			$out_header{$header}{$label}{$file} = "$label_out$header";			
			$id_header++;
		    } else {
			if (@files > 1) {
			    my $basename = basename $file;
			    if ($basename =~ /^merge_(.*?)_(.*?)_/) {
				$basename = "$1 vs. $2";
			    }
			    $out_header{$header}{$label}{$file} = $basename." $label_out$header";
			} else {
			    $out_header{$header}{$label}{$file} = "$label_out$header";
			}
		    }
		}
	    }
	}
	# if header doesn't exist for any column, remove from columns list
	if ($found == 0
	    and
	    defined $columns{$header}) {	    
	    $columns{$header} = 0;
	}
    }

    # print headers to file (including coordinates column)
    my @headers = ();
    foreach my $header ('Probe ID', 'i', 'MROW', 'MCOL', 'ROW', 'COL') {

	# skip unless column was selected
	next unless (defined $columns{$header});

	push @headers, $header;
    }
	
    foreach my $header (@output) {

	# skip unless column was selected
	next unless (defined $columns{$header}
		     and
		     $columns{$header});
	
	foreach my $label (sort keys %{$out_header{$header}} ) {

	    foreach my $file (@files) {
		if (defined $out_header{$header}{$label}{$file}) {
		    my $header = $out_header{$header}{$label}{$file};
		    if ($header =~ /^(CH[1-2]) /) {
			my $label = $1;
			if (defined $file_specs{$file}{"${label}_LABEL"}) {
			    my $new_label = $file_specs{$file}{"${label}_LABEL"};
			    $header =~ s/^$label /$new_label /;
			}
		    }
		    push @headers, $header;
		}
	    }
	}
    }

    push @headers, 'Description', if (defined $columns{'Description'}
				      and
				      $columns{'Description'});
    	
    # special output header for TIGR's MEV format
    if ($output_format eq 'TIGR MEV') {
	@headers = ('UID','IA','IB','R','C','SR','SC','MR','MC');
    } elsif ($output_format eq 'MEV_annotation') {
	@headers = ('UID','row','col', $SPOT_ID_TERM);
    }

    # add headers for annotation
    my @annotation_headers = ();
#    if (%annotation) {
#	my %done = ();
#	foreach my $item (@{$var_def_values{'annotate_items_probelynx'}}, @{$var_def_values{'annotate_items_funcdb'}}) {
#	    # skip dividers
#	    next if ($item =~ /--.*--/);

#	    # skip those items that weren't selected	    
#	    next unless (defined $annotation_select_probelynx{$item}
#			 or
#			 defined $annotation_select_funcdb{$item});
#	    # skip those items that had no entries
#	    next unless (defined $annotation_done{$item});

#	    # report each column only once:
#	    unless (defined $done{$item}) {
#		$done{$item}++;
#	    } else {
#		next;
#	    }
#	    
#	    if (defined $annotation_item_match{$item}) {
#		my $item2 = $annotation_item_match{$item};
#		unless (defined $done{$item2}) {
#		    $done{$item2}++;
#		} else {
#		    next;
#		}
#	    }

#	    # add other items to the header
#	    push @annotation_headers, $item;
#	}
#    }

    my @extra_column = ();
    my %column_usage = ();
    foreach my $key (keys %{$file_specs{$files[0]}}) {
	next if ($key eq 'ADD');
	if ($key eq 'FLAG') {
	    foreach ( @{$file_specs{$files[0]}{$key}} ) {
		$column_usage{$_}++;
	    } 
	} else {
	    $column_usage{$file_specs{$files[0]}{$key}}++;
	}
    }
    if (defined $file_specs{$files[0]}{'ADD'}) {
	foreach my $header (@{$file_specs{$files[0]}{'ADD'}}) {
	    if ($column_usage{$header}) {
		next;
	    }
	    push @extra_column, $header;
	}
    }

    
    if ($output_format eq 'TIGR MEV') {
	# mev data files do not contain annotation
	@annotation_headers = ();
    } else {
	@annotation_headers = @annotation_headers_all;
	if ($output_format eq 'MEV_annotation') {
	    # mev annotation files do not contain the extra data columns
	    @extra_column = ();    
	    # we're going to read the spot id's from the first file in the data set:
	    @files = (sort keys %data)[0];
	    @check_files = (sort keys %data)[0];

	    if ($output_format eq 'MEV_annotation') {
		%columns = ($SPOT_ID_TERM, 1);
		@output = ();
		@labels = ('__NO_LABEL__');
	    }

	}
    }

    my @gene_add = ();
    unless ($output_format eq 'TIGR MEV'
	    or
	    $output_format eq 'MEV_annotation') {
	if (defined $annotation{$COMMON_GENE_NAME}) {
#	@headers = ($COMMON_GENE_NAME, @headers);
	    push @gene_add, $COMMON_GENE_NAME;
	    push @gene_add, $COMMON_GENE_NAME_SOURCE;
	}
	if (defined $annotation{$COMMON_GENE_DESCRIPTION}) {
#	@headers = ($COMMON_GENE_DESCRIPTION, @headers);
	    push @gene_add, $COMMON_GENE_DESCRIPTION;
	    push @gene_add, $COMMON_GENE_DESCRIPTION_SOURCE;
	}
    }
    print OUT join( "\t", @gene_add,@headers,@annotation_headers,@extra_column)."\n";
	
    my $data_content_header = '';
    my $last_column_used = '';
    if ($last_column_used) {
	$last_column_used = $last_column_used;
    }
    OUTPUT:foreach my $column ($last_column_used, @{$item_data_column{$action}}) {
	foreach my $label (@labels) {
	    if ($data_column_available{$files[0]}{$label}{$column}) {
		$data_content_header = $column;
		last OUTPUT;
	    }
	}
    }
     # check for specific settings:
    $data_content_header = &get_user_settings($data_content_header,'output_data_column',$force_global_opt,$action_num);
    
    # other variables:
#    my $skip_empty = &get_user_settings($var_def_default{'output_skip_empty'},'output_skip_empty',$force_global_opt,$action_num,$files[0]);
    my $skip_empty = 0;
    my $rounding = &get_user_settings($var_def_default{'output_rounding'},'output_rounding',$force_global_opt,$action_num,$files[0]);
    
    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'output_skip_flags'},'output_skip_flags',$force_global_opt,$action_num,$files[0]);
    %skip_flags = &get_skip_flags(@flags);

    my %out = ();
    my %lines = ();
    my %skip_lines = ();
    my ($mrow,$mcol,$row,$col,$data);


    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@check_files);

    foreach $mrow (1..$MROW) {
	foreach $mcol (1..$MCOL) {
	    foreach $row (1..$ROW) {
		foreach $col (1..$COL) {
		    
		    # calculate line number
		    my $line = ($row - 1 + ($mrow - 1) * $ROW) * $MCOL*$COL + ($mcol - 1) * $COL + $col;
		    # keep track of conversion
		    $lines{$line} = "$mrow,$mcol,$row,$col" unless (defined $lines{$line});
		    
		    # at least one of the following headers should be defined
		    CHECK:foreach my $column ($SPOT_ID_TERM, $FG_SIGNAL_TERM, $FLAG_CHANNEL_TERM) {
			foreach my $label (@labels) {
			    foreach my $file (@check_files) {
				if (defined $data{$file}{'__DATA__'}{$label}{$column}{$mrow}{$mcol}{$row}{$col}) {
				    $out{$line}++;
				    last CHECK;
				}
			    }
			}
		    }
		}
	    }
	}
    }

    # check for flags set:
    if (%skip_flags) {
	my $header = $FLAG_CHANNEL_TERM;

	my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@check_files);

	foreach my $label (@labels) {
	    next unless (defined $out_header{$header}{$label}
			 or
			 $output_format eq 'MEV_annotation'
			 );

	    foreach $mrow (1..$MROW) {
		foreach $mcol (1..$MCOL) {
		    foreach $row (1..$ROW) {
			foreach $col (1..$COL) {

			    # calculate line number
			    my $line = ($row - 1 + ($mrow - 1) * $ROW) * $MCOL*$COL + ($mcol - 1) * $COL + $col;

			    my $skip = 0;
			  FILE:foreach my $file (@check_files) {
			      next unless (defined $out_header{$header}{$label}{$file}
					   or
					   $output_format eq 'MEV_annotation'
					   );
			      # skip flagged spots
			      my @flag_labels = ('__NO_LABEL__', 'CH1', 'CH2');
			      my $flag_defined = 0;
			      my @flags = ();
			      foreach my $flag_label (@flag_labels) {
				  if (defined $data{$file}{'__DATA__'}{$flag_label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
				      push @flags, &get_flags($data{$file}{'__DATA__'}{$flag_label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}, 'add_zero');
				      $flag_defined++;
				  }
			      }

			      if ($flag_defined) {
				  # if we have flag '0' and others, remove '0'
				  my %tmp = ();
				  foreach (@flags) {
				      $tmp{$_}++;
				  }
				  my @tmp = sort keys %tmp;
				  if (@tmp > 1
				      and 
				      $tmp[0] == 0) {
				      shift @tmp;				      
				  }
				  @flags = @tmp;

				  foreach (@flags) {
				      if (defined $skip_flags{$_}) {
					  $skip = 1;
					  last FILE;
				      }
				  }
			      } else {
				  if (defined $skip_flags{0}) {
				      $skip = 1;
				      last FILE;
				  }
			      }
			  }
			    if ($skip) {
				$skip_lines{$line}++;
			    }
			}
		    }
		}
	    }
	}
    }


    foreach my $header (@output) {
	
	# check if content needs to be present
	my $check_content = 0;
	if ($header eq $data_content_header
	    and
	    $skip_empty) {

	    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@check_files);
	    foreach $mrow (1..$MROW) {
		foreach $mcol (1..$MCOL) {
		    foreach $row (1..$ROW) {
			foreach $col (1..$COL) {
			    
    			    # calculate line number
			    my $line = ($row - 1 + ($mrow - 1) * $ROW) * $MCOL*$COL + ($mcol - 1) * $COL + $col;
			    
			    next if (defined $skip_lines{$line});
				     
			    foreach my $label (@labels) {
				next unless (defined $out_header{$header}{$label});
				my $got_val = 0;
				my $skip = 0;
				foreach my $file (@check_files) {
				    if (!defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}) {
					$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col} = '';
				    }
				    if ($data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col} eq '') {
					$skip++;
				    } else {
					$got_val++;
				    }
				}
				if ($output_format eq 'TIGR MEV') {
				    # one value is enough for TIGR MEV - report them for all files
				    if ($got_val) {
					$out{$line}++;
				    } else {					
					$skip_lines{$line}++;
				    }
				} else {
				    # if one value is undefined remove entries from all files
				    if ($skip) {					
					$skip_lines{$line}++;
				    } else {
					$out{$line}++;
				    }			    
				}
			    }
			}
		    }
		}
	    }
	}
    }
    
    # get olido_ids
    if (defined $columns{'Probe ID'}) {
	&get_probe_ids($files[0],$MROW,$MCOL,$ROW,$COL);
    }

    my $counter = 0;
    my $val = '';

    my @output_headers = @output;
    
    if ($output_format eq 'TIGR MEV') {

	my @MEV_labels = ();
	foreach (@labels) {
	    next if ($_ eq '__NO_LABEL__');
	    push @MEV_labels, $_;
	}

	# check if control and treatment have been identified
	if (defined $cgi->param("CH1_ANNOTATE_$files[0]")
	    and
	    $cgi->param("CH1_ANNOTATE_$files[0]") =~ /^T/) {
	    @labels = reverse @labels;
	}	    
	
	# get the best suitable output headers:
	my $header = $FG_SIGNAL_TERM;
	HEAD:foreach my $column (@{$item_data_column{$output_format}}) {
	    foreach my $label (sort {$b cmp $a} keys %{$data{$files[0]}{'__DATA__'}}) {
		if (defined $data_column_available{$files[0]}{$label}{$column}) {
		    $header = $column;
		    last HEAD;
		}
	    }
	}
	@output_headers = ($header);

	$extra_info .= ", header = <B>$header</B>\n";
	if ($header =~ /_L/) {
	    $extra_info .= " (log has been reversed)\n";
	}
	    
	
	foreach my $line (sort { $a <=> $b } keys %out) {
	    
	    next unless ($line =~ /\d/);
	    
	    next if (defined $skip_lines{$line});
	    
	    my ($mrow,$mcol,$row,$col) = split /,/, $lines{$line};
	    my $started = 0;
	    
	    # WARNING: something specific to our array
	    # split spot labels at '^' into id and description )
	    my $id = '';
	    my $descr = '';;
	    if (defined $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col}) {
		if ($FPMI_specific) {
		    if ( $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col} =~ /\^/ ) {
			my $val = $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col};
			$val =~ s/^\^+//g;
			($id,$descr) = split /\^/, $val, 2;
		    } else {
			$id = $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col};
		    }
		} else {
		    $id = $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col};
		}
	    } else {
		$id = '';
		$descr = '';
	    }
	    
	    my @out = ();
	    my $got_val = 0;
	    my $data_file = $files[0];
	    foreach my $header (@output_headers) {
		foreach my $label (@MEV_labels) {
		    # make sure at least one of the check files has data:
		    foreach my $file (@files) {
			if (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}
			    and
			    $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col} =~ /\d/) {
			    $val = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
			    $got_val++;
			    if ($header =~ /_L/) {
				# undo log2:
				$val = 2**$val;
			    }
			    if ($val =~ /\.(\d+)/) {
				if (length($1) > $rounding) {
				    $val = sprintf "%.${rounding}f", $val;
				}
			    }
			} else {
			    $val = 0; # if both channels are '0' this will result in grey boxes in MEV
#			    $val = 'NA';
			}

			push @out, $val;
		    }
		}
	    }
	    
	    # skip lines where all entries are empty or undefined
	    unless ($got_val) {
#		next;
	    }
	    
	    # print Probe ID if available
#	    if (%probe_ids) {
#		if (defined $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col}) {
#		    print OUT $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col};
#		} else {
#		    print OUT $id;
#		}
#	    } else {
		print OUT $id;
#	    }

	    # print values
	    print OUT "\t".(join "\t", @out);
	
	    # print coordinates
	
	    print OUT "\t$row";
	    print OUT "\t$col";
	    print OUT "\t$mrow";	    
	    print OUT "\t$mcol";
	    print OUT "\t$mrow";	    
	    print OUT "\t$mcol";

	    print OUT "\n";
	    $counter++;
	
	}
	close OUT;

    } else {

	foreach my $line (sort { $a <=> $b } keys %out) {
	    
	    next unless ($line =~ /\d/);
	    
	    next if (defined $skip_lines{$line});
	    
	    my ($mrow,$mcol,$row,$col) = split /,/, $lines{$line};
	    my $started = 0;
	    
	    # WARNING: something specific to our array
	    # split spot labels at '^' into id and description )
	    my $id = '';
	    my $descr = '';
	    if (defined $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col}) {
		my $val = $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col};
		$val =~ s/^\^+//;
		($id,$descr) = split /\^/, $val, 2;
	    }
	    unless (defined $id) {
		$id = '';
	    }
	    # remove number at end of description
	    unless (defined $descr) {
		$descr = '';
	    }
	    $descr =~ s/ \(\d+\)\s*$//;
	    
	    # print reporter id if we are producing a MEV annotation file
	    if ($output_format eq 'MEV_annotation') {
		print OUT "$id\t$row\t$col\t$id";
		$started = 1;
	    }
	    
	    # print common gene name
	    if (defined $annotation{$COMMON_GENE_NAME}) {
		
		if ($started == 1) {
		    print OUT "\t";
		} else {
		    $started = 1;
		}	    

		# get probe
		my $probe = '';
		if (defined $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col}) {
		    $probe = $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col};
		}
		
		my $annotation_item = $COMMON_GENE_NAME;
		my $annotation = 'unknown';
		my $annotation_type = 'unknown';

		if ($probe
		    and		    
		    defined $annotation{$annotation_item}{$probe}
		    and
		    @{$annotation{$annotation_item}{$probe}} > 0) {
		    my @tmp = ();
		    my %tmp = ();
		    # remove duplicate entries
		    foreach (@{$annotation{$annotation_item}{$probe}}) {
			next unless (defined $_);
			$tmp{$_}++;
		    }
		    # and sort
		    @tmp = (sort keys %tmp);
		    $annotation = join '; ', @tmp;
		    $annotation_type = $annotation{$COMMON_GENE_NAME_SOURCE}{$probe};
		} 
		print OUT "$annotation\t$annotation_type";
	    }

	    # print common gene description
	    if (defined $annotation{$COMMON_GENE_DESCRIPTION}) {
		
		if ($started == 1) {
		    print OUT "\t";
		} else {
		    $started = 1;
		}	    

		# get probe
		my $probe = '';
		if (defined $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col}) {
		    $probe = $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col};
		}
		
		my $annotation_item = $COMMON_GENE_DESCRIPTION;
		my $annotation = 'unknown';
		my $annotation_type = 'unknown';

		if ($probe
		    and		    
		    defined $annotation{$annotation_item}{$probe}
		    and
		    @{$annotation{$annotation_item}{$probe}} > 0) {
		    my @tmp = ();
		    my %tmp = ();
		    # remove duplicate entries
		    foreach (@{$annotation{$annotation_item}{$probe}}) {
			next unless (defined $_);
			$tmp{$_}++;
		    }
		    # and sort
		    @tmp = (sort keys %tmp);
		    $annotation = join '; ', @tmp;
		    $annotation_type = $annotation{$COMMON_GENE_DESCRIPTION_SOURCE}{$probe};
		} 
		print OUT "$annotation\t$annotation_type";
	    }

	    # print Probe ID if available
	    if (defined $columns{'Probe ID'}) {
		if ($started == 1) {
		    print OUT "\t";
		} else {
		    $started = 1;
		}	    
		if (%probe_ids) {
		    if (defined $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col}) {
			print OUT $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col};
		    } else {
			if ($output_format eq 'BASE') {
			    print OUT $id;
			} else {
			    print '';
			}
		    }
		} else {
		    print OUT $id;
		}
	    }

	    # print coordinates
	    if (defined $columns{'i'}
		and
		$columns{'i'}) {	    
		if ($started == 1) {
		    print OUT "\t";
		} else {
		    $started = 1;
		}
		print OUT "$line";
	    }
	    
	    if (defined $columns{'MROW'}
		and
		$columns{'MROW'}) {	
		if ($started == 1) {
		    print OUT "\t";
		} else {
		    $started = 1;
		}
		if ($output_format eq 'BASE') {
		    #combine MROW and MCOL into a single block nominator:
		    if (defined $columns{'MCOL'}
			and
			$columns{'MCOL'}) {	
			print OUT "$mrow,$mcol";
		    } else {
			print OUT "$mrow";
		    }
		} else {		    
		    print OUT "$mrow";	    
		}
	    }

	    if (defined $columns{'MCOL'}
		and
		$columns{'MCOL'}) {	
		if ($started == 1) {
		    print OUT "\t";
		} else {
		    $started = 1;
		}
		if ($output_format eq 'BASE') {
		    # see above
		    print OUT '';
		} else {
		    print OUT "$mcol";
		}
	    }

	    if (defined $columns{'ROW'}
		and
		$columns{'ROW'}) {	
		if ($started == 1) {
		    print OUT "\t";
		} else {
		    $started = 1;
		}
		print OUT "$row";
	    }

	    if (defined $columns{'COL'}
		and
		$columns{'COL'}) {	
		if ($started == 1) {
		    print OUT "\t";
		} else {
		    $started = 1;
		}
		print OUT "$col";
	    }

	    foreach my $header (@output_headers) {

		# skip unless column was selected
		next unless (defined $columns{$header}
			     and $columns{$header});

		if ($header eq $SPOT_ID_TERM) {
		    if ($started == 1) {
			print OUT "\t";
		    } else {
			$started = 1;
		    }		
		    print OUT "$id";
		    next;
		}
		foreach my $label (@labels) {
		    foreach my $file (@files) {
			next unless (defined $out_header{$header}{$label}{$file});
			$val = '';
			if (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}
			    and
			    $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col} =~ /\w/) {
			    $val = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
			    my @val = ($val);
			    if ($val =~ /\d/
				and
				$val =~ /;/) {
				@val = split /; /, $val;
			    }
			    foreach (@val) {
				if (/\.(\d+)$/) {
				    if (length($1) > $rounding) {
					$_ = sprintf "%.${rounding}f", $_;
				    }
				}
			    }
			    $val = join '; ', @val;
			}
			if ($started == 1) {
			    print OUT "\t";
			} else {
			    $started = 1;
			}
			print OUT "$val";
		    }
		}
	    }

	    if (defined $columns{'Description'}
		and 
		$columns{'Description'}) {
		if ($started == 1) {
		    print OUT "\t";
		} else {
		    $started = 1;
		}

		print OUT "$descr";	    
	    }

	    
	    # add annotation
	    if (@annotation_headers) {
		# get the probe for this position
		my $probe = '';
		if (defined $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col}) {
		    $probe = $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col};
		}
		foreach my $annotation_item (@annotation_headers) {
		    if ($started == 1) {
			print OUT "\t";
		    } else {
			$started = 1;
		    }

		    my $annotation = ' ';
		    if ($annotation_item eq 'TRANSCRIPT HITS'
			or
			$annotation_item eq 'GENE HITS') {
			if (defined $annotation{$annotation_item}{$probe}) {
			    $annotation = @{$annotation{$annotation_item}{$probe}}[0];
			}
		    } elsif ($probe) {
			if (defined $annotation{$annotation_item}{$probe}
			    and
			    @{$annotation{$annotation_item}{$probe}} > 0) {
			    my @tmp = ();
			    my %tmp = ();
			    if ($annotation_item eq 'GO'
				or
				$annotation_item eq 'go term'
				or
				$annotation_item eq 'go description') {
				foreach (@{$annotation{$annotation_item}{$probe}}) {
				    # report each GO term and description only once
				    unless ($tmp{$_}) {
					$tmp{$_}++;
					push @tmp, $_ ;
				    }
				}
			    } else {
				# remove duplicate entries
				foreach (@{$annotation{$annotation_item}{$probe}}) {
				    next unless (defined $_);
				    $tmp{$_}++;
				}
				# and sort
				@tmp = (sort keys %tmp);
				
				# DON'T do that for certain annotation items
				# because, for example, you won't know which hit quality matches to which hit!
				if (
				    $annotation_item eq 'TRANSCRIPT'
				    or
				    $annotation_item eq 'HIT COVERAGE'
				    or
				    $annotation_item eq 'HIT IDENTITY'
				    or
				    $annotation_item eq 'TC'
				    or
				    $annotation_item eq 'TC description'
				    or
				    $annotation_item eq 'hit identity'
				    or
				    $annotation_item eq 'hit coverage'
				    or
				    $annotation_item eq 'hit gaps'
				    or
				    $annotation_item eq 'hit flag'
				    or
				    $annotation_item eq 'putative ortholog Ensembl transcript'
				    or
				    $annotation_item eq 'putative ortholog cover'
				    or
				    $annotation_item eq 'putative ortholog ident'
				    or
				    $annotation_item eq 'putative ortholog gaps'
				    or
				    $annotation_item eq 'putative ortholog status'
				    ) {
				    # re-assign original annotation:
				    @tmp = @{$annotation{$annotation_item}{$probe}};
				}
			    }
			    $annotation = join '; ', @tmp;
			}
		    }
		    
		    print OUT "$annotation";
		}
	    }

	    # add extra column
	    if (@extra_column) {
		foreach my $extra_item (@extra_column) {
		    if ($started == 1) {
			print OUT "\t";
		    } else {
			$started = 1;
		    }

		    my $extra = ' ';
		    if (defined $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$extra_item}{$mrow}{$mcol}{$row}{$col}) {
			$extra = $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$extra_item}{$mrow}{$mcol}{$row}{$col};
			$extra = ' ' if ($extra eq '');
		    }
		    print OUT "$extra";
#		    warn "#KH: adding '$extra'";
		}
	    }

	    print OUT "\n";

	    $counter++;
	}

	close OUT;
    }

    # make button to generate HTML table
    my $html_table = '';
    my $prog_table = $cgi->url();
    my $basename = basename($prog_table);
    my $table_generator = 'table_gen.pl';
    $prog_table =~ s/(.*)$basename/$1$table_generator/;

    # retrieve a sample of $sample_lines lines
    open (IN, "$out_file")
	or die "Can't read from $out_file: $!";

    my @lines = ();
    my $header = <IN>;
    while ($header =~ /^\#/) {
	$header = <IN>;
    }
    chomp $header;
    my $cell_size_limit = 38;
    
    my %fold_change_header_tmp = ();
    foreach (keys %fold_change_header) {
	$fold_change_header_tmp{$fold_change_header{$_}}++;
    }
    foreach (keys %ratio_header) {
	$fold_change_header_tmp{$ratio_header{$_}}++;
    }
    my $fold_change_header_out = (join ',', sort keys %fold_change_header_tmp);
    
#    my $header_info = "$FG_SIGNAL_TERM = foreground, $BG_SIGNAL_TERM = background, $FLAG_CHANNEL_TERM = flag, $FG_COR_SIGNAL_TERM = background-corrected foreground, '_L' = log-transferred, '_R' = ratio, '_I' = intensity product, '_M' = merged\n";
    my $sample = "<B>Sample lines:</B><BR>(reduced to a length of $cell_size_limit characters per cell)<BR>Hover with mouse over green column headers to get a popup box with more info.\n";
    if (%fold_change_header) {
#	$header_info .= "\nFold changes are derived from <B>$fold_change_header_out</B>.";
	$sample .= "<BR>Fold changes and ratios are derived from <B>$fold_change_header_out</B>.";
    }
    $sample .= "<TABLE BORDER=1 CELLPADDING=3><TR><TH>";
    my @header_out = ();
    foreach (split /\t/, $header) {
	push @header_out, &proper_header($_);
    }
    $sample .= join '</TH><TH>', @header_out;
    $sample .= "</TH></TR>\n";
    my $sample_lines = &get_user_settings($var_def_default{'output_sample_lines'},'output_sample_lines',$force_global_opt,$action_num,$files[0]);
    my $sample_lines_tmp = $sample_lines;
    unless ($sample_lines) {
	$sample_lines_tmp = 40;
    }
    for (1..$sample_lines_tmp) {
	$_ = <IN>;
	last unless (defined $_);
	# skip comments
	while ($_ =~ /^\#/) {
	    $_ = <IN>;
	}
	chomp;
	$sample .= "<TR>\n";
	my @tmp = (split /\t/, $_);
	foreach (@tmp) {
	    if ($_ eq '') {
		$_ = '&nbsp;';
	    } elsif (length($_) > $cell_size_limit) {
		# reduce size of output fields to $cell_size_limit
		$_ = substr $_, 0, ($cell_size_limit-4);
		$_ .= '...';
	    }
	    my $align = ' ALIGN=RIGHT';
	    if (/[a-z]/i) { # entry contains a letter
		$align = '';
	    }
	    $sample .= "<TD$align>$_</TD>\n";
	}
#	$sample .= join '</TD><TD>', @tmp;
	$sample .= "</TR>\n";
    }
    close IN;
    $sample .= "</TABLE>\n";
    my $sample_file = '';
    unless ($sample_lines) {
	# write sample lines to file
	$sample_file = "${out_file}_sample.html";
	if (open (SAMPLE, ">$sample_file")) {
	    my $title = "$sample_lines_tmp sample lines for ".(basename $out_file);
	    print SAMPLE $cgi->start_html(-title=>"$title",
				   -author=>"$operator_email",
				   -BGCOLOR=>'white',
				   -style=>{-code=>$css_style},                         
				   -script=>$help_javascript,
				   -onload=>"setFocus()"
				   );
	    print SAMPLE $sample;
	    print SAMPLE "\n</BODY>\n</HTML>\n";
	    close SAMPLE;
	}
	$sample = '';
    }
    
    my $output_info = "<input type=\"hidden\" name=\"output_info\" value=\"$out_file";
    foreach my $file (@files) {
	$output_info .= "|$file";
    }
    $output_info .= "\">";
    my $return_string = "$output_info<B>$extra_info</B>:<BR>$sample<B>Download:</B> <A HREF=\"$out_file_html\">".(basename $out_file)."</A> ($counter entries, ".(&file_size($out_file)).")<BR>\n";
    if ($sample eq ''
	and 
	$sample_lines_tmp > 0) {
	$return_string .= "Sample of the <A HREF=\"".(basename $sample_file)."\">first $sample_lines_tmp lines</A>.\n<BR>\n";
    }
    $return_string .= "<BR>\n";
    return $return_string;
}


sub get_probe_ids {

    my $file = shift;
    my $id_extract = 0; # special treatment for FPMI Q2 slides

    my ($MROW,$MCOL,$ROW,$COL) = @_;

    # skip if done already
    return if (defined $probe_ids{"$MROW,$MCOL,$ROW,$COL"});

    if ($MROW == 12
	and
	$MCOL == 4
	and 
	$ROW == 22
	and 
	$COL == 22
	and
	$FPMI_specific
	) {
	# we presume the JB 21K array with Qiagen v2 probes, for which we can extract the Probe ID from the spot ID:
	$id_extract = 1;
    }

    if ($MROW == 12
	and
	$MCOL == 4
	and 
	$ROW == 20
	and 
	$COL == 20
	and
	$FPMI_specific
	) {
	# we presume OCI files
	# for these we have a special file to match probe id's to probe id's:
	if (open (IN, "/www/koch.pathogenomics.ca/data/arraypipe/internal/annotation/OCI_probe_linkage.tab")) {
	    while (<IN>) {
		chomp;
		my @tmp = split /\t/, $_;
		my ($mrow,$mcol,$row,$col) = @tmp[0..3];
		my $probe = $tmp[-1];
		$probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col} = $probe;
		$probe_names{$probe}++;
	    }
	    close IN;
	} else {
	    warn "#KH: Can't open annotation file!";
	}
	return;
    } elsif ($MROW == 12
	     and
	     $MCOL == 4
	     and 
	     $ROW == 23
	     and 
	     $COL == 26
	     and
	     $FPMI_specific
	     ) {
	# we presume the array with Qiagen v1 probes, for which we have a matching file:
	if (open (IN, "/www/koch.pathogenomics.ca/data/arraypipe/internal/annotation/Quiagen_v1_probe_linkage.tab")) {
	    while (<IN>) {
		chomp;
		my ($probe,$mrow,$mcol,$row,$col) = split /\t/, $_;
		$probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col} = $probe;
		$probe_names{$probe}++;
	    }
	    close IN;
	} else {
	    warn "#KH: Can't open annotation file!";
	}
	return;
	
    } else {
	# because we are getting the annotation from ProbeLynx
	# we presume that spot ids are also probe ids
	foreach my $mrow (1..$MROW) {
	    foreach my $mcol (1..$MCOL) {
		foreach my $row (1..$ROW) {
		    foreach my $col (1..$COL) {
			if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col}
			    and
			    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col} =~ /\w/
			    ) {
			    my $probe_id = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col};
			    if ($id_extract) {
				# special treatment for FPMI Q2 slides:
				# (the probe id is part of the spot id)
				my @tmp = split /\^/, $probe_id;
				$probe_id = '';
				if (defined $tmp[4]) {
				    $probe_id = $tmp[4];
				}
			    }
			    $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col} = $probe_id;
			    $probe_names{$probe_id}++;
			}
		    }
		}
	    }
	}
				
	return;
    }

    # this has been used for specific arrays and is now probably obsolete:
    unless ($MROW == 12
	    and
	    $MCOL == 4
	    and 
	    $ROW == 23
	    and 
	    $COL == 26) {
	# currently probe ids are only available for human array
	return;
    }
    
    # db details
    require DBI;
    my $DBD = 'mysql';
    my $database = 'hostdb';
    my $db_host = 'c003';
    my $user = 'pipeline';
    my $password = '9aleDL4k';
    my $table = 'oligo_position';
    
    my $dbh;
    eval { $dbh = DBI->connect("DBI:$DBD:database=$database;host=$db_host", $user, $password,{ RaiseError => 1, AutoCommit => 0 }); };
    if ($@) {
	warn $@;
	return $@;
    }

    unless (defined $dbh) {
	die "Can't connect to database: $DBI::errstr!\n";
    }
    my $statement = "select probe_id,mrow,mcol,row,col from $table";
    my $sth = $dbh->prepare($statement);
    die "Can't prepare '$statement': $DBI::errstr" if ($DBI::errstr);
    
    my $rv = $sth->execute();
    while (my ($probe,$mrow,$mcol,$row,$col) = $sth->fetchrow_array) {
	$probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col} = $probe;
	$probe_names{$probe}++;
    }

    $dbh->disconnect;

}

sub process_command_line {

    my $help = '';
    my $version = '';
    my $copyright = '';
    my @input_files = ();
    my @input_lists = ();
    my $action_profile = '';
    my @actions = ();
    my $out_dir = '';

    &GetOptions('help|h|?', \$help,
		'copy', \$copyright,
		'version', \$version,
		'BATCH', \$batch_processing,
		'file=s' => \@input_files,
		'list=s' => \@input_lists,
		'action=s' => \@actions,
		'profile=s' => \$action_profile,
		'listing|l:s' => \$modules_listing,
		'out=s' => \$out_dir,
		);
    
    # allow multiple files specified as comma-separated list
    @input_files = split(/,/,join(',',@input_files));
    @input_lists = split(/,/,join(',',@input_lists));
    @actions = split(/,/,join(',',@actions));
    
    
    if ($help) {
	print $help_message;
	exit;
    } 
    if ($copyright) {
	print $license;
	exit;
    } 
    if ($version) {
	print "\nThis is ArrayPipe version $revision_number.\n\n";
	exit;
    }

    if ($modules_listing eq '') {
	$modules_listing = 1;
    }

    if (@ARGV
	and
	$ARGV[0] eq 'BATCH'
	or
	$modules_listing) {
	$batch_processing = 1;
    }

    # convert arguments into URL-style command line:
    @CMD_ARG = ();
    
    my @tmp = ();
    my $file_counter = 0;
    foreach (@input_files) {
	$file_counter++;
	print "# checking format of file $file_counter: '$_'...";
#	my ($format_check,$file_format) = &check_file_format($_);
#	unless ($format_check eq 'ok') {
#	    print "\n# WARNING: file '$_' has invalid format: $format_check - skipping!";
#	} else {
	    print "o.k.\n";
	    push @tmp, $_;
	    $CMD_ARG[0] .= "\&work_files=$_";
#	}
    }
    if (@tmp) {
	@input_files = @tmp;
	if (@input_files > $max_files) {
	    $max_files = @input_files;
	}
    }
				   
    # check for mandatory parameters
    unless (@input_files) {
	&batch_die("\nNo input files provided!\nPlease use --help option to see more info!\n") unless ($modules_listing);
    }
    
    foreach (@input_lists) {
	$CMD_ARG[0] .= "\&work_lists=$_";
    }

    for (my $i = 0; $i <= $#actions; $i++) {
	$CMD_ARG[0] .= "\&action$i=$actions[$i]";
    }

    # load profile settings
    if ($action_profile) {
	my $file = '';
	if (-r $action_profile) {
	    $file = $action_profile;
	} elsif (-r "$save_settings_dir/$action_profile") {
	    $file = "$save_settings_dir/$action_profile";
	}

	unless ($file) {
	    &batch_die("Can't read from $action_profile!");
	}
	
	open (IN, $file)
	    or &batch_die("Can't open $file: $!");
	while (<IN>) {
	    chomp;
	    if (/^\# ACTION (\d+) (.*)/) {
		my $action_num = $1;
		my $action = $2;
		$CMD_ARG[0] .= "\&action$action_num=$action";
	    }
	    if (/^\# VARIABLE (\w+) (.*)/) {
		my $var = $1;
		my $value = $2;
		unless (defined $value) {
		    $CMD_ARG[0] .= "\&var";
		} else {
		    $CMD_ARG[0] .= "\&var=$value";
		}
	    }	
	}
	close IN;
    }

    # add the submit button
    $CMD_ARG[0] .= "\&submit=Do it!";

    return;
}


sub batch_die {
    my $message = shift;
    print STDERR "$message";
    exit $FAILURE;
}

sub set_cgi_params {

    # convert command line arguments into CGI parameters
    # so that the program can run as if it had been called from the web

#    $CMD_ARG[0] =~ s/\\//g;
    #warn "# KH cmd line:: $CMD_ARG[0]\n";

    # if a configuration file was supplied, use that first
    my $args = '';
    my $cmd_config_file_in = '';

    if (defined $CMD_ARG[1]
	and
	$CMD_ARG[1] =~ /cfg/) {
	$cmd_config_file_in = $CMD_ARG[1];
    }
    if ($cmd_config_file_in) {
	if (open (IN, $cmd_config_file_in)) {
	    while (<IN>) {
		chomp;
		if (/=/) {
		    $args .= '&'.$_;
		    if (/results_file=(.*)/) {
			my $results_file = $1;
			# first thing should be writing something to the results file
			unless (-e $results_file) {
			    system "echo 'No results available yet, please reload!' > $results_file; chmod 666 $results_file 2> /dev/null";
			    $new_results_file = $results_file;
			}
		    }
		}
	    }
	    close IN;
	    $CMD_ARG[0] = $args.$CMD_ARG[0];
	}
    }

    my @tmp = split /\&/, $CMD_ARG[0];
    my %vars = ();

    # remove possible program call from first element
    $tmp[0] =~ s/.*\??(.*)/$1/;
    foreach (@tmp) {
	next unless (/\w/);
	my ($name,$value) = split /\=/, $_, 2;
	$name =~ s/\+/ /g;
	
	$value = '' unless (defined $value);

	$value =~ s/\+/ /g;

	push @{$vars{$name}}, $value;
    }

    foreach my $key (sort keys %vars) {
	my @list = @{$vars{$key}};
	my $name = $key;

	$cgi->param(-name=>$name,
		    -values=>[@list]);
    }
}


sub add_config {
    my $file = shift;
    my $add = shift;
    unless (open (CFG, ">>$file")) {
	warn "Can't add to config_file ($file): $!";
	return;
    } else {
	print CFG "$add\n";
    }
    
    close CFG;
}

sub get_cgi_params {
    my $cmd = (basename $0).'?';

    my $cwd = `pwd`;
    chomp $cwd;
    my $tmp_dir = $cgi->param('tmp_dir') || '';
    my $suffix = $$;
    unless ($tmp_dir =~ /\w/
	    and
	    -d $tmp_dir) {
	$tmp_dir = "/home/khokamp/arraypipe_tmp/";
    } else {	
	$suffix = $cgi->param('action_step');
    }
    my $config_file = "$tmp_dir/arraypipe_cfg.$suffix";
#    my $config_file = "arraypipe_cfg.$$";
    open (CFG, ">$config_file")
	or die "Can't write config_file ($config_file) in '$cwd': $!";
    system "chmod 666 $config_file 2> /dev/null";
    print CFG "prog=".(basename $0)."\n";;

    my $spreadsheet = 0;
    my $submit = $cgi->param('submit') || '';
    if ($submit =~ /Spreadsheet/) {
	# exclude all of the parameters that are not necessary
	$spreadsheet = 1;
    }


    # todo:
    # instead of passing on all parameters, we could exclude the ones from actions that have not been used
    my %skip_parameters = ();
    my %used_actions = ();
    foreach (@actions) {
#	warn "# using '$_'";
	$used_actions{$_}++;
    }
    foreach my $item (keys %item_var) {
	unless ($spreadsheet) {
	    next if (defined $used_actions{$item});
	}
#	warn "# skipping '$item'";
	foreach (@{$item_var{$item}}) {
	    $skip_parameters{$_}++;
	}
    }

    my @names = $cgi->param;

    # print results_file parameter first:
    if (defined $cgi->param('results_file')) {
	my $val = $cgi->param('results_file');
	print CFG 'results_file='.$cgi->param('results_file')."\n";
    }

#    foreach my $file (keys %file_specs) {
#	warn "#KH file: '$file'";
#    }

    if (defined $file_specs{''}) {
	delete $file_specs{''};
    }

    foreach my $param (sort @names) {
	# skip parameters that weren't used
	if ($skip_parameters{$param}) {
#	    warn "skipping parameter '$param'";
	    next;
	}
	foreach my $value ($cgi->param($param)) {
	    next unless (defined $value);
	    next if ($value eq '');
	    # skip parameters that are the same as the default
	    if (defined $var_def_default{$param}) {
		if ($value =~ /\D/) {
		    next if ($value eq $var_def_default{$param});
		} else {
		    unless ($var_def_default{$param} =~ /\D/) {
			next if ($value == $var_def_default{$param});
		    }
		}
	    }
	    my $name = $param;

	    if ($spreadsheet) {
		if ($name =~ /^action/) {
		    next;
		} elsif ($name =~ /^step_label/) {
		    next;
		} elsif ($name =~ /^tech_rep/) {
		    next;
		} elsif ($name =~ /^bio_rep/) {
		    next;
		}
	    }
	    $cmd .= '&'."$name=$value";
	    print CFG "$name=$value\n";
	}
    }

    close CFG;

    return $cmd, $config_file;
}


sub set_variables {

    # the process id number might be supplied from the previous call to the script

    if (defined $cgi->param('swap_dir') 
	and
	defined $cgi->param('dir_swap')) {
	$from_dir = $cgi->param('swap_dir');
	$to_dir = $cgi->param('dir_swap');
    }

    if (defined $cgi->param('results_file')
	and
	$cgi->param('results_file') =~ /\w/) {
	$results_file = $cgi->param('results_file');
 	if ($to_dir) {
	    $results_file =~ s/$from_dir/$to_dir/;
	}
   }

    if ($cgi->param('swap_channels')) {
	my @files = $cgi->param('swap_channels');
	foreach (@files) {
	    $channel_swaps{$_}++;
	}
    }

    if (defined $cgi->param('spot_limit')
	and
	$cgi->param('spot_limit') =~ /\d/) {
	$spot_limit = $cgi->param('spot_limit');
    }

    if (defined $cgi->param('spot_limit_start')
	and
	$cgi->param('spot_limit_start') =~ /\d/) {
	$spot_limit_start = $cgi->param('spot_limit_start');
    }

    if (defined $cgi->param('subgrid_start')
	and
	$cgi->param('subgrid_start') =~ /\d/) {
	$subgrid_start = $cgi->param('subgrid_start');
    }

    if (defined $cgi->param('subgrid_end')
	and
	$cgi->param('subgrid_end') =~ /\d/) {
	$subgrid_end = $cgi->param('subgrid_end');
    }

    if (defined $cgi->param('spotid_filter')
	and
	$cgi->param('spotid_filter') =~ /\d/) {
	$spotid_filter = $cgi->param('spotid_filter');
    }

    if (defined $cgi->param('pid')
	and
	$cgi->param('pid') =~ /^\d+$/) {
	$PID = $cgi->param('pid');
    } elsif ($results_file =~ /(\d+)\.html/) {
	$PID = $1;
    }

    if (defined $cgi->param('action_step')
	and
	$cgi->param('action_step') =~ /\w/) {
	$action_step = $cgi->param('action_step');
	if ($action_step =~ /(\d+)(\..+)$/) {
	    $action_step_num = $1;
	    $action_step_ext = $2;
	} else {
	    $action_step_num = $action_step;
	}
    }

    foreach my $config_file (@config_files) {
	if (-r $config_file) {
	    unless (open (CONF, "$config_file")) {
		warn "Can't read $config_file: $!";
		next;
	    }
	    while (<CONF>) {
		# remove comments (everything following a hash)
		my ($line,$rest) = split /\#/; 
		if ($line =~ /example_dir = '(.+)';/) {
		    $example_dir = $1;
		} elsif ($line =~ /example_dir_html = '(.+)';/) {
		    $example_dir_html = $1;
		} elsif ($line =~ /log_file = '(.+)';/) {
		    $log_file = $1;
		} elsif ($line =~ /tmp_dir = '(.+)';/) {
		    $tmp_dir = $1;
		} elsif ($line =~ /tmp_dir_html = '(.+)';/) {
		    $tmp_dir_html = $1;
		} elsif ($line =~ /results_file = '(.+)';/) {
		    $results_file = $1;
		} elsif ($line =~ /results_file_html = '(.+)';/) {
		    $results_file_html = $1;
		} elsif ($line =~ /save_settings_dir = '(.+)';/) {
		    $save_settings_dir = $1;
		} elsif ($line =~ /pointer_image = '(.+)';/) {
		    $pointer_image = $1;
		} elsif ($line =~ /remote_user = '(.+)';/) {
		    $remote_user = $1;
		} elsif ($line =~ /remote_config_file = '(.+)';/) {
		    $remote_config_file = $1;
		} elsif ($line =~ /public_username = '(.+)';/) {
		    $public_username = $1;
		} elsif ($line =~ /operator_email = '(.+)';/) {
		    $operator_email = $1;
		} elsif ($line =~ /home_link = '(.+)';/) {
		    $home_link = $1;
		} elsif ($line =~ /cgi_link = '(.+)';/) {
		    $cgi_link = $1;
		} elsif ($line =~ /user_instructions_link = '(.+)';/) {
		    $user_instructions_link = $1;
		} elsif ($line =~ /project_logo = '(.+)';/) {
		    $project_logo = $1;
		} elsif ($line =~ /tutorial_link = '(.+)';/) {
		    $tutorial_link = $1;
		} elsif ($line =~ /lynx_location = '(.+)';/) {
		    $lynx_location = $1;
		} elsif ($line =~ /allow_web_file_load = '(.+)';/) {
		    $allow_web_file_load = $1;
		} elsif ($line =~ /check_example_files = '(.+)';/) {
		    $check_example_files = $1;
		} elsif ($line =~ /help_file_URL = '(.+)';/) {
		    $help_file_URL = $1;
		} elsif ($line =~ /R_executable = '(.+)';/) {
		    $R_executable = $1;
		} elsif ($line =~ /allow_debug = '(.+)';/) {
		    $allow_debug = $1;
		} elsif ($line =~ /transpath_user = '(.+)';/) {
		    $transpath_user = $1;
		} elsif ($line =~ /transpath_URL = '(.+)';/) {
		    $transpath_URL = $1;
		} elsif ($line =~ /FPMI_specific = '(.+)';/) {
		    $FPMI_specific = $1;
		} elsif ($line =~ /use_cluster_nodes = '(.+)';/) {
		    $use_cluster_nodes = $1;
		} elsif ($line =~ /common_user_group = '(.+)';/) {
		    $common_user_group = $1;
		} elsif ($line =~ /data_db_setup = '(.+)';/) {
		    $data_db_setup = $1;
		} elsif ($line =~ /data_db_user = '(.+)';/) {
		    $data_db_user = $1;
		} elsif ($line =~ /data_db_passwd = '(.+)';/) {
		    $data_db_passwd = $1;
		} elsif ($line =~ /data_db_host = '(.+)';/) {
		    $data_db_host = $1;
		} elsif ($line =~ /data_db_type = '(.+)';/) {
		    $data_db_type = $1;
		}
	    }  
	    close CONF;
	}
    }

    if (defined $cgi->param('submit')) {
	
	# make sure that we are logging in with the public username

	if ($cgi->param('submit') eq 'Enter') {
	    # the 'Enter' button always defaults the user to the public name
	    $cgi->param(-name=>'pipeline_user',
			-value=>$public_username);
	} elsif ($cgi->param('submit') eq 'Login') {
	    # if 'Login' was pressed with empty or no user name,
	    # then this defaults to the public name as well
	    unless (defined $cgi->param('pipeline_user')) {
		$cgi->param(-name=>'pipeline_user',
			    -value=>$public_username);
	    } else {
		if ($cgi->param('pipeline_user') eq '') {
		    $cgi->param(-name=>'pipeline_user',
				-value=>$public_username);
		}
	    }
	}
    }

    $username = $cgi->param('pipeline_user') if (defined $cgi->param('pipeline_user'));

    my $dir_add = '';
    my $add = '';
    my $old_tmp_dir = '';

    # check if this is the first of the second call
    if (defined $cgi->param('second_call')) {
	$second_call = $cgi->param('second_call');
	if ($second_call) {
	    $first_call = 0;
	}
    }

    # check if a tmp_dir is specified
    if (defined $cgi->param('tmp_dir')
	and
	$cgi->param('tmp_dir') =~ /\w/) {
	$tmp_dir = $cgi->param('tmp_dir');

	if ($to_dir) {
	    $tmp_dir =~ s/$from_dir/$to_dir/;
	}

	if (defined $cgi->param('tmp_dir_html')
	    and
	    $cgi->param('tmp_dir_html') =~ /\w/) {
	    $tmp_dir_html = $cgi->param('tmp_dir_html');
	}

#	warn "# KH: second call: '$second_call', save_dir_add: '".($cgi->param('save_dir_add'))."'";

	if (defined $cgi->param('save_dir_add')
	    and
	    $cgi->param('save_dir_add') ne ''
	    and
	    # only do the renaming once - in the first call of the program!
	    $second_call eq '') {

	    # remove original appendix:
	    $old_tmp_dir = $tmp_dir;
	    $tmp_dir =~ s/(.*\/)(.*)/$1/;
	    my $prev_add_dir = $2;
	    my $old_tmp_dir_html = $tmp_dir_html;
	    $tmp_dir_html =~ s/(.*\/)(.*)/$1/;
	    my $prev_add_dir_html = $2;
	    $dir_add = $cgi->param('save_dir_add');
	    $dir_add =~ s/ /_/g;
	    $dir_add =~ s/[\W|\.]//g;
	    while (-d "${tmp_dir}$dir_add$add") {
		warn "#Found '${tmp_dir}$dir_add$add'";
		$add++;
	    }
	    $dir_add .= $add;
	    $tmp_dir .= $dir_add;
	    $tmp_dir_html .= $dir_add;

	    warn "#KH: renaming $old_tmp_dir into $tmp_dir as user $ENV{USER}, $ENV{HOME}\n";
	    # rename the previous directory to the new one:
	    rename $old_tmp_dir, $tmp_dir
		or die "Can't rename $old_tmp_dir into $tmp_dir: $!\n";
	    $old_tmp_dir = '';

	    $results_file =~ s/$prev_add_dir/$dir_add/;
	    $results_file_html =~ s/$prev_add_dir_html/$dir_add/;
	}

	# create unless not existent

	unless (-d $tmp_dir) {
	    # try to create it:
	    my $date = localtime;
#	    warn "#KH: mkdir $tmp_dir at $date";
	    system "mkdir -p '$tmp_dir'";
	    system "chgrp $common_user_group '$tmp_dir' 2> /dev/null";
	    system "chmod 777 '$tmp_dir' 2> /dev/null";
	}
	
    } else {

	# make a temporary directory for this session:
	unless ($tmp_dir =~ /\w/) {
	    $tmp_dir = "/tmp/_${progname}_tmp";
	    $tmp_dir_html = "file://$tmp_dir";
	}
	unless (-d $tmp_dir) {
	    # try to create it:
	    system "mkdir -p '$tmp_dir'";
	    system "chgrp $common_user_group '$tmp_dir' 2> /dev/null";
	    system "chmod 777 '$tmp_dir' 2> /dev/null";
	}

	# add extension
	if (-d $tmp_dir) {
#	    my $dir_add = '_'.CGI::remote_host().'_';
	    my $dir_add = $username.'/'.'_';
	    my $add = "$$";

	    if (defined $cgi->param('save_dir_add')) {
		$dir_add = $cgi->param('save_dir_add');
		$add = '';
	    } else {
		$dir_add .= $add;
	    }
	    while (-d "$tmp_dir/$dir_add$add") {
		$add++;
	    }

	    system "mkdir -p '$tmp_dir/$dir_add'";
	    system "chgrp $common_user_group '$tmp_dir/$dir_add' 2> /dev/null";
	    system "chmod 777 '$tmp_dir/$dir_add' 2> /dev/null";
	    $tmp_dir = "$tmp_dir/$dir_add";
	    $tmp_dir_html .= "/$dir_add";
	    if ($username eq $public_username) {
		system "cp -p /www/koch.pathogenomics.ca/data/cgi-bin/FPMI/.htaccess $tmp_dir";
		system "chgrp $common_user_group '$tmp_dir/.htaccess'";
		system "chgrp $common_user_group '$tmp_dir/.htaccess' 2> /dev/null";
#		system "chmod 777 '$tmp_dir/.htaccess' 2> /dev/null";
	    }
	}
	
	# die if directory wasn't created
	unless (-d $tmp_dir) {
	    die "Couldn't create directory '$tmp_dir'!";
	}
    }
 
    # check if a results file is specified
    if (defined $cgi->param('results_file')
	and
	$cgi->param('results_file') =~ /\w/
	and
	$dir_add eq '') {
	$results_file = $cgi->param('results_file');
 	if ($to_dir) {
	    $results_file =~ s/$from_dir/$to_dir/;
	}
	if (defined $cgi->param('results_file_html')
	    and
	    $cgi->param('results_file_html') =~ /\w/) {
	    $results_file_html = $cgi->param('results_file_html');
	} else {
	    $results_file_html = $tmp_dir_html . '/' . (basename $results_file);
	}
	unless ($results_file =~ /^\//) {
	    # no full path specified, prepend tmp_dir
	    $results_file = "$tmp_dir/$results_file";
	}
    }
    
    # set defaults for output file   
    $results_file = "$tmp_dir/results_step_$action_step.html" unless ($results_file);
    $results_file_html = "$tmp_dir_html/results_step_$action_step.html" unless ($results_file_html);

    # if $results_file exists already, change action_step
#    unless ($second_call) {
    unless ($new_results_file eq $results_file) {
	# make sure we're not overwriting a previous results file:
	if (-e $results_file) {
	    $action_step_ext .= '.0';
	}
	
	while (-e $results_file) {
	    $action_step_ext =~ /(\d+)$/;
	    my $num = $1 + 1;
	    $action_step_ext =~ s/\d+$/$num/;
	    $action_step = $action_step_num.$action_step_ext;
	    $results_file =~ s/_step_\d+(\.\d+)*/_step_${action_step}/;
	    $results_file_html = $results_file;
	    $results_file_html =~ s/$tmp_dir/$tmp_dir_html/;
	}
    }

# %example_files = ('/www/www.brinkman.mbb.sfu.ca/data/mx/tmp/Cy3_110402_1029-PRHUO4-S1-0002-CI008-002.txt', 'Cy3_110402_1029-PRHUO4-S1-0002-CI008-002.txt', 
#		     '/www/www.brinkman.mbb.sfu.ca/data/mx/tmp/Cy5_110402_1029-PRHUO4-S1-0002-CI008-002.txt', 'Cy5_110402_1029-PRHUO4-S1-0002-CI008-002.txt', 
#		     );

    # other general settings:
    if (defined $cgi->param('SCALE_IMAGE')
	and
	$cgi->param('SCALE_IMAGE') =~ /\d+/) {
	my $new_scale = $cgi->param('SCALE_IMAGE');
	if ($new_scale > 0
	    and
	    $new_scale <= 4) {
	    $SCALE_IMAGE = $new_scale;
	}
    }

    unless ($save_settings_dir) {
	$save_settings_dir = "$tmp_dir/.save_settings";
    }

    &write_help_file;
}


sub write_help_file {
    $help_file = "$tmp_dir/help.html";
    $help_file_link = 'help.html';
    return if (-e $help_file);

    unless (open (HELP, ">$help_file")) {
	die "Can't write help file at $help_file: $!";
    }
    
    print HELP "<HTML>
<HEAD>
<TITLE>ArrayPipe Help</TITLE>
<STYLE>
$css_style
</STYLE>
</HEAD>
<BODY BGCOLOR='WHITE'>
<DIV>
<TABLE WIDTH=$main_width>
<TR>
<TD>
<H1>ArrayPipe Help</H1>
<UL>
";

    foreach my $topic (sort keys %pipeline_help_title) {
	my $title = $pipeline_help_title{$topic};
	my $text = $pipeline_help_text{$topic};
	next unless ($text);
	
	print HELP "
<BR><LI><B><A NAME='$topic'>$title</A></B><BR>
$text
</LI>
<BR>
";
    }
    
    print HELP "\n</UL>";
    print HELP "</TD></TR></TABLE>\n</DIV>\n";
    print HELP &ending;
    close HELP;
}
	
sub initialize_output {
    my $refresh = shift;
    my $first = shift || '';

    unless ($refresh =~ /\d/) {
	$refresh = 4;
    }
    # set minimum refresh rate to 4 seconds
    if ($refresh < 4) {
	$refresh = 4;
    }
    # write a new header, which contains a refresh statement
    open (TMP_RES, ">$results_file")
	or die "Can't write to $results_file: $!";
    print TMP_RES &html_head_refresh("$progname (step $action_step, in progress)",$refresh,'-1');
    close TMP_RES;

    # write all results to a temporary file
    if ($first) {
	open (RESULT, ">$tmp_results_file")
	    or die "Can't write result to $tmp_results_file: $!";
    } else {
	open (RESULT, ">>$tmp_results_file")
	    or die "Can't write result to $tmp_results_file: $!";
    }
    select RESULT;
    $| = 1;

    if ($first) {
	# create an index file for that directory:
	my @html_files = ();
	# collect all html files
	if (opendir(DIR, $tmp_dir)) {
	    while (defined(my $file = readdir(DIR))) {
		if (-f "$tmp_dir/$file"
		    and
		    $file =~ /\.html$/) {
		    next if ($file eq 'index.html');	    
		    push @html_files, $file;
		}
	    }
	}
	close DIR;

	unless (open (INDEX, ">$tmp_dir/index.html")) {
	    warn "#$progname: can't write index.html in $tmp_dir: $!";
	    return;
	}

	my $title = "$progname - temporary results";
	print INDEX $cgi->start_html(-title=>"$title",
				     -author=>"$operator_email",
				     -BGCOLOR=>'white',
				     -script=>$help_javascript,
				     -style=>{-code=>$css_style}
				     );
	print INDEX "<DIV><H1>Index of results files:</H1>\n<OL>\n";
	
	# sort by date
	my %tmp = ();
	foreach (@html_files) {
	    my $date = (stat "$tmp_dir/$_")[9];
	    push @{$tmp{$date}}, $_;
	}

	my $files = @html_files;
	foreach my $date (sort { $a <=> $b } keys %tmp) {
	    foreach (@{$tmp{$date}}) {
		$files--;
		my $size = &commify(sprintf "%.0f", (stat "$tmp_dir/$_")[7]/1024);
		my $step = '';
		if ($_ =~ /step_(.*)\.html/) {
		    $step = $1;
		}
		my $label = '';
		if (defined $cgi->param("step_label$step")) {
		    $label = $cgi->param("step_label$step");
		}
		print INDEX "<LI><A HREF=\"$_\">$_</A>&nbsp;<B>$label</B><BR>(created ".(localtime($date))." ";
		if ($files) {
		    print INDEX ", size $size KB + pictures\n";
		} else {
		    print INDEX ", latest results)\n";
		}
		print INDEX ")\n";
	    }
	}
	print INDEX "</form>\n</CENTER>
        <BR>
        <HR>
        <address>automatically generated at ".(localtime)."
        <BR>for questions or remarks e-mail <A HREF=\"MAILTO:$operator_email\">$operator_email</A>
        </address>
        </DIV>
        </BODY>
        </HTML>\n";
	close INDEX;
	system "chgrp $common_user_group '$tmp_dir/index.html' 2> /dev/null";
	system "chmod 666 '$tmp_dir/index.html' 2> /dev/null";	
    }
}


sub finalize_output {
    # write a new header, which contains a refresh statement
    open (TMP_RES, ">$results_file.2")
	or die "Can't write to $results_file.2: $!";
    select TMP_RES;
    print &html_head("$progname - Step $original_action_step",'+1m');
    close TMP_RES;
    
    # copy everything collected so far to the refreshed file
    system "cat '$tmp_results_file' >> '$results_file.2'";
    system "mv '$results_file.2' '$results_file'";

}


sub format_output {

    my $ref = shift;
    my %args = %$ref;

    my $prefix = $args{'prefix'} || '';
    my $action_num = '';
    if (defined $args{'num'}) {
	$action_num = $args{'num'};
    }
    my $file = $args{'file'} || '';
    my $label = $args{'label'} || '';
    my $action = $args{'action'} || '';
    my $image = $args{'image'} || '';
    my $info = $args{'info'} || '';
    my $min_val = '';
    if (defined $args{'min'}) {
	$min_val = $args{'min'};
    }
    my $max_val = '';
    if (defined $args{'max'}) {
	$max_val = $args{'max'};
    }

    unless ($label) {
	$label = '__NO_LABEL__';
    }

    if ($label eq 'ratio') {
	$label = '__NO_LABEL__';
    }
	
    if ($label eq '-') {
	$label = '';
    }

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    # add data_column parameter
    my $param_add = $prefix.'data_column';
    unless ($param_added{$prefix.'data_column'}) {
	$param_added{$prefix.'data_column'} = 1;
    }
	
    $var_def_type{$param_add} = 'select';
    $var_def_text{$param_add} = '<BR>Data column: ';
    my $default = '';

    LABEL1:foreach my $column (@{$item_data_column{$action}}) {
	my $file_tmp = $file;
	unless ($file_tmp) {
	    $file_tmp = (keys %data)[0];
	}
	my $label_tmp = $label;
	unless ($label_tmp) {
	    $label_tmp = '__ALL__';
	}
	if ($label_tmp eq '__ALL__') {
	    foreach my $label (keys %{$data{$file_tmp}{'__DATA__'}}) {
		if ($data_column_available{$file_tmp}{$label}{$column}) {
		    $default = $column;
		    last LABEL1;
		}
	    }
	} else {
	    if ($data_column_available{$file_tmp}{$label}{$column}) {
		$default = $column;
		last LABEL1;
	    }
	}
    }

    my @default = &get_user_settings($var_def_default{$param_add},$param_add,$force_global_opt,$action_num,$file,$label);
    if (@default > 1) {
	$default = \@default;
    }
    if ($last_column_used) {
	$default = $last_column_used;
    }
    $var_def_default{$param_add} = 'default';

    my $label_check = $label;
    if ($label eq '__ALL__'
	or
	$label eq '') {
	$label_check = 'CH1';
    }
    my $file_tmp = $file;
    unless ($file) {
	$file_tmp = (keys %data)[0];
    }

    my %column_lst = ();
    foreach (@{$item_data_column{$action}}, @{$data_column_available{$file_tmp}{$label_check }{'__ORDER__'}}) {
	$column_lst{$_}++;
    }
    my @column_lst = sort keys %column_lst;

    $var_def_values{$param_add} = ['default', @column_lst];

    my @infos = ();
    if (ref($info) eq 'ARRAY') {
	@infos = @$info;
    } else {
	@infos = ($info);
    }

    my @images = ();
    if (ref($image) eq 'ARRAY') {
	@images = @$image;
    } else {
	@images = ($image);
    }

    my $image_link = '';
    my @image_links = ();
    if (defined $args{'image_link'}) {
	$image_link = $args{'image_link'};
    }
    if (ref($image_link) eq 'ARRAY') {
	@image_links = @$image_link;
    } elsif ($image_link) {
	@image_links = ($image_link);
    }

    # compile HTML code for integration of image    
    my $label_out = '';
    my $default_out = '';
    if ($label eq 'CH1') {
	if (defined $file_specs{$file}{'CH1_LABEL'}) {
	    $label_out = "$file_specs{$file}{'CH1_LABEL'}";
	} else {
	    $label_out = $label;
	}
	if ($default eq $FG_SIGNAL_TERM
	    and
	    defined $file_specs{$file}{'CH1FG'}) {
	    $default_out = " ($file_specs{$file}{'CH1FG'})";
	} elsif ($default eq $BG_SIGNAL_TERM
		 and
		 defined $file_specs{$file}{'CH1BG'}) {
	    $default_out = " ($file_specs{$file}{'CH1BG'})";
	}
    } elsif ($label eq 'CH2') {
	if (defined $file_specs{$file}{'CH2_LABEL'}) {
	    $label_out = "$file_specs{$file}{'CH2_LABEL'}";
	} else {
	    $label_out = $label;
	}
	if ($default eq $FG_SIGNAL_TERM
	    and
	    defined $file_specs{$file}{'CH2FG'}) {
	    $default_out = " ($file_specs{$file}{'CH2FG'})";
	} elsif ($default eq $BG_SIGNAL_TERM
		 and
		 defined $file_specs{$file}{'CH2BG'}) {
	    $default_out = " ($file_specs{$file}{'CH2BG'})";
	}
    } elsif ($label eq '__ALL__'
	     or
	     $label eq '__NO_LABEL__'
	     or
	     $label eq '') {

	if (defined $file_specs{$file}{'CH1_LABEL'}) {
	    $label_out = "$file_specs{$file}{'CH1_LABEL'}";
	} else {
	    $label_out = 'CH1';
	}
	if (defined $file_specs{$file}{'CH2_LABEL'}) {
	    $label_out .= ",$file_specs{$file}{'CH2_LABEL'}";
	} else {
	    $label_out .= ',CH2';
	}

	if ($default eq $FG_SIGNAL_TERM
	    and
	    defined $file_specs{$file}{'CH1FG'}
	    and
	    defined $file_specs{$file}{'CH2FG'}
	    ) {
	    $default_out = " ($file_specs{$file}{'CH1FG'}, $file_specs{$file}{'CH2FG'})";
	} elsif ($default eq $BG_SIGNAL_TERM
		 and
		 defined $file_specs{$file}{'CH1BG'}
		 and
		 defined $file_specs{$file}{'CH2BG'}) {
	    $default_out = " ($file_specs{$file}{'CH1BG'}, $file_specs{$file}{'CH2BG'})";
	}
    }

    if ($default eq $SPOT_ID_TERM
	and
	defined $file_specs{$file}{'SPOT_ID'}) {
	$default_out = " ($file_specs{$file}{'SPOT_ID'})";
    }

    my $image_lnk = "<BR>";
    if ($file eq '') {
	$image_lnk .= 'L' unless ($label eq '');
    } else {
	$image_lnk .= "File <B>'".(basename $file)."'</B>, l";
    }
    if ($label eq '') {
	$image_lnk .= 'D';
    } else {
	$image_lnk .= "abel <B>$label_out</B>, d";
    }
    $image_lnk .= "ata column: <B>'$default'$default_out</B>\n";

    if ($refine_parameters) {
	$image_lnk = "<BR>File <B>'".(basename $file)."'</B>, label <B>$label_out</B>, data column: <B>undetermined, please choose below</B>\n";
    }
    if (defined $args{'outtext'}) {
	$image_lnk = $args{'outtext'};
    }

    $image_lnk .= "<TABLE>\n";
    # place images beside information one pair per row:
    foreach my $image (@images) {
	# check if image has been successfully created:
	if ($image =~ /Execution of R-job failed/) {
	    # no image - replace info with error message
	    $infos[0] = $image;
	    $image_lnk .= "<TR VALIGN=TOP><TD>\&nbsp;";
	} elsif ($image eq '') {
	    $image_lnk .= "<TR VALIGN=TOP><TD>\&nbsp;";
	} else {
	    
	    my $image_file = "$tmp_dir/$image";
	    # get dimensions
	    my ($width,$height) = (0,0);
	    my $src_image = '';
	    if ($image_file =~ /\.png/
		and
		-r $image_file) {
		my $src_image = newFromPng GD::Image($image_file);
		($width,$height) = $src_image->getBounds();
	    }
	    
#	    warn "dimensions for '$image_file': ($width x $height) ";

	    # scale image if requested
	    if ($SCALE_IMAGE
		and
		$SCALE_IMAGE != 1
		and 
		$image_file =~ /\.png/
		and
		-r $image_file) {
		my $new_width = sprintf "%.0f", $width * $SCALE_IMAGE;
		my $new_height = sprintf "%.0f", $height * $SCALE_IMAGE;
		# generate the new, scaled image:
		my $new_image = new GD::Image($new_width,$new_height);
		$new_image->copyResized($src_image,0,0,0,0,$new_width,$new_height,$width,$height);
		# overwrite image
		open(PNG, ">$image_file")
		    or die;
		binmode PNG;
		print PNG $new_image->png;
		close PNG;
		($width,$height) = ($new_width,$new_height);
	    }
#	    warn "new dimensions for '$image_file': ($width x $height) ";

	    if (@image_links) {
		my $image_link = shift @image_links;
		push @image_links, $image_link unless (@image_links);
		$image_lnk .= "<TR VALIGN=TOP><TD>$image_link<IMG SRC=\"$image\" BORDER=0 HEIGHT=\"$height\" WIDTH=\"$width\"></A>" if ($image);
	    } else {
		$image_lnk .= "<TR VALIGN=TOP><TD><IMG SRC=\"$image\" HEIGHT=\"$height\" WIDTH=\"$width\">" if ($image);
	    }
	}
	my $info = shift @infos || '&nbsp;';
	$image_lnk .= "</TD>\n<TD>\n$info\n</TD></TR>\n";
    }

    if ($show_parameters) {
	
	# add a row containing a table with parameters:

	$image_lnk .= "<TR><TD>
 <TABLE>
";

	$image_lnk .= "<TR BGCOLOR=#AAAAAA><TH COLSPAN=2>Parameters</TH></TR>\n";
	
	# list parameters:
	
	if ($action =~ /filter_by_value/) {
	    $param_add = '';
	}
	
	foreach my $param (@{$item_var{$action}},$param_add) {
	    next if ($param eq '');
	    my @default = &get_user_settings($var_def_default{$param},$param,$force_global_opt,$action_num,$file,$label);
	    if (@default > 1) {
		$default = \@default;
	    } else {
		$default = $default[0];
	    }
	    
	    $image_lnk .= "\n<TR VALIGN=TOP><TD>$var_def_text{$param}</TD><TD>".
		(&var_input("action$action_num$file$label$param",$param,$var_def_type{$param},$default)).
		" ".
		(defined $var_def_add{$param} ? $var_def_add{$param} : '').
		"</TD></TR>\n";
	}
	
	
	$image_lnk .= '<TD COLSPAN=2 ALIGN=RIGHT>--->'.
	    $cgi -> submit(-name=>'submit', 
			   -value=>'Do it!')
	    .'</TD></TR></TABLE></TD></TR>';
    }

    $image_lnk .=  "</TABLE>\n";

    if ($individual_results_pages) {
	# write output to a separate file:
	my $out_file = "$tmp_dir/result_$action.$action_num.$action_step.".(basename $file).".$label.html";
	open (OUT, ">$out_file")
	    or return "Can't write results: $!\n";
	print OUT $image_lnk;
	close OUT;
	
	return "<!-- start results ".(basename $out_file)." -->\nSee results for this module in <A HREF=\"".(basename $out_file)."\">separate page</A><BR>\n<!-- end results ".(basename $out_file)." -->\n";
	
    } else {
	return $image_lnk;
    }

}



sub signal_plotting_setup {
    my $action_num = shift;
    my $action = shift;

    # carry out action on each file ...
    foreach my $file (sort keys %data) {
	
	# put all the main calculations into a separate subroutine
	# the return value should be HTML code that will be directly
	# included into the results page

	my @images = ();
	my @info = ();


	# determine number of plots and subgrids per plots:

	my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);

	my @plots = ();

	my $force_global_opt = 0;
	if (defined $cgi->param("${action}_force")
	    and
	    $cgi->param("${action}_force") == 1) {
	    $force_global_opt = 1;
	}

	my $plot_layout = &get_user_settings($var_def_default{'signal_layout'},'signal_layout',$force_global_opt,$action_num,$file,'__ALL__');
	my $grids_per_plot = &get_user_settings($var_def_default{'signal_grids_per_plot'},'signal_grids_per_plot',$force_global_opt,$action_num,$file,'__ALL__');

	if ($grids_per_plot eq 'all') {
	    $grids_per_plot = $MROW * $MCOL;
	} elsif ($grids_per_plot eq 'MROW') {
	    $grids_per_plot = $MROW;
	} elsif ($grids_per_plot eq 'MCOL') {
	    $grids_per_plot = $MCOL;
	}
	
	# set up layout
	my $dim_row = 1;
	my $dim_col = 1;
	my $add_info = 0;

	if ($grids_per_plot == $MROW * $MCOL) {
	    # doesn't matter how many plots per sheets were specified, we only need one plot:
	    $plots[0] = ["1,1,$MROW,$MCOL"];
#	    if ($plot_layout =~ /(\d+)x(\d+)/) {
#		# makes no sense, but it is requested by the user:
#		($dim_row, $dim_col) = ($1,$2);
#	    }
	    $add_info = 1;
	} else {
	    # check how many plots are necessary
	    my $required_plots = int($MROW * $MCOL / $grids_per_plot);
	    if ($required_plots * $grids_per_plot < $MROW * $MCOL) {
		$required_plots++;
	    }

	    if ($plot_layout =~ /(\d+)x(\d+)/) {
		($dim_row, $dim_col) = ($1,$2);
	    } else {

		foreach my $factor (4,3,2,1) {
		    # check if number is divisible by factor:
		    my $dim_tmp = int($required_plots / $factor);
		    if ($dim_tmp * $factor == $required_plots) {
			$dim_col = $factor;
			$dim_row = $dim_tmp;
			if ($dim_row > 10) {
			    # break this into a sensible number
			    $dim_row = 6;
			} elsif ($dim_row > 8) {
			    $dim_row = 5;
			} elsif ($dim_row > 6) {
			    $dim_row = 4;
			}
			last;
		    }
		}
		if ($dim_col > $dim_row) {
		    ($dim_col, $dim_row) = ($dim_row, $dim_col);
		}
		if ($required_plots == 4) {
		    $dim_col = 2;
		    $dim_row = 2;
		}
	    }
	
	    my ($mrow_run, $mcol_run) = (1,1);
	    my ($mrow_last, $mcol_last) = (1,1);

	  OUTER:for (my $k = 1; $k <= $required_plots; $k++) {
		for (my $i = 1; $i <= $dim_col*$dim_row; $i++) {
		    for (my $j = 1; $j < $grids_per_plot; $j++) {
			$mcol_run++;
			if ($mcol_run > $MCOL) {
			    $mcol_run = 1;
			    $mrow_run++;
			    if ($mrow_run > $MROW) {
				last;
			    }
			}
		    }
		    push @{$plots[$k-1]}, "$mrow_last,$mcol_last,$mrow_run,$mcol_run";
		    # next plot starts with the next block
		    $mcol_run++;
		    if ($mcol_run > $MCOL) {
			$mcol_run = 1;
			$mrow_run++;
		    }
		    ($mrow_last,$mcol_last) = ($mrow_run, $mcol_run);
		    if ($mrow_run > $MROW) {
			last OUTER;
		    }
		}
	    }
	}
    
	
	my $round = 0;

	my ($min,$max) = (0,0);

	foreach my $element (@plots) {
	    $round++;
	    my ($image,$info) = &signal_plotting($action_num,$file,$action,$round,$dim_row,$dim_col,$add_info,$min,$max,@$element);
	    push @images, $image;
	    push @info, $info;

	}

	print &format_output(
			     {
				 'prefix' => 'signal_plot_',
				 'num' => $action_num,
				 'file' => $file,
				 'label' => '__ALL__',
				 'action' => $action,
				 'image' => \@images,
				 'info' => \@info
				 }
			     );
    }
}


sub signal_plotting {
    my $action_num = shift;
    my $file = shift;
    my $action = shift;
    my $round = shift;
    my $dim_row = shift;
    my $dim_col = shift;
    my $add_info = shift;
    my $min_coord = shift;
    my $max_coord = shift;
    my @plots = @_;

    if ($refine_parameters) {
	return '', '', 'Paramter Refinement!';
    }

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $header = '';
    foreach my $column (@{$item_data_column{$action}}) {

	my $occ = 0;
	foreach my $label (sort {$b cmp $a} keys %{$data{$file}{'__DATA__'}}) {
	    next if ($label eq '__NO_LABEL__');

	    if ($data_column_available{$file}{$label}{$column}) {
		$header = $column;
		$occ++;
	    }
	}
	if ($occ == 2) {
	    last;
	} else {
	    $header = '';
	}
    }

    # check for specific settings:
    $header = &get_user_settings($header,'signal_plot_data_column',$force_global_opt,$action_num,$file,'__ALL__');

    # check for specific settings:
    my $take_log = &get_user_settings($var_def_default{'signal_log'},'signal_log',$force_global_opt,$action_num,$file,'__ALL__');
    $take_log = $take_log eq 'yes' ? 1 : 0;

    if ($header =~ /_L/) {
	$take_log = 0;
    }

    $last_column_used = $header;

    # get labels for channel 1 (
    # check for user specific settings
    my $ch1 = &get_user_settings($var_def_default{'signal_x'},'signal_x',$force_global_opt,$action_num,$file,'__ALL__');

    # collect labels
    my %labels = ();

    foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	
	# skip columns that are not label specific ('spot labels')
	next if ($label eq '__NO_LABEL__');
	$labels{$label}++;
    }

    unless ($labels{$ch1}) {
	warn "# ERROR: Label '$ch1' is not defined for ".(basename $file)."!";
	die "# ERROR: Label '$ch1' is not defined for ".(basename $file)."!\n";
    }

    my $ch2 = '';
    foreach (keys %labels) {
	next unless (/\w/);
	unless ($ch1 eq $_) {
	    $ch2 = $_;
	    last;
	}
    }

    unless ($ch2) {
	die "# ERROR: No channel 2 defined for ".(basename $file)." (Ch1 = '$ch1', labels available: '".(join "','", keys %labels)."')!\n";
    }

    my ($ch1_out, $ch2_out) = ($ch1,$ch2);
    if (defined $file_specs{$file}{'CH1_LABEL'}) {
	$ch1_out = $file_specs{$file}{'CH1_LABEL'};
    }
    if (defined $file_specs{$file}{'CH2_LABEL'}) {
	$ch2_out = $file_specs{$file}{'CH2_LABEL'};
    }

    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'signal_skip_flags'},'signal_skip_flags',$force_global_opt,$action_num,$file,'__ALL__');
    %skip_flags = &get_skip_flags(@flags);

    my %flag2col_tmp = ();
    foreach my $flag (0, (sort { $a <=> $b } keys %flag_values_num)) {
 
	my $default = '';
	if (defined $pref_flag2col{$flag}) {
	    $default = $pref_flag2col{$flag};
	}
	# check for user specific settings
	my $flag_colour = &get_user_settings($default,'signal_flag_colour'.$flag,$force_global_opt,$action_num,$file,'__ALL__');
	$flag2col_tmp{$flag} = $flag_colour if ($flag_colour =~ /\w/);	    
    }

    my %values = ();
    my ($mrow,$mcol,$row,$col);
    my %skipped = ();
    my %flags_num = ();
    my $no_flag = 0;
    my $dots = 0;

    my $out_header1 = "$ch1.signal_plot";
    my $out_header2 = "$ch2.signal_plot";

    # initialize field for flags
#    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{'__FLAGS__'} = ();

    my %data_files = ();
    my %coords = ();
    my %flag_files = ();

    my ($all_minx, $all_miny) = (99999,99999);
    my ($all_maxx, $all_maxy) = (-99999,-99999);
    foreach my $coord (@plots) {
	my ($start_row,$start_col,$end_row,$end_col) = split /,/, $coord;
	
	# go through each field	
	foreach $mrow ($start_row .. $end_row) {
	    foreach $mcol ($start_col .. $end_col) {

		my %flags = ();

		# define data file:
		my $data_file = "R_in.".(basename ($file)).".$ch1.$ch2.$mrow,$mcol,$action.$action_num.$action_step.$PID";
		$data_file =~ s/\//_/g;
		$data_file =~ s/ /_/g;
		$data_file = "$tmp_dir/$data_file";
		open (DATA, ">$data_file") 
		    or die "Can't write to $data_file: $!";
		my $spots_out = 0;
		
		push @{$data_files{$coord}}, $data_file;
		push @{$coords{$coord}}, "$mrow:$mcol";

		foreach $row (keys %{$data{$file}{'__DATA__'}{$ch1}{$header}{$mrow}{$mcol}}) {
		    foreach $col (keys %{$data{$file}{'__DATA__'}{$ch1}{$header}{$mrow}{$mcol}{$row}}) {
			
			my $val1 = '';
			my $val2 = '';
			
			# initialize output field:
			$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header1}{$mrow}{$mcol}{$row}{$col} = $val1;
			$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header2}{$mrow}{$mcol}{$row}{$col} = $val2;
			if ($take_log) {
			    $data{$file}{'__DATA__'}{$ch1}{"${header}_L"}{$mrow}{$mcol}{$row}{$col} = $val1;
			    $data{$file}{'__DATA__'}{$ch2}{"${header}_L"}{$mrow}{$mcol}{$row}{$col} = $val2;
			}
			
			# retrieve values from both channels
			if (defined $data{$file}{'__DATA__'}{$ch1}{$header}{$mrow}{$mcol}{$row}{$col}) {
			    $val1 = $data{$file}{'__DATA__'}{$ch1}{$header}{$mrow}{$mcol}{$row}{$col};
			}
			if (defined $data{$file}{'__DATA__'}{$ch2}{$header}{$mrow}{$mcol}{$row}{$col}) {
			    $val2 = $data{$file}{'__DATA__'}{$ch2}{$header}{$mrow}{$mcol}{$row}{$col};
			}
			
			# skip unless both values set
			unless ($val1 =~ /\d/
				and
				$val2 =~ /\d/) {
			    $skipped{'at least one of the values is undefined'}++;
			    next;
			}
			
			if (($val2 <= 0
			     or
			     $val1 <= 0)
			    and
			    $take_log > 0) {
			    $skipped{'at least one of the values is <= 0 (no log possible)'}++;
			    next;
			} else {
			    if ($take_log) {
				$val1 = &log_base(2,$val1);
				$val2 = &log_base(2,$val2);
			    }
			}
			
			# get flags for first channel:
			my @flags = ();
			if (defined $data{$file}{'__DATA__'}{$ch1}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			    @flags = &get_flags($data{$file}{'__DATA__'}{$ch1}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
			}
			
			# get flags for second channel as well
			if (defined $data{$file}{'__DATA__'}{$ch2}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			    push @flags, (&get_flags($data{$file}{'__DATA__'}{$ch2}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
			}
			
			# and also flag from both channels
			if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			    push @flags, (&get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
			}
			
			
			# skip flagged spots
			my $skip = 0;
			if (%skip_flags) {
			    
			    foreach (@flags) {
				if (defined $skip_flags{$_}) {
				    $skip = 1;
				    last;
				}
			    }			
			}
			
			if ($skip) {
			    $skipped{'at least one of the values is flagged'}++;
			    next;
			}
			
			# if we have flags, put the data into a special bin
			my $flag_print = 0;
			if (@flags) {
			    # take the lowest-numbered flag
			    my $flag = (sort { $a <=> $b } @flags)[0];
			    $flags{$flag}{$val1}{$val2}++;
			    $flags_num{$flag}++;
			    $flag_print = 1;
			} else {
			    $no_flag++;
			}
			
			# store ratio in specified destination
			$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header1}{$mrow}{$mcol}{$row}{$col} = $val1;
			$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header2}{$mrow}{$mcol}{$row}{$col} = $val2;
			if ($take_log) {
			    $data{$file}{'__DATA__'}{$ch1}{"${header}_L"}{$mrow}{$mcol}{$row}{$col} = $val1;
			    $data{$file}{'__DATA__'}{$ch2}{"${header}_L"}{$mrow}{$mcol}{$row}{$col} = $val2;
			}
			
			# store values in specified destination
			unless ($flag_print) {
			    print DATA "$val1\t$val2\n";
			    $spots_out++;
			}
			
			if ($val1 > $all_maxx) {
			    $all_maxx = $val1;
			}
			if ($val1 < $all_minx) {
			    $all_minx = $val1;
			}
			if ($val2 > $all_maxy) {
			    $all_maxy = $val2;
			}
			if ($val2 < $all_miny) {
			    $all_miny = $val2;
			}

			# count number of dots
			$dots++;
		    }
		}
		
		close DATA;
		
		# write flag files:		
		if (%flags) {
		    foreach my $flag (sort { $a <=> $b } keys %flags) {
			my $flag_file = "R_in.".(basename ($file)).".$ch1.$ch2.$mrow,$mcol,$action.$action_num.$action_step.$PID.$flag";
			$flag_file =~ s/\//_/g;
			$flag_file =~ s/ /_/g;
			$flag_file = "$tmp_dir/$flag_file";
			open (FLAG, ">$flag_file")
			    or die "Can't write to $flag_file: $!";
			push @{$flag_files{$coord}{$flag}}, $flag_file;
			foreach my $val1 (keys %{$flags{$flag}}) {
			    foreach my $val2 (keys %{$flags{$flag}{$val1}}) {
				print FLAG "$val1\t$val2\n";
			    }
			}
			close FLAG;
		    }
		}
	    }	    
	}
    }

    # overwrite with preferred flag colours
    foreach my $flag (sort { $a <=> $b } keys %flags_num) {
	if (defined $flag2col_tmp{$flag}) {
	    unless (defined $pref_flag2col{$flag}) {
		$pref_flag2col{$flag} = $flag2col_tmp{$flag};
		$pref_flag2col{$flag2col_tmp{$flag}} = $flag;
	    }
	} else {
	    if (defined $pref_flag2col{$flag}) {
		$flag2col_tmp{$flag} = $pref_flag2col{$flag};
	    } else {
		foreach my $colour (@pref_flag_colours) {
		    unless (defined $pref_flag2col{$colour}) {
			$flag2col_tmp{$flag} = $colour;
			$pref_flag2col{$colour} = $flag;
			$pref_flag2col{$flag} = $colour;
			last;
		    }
		}
	    }
	}
    }

    # create scatterplot    

    my $unflagged_colour = 'black';
    my $file_tmp = basename $file;
    $file_tmp =~ s/\//_/g;
    $file_tmp =~ s/ /_/g;
    my $R_commands_file = "$tmp_dir/R_cmd.$file_tmp.$round.$action.$action_num.$action_step.$PID";
    my $R_out = "$tmp_dir/R_out.$file_tmp.$round.$action.$action_num.$action_step.$PID";
    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";
    print RH "postscript(\"$R_out\",horizontal=FALSE)\n";

    print RH "nf <- layout(matrix(c(1:".($dim_row*$dim_col)."),$dim_row,$dim_col,byrow=TRUE))\n";
    
    my $colour_subgrids = &get_user_settings($var_def_default{'signal_coloured'},'signal_coloured',$force_global_opt,$action_num,$file,'__ALL__');
    my $first_grid = '';
    my $last_grid = '';
    my @remove_files = ();
    foreach my $coord (@plots) {
	my ($start_row,$start_col,$end_row,$end_col) = split /,/, $coord;
	my $plotted = 0;
 
	my @colours = ('black','orange','green','blue','cyan','magenta','yellow','grey');
#	if ($colour_subgrids eq 'yes') {
#	    $add_info = 0;
#	}
	if ($colour_subgrids eq 'yes') {
	    $unflagged_colour = 'various';
	} else {
	    $unflagged_colour = &get_user_settings($unflagged_colour,'signal_flag_colour0',$force_global_opt,$action_num,$file,'__ALL__');
	    if ($unflagged_colour eq 'various') {
		$unflagged_colour = 'black';
	    }
	    @colours = ($unflagged_colour);
	}
	my @labels = ("'quartiles'");
	my @label_colours = ("'red'");
	my $pch = 'X';

	while (my $data_file = shift @{$data_files{$coord}}) {
	    
	    if (-s $data_file) {

		# read in data:
		print RH "fm <- read.table(\"$data_file\")\n";
		print RH "attach(fm)\n";
		
		my $colour = shift @colours;
		push @colours, $colour;
		push @label_colours, "'$colour'" if ($colour_subgrids eq 'yes');
		
		# draw scatterplot
		unless ($plotted) {
		    print RH "plot(V1,V2, xlab = \"$ch1\", ylab = \"$ch2\", pch = \".\", main = \"\", col = \"$colour\", xlim=range($all_minx,$all_maxx), ylim=range($all_miny,$all_maxy))\n";
		    print RH "allx = V1\n";
		    print RH "ally = V2\n";
		    $plotted++;
		} else {
		    print RH "points(V1,V2, pch = \".\", col = \"$colour\")\n";
		    print RH "allx = c(allx,V1)\n";		
		    print RH "ally = c(ally,V2)\n";		
		}
		$pch .= '.';
		push @labels, "'".(shift @{$coords{$coord}})."'" if ($colour_subgrids eq 'yes'); 
		
	    }
	    push @remove_files, $data_file;
	}

	# initialize flag arrays with values
	unless ($plotted) {
	    print RH "flagx = 0\n";		
	    print RH "flagy = 0\n";
	} else {		
	    print RH "flagx = min(allx)\n";		
	    print RH "flagy = min(ally)\n";		
	}

	# if we have flags, print them in different colours on top of the plot
	if (defined $flag_files{$coord}) {
	    foreach my $flag (sort { $a <=> $b } keys %{$flag_files{$coord}}) {
		foreach my $flag_file (@{$flag_files{$coord}{$flag}}) {
		    # read in data
		    print RH "flag$flag <- read.table(\"$flag_file\")\n";
		    unless ($plotted) {
			print RH "plot(flag$flag\$V1,flag$flag\$V2, xlab = \"$ch1\", ylab = \"$ch2\", pch = \"*\", main = \"\", col = \"$flag2col_tmp{$flag}\", xlim=range($all_minx,$all_maxx), ylim=range($all_miny,$all_maxy))\n";
			print RH "allx = flag$flag\$V1\n";
			print RH "ally = flag$flag\$V2\n";
			$plotted++;
			# keep track of the values for positioning of the legend
			print RH "flagx = c(flagx,flag$flag\$V1)\n";
			print RH "flagy = c(flagy,flag$flag\$V2)\n";

		    } else {
			print RH "points(flag$flag\$V1,flag$flag\$V2, col = \"$flag2col_tmp{$flag}\", pch = \"*\")\n";
			# keep track of the values for positioning of the legend
			print RH "flagx = c(flagx,flag$flag\$V1)\n";
			print RH "flagy = c(flagy,flag$flag\$V2)\n";
		    }
		}
	    }
	}

	unless ($plotted) {
	    print RH "plot(0,0, xlab = \"$ch1\", ylab = \"$ch2\", pch = \".\", main = \"\", col = \"$flag2col_tmp{0}\", xlim=range($all_minx,$all_maxx), ylim=range($all_miny,$all_maxy))\n";
	    print RH "allx = 0\n";
	    print RH "ally = 0\n";
	}
	print RH "abline(0,1, col='red')\n";
	
	# TODO: lines for two-fold change
	# make optional 
	if ($take_log
	    or
	    $header =~ /_L/) {
	    print RH "abline(1,1, col='grey')\n";
	    print RH "abline(-1,1, col='grey')\n";
	}

	# add red crosses for 25% and 75% quartile:
	print RH "q1 <- quantile(allx,c(0.25,0.75))\n";
	print RH "q2 <- quantile(ally,c(0.25,0.75))\n";
	print RH "points(q1[1],q2[1], pch='X', col='red')\n";
	print RH "points(q1[2],q2[2], pch='X', col='red')\n";
	
	# add legend (include flag points for finding best destination)       
	print RH "allx = c(allx,flagx)\n";
	print RH "ally = c(ally,flagy)\n";
	print RH "lx = min(allx)\n";
	print RH "ly = max(ally) - (max(ally) - min(ally))/10\n";	
#	print RH "ly = max(ally)\n";	
	if ($colour_subgrids
	    and
	    @labels > 2) {
	    print RH "legend(lx,ly, c(".(join ',', @labels)."), pch='$pch', col=c(".(join ',', @label_colours)."))\n";
	} else {
	    print RH "legend(lx,ly, 'quartiles', pch='X', col='red')\n";
	}
 
	if ("$start_row:$start_col" eq "$end_row:$end_col") {
	    print RH "title(main = \"subgrid $end_row:$end_col\", line=-1)\n";
	} else {
	    print RH "title(main = \"subgrids $start_row:$start_col - $end_row:$end_col\", line=-1)\n";
	}

	unless ($first_grid) {
	    $first_grid = "$start_row:$start_col";
	}
	$last_grid = "$end_row:$end_col";
    }

    print RH "mtext(\"Signal scatter plot for ".(basename $file).", $first_grid - $last_grid\",side=3,line=-1.5,font=2,outer=TRUE)\n";
    #print RH "title(main = \"Signal scatter plot for ".(basename $file)."\", line=-1)\n";

    print RH "dev.off()\n";
    close RH;
   
    my $image = &get_R_output($R_commands_file, $R_out, $action_num, $file);
    
    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	foreach my $data_file (@remove_files) {
	    unlink $data_file;
	}
    }

    my $info = "<B>$dots</B> dots plotted<BR>\n";
    if (%flags_num) {
	$info .=  "<BR>
     <TABLE>
      <TR>
       <TH>flag</TH><TH>#</TH><TH>colour</TH>
      </TR>\n";

	my $colour_field = '';
	$colour_field = "<SELECT name=\"action$action_num${file}signal_flag_colour0\" size=\"1\">";
	foreach my $col ($unflagged_colour, @pref_flag_colours) {
	    $colour_field .= "<OPTION".($col eq $unflagged_colour ? ' SELECTED' : '').">$col</OPTION>\n";
	}
	$colour_field .= "\n</select>\n";

	$info .= "<TR><TD>unflagged</TD><TD><B>$no_flag</B></TD><TD ALIGN=RIGHT>$colour_field</TD></TR>\n";
	foreach my $flag (sort { $a <=> $b } keys %flags_num) {
	    # make field to choose colour:
	    my $colour_field = '';
	    $colour_field = "<SELECT name=\"action$action_num${file}signal_flag_colour$flag\" size=\"1\">";
	    foreach my $col (@pref_flag_colours) {
		my $selected = '';
		if (defined $flag2col_tmp{$flag}
		    and
		    $flag2col_tmp{$flag} eq $col) {
		    $selected = ' SELECTED';
		}
		$colour_field .= "<OPTION$selected>$col</OPTION>\n";
	    }
	    $colour_field .= "\n</select>\n";
	    
	    $info.=  "
      <TR>
       <TD ALIGN=RIGHT>$flag: $flag_values_num{$flag}</TD><TD ALIGN=RIGHT><B>$flags_num{$flag}</B></TD><TD ALIGN=RIGHT>$colour_field</TD>
      </TR>";
	}
	$info .=  "
     </TABLE>
";
    }
	    
    foreach (keys %skipped) {
	$info .= "<B>$skipped{$_}</B> spots were skipped because<BR>$_!<BR>\n";
    }

    unless ($add_info) {
	$info = '';
    }
    
   # take note of available columns for later

    if ($take_log) {
	foreach my $ch ($ch1,$ch2) {
	    # this is not a good idea, for example, if the unlogged value is calculated again by a different method,
	    # then the logged value has probably preference but doesn't reflect the latest status anymore!
#	    unless (defined $data_column_available{$file}{$ch}{"${header}_L"}) {
#		$data_column_available{$file}{$ch}{"${header}_L"} = 1;
#		push @{$data_column_available{$file}{'__NO_LABEL__'}{'__ORDER__'}}, "${header}_L";
#	    }
#	    $data_done{$file}{$ch}{"${header}_L"}++;
	}
    }

    return $image, $info;
} 


sub rep_signal_plotting {
    my $action_num = shift;
    my $file1 = shift;
    my $ch1 = shift;
    my $file2 = shift;
    my $ch2 = shift;
    my $assign = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $header = '';
    foreach my $column (@{$item_data_column{$action}}) {

	my $occ = 0;

	if ($data_column_available{$file1}{$ch1}{$column}) {
	    $header = $column;
	    $occ++;
	}

	if ($data_column_available{$file2}{$ch2}{$column}) {
	    $header = $column;
	    $occ++;
	}

	if ($occ == 2) {
	    last;
	} else {
	    $header = '';
	}
    }

    # check for specific settings:
    ($header) = &get_user_settings($header,'rep_signal_scatter_plot_data_column',$force_global_opt,$action_num,'','__ALL__');

    # check for specific settings:
    my ($take_log) = &get_user_settings($var_def_default{'rep_signal_log'},'rep_signal_log',$force_global_opt,$action_num,'','__ALL__');
    $take_log = $take_log eq 'yes' ? 1 : 0;

    if ($header =~ /_L/) {
	$take_log = 0;
    }

    $last_column_used = $header;

    # collect labels
    my %labels1 = ();
    my %labels2 = ();

    foreach my $label (sort keys %{$data{$file1}{'__DATA__'}}) {
	
	# skip columns that are not label specific ('spot labels')
	next if ($label eq '__NO_LABEL__');
	$labels1{$label}++;
    }

    foreach my $label (sort keys %{$data{$file2}{'__DATA__'}}) {
	
	# skip columns that are not label specific ('spot labels')
	next if ($label eq '__NO_LABEL__');
	$labels2{$label}++;
    }

    unless ($labels1{$ch1}) {
	warn "# ERROR: Label '$ch1' is not defined for ".(basename $file1)."!";
	die "# ERROR: Label '$ch1' is not defined for ".(basename $file1)."!\n";
    }

    unless ($labels2{$ch2}) {
	warn "# ERROR: Label '$ch2' is not defined for ".(basename $file2)."!";
	die "# ERROR: Label '$ch2' is not defined for ".(basename $file2)."!\n";
    }


#    my $file = $files[0];

    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'rep_signal_skip_flags'},'rep_signal_skip_flags',$force_global_opt,$action_num,'','__ALL__');
    %skip_flags = &get_skip_flags(@flags);

    my %flag2col_tmp = ();
    foreach my $flag (0, (sort { $a <=> $b } keys %flag_values_num)) {
 
	my $default = '';
	if (defined $pref_flag2col{$flag}) {
	    $default = $pref_flag2col{$flag};
	}
	# check for user specific settings
	my $flag_colour = &get_user_settings($default,'rep_signal_scatter_plot_flag_colour'.$flag,$force_global_opt,$action_num,'','__ALL__');
	$flag2col_tmp{$flag} = $flag_colour if ($flag_colour =~ /\w/);	    
    }

    my %values = ();
    my ($mrow,$mcol,$row,$col);
    my %skipped = ();
    my %flags = ();
    my %flag_num = ();
    my $no_flag = 0;
    my $dots = 0;

    # define data file:
    my $data_file = "R_in.".(basename ($file1)).'.'.(basename ($file2)).".$ch1.$ch2.$action.$action_num.$action_step.$PID";
    $data_file =~ s/\//_/g;
    $data_file =~ s/ /_/g;
    $data_file = "$tmp_dir/$data_file";
    open (DATA, ">$data_file") 
	or die "Can't write to $data_file: $!";
	
    # go through each field
    foreach $mrow (keys %{$data{$file1}{'__DATA__'}{$ch1}{$header}}) {
	foreach $mcol (keys %{$data{$file1}{'__DATA__'}{$ch1}{$header}{$mrow}}) {
	    foreach $row (keys %{$data{$file1}{'__DATA__'}{$ch1}{$header}{$mrow}{$mcol}}) {
		foreach $col (keys %{$data{$file1}{'__DATA__'}{$ch1}{$header}{$mrow}{$mcol}{$row}}) {

		    my $val1 = '';
		    my $val2 = '';

		    # retrieve values from both channels
		    if (defined $data{$file1}{'__DATA__'}{$ch1}{$header}{$mrow}{$mcol}{$row}{$col}) {
			$val1 = $data{$file1}{'__DATA__'}{$ch1}{$header}{$mrow}{$mcol}{$row}{$col};
		    }
		    if (defined $data{$file2}{'__DATA__'}{$ch2}{$header}{$mrow}{$mcol}{$row}{$col}) {
			$val2 = $data{$file2}{'__DATA__'}{$ch2}{$header}{$mrow}{$mcol}{$row}{$col};
		    }

		    # skip unless both values set
		    unless ($val1 =~ /\d/
			    and
			    $val2 =~ /\d/) {
			$skipped{'at least one of the values is undefined'}++;
			next;
		    }

		    if (($val2 <= 0
			 or
			 $val1 <= 0)
			and
			$take_log > 0) {
			# set flag
			$skipped{'at least one of the values is <= 0 (no log possible)'}++;
			next;
		    } else {
			if ($take_log) {
			    $val1 = &log_base(2,$val1);
			    $val2 = &log_base(2,$val2);
			}
		    }

		    # get flags for first channel:
		    my @flags = ();
		    if (defined $data{$file1}{'__DATA__'}{$ch1}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			@flags = &get_flags($data{$file1}{'__DATA__'}{$ch1}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
		    }

		    # get flags for second channel as well
		    if (defined $data{$file2}{'__DATA__'}{$ch2}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			push @flags, (&get_flags($data{$file2}{'__DATA__'}{$ch2}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
		    }

		    # and also flag from both channels
		    if (defined $data{$file1}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			push @flags, (&get_flags($data{$file1}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
		    }
		    if (defined $data{$file2}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			push @flags, (&get_flags($data{$file2}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
		    }
		    

		    # skip flagged spots
		    my $skip = 0;
		    if (%skip_flags) {

			foreach (@flags) {
			    if (defined $skip_flags{$_}) {
				$skip = 1;
				last;
			    }
			}			
		    }
			
		    if ($skip) {
			$skipped{'at least one of the values is flagged'}++;
			next;
		    }

		    # if we have flags, put the data into a special bin
		    my $flag_print = 0;
		    if (@flags) {
			# take the lowest-numbered flag
			my $flag = (sort { $a <=> $b } @flags)[0];
			$flags{$flag}{$val1}{$val2}++;
			$flag_num{$flag}++;
			$flag_print = 1;
		    } else {
			$no_flag++;
		    }

		    # store values in specified destination
		    unless ($flag_print) {
			print DATA "$val1\t$val2\n";
		    }

		    # count number of dots
		    $dots++;
		}
	    }
	}
    }

    close DATA;

    my %flag_files = ();
    if (%flags) {
	foreach my $flag (sort { $a <=> $b } keys %flags) {
	    my $flag_file = "$data_file.$flag";
	    open (FLAG, ">$flag_file")
		or die "Can't write to $flag_file: $!";
	    $flag_files{$flag} = $flag_file;
	    foreach my $val1 (keys %{$flags{$flag}}) {
		foreach my $val2 (keys %{$flags{$flag}{$val1}}) {
		    print FLAG "$val1\t$val2\n";
		}
	    }
	    close FLAG;

	    # get flag colours
	    if (defined $flag2col_tmp{$flag}) {
		unless (defined $pref_flag2col{$flag}) {
		    $pref_flag2col{$flag} = $flag2col_tmp{$flag};
		    $pref_flag2col{$flag2col_tmp{$flag}} = $flag;
		}
	    } else {
		if (defined $pref_flag2col{$flag}) {
		    $flag2col_tmp{$flag} = $pref_flag2col{$flag};
		} else {
		    foreach my $colour (@pref_flag_colours) {
			unless (defined $pref_flag2col{$colour}) {
			    $flag2col_tmp{$flag} = $colour;
			    $pref_flag2col{$colour} = $flag;
			    $pref_flag2col{$flag} = $colour;
			    last;
			}
		    }
		}
	    }
	}
    }

      
    # create scatterplot
    my $draw_subgrids = '';

    my $file_short1 = basename ($file1);
    my $file_short2 = basename ($file2);
    # get rid of file ending
    $file_short1 =~ s/\.\w{3}$//;
    if (length($file_short1) > 8) {
	$file_short1 =~ s/.*(.{8})$/$1/;
    }
    # get rid of file ending
    $file_short2 =~ s/\.\w{3}$//;
    if (length($file_short2) > 8) {
	$file_short2 =~ s/.*(.{8})$/$1/;
    }
    my $R_commands_file = "$tmp_dir/R_cmd.$file_short1.$file_short2.$ch1.$ch2.$action.$action_num.$action_step.$PID";    
    my $R_out = "$tmp_dir/R_out.$file_short1.$file_short2.$ch1.$ch2.$action.$action_num.$action_step.$PID";
    my $ending = '';
    while (-e "$R_out$ending") {
	$ending++;
    }
    $R_out .= $ending;

    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";
    print RH "postscript(\"$R_out\",horizontal=FALSE)\n";

    # read in data
    print RH "fm <- read.table(\"$data_file\")\n";
    print RH "attach(fm)\n";
    
    # draw scatterplot
    print RH "plot(V1,V2, xlab = \"".(basename $file1).",$ch1\", ylab = \"".(basename $file2).",$ch2\", pch = \".\", main = \"\", col = \"$flag2col_tmp{0}\")\n";

    # if we have flags, print them in different colours on top of the plot
    if (%flag_files) {
	foreach my $flag (sort { $a <=> $b } keys %flag_files) {
	    my $flag_file = $flag_files{$flag};
	    # read in data
	    print RH "flag$flag <- read.table(\"$flag_file\")\n";
	    print RH "points(flag$flag\$V1,flag$flag\$V2, col = \"$flag2col_tmp{$flag}\", pch = \"*\")\n";
	}
    }
    
    print RH "abline(0,1, col='grey')\n";

    # add red crosses for 25% and 75% quartile:
    print RH "q1 <- quantile(V1,c(0.25,0.75))\n";
    print RH "q2 <- quantile(V2,c(0.25,0.75))\n";
    print RH "points(q1[1],q2[1], pch='X', col='red')\n";
    print RH "points(q1[2],q2[2], pch='X', col='red')\n";
    
    # add legend
    print RH "legend(min(V1), max(V2), 'quartiles', pch='X', col='red')\n";

    print RH "title(main = \"Inter-slide scatter plot for $assign\", line=-1)\n";

    print RH "dev.off()\n";
    close RH;

    my $image = &get_R_output($R_commands_file, $R_out, $action_num, $file1);
    
    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	unlink $data_file;
    }

    my $info = "<B>$dots</B> dots plotted<BR>\n";
    if (%flags) {
	$info .=  "<BR>
     <TABLE>
      <TR>
       <TH>flag</TH><TH>#</TH><TH>colour</TH>
      </TR>\n";

	my $colour_field = '';
	$colour_field = "<SELECT name=\"action$action_num${file1}signal_flag_colour0\" size=\"1\">";
	foreach my $col ('black', @pref_flag_colours) {
	    $colour_field .= "<OPTION".($col eq 'black' ? ' SELECTED' : '').">$col</OPTION>\n";
	}
	$colour_field .= "\n</select>\n";

	$info .= "<TR><TD>unflagged</TD><TD><B>$no_flag</B></TD><TD ALIGN=RIGHT>$colour_field</TD></TR>\n";
	foreach my $flag (sort { $a <=> $b } keys %flags) {
	    # make field to choose colour:
	    my $colour_field = '';
	    $colour_field = "<SELECT name=\"action$action_num${file1}signal_flag_colour$flag\" size=\"1\">";
	    foreach my $col (@pref_flag_colours) {
		my $selected = '';
		if (defined $flag2col_tmp{$flag}
		    and
		    $flag2col_tmp{$flag} eq $col) {
		    $selected = ' SELECTED';
		}
		$colour_field .= "<OPTION$selected>$col</OPTION>\n";
	    }
	    $colour_field .= "\n</select>\n";
	    
	    $info.=  "
      <TR>
       <TD ALIGN=RIGHT>$flag: $flag_values_num{$flag}</TD><TD ALIGN=RIGHT><B>$flag_num{$flag}</B></TD><TD ALIGN=RIGHT>$colour_field</TD>
      </TR>";
	}
	$info .=  "
     </TABLE>
";
    }
	    
    foreach (keys %skipped) {
	$info .= "<B>$skipped{$_}</B> spots were skipped because<BR>$_!<BR>\n";
    }
    
    return $image, $info;
} 


sub rep_signal_plotting_setup {
    my $action_num = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my @files = sort keys %data;

    my %channel_assign = &channel_assign(@files);  
    
    foreach my $assign (sort keys %channel_assign) {
	my @files = sort keys %{$channel_assign{$assign}};
	next unless (@files);
	unless (@files > 1) {
	    # we need at least 2 files to plot two channels against each other
	    print &format_output( # 'rep_signal_scatter_plot_',$action_num,'',$assign,$action,'-',"Only 1 file for '$assign': ".(basename $files[0])
				 {
				      'prefix' => 'rep_signal_scatter_plot_',
				      'num' => $action_num,
				      'label' => $assign,
				      'action' => $action,
				      'image' => '-',
				      'info' => "Only 1 file for '$assign': ".(basename $files[0])
				      }
				  );
	    next;
	}

	# plot each channel against each other:
	for (my $i = 0; $i < $#files; $i++) {
	    for (my $j = $i+1; $j <= $#files; $j++) {
		my ($file1,$file2) = ($files[$i], $files[$j]);
		my $ch1 = $channel_assign{$assign}{$file1}{'CH'};
		my $ch2 = $channel_assign{$assign}{$file2}{'CH'};
		# carry out action on file combination
#		warn "plot $file1, $ch1 against $file2, $ch2";
		my ($image,$info) = &rep_signal_plotting($action_num,$file1,$ch1,$file2,$ch2,$assign,$action);
		
		print &format_output( # 'rep_signal_scatter_plot_',$action_num,$file1,'__ALL__',$action,$image,$info
				     {
					 'prefix' => 'rep_signal_scatter_plot_',
					 'num' => $action_num,
					 'file' => $file1,
					 'label' => '__ALL__',
					 'action' => $action,
					 'image' => $image,
					 'info' => $info
					 }
				     );
	    }
	}
    }
}


sub ratio_signal_plotting_setup {
    my $action_num = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my @files = &get_user_settings($var_def_default{'ratio_signal_files'},'ratio_signal_files',$force_global_opt,$action_num,'','__ALL__');
    my %file_setup = ();

    if (grep /^auto$/, @files) {
	# apply the best setting
	my $choice = '';
	foreach (@files) {
	    if (/^$MERGED_FILE_NAME.*__/) {
		$choice = 'merged between groups';
		last;
	    } elsif (/^$MERGED_FILE_NAME/) {
		$choice = 'merged';
	    }
	}
	if ($choice eq 'merged') {
	    $choice = 'merged within groups';
	}
	unless ($choice) {
	    $choice = 'between technical replicates';
	}
	@files = ($choice);
    }
		
    if (grep /^all$/, @files) {

	@files = sort keys %data;
	@{$file_setup{'01 all vs all'}} = @files;

    } elsif (grep /^merged between groups$/, @files) {
	@files = ();
	foreach my $file (sort keys %data) {
	    my $base = basename($file);
	    next unless ($base =~ /^$MERGED_FILE_NAME.*__/);
	    push @files, $file;
	}
	@{$file_setup{'02 merged between groups'}} = @files;
    } elsif (grep /between groups$/, @files) {
	my $merged_only = 0;
	@files = sort keys %data;

	my $counter = 0;
	my $counter_out = 0;
	my %counter = ();
	foreach my $file (@files) {
	    
	    my $full_file = $file;

	    if ($merged_only) {
		my $base = basename($file);
		next unless ($base =~ /^$MERGED_FILE_NAME.*__/);
	    }
	    
	    my ($ch1_assign,$ch2_assign);
	    
	    if (defined $cgi->param("CH1_ANNOTATE_$full_file")) {
		$ch1_assign = $cgi->param("CH1_ANNOTATE_$full_file");
	    }
	    if (defined $cgi->param("CH2_ANNOTATE_$full_file")) {
		$ch2_assign = $cgi->param("CH2_ANNOTATE_$full_file");
	    }
	    
	    if ($counter{"$ch1_assign vs $ch2_assign"}) {
		$counter_out = $counter{"$ch1_assign vs $ch2_assign"};
	    } else {
		$counter++;
		$counter_out = $counter;
		until ((length $counter_out) >= 3) {
		    $counter_out = '0'.$counter_out;
		}
		$counter{"$ch1_assign vs $ch2_assign"} = $counter_out;
	    }
	    push @{$file_setup{"$counter_out $ch1_assign vs $ch2_assign"}}, $file;
	}
    } elsif (grep /^within groups$/, @files
	     or
	     grep /^between technical replicates$/, @files) {

	my $merged_only = 0;
	if (grep /^merged within groups$/, @files) {
	    $merged_only = 1;
	}
	my $tech_reps = 0;
	if (grep /^between technical replicates$/, @files) {
	    $tech_reps = 1;
	}

        @files = sort keys %data;

        my $counter = 0;
        my $counter_out = 0;
        my %counter = ();
        foreach my $file (@files) {

            my $full_file = $file;

	    if ($merged_only) {
		my $base = basename($file);
		next unless ($base =~ /^$MERGED_FILE_NAME/);
	    }
	    
            my ($ch1_assign,$ch2_assign, $bio_rep);

            if (defined $cgi->param("CH1_ANNOTATE_$full_file")) {
                $ch1_assign = $cgi->param("CH1_ANNOTATE_$full_file");
            }
            if (defined $cgi->param("CH2_ANNOTATE_$full_file")) {
                $ch2_assign = $cgi->param("CH2_ANNOTATE_$full_file");
            }

	    if ($tech_reps) {
		# consider technical replicates per biological sample:
		if (defined $cgi->param("BIO_REP_$full_file")) {
		    $bio_rep = $cgi->param("BIO_REP_$full_file");
		}
	    }

	    if ($bio_rep eq '') {
		$bio_rep = ', bio_rep1';
	    } else {
		if ($bio_rep =~ /^\d/) {
		    $bio_rep = ', B'.$bio_rep;
		} else {
		    $bio_rep = ", $bio_rep";
		}
	    }

            if ($counter{"$ch1_assign vs $ch2_assign$bio_rep"}) {
                $counter_out = $counter{"$ch1_assign vs $ch2_assign$bio_rep"};
            } else {
                $counter++;
                $counter_out = $counter;
                until ((length $counter_out) >= 3) {
                    $counter_out = '0'.$counter_out;
                }
                $counter{"$ch1_assign vs $ch2_assign$bio_rep"} = $counter_out;
            }
            push @{$file_setup{"$counter_out $ch1_assign vs $ch2_assign$bio_rep"}}, $file;
        }
    } else {
	@{$file_setup{'01 selection vs selection'}} = @files;
    }

    my $out_header = $RATIO_DIFF_TERM;
    # initialize output field:
    foreach (sort keys %file_setup) {
	@files = @{$file_setup{$_}};
	foreach my $file (@files) {
	    unless (defined $data_column_available{$file}{'__NO_LABEL__'}{$out_header}) {
		$data_column_available{$file}{'__NO_LABEL__'}{$out_header} = 1;
		push @{$data_column_available{$file}{'__NO_LABEL__'}{'__ORDER__'}}, $out_header;
	    }
	    $data_done{$file}{'__NO_LABEL__'}{$out_header}++;

	    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);
	    foreach my $mrow (1..$MROW) {
		foreach my $mcol (1..$MCOL) {
		    foreach my $row (1..$ROW) {
			foreach my $col (1..$COL) {
			    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header}{$mrow}{$mcol}{$row}{$col} = '';
			}
		    }
		}
	    }
	}
    }

    foreach (sort keys %file_setup) {
	@files = @{$file_setup{$_}};
	my ($num,$title) = split / /, $_, 2;
    
	unless (@files > 1) {
	    next unless ($files[0] =~ /\d/);
	    # we need at least 2 files to plot two channels against each other
	    print &format_output( # 'ratio_signal_scatter_plot_',$action_num,'',$assign,$action,'-',"Only 1 file for '$assign': ".(basename $files[0])
				 {
				      'prefix' => 'ratio_signal_scatter_plot_',
				      'num' => $action_num,
				      'label' => '__NO_LABEL__',
				      'action' => $action,
				      'image' => '-',
				      'info' => "Only 1 file $title: ".(basename $files[0])
				      }
				  );
	    next;
	}

	# plot each channel against each other:
	my $plot_counter = 0;
	for (my $i = 0; $i < $#files; $i++) {
	    for (my $j = $i+1; $j <= $#files; $j++) {
		my ($file1,$file2) = ($files[$i], $files[$j]);
		# carry out action on file combination
#		warn "plot $file1, $ch1 against $file2, $ch2";
		$plot_counter++ if (@files > 2);
		my ($image,$info) = &ratio_signal_plotting($action_num,$file1,$file2,$title,$action, $out_header, $plot_counter);
		
		print &format_output( # 'ratio_signal_scatter_plot_',$action_num,$file1,'__ALL__',$action,$image,$info
				     {
					 'prefix' => 'ratio_signal_scatter_plot_',
					 'num' => $action_num,
					 'file' => $file1,
					 'label' => '__ALL__',
					 'action' => $action,
					 'image' => $image,
					 'info' => $info
					 }
				     );
	    }
	}
    }
}


sub ratio_signal_plotting {
    my $action_num = shift;
    my $file1 = shift;
    my $file2 = shift;
    my $assign = shift;
    my $action = shift;
    my $out_header = shift;
    my $plot_counter = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    # get the best header per file
    my %header = ();    
    my $header = '';
    my $check_ch = '__NO_LAEL__';
    my %take_log = ();
    foreach my $file ($file1, $file2) {
	foreach my $column (@{$item_data_column{$action}}) {
	    
	    my $occ = 0;
	    
	    if ($data_column_available{$file}{$check_ch}{$column}) {
		$header = $column;
		$header{$file} = $header;
		if ($header =~ /_L/) {
		    $take_log{$file} = 0;
		} else {
		    $take_log{$file} = 1;
		}
		
		last;
	    }
	}
    }
    
    unless (scalar keys %header == 2) {
	# didn't find header for all files
	foreach my $file ($file1, $file2) {
	    next if (defined $header{$file});
	    
	    # find the best single channel from which a ratio will be calculated
	  COLUMN:foreach my $column (@{$item_data_column{$action}}) {
	      $column =~ s/_R//;
	      my $occ = 0;
	      foreach my $ch ('CH1', 'CH2') {
		  
		  if ($data_column_available{$file}{$ch}{$column}) {
		      $header = $column;
		      $occ++;
		  }
		  if ($occ == 2) {
		      $header{$file} = $header;
		      $check_ch = 'CH1';
		      
		      if ($header =~ /_L/) {
			  $take_log{$file} = 0;
		      } else {
			  $take_log{$file} = 1;
		      }
		      
		      last COLUMN;
		  }
	      }
	  }
	}
    }

    # check for specific settings:
    ($header) = &get_user_settings($header,'ratio_signal_scatter_plot_data_column',$force_global_opt,$action_num,'','__ALL__');
    foreach my $file ($file1, $file2) {
	$header{$file} = $header;
	if ($header =~ /_L/) {
	    $take_log{$file} = 0;
	} else {
	    $take_log{$file} = 1;
	}
    }
    $last_column_used = $header;

    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'ratio_signal_skip_flags'},'ratio_signal_skip_flags',$force_global_opt,$action_num,'','__ALL__');
    %skip_flags = &get_skip_flags(@flags);

    my %flag2col_tmp = ();
    foreach my $flag (0, (sort { $a <=> $b } keys %flag_values_num)) {
 
	my $default = '';
	if (defined $pref_flag2col{$flag}) {
	    $default = $pref_flag2col{$flag};
	}
	# check for user specific settings
	my $flag_colour = &get_user_settings($default,'ratio_signal_scatter_plot_flag_colour'.$flag,$force_global_opt,$action_num,'','__ALL__');
	$flag2col_tmp{$flag} = $flag_colour if ($flag_colour =~ /\w/);	    
    }

    my %values = ();
    my ($mrow,$mcol,$row,$col);
    my %skipped = ();
    my %flags = ();
    my %flag_num = ();
    my $no_flag = 0;
    my $dots = 0;

    # define data file:
    my $data_file = "R_in.".(basename ($file1)).'.'.(basename ($file2)).".$action.$action_num.$action_step.$PID";
    $data_file =~ s/\//_/g;
    $data_file =~ s/ /_/g;
    $data_file = "$tmp_dir/$data_file";
    open (DATA, ">$data_file") 
	or die "Can't write to $data_file: $!";

    # go through each field
    foreach $mrow (keys %{$data{$file1}{'__DATA__'}{$check_ch}{$header{$file1}}}) {
	foreach $mcol (keys %{$data{$file1}{'__DATA__'}{$check_ch}{$header{$file1}}{$mrow}}) {
	    foreach $row (keys %{$data{$file1}{'__DATA__'}{$check_ch}{$header{$file1}}{$mrow}{$mcol}}) {
		COL:foreach $col (keys %{$data{$file1}{'__DATA__'}{$check_ch}{$header{$file1}}{$mrow}{$mcol}{$row}}) {

		    my %vals = ();

		    # retrieve values from both channels
		    if ($check_ch eq '__NO_LABEL__') {
			foreach my $file ($file1, $file2) { 
			    if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$header{$file}}{$mrow}{$mcol}{$row}{$col}
				and $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$header{$file}}{$mrow}{$mcol}{$row}{$col} =~ /\d/) {
				my $val = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$header{$file}}{$mrow}{$mcol}{$row}{$col};
				if ($take_log{$file}) {
				    if ($val <= 0) {
					$skipped{'at least one of the values is <= 0 (no log possible)'}++;
					next COL;
				    }
				    $val = &log_base(2,$val);
				}
				$vals{$file}{'ratio'} = $val;
			    } else {
				$skipped{'at least one of the values is undefined'}++;
				next COL;
			    }
			}
		    } else {
			# we need to calculate the ratio first
			foreach my $file ($file1, $file2) { 
			    foreach my $ch ('CH1', 'CH2') {
				if (defined $data{$file}{'__DATA__'}{$ch}{$header{$file}}{$mrow}{$mcol}{$row}{$col}
				    and
				    $data{$file}{'__DATA__'}{$ch}{$header{$file}}{$mrow}{$mcol}{$row}{$col} =~ /\d/
				    ) {
				    $vals{$file}{$ch} = $data{$file}{'__DATA__'}{$ch}{$header{$file}}{$mrow}{$mcol}{$row}{$col};
				} else {
				    $skipped{'at least one of the values is undefined'}++;
				    next COL;
				}
			    }
			    if ($take_log{$file}) {
				if ($vals{$file}{'CH1'} == 0) {
				    $skipped{'at least one of the values has channel 1 of zero (no ratio possible)'}++;
				    next COL;
				}
				my $ratio = $vals{$file}{'CH2'} / $vals{$file}{'CH1'};
				if ($ratio <= 0) {
				    $skipped{'at least one of the values is <= 0 (no log possible)'}++;
				    next COL;
				}
				$vals{$file}{'ratio'} = &log_base(2,$ratio);
			    } else {
				$vals{$file}{'ratio'} = $vals{$file}{'CH2'} - $vals{$file}{'CH1'};
			    }
			}
		    }

		    # get flags for first channel:
		    my @flags = ();
		    foreach my $file ($file1,$file2) {
			foreach my $ch ('CH1', 'CH2', '__NO_LABEL__') {
			    if (defined $data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
				push @flags, &get_flags($data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
			    }
			}
		    }

		    # skip flagged spots
		    my $skip = 0;
		    if (%skip_flags) {

			foreach (@flags) {
			    if (defined $skip_flags{$_}) {
				$skip = 1;
				last;
			    }
			}			
		    }
			
		    if ($skip) {
			$skipped{'at least one of the values is flagged'}++;
			next;
		    }

		    # if we have flags, put the data into a special bin
		    my $val1 = $vals{$file1}{'ratio'};
		    my $val2 = $vals{$file2}{'ratio'};

		    # calculate the difference:
		    my $diff = 2**$val1 / 2**$val2;

#		    $val1 = 2**$val1;
#		    $val2 = 2**$val2;

		    foreach my $file ($file1,$file2) {
			if ($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header}{$mrow}{$mcol}{$row}{$col} =~ /\d/) {
			    
			    if (abs($diff) > abs($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header}{$mrow}{$mcol}{$row}{$col}) ) {
				$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header}{$mrow}{$mcol}{$row}{$col} = $diff;
			    }
			} else {
			    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header}{$mrow}{$mcol}{$row}{$col} = $diff;
			}
		    }

		    my $flag_print = 0;

		    if (@flags) {
			# take the lowest-numbered flag
			my $flag = (sort { $a <=> $b } @flags)[0];
			$flags{$flag}{$val1}{$val2}++;
			$flag_num{$flag}++;
			$flag_print = 1;
		    } else {
			$no_flag++;
		    }

		    # store values in specified destination
		    unless ($flag_print) {
			print DATA "$val1\t$val2\n";
		    }

		    # count number of dots
		    $dots++;
		}
	    }
	}
    }

    close DATA;

    my %flag_files = ();
    if (%flags) {
	foreach my $flag (sort { $a <=> $b } keys %flags) {
	    my $flag_file = "$data_file.$flag";
	    open (FLAG, ">$flag_file")
		or die "Can't write to $flag_file: $!";
	    $flag_files{$flag} = $flag_file;
	    foreach my $val1 (keys %{$flags{$flag}}) {
		foreach my $val2 (keys %{$flags{$flag}{$val1}}) {
		    print FLAG "$val1\t$val2\n";
		}
	    }
	    close FLAG;

	    # get flag colours
	    if (defined $flag2col_tmp{$flag}) {
		unless (defined $pref_flag2col{$flag}) {
		    $pref_flag2col{$flag} = $flag2col_tmp{$flag};
		    $pref_flag2col{$flag2col_tmp{$flag}} = $flag;
		}
	    } else {
		if (defined $pref_flag2col{$flag}) {
		    $flag2col_tmp{$flag} = $pref_flag2col{$flag};
		} else {
		    foreach my $colour (@pref_flag_colours) {
			unless (defined $pref_flag2col{$colour}) {
			    $flag2col_tmp{$flag} = $colour;
			    $pref_flag2col{$colour} = $flag;
			    $pref_flag2col{$flag} = $colour;
			    last;
			}
		    }
		}
	    }
	}
    }

      
    # create scatterplot
    my $draw_subgrids = '';

    my $file_short1 = basename ($file1);
    my $file_short2 = basename ($file2);
    # get rid of file ending
    $file_short1 =~ s/\.\w{3}$//;
    $file_short2 =~ s/\.\w{3}$//;
#    if (length($file_short2) > 8) {
#	$file_short2 =~ s/.*(.{8})$/$1/;
#    }
    my $R_commands_file = "$tmp_dir/R_cmd.$assign.$action.$action_num.$action_step.$plot_counter.$PID";    
    my $R_out = "$tmp_dir/R_out.$assign.$action.$action_num.$action_step.$plot_counter.$PID";
    my $ending = '';
    while (-e "$R_out$ending") {
	$ending++;
    }
    $R_out .= $ending;

    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";
    print RH "postscript(\"$R_out\",horizontal=FALSE)\n";

    # read in data
    print RH "fm <- read.table(\"$data_file\")\n";
    print RH "attach(fm)\n";
    
    # draw scatterplot
    print RH "plot(V1,V2, xlab = \"log-ratio $file_short1\", ylab = \"log-ratio $file_short2\", pch = \".\", main = \"\", col = \"$flag2col_tmp{0}\")\n";

    # if we have flags, print them in different colours on top of the plot
    if (%flag_files) {
	foreach my $flag (sort { $a <=> $b } keys %flag_files) {
	    my $flag_file = $flag_files{$flag};
	    # read in data
	    print RH "flag$flag <- read.table(\"$flag_file\")\n";
	    print RH "points(flag$flag\$V1,flag$flag\$V2, col = \"$flag2col_tmp{$flag}\", pch = \"*\")\n";
	}
    }
    
    print RH "abline(0,1, col='red')\n";
    print RH "abline(1,1, col='grey')\n";
    print RH "abline(-1,1, col='grey')\n";

    # add red crosses for 25% and 75% quartile:
    print RH "q1 <- quantile(V1,c(0.25,0.75))\n";
    print RH "q2 <- quantile(V2,c(0.25,0.75))\n";
    print RH "points(q1[1],q2[1], pch='X', col='red')\n";
    print RH "points(q1[2],q2[2], pch='X', col='red')\n";
    
    # add legend
    print RH "legend(min(V1), max(V2), 'quartiles', pch='X', col='red')\n";

    if ($plot_counter) {
	$plot_counter = " ($plot_counter)";
    } else {
	$plot_counter = '';
    }

    print RH "title(main = \"Inter-slide ratio scatter plot for $assign$plot_counter\", line=-1)\n";

    print RH "dev.off()\n";
    close RH;

    my $image = &get_R_output($R_commands_file, $R_out, $action_num, $file1);
    
    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	unlink $data_file;
    }

    my $info = "<B>$dots</B> dots plotted<BR>\n";
    if (%flags) {
	$info .=  "<BR>
     <TABLE>
      <TR>
       <TH>flag</TH><TH>#</TH><TH>colour</TH>
      </TR>\n";

	my $colour_field = '';
	$colour_field = "<SELECT name=\"action$action_num${file1}signal_flag_colour0\" size=\"1\">";
	foreach my $col ('black', @pref_flag_colours) {
	    $colour_field .= "<OPTION".($col eq 'black' ? ' SELECTED' : '').">$col</OPTION>\n";
	}
	$colour_field .= "\n</select>\n";

	$info .= "<TR><TD>unflagged</TD><TD><B>$no_flag</B></TD><TD ALIGN=RIGHT>$colour_field</TD></TR>\n";
	foreach my $flag (sort { $a <=> $b } keys %flags) {
	    # make field to choose colour:
	    my $colour_field = '';
	    $colour_field = "<SELECT name=\"action$action_num${file1}signal_flag_colour$flag\" size=\"1\">";
	    foreach my $col (@pref_flag_colours) {
		my $selected = '';
		if (defined $flag2col_tmp{$flag}
		    and
		    $flag2col_tmp{$flag} eq $col) {
		    $selected = ' SELECTED';
		}
		$colour_field .= "<OPTION$selected>$col</OPTION>\n";
	    }
	    $colour_field .= "\n</select>\n";
	    
	    $info.=  "
      <TR>
       <TD ALIGN=RIGHT>$flag: $flag_values_num{$flag}</TD><TD ALIGN=RIGHT><B>$flag_num{$flag}</B></TD><TD ALIGN=RIGHT>$colour_field</TD>
      </TR>";
	}
	$info .=  "
     </TABLE>
";
    }
	    
    foreach (keys %skipped) {
	$info .= "<B>$skipped{$_}</B> spots were skipped because<BR>$_!<BR>\n";
    }
    
    return $image, $info;
} 



sub signal_boxplotting {
    my $action_num = shift;
    my $action = shift;

    if ($refine_parameters) {
	return '', '', 'Paramter Refinement!';
    }

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my @files = ();
    foreach (sort keys %data) {
	next if ($_ eq '');
	push @files, $_;
    }

    my $header = '';
    BOXPLOT_COL:foreach my $column (@{$item_data_column{$action}}) {

	my $occ = 0;
	foreach my $label (sort {$b cmp $a} keys %{$data{$files[0]}{'__DATA__'}}) {
	    my $not_found = 0;
	    foreach my $file (@files) {
		unless (defined $data_column_available{$file}{$label}{$column}) {
		    $not_found = 1;
		}
	    }
	    
	    if ($not_found == 0) {
		$header = $column;
		last BOXPLOT_COL;
	    }
	}
    }

    # check for specific settings:
    $header = &get_user_settings($header,'boxplot_data_column',$force_global_opt,$action_num,'__ALL__');

    my @work_labels = &get_user_settings($var_def_default{'boxplot_labels'},'boxplot_labels',$force_global_opt,$action_num,'__ALL__');
    foreach (@work_labels) {
	if (/auto/) {
	    if ($header =~ /^Ratio_/
		or
		$header =~ /_R$/
		or
		$header =~ /_R_M$/) {
		@work_labels = ('__NO_LABEL__');
	    } else {
		@work_labels = ('CH1', 'CH2');
	    }
	}
    }
    my $combined = 0;
    foreach (@work_labels) {
	if (/ratio/) {
	    @work_labels = ('__NO_LABEL__');
	    last;
	}
	if (/combined/) {
	    @work_labels = ('CH1', 'CH2');
	    $combined = 1;
	    last;
	}
    }

    # check for specific settings:
    my $take_log = &get_user_settings($var_def_default{'boxplot_log'},'boxplot_log',$force_global_opt,$action_num,'__ALL__');
    $take_log = $take_log eq 'yes' ? 1 : 0;

    if ($header =~ /_L/) {
	$take_log = 0;
    }

    $last_column_used = $header;


    my @flags = &get_user_settings($var_def_default{'boxplot_skip_flags'},'boxplot_skip_flags',$force_global_opt,$action_num,'__ALL__');
    my %skip_flags = ();
    %skip_flags = &get_skip_flags(@flags);

    my %values = ();
    my ($mrow,$mcol,$row,$col);
    my %skipped = ();
    my %flags = ();
    my $no_flag = 0;
    my %dots = ();

    
    my $ch1 = $work_labels[0];
    my $ch2 = $work_labels[1];
    my $use_ratios = 0;

    if ($ch1 eq '__NO_LABEL__') {
	$ch2 = '';
	$use_ratios = 1;
    }

    my $out_header1 = "$ch1.signal_plot";
    my $out_header2 = "$ch2.signal_plot";

    my %skip_coord = ();

    my @data_files = ();
    my $file_counter = 0;
    my $min_val = 99999;
    my $max_val = -99999;
    my %order = ();
    my @plot_labels = ();

    # go through each file
    foreach my $file (@files) {

	my %ch1 = ();

	foreach my $ch (sort keys %{$data{$file}{'__DATA__'}}) {
	    next if ($ch eq '__NO_LABEL__');

	    # open data file:
	    my $data_file = "data_f${file_counter}_${ch}_$header.$action.$action_num.$action_step.$PID.tmp";
	    $data_file =~ s/ /_/g;
	    $data_file =~ s/-/_/g;
	    $data_file = "$tmp_dir/$data_file";

	    unless ($use_ratios 
		    and
		    $ch eq 'CH1') {
		push @data_files, $data_file;
		$file_counter++;

		open (OUT, ">$data_file")
		    or return "Can't write to $data_file: $!";
		
		$order{$file_counter}{'DATA'} = $data_file;
		$order{$file_counter}{'FULL_NAME'} = $file;
		$order{$file_counter}{'NAME'} = basename $file;
		$order{$file_counter}{'CHANNEL'} = $ch;
		if ($ch eq 'CH1') {
		    if (defined $file_specs{$file}{'CH1_LABEL'}) {
			$order{$file_counter}{'CHANNEL'} = $file_specs{$file}{'CH1_LABEL'};
		    }
		} else {
		    if (defined $file_specs{$file}{'CH2_LABEL'}) {
			$order{$file_counter}{'CHANNEL'} = $file_specs{$file}{'CH2_LABEL'};
		    }
		}
		if ($use_ratios) {
		    if ($header =~ /_L/
			or
			$take_log) {
			push @plot_labels, 'CH2 - CH1';
		    } else {
			push @plot_labels, 'CH2 / CH1';
		    }
		} else {
		    push @plot_labels, $order{$file_counter}{'CHANNEL'};
		}
	    }

	    # go through each field
	    foreach $mrow (keys %{$data{$file}{'__DATA__'}{$ch}{$header}}) {
		foreach $mcol (keys %{$data{$file}{'__DATA__'}{$ch}{$header}{$mrow}}) {
		    foreach $row (keys %{$data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}}) {
			foreach $col (keys %{$data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}}) {
			    
			    # retrieve value
			    my $val = '';
			
			    if (defined $data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col}) {
				$val = $data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col};
			    }

			    # skip unless proper value
			    unless ($val =~ /\d/) {
				$skipped{$file}{$ch}{'values not a number'}++;
				next;
			    }
			    
			    if ($take_log
				and
				($val <= 0
				 or 
				 $val eq '')
				) {
				# set flag
				$skipped{$file}{$ch}{'values <= 0 (no log)'}++;
				next;
			    } else {
				if ($take_log) {
				    $val = &log_base(2,$val);
				}
			    }

			    # get flags for this channel
			    my @flags = ();
			    if (defined $data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
				@flags = &get_flags($data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
			    }
			
			    # and also flag from both channels
			    if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
				push @flags, (&get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
			    }

		    
			    # skip flagged spots
			    my $skip = 0;
			    if (%skip_flags) {
				
				foreach (@flags) {
				    if (defined $skip_flags{$_}) {
					$skip = 1;
					last;
				    }
				}			
			    }
			    
			    if ($skip) {
				$skipped{$file}{$ch}{'flagged'}++;
				next;
			    }

			    if ($use_ratios) {
				if ($ch eq 'CH1') {
				    $ch1{$mrow}{$mcol}{$row}{$col} = $val;
				} else {
				    unless (defined $ch1{$mrow}{$mcol}{$row}{$col}) {
					# value missing - no ratio possible!
					next;
				    }
				    my $val1 = $ch1{$mrow}{$mcol}{$row}{$col};
				    if ($header =~ /_L/
					or
					$take_log) {
					$val = $val - $val1;
				    } else {
					if ($val1 == 0) {
					    # no ratio possible!
					    next;
					}
					$val = $val / $val1;
				    }
				}
			    }

			    unless ($use_ratios 
				    and
				    $ch eq 'CH1') {				    
				if ($val < $min_val) {
				    $min_val = $val;
				}
				
				if ($val > $max_val) {
				    $max_val = $val;
				}
			    }
			    
			    # write value to file
			    unless ($use_ratios 
				    and
				    $ch eq 'CH1') {
				print OUT "$val\n";

				# count number of dots
				$dots{$file}{$ch}++;
			    }
			}
		    }
		}
	    }
	    unless ($use_ratios 
		    and
		    $ch eq 'CH1') {
		close OUT;
	    }
	}
    }
       
    # create boxplot
    
    my $R_commands_file = "$tmp_dir/R_cmd.$action.$action_num.$action_step.$PID";
    my $file_tmp = basename $files[0];
    $file_tmp =~ s/\//_/g;
    $file_tmp =~ s/ /_/g;
    my $R_out = "$tmp_dir/R_out.$file_tmp.$action.$action_num.$action_step.$PID";
    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";
    print RH "postscript(\"$R_out\",horizontal=FALSE)\n";

    foreach my $i (1..$file_counter) {
	print RH "f$i <- read.table(\"$order{$i}{'DATA'}\")\n";
    }

#    if ($work_labels[0] eq '__NO_LABEL__') {
#	# combine both channels per file into one ratio channel:
#	foreach my $i (1..$file_counter) {
#	    my $h = $file_counter * 2 - 1;
#	    my $k = $h + 1;
#	    print RH "f$h = $f$h
    # draw boxplot
    print RH "boxplot(";
    foreach my $i (1..$file_counter) {
	print RH "f$i\$V1,";
    }

    print RH " col = \"grey\", pch = \".\")\n";

    # draw labels
    print RH "axis(1,1:$file_counter,c('".(join "','", @plot_labels)."'))\n";
    my @ab_positions = ();
    if (@plot_labels > @work_labels) {
	foreach my $i (1..(int $#plot_labels)) {
	    push @ab_positions, $i*@work_labels + 0.5;
	}
	# draw separation line between files
	print RH "abline(v=c(".(join ',', @ab_positions)."), col = \"grey\")\n";
    }    

    print RH "par(srt = 90)\n";

    if ($use_ratios) {
	push @plot_labels, 'add';
    }
    for (my $i = 1; $i < @plot_labels; $i += @work_labels) {
	my $file = $order{$i}{'NAME'};
	my $x = $i - 0.4;
	my $y = &median($min_val,$max_val);
	$y = int(&median($y,$max_val));
	print RH "text($x,$y,'$file', col=\"blue\")\n";
    }
    print RH "par(srt = 0)\n";

    print RH "title(main = \"Box Plots for '$header'\", line=-1)\n";

    print RH "dev.off()\n";
    close RH;

    my $image = &get_R_output($R_commands_file, $R_out, $action_num, '');
    
    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	foreach my $i (1..$file_counter) {
	    unlink $order{$i}{'DATA'}
	}
    }

    my $info = "<TABLE BORDER=2><TR><TH>File</TH><TH>CH</TH><TH>Category</TH><TH># spots</TH></TR>\n";

    my %done = ();
    for my $i (1..$file_counter) {
	my $file = $order{$i}{'FULL_NAME'};
	
	if (defined $done{$file}) {
	    next;
	} else {
	    $done{$file}++;
	}
	
	my $name = $order{$i}{'NAME'};
	my $rowspan = 0;
	foreach (keys %{$data{$file}{'__DATA__'}}) {
	    next if ($_ eq '__NO_LABEL__');
	    $rowspan++;
	}
	
	my %extra = ();
	if (defined $skipped{$file}) {
	    foreach my $ch (keys %{$skipped{$file}}) {
		next if ($ch eq '__NO_LABEL__');
		$extra{$ch} = scalar (keys %{$skipped{$file}{$ch}});
		$rowspan += $extra{$ch};
	    }
	}
	
	$info .= "<TR><TD ROWSPAN=$rowspan>$name</TD>\n";
	my $row = 1;
	foreach my $ch (sort keys %{$data{$file}{'__DATA__'}}) {
	    next if ($ch eq '__NO_LABEL__');
	    if (defined $extra{$ch}) {
		if ($row > 1) {
		    $info .= "<TR>";
		} else {
		    $row++;
		}
		$info .= "<TD ROWSPAN=".($extra{$ch} + 1);
	    } else {
		$info .= "<TD";
	    }
	    my $ch_out = $ch;
	    if ($ch eq 'CH1') {
		if (defined $file_specs{$file}{'CH1_LABEL'}) {
		    $ch_out = $file_specs{$file}{'CH1_LABEL'};
		}
	    } else {
		if (defined $file_specs{$file}{'CH2_LABEL'}) {
		    $ch_out = $file_specs{$file}{'CH2_LABEL'};
		}
	    }
	    $info .= ">$ch_out</TD><TD>plotted</TD><TD ALIGN=RIGHT>$dots{$file}{$ch}</TD></TR>\n";
	    if (defined $skipped{$file}{$ch}) {
		foreach my $skip (keys %{$skipped{$file}{$ch}}) {
		    $info .= "<TR><TD>$skip</TD><TD ALIGN=RIGHT>$skipped{$file}{$ch}{$skip}</TD></TR>\n";
		}
	    }
	}
    }
    $info .= "</TABLE>\n";
    
    return $image, $info;
} 


sub signal_boxplotting_printtip {
    my $action_num = shift;
    my $action = shift;
    my $header = shift;
    my $labels = shift;
    my @labels = split /,/, $labels;
    my $combined = shift;
    my $ratio = shift;
    my @files = @_;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    # check for specific settings:
    my $use_colours = &get_user_settings($var_def_default{'boxplot_pt_colour'},'boxplot_pt_colour',$force_global_opt,$action_num,'__ALL__');
    $use_colours = $use_colours eq 'yes' ? 1 : 0;
    my @colours = ('orange','green','blue','cyan','magenta','yellow','grey','black');
    my @bp_colours = ();

    my $take_log = &get_user_settings($var_def_default{'boxplot_pt_log'},'boxplot_pt_log',$force_global_opt,$action_num,'__ALL__');
    $take_log = $take_log eq 'yes' ? 1 : 0;

    if ($header =~ /_L/) {
	$take_log = 0;
    }
    
    my $orig_header = $header;
    my @ratio_ch = ();
    if ($ratio) {
	if ($header =~ /_R/
	    or
	    $header =~ /$MERGED_RATIOS_TERM/) {
	    $ratio = '';
	} else {
	    
	    # make sure we have the _L ending if log is required:
	    if ($take_log) {
		$header .= '_L';
	    }

	    # add the _R ending for Ratio:
	    $header .= '_R';  # WARNING - this mightn't work for _M headers!
	    foreach my $ch (sort keys %{$data{$files[0]}{'__DATA__'}}) {
		next if ($ch eq '__NO_LABEL__');
		push @ratio_ch, $ch;
	    }
	    # take note of available columns for later
	    
	    foreach my $file (@files) {
		unless (defined $data_column_available{$file}{'__NO_LABEL__'}{$header}) {
		    $data_column_available{$file}{'__NO_LABEL__'}{$header} = 1;
		    push @{$data_column_available{$file}{'__NO_LABEL__'}{'__ORDER__'}}, $header;
		}
		$data_done{$file}{'__NO_LABEL__'}{$header}++;
	    }
	}
    }


    # if channels other than '__NO_LABEL__' are specified
    # and the header contains _R (for ratio), bring it down
    # to the separate channels:
    unless (grep /__NO_LABEL__/, @labels) {
	if ($header =~ /_R/) {
	    my $tmp_header = $header;
	    $tmp_header =~ s/_R//;
	    my $replace = 1;
	    foreach my $label (@labels) {
		unless (defined $data_column_available{$files[0]}{$label}{$tmp_header}) {
		    $replace = 0;
		    last;
		}
	    }
	    if ($replace) {
		$header = $tmp_header;
	    }
	}
    }
			

    $last_column_used = $header;


    my @flags = &get_user_settings($var_def_default{'boxplot_pt_skip_flags'},'boxplot_pt_skip_flags',$force_global_opt,$action_num,'__ALL__');
    my %skip_flags = ();
    %skip_flags = &get_skip_flags(@flags);

    my %values = ();
    my ($mrow,$mcol,$row,$col);
    my %skipped = ();
    my %flags = ();
    my $no_flag = 0;
    my %dots = ();

    
    my %skip_coord = ();

    # write data for each file/printtip/label into a separate file
    my @data_files = ();
    my $file_counter = 0;
    my $min_val = 99999;
    my $max_val = -99999;
    my %order = ();
    my @plot_labels = ();
    my @labels_out = ();
    my %labels_done = ();
    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@files);

    my @other_labels = ();
    if ($combined) {
	@other_labels = @labels;
	@labels = ($labels[0]);
    }

    # go through each field
    foreach my $file (sort @files) {
	foreach $mrow (1..$MROW) {
	    foreach $mcol (1..$MCOL) {

		# determine colour for this grid
		my $colour_counter = ($mrow - 1) % (scalar @colours);
		my $bp_colour = $colours[$colour_counter];
#		warn "#KH: ".(scalar @colours)." % $mrow = $colour_counter, colour: $bp_colour";
		
		foreach my $ch (@labels) {

		    # open file
		    $file_counter++;
		    my $data_file = "data_f${file_counter}_${ch}_$header.$action.$action.num.$action_step.${PID}_$mrow,${mcol}.tmp";
		    $data_file =~ s/ /_/g;
		    $data_file =~ s/-/_/g;
		    $data_file = "$tmp_dir/$data_file";

		    open(OUT, ">$data_file")
			or return "Error: Can't write to file '$data_file': $!\n";
		    my $output_counter = 0;
		    push @data_files, $data_file;
		    push @bp_colours, $bp_colour;

		    $order{$file_counter}{'CHANNEL'} = $ch;
		    $order{$file_counter}{'DATA'} = $data_file;
		    $order{$file_counter}{'FULL_NAME'} = $file;
		    if ($ch eq 'CH1') {
			if (defined $file_specs{$file}{'CH1_LABEL'}) {
			    $order{$file_counter}{'CHANNEL'} = $file_specs{$file}{'CH1_LABEL'};
			}
		    } else {
			if (defined $file_specs{$file}{'CH2_LABEL'}) {
			    $order{$file_counter}{'CHANNEL'} = $file_specs{$file}{'CH2_LABEL'};
			}
		    }
		    unless ($labels_done{$order{$file_counter}{'CHANNEL'}}) {
			push @labels_out, $order{$file_counter}{'CHANNEL'};
			$labels_done{$order{$file_counter}{'CHANNEL'}}++;
		    }
		    $order{$file_counter}{'NAME'} = "$mrow,$mcol";
		    my $plot_label = "$mrow,$mcol";
		    if (@labels > 1) {
			$plot_label .= ", $order{$file_counter}{'CHANNEL'}";
		    }
		    push @plot_labels, $plot_label;

		    unless ($combined) {
			@other_labels = ($ch);
		    }

		    foreach $row (1..$ROW) {
			foreach $col (1..$COL) {
			    
			    foreach my $ch (@other_labels) {
				# retrieve value
				my $val = '';
					
				if ($ratio) {

				    # initialize new output:
				    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$header}{$mrow}{$mcol}{$row}{$col} = '';

				    my ($val1,$val2);
				    if (defined $data{$file}{'__DATA__'}{'CH1'}{$orig_header}{$mrow}{$mcol}{$row}{$col}
					and
					defined $data{$file}{'__DATA__'}{'CH1'}{$orig_header}{$mrow}{$mcol}{$row}{$col} =~ /\d/) {
					$val1 = $data{$file}{'__DATA__'}{'CH1'}{$orig_header}{$mrow}{$mcol}{$row}{$col};
				    }
				    if (defined $data{$file}{'__DATA__'}{'CH2'}{$orig_header}{$mrow}{$mcol}{$row}{$col}
					and
					defined $data{$file}{'__DATA__'}{'CH2'}{$orig_header}{$mrow}{$mcol}{$row}{$col} =~ /\d/) {
					$val2 = $data{$file}{'__DATA__'}{'CH2'}{$orig_header}{$mrow}{$mcol}{$row}{$col};
				    }
				    unless ($val1 =~ /\d/
					    and
					    $val2 =~ /\d/) {
					    $skipped{"$mrow,$mcol"}{$ch}{'at least one channel undefined'}++;
					next;
				    }
				    if ($take_log) {
					unless ($val1 > 0
						and
						$val2 > 0) {
					    $skipped{"$mrow,$mcol"}{$ch}{'values <= 0 (no log)'}++;
					    next;
					} else {
					    $val1 = &log_base(2,$val1);
					    $val2 = &log_base(2,$val2);
					}
				    }

				    # take ratio:
				    if ($take_log
					or
					$header =~ /_L/) {
					$val = $val2 - $val1;
				    } else {
					if ($val1 == 0) {
					    
					    $skipped{"$mrow,$mcol"}{$ch}{'denominator equals zero (no ratio)'}++;
					    next;
					}
					$val = $val2 / $val1;
				    }
				} else {
				    if (defined $data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col}) {
					$val = $data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col};
				    }
				}
				
				# skip unless proper value
				unless ($val =~ /\d/) {
				    $skipped{"$mrow,$mcol"}{$ch}{'values not a number'}++;
				    next;
				}
				
				unless ($ratio) {
				    # log-transfer for ratios has been done already
				    if ($take_log) {
					if (($val <= 0
					     or 
					     $val eq '')
					    ) {
					    # set flag
					    $skipped{"$mrow,$mcol"}{$ch}{'values <= 0 (no log)'}++;
					    next;
					}
					$val = &log_base(2,$val);
				    }
				}
				
				# get flags for this channel
				my @flags = ();
				if (defined $data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
				    @flags = &get_flags($data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
				}
				# if this channel is '__NO_LABEL__', check the others as well
				if ($ch eq '__NO_LABEL__') {
				    foreach my $ch ('CH1', 'CH2') {
					if (defined $data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
					    push @flags, (&get_flags($data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
					}
				    }
				} else {				
				    # and also flag from both channels
				    if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
					push @flags, (&get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
				    }
				}
				
				
				# skip flagged spots
				my $skip = 0;
				if (%skip_flags) {
				    
				    foreach (@flags) {
					if (defined $skip_flags{$_}) {
					    $skip = 1;
					    last;
					}
				    }			
				}
				
				if ($skip) {
				    $skipped{"$mrow,$mcol"}{$ch}{'flagged'}++;
				    next;
				}
				
				# write value to file
				print OUT "$val\n";
				$output_counter++;
				
				if ($val < $min_val) {
				    $min_val = $val;
				}
				
				if ($val > $max_val) {
				    $max_val = $val;
				}
				
				if ($ratio) {
				    # store new output:
#				    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$header}{$mrow}{$mcol}{$row}{$col} = $val;
				}

				# count number of dots
				$dots{"$mrow,$mcol"}{$ch}++;
			    }
			}
		    }
		    unless ($output_counter) {
			# no spot plotted - this causes problems for the output
			# temporary solution: insert one value:
			print OUT "1\n";
			# TODO: leave printtip group empty but avoid error in R code!
		    }
		    close OUT;
		    unless (defined $dots{"$mrow,$mcol"}{$ch}) {
			$order{$file_counter}{'SKIP'}++;
		    }
		}
	    }
	}
    }

    if ($combined) {	
	@labels_out = ();
	foreach my $ch (@other_labels) {
	    if ($ch eq 'CH1') {
		if (defined $file_specs{$files[0]}{'CH1_LABEL'}) {
		    push @labels_out, $file_specs{$files[0]}{'CH1_LABEL'};
		}
	    } else {
		if (defined $file_specs{$files[0]}{'CH2_LABEL'}) {
		    push @labels_out, $file_specs{$files[0]}{'CH2_LABEL'};
		}
	    }  
	}
	@labels = @other_labels;
    }

    # create printtip boxplot
    
    my $R_commands_file = "$tmp_dir/R_cmd.$action.$action_num.$action_step.$PID";
    my $file_tmp = basename $files[0];
    $file_tmp =~ s/\//_/g;
    $file_tmp =~ s/ /_/g;
    my $R_out = "$tmp_dir/R_out.$file_tmp.$action.$action_num.$labels.$action_step.$PID";
    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";
    print RH "postscript(\"$R_out\",horizontal=FALSE, width = ".($IMAGE_HEIGHT/72).", height = ".($IMAGE_WIDTH/72).", paper = \"special\")\n";

    foreach my $i (1..$file_counter) {
	print RH "f$i <- read.table(\"$order{$i}{'DATA'}\")\n";
    }
    # draw boxplot
    if ($use_colours) {
	$use_colours = 'c("'.(join '","', @bp_colours).'")';
    } else {
	$use_colours = '"grey"';
    }
    print RH "boxplot(";
    foreach my $i (1..$file_counter) {
	next if (defined $order{$file_counter}{'SKIP'});
	print RH "f$i\$V1,";
    }

    print RH " col = $use_colours, pch = \".\")\n";

    print RH "abline(0,0, col = \"grey\")\n";

    # draw labels
    print RH "axis(1,1:$file_counter,c('".(join "','", @plot_labels)."'),las=3)\n";

    if ($combined eq ''
	and
	@labels_out > 1) {
	$combined = ', together';
    }

    if ($take_log) {
	$take_log = ', log-transferred';
    }
    print RH "title(main = \"PrintTip Box Plots for '$header', ".(join ',', @labels_out)."$combined$take_log\", line=-1)\n";

    print RH "dev.off()\n";
    close RH;

    my $image = &get_R_output($R_commands_file, $R_out, $action_num,$files[0],'','landscape');
    
    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	foreach my $i (1..$file_counter) {
	    unlink $order{$i}{'DATA'}
	}
    }

    my $info = '';
    # calculate how many spots were plotted in a grid without skips:
    my $plot_sum = 0;
    my @full_grids = ();
    foreach my $grid (keys %dots) {
	next if (defined $skipped{$grid});
	foreach my $label (keys %{$dots{$grid}}) {
	    $plot_sum += $dots{$grid}{$label};
	}
	last;
    }
    foreach my $grid (keys %dots) {
	next if (defined $skipped{$grid});
	push @full_grids, $grid;
    }
    
    my $full_grids = 'each grid';
    if (%skipped) {
#	$full_grids = 'the following grids: '.(join ', ', @full_grids);
	$full_grids = 'each but the following grids';
    }

    my $add_info_table = &get_user_settings($var_def_default{'boxplot_pt_add_info'},'boxplot_pt_add_info',$force_global_opt,$action_num,'__ALL__');
    $add_info_table = $add_info_table eq 'yes' ? 1 : 0;

    if ($add_info_table) {
	$info = "For $full_grids <B>$plot_sum</B> spots were plotted.<BR>\n";
	if ($plot_sum == 0) {
	    $info = "Overview of plotted and skipped spots:<BR>\n";
	}
	if (%skipped) {
	    $info .= "<TABLE BORDER=2><TR><TH>Grid</TH><TH>CH</TH><TH>Category</TH><TH># spots</TH></TR>\n";
	}
	
	my %done = ();
	for my $i (1..$file_counter) {
	    next if (defined $order{$file_counter}{'SKIP'});
	    my $file = $order{$i}{'FULL_NAME'};
	    
	    my $name = $order{$i}{'NAME'};
	    if (defined $done{$name}) {
		next;
	    } else {
		$done{$name}++;
	    }
	    
	    my $rowspan = scalar @labels;
#	if (@other_labels) {	    
#	    $rowspan = scalar @other_labels;
#	}
	    
	    my %extra = ();
	    if (defined $skipped{$name}) {
		foreach my $ch (keys %{$skipped{$name}}) {
		    $extra{$ch} = scalar (keys %{$skipped{$name}{$ch}});
		    $rowspan += $extra{$ch};
		}
	    } else {
		next;
	    }
	    
	    $info .= "<TR><TD ROWSPAN=$rowspan>$name</TD>\n";
	    my $row = 1;
	    foreach my $ch (@labels) {
		if (defined $extra{$ch}) {
		    if ($row > 1) {
			$info .= "<TR>";
		    } else {
			$row++;
		    }
		    $info .= "<TD ROWSPAN=".($extra{$ch} + 1);
		} else {
		    $info .= "<TD";
		}
		my $ch_out = $ch;
		if ($ch eq 'CH1') {
		    if (defined $file_specs{$file}{'CH1_LABEL'}) {
			$ch_out = $file_specs{$file}{'CH1_LABEL'};
		    }
		} else {
		    if (defined $file_specs{$file}{'CH2_LABEL'}) {
			$ch_out = $file_specs{$file}{'CH2_LABEL'};
		    }
		}
		unless ($dots{$name}{$ch}) {
		    $dots{$name}{$ch} = 0;
		}
		$info .= ">$ch_out</TD><TD>plotted</TD><TD ALIGN=RIGHT>$dots{$name}{$ch}</TD></TR>\n";
		if (defined $skipped{$name}{$ch}) {
		    foreach my $skip (keys %{$skipped{$name}{$ch}}) {
			$info .= "<TR><TD>$skip</TD><TD ALIGN=RIGHT>$skipped{$name}{$ch}{$skip}</TD></TR>\n";
		    }
		}
	    }
	}
	$info .= "</TABLE>\n" if (%skipped);
    }

    return $image, $info;
} 

sub get_flags {
    my $flag = shift || '';
    my $add_zero = shift || '';
    my @out = ();

    # treat no number ('') as no flag (0)
    unless ($flag =~ /\d/) {
	$flag = 0;
    }
    
    unless ($add_zero) {
	return () unless ($flag > 0);
    }

    if ($add_zero
	and
	$flag == 0
	or
	$flag eq '') {
	return (0);
    }
    
    if ($flag > $highest_flag_num) {
	my $count = 0;
	foreach (@flag_names) {
	    if (/list(\d+)/) {
		my $count2 = $1;
		if ($count2 > $count) {
		    $count = $count2;
		}
	    }
	}
	my $flag_name = '';
	while ($highest_flag_num < $flag) {
	    $count++;
	    $flag_name = "list$count";
	    $highest_flag_num *= 2;
	    $flag_values{$flag_name} = $highest_flag_num;
	    $flag_values_num{$highest_flag_num} = $flag_name;
	    push @all_flags, $flag_name;
	    unshift @all_flags_num, $highest_flag_num;
	}
	$used_flags{$flag_name}++;
    }

    foreach my $num (@all_flags_num) {
	if ($num > $flag) {
	    next;
	}
	$flag = int($flag - $num);
	push @out, $num;
    }

    return @out;
}


sub MA_plotting {
    my $action_num = shift;
    my $file = shift;
    my $action = shift;

    if ($refine_parameters) {
	return '', '', 'Paramter Refinement!';
    }

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $header = '';
    foreach my $column (@{$item_data_column{$action}}) {

	my $occ = 0;
	foreach my $label (sort {$b cmp $a} keys %{$data{$file}{'__DATA__'}}) {
	    if ($data_column_available{$file}{$label}{$column}) {
		$header = $column;
		if ($label eq '__NO_LABEL__') {
		    $occ = 2;
		} else {
		    $occ++;
		}
	    }
	}
	if ($occ == 2) {
	    last;
	} else {
	    $header = '';
	}
    }
    # check for specific settings:
    $header = &get_user_settings($header,'ma_plot_data_column',$force_global_opt,$action_num,$file,'__ALL__');

    my $logtransferred = 0;
    if ($header =~ /_L/) {
	$logtransferred = 1;
    }

    $last_column_used = $header;

    # get labels for channel 1 (
    my $ch_up = $var_def_default{'ma_plot_ch1'};
    
    # check for user specific settings
    $ch_up = &get_user_settings($ch_up,'ma_plot_ch1',$force_global_opt,$action_num,$file,'__ALL__');

    # collect labels
    my %labels = ();

    foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	
	# skip columns that are not label specific ('spot labels')
	next if ($label eq '__NO_LABEL__');
	$labels{$label}++;
    }

    unless ($labels{$ch_up}) {
	warn "# ERROR: Label '$ch_up' is not defined for ".(basename $file)."!";
	die "# ERROR: Label '$ch_up' is not defined for ".(basename $file)."!\n";
    }

    my $ch_down = '';
    foreach (keys %labels) {
	next unless (/\w/);
	unless ($ch_up eq $_) {
	    $ch_down = $_;
	    last;
	}
    }
    unless ($ch_down) {
	die "# ERROR: No channel 2 defined for ".(basename $file)." (Ch1 = '$ch_up', labels available: '".(join "','", keys %labels)."')!\n";
    }

    my ($ch_up_out, $ch_down_out) = ($ch_up,$ch_down);
    if ($ch_up eq 'CH1') {
	if (defined $file_specs{$file}{'CH1_LABEL'}) {
	    $ch_up_out = $file_specs{$file}{'CH1_LABEL'};
	}
	if (defined $file_specs{$file}{'CH2_LABEL'}) {
	    $ch_down_out = $file_specs{$file}{'CH2_LABEL'};
	}
    } else {
	if (defined $file_specs{$file}{'CH2_LABEL'}) {
	    $ch_up_out = $file_specs{$file}{'CH2_LABEL'};
	}
	if (defined $file_specs{$file}{'CH1_LABEL'}) {
	    $ch_down_out = $file_specs{$file}{'CH1_LABEL'};
	}
    }

    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'ma_skip_flags'},'ma_skip_flags',$force_global_opt,$action_num,$file,'__ALL__');
    %skip_flags = &get_skip_flags(@flags);

    my $flag_no_ratio = &get_user_settings($var_def_default{'ma_flag_no_ratio'},'ma_flag_no_ratio',$force_global_opt,$action_num,$file,'__ALL__');

    my %flag2col_tmp = ();
    foreach my $flag (0, (sort { $a <=> $b } keys %flag_values_num)) {
 
	my $default = '';
	if (defined $pref_flag2col{$flag}) {
	    $default = $pref_flag2col{$flag};
	}
	# check for user specific settings
	my $flag_colour = &get_user_settings($default,'MA_flag_colour'.$flag,$force_global_opt,$action_num,$file,'__ALL__');
	$flag2col_tmp{$flag} = $flag_colour if ($flag_colour =~ /\w/);	    
    }

    # compute MA columns
    my $out_header1 = $header.'_I';
    my $out_header2 = $header.'_R';
    my $out_header3 = '';
    if ($logtransferred == 0) {
	$out_header1 = $header.'_L_I';
	$out_header2 = $header.'_L_R';
	$out_header3 = $header.'_L';
    }

    my %values = ();
    my ($mrow,$mcol,$row,$col);
    my %skipped = ();
    my %flags = ();
    my $no_flag = 0;
    my $dots = 0;
    my %quantiles = ();
    my @quantiles_x = ();
    my @quantiles_y = ();

    if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{'__FLAGS__'}) {
	%{$data{$file}{'__DATA__'}{'__NO_LABEL__'}{'__FLAGS__'}} = ();
    }

    # calculate ratio and store in array
    foreach $mrow (keys %{$data{$file}{'__DATA__'}{$ch_up}{$header}}) {
	foreach $mcol (keys %{$data{$file}{'__DATA__'}{$ch_up}{$header}{$mrow}}) {
	    foreach $row (keys %{$data{$file}{'__DATA__'}{$ch_up}{$header}{$mrow}{$mcol}}) {
		foreach $col (keys %{$data{$file}{'__DATA__'}{$ch_up}{$header}{$mrow}{$mcol}{$row}}) {

		    # retrieve values from both channels
		    my ($val_up,$val_down,$ratio,$product);
		    $val_up = '';
		    $val_down = '';

		    # initialize output field:
		    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header1}{$mrow}{$mcol}{$row}{$col} = '' 
			unless (
				defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header1}{$mrow}{$mcol}{$row}{$col}
				and
				$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header1}{$mrow}{$mcol}{$row}{$col} =~ /\d/);
		    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header2}{$mrow}{$mcol}{$row}{$col} = '' 
			unless (
				defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header2}{$mrow}{$mcol}{$row}{$col}
				and
				$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header2}{$mrow}{$mcol}{$row}{$col} =~ /\d/);
		    

		    if (defined $data{$file}{'__DATA__'}{$ch_up}{$header}{$mrow}{$mcol}{$row}{$col}) {
			$val_up = $data{$file}{'__DATA__'}{$ch_up}{$header}{$mrow}{$mcol}{$row}{$col};
		    }
		    if (defined $data{$file}{'__DATA__'}{$ch_down}{$header}{$mrow}{$mcol}{$row}{$col}) {
			$val_down = $data{$file}{'__DATA__'}{$ch_down}{$header}{$mrow}{$mcol}{$row}{$col};
		    }

		    # insert log'ed values
		    if ($logtransferred < 1) {
			
			$data{$file}{'__DATA__'}{$ch_up}{$out_header3}{$mrow}{$mcol}{$row}{$col} = '';
			$data{$file}{'__DATA__'}{$ch_down}{$out_header3}{$mrow}{$mcol}{$row}{$col} = '';
			
			if ($val_up =~ /\d/
			    and
			    $val_up > 0) {
			    my $loged = &log_base(2,$val_up);
			    $data{$file}{'__DATA__'}{$ch_up}{$out_header3}{$mrow}{$mcol}{$row}{$col} = $loged;
			}
			if ($val_down =~ /\d/
			    and
			    $val_down > 0) {
			    my $loged = &log_base(2,$val_down);
			    $data{$file}{'__DATA__'}{$ch_down}{$out_header3}{$mrow}{$mcol}{$row}{$col} = $loged;
			}
		    }

		    # skip unless both values set
		    unless ($val_up =~ /\d/
			    and
			    $val_down =~ /\d/) {
			$skipped{'at least one of the values is undefined'}++;
			next;
		    }

		    if (($val_down <= 0
			 or
			 $val_up <= 0)
			and
			$logtransferred < 1) {
			# set flag
			if ($flag_no_ratio) {
			    $data{$file}{'__DATA__'}{$ch_up}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'no_ratio'};
			    $data{$file}{'__DATA__'}{$ch_down}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'no_ratio'};
			}
			$skipped{'at least one of the values is <= 0'}++;
			next;
		    } else {
			if ($logtransferred) {
			    $ratio = $val_up - $val_down;
			} else {			    
			    if ($val_down <= 0
				or
				$val_up <= 0) {
				if ($flag_no_ratio) {
				    $data{$file}{'__DATA__'}{$ch_up}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'no_ratio'};
				    $data{$file}{'__DATA__'}{$ch_down}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'no_ratio'};
				}
				$skipped{'at least one of the values is <= 0'}++;
				next;
			    }
			    $ratio = $val_up / $val_down;
			    $ratio = &log_base(2,$ratio);
			}
		    }


		    if ($logtransferred) {
			$product = $val_up + $val_down;
		    } else {
			$product = $val_up * $val_down;
			if ($product <= 0) {
			    if ($flag_no_ratio) {
				$data{$file}{'__DATA__'}{$ch_up}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'no_ratio'};
				$data{$file}{'__DATA__'}{$ch_down}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'no_ratio'};
			    }
			    $skipped{'the product is < 0 (no log)'}++;
			    next;
			}			    
#			$product = &log_base(10,$product);
			$product = &log_base(2,$product);
		    }

		    $product *= 0.5;

		    # get flags for first channel:
		    my @flags = ();
		    if (defined $data{$file}{'__DATA__'}{$ch_up}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			@flags = &get_flags($data{$file}{'__DATA__'}{$ch_up}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
		    }

		    # get flags for second channel as well
		    if (defined $data{$file}{'__DATA__'}{$ch_down}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			push @flags, (&get_flags($data{$file}{'__DATA__'}{$ch_down}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
		    }

		    # and flags for common channel
		    if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			push @flags, (&get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
		    }

		    # skip flagged spots
		    my $skip = 0;
		    if (%skip_flags) {

			foreach (@flags) {
			    if (defined $skip_flags{$_}) {
				$skip = 1;
				last;
			    }
			}			
		    }
			
		    if ($skip) {
			$skipped{'at least one of the values is flagged'}++;
			next;
		    }

		    # if we have flags, put the data into a special bin
		    if (@flags) {
			# take the lowest-numbered flag
			my $flag = (sort { $a <=> $b } @flags)[0];
			$data{$file}{'__DATA__'}{'__NO_LABEL__'}{'__FLAGS__'}{$flag}{$out_header1}{$mrow}{$mcol}{$row}{$col} = $product;
			$data{$file}{'__DATA__'}{'__NO_LABEL__'}{'__FLAGS__'}{$flag}{$out_header2}{$mrow}{$mcol}{$row}{$col} = $ratio;
			$flags{$flag}++;
		    } else {
			$no_flag++;
		    }


		    # store ratio in specified destination
		    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header1}{$mrow}{$mcol}{$row}{$col} = $product unless ($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header1}{$mrow}{$mcol}{$row}{$col} =~ /\d/);
		    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header2}{$mrow}{$mcol}{$row}{$col} = $ratio unless ($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header2}{$mrow}{$mcol}{$row}{$col} =~ /\d/);
		    push @{$quantiles{$product}}, $ratio;
		    push @quantiles_x, $product;
		    # count number of dots
		    $dots++;
		}
	    }
	}
    }

    # calculate quantile coordinates
    my ($quantile1_x, $quantile1_y, $quantile2_x, $quantile2_y);
    foreach (sort { $a <=> $b } keys %quantiles) {
	push @quantiles_y, @{$quantiles{$_}};
    }
#    @quantile_x = (sort { $a <=> $b } @quantile_x);
    (undef,$quantile1_x, undef,$quantile2_x,undef) = &quantile(0.25, \@quantiles_x);
    ($quantile1_y,$quantile2_y) = &quantile_average([0.25, 0.75], \@quantiles_y);
  #  ($quantile1_y, $quantile2_y) = &quantile_y([0.25, 0.75, \@quantile_y);

    my $data_tmp_file = &write_data_to_file('','',$file, '__NO_LABEL__', 'tab', "R_$action_step.$action_num", $out_header1, $out_header2);
    
#    foreach $mrow (keys %{$data{$file}{'__DATA__'}{$ch_up}{$header}}) {
#	foreach $mcol (keys %{$data{$file}{'__DATA__'}{$ch_up}{$header}{$mrow}}) {
#	    foreach $row (keys %{$data{$file}{'__DATA__'}{$ch_up}{$header}{$mrow}{$mcol}}) {
#		foreach $col (keys %{$data{$file}{'__DATA__'}{$ch_up}{$header}{$mrow}{$mcol}{$row}}) {
#		    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header1}{$mrow}{$mcol}{$row}{$col} = '';
#		    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header2}{$mrow}{$mcol}{$row}{$col} = '';
#		}
#	    }
#	}
#    }
		    
    my $R_commands_file = "$tmp_dir/R_cmd.$action.$action_num.$action_step.$PID";
    my $file_tmp = basename $file;
    $file_tmp =~ s/\//_/g;
    $file_tmp =~ s/ /_/g;
    my $R_out = "$tmp_dir/R_out.$file_tmp.$action.$action_num.$action_step.$PID";
    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";
    print RH "postscript(\"$R_out\",horizontal=FALSE, width = ".($IMAGE_HEIGHT/72).", height = ".($IMAGE_WIDTH/72).", paper = \"special\")\n";
#    print RH "png(\"$R_out\",width = $IMAGE_HEIGHT, height = $IMAGE_WIDTH, pointsize = 12, bg = \"white\")\n";
    # unfortunately the direct png output doesn't work because it needs an X11-device

    # read in data
    print RH "fm <- read.table(\"$data_tmp_file\")\n";
    print RH "attach(fm)\n";

    # draw histogram
    print RH "plot(V1,V2, xlab = \"log2($ch_down_out*$ch_up_out)\", ylab = \"log2($ch_up_out/$ch_down_out)\", pch= \".\", main = \"\", col = \"$flag2col_tmp{0}\")\n";

    # add red crosses for 25% and 75% quartile:
#    print RH "q1 <- quantile(V1,c(0.25,0.75))\n";
#    print RH "q2 <- quantile(V2,c(0.25,0.75))\n";
#    print RH "points(q1[1],q2[1], pch='X', col='red')\n";
#    print RH "points(q1[2],q2[2], pch='X', col='red')\n";
    print RH "points($quantile1_x,$quantile1_y, pch='X', col='red')\n";
    print RH "points($quantile2_x,$quantile2_y, pch='X', col='red')\n";
    print RH "lx = min(V1)\n";
    print RH "ly = max(V2) - (max(V2) - min(V2))/10\n";	
    print RH "legend(lx,ly, 'quartiles', pch='X', col='red')\n";
    
    print RH "abline(0,0, col = \"red\")\n";
    print RH "abline(1,0, col = \"grey\")\n";
    print RH "abline(-1,0, col = \"grey\")\n";
#    print RH "abline(32,1, col = \"grey\")\n";
#    print RH "abline(-32,-1, col = \"grey\")\n";
#    print RH "abline(32,-1, col = \"grey\")\n";
#    print RH "abline(-32,1, col = \"grey\")\n";

    # if we have flags, print them in different colours on top of the plot
    if (%flags) {
	my $counter = 2;
	foreach my $flag (sort { $a <=> $b } keys %flags) {
	    # get flag colours
	    if (defined $flag2col_tmp{$flag}) {
		unless (defined $pref_flag2col{$flag}) {
		    $pref_flag2col{$flag} = $flag2col_tmp{$flag};
		    $pref_flag2col{$flag2col_tmp{$flag}} = $flag;
		}
	    } else {
		if (defined $pref_flag2col{$flag}) {
		    $flag2col_tmp{$flag} = $pref_flag2col{$flag};
		} else {
		    foreach my $colour (@pref_flag_colours) {
			unless (defined $pref_flag2col{$colour}) {
			    $flag2col_tmp{$flag} = $colour;
			    $pref_flag2col{$colour} = $flag;
			    $pref_flag2col{$flag} = $colour;
			    last;
			}
		    }
		}
	    }

	    $counter++;
	    print RH "points(V$counter,V";
	    $counter++;
	    print RH "$counter, col = \"$flag2col_tmp{$flag}\", pch = \"*\")\n";
	}
    }
    
    print RH "title(main = \"MA plot for ".(basename $file)."\", line=-1)\n";

    # add legend
#    print RH "legend(min(V1), max(V2), 'boundaries', lty = 1, col='grey')\n";

    print RH "dev.off()\n";
    close RH;


    my $image = &get_R_output($R_commands_file, $R_out, $action_num, $file,'','landscape');
    
    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	unlink $data_tmp_file;
    }
#    $image = "$tmp_dir_html/".(basename $R_out);

    my $info = "<B>$dots</B> dots plotted<BR>\n";
    if (%flags) {
	$info .=  "<BR>
     <TABLE>
      <TR>
       <TH>flag</TH><TH>#</TH><TH>colour</TH>
      </TR>\n";

	my $colour_field = '';
	$colour_field = "<SELECT name=\"action$action_num${file}MA_flag_colour0\" size=\"1\">";
	foreach my $col ('black', @pref_flag_colours) {
	    $colour_field .= "<OPTION".($col eq 'black' ? ' SELECTED' : '').">$col</OPTION>\n";
	}
	$colour_field .= "\n</select>\n";

	$info .= "<TR><TD>unflagged</TD><TD><B>$no_flag</B></TD><TD ALIGN=RIGHT>$colour_field</TD></TR>\n";
	foreach my $flag (sort { $a <=> $b } keys %flags) {
	    # make field to choose colour:
	    my $colour_field = '';
	    $colour_field = "<SELECT name=\"action$action_num${file}MA_flag_colour$flag\" size=\"1\">";
	    foreach my $col (@pref_flag_colours) {
		my $selected = '';
		if (defined $flag2col_tmp{$flag}
		    and
		    $flag2col_tmp{$flag} eq $col) {
		    $selected = ' SELECTED';
		}
		$colour_field .= "<OPTION$selected>$col</OPTION>\n";
	    }
	    $colour_field .= "\n</select>\n";
	    
	    $info.=  "
      <TR>
       <TD ALIGN=RIGHT>$flag: $flag_values_num{$flag}</TD><TD ALIGN=RIGHT><B>$flags{$flag}</B></TD><TD ALIGN=RIGHT>$colour_field</TD>
      </TR>";
	}
	$info .=  "
     </TABLE>
";
    }
	    
    foreach (keys %skipped) {
	$info .= "<B>$skipped{$_}</B> spots were skipped because<BR>$_!<BR>\n";
    }

    # take note of available columns for later

    unless (defined $data_column_available{$file}{'__NO_LABEL__'}{$out_header1}) {
	$data_column_available{$file}{'__NO_LABEL__'}{$out_header1} = 1;
	push @{$data_column_available{$file}{'__NO_LABEL__'}{'__ORDER__'}}, $out_header1;
    }
    $data_done{$file}{'__NO_LABEL__'}{$out_header1}++;

    unless (defined $data_column_available{$file}{'__NO_LABEL__'}{$out_header2}) {
	$data_column_available{$file}{'__NO_LABEL__'}{$out_header2} = 1;
	push @{$data_column_available{$file}{'__NO_LABEL__'}{'__ORDER__'}}, $out_header2;
    }
    $data_done{$file}{'__NO_LABEL__'}{$out_header2}++;
    
    if ($logtransferred < 1) {
#	$data_column_available{$file}{$ch_up}{$out_header3} = 1;
#	push @{$data_column_available{$file}{$ch_up}{'__ORDER__'}}, $out_header3;
#	$data_done{$file}{$ch_up}{$out_header3}++;
#	$data_column_available{$file}{$ch_down}{$out_header3} = 1;
#	push @{$data_column_available{$file}{$ch_down}{'__ORDER__'}}, $out_header3;
#	$data_done{$file}{$ch_down}{$out_header3}++;
    }


    return $ch_up, $image, $info;
}

sub R_histogram_plotting {
    my $action_num = shift;
    my $file = shift;
    my $label = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $header = '';
    foreach my $column (@{$item_data_column{$action}}) {
	if ($data_column_available{$file}{$label}{$column}) {
	    $header = $column;
	    last;
	}
    }
    # check for specific settings:
    $header = &get_user_settings($header,'hist_data_column',$force_global_opt,$action_num,$file,$label);

    my $data_tmp_file=&get_data_for_R($action_num, $file, $label, $header);
    
    my $R_commands_file = "$tmp_dir/R_cmd.$action.$action_num.$action_step.$PID";
    my $file_tmp = basename $file;
    $file_tmp =~ s/\//_/g;
    $file_tmp =~ s/ /_/g;
    my $R_out = "$tmp_dir/R_out.$file_tmp.$label.$action.$action_num.$action_step.$PID";
    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";
    print RH "postscript(\"$R_out\",horizontal=FALSE)\n";

    # read in data
    print RH "fm <- read.table(\"$data_tmp_file\")\n";
    print RH "attach(fm)\n";

    # draw histogram
    print RH "hist(V1, col=\"gray\", labels = TRUE, xlab = \"$header\", main = \"\")\n";
    print RH "title(main = \"Histogram of $label intensities in ".(basename $file)."\", line=-1)\n";
    print RH "dev.off()\n";
    close RH;

    my $image = &get_R_output($R_commands_file, $R_out, $action_num, $file);
    
    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	unlink $data_tmp_file;
    }

    return $image;
}

sub R_quintile_histogram {

    my $action_num = shift;
    my $file = shift;
    my $label = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $header = '';

    my $plot_ratio = 0;
    if ($label eq '__NO_LABEL__') {
	$plot_ratio = 1;
    }

    foreach my $column (@{$item_data_column{$action}}) {
	if ($plot_ratio) {
	    $column .= '_R' unless ($column =~ /_R/);
	}
	warn "# column: $column";
	if ($data_column_available{$file}{$label}{$column}) {
	    $header = $column;
	    last;
	}
    }

#    warn "# header: $header";

    # check for specific settings:
    $header = &get_user_settings($header,'histo_data_column',$force_global_opt, $action_num,$file,$label);

#    warn "# header: $header";

#    my $data_tmp_file=&get_data_for_R($action_num, $file, $label, $header);
    
    my $min_val = &get_user_settings($var_def_default{'histo_take_min_val'},'histo_take_min_val',$force_global_opt,$action_num,$file,$label);

    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'histo_skip_flags'},'histo_skip_flags',$force_global_opt,$action_num,$file,'__ALL__');
    %skip_flags = &get_skip_flags(@flags);

    my $take_log = &get_user_settings($var_def_default{'histo_take_log'},'histo_take_log',$force_global_opt,$action_num,$file,'__ALL__');
    $take_log = $take_log eq 'yes' ? 1 : 0;

    if ($header =~ /_L/) {
	$take_log = 0;
    }

    my $add_quint = &get_user_settings($var_def_default{'histo_add_quintile'},'histo_add_quintile',$force_global_opt,$action_num,$file,'__ALL__');
    $add_quint = $add_quint eq 'yes' ? 1 : 0;

    # read data:
    my $data_tmp_file = "data.".(basename $file)."$action.$action_num.$action_step.$label.$PID";
    $data_tmp_file =~ s/ /_/g;
    $data_tmp_file = "$tmp_dir/$data_tmp_file";

    open (DATA, ">$data_tmp_file") 
	or die "Can't write data to file: $!\n";
    
    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);

    my @data = ();

    foreach my $mrow (1..$MROW) {
	foreach my $mcol (1..$MCOL) {
	    foreach my $row (1..$ROW) {
		COL:foreach my $col (1..$COL) {

		    # skip this unless data is available:
		    next unless (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col});
		    next unless ($data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col} =~ /\d/);

		    my @flags = ();
		    if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}
			or
			defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			    @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}, 'add_zero');
			}
			if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			    push @flags, &get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}, 'add_zero');
			}
			foreach (@flags) {
			    if (defined $skip_flags{$_}) {
				next COL;
			    }
			}
		    }

		    my $val = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
		    
		    if ($take_log) {
			if ($val <= 0) {
			    next COL;
			}
		    }
		    
		    push @data, $val;
		}
	      }
	}
    }

#    if ($min_val =~ /\d/) {
#	# adjust the values
#	my @tmp = @data;
#	my $min = (sort { $a <=> $b } @tmp)[0];
#	foreach (@data) {
#	    $_ = $_ - $min + $min_val;
#	}
#    }
    
    foreach my $val (@data) {
	# print data
	print DATA "$val\n";
    }
    close DATA;

    if ($plot_ratio) {
	$label = 'ratios';
    }

    my $output_label = $header;
    if ($take_log) {
	$output_label = 'log2 of '.$header;
    }
    
    if ($min_val =~ /\d/) {
	$output_label .= ", shifted to $min_val";
    }

    my $R_commands_file = "$tmp_dir/R_cmd.$action.$action_num.$action_step.$label.$PID";
    my $file_tmp = basename $file;
    $file_tmp =~ s/\//_/g;
    $file_tmp =~ s/ /_/g;
    my $R_out = "$tmp_dir/R_out.$file_tmp.$label.$action.$action_num.$action_step.$PID";
    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";
    print RH "postscript(\"$R_out\",horizontal=FALSE)\n";
    
    # define function 
    print RH "
## This function takes a vector of intensities and plots the
## resulting values in  a histogram.

## Created by Dana from UBC Stats Dept.
## modified by Karsten Hokamp, Brinkman Lab

hist.qc=function(x, m.lab=\"Use the label!\", x.lab=\"$output_label\"){

        # gives quick look at distribution

";
    if ($min_val =~ /\d/) {
	print RH"
        # if x has negative values, correct it

        x=x-min(x)+$min_val;
";
    }
    
    if ($take_log) {
	print RH "
        # work on log2 scale
        x=log2(x);
";
    }

    print RH "
        x.quins=quantile(x, probs=seq(0., 1.0, 0.2))
";
    if ($take_log
	or
	$header =~ /_L/) {
	print RH "
        x.diffs=c(x.quins,log2(65535))-c(0, x.quins)
";
    } else {
	print RH "
        x.diffs=c(x.quins,65535)-c(min(x), x.quins)
";
    }

    print RH "
#        tmpX=hist(x, col=\"gray\", labels = TRUE, xlab=x.lab, main=\"\")
        tmpX=hist(x, col=\"gray\", labels = FALSE, xlab=x.lab, main=\"\", breaks=100)
        title(main=m.lab)

";

    if ($add_quint) {
	print RH "
        legend((2/3)*(max(x)-min(x)), quantile(tmpX\$counts,0.9), \"Quintile lines\", lty=2, col=\"blue\")
";
    }

	print RH "

        # David, \"2\" and \"3\" are chosen pretty arbitrarily here.  This could stand some tuning
        # if we actually find the function being used
        x.message=ifelse(sum(x.diffs[2:6]<2)>=3,\"Your x-data isn't spread very well\", \"The spread of your x-data is OK\")

        # KH: the above test means: check for the difference between adjacent quintiles. If three or more of the
        #     differences are smaller than '2', then issue a warning.
";

    if ($add_quint) {
	print RH "
        abline(v=x.quins, lty=2, col=\"blue\")
";
    }

	print RH "
}
";


    # read in data
    print RH "fm <- read.table(\"$data_tmp_file\")\n";
    print RH "attach(fm)\n";
    
    my $slide = basename $file;
    $slide =~ s/.txt//g;
   
    print RH "new.label = paste\(\"Histogram for $output_label values for slide\",\"$slide\",\"channel\:\",\"$label\")\n";

    print RH "hist.qc(V1, m.lab=\"$slide, channel\:$label\", x.lab=\"$output_label\")";

    print RH "\ndev.off()\n";
    close RH;

    my $image = &get_R_output($R_commands_file, $R_out, $action_num, $file);
    
    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	unlink $data_tmp_file;
    }

    return $image;
}

#######################################################################
##
## Subroutine for Dana's QA Array Sub-grid plotting function
##
## Takes values for the Ag or Au channel and plots the average
## intensities for each sub-grid on a relative scale.
## Adapted from the CGI script for the web interface.
## 
########################################################################

sub R_ArrayPlot {

    my $action_num = shift;
    my $file = shift;
    my $label = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $header = '';
    foreach my $column (@{$item_data_column{$action}}) {
	if ($data_column_available{$file}{$label}{$column}) {
	    $header = $column;
	    last;
	}
    }
    # check for specific settings:
    $header = &get_user_settings($header,'data_column',$force_global_opt,$action_num,$file,$label);

    ## Get the data file from the sub-routine that handles collection of data from the ARV file.

    my $data_tmp_file=&get_data_for_R($action_num, $file, $label, $header);

    my $R_commands_file = "$tmp_dir/R_cmd.$action.$action_num.$action_step.$PID";

    my $file_tmp = basename $file;
    $file_tmp =~ s/\//_/g;
    $file_tmp =~ s/ /_/g;
    my $R_out = "$tmp_dir/R_out.$file_tmp.$label.$action.$action_num.$action_step.$PID";
    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";
    print RH "postscript(\"$R_out\",horizontal=FALSE)\n";

     # read in data
    print RH "fm <- read.table(\"$data_tmp_file\")\n";
    print RH "attach(fm)\n";
 
    my $slide = basename $file;
    $slide =~ s/.txt//g;
   
	print RH "new.label = paste\(\"Subgrid plot for slide\",\"$slide\",\"channel\:\",\"$label\")\n";
    
    # Do Array Plot
    print RH "plot.w.col(grid.positions, V1, label = new.label)\n";
    
    print RH "dev.off()\n";
    close RH;

    ## Pass R script over to function that handles incorporation of graphical data.

    my $image = &get_R_output($R_commands_file, $R_out, $action_num, $file);
    
    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	unlink $data_tmp_file;
    }

    return "<IMG SRC=\"$image\">";
}

## Retrieves data from named ARV file & creates a temporary file that R can use to get the intensity values

sub get_data_for_R {

    my $action_num = shift;
    my $file = shift;
    my $label = shift;
    my @headers = @_;

    &read_data($file,@headers);

    # data needs to be stored in temp file for R
    my $data_tmp_file = '';

    # check if data has been written out already  
    my $format = 'space';
    $data_tmp_file = &write_data_to_file('','',$file,$label,$format,"R_$action_step.$action_num",@headers);

    return $data_tmp_file;

}


## Executes given R script, collects output, converts it to .png from .ps, and incorporates this into the final HTML output

sub get_R_output {

    my $R_Script = shift;
    my $R_out = shift;
    my $action_num = shift;
    my $file = shift;
    my $end_signal = shift || '';

    my $orientation = shift || '';

    my $no_image = shift || '';

    # change to tmp directory
    # so that we are allowed to write Rout file
    my $cwd = cwd;
    chdir "$tmp_dir"
	or die "Can't change into $tmp_dir: $!";

    # careful: R BATCH (1.7.1) can't deal with spaces in filenames, 
    # make sure that at least the basename is space-free!
    my $cmd_base = basename $R_Script;
    my $cmd_dir  = dirname $R_Script;
    if ($cmd_base =~ / /) {
	$cmd_base =~ s/ /_/g;
	my $new_name = "$cmd_dir/$cmd_base";
	rename $R_Script, $new_name
	    or die "Can't rename '$R_Script' to '$new_name': $!";
	$R_Script = $new_name;
    }

    # execute R commands
    my $job = "$R_executable BATCH --slave '$R_Script'";
    if ($end_signal) {
	$job .= ' &';
    }

    system "$job";
    if ($end_signal) {
	until (-s $end_signal) {
	    my $failure = 0;
	    if (-s "$R_Script.Rout") {
		$failure = `grep -c 'Execution halted' '$R_Script.Rout'`;
		chomp $failure;
	    }
	    if ($failure > 0) {
		system "echo '$failure' > '$end_signal'";
		$failure = `cat '$R_Script.Rout'`;
		# change back into previous directory
		chdir "$cwd";
		return "Execution of R-job failed:\n<BR><PRE>$failure</PRE><BR>\n";
	    } else {
		sleep 5;
	    }
	}
	if ($clean_up_files) {
	    unlink $end_signal;
	}
    }

    # check for failure:
    my $failure = 0;
    if (-s "$R_Script.Rout") {
	$failure = `grep -c 'Execution halted' '$R_Script.Rout'`;
	chomp $failure;
    }
    if ($failure > 0) {
	$failure = `cat '$R_Script.Rout'`;
	if ($clean_up_files) {
	    unlink "$R_Script.Rout";
	}
	# change back into previous directory
	chdir "$cwd";
	return "Execution of R-job failed:\n<BR><PRE>$failure</PRE><BR>\n";
    }
    
    if ($clean_up_files) {
	unlink "$R_Script.Rout";
    }

    # change back into previous directory
    chdir "$cwd";

    my @R_out = ();

    if (ref($R_out) eq 'ARRAY') {
	@R_out = @$R_out;
    } else {
	@R_out = ($R_out);
    }

    my $image = '';
    my @images = ();

    foreach my $out_file (@R_out) {
	# make sure output has been produced:
	unless (-s $out_file) {
	    $image = "<B>WARNING: no output produced!</B><BR>\n";
	    
	    # get warning messages:
	    warn "#KH: checking for '$R_Script.Rout'";
	    if (-r "$R_Script.Rout") {
		$image .= "Output from R:<BR>\n<PRE>\n";
		open (IN, "$R_Script.Rout");	    
		while (<IN>) {
		    $image .= $_;
		}
		close IN;
		$image .= "</PRE><BR>\n";
	    }

	    if (ref($R_out) eq 'ARRAY') {
		push @images, $image;
		next;
	    } else {
		return $image;
	    }
	}
	
	unless ($no_image) {
	    # convert to png:
	    my $png_file = "$out_file.png";
	    my $WIDTH = $IMAGE_WIDTH;
	    my $HEIGHT = $IMAGE_HEIGHT;
	    if ($orientation eq 'landscape') {
		$WIDTH = $IMAGE_HEIGHT;
		$HEIGHT = $IMAGE_WIDTH;
	    }
	    $job = "gs -dNOPAUSE -sDEVICE=png256 -g${WIDTH}x${HEIGHT} -dBATCH -q -sOutputFile='$png_file' '$out_file'";
	    system "$job";
	    
	    # compile HTML code for integration of image
	    $image = basename("$png_file");

	    push @images, $image;
	}
    }

    if (ref($R_out) eq 'ARRAY') {
	return @images;
    } else {
	return $image;
    }
}

sub get_required_columns {

    # call each action subroutine with the '__columns__' argument
    # this should return the columns required for its execution

    my %cols = ();
    my @cols = ();

    for my $i (1..100) {

	# skip undefined actions
        next unless ($cgi->param("action$i"));
	
	# call the sub routine named after the action
	my $action = $cgi->param("action$i");
	
	unless ($function{$action}) {
	    warn "No action '$action' defined!";
	    next;
	}

	eval { @cols = $function{$action}->('__columns__') };
	    
	# warn if action didn't execute properly
	if ($@) {
	    warn "Function $action couldn't be executed: $@";
	}

	foreach (@cols) {
	    $cols{$_}++;
	}
    }

    return keys %cols;
}  

    
sub change_headers {
    my $file = shift;
    open (IN, $file)
	or return "Can't read file '$file': $!\n";
    $_ = <IN>;
    chomp;
    my @tmp = split /\t/, $_;
    my %tmp = ();
    my @out = ();
    foreach my $column (@tmp) {
	if (defined $tmp{$column}) {
	    $tmp{$column}++;
	    $column .= " ($tmp{$column})";
	} else {
	    $tmp{$column}++;
	}
	push @out, $column;
    }
    my $new_header = (join "\t", @out)."\n";
    my $new_file = $file.'.new';
    warn "# printing '$new_header' to $new_file";
    unless (open (OUT, ">$new_file")) {
	return "Can't write file '$new_file': $!\n";
    }
    print OUT $new_header;
    close OUT;
    system "tail +2 $file >> $new_file";
    system "mv $new_file $file";
    return 'ok';
}
    
    
sub check_file_format {
    
    # get the filename
    my $file = shift;
    my $type = shift || '';
    my $counter = 0;
    my $format = 'ArrayVision';
    my $comment_lines = 0;
    my $header_lines = 0;    
    my @header = ();
    my $spot_id = '';
    my $ch1fg = ''; 
    my $ch1bg = '';
    my $ch2fg = '';
    my $ch2bg = '';
    my @flags = ();
    my $size = (stat $file)[7];
    $file_specs{$file}{'SIZE'} = $size;
    
    # see if it has been checked before
    # this is just a quick check for the list of example files
    # if the size of these hasn't changed a thourough check is not necessary
    if (defined $cgi->param("example_$file")) {
	my $check_size = $cgi->param("example_$file");
	if ($check_size == $size) {
	    return 'ok' unless ($type =~ /force/);
	}
    }

    if ($to_dir) {
	$file =~ s/$from_dir/$to_dir/;
    }

    # open the file:
    open (IN, $file)
	or return "Can't read file '$file': $!\n";

    if ($type =~ /reporter_list/) {

	$file_specs{$file}{'TYPE'} = 'reporter_list';

	my $data = '';

	# the first line might specify which columns are used for identifying a spot

	$_ = <IN>;

	# remove comments
	if (/#/) {
	    s/#.*//;
	}

	until (/\w/) {
	    $_ = <IN>;
	    # remove comments
	    if (/#/) {
		s/#.*//;
	    }
	    last if (eof(IN));
	}


	unless (/\w/) {
	    return "No data in file '$file'!\n";
	}
	
	my @headers = ();

	if (/Probe ID\t/
	    or
	    /ROW\t/
	    or
	    /Spot labels\t/
	    or
	    /$FG_SIGNAL_TERM\t/
	    or
	    /Bkgd\t/
	    or
	    /Flag\t/
	    ){
	    # looks like output from the pipeline
	    # specify keyword for recognition of column headers:
	    $_ = 'COLUMNS:'.$_;
	}

	unless (/COLUMNS:/) {
	    @headers = split /\t/, $_;
	    unless (@headers == 4) {
		# unless grid positions are provided
		# the default is to match with Spot labels
		# which means that only one column is expected
		warn "#KH: headers = @headers";
		unless (grep /ROW\t/, @headers) {
		    @headers = ($SPOT_ID_TERM);
		}
	    }
	    $data++;
	    
	} else {
	    
	    # remove tag
	    s/COLUMNS:\s*//;
	    
	    # get rid of trailing white space
	    while (/\s$/) {
		chop;
	    }
	    @headers = split /\t/, $_;
	}

	
	# make sure that we have some data in the file
	while (<IN>) {
	    # remove comments
	    if (/^#/) {
		s/#.*//;
	    }
	    if (/\w/) {
		$data++;
		my @tmp = split /\t/, $_;
		unless (@tmp == @headers) {
		    return "Mismatch in number of columns ('".(join "\t", @tmp)."' ne '".(join "\t", @headers)."')\n";
		}
	    }
	}

	close IN;

	if ($data) {
	    return 'ok';
	} else {
	    return "No data in file '$file'!\n";
	}
	$file_specs{$file}{'FINISHED'} = 1;
    }

    # the first line might let us guess which format the file is in:
    $_ = <IN>;

    # get rid of trailing white space
    chomp;
    s/\cM$//;

    if (/^probe_id\t/) {
	# this is probably an annotation file from ProbeLynx
	$file_specs{$file}{'TYPE'} = 'annotation_file';
	$web_link_annotation_file = $file;
	push @annotation_files, $web_link_annotation_file;
	return 'ok', 'probelynx_annotation';
    }


    # if we got here, then we are not dealing with a reporter list or annotation file but with a data file:
    $file_specs{$file}{'TYPE'} = 'data_file';

    if (/^# output from .* version .* at/) {
	$comment_lines++;

	# file comes from this program

	$format = 'ArrayPipe';
	$header_lines = 1;
	my %new_set = ();

	# skip through comments
	while (<IN>) {
	    if (/^\#/) {
		$comment_lines++;
		if (/# (.*?) = (.*)\s+$/) {
		    # some previously made specifications can be read from the file:
		    my $key = $1;
		    my $val = $2;
		    next if ($key eq 'START_LINE');
		    if ($val =~ /\w/) {
			my @cgi_val = ();
			if ($val =~ /\|/
			    or
			    $key eq 'FLAG'
			    or
			    $key eq 'ADD') {
			    # here we have an array of values
			    my @val = split / \| /, $val;
			    @{$file_specs{$file}{$key}} = @val;
			    @cgi_val = @val;
			} else {
			    $file_specs{$file}{$key} = $val;
			    @cgi_val = ($val);
			}
			my $name = "${key}_$file";
#			warn " #KH adding ${key}_$file: @cgi_val";
			# setting the cgi-parameter should only be done if no other value was submitted
			unless (defined $cgi->param($name)) {
			    $cgi->param(-name=>$name,
					-values=>[@cgi_val]);
			    $new_set{$name}++;
			}
		    }
		}
	    } else {
#		$counter++;
		# we have reached the header
		# overwrite file_specs with proper choice of column
		chomp;
		my @headers = split /\t/, $_;
		foreach my $column (@headers) {
		    if ($column eq $SPOT_ID_TERM) {
			my $type = 'SPOT_ID';
			$file_specs{$file}{$type} = $column;
			unless (defined $cgi->param("${type}_$file")) {
			    $cgi->param(-name=>"${type}_$file",
					-values=>[($column)]);
			} elsif (defined $new_set{"${type}_$file"}) {
			    $cgi->param(-name=>"${type}_$file",
					-values=>[($column)]);
			}			    
			next;
		    }
		    my $type = '';		    
		    my $ch = '';
		    my @tmp = split / /, $column;
		    if (
			$tmp[-1] eq $FG_SIGNAL_TERM 
			) {
			$type = 'FG';
		    } elsif (
			$tmp[-1] eq $BG_SIGNAL_TERM
			) {
			$type = 'BG';
		    }
		    if ($tmp[0] eq 'Cy3'
			or
			$tmp[0] eq 'Ag'
			or
			$tmp[0] eq 'CH1'
			) {
			$ch = 'CH1';
		    } else {
			$ch = 'CH2';
		    }
		    if ($type
			and
			$ch) {
			$file_specs{$file}{$ch.$type} = $column;
#			warn " #KH adding $ch${type}_file: $column";
			unless (defined $cgi->param("$ch${type}_$file")) {
			    $cgi->param(-name=>"$ch${type}_$file",
					-values=>[($column)]);
			} elsif (defined $new_set{"$ch${type}_$file"}) {
			    $cgi->param(-name=>"$ch${type}_$file",
					-values=>[($column)]);
			}
			
		    }
		}
		last;
	    }
	}
    } elsif (/^MROW\tMCOL\tROW\tCOL\t$SPOT_ID_TERM\t.*\t\w* ?$FLAG_CHANNEL_TERM/) {

	$format = 'ArrayPipe';
	$header_lines = 1;
	
    } elsif (/HEADER\tSPOT/) {
	# this could be a ScanAlyze file
	$format = 'ScanAlyze';
	$header_lines = 1;
	$spot_id = 'SPOT';
	$ch1fg = 'CH1I';
	$ch1bg = 'CH1B';
	$ch2fg = 'CH2I';
	$ch2bg = 'CH2B';

	while (<IN>) {
	    if (/^REMARK/) {
		$comment_lines++;
	    } else {
		$counter++;
		last;
	    }
	}
	$file_specs{$file}{'COMMENT_KEY'} = '^REMARK';
    } elsif (/^ATF/) {
	# this could be a GenePix file
	$format = 'GenePix';
	$header_lines = 1;
	$spot_id = '"ID"';
	$ch1fg = '"F635 Median"';
	$ch1bg = '"B635 Median"';
	$ch2fg = '"F532 Median"';
	$ch2bg = '"B532 Median"';
	push @{$file_specs{$file}{'FLAG'}}, '"Flags"';
	$file_specs{$file}{'START_LINE'} = '"Supplier';

	while (<IN>) {
	    $comment_lines++;
	    if (/^$file_specs{$file}{'START_LINE'}/) {
		<IN>;
		last;
	    }
	}
#	$file_specs{$file}{'COMMENT_KEY'} = '^REMARK';
    } elsif (/Begin Header/) {
	$comment_lines++;	    
	# this could be an ImaGene file
	$format = 'ImaGene';
	$header_lines = 1;
	$spot_id = 'Gene ID';
	$ch1fg = 'Signal Median';
	$ch1bg = 'Background Median';
	$ch2fg = 'Signal Median';
	$ch2bg = 'Background Median';
	$file_specs{$file}{'START_LINE'} = 'Begin Raw Data';

	while (<IN>) {
	    $comment_lines++;
	    if (/^Begin Raw Data/) {
		<IN>;
		last;
	    }
	}
	$file_specs{$file}{'COMMENT_KEY'} = '^End';
    } elsif ((/\tAg\t/
	 and
	 /\tAu\t/)
	or
	(/\t.*?_Ag60\t/
	 and
	 /\t.*?_Au80\t/)
	) {
	# probably an ArrayVision file
	$format = 'maybe ArrayFision';
	$header_lines = 2;
	push @header, $_;
	$_ = <IN>;
	push @header, $_;	
    } elsif (/^\# V\d.*User:.*Total number of Spots/i) {
	# this could be an TMEV file
	$comment_lines++;	    
	$format = 'TMEV';
	$header_lines = 1;
	$spot_id = 'UID';
	$ch1fg = 'IA';
	$ch1bg = 'BkgA';
	$ch2fg = 'IB';
	$ch2bg = 'BkgB';

	while (<IN>) {
	    if (/^\#/) {
		$comment_lines++;
		next;
	    }
	    chomp;
	    my @headers = split;
	    my $column_counter = 0;
	    my %headers = ();
	    foreach my $field (@headers) {
		$headers{$field} = $column_counter;
		$column_counter++;		    
	    }
	    
	    my $line_counter = 0;
	    my $mrow = 0;
	    my $mcol = 0;
	    my $row = 0;
	    my $col = 0;
	    unless (defined $cgi->param("MROW_$file")) {
		while (<IN>) {
		    chomp;
		    $line_counter++;
		    my @tmp = split;
		    if ($tmp[$headers{'MR'}] > $mrow) {
			$mrow = $tmp[$headers{'MR'}];
		    }
		    if ($tmp[$headers{'MC'}] > $mcol) {
			$mcol = $tmp[$headers{'MC'}];
		    }
		    if ($tmp[$headers{'R'}] > $row) {
			$row = $tmp[$headers{'R'}];
		    }
		    if ($tmp[$headers{'C'}] > $col) {
			$col = $tmp[$headers{'C'}];
		    }
		}

		if ($col * $row == $line_counter) {
		    # calculate col from the other numbers
		    $col /= $mcol;
		    $row /= $mrow;
		}

		# set parameter as if they had been supplied by the user
		unless (defined $cgi->param("MROW_$file")) {
		    $cgi->param(-name=>"MROW_$file",
				-value=>$mrow);
		}
		unless (defined $cgi->param("MCOL_$file")) {
		    $cgi->param(-name=>"MCOL_$file",
				-value=>$mcol);
		}
		unless (defined $cgi->param("ROW_$file")) {
		    $cgi->param(-name=>"ROW_$file",
				-value=>$row);
		}
		unless (defined $cgi->param("COL_$file")) {
		    $cgi->param(-name=>"COL_$file",
				-value=>$col);
		}
	    }
	    last;
	}
	
    } else {
	$format = 'unknown';

	my @tmp = ();
	foreach (split /\t/, $_) {
	    # remove leading and trailing white-space (DANGEROUS!)
	    s/^\s*//;
	    s/\s*$//;
	    s/  / /g;
	    push @tmp, $_;
	}

	if (@tmp == 1) {
	    close IN;
	    return 'First line does not contain tab-delimited headers!';
	}
	# check how many duplicate entries we have:
	my %tmp = ();
	foreach my $column (@tmp) {
	    if (defined $tmp{$column}
		and
		$column eq 'SPOT'
		) {
		$column = 'SPOT2';
	    }
	    $tmp{$column}++;
	}
	my $diff = @tmp - (keys %tmp);
	if ($diff) {
	    # duplicate headers - we might have an error here
	    unless ($diff >= (keys %tmp)) {
		close IN;
		# change headers so they are not duplicated anymore
#		warn " # WARNING: Found duplicate columns in header of $file!";
		my $rv = &change_headers($file);
		if ($rv eq 'ok') {
		    return 'Run again';
		} else {
		    return "Duplicate headers couldn't be changed ($rv)";
		}
	    }
	    $header_lines = 2;
	    push @header, $_;
	    $_ = <IN>;
	    push @header, $_;
	} else {
	    # probably only one header line
	    $header_lines = 1;

	    # if we have more data columns than text columns, this is possible a file without a header
	    my $data_count = 0;
	    my $text_count = 0;
	    foreach (@tmp) {
		if (/[a-z]/i) {
		    $text_count++;
		} else {
		    $data_count++;
		}
	    }
	    if ($data_count > 1) {
		close IN;
		return 'File doesn\'t seem to have a header line!';
	    }	    
	}
    }

    if ($header_lines == 2) {
	# second line contains the column headers
	$_ = $header[1];
	# make sure we have some of the expected columns
	if (/^Spot labels\t/
	    and
	    /\tMTM Dens - Levels\t/
#	    and
#	    /\tFlag\t/
		) {
	    $format = 'ArrayVision';
	} else {
	    # if we have more data columns than text columns, this is possibly a file without a header
	    my @tmp = split /\t/, $_;
	    my $data_count = 0;
	    my $text_count = 0;
	    foreach (@tmp) {
		if (/[a-z]/i) {
		    $text_count++;
		} else {
		    $data_count++;
		}
	    }
	    if ($data_count > 1) {
		close IN;
		return 'File doesn\'t seem to have a second header line!';
	    }
	    $format = 'unknown';	    
	}
    }

    while (<IN>) {
	# remove comments
	my $comment_key = '^#';
	if (defined $file_specs{$file}{'COMMENT_KEY'}) {
	    $comment_key = $file_specs{$file}{'COMMENT_KEY'};
	}
	if (/$comment_key/) {
	    $comment_lines++;
	    next;
	} 

	unless (/\w/) {
	    $comment_lines++;
	    next;
	}
	
	$counter++;
    }

    $file_specs{$file}{'LINES'} = $counter;
    $file_specs{$file}{'COMMENTS'} = $comment_lines;
    $file_specs{$file}{'FORMAT'} = $format;
    $file_specs{$file}{'HEADER'} = $header_lines;
    $file_specs{$file}{'ORGANIZATION'} = 'unknown';

    if ($spot_id =~ /\w/) {
	$file_specs{$file}{'SPOT_ID'} = $spot_id;
    }
    if ($ch1fg =~ /\w/) {
	$file_specs{$file}{'CH1FG'} = $ch1fg;
    }
    if ($ch1bg =~ /\w/) {
	$file_specs{$file}{'CH1BG'} = $ch1bg;
    }
    if ($ch2fg =~ /\w/) {
	$file_specs{$file}{'CH2FG'} = $ch2fg;
    }
    if ($ch2bg =~ /\w/) {
	$file_specs{$file}{'CH2BG'} = $ch2bg;
    }

    my ($mrow,$mcol,$row,$col,$alias) = ('?','?','?','?','?');
    if ($counter == 28704) {
	($mrow,$mcol,$row,$col) = (12,4,23,26);
    } elsif ($counter == 27648) {
	($mrow,$mcol,$row,$col) = (12,4,24,24);
    } elsif ($counter == 23232) {
	($mrow,$mcol,$row,$col) = (12,4,22,22);
    } elsif ($counter == 19200) {
	($mrow,$mcol,$row,$col) = (12,4,20,20);
    } elsif ($counter == 18432) {
	($mrow,$mcol,$row,$col) = (8,4,24,24);
    } elsif ($counter == 16896) {
	($mrow,$mcol,$row,$col) = (8,4,22,24);
    } elsif ($counter == 16416) {
	($mrow,$mcol,$row,$col) = (12,4,18,19);
    } elsif ($counter == 13056) {
	($mrow,$mcol,$row,$col) = (12,4,16,17);
    } elsif ($counter == 12288) {
	($mrow,$mcol,$row,$col) = (12,4,16,16);
    } elsif ($counter == 10752) {
	($mrow,$mcol,$row,$col) = (12,4,14,16);
    } elsif ($counter == 9216) {
	($mrow,$mcol,$row,$col) = (4,4,24,24);
    } elsif ($counter == 8832) {
	($mrow,$mcol,$row,$col) = (2,2,48,46);
    } elsif ($counter == 8448) {
	($mrow,$mcol,$row,$col) = (2,2,48,44);
    } elsif ($counter == 6528) {
	($mrow,$mcol,$row,$col) = (4,4,17,24);
    } elsif ($counter == 5472) {
	($mrow,$mcol,$row,$col) = (4,4,18,19);
    } elsif ($counter == 4608) {
	($mrow,$mcol,$row,$col) = (8,6,8,12);
    }

    if (defined $cgi->param("MROW_$file")
	and
	$cgi->param("MROW_$file") =~ /\d/) {
	$mrow = $cgi->param("MROW_$file");
    } elsif (defined $file_specs{$file}{'MROW'}) {
	$mrow = $file_specs{$file}{'MROW'};
    }
    if (defined $cgi->param("MCOL_$file")
	and
	$cgi->param("MCOL_$file") =~ /\d/) {
	$mcol = $cgi->param("MCOL_$file");
    } elsif (defined $file_specs{$file}{'MCOL'}) {
	$mcol = $file_specs{$file}{'MCOL'};
    }
    if (defined $cgi->param("ROW_$file")
	and
	$cgi->param("ROW_$file") =~ /\d/) {
	$row = $cgi->param("ROW_$file");
    } elsif (defined $file_specs{$file}{'ROW'}) {
	$row = $file_specs{$file}{'ROW'};
    }
    if (defined $cgi->param("COL_$file")
	and
	$cgi->param("COL_$file") =~ /\d/) {
	$col = $cgi->param("COL_$file");
    } elsif (defined $file_specs{$file}{'COL'}) {
	$col = $file_specs{$file}{'COL'};
    }

    if (defined $cgi->param("ALIAS_$file")
	and
	$cgi->param("ALIAS_$file") =~ /\w/) {
	$alias = $cgi->param("ALIAS_$file");
    } elsif (defined $file_specs{$file}{'ALIAS'}) {
	$alias = $file_specs{$file}{'ALIAS'};
    }
    $file_specs{$file}{'MROW'} = $mrow;
    $file_specs{$file}{'MCOL'} = $mcol;
    $file_specs{$file}{'ROW'} = $row;
    $file_specs{$file}{'COL'} = $col;
    $file_specs{$file}{'ALIAS'} = $alias;

#    my $required_lines = $MROW * $MCOL * $ROW * $COL + 2;

#    unless ($counter == $required_lines) {
#	close IN;
#	return "File only has $counter lines (instead of $MROW * $MCOL * $ROW * $COL + 2 = $required_lines)!";
#    }


    return 'ok', $format;
 
}


sub read_data {

    # get the filename
    my $file = shift;
    my $data_file = $file;
    if (defined $first_data_file{$file}) {
	$data_file = $first_data_file{$file};
    }

    my %spot_label_occurence = ();

    # get column headers
    # (if the first element is set to 'all_columns')
    # we are going to store all columns)

    # read in label and fg intensities by default

    unless (@_) {
	@_ = ($SPOT_ID_TERM,$FG_SIGNAL_TERM);
    }

    my %extra_column = ();
    my %column_usage = ();
     foreach my $key (keys %{$file_specs{$file}}) {
	next if ($key eq 'ADD');
	if ($key eq 'FLAG') {
	    foreach ( @{$file_specs{$file}{$key}} ) {
		$column_usage{$_}++;
	    } 
	} else {
	    $column_usage{$file_specs{$file}{$key}}++;
	}
    }

    if (defined $file_specs{$file}{'ADD'}) {
	foreach (@{$file_specs{$file}{'ADD'}}) {
	    if ($column_usage{$_}) {
		next;
	    }
	    push @_, $_;
	    $extra_column{$_}++;
	}
#	warn "#KH: pushing extra columns: @{$file_specs{$file}{'ADD'}}";
    }

    if ($file_specs{$file}{'FORMAT'} eq 'TMEV') {
	push @_, $SPOT_QUALITY_TERM;
    }

    my @columns = ();
    foreach my $header (@_) {
	my $found = 0;
	foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	    if (defined $data_done{$file}{$label}{$header}) {
		$found++;
	    }
	}
	unless ($found) {
	    push @columns, $header;
	}
    }

    unless (@columns) {
	# all headers have been read in already
	return 'ok';
    }
    
#    warn "#KH: columns: '@columns'";

    # counters for column headers:
    my @counters = ();

    my ($format_check,$file_format);
    # check file format if requested:
    if ($cgi->param('check_files')) {

	($format_check,$file_format) = &check_file_format($file);
	unless ($format_check eq 'ok') {
	    warn "# WARNING: file '$file' has invalid format: $format_check - skipping!";
	    $wrong_format_files{$file} = $format_check;
	    return 'invalid format';
	}
    }

    my $skip_counter = 0;

    if ($to_dir) {
	$file =~ s/$from_dir/$to_dir/;
    }

   # open the file:
    open (IN, $file)
	or return "Can't read file '$file': $!\n";

    # specify label names for each column:

    my %label = ();
    $_ = <IN>;  
    while (/^\#/) {
	last if (/^# output from .* version .* at/);
	$skip_counter++;
	$_ = <IN>;
    }

    # get rid of trailing white space
    chomp;
    s/\cM$//;

    my @tmp = split /\t/, $_;
    # extra measurement for trailing tab:
    while (/\t$/) {
	chop;
	push @tmp, '';
    } 
    my @tmp2 = ();
    foreach (@tmp) {
	# remove leading and trailing white-space (DANGEROUS!)
	s/^\s*//;
	s/\s*$//;
	s/  / /g;
	push @tmp2, $_;
    }
    @tmp2 = @tmp;

    my $pipeline_data = 0;


    # get channel label for each column:

    if (/^# output from .* version .* at/) {

	$pipeline_data = 1;

	# file comes from this program
	# make sure we have some data in it
	
	# skip comments
	while (/^\#/) {
#	    $skip_counter++;
	    $_ = <IN>;
	}

	# get rid of trailing white space
	while (/\s$/) {
	    chop;
	}

	# leave only label names in fields
	@tmp = ();
	foreach (split /\t/, $_) {
	    # remove leading and trailing white-space (DANGEROUS!)
	    s/^\s*//;
	    s/\s*$//;
	    s/  / /g;
	    push @tmp, $_;
	}

	foreach (@tmp) {
	    if (/^(Ag|Au) /) {
		s/^(Ag|Au).*/$1/;
	    } elsif (/^(Cy3|Cy5) /) {
		s/^(Cy3|Cy5).*/$1/;
	    } else {
		$_ = '';
	    }
	}
	# go back to start of file
	seek IN,0,0;
    } elsif ($file_specs{$file}{'FORMAT'} eq 'ImaGene') {
	# only one label per file!
	until (/$file_specs{$file}{'START_LINE'}/) {
	    $_ = <IN> ;
	}
	$_ = <IN>;

	my $default_label = 'CH1';
	if (defined $first_data_file{$file}) {
	    # this is the second part of a multi-part data set:
	    $default_label = 'CH2' unless ($channel_swaps{$first_data_file{$file}});
	} else {
	    # it's the first data file, default is set to 'CH1'   
	    if ($channel_swaps{$file}) {
		# if file swap was specified we assign CH2:
		$default_label = 'CH2';
	    }
	}

	# get rid of trailing white space
	chomp;
	s/\cM$//;

	@tmp = ();
	foreach (split /\t/, $_) {
	    # remove leading and trailing white-space (DANGEROUS!)
	    s/^\s*//;
	    s/\s*$//;
	    s/  / /g;
	push @tmp, $_;
	}

	foreach my $counter (0..$#tmp) {
	    my $header = $tmp[$counter];
	    my $label = $default_label;
	    if ($header eq $file_specs{$file}{'SPOT_ID'}) {
		$label = '__NO_LABEL__';
	    }
	    $label{$counter} = $label;
	    $labels{$file}{$label}++;
	}
	# go back to start of file
	seek IN,0,0;
	# and to position just before header
	until (/$file_specs{$file}{'START_LINE'}/) {
	    $_ = <IN> ;
	}
    } elsif ($file_specs{$file}{'FORMAT'} eq 'GenePix') {
	until (/$file_specs{$file}{'START_LINE'}/) {
	    $_ = <IN> ;
	}
	$_ = <IN>;

	# get rid of trailing white space
	chomp;
	s/\cM$//;
	@tmp = split /\t/, $_;

	foreach my $counter (0..$#tmp) {
	    my $header = $tmp[$counter];
	    my ($header2,$label) = &match_header($header,$file);
	    if ($extra_column{$header}) {
		$label = '__NO_LABEL__';
	    }
	    $label{$counter} = $label;
	    $labels{$file}{$label}++;
	}

	# go back to start of file
	seek IN,0,0;
	# and to position just before header
	until (/$file_specs{$file}{'START_LINE'}/) {
	    $_ = <IN> ;
	}
    } elsif ($file_specs{$file}{'HEADER'} == 2) {

	# presume ArrayVision by default:
	# the first line specifies which columns are gold and which are silver:

	my $label_keep = '';
	my $first_label = '';
	my %label_match = ();
	foreach my $counter (0..$#tmp) {
#	    my $label = $tmp[$counter] || '__NO_LABEL__';
	    my $label = $tmp[$counter];

	    if ($label =~ /\w/) {
		if ($first_label eq '') {
		    $first_label = $label;
		}

		$label_keep = $label_match{$label};

		if ($label eq $first_label) {
		    $label_match{$label} = 'CH1';
		    $label = 'CH1';
		} else {
		    $label_match{$label} = 'CH2';
		    $label = 'CH2';
		}
	    } else {
		if ($label_keep) {
		    # some VIDO columns have no label 
		    # but really should have the same 
		    # as the previous
		    $label = $label_keep;
		} else {
		    $label = '__NO_LABEL__';
		}
	    }
	    $label{$counter} = $label;
	    $labels{$file}{$label}++;
	}
    } else {
	foreach my $counter (0..$#tmp) {
	    my $header = $tmp[$counter];
	    my ($header2,$label) = &match_header($header,$file);
	    if ($extra_column{$header}) {
		$label = '__NO_LABEL__';
	    }
	    $label{$counter} = $label;
	    $labels{$file}{$label}++;
	}
	# go back to start of file
	seek IN,0,0;
	for (1..$skip_counter) {
	    $_ = <IN>;
	}
    }

    if ($channel_swaps{$file}
	and
	$file_specs{$file}{'FORMAT'} ne 'ImaGene') {
	my %swap = ();
	my @tmp = ();
	foreach (keys %{$labels{$file}}) {
	    next if (/__NO_LABEL__/);
	    next unless ($_);
	    push @tmp, $_;
	}

	# channel swapping of ImaGene files is taken care of above already
#	unless (@tmp > 1) {
#	    # we might be dealing with ImaGene files, that come in two parts
#	    # we have to set the second label manually:
#	    if ($tmp[0] eq 'CH1') {
#		$tmp[1] = 'CH2';
#	    } else {
#		$tmp[1] = 'CH1';
#	    }
#	}

	$swap{$tmp[0]} = $tmp[1];
	$swap{$tmp[1]} = $tmp[0];

	foreach (keys %label) {
	    next unless (defined $swap{$label{$_}});
	    $label{$_} = $swap{$label{$_}};
	}
    }
    
    # next line contains the column headers
    $_ = <IN>;

    # get rid of trailing white space
    while (/\s$/) {
	chop;
    }
    @tmp = ();
    foreach (split /\t/, $_) {
	# remove leading and trailing white-space (DANGEROUS!)
	s/^\s*//;
	s/\s*$//;
	s/  / /g;
	push @tmp, $_;
    }

    my @pipeline_data_columns = ();
    my $arraypipe_file = 0;
    if (/^# output from .* version .* at/) {
	# file comes from this program
	# make sure we have some data in it

	# skip comments
	while (/^\#/) {
	    $skip_counter++;
	    $_ = <IN>;
	}
	

	# get rid of trailing white space
	while (/\s$/) {
	    chop;
	}

	@tmp = split /\t/, $_;
	@columns = ();
	$counter = 0;
	my %columns = ();
	foreach my $col (@tmp) {
	    my $col2 = $col;
	    my $label = '__NO_LABEL__';
	    if ($col =~ /^(Ag|Cy3|CH1) (.*)/i) {
		$label = 'CH1';
		$col2 = $2;
	    } elsif ($col =~ /^(Cy5|Au|CH2) (.*)/i) {
		$label = 'CH2';
		$col2 = $2;
	    }
	    $columns{$col2}++;
	    $label{$counter} = $label;
	    $counter++;
	}
	@columns = keys %columns;
	$arraypipe_file = 1;
    }

    # use all columns if requested
    my $all_columns = 0;
    if (defined $columns[0]
	and
	$columns[0] eq 'all_columns') {
	$all_columns = 1;
	my %tmp = ();
	foreach (@tmp) {
	    $tmp{$_}++;
	}
	@columns = keys %tmp;
    }

    # unless we have markers/controls defined
    # determine them from the file
    unless (%control_labels) {	
	# make sure we have spot labels amongst the headers
	unless (grep /^$SPOT_ID_TERM$/, @columns) {
	    push @columns, $SPOT_ID_TERM;
	}
    }

    # drop the columns for which we have data already
    my @columns_tmp = ();
    foreach my $header (@columns) {
	my $found = 0;
	foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	    if (defined $data_done{$file}{$label}{$header}) {
		$found++;
	    }
	}
	unless ($found) {
	    push @columns_tmp, $header;
	}
    }

    @columns = @columns_tmp;

    unless (@columns) {
	# all headers have been read in already
	return 'ok';
    }

#    warn " #KH: reading in @columns";

    # headers hash
    my %headers = ();

    # collect positions for each header
    foreach my $counter (0..$#tmp) {

	my $header = $tmp[$counter];
	# remove leading and trailing white-space (DANGEROUS!)
	$header =~ s/^\s*//;
	$header =~ s/\s*$//;
	$header =~ s/  / /g;
	
	unless (defined $header) {
	    warn "# KH: No header at counter $counter in file $file!";
	    next;
	}
		
	my $label = '';
	$label = $label{$counter};
	unless (defined $label) {
	    warn "# KH: No label at counter $counter in file $file!";
	    next;
	}

	($header) = &match_header($header,$file);
	
	if ($arraypipe_file) {
	    # remove label names from fields	
	    if ($header =~ /^(Ag|Au|Cy3|Cy5|CH1|CH2) /i) {
		$header =~ s/^(Ag|Au|Cy3|Cy5|CH1|CH2) (.*)/$2/i;
	    }
	}
	
	# get position for requested columns
	# and associate with label
	foreach my $col (@columns) {

	    if ($header eq $col) {	
		if (defined $headers{$col}{$label}
		    and
		    keys %{$headers{$col}{$label}}
		    and
		    $header ne '') {
		    warn "# WARNING: header '$col' found at column ".((keys %{$headers{$col}{$label}})[0])." and $counter of $data_file!";
		    if ($header eq $SPOT_ID_TERM
			or
			$header eq 'Probe ID') {
			next;
		    }
		}
		$headers{$col}{$label}{$counter}++;
	    
		# take note of available columns for later
#		if ($arraypipe_file) {
#		    # remove label names from fields	
#		    if ($header =~ /^(Ag|Au|Cy3|Cy5|CH1|CH2) /i) {
#			$header =~ s/^(Ag|Au|Cy3|Cy5|CH1|CH2) (.*)/$2/i;
#		    }
#		}

		unless (defined $data_column_available{$data_file}{$label}{$header}) {
		    $data_column_available{$data_file}{$label}{$header} = 1;
		    push @{$data_column_available{$data_file}{$label}{'__ORDER__'}}, $header;
		    $data_column_available{$data_file}{'__ALL__'}{$header} = 1 unless ($label eq '__NO_LABEL__');
		}
	    }
	}
    }
    
    # make sure we found all headers:
    my %headers_not_found = ();
    foreach my $header (@columns) {
	unless ($headers{$header}) {
	    $headers_not_found{$header} = 'not found';
	}
	my $label_count = 0;
	foreach my $label (keys %{$headers{$header}}) {
	    if ($label eq '__NO_LABEL__') {
		$label_count += 2;
		next;
	    } else {
		$label_count++;
	    }
	}
	unless ($label_count == 2) {
	    $headers_not_found{$header} = 'only one label' unless (defined $headers_not_found{$header}
								   or
								   $file_specs{$file}{'FORMAT'} eq 'ImaGene');
	}
    }
		
#    warn "#KH: columns: '".(join ',', sort keys %headers)."'";
#    warn "#KH: labels: '".(join ',', sort keys %{$headers{'TYPE'}})."'";

    # warn if some headers were not found:
    my $return_message = 'ok';
    if (%headers_not_found) {
#	$return_message = "\nThe following headers produced warnings:<BR>\n<UL>\n";
#	foreach (sort keys %headers_not_found) {
#	    $return_message .= "<LI>$_: $headers_not_found{$_}</LI>\n";
#	}
#	$return_message .= "</UL>\n";
#	return "# ERROR: The following headers were not found: '".(join '\',\'', sort keys %headers_not_found)."'!\n";
    }


    # read in info for each spot
    my ($index,$row,$col,$mrow,$mcol);
    $index = -1;
    my $spot_limit_counter = 0;
    my %missing = ();
    my %flags = ();
    my $found_start = 0;
#    my %index = ();

    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);

    while (<IN>) {

	my $comment_key = '^#';
	if (defined $file_specs{$file}{'COMMENT_KEY'}) {
	    $comment_key = $file_specs{$file}{'COMMENT_KEY'};
	}
	if (/$comment_key/) {
	    next;
	} 
	
	# skip empty lines
	next unless (/\w/);
		
	$index++;

	if ($spot_limit_start) {
	    if ($index+1 >= $spot_limit_start) {
		$spot_limit_counter++;
	    }
	} else {
	    $spot_limit_counter++;
	}

	# skip the first lines before $spot_limit_start
	next if ($spot_limit_start
		 and
		 $spot_limit_start > $index+1);

	# skip the lines after spot_limit
	last if ($spot_limit =~ /\d/
		 and
		 $spot_limit_counter > $spot_limit);

	# figure out position within subgrid
	# index = ($row + $mrow * $ROW) * $MCOL * $COL
        #         + $mcol * $COL + $col - 1

	# define subgrids that need to be debugged:
	my %subgrid_debug = ();
	if ($subgrid_start) {
	    my ($mrow_start,$mcol_start);
	    if ($subgrid_start =~ /,/) {
		($mrow_start,$mcol_start) = split /,/, $subgrid_start;		
		$mrow_start =~ s/\D//g;
		$mcol_start =~ s/\D//g;
		$subgrid_start = ($mrow_start - 1) * $MCOL + $mcol_start;		
	    } else {
		$subgrid_start =~ s/\D//g;
#		$mrow_start = int($subgrid_start / $MCOL) + 1;
#		$mcol_start = $subgrid_start - ($mrow_start - 1) * $MCOL;
	    }
	    my ($mrow_end,$mcol_end) = ($mrow_start,$mcol_start);
	    if ($subgrid_end =~ /\d/) {
		if ($subgrid_end =~ /,/) {
		    ($mrow_end,$mcol_end) = split /,/, $subgrid_end;		
		    $mrow_end =~ s/\D//g;
		    $mcol_end =~ s/\D//g;
		    $subgrid_end = ($mrow_end - 1) * $MCOL + $mcol_end;		
		} else {
		    $subgrid_end =~ s/\D//g;
#		    $mrow_end = int($subgrid_end / $MCOL) + 1;
#		    $mcol_end = $subgrid_end - ($mrow_end - 1) * $MCOL;
		}
	    } else {
		$subgrid_end = $subgrid_start;
	    }

#	    for (my $i = $mrow_start; $i <= $mrow_end; $i++) {
#		for (my $j = $mcol_start; $i <= $mcol_end; $i++) {
	    for (my $i = $subgrid_start; $i <= $subgrid_end; $i++) {
		    $subgrid_debug{$i}++;
		}
#	    }
	}
	

	# figure out the row,col, mrow and mcol position
	# according to the line number:
	unless (defined $file_specs{$file}{'ORGANIZATION'}) {
	    $file_specs{$file}{'ORGANIZATION'} = 'unknown';
	}

	if ($file_specs{$file}{'ORGANIZATION'} eq 'by-subgrid'
	    or
	    $file_specs{$file}{'FORMAT'} eq 'ScanAlyze'
	    or
	    $file_specs{$file}{'FORMAT'} eq 'ImaGene'
	    or
	    $file_specs{$file}{'FORMAT'} eq 'GenePix') {
				
            # the listing here does not go from top left to bottom right for the whole array
	    # but top left to bottom right by subgrid, one subgrid after another

	    # first check which subgrid we are in:
	    my $mbox = int ($index / ($ROW * $COL));

	    # from that calculate mrow and mcol:
	    $mrow = int ($mbox / $MCOL);
	    $mcol = $mbox - $mrow * $MCOL;

	    # get new index (number of spots in last subgrid
	    my $new_index = $index - ($mrow * $MCOL + $mcol) * ($ROW * $COL);
	    
	    # calculate row and col from that
	    $row = int ($new_index / $COL);
	    $col = $new_index - ($row * $COL);
	    
	} else {

	    # go from top left to bottom right, row by row:
	    $row = int ($index / ($MCOL * $COL));
	    $mrow = int ($row / $ROW);
	    if ($mrow > 0) {
		$row =  $row % ($mrow*$ROW);
	    }
	    # subtract spots in full rows from index
	    $col = $index - (($row + $mrow * $ROW) * $MCOL * $COL);
	    $mcol = int ($col / $COL);
	    if ($mcol > 0) {
		$col = $col % ($mcol*$COL);
	    }
	}
	    
	# all indeces were counted starting from 0,
	# shift them all up by one
	$row++;
	$col++;
	$mrow++;
	$mcol++;
    
	if (%subgrid_debug) {
	    my $subgrid = ($mrow - 1) * $MCOL + $mcol;		

	    next unless (defined $subgrid_debug{$subgrid});
	}

	# delete trailing white-space
	chomp;
	s/\cM//;

	# extract required columns from line
	# split at tabs
	my @tmp = split /\t/, $_;

	# add empty field at the end:
	if ($_ =~ /\cI$/) {
	    push @tmp, '';
	}

	# set explicitely if specified in input file
	if (defined $headers{'MROW'}
	    and
	    defined $headers{'MROW'}{'__NO_LABEL__'}) {
	    my $pos = (keys %{$headers{'MROW'}{'__NO_LABEL__'}})[0];
	    $mrow = $tmp[$pos];
	    $pos = (keys %{$headers{'MCOL'}{'__NO_LABEL__'}})[0];
	    $mcol = $tmp[$pos];
	    $pos = (keys %{$headers{'ROW'}{'__NO_LABEL__'}})[0];
	    $row = $tmp[$pos];
	    $pos = (keys %{$headers{'COL'}{'__NO_LABEL__'}})[0];
	    $col = $tmp[$pos];
	}

	# filter for a specific spot id
#	my $filter_spot = '173057:H20004:Hs.31714:WD repeat domain 8:WDR8:1:1p36.3:0:173057';
	if ($spotid_filter) {
	    my $position = (keys %{$headers{$SPOT_ID_TERM}{'__NO_LABEL__'}})[0];
	    my $id = $tmp[$position];
	    $id =~ s/\s+$//;
	    unless ($id eq $spotid_filter) {
		next;
	    }
	}
	
	foreach my $header (keys %headers) {
	    LABEL:foreach my $label (keys %{$headers{$header}}) {

		my @values = ();

		foreach my $position (keys %{$headers{$header}{$label}}) {
		
		    # make sure requested column is defined
		    unless (defined $tmp[$position]) {
			$missing{$position}{"$label,$header"}++;
			warn "# WARNING: field $position ($label,$header) not defined in line ".($index+1)." '$_', '".(join "','", @tmp)."'!\n";
			next;
		    }
		    
#		    if ($arraypipe_file) {
#			# remove channel info from header:
#			if ($header =~ /^(Ag|Au|Cy3|Cy5|CH1|CH2) /i) {
#			    $header =~ s/^(Ag|Au|Cy3|Cy5|CH1|CH2) (.*)/$2/i;
#			}
#		    }

		    # die if the same field is defined multiple times
		    if (defined $data{$data_file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}) {

			# if this is second part of two-file dataset
			# make sure fields are matching
			if (defined $first_data_file{$file}) {
			    # TODO: check why this isn't working:
#			    unless ($tmp[$position] =~ /$data{$data_file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}/) {
#				return "Mismatch for file $data_file and $file at header '$header', label '$label', position '$mrow,$mcol,$row,$col': '$data{$data_file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}', '$tmp[$position]'!\n";
#			    } else {
				next LABEL;
#			    }
			} else {
			    die "# WARNING: field '$label'->'$header'->'$mrow'->'$mcol'->'$row'->'$col' is defined twice in file '$data_file': '$data{$data_file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}', '$tmp[$position]'!";
			}
		    }

		    my $value = $tmp[$position];

		    # take note of multiple lables (for marker/controls flagging)
		    if ($header eq $SPOT_ID_TERM) {
			# remove occurence number in brackets, that some software packages add
			my $value2 = $value;
			$value2 =~ s/ ?\(\d+\)$//;
			if ($FPMI_specific
			    and
			    $value2 =~ /\^/) {
			    my @tmp = split /\^/, $value2;
			    $value2 = 'undefined';
			    # check the first three fields for annotation:
			    if (defined $tmp[0]
				and
				$tmp[0] ne '') {
				$value2 = $tmp[0];			    
			    } elsif (defined $tmp[1]
				     and
				     $tmp[1] ne '') {
				$value2 = $tmp[1];
			    } elsif (defined $tmp[2]
				     and
				     $tmp[2] ne '') {
				$value2 = $tmp[2];
			    } elsif (defined $tmp[3]
				     and
				     $tmp[3] ne '') {
				$value2 = $tmp[3];
			    } elsif (defined $tmp[4]
				     and
				     $tmp[4] ne '') {
				$value2 = $tmp[4];
			    } elsif (defined $tmp[5]
				     and
				     $tmp[5] ne '') {
				$value2 = $tmp[5];
			    } elsif (defined $tmp[-1]
				     and
				     $tmp[-1] ne '') {
				$value2 = $tmp[-1];
			    }
			    # for the JB chips we can extract the probe id from the spot id:
			    if ($MROW == 12
				and
				$MCOL == 4
				and 
				$ROW == 22
				and 
				$COL == 22
				) {
				my $probe_id = '';
				if (defined $tmp[4]) {
				    $probe_id = $tmp[4];
				}
				$probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col} = $probe_id;
				$probe_names{$probe_id}++;
			    }
			}
			unless ($value2 =~ /\w/) {
			    $value2 = 'undefined';
			}
			$spot_label_occurence{$value2}++;
			$value = $value2;
		    }

		    # make sure that the value is a number in case we are reading in a FG intensity:
		    if ($header eq $FG_SIGNAL_TERM) {
			unless ($value =~ /\d/) {
			    # no digit in value, flag it!
			    if ($value eq '') {
				push @{$flags{$mrow}{$mcol}{$row}{$col}{$label}}, 'undefined';
			    } else  {
				$value = '';
				push @{$flags{$mrow}{$mcol}{$row}{$col}{$label}}, 'error';
			    }
			}
		    } elsif ($header eq $FLAG_CHANNEL_TERM) {
			if ($file_specs{$file}{'FORMAT'} eq 'ImaGene') {
			    # special flagging values for ImaGene:
#			    if ($value > 2) {
			    if ($value =~ /\D/) {
				$value = $flag_values{'error'};
			    } else {
				if ($value == 0) {
				    # do nothing
				} elsif ($value == 1
					 or
#					 $value == 2
#					 or
					 $value == 5) {
				    $value = $flag_values{'automatic'};
				} elsif ($value == 2) {
				    # this means empty automatic
				    # and would result in too many flagged spots
				    # so we decided to keep them in
				    $value = 0;
				} else {
				    $value = $flag_values{'warning'};
				}
			    }
			} elsif ($file_specs{$file}{'FORMAT'} eq 'GenePix') {
			    # special flagging values for GenePix:
			    if ($value < 0) {
				$value = $flag_values{'automatic'};
			    }
			}
		    }
			

		    # some flags might be 'P' and 'A' (present and absent)
		    # convert them to 'no flag' and 'user1'
		    if ($header eq $FLAG_CHANNEL_TERM) {
			
			# if we are dealing with TIGR's MEV format we have a variety of flags
			if ($file_specs{$file}{'FORMAT'} eq 'TMEV') {
			    if ($value eq 'Z') {
				$value = $flag_values{'error'};
			    } elsif ($value eq 'Y') {
				$value = $flag_values{'floor'};
			    } elsif ($value eq 'X') {
				$value = $flag_values{'automatic'};
			    } elsif ($value eq 'S') {
				$value = $flag_values{'ceiling'};
			    } elsif ($value eq 'C') {
				$value = 0;
			    } elsif ($value eq 'B') {
				$value = $flag_values{'warning'};
			    } elsif ($value eq 'A') {
				$value = $flag_values{'low_quality'};
			    }
			} elsif ($value =~ /^o/) {
			    if ($value eq 'ok') {
				$value = $flag_values{'no flag'};
			    } else {
				$value = $flag_values{'automatic'};
			    }
			} elsif ($value =~ /(a|p)/i) {
			    if ($value =~ /a/i) {
				$value = $flag_values{'automatic'};
			    } else {
				$value = $flag_values{'no flag'};
			    }
			}
			
			# if flag is a list that is not known yet, register it:
			if ($value > $highest_flag_num) {
			    my $count = 0;
			    foreach (@flag_names) {
				if (/list(\d+)/) {
				    my $count2 = $1;
				    if ($count2 > $count) {
					$count = $count2;
				    }
				}
			    }
			    my $flag = '';
			    while ($highest_flag_num < $value) {
				$count++;
				$flag = "list$count";
				$highest_flag_num *= 2;
				$flag_values{$flag} = $highest_flag_num;
				$flag_values_num{$highest_flag_num} = $flag;
				push @all_flags, $flag;
				unshift @all_flags_num, $highest_flag_num;
			    }
			    $used_flags{$flag}++;
			}

		    } elsif ($header eq $SPOT_QUALITY_TERM) {
			if ($file_specs{$file}{'FORMAT'} eq 'TMEV') {
			    if ($value < $SPOT_QUALITY_THRESHOLD) {
				$data{$data_file}{'__DATA__'}{'CH1'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} = $flag_values{'automatic'};
				$data{$data_file}{'__DATA__'}{'CH2'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} = $flag_values{'automatic'};
#				$data{$data_file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} = $flag_values{'automatic'};
			    }
			}
		    }

		    push @values, $value;

		}
		    
		# merge multiple values (this should really only occur for flags) 
		if (@values > 1) {
		    # make sure flag isn't set yet
		    my %flags = ();
		    foreach (@values) {
			$flags{$_}++;
		    }
		    @values = ();
		    foreach (keys %flags) {
			$values[0] += $_;
		    }
		}

		# set default to '' if no input retrieved:
		unless (@values) {
		    $values[0] = '';
		}

		# chop off trailing white space
		$values[0] =~ s/\s+$//;

		# fill data into hash:
		$data{$data_file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col} = $values[0]; 

		# did some checks of required memory and system time for 
		# different data structures:

		# above:
		# -> 91 MB, 0.7 s for 28704 rows

		# positions first:
#		$data{$data_file}{'__DATA__'}{$mrow}{$mcol}{$row}{$col}{$label}{$header} = $tmp[$headers{$label}{$header}]; 
		# -> 99 MB, 0.7 s for 28704 rows

		# positions combined into one key:
#		$data{$data_file}{'__DATA__'}{$label}{$header}{$mrow,$mcol,$row,$col} = $tmp[$headers{$label}{$header}]; 
		# -> 87 MB, 0.8 s for 28704 rows
	    }
	}
#	$index{$index+1} = "'$mrow'->'$mcol'->'$row'->'$col'";
    }

    close IN;

    # take note of markers that occur more than twice
    unless ($occurence_threshold) {
	$occurence_threshold = $MCOL * $MROW * 0.5;	
    }
    unless ($occurence_threshold > 4) {
	$occurence_threshold = 5;	
    }

    foreach my $label (keys %spot_label_occurence) {
	if ($spot_label_occurence{$label} >= $occurence_threshold) {
	    $control_labels{$label} = $spot_label_occurence{$label};
	    warn "#Spot Label Occurence: $label: $spot_label_occurence{$label}\n";
	}
	$control_labels{'done'}++;
    }

    # add flags for spots were invalid foreground signal was detected
    foreach my $mrow (keys %flags) {
	foreach my $mcol (keys %{$flags{$mrow}}) {
	    foreach my $row (keys %{$flags{$mrow}{$mcol}}) {
		foreach my $col (keys %{$flags{$mrow}{$mcol}{$row}}) {
		    foreach my $label (keys %{$flags{$mrow}{$mcol}{$row}{$col}}) {
			foreach my $flag (@{$flags{$mrow}{$mcol}{$row}{$col}{$label}}) {
			    my $flag_value = $flag_values{$flag};
			    $data{$data_file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_value;
			}
		    }
		}
	    }
	}
    }
    
    # record number of spots
    $data{$data_file}{'__SPOTS__'} = $index+1;

    # keep track of the headers that have been read in already
    my %labels = ();
    foreach my $header (keys %headers) {
	my $header_tmp = $header;
#	if ($arraypipe_file) {
#	    # remove channel info from header:
#	    if ($header_tmp =~ /^(Ag|Au|Cy3|Cy5|CH1|CH2) /i) {
#		$header_tmp =~ s/^(Ag|Au|Cy3|Cy5|CH1|CH2) (.*)/$2/i;
#	    }
#	}
	
	foreach my $label (keys %{$headers{$header}}) {
	    $labels{$label}++;
	    $data_done{$data_file}{$label}{$header_tmp}++;
	    $file_input{$file}{$label}{$header_tmp}++;
	}
    }

    # add flags unless specified
    foreach my $label (keys %labels) {
	my $add_on = '';
#	if ($arraypipe_file) {
#	    $add_on = "$label ";
#	} else {
#	    next if ($label eq '__NO_LABEL__');
#	}
	unless (keys %{$headers{$add_on.$FLAG_CHANNEL_TERM}{$label}}) {
	    $data_done{$data_file}{$label}{$FLAG_CHANNEL_TERM}++;
	    $data_column_available{$data_file}{$label}{$FLAG_CHANNEL_TERM}++;
	    foreach $mrow (1..$MROW) {
		foreach $mcol (1..$MCOL) {
		    foreach $row (1..$ROW) {
			foreach $col (1..$COL) {
			    if (defined $data{$data_file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col}) {
				unless (defined $data{$data_file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
				    $data{$data_file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} = 0;
				}
			    }
			}
		    }
		}
	    }
	    $file_input{$file}{$label}{$FLAG_CHANNEL_TERM}++;
	} else {
	    # flags are valid for all channels if set in one
	    $data_done{$data_file}{$label}{$FLAG_CHANNEL_TERM}++;
	}
    }
#    $data_done{$data_file}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}++;
#    $data_column_available{$data_file}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}++;

    # make sure that necessary parameters are set:
    unless (defined $file_specs{$file}{'CH1FG'}) {
	return 'Foreground column for channel 1 is not defined!';
    }
    unless (defined $file_specs{$file}{'CH1BG'}) {
#	return 'Background column for channel 1 is not defined!';
    }
    unless (defined $file_specs{$file}{'CH2FG'}) {
	return 'Foreground column for channel 2 is not defined!';
    }
    unless (defined $file_specs{$file}{'CH2BG'}) {
#	return 'Background column for channel 2 is not defined!';
    }
    unless (defined $file_specs{$file}{'SPOT_ID'}) {
	 $return_message .= '; column for spot ID is not defined';
	 $file_specs{$file}{'SPOT_ID'} = '- undefined -';
    }

    my $alias = '';
    if (defined $cgi->param("ALIAS_$file")
	and
	$cgi->param("ALIAS_$file") =~ /\w/) {
	$alias = $cgi->param("ALIAS_$file");
    }
    $file_specs{$file}{'ALIAS'} = $alias;

    if (%missing) {
	foreach my $pos (sort keys %missing) {
	    foreach my $field (sort keys %{$missing{$pos}}) {
		my $nums = $missing{$pos}{$field};
		$return_message .= "; at position $pos ($field) $nums values are missing ($file)";
	    }
	}
    }    

    return $return_message;
}


sub match_header {
    my $header = shift;
    my $file = shift;

    if (defined $cgi->param("CH1FG_$file")
	and
	$cgi->param("CH1FG_$file") =~ /\w/) {
	$file_specs{$file}{'CH1FG'} = $cgi->param("CH1FG_$file");
    }

    if (defined $cgi->param("CH1BG_$file")
	and
	$cgi->param("CH1BG_$file") =~ /\w/) {
	$file_specs{$file}{'CH1BG'} = $cgi->param("CH1BG_$file");
    }

    if (defined $cgi->param("CH2FG_$file")
	and
	$cgi->param("CH2FG_$file") =~ /\w/) {
	$file_specs{$file}{'CH2FG'} = $cgi->param("CH2FG_$file");
    }

    if (defined $cgi->param("CH2BG_$file")
	and
	$cgi->param("CH2BG_$file") =~ /\w/) {
	$file_specs{$file}{'CH2BG'} = $cgi->param("CH2BG_$file");
    }

    if (defined $cgi->param("SPOT_ID_$file")
	and
	$cgi->param("SPOT_ID_$file") =~ /\w/) {
	$file_specs{$file}{'SPOT_ID'} = $cgi->param("SPOT_ID_$file");
    }

    if (defined $cgi->param("FLAG_$file")) {
	foreach my $flag ($cgi->param("FLAG_$file")) {		
	    if ($flag =~ /\w/) {
		if (defined $flag_done{$file}{$flag}) {
		    next;
		} else {
		    push @{$file_specs{$file}{'FLAG'}}, $flag;
		    $flag_done{$file}{$flag}++;
		}
	    }
	}
    }

    if ($file_specs{$file}{'CH1FG'} eq $header) {
	return $FG_SIGNAL_TERM,'CH1';
    } elsif (defined $file_specs{$file}{'CH1BG'}
	     and
	     $file_specs{$file}{'CH1BG'} eq $header) {
	return $BG_SIGNAL_TERM,'CH1';
    } elsif ($file_specs{$file}{'CH2FG'} eq $header) {
	return $FG_SIGNAL_TERM,'CH2';
    } elsif (defined $file_specs{$file}{'CH2BG'}
	     and
	     $file_specs{$file}{'CH2BG'} eq $header) {
	return $BG_SIGNAL_TERM,'CH2';
    } elsif ($file_specs{$file}{'SPOT_ID'} eq $header) {
	return $SPOT_ID_TERM,'__NO_LABEL__';
    } elsif (defined $file_specs{$file}{'SPOT_QUALITY'}
	     and
	     $file_specs{$file}{'SPOT_QUALITY'} eq $header) {
	return $SPOT_QUALITY_TERM,'__NO_LABEL__';
    } elsif (defined $file_specs{$file}{'WEIGHTS'}
	     and
	     $file_specs{$file}{'WEIGHTS'} eq $header) {
	return $WEIGHTS_TERM,'__NO_LABEL__';
    } elsif (defined $file_specs{$file}{'VALID RATIOS'}
	     and
	     $file_specs{$file}{'VALID_ENTRIES'} eq $header) {
	return $VALID_ENTRIES_TERM,'__NO_LABEL__';
    } elsif ($header eq $SPOT_QUALITY_TERM) {
	return $SPOT_QUALITY_TERM,'__NO_LABEL__';
    } else {
	foreach (@{$file_specs{$file}{'FLAG'}}) {
	    if ($_ eq $header) {
		return $FLAG_CHANNEL_TERM,'__NO_LABEL__';
	    }
	}
	# if we got here, than none of the headers matched
	return $header, '';
    }
}
    
      

sub read_reporter_list {
    
    # get the filename
    my $file = shift;
    my $flag = shift;

    # counters for column headers:
    my @counters = ();

    # check file format if requested:
    if ($cgi->param('check_files')) {

	my ($format_check) = &check_file_format($file, 'reporter_list');
	unless ($format_check eq 'ok') {
	    warn "# WARNING: file '$file' has invalid format: $format_check - skipping!";
	    $wrong_format_files{$file} = $format_check;
	    return 'invalid format';
	}
    }

    if ($to_dir) {
	$file =~ s/$from_dir/$to_dir/;
    }

    # open the file:
    open (IN, $file)
	or return "Can't read file '$file': $!\n";

    my @headers = ();
    my ($mrow,$mcol,$row,$col,$mrow_pos,$mcol_pos,$row_pos,$col_pos);
    my %matches = ();
    my %skip_i = ();
	
    # the first line specifies which columns are used for identifying a spot
    $_ = '';
    until (/\w/) {
	$_ = <IN>;
	
	# remove comments
	if (/#/) {
	    s/#.*//;
	}
	
	# delete trailing white-space
	s/\s*$//;
	
	last if (eof(IN));
    }

    if (/Probe ID\t/
	or
	/ROW\t/
	or
	/Spot labels\t/
	or
	/$FG_SIGNAL_TERM\t/
	or
	/Bkgd\t/
	or
	/Flag\t/
	){
	# looks like output from the pipeline
	# specify keyword for recognition of column headers:
	$_ = 'COLUMNS:'.$_;
    }

    unless (/COLUMNS:/) {
	@headers = split /\t/, $_;
	if (@headers == 4) {
	    $mrow_pos = 0;
	    $mcol_pos = 1;
	    $row_pos = 2;
	    $col_pos = 3;

	    # go back to start of file
	    seek IN,0,0;
	} else {
	    # default is to match with Spot labels
	    @headers = ($SPOT_ID_TERM);
	    # if probe nucleotide specified, use that column:
	    if (/H0.*_01/) {
		@headers = ('Probe ID');
		my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);
		&get_probe_ids($file,$MROW,$MCOL,$ROW,$COL);
	    }
	    $matches{$headers[0]}{$_}++ if (/\w/);
	}

	# first line might contain data, so we go back to beginning of file
	seek IN,0,0;

    } else {
	
	# remove tag
	s/COLUMNS:\s*//g;

	@headers = ();
	my $skip = 0;
	my @tmp = split /\t/, $_;
	for ( my $i = 0; $i <= $#tmp; $i++) {
	    if ($tmp[$i] eq 'SKIP') {
		$skip++;
	    } else {
		push @headers, $tmp[$i];
	    }
	}

	if (@headers == 4) {
	    for (my $i = 0 ; $i <= $#headers; $i++) {
		if ($headers[$i] eq 'MROW') {
		    $mrow_pos = $i + $skip;
		} elsif ($headers[$i] eq 'MCOL') {
		    $mcol_pos = $i + $skip;
		} elsif ($headers[$i] eq 'ROW') {
		    $row_pos = $i + $skip;
		} elsif ($headers[$i] eq 'COL') {
		    $col_pos = $i + $skip;
		}
	    }

	    # make sure we have positions for each key
	    my @headers_tmp = ('mrow','mcol','row','col');
	    foreach my $pos ($mrow_pos,$mcol_pos,$row_pos,$col_pos) {
		my $header = shift @headers_tmp;
		unless ($pos =~ /\d/) {
		    return "No position for key '$header', headers found: '".(join "','", @headers)."'!\n";
		}
	    }
	} else {
	    # get some important columns
	    my %headers = ();
	    for (my $i = 0; $i <= $#headers; $i++) {
		my $header = $headers[$i];
		$headers{$header} = $i;
	    }
	    if (defined $headers{'MROW'}
		and
		defined $headers{'MCOL'}
		and
		defined $headers{'ROW'}
		and
		defined $headers{'COL'}) {
		$mrow_pos = $headers{'MROW'} + $skip;
		$mcol_pos = $headers{'MCOL'} + $skip;
		$row_pos = $headers{'ROW'} + $skip;
		$col_pos = $headers{'COL'} + $skip;
		@headers = (1,1,1,1);
	    } elsif (defined $headers{'Probe ID'}) {
		for(my $i = 0; $i <= $#headers; $i++) {
		    my $header = $headers[$i];
		    unless ($header eq 'Probe ID') {
			$skip_i{$i}++;
		    }
		}
	    } elsif (defined $headers{$SPOT_ID_TERM}) {
		for(my $i = 0; $i <= $#headers; $i++) {
		    my $header = $headers[$i];
		    unless ($header eq $SPOT_ID_TERM) {
			$skip_i{$i}++;
		    }
		}
	    }
	}
    }

    my %done = ();
    while (<IN>) {	

	# remove comments
	if (/#/) {
	    s/#.*//;
	}

	# skip empty lines
	next unless (/\w/);
		
	# delete trailing white-space
	s/\s*$//;

	# extract required columns from line
	# split at tabs
	my @tmp = split /\t/, $_;

	if (@headers == 4) {
	    $mrow = $tmp[$mrow_pos];
	    $mcol = $tmp[$mcol_pos];
	    $row = $tmp[$row_pos];
	    $col = $tmp[$col_pos];
	    foreach my $data_file (keys %data) {
		next if ($done{$data_file}{$mrow}{$mcol}{$row}{$col});
		foreach my $label (keys %{$data{$data_file}{'__DATA__'}}) {
		    next if ($label eq '__NO_LABEL__');
		    $data{$data_file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag;
		}
		$done{$data_file}{$mrow}{$mcol}{$row}{$col}++;
		$list_match{$file}{$data_file}++;
	    }
	} else {
	    for (my $i = 0 ; $i <= $#tmp; $i++) {
		next if ($skip_i{$i});
		$matches{$headers[$i]}{$tmp[$i]}++;
	    }
	}
    }

    close IN;

#    foreach (sort keys %matches) {
#	warn "Match: $_";
#    }

    if (%matches) {
	my %pattern_match = ();
	my %patterns = ();
 	# cycle through each spot and find matches
	foreach my $data_file (sort keys %data) {
	    # make sure the according columns are available
	    my $rc = &read_data($data_file,(keys %matches));
	    unless ($rc =~ /ok/) {
		warn "# File $data_file has error: $rc";
		next;
	    }

	    $list_match{$file}{$data_file} = 0;
	    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($data_file);
	    foreach $mrow (1..$MROW) {
		foreach $mcol (1..$MCOL) {
		    foreach $row (1..$ROW) {
			foreach $col (1..$COL) {
			    
			    foreach my $header (@headers) {
				
				my $flag_spot = '';
					
				if ($header eq 'Probe ID') {
				    foreach my $pattern (keys %{$matches{$header}}) {
					if (defined $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col}
					    and
					    $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col} eq $pattern) {
					    $flag_spot++;
					    last;
					}
				    }
				} else {
				    
				    if ($header eq $SPOT_ID_TERM) {
					my $spot_label = $data{$data_file}{'__DATA__'}{'__NO_LABEL__'}{$header}{$mrow}{$mcol}{$row}{$col};
					if (defined $matches{$header}{$spot_label}) {
					    $flag_spot++;
					    $patterns{$spot_label}++;
					    $pattern_match{$spot_label}++;
					} else {					
					    ($spot_label) = split /\^/, $spot_label;
					    if (defined $matches{$header}{$spot_label}) {
						$flag_spot++;
						$patterns{$spot_label}++;
						$pattern_match{$spot_label}++;
					    } else {					
						$spot_label =~ s/ +$//g;
						if (defined $matches{$header}{$spot_label}) {
						    $flag_spot++;
						    $patterns{$spot_label}++;
						    $pattern_match{$spot_label}++;
						} else {						    
						    $spot_label =~ s/\s*\(\d+\)$//g;
						    if (defined $matches{$header}{$spot_label}) {
							$flag_spot++;
							$patterns{$spot_label}++;
							$pattern_match{$spot_label}++;
						    }
						}
					    }
					}
#					warn "#MATCH for $spot_label: $pattern_match{$spot_label}" if (defined $pattern_match{$spot_label});
				    } else {
				    
				      LABEL:foreach my $label (keys %{$data{$data_file}{'__DATA__'}}) {
					  
					  next unless (defined $data{$data_file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col});
					  
					  foreach my $pattern (keys %{$matches{$header}}) {
					      
					      $patterns{$pattern}++;
					      
					      if ($header eq $SPOT_ID_TERM
						  and
						  $FPMI_specific
						  and
						  $pattern !~ /\^/) {
						  my $spot_label = $data{$data_file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
						  ($spot_label) = split /\^/, $spot_label;
						  unless ($pattern =~ / $/) {
						      $spot_label =~ s/ +$//g;
						  }
#						if ($spot_label =~ /^$pattern/) {
#						warn "comparing '$spot_label' with '$pattern'\n";
						  if ($spot_label eq $pattern) {
						      $pattern_match{$pattern}++;
						      $flag_spot++;
						      last LABEL;
						  }
					      } else {
						  my $spot_label = $data{$data_file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
						  # remove trailing '(1)' for duplicate labels
						  # unless it is specified in the list as well
						  unless ($pattern =~ /\(\d+\)$/) {
						      $spot_label =~ s/\s*\(\d+\)$//g;
						  }
						  if ($spot_label eq $pattern) {
						      $pattern_match{$pattern}++;
						      # flag data if matched up:
						      $flag_spot++;
						    last LABEL;
						  }
					      }
					  }
				      }
				    }
				}

				# set flag if necessary
				if ($flag_spot) {
				    foreach my $label (keys %{$data{$data_file}{'__DATA__'}}) {
					next if ($label eq '__NO_LABEL__');
					$data{$data_file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag;
				    }
				    $list_match{$file}{$data_file}++;
				}
			    }
			}
		    }
		}
	    }
	}	
	foreach my $pattern (keys %patterns) {
	    unless (defined $pattern_match{$pattern}) {
		warn "#KH: No match for '$pattern'";
	    }
	}
    }

    return 'ok';
}


sub list_action_items {

    # get all saved settings:
    my @saved_settings = ();
    if (opendir(DIR, $save_settings_dir)) {
	while (defined(my $file = readdir(DIR))) {
	    if (-f "$save_settings_dir/$file"
		and
		-r "$save_settings_dir/$file") {
		push @saved_settings, $file;
	    }
	}
	close DIR;
    } else {
#	warn "# $progname WARN: can't open '$save_settings_dir': $!";
    }	
    
    if (@saved_settings) {
	@saved_settings = sort @saved_settings;
	my $override = 0;
	my $default = '';

	if ($submit eq 'Save Settings!') {
	    $override = 1;
	    $default = $cgi->param('save_settings_name');
	}
	$saved_settings_html = "<BR>Load previously saved settings: "
	    . $cgi -> popup_menu(-name     => 'load_settings',
				 -override => $override,
				 -values   => [ @saved_settings ],
				 -default  => $default)
	    . $cgi -> submit(-name=>'submit',
			     -value=>'Load Settings!')
	    . $cgi -> submit(-name=>'submit',
			     -value=>'Load Settings and Do it!')
	    ."<BR>\n";
#	print "$saved_settings_html";
    }

    $output_collect{'ACTION SETUP'} .= &print_action_section;

#    warn "# KH: actions = @actions";
    $output_collect{'SETTINGS'} .= &print_settings_section;

#    $output_collect{'FINISH'} .=  "<BR><BR><A HREF=\"#INFO\">[info section]</A>&nbsp;<A HREF=\"#ACTION SETUP\">[action section]</A>&nbsp;<A HREF=\"#SETTINGS\">[settings]</A>&nbsp;<BR>\n"
#	. "<TABLE><TR><TD><TABLE CELLPADDING=5><TR><TD BGCOLOR=RED>"
#	. $cgi -> submit(-name=>'submit', 
#			 -value=>'Do it!')
#	. '</TD></TR></TABLE>'
#	. '</TD><TD>'.&help('do_it_button').'</TD></TR></TABLE>'
#	."\n";

    if ($action_step_num > 1) {
	$output_collect{'FINISH'} .=  "<BR><BIG><B><A NAME=\"previous_steps\">Previous Steps:</A></B></BIG>&nbsp;".&help('previous_steps')."<BR>\n<TABLE BORDER=0 CELLSPACING=5>\n";
	for (my $i = 1; $i < $action_step_num; $i++) {
	    my $extension = $action_step_ext;
	    my $file_html = $results_file_html;
	    my $file = $results_file;
	    $file =~ s/_step_\d+(\.\d+)*/_step_$i$extension/;
	    while ($extension) {
		if (-e $file) {		    
		    last;
		} else {
		    $extension =~ s/\.\d+$//;
		    $file =~ s/_step_\d+(\.\d+)*/_step_$i$extension/;
		}
	    }
		
	    $file_html = basename($file);
	    my $default_text = '';
	    if (defined $cgi->param("step_label$i$extension")) {
		$default_text = ' VALUE="'.$cgi->param("step_label$i$extension").'"';
	    }
	    $output_collect{'FINISH'} .=  "<TR><TD><A HREF=\"$file_html\">Step $i$extension</A>:</TD><TD><INPUT SIZE=80 TYPE=TEXT NAME=\"step_label$i$extension\"$default_text></TD></TR>\n";
	}
	
	# current step:
	my $default_text = $cgi->param('submit');
	if (@actions
	    and
	    $cgi->param('submit') =~ /Do it/i) {
	    $default_text = 'Actions: ';
	}
	if ($cgi->param('submit') =~ /do it/i) {
	    my @descriptions = ();
	    foreach (@actions) {
		my $description = $_;
		if (defined $action_items{$_}) {
		    $description = $action_items{$_};
		}
		push @descriptions, $description;
	    }
	    $default_text .= join ', ', @descriptions;
	}

	if (defined $cgi->param("new_step")
	    and
	    $cgi->param("new_step") =~ /\w/) {
	    $default_text = $cgi->param("new_step");
	}
	$default_text = ' VALUE="'.$default_text.'"';
	$output_collect{'FINISH'} .=  "<TR><TD>Step $action_step:</TD><TD><INPUT SIZE=80 TYPE=TEXT NAME=\"step_label$action_step\"$default_text></TD></TR>\n";
	$output_collect{'FINISH'} .=  "<TR><TD>Next step:</TD><TD><INPUT SIZE=80 TYPE=TEXT NAME=\"new_step\"></TD></TR>\n";
	
	$output_collect{'FINISH'} .=  "</TABLE>\n<BR><A HREF=\"index.html\">Index</A> of all HTML files generated.<BR>\n";
    } else {
	$output_collect{'FINISH'} .=  "<input type=\"hidden\" name=\"step_label$action_step\" value=\"$submit\">\n";
    }
}


sub print_settings_section {

    my $out = "<!-- INFO START of SETTINGS section -->\n";

    # check what view has been selected:
    if (defined $cgi->param('section_view_SETTINGS')) {
	$section_view{'SETTINGS'} = $cgi->param('section_view_SETTINGS');
    }
    # overwrite with the button that has been pressed
    if ($submit =~ /^Hide (.*)/) {
        &section_view($1, 'hide');
    } elsif ($submit =~ /^Show (.*)/) {
        &section_view($1, 'show');
    }  

    # if no sections are to be shown make sure the selected values are stored in hidden fields:
    if ($section_view{'SETTINGS'} eq 'hide all') {
	$out .= "Settings are hidden - click button to show: "
	    . $cgi -> submit(-name=>'submit', 
			     -value=>'Show all settings!')
	    . $cgi -> submit(-name=>'submit', 
			     -value=>'Show used settings!')
	    . "<BR><BR>\n";
	# add all the values for variables where something other than the default was selected:
	foreach (@actions) {
	    foreach (keys %action_items) {
		foreach my $param (keys %item_var) {
		    if (defined $cgi->param($param)) {
			foreach my $value ($cgi->param($param)) {
			    if ($var_def_default{$param} eq $value) {
				next;
			    } else {
				$out .= "<input type=\"hidden\" name=\"param\" value=\"$value\">\n";
			    }
			}
		    }
		}
	    }
	}    
	return $out;
    }

    my %selected_settings = ();
    if ($section_view{'SETTINGS'} eq 'show all') {
#	$out .= $cgi -> submit(-name=>'submit', 
#			       -value=>'Hide all settings!')
#	    . $cgi -> submit(-name=>'submit', 
#			     -value=>'Hide unused settings!')
#	    . "<BR><BR>\n";
    } else {
	# compile which settings are to be shown:
	if ($section_view{'SETTINGS'} eq 'hide unused'
	    or
	    $section_view{'SETTINGS'} eq 'show used') {
	    foreach (@actions) {
		$selected_settings{$_}++;
	    }
	}

#	$out .= $cgi -> submit(-name=>'submit', 
#			       -value=>'Hide all settings!')
#	    . $cgi -> submit(-name=>'submit', 
#			     -value=>'Show all settings!')
#	    . $cgi -> submit(-name=>'submit', 
#			     -value=>'Show used settings!')

	$out .= "Only the settings for the functions in use are shown. To see more, choose additional modules in the Action Setup above and click on the 'Refresh' button: "
	    . $cgi -> submit(-name=>'submit', 
			     -value=>'Refresh!')
	    . "<BR><BR>\n";
    }
    
    $out .= "\n<H3><A NAME=\"general_settings\">General:</A></H3>\n";
    $out .= &pointer."You can adjust the following general settings:<BR><BR>\n";
    $out .= "<TABLE BORDER=0 CELLSPACING=8>\n";
    $out .= "<TR>\n<TH colspan=2>scale images</TH>\n</TR>\n";
    $out .= "<TR>\n<TD>scale images</TD>\n<TD><input type=\"text\" size=4 name=\"SCALE_IMAGE\" value=\"$SCALE_IMAGE\"> (0 < scale <= 4)</TD>\n</TR>\n";
    $out .= "</TABLE>\n";


    $out .= "<H3><A NAME=\"module_settings\">Module specific:</A></H3>\n";
    $out .= &pointer."Modify the parameters for the action steps that you plan to apply. Once you're finished, click on the "	
	. $cgi -> submit(-name=>'submit', 
			 -value=>'Do it!')
	. " button in the <A HREF=\"#action_start\">action section</A> to start the processing.<BR> The following modules are available:<BR><BR>\n";

    # print menu of action items:

    $out .= "<TABLE BORDER=0 CELLSPACING=8>\n<TR>\n";
    my $columns = 0;;
    foreach my $category (sort keys %module_category) {
	# go to next row after four entries
	if ($columns == 4) {
	    $out .= "</TR>\n<TR>\n";
	    $columns = 0;
	}

	$out .= "<TD VALIGN=TOP>\n<TABLE BORDER=0><TR><TH>$category</TH></TR>\n";
	my @items = sort @{$module_category{$category}};
	if ($category =~ /often/) {
	    # don't sort the common modules,
	    # assume they are in a specific order already:
	    @items = @{$module_category{$category}};
	}
	foreach my $item (@items) {
	    if (%selected_settings) {
		# skip setting for actions that weren't used
		unless ($selected_settings{$item}) {
		    next;
		}
	    }
	    my $label = '';
	    if (defined $action_items{$item}) {
		$label = $action_items{$item};
	    }	    
	    $out .= "<TR><TD><A HREF=\"#$item\">$label</A></TD></TR>\n";
	}
	$out .= "</TABLE></TD>\n";
	$columns++;
    }
    $out .= "</TR></TABLE>\n";


    $out .= "<TABLE BORDER=2 CELLPADDING=5>\n";
    $out .= "<TR BGCOLOR=#AAAAAA><TH>Function</TH><TH>Parameters</TH><TH>Description</TH></TR>\n";

    if (scalar(@actions) >= $max_actions + 1) {
	$max_actions = (scalar @actions) + 2;
    }

    foreach my $action (sort keys %action_items) {

	if (%selected_settings) {
	    # skip setting for actions that weren't used
	    unless ($selected_settings{$action}) {
		next;
	    }
	}

	# list action
	$out .= "<TR><TD VALIGN=TOP><A NAME=\"$action\">$action_items{$action}</A><br>\n";
	$out .= "</TD><TD>\n<TABLE>";
	# list parameters:
	my $param_count = 0;
	foreach my $param (@{$item_var{$action}}) {
	    my $override = 0;
	    my $default = $var_def_default{$param};
	    if ($cgi->param("$param")) {
		$default = $cgi->param("$param");
	    }
	    if (defined $saved_settings_var{$param}) {
		$override = 1;
		if (@{$saved_settings_var{$param}} > 1) {
		    my @default = @{$saved_settings_var{$param}};
		    $default = \@default;
		} else {
		    $default = shift @{$saved_settings_var{$param}};
		}
		if ($default eq '') {
		    $default = 'force_empty';
		}
	    }
	    $out .= "<TR VALIGN=TOP><TD>$var_def_text{$param}</TD><TD>".
		(&var_input($param,$param,$var_def_type{$param},$default,$override)).
		" ".
		(defined $var_def_add{$param} ? $var_def_add{$param} : '').
		"</TD></TR>\n";
	    $param_count++;
	}

	# add 'force' checkbox
	if ($param_count
	    and
	    $show_parameters) {
	    $out .= "<TR><TD>force</TD><TD>".
		(&var_input("${action}_force","${action}_force",'check_single','')).
		"</TD></TR>\n";
	}
	    
	unless (defined $item_description{$action}) {
	    warn "# WARNING: no description for $action!";
	    $item_description{$action} = 'no description available';
	}
	$out .= "</TABLE></TD><TD><TABLE WIDTH=300><TR><TD>$item_description{$action}<br><A HREF=\"#SETTINGS\">[back to settings]</A></TD></TR></TABLE>\n";
	$out .= "</TD></TR>\n";
    }
    
    $out .= "</TABLE><BR>\n";
    
    return $out;
}


sub print_action_section {

    my $out = '';

#    $out .= "\n<BR><TABLE WIDTH=$main_width cellpadding=\"6\"><TR BGCOLOR=lightgrey><TD><CENTER><BIG><B><A NAME=\"ACTION SETUP\">ACTION SETUP</A></B></BIG></CENTER></TD></TR></TABLE><BR>\n";
#    $out .= "$back_to_info&nbsp;<A HREF=\"#SETTINGS\">[forward to settings]</A>&nbsp;<A HREF=\"#action_start\">[action buttons]</A><BR>\n";
    $out .= &pointer."Please ".($settings_loaded ? 'revise' : 'configure')." your action settings and then proceed to the <A HREF=\"#action_start\">action buttons</A>.<BR>\n";
    $out .= "$saved_settings_html";
    if ($settings_loaded
	or
	$submit eq 'Save Settings!') {
	my $description_text = '';
	my $file = '';

	if ($submit eq 'Save Settings!'
	    and
	    defined $cgi->param('save_settings_name')) {
	    $cgi->param('save_settings_name');
	}

	if ($settings_loaded
	    and
	    defined $cgi->param('load_settings')) {
	    $file = $cgi->param('load_settings');
	}

	if (open (IN, "$save_settings_dir/$file")) {
	    while (<IN>) {
		if (/^\# NOTE\:/) {
		    while (<IN>) {
			last if (/^\# /);
			chomp;
			 if ($_) {
			     $description_text .= '<BR>' unless (/^<.*>/);
			     $description_text .= $_."\n";
			 }
		    }
		}
	    }
	    close IN;
	    $description_text =~ s/<BR>$//s;
	}
	if ($description_text) {
	    $out .= "<BR>Description for '".($cgi->param('load_settings'))."' :<BR>\n<TABLE BORDER=2 WIDTH=$main_width><TR><TD>$description_text</TD></TR></TABLE>\n";
	}
	$out .= "<BR>The action steps are listed below and can be modified.".&help('action_chain')."\n<BR>\n<TABLE BORDER=0 CELLPADDING=4>\n";
    } else {
	if ($saved_settings_html) {
	    $out .= "<B>or</B><BR>\n";
	}
	$out .= "Specify the action steps you plan to apply. ".&help('action_chain')."\n<BR>\n<TABLE BORDER=0 CELLPADDING=4>\n";
    }
    if (defined $saved_settings_var{'max_actions'}) {
	$max_actions = @{$saved_settings_var{'max_actions'}}[0];
    }
    if (defined $saved_settings_var{'action_alternatives'}) {
	$action_alternatives = @{$saved_settings_var{'action_alternatives'}}[0];
    }
    $out .= "<BR>Layout: "
	. $cgi ->  textfield(-name => 'max_actions', 
			     -override => 1,
			      -default => $max_actions, 
			      -size => 3)
	." x "
	. $cgi ->  textfield(-name => 'action_alternatives', 
			     -override => 1,
			     -default => $action_alternatives, 
			     -size => 2)
	." &rarr; "
	. $cgi -> submit(-name=>'submit',
			  -value=>'Refresh!')
	."<BR><BR>\n";

    my $action_counter = 0;
    for my $i (1..$max_actions) {
	$out .= "<TR BGCOLOR=#AAAAAA>\n";
	
	for my $k (1..$action_alternatives) {
	    my $default = '';
	    my $td_colour = '';
	    $action_counter++;
	    if (defined $saved_settings_action{$action_counter}) {
		$default = $saved_settings_action{$action_counter};
		$td_colour = ' BGCOLOR=#FFFF00';
	    } elsif ($cgi->param("action$action_counter")) {
		# avoid mixing of previous steps and saved settings steps:
		if (%saved_settings_action) {
		    $default = '';
		} else {
		    $default = $cgi->param("action$action_counter");
		    $td_colour = ' BGCOLOR=#FFFF00';
		}
	    }
	    if ($submit eq 'Clear!') {
		$default = '';
		$td_colour = ' BGCOLOR=#AAAAAA';
	    }
	    $out .= "<TD$td_colour>\n<SELECT name=\"action$action_counter\" size=\"1\">\n";
	    $out .= "<OPTION VALUE=\"\">  -  </OPTION>\n";
	    
	    my $beginning = 1;
	    foreach my $category (sort keys %module_category) {
		# for any but the first category, include a blank line in the menu:
		if ($beginning) {
		    $beginning = 0;
		} else {
		    $out .= "<OPTION VALUE=\"\">   </OPTION>";
		}
		$out .= "<OPTION VALUE=\"\">-- $category --</OPTION>\n";
		my @items = sort @{$module_category{$category}};
		if ($category =~ /often/) {
		    # don't sort the common modules,
		    # assume they are in a specific order already:
		    @items = @{$module_category{$category}};
		}
		foreach (@items) {
		    my $label = '';
		    if (defined $action_items{$_}) {	
			$label = $action_items{$_};
		    }
		    $out .= "<OPTION VALUE=\"$_\"".($_ eq $default ? ' SELECTED' : '').">$label</OPTION>\n";
		}
	    }

	    $out .= "</select>\n";
	    $out .= "&rarr;</TD>\n";
	}
	$out .= "</TR>\n";
    }
    
	
    my $settings_name = 'usersettings ';
    # check how many we have already that start with this string
    
    if (opendir(DIR, $save_settings_dir)) {
	my @nums = ();
	while (defined(my $file = readdir(DIR))) {
	    if ($file =~ /$settings_name\s*(\d+)/) {
		push @nums, $1;
	    }
	}
	close DIR;
	if (@nums) {
	    @nums = sort { $a <=> $b } @nums;
	    $settings_name .= $nums[-1] + 1;
	} else {
	    $settings_name .= 1;
	}
    } else {
	$settings_name .= 1;
    }
    
    if ($settings_loaded) {
	$settings_name = $cgi->param('load_settings');
	if ($settings_name =~ /(\d+)$/) {
	    my $num = $1;
	    $num++;
	    $settings_name =~ s/(\d+)$/$num/;
	}
    } elsif (defined $cgi->param('save_settings_name')) {
	$settings_name = $cgi->param('save_settings_name');
	if ($settings_loaded) {
	    if ($settings_name =~ /(\d+)$/) {
		my $num = $1;
		$num++;
		$settings_name =~ s/(\d+)$/$num/;
	    }
	}
    } else {
	if (defined $cgi->param('load_settings')) {
	    $settings_name = $cgi->param('load_settings');
	}
    }

    $out .= "</TR>\n</TABLE>\n";

    $out .= "$back_to_info&nbsp;<SMALL><A HREF=\"#SETTINGS\">[SETTINGS]</A></SMALL>\n";

    # colour the submit button red
    $out .= "\n<H3><A NAME=\"action_start\">Actions:</A></H3>\n<TABLE><TR VALIGN=\"TOP\" ALIGN=CENTER>"
    	
#    $out .= $cgi->defaults('Defaults');
	. "<TD>"
	. "<TABLE CELLPADDING=5><TR><TD>"
	. $cgi-> reset(-name=>'Reset!', 
		       -value=>'Reset!')
	. '</TD></TR></TABLE>'
	. &help('reset_button')
	. "</TD>\n<TD>"
	. "<TABLE CELLPADDING=5><TR><TD>"
	. $cgi-> submit(-name=>'submit', 
		       -value=>'Clear!')
	. '</TD></TR></TABLE>'
	. &help('clear_button')
	. "</TD>\n<TD>"
	. "<TABLE CELLPADDING=5><TR><TD BGCOLOR=BLUE>"
	. $cgi -> submit(-name=>'submit', 
			 -value=>'Refine!')
	. '</TD></TR></TABLE>'
	. &help('refine_button')
	. "</TD>\n<TD>"
	. "<TABLE CELLPADDING=5><TR><TD BGCOLOR=RED>"
	. $cgi -> submit(-name=>'submit', 
			 -value=>'Do it!')
	. '</TD></TR></TABLE>'
	. &help('do_it_button')
	. "</TD>\n<TD>"
	. "<TABLE CELLPADDING=5><TR><TD BGCOLOR=GREEN>"
	. $cgi-> submit(-name=>'submit', 
		       -value=>'Save Settings!')
	. '</TD></TR></TABLE>'
	. &help('save_settings_button')
	. "</TD>\n<TD>"
	. "<TABLE CELLPADDING=5><TR><TD VALIGN=TOP BGCOLOR=GREEN>"
	. '<B>name:</B></TD><TD VALIGN=TOP BGCOLOR=GREEN>'
	.$cgi ->  textfield(-name => 'save_settings_name', 
			    -override=>1,
			    -default => $settings_name, 
			    -size => (length($settings_name)+5))
	. '</TD></TR></TABLE>'
	. "</TD>\n<TD>"
	. "<TABLE CELLPADDING=5><TR>"
	. '<TD VALIGN=TOP>'
	.$cgi->textarea(-name=>'save_settings_note',
			-value=>'<Add Description>',
			-rows=>2,
			-columns=>30)
	. '</TD></TR></TABLE>'
	. "</TD>"
	. "\n</TR>"
	. "\n<TR>"
	. "<TD COLSPAN=5>Change output name: "
	. $cgi ->  textfield(-name => 'save_dir_add', 
			     -override=>1,
			     -default => '', 
			     -size => 15
			     )
	. &help('change_output_name')
	."\n";

    $out .= "</TD></TR>\n</TABLE>\n";

    return $out;
}


sub var_input {
    
    my $var_name = shift;
    my $var_key = shift;
    my $type = shift;
    my $default = shift || $var_def_default{$var_key} || '';
    my $override = shift || 0;

    if (@_) {
	my @tmp = ($default, @_);
	$default = \@tmp;
    }

    unless (%added_parameters) {
	foreach (sort keys %work_files) {
#	    next if (defined $second_data_file{$_});
	    push @{$added_parameters{'work_files'}}, $_;
	}
	foreach (keys %work_files) {
#	    next if (defined $second_data_file{$_});
	    push @{$added_parameters{'work_files_base'}}, ($_, basename $_);
	}
	if ($last_column_used) {
	    $added_parameters{'last_column_used'} = $last_column_used;
	} else {
	    $added_parameters{'last_column_used'} = $FG_SIGNAL_TERM;
	}
    }

    if (defined $default
	and
	$default =~ /^_X__/) {
	my @tmp = split /_X__/, $default;
	my @tmp2 = ();
	foreach (@tmp) {
	    next unless (/\w/);
	    if (/^\@(.*)/) {
		push @tmp2, @{$added_parameters{$1}};
	    } elsif (/^\$(.*)/) {
		push @tmp2, $added_parameters{$1};
	    } else {
		push @tmp2, $_;
	    }
	}
	$default = \@tmp2;
	$var_def_default{$var_key} = $default;
    }

    if (defined $var_def_values{$var_key}
	and 
	$var_def_values{$var_key} =~ /^_X__/) {
	my @tmp = split /_X__/, $var_def_values{$var_key};
	my @tmp2 = ();
	foreach (@tmp) {
	    next unless (/\w/);
	    if (/^\@(.*)/) {
		if (defined @{$added_parameters{$1}}) {
		    push @tmp2, @{$added_parameters{$1}};
		}
	    } else {
		push @tmp2, $_;
	    }
	}
	$var_def_values{$var_key} = \@tmp2;
    }

    if (defined $var_def_labels{$var_key}
	and 
	$var_def_labels{$var_key} =~ /^_X__/) {
	my @tmp = split /_X__/, $var_def_labels{$var_key};
	my %tmp2 = ();
	foreach (@tmp) {
	    next unless (/\w/);
	    if (/^\%(.*)/) {
		if (defined @{$added_parameters{$1}}) {
		    %tmp2 = @{$added_parameters{$1}};
		    $var_def_labels{$var_key} = \%tmp2;
		    last;
		}
	    }
	}
    }

    if ($default eq 'force_empty') {
	$default = '';
    }

    if ($type eq 'header') {
	return '';
    } elsif ($type eq 'text') {
	return $cgi ->  textfield(-name => $var_name, 
				  -override => $override,
				  -default => $default, 
				  -size => $var_def_size{$var_key});
    } elsif ($type eq 'radio') {
	return $cgi -> radio_group(-name    => $var_name,
				   -override => $override,
				   -values  => $var_def_values{$var_key},
				   -default => $default);
    } elsif ($type eq 'check_single') {
	my $label = '';
	return $cgi -> checkbox(-name => $var_name,
				-override => $override,
				-value => 1,
				-checked => $default,
				-label => $label);

    } elsif ($type eq 'select') {

	if (defined $var_def_labels{$var_key}) {
	
	    return $cgi -> popup_menu(-name    => $var_name,
				      -override => $override,
				      -values  => $var_def_values{$var_key},
				      -default => $default,
				      -labels => $var_def_labels{$var_key});
	} else {
	    return $cgi -> popup_menu(-name    => $var_name,
				      -override => $override,
				      -values  => $var_def_values{$var_key},
				      -default => $default);
	}
    } elsif ($type eq 'list') {

	my @default = ($default);
	if (ref($default) eq 'ARRAY') {
	    @default = @$default;
	}

	if (defined $var_def_labels{$var_key}) {	    
	    return $cgi->scrolling_list(-name     => $var_name,
					-override => $override,
					-values   => $var_def_values{$var_key},
					-default  => [ @default ],
					-size     => $var_def_size{$var_key},
					-multiple => 'true',
					-labels => $var_def_labels{$var_key});
	} else {
	    return $cgi->scrolling_list(-name     => $var_name,
					-override => $override,
					-values   => $var_def_values{$var_key},
					-default  => [ @default ],
					-size     => $var_def_size{$var_key},
					-multiple => 'true');
	}
    } else {
	return "Unknown type for '$var_key': '$type'\n";
    }
}


sub list_files {

    my $out = '';
    my $out1 = '';
    my $out2 = '';
    my $out3 = '';

    # gather all the files to be skipped:
    foreach ($cgi->param('skip_files')) {
	$skip_files{$_}++;
    }

    # first the working files

    foreach my $file ($cgi->param('work_files')) {

	$work_files{$file}++ unless ($unload_files{$file});

	if (defined $cgi->param("file_specs$file")
	    and
	    $cgi->param("file_specs$file") =~ /\w/) {
	    my $specs = $cgi->param("file_specs$file");
	    $specs =~ s/_DQ_/\"/g;
	    my @tmp = split /\|/, $specs;
	    foreach my $spec (@tmp) {		
		my ($name, $value) = split /==/, $spec;
		if ($name eq 'FLAG') {
		    my @vals = split /,/, $value;
		    if (defined $cgi->param("${name}_$file")
			and
			$cgi->param("${name}_$file") =~ /\w/) {
			@vals = $cgi->param("${name}_$file");
		    }
		    push @{$file_specs{$file}{$name}}, @vals;
		} else {
		    if (defined $cgi->param("${name}_$file")
			and
			$cgi->param("${name}_$file") =~ /\w/) {
			$value = $cgi->param("${name}_$file");
		    }	
		    if ($name eq 'ADD') {			
			push @{$file_specs{$file}{$name}}, $value;
		    } else {
			$file_specs{$file}{$name} = $value;
		    }
		}
	    }
	} else {
	    my ($format_check) = &check_file_format($file, 'force');
	}
    }

    if (scalar keys %work_files <= 0
	and
	scalar @files <= 0
	) {
	return $out;
    }
    $out .= &pointer. "Please check the list of the loaded files and make sure the settings (particularly the <B>array layout</B> and the <B>column headers</B>) are correct.<BR>The first table contains fields to assign files to treatment groups and label them as biological or technical replicates.<BR>The second table contains the file content settings. After that proceed to the <A HREF=\"#ACTION SETUP\">action section</A>.<BR><BR>\n";
    $out1 = "<P>File annotation settings:<TABLE BORDER=1 CELLPADDING=5>\n<TR><TH>#</TH><TH>name</TH><TD><CENTER>"
	.&help('file_handle', '<B>alias</B>')
	."</CENTER></TD><TD><CENTER>"
	.&help('unload_box', '<B>unload</B>')
	."</CENTER></TD><TD><CENTER>"
	.&help('skip_file', '<B>skip</B>')
	."</CENTER></TD><TD><CENTER>"
	.&help('swap_channels', '<B>swap channels</B>')
	."</CENTER></TD><TD><CENTER>"
	.&help('ch_assign', '<B>CH1</B>')
	."</CENTER></TD><TD><CENTER>"
	.&help('ch_assign', '<B>CH2</B>')
	."</CENTER></TD><TD><CENTER>"
	.&help('bio_rep', '<B>biological</B>')
	."</CENTER></TD><TD><CENTER>"
	.&help('tech_rep', '<B>technical</B>')
	."</CENTER></TD></TR>\n";
    $out2 = "<P>File content settings:<BR><TABLE BORDER=1 CELLPADDING=5>\n<TR><TH>#</TH><TH>name</TH><TD><CENTER>"
	.&help('lines_column', '<B>info</B>')
	."</CENTER></TD><TD><CENTER>"
	.&help('array_layout','<B>array layout</B>')
	."</CENTER></TD><TD><CENTER>"
	.&help('column_headers', '<B>column headers</B>')
	."</CENTER></TD><TD><CENTER>"
	.&help('flag_column', '<B>flag</B>')
	."</CENTER></TD><TD><CENTER>"
	.&help('add_column', '<B>add</B>')
	."</CENTER></TD></TR>\n";
    $counter = 0;
    my @nonexist;
    my @nonread;

    for (my $i = 0; $i <= $#files; $i++) {
	my $file = $files[$i];

	unless (-e $file) {
	    push @nonexist, $file;
	    next;
	}
	unless (-r $file) {
	    push @nonread, $file;
	    next;
	}
	my ($format_check,$file_format) = &check_file_format($file, 'force');
	unless ($format_check =~ /^ok/) {
	    warn "# Invalid format for file '$file': '$format_check'";
	    next;
	}
	$work_files{$file}++ unless ($file_format eq 'probelynx_annotation');
    }

    my $prev_format = '';
    my $assign_counter = 0;

    foreach my $full_file (sort keys %work_files) {
 	if ($to_dir) {
	    $full_file =~ s/$from_dir/$to_dir/;
	}

	if (defined $first_data_file{$full_file}) {
	    # this one has been treated together with the first data part
	    next;
	}

	my %second_part = ();
	my $target = '';
	my $file = basename $full_file;	
	my $size = &commify((stat($full_file))[7]);

	unless (defined $file_specs{$full_file}{'HEADER'}) {
	    my ($message) = &check_file_format($full_file,'force');
	}

	if ($file_specs{$full_file}{'FORMAT'} eq 'ImaGene') {
	    # unless we have a file for each label, skip it
	    if (defined $second_data_file{$full_file}) {
		# it seems that the first one is defined
		unless (grep /$second_data_file{$full_file}/, (keys %work_files)) {
		    warn "# KH: First file for '$full_file' not listed!";
		    $wrong_format_files{$full_file} = "# KH: First file for '$full_file' not listed!";
		    undef $work_files{$full_file};
		}
		# nothing left to do with this file
		next;
	    } else {
		# make sure we have the second data file for this
		$target = $full_file;
		if ($full_file =~ /(.*)532/) {
		    $target = $1.'635';
		} elsif ($full_file =~ /(.*)635/) {
		    $target = $1.'532';
		} elsif ($full_file =~ /(.*cy)3/i) {
		    $target = $1.'5';
		} elsif ($full_file =~ /(.*cy)5/i) {
		    $target = $1.'3';
		} else {
		    warn "# KH: No label identifier found in $full_file, please include 'Cy3' or 'Cy5' in the file name!";
		    $wrong_format_files{$full_file} = "# KH: No label identifier found in $full_file, please include 'Cy3' or 'Cy5' in the file name!";
		    undef $work_files{$full_file};
		    next;
		}
		my @targets = `ls $target*`;
		if (@targets == 0) {
		    warn "# KH: no file matching $target*!";
		    $wrong_format_files{$full_file} = "# KH: no file matching $target*!";
		    undef $work_files{$full_file};
		} elsif (@targets > 1) {
		    warn "# KH: ".(scalar @targets)." files matching $target*!";
		    $wrong_format_files{$full_file} = "# KH: ".(scalar @targets)." file matching $target*!";
		    undef $work_files{$full_file};
		} else {
		    $target = $targets[0];
		    chomp $target;
		}

		for my $rest_files (keys %work_files) {
		    if ($rest_files eq $target) {
			$second_data_file{$full_file} = $target;
			$first_data_file{$target} = $full_file;
			last;
		    }
		}
		unless (defined $second_data_file{$full_file}) {
		    warn "# KH: Second file for '$full_file' not listed!";
		    $wrong_format_files{$full_file} = "# KH: Second file for '$full_file' not listed!";
		    undef $work_files{$full_file};
		    next;
		}
		# get specs for second file:
		$second_part{'full_file'} = $target;
		$second_part{'file'} = basename $target;	
		$second_part{'size'} = &commify((stat($target))[7]);
		
		unless (defined $file_specs{$target}{'HEADER'}) {
		    my ($message, $format) = &check_file_format($target,'force');
		    unless ($message eq 'ok') {
			die "# KH: second data file not valid!";
		    }
		    unless ($format eq 'ImaGene') {
			die "# KH: second data file not in ImaGene format ($format)!";
		    }
		}
	    }
	}

	my $lines = "$file_specs{$full_file}{'LINES'}\&nbsp;+\&nbsp;$file_specs{$full_file}{'HEADER'}";
	if ($file_specs{$full_file}{'COMMENTS'}) {
	    $lines .= " + $file_specs{$full_file}{'COMMENTS'}";
	}
	if (%second_part) {
	    $lines .= "<BR>$file_specs{$second_part{'full_file'}}{'LINES'}\&nbsp;+\&nbsp;$file_specs{$second_part{'full_file'}}{'HEADER'}";
	    if ($file_specs{$second_part{'full_file'}}{'COMMENTS'}) {
		$lines .= " + $file_specs{$second_part{'full_file'}}{'COMMENTS'}";
	    }
	    $size .= "<BR>".(&commify((stat($second_part{'full_file'}))[7]));
	}

	# get file format and organization
	my $format = $file_specs{$full_file}{'FORMAT'};
	my $organization = 'by-row';
	if ($format eq 'ScanAlyze'
	    or
	    $format eq 'ImaGene'
	    or
	    $format eq 'GenePix') {
	    $organization = 'by-subgrid';
	}
	if (defined $cgi->param("ORGANIZATION_$full_file")) {
	    $organization = $cgi->param("ORGANIZATION_$full_file");
	} elsif (defined $file_specs{$full_file}{'ORGANIZATION'}) {
	    $organization = $file_specs{$full_file}{'ORGANIZATION'};
	}
	$file_specs{$full_file}{'ORGANIZATION'} = $organization;

	my ($mrow,$mcol,$row,$col) = ($file_specs{$full_file}{'MROW'},
				      $file_specs{$full_file}{'MCOL'},
				      $file_specs{$full_file}{'ROW'},
				      $file_specs{$full_file}{'COL'});
	
	my $grid = "<TABLE><TR><TD>MROW:</TD><TD><input type=\"text\" name=\"MROW_$full_file\" value=\"$mrow\" size=3></TD></TR><TR><TD>MCOL:</TD><TD><input type=\"text\" name=\"MCOL_$full_file\" value=\"$mcol\" size=3></TD></TR><TR><TD>ROW:</TD><TD><input type=\"text\" name=\"ROW_$full_file\" value=\"$row\" size=3></TD></TR><TR><TD>COL:</TD><TD><input type=\"text\" name=\"COL_$full_file\" value=\"$col\" size=3></TD></TR></TABLE>";
	my @headers = ();
	if ($file_specs{$full_file}{'HEADER'} == 1) {
	    open (IN, $full_file)
		or die "Can't open file '$full_file': $!";
	    my $line = <IN>;
	    while ($line =~ /^\#/
		   or $line eq '') {
		$line = <IN>;
	    }
	    
	    if (defined $file_specs{$full_file}{'START_LINE'}) {
		until ($line =~ /$file_specs{$full_file}{'START_LINE'}/) {
		    $line = <IN>;
		    last if (eof(IN));
		}
		$line = <IN>;
	    }

	    chomp $line;
	    $line =~ s/\cM$//;

	    foreach (split /\t/, $line) {
		# remove leading and trailing white-space (DANGEROUS!)
		s/^\s*//;
		s/\s*$//;
		s/  / /g;
		push @headers, $_;
	    }

	    close IN;
	} elsif ($file_specs{$full_file}{'HEADER'} == 2) {
	    my %headers = ();
	    # match up labels with headers
	    open (IN, $full_file)
		or die "Can't open file '$full_file': $!";
	    my $line = <IN>;
	    while ($line =~ /^\#/
		   or $line eq '') {
		$line = <IN>;
	    }
	    chomp $line;
	    $line =~ s/\cM$//;

	    # labels first
	    my @labels = split /\t/, $line;
	    foreach (@labels) {
		s/^(Ag|Au).*/$1/;
		s/^(Cy3|Cy5).*/$1/;
	    }
	    
	    # headers next
	    $line = <IN>;
	    while ($line =~ /^\#/
		   or $line eq '') {
		$line = <IN>;
	    }
	    chomp $line;
	    $line =~ s/\cM$//;
	    my @tmp = split /\t/, $line;

#	    foreach (@headers) {
#		my $label = shift @labels;
#		if ($label =~ /\w/) {
#		    $_ = "$label $_";
#		}
#	    }

	    @headers = ();
	    foreach (@tmp) {
		# remove leading and trailing white-space and collaps multiple blanks (DANGEROUS!)
		s/^\s*//;
		s/\s*$//;
		s/  / /g;
		$headers{$_}++;
		if ($headers{$_} == 1) {
		    next unless (/\w/);
		    push @headers, $_;
		}
	    }
	    
	    
	    close IN;
	} else {
	    die "Can't deal with $file_specs{$full_file}{'HEADER'} headers!";
	}
	
	# Spot id column
	my $selected = 0;
	my $spot_id = "<TR><TD><nobr>Spot ID:</nobr></TD><TD><SELECT name=\"SPOT_ID_$full_file\"><OPTION>- undefined -</OPTION>\n";
	my $default = '';
	if (defined $cgi->param("SPOT_ID_$full_file")
	    and
	    $cgi->param("SPOT_ID_$full_file") =~ /\w/) {
	    $default = $cgi->param("SPOT_ID_$full_file");
	    $file_specs{$full_file}{'SPOT_ID'} = $default;
	} elsif (defined $file_specs{$full_file}{'SPOT_ID'}) {
	    $default = $file_specs{$full_file}{'SPOT_ID'};
	} elsif ($format eq 'ArrayPipe') {
	    $default = $SPOT_ID_TERM;
	}

	foreach (@headers) {
	    my $select = '';
	    if ($default) {
		# default overrules
		if ($_ eq $default) {
		    $select = ' SELECTED';
		}
	    } else {
		if ($format eq 'ArrayVision'
		    and
		    $_ eq 'Spot labels') {
		    $select = ' SELECTED';
		} elsif ($format eq 'unknown'
			 and
			 ($_ =~ /^gene id$/i
			  or
			  $_ eq 'Spot labels'
			  or
			  $_ eq 'Gene Name'
			  or
			  $_ eq 'Well ID'
			  or
			  $_ eq 'ID_REF'
			  or
			  $_ eq 'NAME'
			  or
			  $_ eq 'ESTs'
			  )
			 ) {
		    $select = ' SELECTED';
		}
	    }
	    if ($select
		and
		$selected == 0) {
		$selected = 1;
		$file_specs{$full_file}{'SPOT_ID'} = $_;
	    }
	    $spot_id .= "<OPTION$select>$_</OPTION>\n";
	}
	$spot_id .= "</SELECT>\n</TD></TR>\n";

	# Channel 1 foreground column
	$selected = 0;
	my $ch1_fg = "<TR><TD>\n<nobr>CH1 FG:</nobr></TD><TD><SELECT name=\"CH1FG_$full_file\"><OPTION>?</OPTION>\n";
	$default = '';
	if (defined $cgi->param("CH1FG_$full_file")
	    and
	    $cgi->param("CH1FG_$full_file") =~ /\w/) {
	    $default = $cgi->param("CH1FG_$full_file");
	    $file_specs{$full_file}{'CH1FG'} = $default;
	} elsif (defined $file_specs{$full_file}{'CH1FG'}) {
	    $default = $file_specs{$full_file}{'CH1FG'};
	}
#	warn "#KH: default: $default";

	foreach (@headers) {	    
	    my $select = '';
	    if ($default) {
		if ($_ eq $default) {
		    $select = ' SELECTED';
		}
	    } else {
		if ($format eq 'ArrayVision'
		    and
		    $_ eq 'MTM Dens - Levels') {
		    $select = ' SELECTED';
		} elsif ($format eq 'unknown'
			 and
			 ($_ eq 'Ctrl'
			  or
			  $_ eq 'Control'
			  or
			  $_ eq 'PBS'
			  or
			  $_ =~ /CY? ctlr Signal/
			  or
			  $_ eq 'Ag MTM Dens - Levels'
			  or
			  $_ eq 'Ctrl MTM Dens - Levels'
			  or
			  $_ eq 'CTRL-MTM'
			  or
			  $_ eq 'Cntl_MTM Dens'
			  or
			  $_ eq 'ch1.Intensity'
			  or		      
			  $_ eq 'CH1I')
			 ) {
		    $select = ' SELECTED';
		} elsif ($format eq 'ArrayPipe'
			 and 
			 $_ =~ /^CH1 (${FG_SIGNAL_TERM}_?M?)$/) {
		    $select = ' SELECTED';
		}
	    }
	    if ($select
		and
		$selected == 0) {
		$selected = 1;
		$file_specs{$full_file}{'CH1FG'} = $_;
	    }
	    $ch1_fg .= "<OPTION$select>$_</OPTION>\n";
	}
	$ch1_fg .= "</SELECT>\n</TD></TR>\n";

	# Channel 2 foreground column
	$selected = 0;
	my $ch2_fg = "<TR><TD>\n<nobr>CH2 FG:</TD><TD><SELECT name=\"CH2FG_$full_file\"><OPTION>?</OPTION>\n";
	$default = '';
	if (defined $cgi->param("CH2FG_$full_file")
	    and
	    $cgi->param("CH2FG_$full_file") =~ /\w/) {
	    $default = $cgi->param("CH2FG_$full_file");
	    $file_specs{$full_file}{'CH2FG'} = $default;
	} elsif (defined $file_specs{$full_file}{'CH2FG'}) {
	    $default = $file_specs{$full_file}{'CH2FG'};
	}
#	warn "#KH: default: '$default'";
	foreach (@headers) {
	    my $select = '';
	    if ($default) {
		if ($_ eq $default) {
		    $select = ' SELECTED';
		}
	    } else {
		if ($format eq 'ArrayVision'
		    and
		    $_ eq 'MTM Dens - Levels') {
		    $select = ' SELECTED';
		} elsif ($format eq 'unknown'
			 and
			 ($_ eq 'Signal'
			  or
			  $_ eq 'Data'
			  or
			  $_ eq 'Sample'
			  or
			  $_ eq 'Rota'
			  or
			  $_ eq 'Corona'
			  or
			  $_ eq 'Sig MTM Dens - Levels'
			  or
			  $_ eq 'Au MTM Dens - Levels'
			  or
			  $_ eq 'MTM Dens - Levels-data'
			  or
			  $_ eq 'Data_MTM Dens'
			  or
			  $_ eq 'ch2.Intensity'
			  or		      
			  $_ =~ /CY? exp Signal/
			  or
			  $_ eq 'CH2I')
			 ) {
		    $select = ' SELECTED';
		} elsif ($format eq 'ArrayPipe'
			 and 
			 $_ =~ /^CH2 (${FG_SIGNAL_TERM}_?M?)$/) {
		    $select = ' SELECTED';
		}
	    }
	    if ($select
		and
		$selected == 0) {
		$selected = 1;
		$file_specs{$full_file}{'CH2FG'} = $_;
	    }
	    $ch2_fg .= "<OPTION$select>$_</OPTION>\n";
	}
	$ch2_fg .= "</SELECT>\n</TD></TR>\n";

	# Channel 1 background column
	my $ch1_bg = "<TR><TD>\n<nobr>CH1 BG:</nobr></TD><TD><SELECT name=\"CH1BG_$full_file\"><OPTION>?</OPTION>\n";
	$default = '';
	if (defined $cgi->param("CH1BG_$full_file")
	    and
	    $cgi->param("CH1BG_$full_file") =~ /\w/) {
	    $default = $cgi->param("CH1BG_$full_file");
	    $file_specs{$full_file}{'CH1BG'} = $default;
	} elsif (defined $file_specs{$full_file}{'CH1BG'}) {
	    $default = $file_specs{$full_file}{'CH1BG'};
	}
	foreach (@headers) {
	    my $select = '';
	    if ($default) {
		if ($_ eq $default) {
		    $select = ' SELECTED';
		}
	    } else {
		if ($format eq 'ArrayVision'
		    and
		    $_ eq 'Bkgd') {
		    $select = ' SELECTED';
		} elsif ($format eq 'unknown'
			 and
			 ($_ =~ /^ctrl.?bk?gd?$/i
			  or
			  $_ eq 'Control Bkgd'
			  or
			  $_ eq 'Bkgd-ctrl'
			  or
			  $_ eq 'ch1.Background'
			  or		      
			  $_ eq 'CY5 BG'
			  or
			  $_ eq 'Ag Bkgd'
			  or
			  $_ eq 'Bkgd_ctrl'
			  or
			  $_ eq 'Bkgd-ctrl'
			  or
			  $_ eq 'Cntl_Bkgd'
			  or
			  $_ eq 'CH1B')
			 ) {
		    $select = ' SELECTED';
		} elsif ($format eq 'ArrayPipe'
			 and 
			 $_ =~ /^(\w+) $BG_SIGNAL_TERM$/) {
		    $select = ' SELECTED';
		}
	    }
	    if ($select
		and
		$selected == 0) {
		$selected = 1;
		$file_specs{$full_file}{'CH1BG'} = $_;
	    }
	    $ch1_bg .= "<OPTION$select>$_</OPTION>\n";
	}
	$ch1_bg .= "</SELECT>\n</TD></TR>\n";

	# Channel 2 background column
	my $ch2_bg = "<TR><TD>\n<nobr>CH2 BG:</nobr></TD><TD><SELECT name=\"CH2BG_$full_file\"><OPTION>?</OPTION>\n";
	$default = '';
	if (defined $cgi->param("CH2BG_$full_file")
	    and
	    $cgi->param("CH2BG_$full_file") =~ /\w/) {
	    $default = $cgi->param("CH2BG_$full_file");
	    $file_specs{$full_file}{'CH2BG'} = $default;
	} elsif (defined $file_specs{$full_file}{'CH2BG'}) {
	    $default = $file_specs{$full_file}{'CH2BG'};
	}
	foreach (@headers) {
	    my $select = '';
	    if ($default) {
		if ($_ eq $default) {
		    $select = ' SELECTED';
		}
	    } else {
		if ($format eq 'ArrayVision'
		    and
		    $_ eq 'Bkgd') {
		    $select = ' SELECTED';
		} elsif ($format eq 'unknown'
			 and
			 ($_ eq 'Signal-bgd'
			  or
			  $_ =~ /^data.?bk?gd?$/i
			  or
			  $_ eq 'Sample Bkgd'
			  or
			  $_ eq 'ch2.Background'
			  or		      
			  $_ eq 'Signal-bg'
			  or
			  $_ eq 'Sig Bkgd'
			  or
			  $_ eq 'Au Bkgd'
			  or
			  $_ eq 'Bkgd'
			  or
			  $_ eq 'Bkgd-data'
			  or
			  $_ eq 'Bkgd_data'
			  or
			  $_ eq 'Data_Bkgd'
			  or
			  $_ eq 'CY3 BG'
			  or
			  $_ eq 'CH2B')
			 ) {
		    $select = ' SELECTED';
		} elsif ($format eq 'ArrayPipe'
			 and 
			 $_ =~ /^(\w+) $BG_SIGNAL_TERM$/) {
		    $select = ' SELECTED';
		}
	    }
	    if ($select
		and
		$selected == 0) {
		$selected = 1;
		$file_specs{$full_file}{'CH2BG'} = $_;
	    }
	    $ch2_bg .= "<OPTION$select>$_</OPTION>\n";
	}
	$ch2_bg .= "</SELECT>\n</TD></TR>\n";

	# Flag column(s)
	my @default = ();
	my %default = ();
	my $flag_warning = 0;
	if ($cgi->param("FLAG_$full_file")) {
	    @{$file_specs{$full_file}{'FLAG'}} = ();
	    @default = $cgi->param("FLAG_$full_file");
	    foreach (@default) {
		if (/\w/
		    or
		    /\?/) {
		    $default{$_}++;
		}
	    }
	    foreach (keys %default) {
		push @{$file_specs{$full_file}{'FLAG'}}, $_;
	    }
	} elsif (defined $file_specs{$full_file}{'FLAG'}) {
	    @default = @{$file_specs{$full_file}{'FLAG'}};
	    foreach (@default) {
		if (/\w/
		    or
		    /\?/) {
		    $default{$_}++;
		}
	    }
	}	    

	my $flag_channel = "<SELECT name=\"FLAG_$full_file\" size=\"4\" multiple><OPTION"
	    . (defined $default{'- undefined -'} ? ' SELECTED' : '')
	    . ">- undefined -</OPTION>\n";

	foreach (@headers) {
	    my $select = '';
	    if (%default) {
		if ($default{$_}) {
		    $select = ' SELECTED';
		}
	    } elsif (/^\"?flag$/i) {
		$select = ' SELECTED';
		$flag_warning++;
		unless (defined $file_specs{$full_file}{'FLAG'}) {
		    push @{$file_specs{$full_file}{'FLAG'}}, $_;
		}
	    } elsif ($format eq 'unknown'
		     and
		     ($_ eq 'FLAG'
		      or
		      $_ eq 'Spot Quality'
		      or
		      $_ eq 'Flags')) {
		$select = ' SELECTED';
		push @{$file_specs{$full_file}{'FLAG'}}, $_;
	    } elsif ($format eq 'ArrayPipe'
		     and 
		     ($_ =~ /^(\w+) $FLAG_CHANNEL_TERM$/
		     or
		     $_ eq $FLAG_CHANNEL_TERM)
		     ) {
		$select = ' SELECTED';
		push @{$file_specs{$full_file}{'FLAG'}}, $_;
	    } elsif ($format eq 'TMEV'
		     and 
		     ($_ =~ /^Flag(A|B)$/)
		     ) {
		$select = ' SELECTED';
		push @{$file_specs{$full_file}{'FLAG'}}, $_;
	    }
	    $flag_channel .= "<OPTION$select>$_</OPTION>\n";
	}
	$flag_channel .= "</SELECT>\n";
	if ($flag_warning > 1) {
	    $flag_channel .= "<BR><FONT COLOR=RED><B>WARNING:</B></FONT> found $flag_warning flag columns!\n";
	}

	# Additional column(s)
	@default = ();
	%default = ();
	if ($cgi->param("ADD_$full_file")) {
	    @{$file_specs{$full_file}{'ADD'}} = ();
	    @default = $cgi->param("ADD_$full_file");
	    foreach (@default) {
		if (/\w/
		    or
		    /\?/) {
		    $default{$_}++;
		}
	    }
	    foreach (keys %default) {
		next if ($_ eq '- none -');
		push @{$file_specs{$full_file}{'ADD'}}, $_;
	    }
	} elsif (defined $file_specs{$full_file}{'ADD'}) {
	    @default = @{$file_specs{$full_file}{'ADD'}};
	    foreach (@default) {
		if (/\w/
		    or
		    /\?/) {
		    $default{$_}++;
		}
	    }
	}	    

	my $add_columns = "<SELECT name=\"ADD_$full_file\" size=\"4\" multiple><OPTION";
	if ($default{'- none -'}
	    or
	    @default == 0) {
	    $add_columns .= ' SELECTED';
	}
	    
	$add_columns .= ">- none -</OPTION>\n";

	foreach (@headers) {
	    my $select = '';
	    if (%default) {
		if ($default{$_}) {
		    $select = ' SELECTED';
		}
#	    } elsif (/^\"?Description$/i) {
#		$select = ' SELECTED';
#		unless (defined $file_specs{$full_file}{'ADD'}) {
#		    push @{$file_specs{$full_file}{'ADD'}}, $_;
#		}
	    }
	    $add_columns .= "<OPTION$select>$_</OPTION>\n";
	}
	$add_columns .= "</SELECT>\n";

	$counter++;
	my $swap_checked = '';
	if ($channel_swaps{$full_file}) {
	    $swap_checked = ' CHECKED';
	}
	if (defined $cgi->param("CHANNEL_SWAP_$full_file")
	    and
	    $cgi->param("CHANNEL_SWAP_$full_file") eq '1') {
	    $swap_checked = ' CHECKED';
	}
	    

	my $skip_checked = '';
	if ($skip_files{$full_file}) {
	    $skip_checked = ' CHECKED';
	}


	# Channel annotation:
	if ($prev_format eq '') {
	    $prev_format = $format;
	} else {
	    if ($prev_format ne $format) {
		$assign_counter++;
		$prev_format = $format;
	    }
	}

#	my $CH1_default = 'C'.$assign_counter;
	my $CH1_default = 'C';
	my $CH2_default = 'T'.$assign_counter;
	if ($assign_counter == 0) {
	    $CH2_default = 'T';
	}

	if (defined $cgi->param("CH1_ANNOTATE_$full_file")) {
	    $CH1_default = $cgi->param("CH1_ANNOTATE_$full_file");
	}
	$file_specs{$full_file}{'CH1_ANNOTATE'} = $CH1_default;

	if (defined $cgi->param("CH2_ANNOTATE_$full_file")) {
	    $CH2_default = $cgi->param("CH2_ANNOTATE_$full_file");
	}
	$file_specs{$full_file}{'CH2_ANNOTATE'} = $CH2_default;
	

	my $tech_rep_default = '';
	if (defined $cgi->param("TECH_REP_$full_file")) {
	    $tech_rep_default = $cgi->param("TECH_REP_$full_file");
	}
	$file_specs{$full_file}{'TECH_REP'} = $tech_rep_default;
	
	my $bio_rep_default = '';
	if (defined $cgi->param("BIO_REP_$full_file")) {
	    $bio_rep_default = $cgi->param("BIO_REP_$full_file");
	}
	$file_specs{$full_file}{'BIO_REP'} = $bio_rep_default;
	

	# Channel labelling:
	my $CH1_label_default = 'Cy3';
	if ($format eq 'ArrayVision') {
	    $CH1_label_default = 'Ag';
	} elsif (
		 $mrow == 12
		 and
		 $mcol == 4
		 and
		 $row == 18
		 and
		 $col == 19) {
	    # Pyxis file, very likely to be scanned with ArrayVision,
	    # but maybe not in ArrayVision format
	    $CH1_label_default = 'Ag';
	}

	# channels should be labeled manually, if not the same as the default;
	# the labelling should not be switched if a dye swap is specified;
	# leave this up to the user;
	# several combinations are possible:
	# data column 1 = CH1 = Cy3 (default)
	# data column 1 = CH1 = Cy5 (manual label swap)
	# data column 1 = CH2 = Cy3 (data channel swap)
	# data column 1 = CH2 = Cy5 (data channel and label swap)

#	if ($channel_swaps{$full_file}) {
#	    if ($format eq 'ArrayVision') {
#		$CH1_label_default = 'Au';
#	    } else {
#		$CH1_label_default = 'Cy3';
#	    }
#	} els
	if (defined $cgi->param("CH1_LABEL_$full_file")) {
	    $CH1_label_default = $cgi->param("CH1_LABEL_$full_file");
	}
	$file_specs{$full_file}{'CH1_LABEL'} = $CH1_label_default;
	
	my $CH1_label = "<SELECT name=\"CH1_LABEL_$full_file\">\n";
	foreach (@known_channel_labels) {
	    my $select = '';
	    if ($CH1_label_default eq $_) {
		$select = ' SELECTED';
	    }
	    $CH1_label .= "<OPTION$select>$_</OPTION>\n";
	}
	$CH1_label .= "</SELECT>\n";
	

	my $CH2_label_default = 'Cy5';
	if ($format eq 'ArrayVision') {
	    $CH2_label_default = 'Au';
	} elsif (
		 $mrow == 12
		 and
		 $mcol == 4
		 and
		 $row == 18
		 and
		 $col == 19) {
	    # Pyxis file, very likely to be scanned with ArrayVision,
	    # but maybe not in ArrayVision format
	    $CH2_label_default = 'Au';
	}

#	if ($channel_swaps{$full_file}) {
#	    if ($format eq 'ArrayVision') {
#		$CH2_label_default = 'Ag';
#	    } else {
#		$CH2_label_default = 'Cy5';
#	    }
#	} els
	if (defined $cgi->param("CH2_LABEL_$full_file")) {
	    $CH2_label_default = $cgi->param("CH2_LABEL_$full_file");
	}

	$file_specs{$full_file}{'CH2_LABEL'} = $CH2_label_default;

	my $CH2_label = "<SELECT name=\"CH2_LABEL_$full_file\">\n";
	foreach (@known_channel_labels) {
	    my $select = '';
	    if ($CH2_label_default eq $_) {
		$select = ' SELECTED';
	    }
	    $CH2_label .= "<OPTION$select>$_</OPTION>\n";
	}
	$CH2_label .= "</SELECT>\n";


	# provide link to input file:
	my $file_link = basename($full_file);
#	$file_link =~ s/$tmp_dir/$tmp_dir_html/;
	if (-r $file_link) {
	    $file_link = "<A HREF=\"$file_link\" TARGET=\"_file\">$file</A>";
	} else {
	    $file_link = $file_link;
	}
	if (%second_part) {
	    my $add_on = $second_part{'full_file'};
	    $file_link = 'CH1:&nbsp;'.$file_link;
	    if ($FPMI_specific
		and
		$add_on =~ /examples/) {
		# not so good to allow download of example file!
		$add_on = $second_part{'file'};
	    } else {
		$add_on = "<A HREF=\"$add_on\" TARGET=\"_file\">$second_part{'file'}</A>";
	    }
	    $file_link .= "<BR>CH2:&nbsp;$add_on";
	}

	# provide shortcut to input file:
	my $file_handle = basename($full_file);
	if (%second_part) {
	    my $add_on = basename $second_part{'file'};
	    # remove the common stem:
	    my @letters1 = split / */, $file_handle;
	    my @letters2 = split / */, $add_on;	    
	    my $l1 = shift @letters1;
	    my $l2 = shift @letters2;
	    my $stem = '';
	    while ($l1 eq $l2) {
		$stem .= $l1;
		$l1 = shift @letters1;
		$l2 = shift @letters2;
	    }

#	    unshift @letters1, $l1;
#	    unshift @letters2, $l2;
	
	    # remove the common suffix
	    my $tail = '';
	    $l1 = pop @letters1;
	    $l2 = pop @letters2;

	    while ($l1 eq $l2) {
		$tail = $l1.$tail;
		$l1 = pop @letters1;
		$l2 = pop @letters2;
#		last if ($l1 eq '.');
		last unless (@letters1);
	    }
	    unless ($l1 eq '.') {
		push @letters1, $l1;
		push @letters2, $l2;
		
	    }
	    # define new file_handle:
	    $file_handle = $stem . (join '', @letters1) . (join '', @letters2) . $tail;
	    $file_handle =~ s/_Cy\b//;
	    $file_handle =~ s/_Cy_/_/;
	}

	# this might be overwritten by a CGI parameter:
	if (defined $cgi->param("ALIAS_$full_file")) {
	    $file_handle = $cgi->param("ALIAS_$full_file");
	}
	my $file_handle_box = "<input type=\"text\" name=\"ALIAS_$full_file\" value=\"$file_handle\" size=".( (length $file_handle) + 3).">";


	my $unload_file = $full_file;
	if (%second_part) {
	    $unload_file .= '__second_part__'.$target;
	}
	
	my $skip_file = $full_file;
	
	my $organization_out = "<SELECT name=\"ORGANIZATION_$full_file\">\n";
	foreach my $choice ('by-row', 'by-subgrid') {
	    my $select = '';
	    if ($choice eq $organization) {
		$select = ' SELECTED';
	    }
	    $organization_out .= "<OPTION$select>$choice</OPTION>\n";
	}
	$organization_out .= "</SELECT>\n";


	$out2 .= "<TR>
<TD>$counter</TD>
<TD>$file_link</TD>
<TD ALIGN=LEFT><B>format:</B><BR>$format<BR><B>size:</B><BR>$size<BR><B>lines:</B><BR>$lines<BR>"
. &help(
	{
	    'topic' => 'data_file_organization',
	    'label' => 'organization',
	    'no_small' => 1
	    }
	)
  .":<BR>$organization_out</TD>
<TD>$grid</TD>
<TD><TABLE>\n$spot_id\n$ch1_fg\n$ch1_bg\n$ch2_fg\n$ch2_bg\n</TABLE></TD>
<TD>$flag_channel</TD>
<TD>$add_columns</TD>
</TR>\n";

	my $text_length1 = 3;
	my $text_length2 = 3;
	if (length($CH1_default) > $text_length1) {
	    $text_length1 = length($CH1_default);
	}
	if (length($CH2_default) > $text_length2) {
	    $text_length2 = length($CH2_default);
	}
	
        $out1 .= "<TR>
<TD>$counter</TD>
<TD>$file_link</TD>
<TD>$file_handle_box</TD>
<TD ALIGN=CENTER><input type=\"checkbox\" name=\"unload_files\" value=\"$unload_file\"></TD>
<TD ALIGN=CENTER><input type=\"checkbox\" name=\"skip_files\" value=\"$skip_file\"$skip_checked></TD>
<TD ALIGN=CENTER><input type=\"checkbox\" name=\"swap_channels\" value=\"$full_file\"$swap_checked></TD>
<TD ALIGN=CENTER><TABLE><TR><TD>Label:</TD><TD>$CH1_label</TD></TR><TR><TD>Annotation:</TD><TD><input type=\"text\" name=\"CH1_ANNOTATE_$full_file\" value=\"$CH1_default\" size=$text_length1></TD></TR></TABLE></TD>
<TD ALIGN=CENTER><TABLE><TR><TD>Label:</TD><TD>$CH2_label</TD></TR><TR><TD>Annotation:</TD><TD><input type=\"text\" name=\"CH2_ANNOTATE_$full_file\" value=\"$CH2_default\" size=$text_length2></TD></TR></TABLE></TD><TD><input type=\"text\" name=\"BIO_REP_$full_file\" value=\"$bio_rep_default\" size=3></TD><TD><input type=\"text\" name=\"TECH_REP_$full_file\" value=\"$tech_rep_default\" size=3></TD></TR>\n";

	if (%second_part) {
	    # a second file is part of this data set
	    # get specs from first file
	    $file_specs{$target}{'MROW'} = $file_specs{$full_file}{'MROW'};
	    $file_specs{$target}{'MCOL'} = $file_specs{$full_file}{'MCOL'};
	    $file_specs{$target}{'ROW'} = $file_specs{$full_file}{'ROW'};
	    $file_specs{$target}{'COL'} = $file_specs{$full_file}{'COL'};
	    @{$file_specs{$target}{'FLAG'}} = @{$file_specs{$full_file}{'FLAG'}};
	    $file_specs{$target}{'CH1FG'} = $file_specs{$full_file}{'CH1FG'};
	    $file_specs{$target}{'CH1BG'} = $file_specs{$full_file}{'CH1BG'};
	    $file_specs{$target}{'CH2FG'} = $file_specs{$full_file}{'CH2FG'};
	    $file_specs{$target}{'CH2BG'} = $file_specs{$full_file}{'CH2BG'};
	    $file_specs{$target}{'SPOT_ID'} = $file_specs{$full_file}{'SPOT_ID'};
	}
    }

    # combine all output parts:
    $out .= "$out1</TABLE>\n$out2</TABLE>\n$out3\n";
    
    $out .= '<BR><BR>'.$cgi -> submit(-name=>'submit', 
				    -value=>'Unload Files!');
    $out .= '&nbsp;'.$cgi -> submit(-name=>'submit', 
				    -value=>'Unload Files and Do it!')
	    . &help('unload_files_and_do_it');

    if (@nonexist) {
	$out .= "<BR>Error - file(s) don't exist:<BR>\n";
	foreach (@nonexist) {
	    $out .= "$_<BR>\n";
	}
    }
    if (@nonread) {
	$out .= "<BR>Error - file(s) not readable:<BR>\n";
	foreach (@nonread) {
	    $out .= "$_<BR>\n";
	}
    }
    foreach (sort keys %work_files) {
	$out .= "<input type=\"hidden\" name=\"work_files\" value=\"$_\">\n";
	my @tmp = ();
	foreach my $key (keys %{$file_specs{$_}}) {
	    my $val = '';
	    if ($key eq 'FLAG'
		and
		defined $file_specs{$_}{$key}) {
		$val = join ',', @{$file_specs{$_}{$key}};
	    } else {
		$val = $file_specs{$_}{$key};
	    }
	    push @tmp, "$key==$val" if (defined $val);
	}
	my $specs = join '|', @tmp;
	$specs =~ s/\"/_DQ_/g;
	$out .= "<input type=\"hidden\" name=\"file_specs$_\" value=\"$specs\">\n";
    }

    # remove skipped files from %work_files
    # skip all files that are ticked for skipping:
    foreach (keys %skip_files) {
        delete $work_files{$_};
#        $out .= "<input type=\"hidden\" name=\"skip_files\" value=\"$_\">\n";
    }
    
    # remove the second parts of multi-file data sets from %work_files
    foreach (keys %first_data_file) {
#	my $second_part = $first_data_file{$_};
#	delete $work_files{$second_part};
	delete $work_files{$_};
    }
    

    # then the reporter lists:

    foreach ($cgi->param('work_lists')) {
	$work_lists{$_}++ unless ($unload_files{$_});
	$work_lists{$_}++ unless ($skip_files{$_});
    }
    if (scalar keys %work_lists <= 0
	and
	scalar @reporter_lists <= 0
	) {
	return $out."<BR>\n";
    }
    $out .= "\n<H2><BR>Working Lists</H2>\n<TABLE BORDER=1 CELLPADDING=5>\n<TR><TH ROWSPAN=2>#</TH><TH ROWSPAN=2>name</TH><TH ROWSPAN=2>size (Bytes)</TH><TD ROWSPAN=2><CENTER>".&help('lines_column', '<B>lines</B>')."</CENTER></TD><TD COLSPAN=".(scalar keys %work_files)."><CENTER>".&help('matches_column', '<B>matches</B>')."</CENTER></TD>";
    $out .= "<TD ROWSPAN=2><CENTER>".&help('flag_box', '<B>flag</B>')."</CENTER></TD>\n";
    $out .= "<TD ROWSPAN=2><CENTER>".&help('unload_box', '<B>unload</B>')."</CENTER></TD></TR>\n";

    $out .= "<TR>";
    $counter = 0;
    foreach my $data_file (sort keys %work_files) {
	unless (defined $data{$data_file}) {
	    %{$data{$data_file}} = ();
	}
	$counter++;
	$out .= "<TH>file\&nbsp;$counter</TH>";
    }
    $out .= "</TR>\n";

    $counter = 0;
    @nonexist = ();
    @nonread = ();

    foreach (@reporter_lists) {
	unless (-e $_) {
	    push @nonexist, $_;
	    next;
	}
	unless (-r $_) {
	    push @nonread, $_;
	    next;
	}
	$work_lists{$_}++;
    }

    my $warning_messages = '';
    my $valid_files = 0;
    my $list_counter = 0;
    foreach my $list (sort keys %work_lists) {
	my $file = basename $list;	
	my $size = &commify((stat($list))[7]);
	my $lines = `grep -c -v '^#' '$list'`; 
	chomp $lines;
#	$lines--; # subtract header
	$counter++;

	# set flags for uploaded lists
	$list_counter++;
#	my $flag = (defined $flag_values{"list$list_counter"} ? $flag_values{"list$list_counter"} : $flag_values{'list1'}); 
	my $flag = "list$list_counter";
	if (defined $cgi->param("flag_$list")
	    and
	    $cgi->param("flag_$list") =~ /\w/) {
	    $flag = $cgi->param("flag_$list");
	}
	unless (defined $flag_values{$flag}) {
	    # create a new list
	    $highest_flag_num *= 2;
	    $flag_values{$flag} = $highest_flag_num;
	    $flag_values_num{$highest_flag_num} = $flag;
	    push @all_flags, $flag;
	    unshift @all_flags_num, $highest_flag_num;
	    # keep track of list flags:
	    if ($flag =~ /^list\d+/) {
		$used_flags{$flag}++;
	    }	    
	}
	my $flag_num = $flag_values{$flag};
        $list_specs{$list}{'FLAG'} = $flag;

	my $rc = &read_reporter_list($list,$flag_num);
	unless ($rc eq 'ok') {
	    $warning_messages .= "<BR><B><FONT COLOR=RED>WARNING: $rc</FONT></B><BR>\n";
	} else {
	    $valid_files++;
	}

my $header_lines = '';
if ($list_specs{$file}{'HEADER'}) {
    $header_lines = " + $list_specs{$file}{'HEADER'}";
    $lines -= $list_specs{$file}{'HEADER'};
}
	$out .= "<TR><TD>$counter</TD><TD><A HREF=\"$file\" TARGET=\"_list\">$file</A></TD><TD ALIGN=RIGHT>$size</TD><TD ALIGN=RIGHT>$lines$header_lines</TD>";
	$list_flag_match{'ENTRIES'}{$list} = $lines;

	foreach my $data_file (sort keys %work_files) {
	    my $matches = $list_match{$list}{$data_file};
	    $out .= "<TD ALIGN=RIGHT>$matches</TD>";
	}
	$out .= "<TD ALIGN=CENTER><input type=\"text\" size=10 name=\"flag_$list\" value=\"$flag\"></TD>\n";
	$out .= "<TD ALIGN=CENTER><input type=\"checkbox\" name=\"unload_files\" value=\"$list\"></TD></TR>\n";
    }
    $out .= "</TABLE>\n";

	foreach my $list (sort keys %work_lists) {
          $out .= "<input type=\"hidden\" name=\"work_lists\" value=\"$list\">\n";
	}
    # then the annotation file:

    unless (@annotation_files) {
	return $out."\n";
    }
 
    $out .= "\n<H2><BR>Annotation File</H2>\n<TABLE BORDER=1 CELLPADDING=5>\n<TR><TH ROWSPAN=2>#</TH><TH ROWSPAN=2>name</TH><TH ROWSPAN=2>size (Bytes)</TH><TD ROWSPAN=2><CENTER>".&help('lines_column', '<B>lines</B>')."</CENTER></TD><TD COLSPAN=".(scalar keys %work_files)."><CENTER>".&help('matches_column', '<B>matches</B>')."</CENTER></TD>";
    $out .= "<TD ROWSPAN=2><CENTER>".&help('unload_box', '<B>unload</B>')."</CENTER></TD></TR>\n";

    $out .= "<TR>";
    $counter = 0;
    foreach my $data_file (sort keys %work_files) {
	unless (defined $data{$data_file}) {
	    %{$data{$data_file}} = ();
	}
	$counter++;
	$out .= "<TH>file\&nbsp;$counter</TH>";
    }
    $out .= "</TR>\n";

    $counter = 0;
    @nonexist = ();
    @nonread = ();

    foreach (@annotation_files) {
	unless (-e $_) {
	    push @nonexist, $_;
	    next;
	}
	unless (-r $_) {
	    push @nonread, $_;
	    next;
	}
	$annotation_files{$_}++;
    }

    $warning_messages = '';
    $valid_files = 0;
    foreach my $list (sort keys %annotation_files) {
	my $file = basename $list;	
	my $size = &commify((stat($list))[7]);
	my $lines = `grep -c -v '^#' '$list'`; 
	$counter++;

	# set flags for uploaded lists
	my $list_counter = 0;
	$list_counter++;
	my $flag = (defined $flag_values{"annotation$list_counter"} ? $flag_values{"annotation$list_counter"} : $flag_values{'annotation1'}); 

# This is currently done in sub annotate
# have to check later if this can be done here already...
#	my $rc = &read_reporter_list($list,$flag);
#	unless ($rc eq 'ok') {
#	    $warning_messages .= "<BR><B><FONT COLOR=RED>WARNING: $rc</FONT></B><BR>\n";
#	} else {
#	    $valid_files++;
#	}

# for the moment, simply presume the annotation files are o.k.:
$valid_files++;

	$out .= "<TR><TD>$counter</TD><TD><A HREF=\"$file\" TARGET=\"_list\">$file</A></TD><TD ALIGN=RIGHT>$size</TD><TD ALIGN=RIGHT>$lines</TD>";
	foreach my $data_file (sort keys %work_files) {
	    my $matches = 'match';
	    $out .= "<TD ALIGN=RIGHT>$matches</TD>";
	}
	$out .= "<TD ALIGN=CENTER><input type=\"checkbox\" name=\"unload_files\" value=\"$list\"></TD></TR>\n";
    }
    $out .= "</TABLE>\n";

    $out .= '<BR><BR>'.$cgi -> submit(-name=>'submit', 
				    -value=>'Unload Files!');
    $out .= '&nbsp;'.$cgi -> submit(-name=>'submit', 
				    -value=>'Unload Files and Do it!');

    if (@nonexist) {
	$out .= "<BR>Error - file(s) don't exist:<BR>\n";
	foreach (@nonexist) {
	    $out .= "$_<BR>\n";
	}
    }
    if (@nonread) {
	$out .= "<BR>Error - file(s) not readable:<BR>\n";
	foreach (@nonread) {
	    $out .= "$_<BR>\n";
	}
    }
    foreach (sort keys %annotation_files) {
	$out .= "<input type=\"hidden\" name=\"annotation_files\" value=\"$_\">\n";
    }
    return $out."<BR>\n";
}


sub commify {
    my $text = reverse $_[0];
    $text =~ s/(\d\d\d)(?=\d)(?!\d*\.)/$1,/g;
    return scalar reverse $text;
}


sub get_files {

    my $out = '';

    for my $i (1..$max_files) {
	if ($cgi->param("file_name$i")) {
	    my $upload_file = ($cgi->param("file_name$i"));

	    if ($webstart) {
		# loads files into $tmp_dir
		# and returns full path to file(s)
		$upload_file = &upload_1_file($cgi, $cgi->param("file_name$i"), $tmp_dir,'');
	    }	
	    
	    # unpack archives or uncompress file if necessary
	    my @upload_files = &unpack_archive($upload_file);	    
	    
	    foreach my $file (@upload_files) {
		my ($format_check) = &check_file_format($file);
		while ($format_check eq 'Run again') {
		    ($format_check) = &check_file_format($file);
		}
#		warn "# check: $format_check for $file";
		unless ($format_check eq 'ok') {
		    $wrong_format_files{$file} = $format_check;
		} else {
		    push @files, $file;
		    $new_files++;
		}
	    }
	}
	if ($cgi->param("web_file_name$i")) {
	    my $upload_file = ($cgi->param("web_file_name$i"));

	    if ($webstart) {
		# loads files into $tmp_dir
		# and returns full path to file(s)
		$upload_file = &download_1_file($cgi, $cgi->param("web_file_name$i"), $tmp_dir,'');
	    }	
	    
	    # unpack archives or uncompress file if necessary
	    my @upload_files = &unpack_archive($upload_file);	    
	    
	    foreach my $file (@upload_files) {
		my ($format_check) = &check_file_format($file);
		while ($format_check eq 'Run again') {
		    ($format_check) = &check_file_format($file);
		}
		unless ($format_check eq 'ok') {
		    $wrong_format_files{$file} = $format_check;
		} else {
		    push @files, $file;
		    $new_files++;
		}
	    }
	}
	if ($cgi->param("reporter_list$i")) {
	    my $upload_file = ($cgi->param("reporter_list$i"));
	    
	    if ($webstart) {
		$upload_file = &upload_1_file($cgi, $cgi->param("reporter_list$i"), $tmp_dir,'');
	    }		

	    # unpack archives or uncompress file if necessary
	    my @upload_files = &unpack_archive($upload_file);	    

	    foreach my $list (@upload_files) {
		my ($format_check) = &check_file_format($list, 'reporter_list');
		while ($format_check eq 'Run again') {
		    ($format_check) = &check_file_format($list, 'reporter_list');
		}
		unless ($format_check eq 'ok') {
		    $wrong_format_files{$list} = $format_check;
		} else {
		    push @reporter_lists, $list;
		}
	    }
	}
    }

    # collect all ticked example files:
    foreach ($cgi->param('example_files')) {
	push @files, $_;
	$new_files++;
    }

    if (%wrong_format_files) {
	$out .= "<BR>
               <H3><FONT COLOR=RED>Invalid files:</FONT></H3>\n<UL>
              ";
	my @invalid_files = sort keys %wrong_format_files;
	foreach my $file (@invalid_files) {
	    $out .= "<LI>".(basename $file).": <B>$wrong_format_files{$file}</B></LI>\n";
	}

	$out .= "</UL>\n<BR>\n";
	
    }
    %wrong_format_files = ();			     

    # unload all files that are ticked for unloading:
    foreach ($cgi->param('unload_files')) {
	$unload_files{$_}++;
    }
    
    # skip all files that are ticked for skipping:
    foreach ($cgi->param('skip_files')) {
	$skip_files{$_}++;
    }
    
    unless ($new_files) {
	$out .= "No new files received!<BR>\n";
    }

    return $out;
}


sub unload_files {
    
    # load all files that are ticked for loading
    for my $i (1..$max_files) {
	if ($cgi->param("file_name$i")) {
	    my $upload_file = ($cgi->param("file_name$i"));

	    if ($webstart) {
		# loads files into $tmp_dir
		# and returns full path to file(s)
		$upload_file = &upload_1_file($cgi, $cgi->param("file_name$i"), $tmp_dir,'');
	    }	
	    
	    # unpack archives or uncompress file if necessary
	    my @upload_files = &unpack_archive($upload_file);	    

	    foreach my $file (@upload_files) {
		my ($format_check) = &check_file_format($file);
		while ($format_check eq 'Run again') {
		    ($format_check) = &check_file_format($file);
		}
		unless ($format_check eq 'ok') {
		    $wrong_format_files{$file} = $format_check;
		} else {
		    push @files, $file;
		    $new_files++;
		}
	    }
	}
    }

    # load all lists that are ticked for loading
    for my $i (1..$max_files) {
	if ($cgi->param("reporter_list$i")) {
	    my $upload_file = ($cgi->param("reporter_list$i"));
	    
	    if ($webstart) {
		$upload_file = &upload_1_file($cgi, $cgi->param("reporter_list$i"), $tmp_dir,'');
	    }		

	    # unpack archives or uncompress file if necessary
	    my @upload_files = &unpack_archive($upload_file);	    

	    foreach my $list (@upload_files) {
		my ($format_check) = &check_file_format($list, 'reporter_list');
		while ($format_check eq 'Run again') {
		    ($format_check) = &check_file_format($list, 'reporter_list');
		}
		unless ($format_check eq 'ok') {
		    $wrong_format_files{$list} = $format_check;
		} else {
		    push @reporter_lists, $list;
		}
	    }
	}
    }

    # unload all files that are ticked for unloading:
    foreach my $file ($cgi->param('unload_files')) {
	my @files = ($file);
	if ($file =~ /__second_part__/) {
	    @files = split /__second_part__/, $file;
	}
	foreach (@files) {
	    $unload_files{$_}++;
	}
    }
}


sub download_1_file {

    my($q, $file_name, $save_dir, $sub_dir_name) = @_;

    if ($file_name =~ /[\||;|\>|\<|\(|\)]/) {
	die "Invalid symbol in filename '$file_name'!";
    }

    my($file_base, $file_path, $file_suffix);
    my @tmp = split /\//, $file_name;
    $file_base = $tmp[-1];
#    warn "#kh: base: $file_base, path: $tmp[-2], server: $tmp[2]";

    # create output directory if it doesn't exist already
    unless (-e "$save_dir$sub_dir_name") {
	mkpath("$save_dir$sub_dir_name")
	    or die "Can't create directory $save_dir$sub_dir_name: $!";
    }

    # the name of the output file:
    my $temp_file_name = "$save_dir/$sub_dir_name$file_base";

    # download file
    
    # try lynx first
    my $lynx = '';
    
    if (defined $lynx_location
	and
	-x $lynx_location) {
	$lynx = $lynx_location;
    } else {
	$lynx = `which lynx`;
        chomp $lynx;
	if ($lynx) {
	    unless (-x $lynx) {
		$lynx = '';
	    }
	}
    }
    
#    warn "#KH: lynx = $lynx";

    if ($lynx) {
	# dump unformatted file
	my $job = "$lynx -source -dump $file_name > $temp_file_name";
	warn "#kh: job = $job";

	system $job;
    } else {
	eval "require LWP::Simple;";
	getstore($file_name, $temp_file_name);
    }


    # Validate the file size.    
    if ((stat $temp_file_name)[7] <= 0) {
	unlink($temp_file_name);
	die 'No data retrieved!'; # Did you use 'start_multipart_form()'?
    } else {
	# make sure the file can be read from other nodes
	if ($$use_cluster_nodes) {
	    system "chmod 644 '$temp_file_name' 2> /dev/null";
	}
    }    	

    return $temp_file_name;
}


sub upload_1_file {

    # KH: I found the basic for this somewhere on the net but unfortunately can't remember where...

    my($q, $file_name, $save_dir, $sub_dir_name) = @_;

    # Validate the file name.
    
    die("Please enter a real file name!") if (! $file_name);
    
    # Try to handle different path separators.
    
    if ($file_name =~ /\\/)
    {
	fileparse_set_fstype("MSDOS");
    }
    elsif ($file_name =~ /:/)
    {
	fileparse_set_fstype("MacOS");
    }
    
    # Note 1 of 2: At this point, we use $file_name as a string.
    
    my($file_base, $file_path, $file_suffix) = fileparse($file_name, '');
    
#    my $remove_files = 0;
#    if ($remove_files) {
#	unlink "$save_dir/$sub_dir_name" or die "can't unlink: $!\n";
#	exit;
#    }

    # Upload and write to a temporary file.
    unless (-e "$save_dir$sub_dir_name") {
	mkpath("$save_dir$sub_dir_name")
	    or die "Can't create directory $save_dir$sub_dir_name: $!";
    }
    my $temp_file_name = "$save_dir/$sub_dir_name$file_base$file_suffix";
    if (-f $temp_file_name) {	
        my $new_file_name = $temp_file_name.'_'.$PID;
   	move($temp_file_name,$new_file_name)
	    or die "Can't move files: $!";
#        warn "Copying $temp_file_name to $new_file_name";
    }
 
    # We don't put the file name in the message for security reasons.
    open(TEMP, "> $temp_file_name") 
	or die "Cannot open temporary file: $!";
    binmode(TEMP);
    
    my($bytes_read, $buf);
    my($total_size) = 0;
    
    # Note 2 of 2: At this point we use $file_name as a file handle.
    
    while ($bytes_read = read($file_name, $buf, 2048) )
    {
	$total_size += $bytes_read;
	print TEMP $buf;
    }
    
    close($file_name);
    close(TEMP);
    

    # Validate the file size.    
    if ((stat $temp_file_name)[7] <= 0) {
	unlink($temp_file_name);
	die 'No data retrieved! Maybe the CGI-tmp partition (/usr/tmp, /var/tmp) is full?'; # Did you use 'start_multipart_form()'?
    } else {
	# make sure the file can be read from other nodes
	if ($use_cluster_nodes) {
	    system "chmod 644 '$temp_file_name' 2> /dev/null";
	}
    }    	

    # make all files readable for all
#    my $job = "chmod 644 $temp_file_name";    
#    system "$job";
# seems to be done already somewhere else...
    
    return $temp_file_name;
}


sub unpack_archive {

    my $file = shift;

#    warn "# unpacking $file";

    my @return_files = ($file);
    
    # unzip archives
    if ($file =~ /\.tgz$/
	or
	$file =~ /\.tar\.gz$/
	or
	$file =~ /\.tar$/
	or
	$file =~ /\.zip$/) {
	# make temporary directory into which to unzip content:
	my $archive_dir = "$tmp_dir/archive_$PID";
	my $ext = 0;
	while (-e "$archive_dir$ext") {
	    $ext++;
	}
	$archive_dir .= $ext;

	system "mkdir -p '$archive_dir'; chmod 775 '$archive_dir'";
	unless (-e $archive_dir) {
	    die "Can't create directory '$archive_dir'!";
	}
	
	# uncompress if necessary and unpack archive
	my $uncompress = '';
	if ($file =~ /gz$/) {
	    $uncompress = 'z';
	}
	my $job = "tar -C '$archive_dir' -${uncompress}xf '$file'";
	
	# slightly different job for ZIP archives:
	if ($file =~ /\.zip/) {
	    $job = "unzip -d '$archive_dir' '$file'";
	}

	# unpack
	system "$job";

	# find all files in the archive directory
	$job = "find '$archive_dir' -type f";
	@return_files = `$job`;
	unless (@return_files) {
	    system "rm -rf '$archive_dir'";
	    die "Couldn't extract files from uploaded archive!";
	}

	# strip off line ending and directory path
	foreach (@return_files) {
	    chomp;
	    $_ = basename($_);
	}

	# remove input file:
	if ($clean_up_files) {
	    unlink $file;
	}

	# make all files readable for all
	$job = "find '$archive_dir' -type f -exec chmod 644 \{\} \\;";

	system "$job";

	# copy all files into output directory
	$job = "find '$archive_dir' -type f -exec mv \{\} '$tmp_dir' \\;";

	system "$job";

	# prepend directory to file names
	foreach (@return_files) {
	    $_ = "$tmp_dir/$_";
	}
	
	# remove archive_dir
	system "rm -rf '$archive_dir'";

    } elsif ($file =~ /\.gz$/) {
	# uncompress uploaded file
	my $rv = `gunzip '$file' 2>&1`;
	if ($rv =~ /not in gzip format/si) {
	    # the file was probably unpacked during download
	    my $file_new = $file;
	    $file_new =~ s/\.gz$//;
	    system "mv $file $file_new";
	}
	$file =~ s/\.gz$//;
	@return_files = $file;
    }

    return @return_files;
    
}	# End of upload_1_file.



sub help_section {

    my $out = '';

    # provides section that allows upload of files
    # or selection of predefined example files

#    warn "$cmd_line";
#    print &pointer."The following paragraph outlines the workflow of the tool:\n<BR><OL>\n";
    $out .= &pointer."Please follow the links and steps below:\n<BR><OL>\n";

    if ($submit =~ /do it/i) {
	$out .= "<LI>Go straight to your <A HREF=\"#RESULTS\">results</A>";
	if (grep /action_50_output/, @actions) {
	    my $counter = 0;
	    foreach (@actions) {
		$counter++;
		if (/action_50_output/) {
		    $out .= " (including <A HREF=\"#action$counter\">data retrieval</A>)";
		    last;
		}
	    }
	} else {
	    $out .= " (no data retrieval requested ".&help('data_retrieval').')';
	}
	$out .= "!<BR>\n";

    } elsif ($submit =~ /Refine/i) {
	$out .= "<LI>Go straight to the <A HREF=\"#RESULTS\">parameters section</A>!<BR>\n";
    }

    unless (%work_files) {
	$out .= "<LI>In the <A HREF=\"#INPUT\">input section</A> specify the data files to work on.".&help('data_input')."<BR>\n"; 
	$out .= "<LI>Click on the <A HREF=\"#load_files_button\">'Load Files'</A> button at the bottom to load the data.<BR>\n";
	$out .= "</OL>\n";
	$out .= "For <B>more help</B>, please refer to the <A HREF=\"$user_instructions_link\">user instructions</A> or the <A HREF=\"$tutorial_link\">tutorial</A>.\n<BR>\n";
	return $out;
    }

#    $out .= "<LI>In the <A HREF=\"#INPUT\">input section</A> add or remove data files.".&help('data_input_modify')."<BR>\n"
#	    unless ($submit eq 'Load Settings!'
#		    or
#		    $submit eq 'Save Settings!'); 

    if ($new_files) {
	$out .= "<LI>In the <A HREF=\"#DATA FILES\">list of loaded files</A> check that the automatically detected arrray layout and data columns are correct.".&help('data_auto_detect')."<BR>\n"; 
    }

    if (@actions
	or
	$settings_loaded
	or
	$submit eq 'Save Settings!') { 
	    $out .= "<LI>In the <A HREF=\"#ACTION SETUP\">action section</A> revise the functions and filters you would like to apply.".&help('action_settings')."<BR>\n"; 
	    $out .= "<LI>In the <A HREF=\"#SETTINGS\">settings section</A> revise the parameters for the actions you plan to carry out.".&help('parameter_settings')."<BR>\n";
#	}
    } else {
	$out .= "<LI>In the <A HREF=\"#ACTION SETUP\">action section</A> specify the functions and filters you would like to apply.".&help('action_settings')."<BR>\n"; 
	$out .= "<LI>In the <A HREF=\"#SETTINGS\">settings section</A> you can change parameters from the defaults for the actions you plan to carry out.".&help('parameter_settings')."<BR>\n"; 
    }
    $out .= "<LI>Click on the "	
	. $cgi -> submit(-name=>'submit', 
			 -value=>'Do it!')
	. " button in the <A HREF=\"#action_start\">action section</A> to start the processing.<BR>\n";
    $out .= "</OL>\n";
    $out .= "For <B>more help</B>, please refer to the <A HREF=\"$user_instructions_link\">user instructions</A> or the <A HREF=\"$tutorial_link\">tutorial</A>.\n<BR>\n";

    return $out;
}




sub plot_pathways {
  #  warn "#KH: '$cmd_line'\n";
#    exit;
    my $pathway_boxplot = 0;
    my $pathway_connectors = 0;
    my $input_file = '';
    unless (defined $cgi->param('formatted_file')) {
	die "No input file specified!";
    }
    $input_file = $cgi->param('formatted_file');
    
    # get the id column
    my $id_column = '';
    unless (defined $cgi->param('pathway_plotter_id_column')) {
	die "No id column chosen for pathway plotter!";
    }
    $id_column = $cgi->param('pathway_plotter_id_column');
    if ($id_column == -1) {
	die "No id column chosen for pathway plotter!";
    }

    # get the value columns
    my @value_column = ();
    unless (defined $cgi->param('pathway_plotter_value_column')) {
	die "No value column chosen for pathway plotter!";
    } else {
	foreach ($cgi->param('pathway_plotter_value_column')) {
	    next if ($_ == -1);
	    push @value_column, $_;
	}
    }
    unless (@value_column) {
	die "No value column(s) chosen for pathway plotter!";
    }

    # get positions of necessary columns
    my %suitable_columns = ('BioCarta', -1,);

    open (IN, "$input_file")
	or die "Can't read from $input_file!\n";
    my %input_data_files = ();
    my $header = <IN>;
    chomp $header;
    my @tmp = split /\t/, $header;
    my $found_column = 0;
    for ( my $i = 0; $i <= $#tmp; $i++) {
	if (defined $suitable_columns{$tmp[$i]}) {
	    $suitable_columns{$tmp[$i]} = $i;
	    $found_column = 1;
	}
	for (my $index = 0; $index <= $#value_column; $index++) {
	    if ($value_column[$index] == $i) {
		$input_data_files{$index} = $tmp[$i];
	    }
	}
    }

    unless ($found_column) {
#	die "No suitable column found (need at least one of '".(join "','", sort keys %suitable_columns)."')";
	die "No BioCarta data available - please go back into ArrayPipe and annotate!\n";
    }

    # create output directory
    my $out_dir = "$tmp_dir/pathway_plot.$PID";
    my $out_dir_html = "$tmp_dir_html/pathway_plot.$PID";
    my $info_out = '';

    if (-d $out_dir) {
	system "rm -rf '$out_dir'";
    }
    system "mkdir '$out_dir'; chmod 777 '$out_dir' 2> /dev/null";

    my %gene_symbol = ();
    my %pathway = ();
    my %pathnum = ();
    my $pathnum = 0;
    my %probes = ();
    my %out = ();

    # collect all data
    while (<IN>) {	

	# remove comments
	if (/#/) {
	    s/#.*//;
	}

	# skip empty lines
	next unless (/\w/);
		
	# delete trailing white-space
	s/\s*$//;

	# extract required columns from line
	# split at tabs
	my @tmp = split /\t/, $_;

	my $biocarta = $tmp[$suitable_columns{'BioCarta'}];
	next unless (defined $biocarta
		     and
		     $biocarta =~ /\w/);
	my @biocarta = ();
	foreach my $pathway (split /\;/, $biocarta) {
	    $pathway =~ s/^\s*//g;
	    $pathway =~ s/\s*$//g;
	    push @biocarta, $pathway;
	    unless (defined $pathnum{$pathway}) {
		$pathnum++;
		$pathnum{$pathway}= $pathnum;
	    }
	}

	my $id = $tmp[$id_column];

	for (my $i = 0; $i <= $#value_column; $i++) {
	    my $data = $tmp[$value_column[$i]];
	    next unless ($data =~ /\w/);
	    push @{$out{$i}{$id}{'BioCarta'}}, @biocarta;
	    push @{$out{$i}{$id}{'Data'}}, $data;
	}
    }
    close IN;
    
    # print data to files
    my %input_files = ();
    foreach my $index (keys %out) {
	my $out = "$out_dir/".(basename($input_data_files{$index})).".pathwayplot.$PID";
	$input_files{$index} = $out;
	open (PATH, ">$out")
	    or die "Can't write to '$out': $!";
	
	foreach my $id (sort keys %{$out{$index}}) {
	    foreach my $pathway (sort @{$out{$index}{$id}{'BioCarta'}}) {		
		foreach my $value (sort @{$out{$index}{$id}{'Data'}}) {
		    print PATH "$id\t$value\t$pathnum{$pathway}\t$pathway\n";
		}
	    }
	}
	close PATH;
    }

    my $job = "perl /home/khokamp/FPMI/tools/pathexpressplot.pl "; 
    if ($pathway_boxplot) {
	$job .= '-box ';
    }
    if (@value_column == 1) {
	$job .= '-single ';
    }

    $pathway_connectors = 1;
    if ($pathway_connectors) {
	$job .= '-lines ';
    }
    my $out_name = 'pathway_out.html';
    my ($action_num,$action) = (1,1);
    my $out = '';
    $job .= "-refer '<A HREF=\"$referral_address\">[back to spreadsheet]</A><P>' ";
    if (@value_column > 1) {
	my $file_label = '-label \'';
	foreach (sort keys %input_files) {
	    $job .= " '$input_files{$_}'";
	    $file_label .= ''.(basename $input_files{$_})."=".(basename $_)." ++ ";
#	    warn "#KH: adding ".(basename $input_files{$_})."=".(basename $_)." ++ ";
	}
	$file_label =~ s/ \+\+ $/\'/;
	$job .= " -out '$out_dir' -name '$out_name' $file_label";
	$out = &pathway_plot_job($action_num,$action,'selected',$job,$out_dir,$out_name,$out_dir_html);
#	warn "#pathway plotter output: $out";
    } else {
	foreach my $file (sort keys %input_files) {
	    my $job2 = "$job -out '$out_dir' -name '$out_name' '$input_files{$file}'";
	    $out = &pathway_plot_job($action_num,$action,$file,$job2,$out_dir,$out_name,$out_dir_html);
#	    warn "#pathway plotter output: $out";
	}
    }
    print $cgi->redirect("$out_dir_html/$out");
    exit;
}
    



sub generate_transpath_output {

    my $input_file = '';
    unless (defined $cgi->param('formatted_file')) {
	die "No input file specified!";
    }
    $input_file = $cgi->param('formatted_file');
    
    my $output_file = "$tmp_dir/".(basename $input_file).'.transpath.html';
    my $output_file_html = "$tmp_dir_html/".(basename $input_file).'.transpath.html';

    # get positions of necessary columns
    my %suitable_columns = ('SWISSPROT', -1,
			    'EMBL', -1,
			    'SPTREMBL', -1,
			    'SP', -1,
			    'dbj', -1,
			    'emb', -1,
			    'gb', -1,
			    'SWISSPROT accession', -1,
			    'EMBL accession', -1,
			    'SPTREMBL accession', -1);
    
    open (IN, "$input_file")
	or die "Can't read from $input_file!\n";
    my $header = <IN>;
    chomp $header;
    my @tmp = split /\t/, $header;
    my $found_column = 0;
    for ( my $i = 0; $i <= $#tmp; $i++) {
	if (defined $suitable_columns{$tmp[$i]}) {
	    $suitable_columns{$tmp[$i]} = $i;
	    $found_column = 1;
	}
    }

    unless ($found_column) {
	die "No suitable column found (need at least one of '".(join "','", sort keys %suitable_columns)."')";
    }

    my $ratio_column = '';
    my $ratio_log = '';
#    for ( my $i = 0; $i <= $#tmp; $i++) {
#	if ($tmp[$i] =~ /_R/
#	    or
#	    $tmp[$i] =~ /$FOLD_CHANGE_TERM/
#	    or
#	    $tmp[$i] =~ /$RATIO_TERM/
#	    or
#	    $tmp[$i] =~ /$MERGED_RATIOS_TERM/) {
#	    $ratio_column = $i;
#	    if ($tmp[$i] =~ /_L/) {
#		$ratio_log = 1;
#	    } else {
#		$ratio_log = 0;
#	    }
#	}
#    }

    if (defined $cgi->param('transpath_column')) {
	$ratio_column = $cgi->param('transpath_column');
    }

    unless ($ratio_column =~ /\d/) {
	die "No column with ratios found!";
    }
    
    my @out = ();
    while (<IN>) {	

	# remove comments
	if (/#/) {
	    s/#.*//;
	}

	# skip empty lines
	next unless (/\w/);
		
	# delete trailing white-space
	s/\s*$//;

	# extract required columns from line
	# split at tabs
	my @tmp = split /\t/, $_;

	my $ratio = $tmp[$ratio_column];
	unless ($ratio =~ /\d/) {
	    $ratio = 0;
	}
	if ($ratio_log) {
	    $ratio = 2**$ratio;
	}
	if ($ratio =~ /\.(\d+)/) {
	    if (length($1) > 1) {
		$ratio = sprintf "%.1f", $ratio;
	    }
	}

	foreach my $key (sort keys %suitable_columns) {
	    next unless ($suitable_columns{$key} >= 0);
	    my $data = $tmp[$suitable_columns{$key}];
	    next unless ($data =~ /\w/);
	    foreach my $data2 (split /\;/, $data) {

		$data2 =~ s/^\s*//g;
		$data2 =~ s/\s*$//g;
		my $key_out = $key;
		if ($key eq 'EMBL'
		    or
		    $key eq 'EMBL accession'
		    or
		    $key eq 'dbj'
		    or
		    $key eq 'gb'
		    or
		    $key eq 'emb') {
		    $key_out = 'DDBJ';
		} elsif ($key =~ /SWISSPROT/
			 or
			 $key eq 'SP') {
		    # skip the SWISSPROT names (which usually contain underscores)
		    # but keep the SWISSPROT IDs:
		    if ($data2 =~ /_/) {
			next;
		    }
		    if ($key eq 'SP') {
			$key_out = 'SWISSPROT';
		    }	
		} 
		$key_out =~ s/ accession//;
		push @out, "$key_out:$data2|$ratio";
	    }
	}
    }
    close IN;

    unless (@out) {
	die "No element found!";
    }
    
my $output_file_menu = $output_file.'.menu';
my $output_file_main = $output_file.'.main';

    open (OUT, ">$output_file")
	or die "Can't write to $output_file: $!\n";

    my $transpath_broken = 1;

    if ($transpath_broken) {
	print OUT "<HTML>\n<HEAD>\n<TITLE>TransPath Input</TITLE>\n</HEAD>\n<BODY BGCOLOR=WHITE>\n<H1>TransPath Input</H1>\nGo to <A HREF=\"http://www.brinkman.mbb.sfu.ca/cgi-bin/biobase/transpath/5.2/bin/start.cgi\" TARGET=TRANSPATH>/www.brinkman.mbb.sfu.ca/cgi-bin/biobase/transpath/5.2/bin/start.cgi</A>, click on <B>Molecules</B> and fill the following info into the <B>New Search Terms</B> box! (Make sure that <B>External Database Hyperlinks</B> is selected under <B>Quick Search Fields</B>    .)<BR><PRE>\n";
	foreach (@out) {
	    print OUT "$_\n";
	}
	
	print OUT "\n</PRE>\n</BODY>\n</HTML>\n";
	close OUT;
	
	print STDOUT $cgi->redirect("$output_file_html");
	
	return '';  
    }

print OUT 
'<html>
<head>
<title>TRANSPATH Professional</title>
</head>
<frameset COLS="160,*" FRAMEBORDER="0" BORDER=0 FRAMESPACING="0" BORDER="no">
	<frame NAME="tp_menu" SRC="'
.(basename $output_file_menu)
.'" MARGINWIDTH="0" MARGINHEIGHT="0" SCROLLING="Auto" FRAMEBORDER="0" FRAMESPACING="0">
	<frame NAME="tp_main" SRC="'
.(basename $output_file_main)
.'" MARGINWIDTH="0" MARGINHEIGHT="0" SCROLLING="Auto" FRAMEBORDER="0" FRAMESPACING="0">
</frameset>
</html>';
close OUT;


open (OUT, ">$output_file_menu")
	or die "Can't write to $output_file_menu: $!\n";

print OUT 
'<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">

<html>
<head>
    <title>Menu</title>
</head>

<body BGCOLOR="#800000" TEXT="#FFFFFF" LINK="#FFFFFF" VLINK="#FFFFFF" ALINK="#FFFFFF">
<table BORDER="0" CELLPADDING="0" ALIGN="CENTER">
	<tr>
		<td HEIGHT="80" ALIGN="CENTER">
			<a HREF="http://www.biobase.de/"target="tp_main"><img SRC="'
. $transpath_URL
    . '/biobase/transpath/4.4/images/logo_small.gif"WIDTH=140 HEIGTH=28 BORDER=0 ALT="LOGO"></a>		
		</td>

	</tr>
	<tr>
		<td  ALIGN="CENTER">
			<a HREF="'
. $transpath_URL
    . '/cgi-bin/biobase/transpath/4.4/bin/start.cgi?PAGE=MAIN"target="tp_main"><img SRC="'
. $transpath_URL
    . '/biobase/transpath/4.4/images/tpsearch.gif"WIDTH=140 HEIGHT=60 BORDER=0 ALT="TRANSPATH search engine TRANSPATH Professional 4.4"></a>
		</td>
	</tr>
	<tr>
		<td HEIGHT="10"  ALIGN="CENTER">
			
			
		</td>

	</tr>
	<tr>
		<td HEIGHT="10" WIDTH=140 ALIGN="CENTER">
			&nbsp;
		</td>
	</tr>
	<tr>
		<td  ALIGN="CENTER">
			<a HREF="'
. $transpath_URL
    . '/cgi-bin/biobase/transpath/4.4/bin/search.cgi?TABLE=molecule"target="tp_main"><img SRC="'
. $transpath_URL
    . '/biobase/transpath/4.4/images/molecule.gif"WIDTH=140 HEIGHT=30 BORDER=0 ALT="Search the MOLECULE table"></a>

		</td>
	</tr>
	<tr>
		<td  ALIGN="CENTER">
			<a HREF="'
. $transpath_URL
    . '/cgi-bin/biobase/transpath/4.4/bin/search.cgi?TABLE=reaction"target="tp_main"><img SRC="'
. $transpath_URL
    . '/biobase/transpath/4.4/images/reaction.gif"WIDTH=140 HEIGHT=30 BORDER=0 ALT="Search the REACTION table"></a>
		</td>
	</tr>
	<tr>
		<td  ALIGN="CENTER">

			<a HREF="'
. $transpath_URL
    . '/cgi-bin/biobase/transpath/4.4/bin/search.cgi?TABLE=gene"target="tp_main"><img SRC="'
. $transpath_URL
    . '/biobase/transpath/4.4/images/gene.gif"WIDTH=140 HEIGHT=30 BORDER=0 ALT="Search the GENE table"></a>
		</td>
	</tr>
<!--
	<TR>
		<TD  ALIGN="CENTER">
			<A HREF="'
. $transpath_URL
    . '/cgi-bin/biobase/transpath/4.4/bin/search.cgi?TABLE=nmolecule"target="tp_main"><IMG DYN_SRC_MENU_4 WIDTH=140 HEIGHT=80 BORDER=0 ALT="Search the NetPro MOLECULE table"></A>
		</TD>
	</TR>
	<TR>
		<TD  ALIGN="CENTER">
			<A HREF="'
. $transpath_URL
    . '/cgi-bin/biobase/transpath/4.4/bin/search.cgi?TABLE=interaction"target="tp_main"><IMG DYN_SRC_MENU_5 WIDTH=140 HEIGHT=80 BORDER=0 ALT="Search the NetPro INTERACTION table"></A>
		</TD>
	</TR>
-->
	<tr>
		<td HEIGHT="10"  ALIGN="CENTER">
			&nbsp;
		</td>
	</tr>

	<tr>
		<td  ALIGN="CENTER">
			<a HREF="'
. $transpath_URL
    . '/biobase/transpath/4.4/doc/doc/browse.html"target="tp_main"><img SRC="'
. $transpath_URL
    . '/biobase/transpath/4.4/images/maps.gif"WIDTH=140 HEIGHT=30 BORDER=0 ALT="Browse Maps"></a>
		</td>
	</tr>
<!-- removed to avoid problems with installation
    <TR>
		<TD  ALIGN="CENTER">
			<A HREF="'
. $transpath_URL
    . '/cgi-bin/biobase/transpath/4.4/bin/search.cgi?TABLE=nmolecule"target="tp_main"><IMG DYN_SRC_MENU_5 WIDTH=140 HEIGHT=30 BORDER=0 ALT="BLAST the MOLECULE table"></A>
		</TD>
	</TR>
//-->
	<tr>
		<td  ALIGN="CENTER">
			<a HREF="'
. $transpath_URL
    . '/cgi-bin/biobase/transpath/4.4/bin/classification.cgi"target="tp_main"><img SRC="'
. $transpath_URL
    . '/biobase/transpath/4.4/images/browse.gif"WIDTH=140 HEIGHT=30 BORDER=0 ALT="Browse Molecules"></a>

		</td>
	</tr>
	<tr>
		<td HEIGHT="10"  ALIGN="CENTER">
			&nbsp;
		</td>
	</tr>
	<tr>
		<td  ALIGN="CENTER">

			<a HREF="'
. $transpath_URL
    . '/cgi-bin/biobase/transpath/4.4/bin/search.cgi?analyzer0=on"target="tp_main"><img SRC="'
. $transpath_URL
    . '/biobase/transpath/4.4/images/ArrayAnalyzer.gif"WIDTH=140 HEIGHT=30 BORDER=0 ALT="ArrayAnalyzer"></a>
		</td>
	</tr>
	<tr>
		<td  ALIGN="CENTER">
			<a HREF="'
. $transpath_URL
    . '/cgi-bin/biobase/transpath/4.4/bin/pwb_start.cgi"target="tp_main"><img SRC="'
. $transpath_URL
    . '/biobase/transpath/4.4/images/PathwayBuilder.gif"WIDTH=140 HEIGHT=30 BORDER=0 ALT="PathwayBuilder"></a>
		</td>
	</tr>
	<tr>

		<td HEIGHT="10"  ALIGN="CENTER">
			&nbsp;
		</td>
	</tr>
	<tr>
		<td HEIGHT="10"  ALIGN="CENTER">
			<a HREF="'
. $transpath_URL
    . '/biobase/transpath/4.4/help.html"target="tp_main"><img SRC="'
. $transpath_URL
    . '/biobase/transpath/4.4/images/info.gif"WIDTH=140 HEIGHT=30 BORDER=0 ALT="Help"></a>
		</td>
	</tr>

	<tr>
		<td  ALIGN="CENTER">
			<a HREF="mailto:support@biobase.de"target="tp_main"><img SRC="'
. $transpath_URL
    . '/biobase/transpath/4.4/images/contact.gif"WIDTH=140 HEIGHT=30 BORDER=0 ALT="Contact us"></a>
		</td>
	</tr>
	<tr>
		<td HEIGHT="10"  ALIGN="CENTER">
			&nbsp;
		</td>

	</tr>
	<tr>
	<td ALIGN="CENTER">
			<b>
            Other installed BIOBASE databases<br>
            <a target=_top HREF="'
. $transpath_URL
    . '/cgi-bin/biobase/transfac/7.4/bin/start.cgi">TRANSFAC</a><br>
            <TRANSPATH>
            <TRANSCOMPEL>

            <PATHODB>
            <SMARTDB>
            <TRANSPRO>
            </b>
		</td>
	</tr>
</table>

</body>
</html>
';

close OUT;



open (OUT, ">$output_file_main")
	or die "Can't write to $output_file_main: $!\n";
    
    print OUT '<!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.0 Transitional//EN">

<html>
<head>
    <title>TRANSPATH Professional 4.4 Search Form</title>
</head>
<body bgcolor="#FFFFFF">
<br>
<form ACTION="'
. $transpath_URL
    . '/cgi-bin/biobase/transpath/4.4/bin/search.cgi" METHOD="POST">
    <input type="hidden" name="TABLE" value="molecule">
    <input type="hidden" name="REM_USR" value="'
    . $transpath_user
    .'">
    <input type="hidden" name="SEARCHFIELD" value="FAST">

    <input type="hidden" name="SEARCHTERM1" value="___VIEW___">
    <input type="hidden" name="SEARCHTERM2" value="">
    <input type="hidden" name="SEARCHTERM3" value="">
    <input type="hidden" name="FASTSEARCHFIELD1" value="AC">
    <input type="hidden" name="FASTSEARCHFIELD2" value="AC">
    <input type="hidden" name="FASTSEARCHFIELD3" value="AC">
    <input type="hidden" name="OUTPUTFIELDS" value="NA">
    <input type="hidden" name="OUTPUTFIELDS" value="TY">
    <input type="hidden" name="LINK1" value="OR">

    <input type="hidden" name="LINK2" value="AND">
    <input type="hidden" name="LINK3" value="AND">
    <table WIDTH="98%" BORDER="0" ALIGN="CENTER" BGCOLOR="#800000" CELLSPACING="0" CELLPADDING="6">
        <tr>
            <td>
                <table WIDTH="100%" BORDER="0" CELLSPACING="0" CELLPADDING="0">
                    <tr>
                        <td COLSPAN="3" ALIGN="CENTER">
                            <font face="sans-serif" color="#FFFFFF"><b>your login name: adm</b></font>

                        </td>
                    </tr>
                </table>
            </td>
        </tr>
        <tr>
            <td>
                <table WIDTH="100%" BORDER="0" CELLSPACING="0" CELLPADDING="0">
                    <tr>

                        <td BGCOLOR="#C0C0C0">
                            <font face="sans-serif">
                            &nbsp;&nbsp;Select a previous search result: &nbsp;
                            <select name="FILENAME">
<option VALUE="last.search">last.search<option VALUE="analysis">analysis                            </select>
                            &nbsp; and &nbsp;
                            </font>

                        </td>
                        <td BGCOLOR="#C0C0C0">
                            <input type="submit" name="SUBMIT" value="VIEW">
                            <font face="sans-serif"> it &nbsp;</font>
                        </td>
                        <td BGCOLOR="#C0C0C0">
                            <input type="submit" name="SUBMIT" value="DELETE">
                            <font face="sans-serif"> it</font>

                        </td>
                    </tr>
                </table>
            </td>
        </tr>
    </table>
</form>
    
<form ACTION="'
. $transpath_URL
    . '/cgi-bin/biobase/transpath/4.4/bin/search.cgi" METHOD="POST" enctype="multipart/form-data">
    <input type="hidden" name="TABLE" value="molecule">

      
    <table WIDTH="98%" ALIGN="CENTER" BGCOLOR="#800000" CELLSPACING="0" CELLPADDING="6">
        <tr>
            <td>
                <table WIDTH="100%" BORDER="0" CELLSPACING="0" CELLPADDING="0" BGCOLOR="#800000">
                    <tr>
                        <td COLSPAN="4" ALIGN="CENTER" VALIGN="MIDDLE" BGCOLOR="#800000">
                            <img SRC="'
    . $transpath_URL
    . '/biobase/transpath/4.4/images/molecule-title.gif" WIDTH=400 HEIGHT=40 BORDER=0 ALT="">
			    <a HREF="'
    . $transpath_URL
    . '/biobase/transpath/4.4/doc/search/TFP_search.html"><font face="sans-serif" color="#FFFFFF">Help</font></a>

                        </td>
                    </tr>
                    <tr>
                        <td WIDTH="12%" VALIGN="TOP" BGCOLOR="#C0C0C0">
                            <table WIDTH="90%" ALIGN="RIGHT">
                                <tr>
                                    <td ALIGN="CENTER">
                                        <font face="sans-serif">Quick Search</font>

                                    </td>
                                </tr>
                                <tr>
                                    <td BGCOLOR="#C0C0C0">
                                        &nbsp;
                                    </td>
                                </tr>
                            </table>
                        </td>

                        <td WIDTH="3%">
                            &nbsp;
                        </td>
                        <td WIDTH="85%" BGCOLOR="#C0C0C0" VALIGN="TOP">
                            <table WIDTH="100%" BORDER="0">
                                <tr>
                                    <td VALIGN="TOP">
                                        <font face="sans-serif"><b>Search Terms</b></font><br>

                                        <input TYPE="radio" checked name="LINK1" value="NEW"> NEW
                                    </td>
                                    <td VALIGN="TOP">
                                        <a HREF="'
    . $transpath_URL
    . '/biobase/transpath/4.4/doc/molecule.html#field"><font face="sans-serif"><b>Quick Search Fields</b></font></a>
                                    </td>
                                    <td VALIGN="TOP">
                                        <a HREF="'
    . $transpath_URL
    . '/biobase/transpath/4.4/doc/molecule.html#field"><font face="sans-serif"><b>Output Fields</b></font></a>
                                    </td>

                                </tr>
                                <tr>
                                    <td VALIGN="TOP">
                                        <p>
					<textarea name=SEARCHTERM1 COLS=22 ROWS=2>
';

    foreach (@out) {
	print OUT "$_\n";
    }

    print OUT '</textarea><br>
<!--                                        <INPUT TYPE="Text" NAME="SEARCHTERM1" VALUE="" SIZE="32"><BR>  -->
                                        <font face="sans-serif">Search Term 1 &#40;can be a list of terms&#41;</font>
					<font face="sans-serif" SIZE=-2><br>

					or upload a file with queries<br>(e.g. accession numbers and expression levels).<br>
					<input TYPE=FILE maxlength=10000000 NAME=QUERYLIST accept="text/*">
					</font>
                                      </p>
                                    </td>
                                    <td VALIGN="TOP" rowspan=2>
                                        <select NAME="FASTSEARCHFIELD1" SIZE="8">
<option VALUE="ALL" >All fields			<option VALUE="AC" >Accession number		<option VALUE="CL" >Classification		<option VALUE="CC" >Comments			<option VALUE="CC0" >Comments:historical		<option VALUE="CC1" >Comments:general		<option VALUE="CC2" >Comments:CONFLICT!		<option VALUE="CC3" >Comments:physiological function<option VALUE="CC4" >Comments:by analogy to	<option VALUE="CC5" >Comments:structure		<option VALUE="CC6" >Comments:experimental inhibitor<option VALUE="CC7" >Comments:kinetics		<option VALUE="CC8" >Comments:mechanism		<option VALUE="CC9" >Comments:methods		<option VALUE="CC10" >Comments:enzyme classification<option VALUE="CC11" >Comments:function<option VALUE="CC12" >Comments:sequence<option VALUE="CC13" >Comments:disease<option VALUE="CX" >Complexes		<option VALUE="ST" >Complex or Modified form of<option VALUE="GE" >Encoding gene<option VALUE="DR" SELECTED>External database hyperlinks<option VALUE="FO" >Feature of<option VALUE="FP" >Feature position<option VALUE="FT" >Features (motifs)<option VALUE="GO" >GO: biological process, molecular function<option VALUE="IP" >Isoelectric point<option VALUE="CN" >Location negative and experiment(s)<option VALUE="CP" >Location positive and experiment(s)<option VALUE="RX" >Medline database hyperlink<option VALUE="MF" >Modified forms	<option VALUE="NA">Molecule name<option VALUE="XC" >Reaction catalyzed<option VALUE="XA" >Reaction downstream<option VALUE="XI" >Reaction inhibited<option VALUE="XB" >Reaction upstream<option VALUE="RA" >Reference authors<option VALUE="RN" >Reference number<option VALUE="RL" >Reference publication<option VALUE="RT" >Reference title<option VALUE="SQ" >Sequence<option VALUE="SZ" >Sequence length, molecular weight<option VALUE="SC" >Sequence source<option VALUE="OS" >Species<option VALUE="HC" >Subfamilies<option VALUE="HP" >Superfamilies<option VALUE="TY" >Type                                        </select>

                                    </td>
                                    <td VALIGN="TOP" rowspan=5>
                                        <select NAME="OUTPUTFIELDS" SIZE="16" multiple>
<option VALUE="AC" >Accession number		<option VALUE="CL" >Classification		<option VALUE="CC" >Comments			<option VALUE="CX" >Complexes		<option VALUE="ST" >Complex or Modified form of<option VALUE="GE" >Encoding gene<option VALUE="DR" >External database hyperlinks<option VALUE="FO" >Feature of<option VALUE="FP" >Feature position<option VALUE="FT" >Features (motifs)<option VALUE="GO" >GO: biological process, molecular function<option VALUE="IP" >Isoelectric point<option VALUE="CN" >Location negative and experiment(s)<option VALUE="CP" >Location positive and experiment(s)<option VALUE="RX" >Medline database hyperlink<option VALUE="MF" >Modified forms	<option VALUE="NA" SELECTED>Molecule name<option VALUE="XC" >Reaction catalyzed<option VALUE="XA" >Reaction downstream<option VALUE="XI" >Reaction inhibited<option VALUE="XB" >Reaction upstream<option VALUE="RA" >Reference authors<option VALUE="RN" >Reference number<option VALUE="RL" >Reference publication<option VALUE="RT" >Reference title<option VALUE="SQ" >Sequence<option VALUE="SZ" >Sequence length, molecular weight<option VALUE="SC" >Sequence source<option VALUE="OS" SELECTED>Species<option VALUE="HC" >Subfamilies<option VALUE="HP" >Superfamilies<option VALUE="TY" SELECTED>Type                                        </select>

                                    </td>
                                </tr>
                                <tr>
                                    <td>
                                        <p>
                                        <input TYPE="radio" checked name="LINK2" value="AND"> AND
                                        <input TYPE="radio" name="LINK2" value="OR"> OR
                                        <input type="radio" name="LINK2" value="NOT"> NOT
                                        </p>

                                    </td>
                                </tr>
                                <tr>
                                    <td VALIGN="TOP">
                                        <p>
                                        <input TYPE="Text" NAME="SEARCHTERM2" VALUE="" SIZE="22"><br>
                                        <font face="sans-serif">Search Term 2</font>
                                        </p>

                                    </td>
                                    <td VALIGN="TOP" rowspan=2>
                                        <select NAME="FASTSEARCHFIELD2" SIZE="5">
<option VALUE="ALL" >All fields			<option VALUE="AC" >Accession number		<option VALUE="CL" >Classification		<option VALUE="CC" >Comments			<option VALUE="CC0" >Comments:historical		<option VALUE="CC1" >Comments:general		<option VALUE="CC2" >Comments:CONFLICT!		<option VALUE="CC3" >Comments:physiological function<option VALUE="CC4" >Comments:by analogy to	<option VALUE="CC5" >Comments:structure		<option VALUE="CC6" >Comments:experimental inhibitor<option VALUE="CC7" >Comments:kinetics		<option VALUE="CC8" >Comments:mechanism		<option VALUE="CC9" >Comments:methods		<option VALUE="CC10" >Comments:enzyme classification<option VALUE="CC11" >Comments:function<option VALUE="CC12" >Comments:sequence<option VALUE="CC13" >Comments:disease<option VALUE="CX" >Complexes		<option VALUE="ST" >Complex or Modified form of<option VALUE="GE" >Encoding gene<option VALUE="DR" >External database hyperlinks<option VALUE="FO" >Feature of<option VALUE="FP" >Feature position<option VALUE="FT" >Features (motifs)<option VALUE="GO" >GO: biological process, molecular function<option VALUE="IP" >Isoelectric point<option VALUE="CN" >Location negative and experiment(s)<option VALUE="CP" >Location positive and experiment(s)<option VALUE="RX" >Medline database hyperlink<option VALUE="MF" >Modified forms	<option VALUE="NA" SELECTED>Molecule name<option VALUE="XC" >Reaction catalyzed<option VALUE="XA" >Reaction downstream<option VALUE="XI" >Reaction inhibited<option VALUE="XB" >Reaction upstream<option VALUE="RA" >Reference authors<option VALUE="RN" >Reference number<option VALUE="RL" >Reference publication<option VALUE="RT" >Reference title<option VALUE="SQ" >Sequence<option VALUE="SZ" >Sequence length, molecular weight<option VALUE="SC" >Sequence source<option VALUE="OS" >Species<option VALUE="HC" >Subfamilies<option VALUE="HP" >Superfamilies<option VALUE="TY" >Type                                        </select>

                                    </td>
                                </tr>
                                <tr>
                                    <td>
                                        <p>
                                        <input TYPE="radio" checked name="LINK3" value="AND"> AND
                                        <input TYPE="radio" name="LINK3" value="OR"> OR
                                        <input type="radio" name="LINK3" value="NOT"> NOT
                                        </p>

                                    </td>
                                </tr>
                                <tr>
                                    <td VALIGN="TOP">
                                        <p>
                                        <input TYPE="Text" NAME="SEARCHTERM3" VALUE="" SIZE="22"><br>
                                        <font face="sans-serif">Search Term 3</font>
                                        </p>

                                    </td>
                                    <td VALIGN="TOP">
                                        <select NAME="FASTSEARCHFIELD3" SIZE="5">
<option VALUE="ALL" >All fields			<option VALUE="AC" >Accession number		<option VALUE="CL" >Classification		<option VALUE="CC" >Comments			<option VALUE="CC0" >Comments:historical		<option VALUE="CC1" >Comments:general		<option VALUE="CC2" >Comments:CONFLICT!		<option VALUE="CC3" >Comments:physiological function<option VALUE="CC4" >Comments:by analogy to	<option VALUE="CC5" >Comments:structure		<option VALUE="CC6" >Comments:experimental inhibitor<option VALUE="CC7" >Comments:kinetics		<option VALUE="CC8" >Comments:mechanism		<option VALUE="CC9" >Comments:methods		<option VALUE="CC10" >Comments:enzyme classification<option VALUE="CC11" >Comments:function<option VALUE="CC12" >Comments:sequence<option VALUE="CC13" >Comments:disease<option VALUE="CX" >Complexes		<option VALUE="ST" >Complex or Modified form of<option VALUE="GE" >Encoding gene<option VALUE="DR" >External database hyperlinks<option VALUE="FO" >Feature of<option VALUE="FP" >Feature position<option VALUE="FT" >Features (motifs)<option VALUE="GO" >GO: biological process, molecular function<option VALUE="IP" >Isoelectric point<option VALUE="CN" >Location negative and experiment(s)<option VALUE="CP" >Location positive and experiment(s)<option VALUE="RX" >Medline database hyperlink<option VALUE="MF" >Modified forms	<option VALUE="NA" SELECTED>Molecule name<option VALUE="XC" >Reaction catalyzed<option VALUE="XA" >Reaction downstream<option VALUE="XI" >Reaction inhibited<option VALUE="XB" >Reaction upstream<option VALUE="RA" >Reference authors<option VALUE="RN" >Reference number<option VALUE="RL" >Reference publication<option VALUE="RT" >Reference title<option VALUE="SQ" >Sequence<option VALUE="SZ" >Sequence length, molecular weight<option VALUE="SC" >Sequence source<option VALUE="OS" >Species<option VALUE="HC" >Subfamilies<option VALUE="HP" >Superfamilies<option VALUE="TY" >Type                                        </select>

                                    </td>
                                </tr>
                                <tr>
                                    <td colspan=2>
                                        <p>
                                        <font FACE="sans-serif" SIZE="-1">
                                        Leave the first search term empty to get a list
                                        of all possible values for the first
                                        selected Quick Search Field
                                        </font>
                                        </p>

                                    </td>
                                    <td>
                                        <font FACE="sans-serif" SIZE="-1">
                                        You can select multiple output fields by using the STRG/CTRL key while clicking
                                        </font>
                                </tr>
                            </table>
                        </td>
                    </tr>

                    <tr HEIGHT=1 BGCOLOR="#800000"><td BGCOLOR="#C0C0C0">&nbsp;</td><td colspan=2 BGCOLOR="#800000">&nbsp;</td></tr>
                    <tr>
                        <td WIDTH="20%" ALIGN="CENTER" VALIGN="MIDDLE" BGCOLOR="#C0C0C0">
                            <p ALIGN="CENTER">
                            <input type="hidden" name="USR" value=XXXUSRXXX>
                            <input type="hidden" name="PASS" value=XXXPASSXXX>
                            <input type="Submit" name="SUBMIT" value="Submit Query"><br>
                            <input type="Reset"  name="RESET"  value="Reset Query">
                            </p>

                        </td>
                        <td WIDTH="3%">
                            &nbsp;
                        </td>
                        <td WIDTH="77%" BGCOLOR="#CCCCCC">
                            <table WIDTH="75%">
                                <tr>
                                    <td VALIGN="TOP">
                                        <font face="sans-serif">Select a previous search you want to refine</font>

                                    </td>
                                    <td VALIGN="TOP">
                                        <font face="sans-serif">and choose an operator</font>
                                    </td>
                                </tr>
                                <tr>
                                    <td VALIGN="TOP">
                                        <select NAME="FILENAME">

<option VALUE="last.search">last.search<option VALUE="analysis">analysis                                        </select>
                                    </td>
                                    <td VALIGN="TOP">
                                        <font face="sans-serif">
                                        <input TYPE="radio" name="LINK1" value="AND"> AND
                                        <input TYPE="radio" name="LINK1" value="OR"> OR
                                        <input type="radio" name="LINK1" value="NOT"> NOT
                                        </font>

                                    </td>
                                </tr>
                                <tr>
                                    <td>
                                        <input type="hidden" name="SEARCHFIELD" value="FAST">
                                        <font face="sans-serif">Store your search as</font>
                                    </td>
                                </tr>

                                <tr>
                                    <td>
                                        <input TYPE="Text" NAME="NEWFILENAME" VALUE="NONE" SIZE="32">
                                    </td>
                                </tr>
                            </table>
                        </td>
                    </tr>
                </table>

            </td>
        </tr>
    </table>
</form>

<p>
<hr>
</p>

</body>
</html>ible values for the first
                                        selected Quick Search Field
                                        </font>
                                        </p>

                                    </td>
                                    <td>
                                        <font FACE="sans-serif" SIZE="-1">
                                        You can select multiple output fields by using the STRG/CTRL key while clicking
                                        </font>
                                </tr>
                            </table>
                        </td>
                    </tr>

                    <tr HEIGHT=1 BGCOLOR="#800000"><td BGCOLOR="#C0C0C0">&nbsp;</td><td colspan=2 BGCOLOR="#800000">&nbsp;</td></tr>
                    <tr>
                        <td WIDTH="20%" ALIGN="CENTER" VALIGN="MIDDLE" BGCOLOR="#C0C0C0">
                            <p ALIGN="CENTER">
                            <input type="hidden" name="USR" value=XXXUSRXXX>
                            <input type="hidden" name="PASS" value=XXXPASSXXX>
                            <input type="Submit" name="SUBMIT" value="Submit Query"><br>
                            <input type="Reset"  name="RESET"  value="Reset Query">
                            </p>

                        </td>
                        <td WIDTH="3%">
                            &nbsp;
                        </td>
                        <td WIDTH="77%" BGCOLOR="#CCCCCC">
                            <table WIDTH="75%">
                                <tr>
                                    <td VALIGN="TOP">
                                        <font face="sans-serif">Select a previous search you want to refine</font>

                                    </td>
                                    <td VALIGN="TOP">
                                        <font face="sans-serif">and choose an operator</font>
                                    </td>
                                </tr>
                                <tr>
                                    <td VALIGN="TOP">
                                        <select NAME="FILENAME">

<option VALUE="last.search">last.search<option VALUE="analysis">analysis                                        </select>
                                    </td>
                                    <td VALIGN="TOP">
                                        <font face="sans-serif">
                                        <input TYPE="radio" name="LINK1" value="AND"> AND
                                        <input TYPE="radio" name="LINK1" value="OR"> OR
                                        <input type="radio" name="LINK1" value="NOT"> NOT
                                        </font>

                                    </td>
                                </tr>
                                <tr>
                                    <td>
                                        <input type="hidden" name="SEARCHFIELD" value="FAST">
                                        <font face="sans-serif">Store your search as</font>
                                    </td>
                                </tr>

                                <tr>
                                    <td>
                                        <input TYPE="Text" NAME="NEWFILENAME" VALUE="NONE" SIZE="32">
                                    </td>
                                </tr>
                            </table>
                        </td>
                    </tr>
                </table>

            </td>
        </tr>
    </table>
</form>

<p>
<hr>
</p>

</body>
</html>
';


#    print OUT "\n</PRE>\n</BODY>\n</HTML>\n";
    close OUT;

    print STDOUT $cgi->redirect("$output_file_html");

    return '';	
}
	    

sub generate_mev_output {

#    die "Sorry - this function is temporarily disabled!\n";

    my $show_command = 0;
#    $show_command = 1;
    if ($show_command) {
	warn "#KH: '$cmd_line'\n";
	open (TMP, ">/tmp/cmd");
	print TMP "perl -d /home/khokamp/FPMI/arraypipe/arraypipe.pl '$cmd_line'\n";
	close TMP;
	exit;
    }
    my $input_file = '';
    unless (defined $cgi->param('formatted_file')) {
	die "No input file specified!";
    }
    $input_file = $cgi->param('formatted_file');
    
    my $output_file = "$tmp_dir/".(basename $input_file).'.mev.html';
    my $output_file_html = "$tmp_dir_html/".(basename $input_file).'.mev.html';

    my $type = '';
    unless (defined $cgi->param('tmev_type')) {
	die "No input type specified!";
    }
    $type = $cgi->param('tmev_type');

    open (IN, "$input_file")
	or die "Can't read from $input_file!\n";
    my $header = <IN>;
    chomp $header;
    my @tmp = split /\t/, $header;
    my $found_column = 0;
    my %columns = ();

    # set up hash that specifies fields indicative of non-annotation columns
    my %annotation_cols = ();
    my %no_annotation = ();
    foreach (@output) {
	next if ($_ eq 'Probe ID'
		 or
		 $_ eq $SPOT_ID_TERM);
	$no_annotation{$_}++;
    }

    my %position_cols = ();
    
    for (my $i = 0; $i <= $#tmp; $i++) {
	if ($tmp[$i] =~ /(.*) ?(ag|au|cy3|cy5|ch1|ch2) $type$/i) {
	    $columns{$1}{$2} = $i;
	    $found_column = 1;
	} elsif ($tmp[$i] =~ /(MROW|MCOL|ROW|COL)/) {
	    $position_cols{$i} = $tmp[$i];
	} else {
	    # if it's an annotation field, keep it
	    my $match = '';
	    if ($tmp[$i] =~ /.* (\w+)$/) {
		$match = $1;
	    }
	    if ($match) {
		if (defined $no_annotation{$match}) {
		    next;
		}
	    }
	    if ($tmp[$i] eq 'line'
		or
		$tmp[$i] eq 'i'
		or
		$tmp[$i] =~ /fold change$/
		or
		$tmp[$i] =~ / Source$/
		or
		$tmp[$i] =~ /^hit /
		or
		$tmp[$i] =~ /^number /
		or
		$tmp[$i] =~ / start$/
		or
		$tmp[$i] =~ / end$/
		) {
		next;
	    }
	    $annotation_cols{$i} = $tmp[$i];
	}
    }

    unless ($found_column) {
	die "No suitable column found (need at least one file with two intensity columns)!";
    }

    # collect info
    my %out = ();
    my %annotation = ();
    my %position = ();
    my $id = 0;
    while (<IN>) {	
	
	# remove comments
	if (/#/) {
	    s/#.*//;
	}

	# skip empty lines
	next unless (/\w/);
		
	# delete trailing white-space
	s/\s*$//;

	# extract required columns from line
	# split at tabs
	my @tmp = split /\t/, $_;

	$id++;

	foreach my $file (keys %columns) {
	    foreach my $ch (keys %{$columns{$file}}) {
		my $pos = $columns{$file}{$ch};
		my $val = $tmp[$pos];
		unless ($val =~ /\d/) {
		    $val = 1;
		}
		if ($type =~ /_L/) {
		    $val = 2**$val;
		}
		$out{$file}{$id}{$ch} = $val;
	    }
	}

	# annotation
	foreach my $pos (keys %annotation_cols) {
	    $annotation{$id}{$pos} = $tmp[$pos];
	}

	# position
	foreach my $pos (keys %position_cols) {
	    $position{$id}{$position_cols{$pos}} = $tmp[$pos];
	}

    }
    close IN;

    unless (%out) {
	die "No element found!";
    }

    my $step = $cgi->param('step');
    my $extension = 0;
    my $appendix = $step;
    while (-e "$tmp_dir/tmev_$appendix.html") {
	$extension++;
	$appendix = "$step.$extension";
    }

    # write MEV files
    my @mev_files = ();
    foreach my $file (sort keys %out) {
	my $file_name = $file;
	if ($file_name eq '') {
	    $file_name = 'tmev';
	}
	my $output_file = "$tmp_dir/${file_name}_$appendix.mev";
	open (OUT, ">$output_file")
	    or die "Can't write to $output_file: $!\n";
	print OUT "# created by ArrayPipe version $revision_number at ".(localtime)."\n";
	print OUT "UID\tIA\tIB\tR\tC\tMR\tMC\n";
	foreach my $id (sort { $a <=> $b } keys %{$out{$file}}) {
	    my $out = "$id\t";
	    foreach my $ch (sort keys %{$out{$file}{$id}}) {
		$out .= "$out{$file}{$id}{$ch}\t";
	    }
	    my ($mcol, $mrow, $col, $row) = (1,1,$id,1);
	    
	    if (defined $position{$id}) {
		if (defined $position{$id}{'MCOL'}) {
		    $mcol = $position{$id}{'MCOL'};
		}
		if (defined $position{$id}{'MROW'}) {
		    $mrow = $position{$id}{'MROW'};
		}
		if (defined $position{$id}{'COL'}) {
		    $col = $position{$id}{'COL'};
		}
		if (defined $position{$id}{'ROW'}) {
		    $row = $position{$id}{'ROW'};
		}
	    }
	    $out .= "$row\t$col\t$mrow\t$mcol\n";	
	    print OUT $out;
	}
	close OUT;
	push @mev_files, basename($output_file);
    }

    # add annotation file:
    if (%annotation) {
	my $output_file = "$tmp_dir/tmev_$appendix.ann";
	open (OUT, ">$output_file")
	    or die "Can't write to $output_file: $!\n";

	# print comments and header
	print OUT "# created by ArrayPipe version $revision_number at ".(localtime)."\n";
	print OUT "UID\tR\tC";
	foreach (sort { $a <=> $b } keys %annotation_cols) {
	    print OUT "\t$annotation_cols{$_}";
	}
	print OUT "\n";

	# print annotation data:
	foreach my $id (sort { $a <=> $b } keys %annotation) {
	    my $out = "$id\t1\t$id";
	    foreach my $pos (sort { $a <=> $b } keys %{$annotation{$id}}) {
		$out .= "\t$annotation{$id}{$pos}";
	    }
	    $out .= "\n";	
	    print OUT $out;
	}
	close OUT;
	push @mev_files, basename($output_file);
    }

	    # create a gzip'ed tar file:
    my $out_file = "$tmp_dir/mev_files_$appendix.tgz";
    my $job = "tar -C $tmp_dir -zcf $out_file '".(join "' '", @mev_files)."'";
#    warn "# KH job: $job";
    system $job;

    # print results page:
    my $results_page = "$tmp_dir/tmev_$appendix.html";
    open (RES, ">$results_page")
	or die "Can't write results page ($results_page): $!";
    print RES &html_head("$progname (TIGR MEV OUTPUT)");
    print RES "<DIV><H1>$progname (TIGR MEV OUTPUT)</H1>\n";
    my $prev_page = '';
    if (defined $cgi->param('prev_page')) {
	$prev_page = $cgi->param('prev_page');
    } else {
	$prev_page = $referral_address;
    }
	    
    print RES "<A HREF=\"$referral_address\">[back to spreadsheet]</A>, <A HREF=\"$prev_page\">[back to ArrayPipe results page]</A><BR>\n";
    print RES "<BR>".(&pointer)."The <B>$type</B> intensities from the following files have been bundled into <a href=\"".(basename $out_file)."\">a compressed tar archive</a> (".(&file_size($out_file)).", right-click and choose to save the file).<BR><OL>\n";
    foreach (sort keys %out) {
	print RES "<LI>".(basename $_)."\n";
    }

    if (%annotation_cols) {
	print RES "</OL>\n<P>An annotation file was added which includes the following information:\n<OL>\n";
	foreach my $pos (sort { $a <=> $b } keys %annotation_cols) {
	    print RES "<LI>$annotation_cols{$pos}\n";
	}
    }
    
    print RES "</OL>\n";
    unless (%position) {
	print RES "<p>Generic spot coordinates were automatically generated because no column 'MCOL', 'MROW', 'COL' or 'ROW' was found.";
    }
    print RES "<P>Download the file and unpack it with <A HREF=\"http://www.winzip.com\">WinZip</A> (Windows) or <FONT COLOR=BLUE><TT>tar -zxvf file</TT></FONT> (Linux), then load into TIGR's <A HREF=\"http://www.tigr.org/software/tm4/mev.html\">TMEV</A> (version 3.0 or higher).</DIV>\n".&ending('start over', 'home', 'tutorial');
    close RES;

    print STDOUT $cgi->redirect("$tmp_dir_html/tmev_$appendix.html");

    return '';	
}
	    

sub plot_barplot {

    # get the column that one wishes to plot:

    my $method = shift;

    my $column = '';
    unless (defined $cgi->param('barplot_column')) {
	die "No column chosen for barplot!";
    }
    $column = $cgi->param('barplot_column');
    if ($column == -1) {
	die "No column chosen for barplot!";
    }

    # locate input file
    my $input_file = '';
    unless (defined $cgi->param('formatted_file')) {
	die "No input file specified!";
    }
    $input_file = $cgi->param('formatted_file');

    # create output file
    my $output_file = $input_file.'.barplot';
    unless (open OUT, ">$output_file") {
	die "Can't write data file ($output_file): $!";
    }

    my $show_lines = 99999999;
    if (defined $cgi->param('barplot_show_lines')) {
	$show_lines = $cgi->param('barplot_show_lines');
	if ($show_lines =~ /all/i) {
	    $show_lines = 99999999;
	} elsif ($show_lines =~ /[a-z]/i) {
	    $show_lines = 0;
	}
    }

    my $hist_breaks = '';
    if (defined $cgi->param('barplot_breaks')) {
	$hist_breaks = $cgi->param('barplot_breaks');
	if ($hist_breaks =~ /\D/i) {
	    $hist_breaks = '';
	}
    }

    my $rounding = '-';
    if (defined $cgi->param('barplot_rounding')) {
	$rounding = $cgi->param('barplot_rounding');
    }

    my $text_input = 0;
    my $range = 'all';
    my $start_line = 0;
    my $end_line = 0;
    if (defined $cgi->param('barplot_range')
	and
	$cgi->param('barplot_range') eq 'visible') {
	$range = 'visible';
	$start_line = $cgi->param('current_line_start');
	$end_line = $cgi->param('current_line_start') + $cgi->param('current_line_numbers') - 1;
    }

    open (IN, $input_file)
	or die "Can't read input from '$input_file': $!";
    my @out = ();
    my %sums = ();
    my $header = <IN>;
    chomp $header;
    my @headers = split /\t/, $header;
    my $column_num = $#headers + 1;
    unless ($column <= $#headers) {
	die "Invalid column ($column) - only $column_num headers!";
    }
    $header = $headers[$column];

    my $line_counter = 0;
    my $skip_empty = 1;
    while (<IN>) {
	chomp;
	$line_counter++;
	if ($range eq 'visible') {
	    next unless ($line_counter >= $start_line);
	    last if ($line_counter > $end_line);
	}
	my @tmp = split /\t/, $_;
#	if (@tmp != $column_num-1) {
#	    if (@tmp != $column_num) {
#		die "Invalid row ($_) - different number of columns (".(scalar @tmp)." instead $column_num)!";
#	    }
#	}
	# the lack of column occurs probably when empty fields are at the end of the line
	# and the split doesn't consider multiple rows of tabs as separate entries...
	while (@tmp != $column_num) {
	    push @tmp, '';
	}

	my $value = $tmp[$column];
#	warn "#: value = '$value'\n";
	if ($value eq ''
	    or
	    $value eq ' ') {

	    # skip empty entries if requested
	    if ($skip_empty) {
		$skip_empty++;
		next;
	    }

	    $value = 'NULL';
	}
	
	if ($rounding =~ /-(\d+)/
	    and
	    $value =~ /\d/) {
	    my $rounding_tmp = -1 * $rounding;
	    $rounding_tmp = 10**$rounding_tmp;
	    $value = ( int ($value / $rounding_tmp) ) * $rounding_tmp;
	} elsif ($rounding =~ /\d/
		 and
		 $value =~ /\.(\d+)/) {
	    if (length($1) > $rounding) {
		$value = sprintf "%.${rounding}f", $value;
	    }
	}
	$sums{$value}++;
	
	# check if we are dealing with text:
	unless ($text_input) {
	    # look for at least two letters in a row
	    if ($value =~ /([a-z]{2,})/i) {
		$text_input = 1;
	    }
	}

#	print OUT "$value\n";
    }

    close IN;

    my $counter = 0;
    my $table_listing = '';
    my @elements = ();
    foreach my $element (sort { $sums{$b} <=> $sums{$a} } keys %sums) {
	push @elements, $element;
	$counter++;
	if ($counter <= $show_lines) {
	    $table_listing .= "<TR><TD ALIGN=RIGHT>$counter</TD><TD>$element</TD><TD ALIGN=RIGHT>$sums{$element}</TD></TR>\n";
	}
	if ($method eq 'Barplot') {
	    print OUT "$element\t$sums{$element}\n";
	} else {
	    for (my $i = 1; $i <= $sums{$element}; $i++) {
		print OUT "$element\n";
	    }
	}
    }

    close OUT;

    # open results page:
    my $step = $cgi->param('step');
    my $tmp_dir = $cgi->param('tmp_dir');
    my $tmp_dir_html = $cgi->param('tmp_dir_html');
    my $extension = 0;
    my $appendix = $step;
    while (-e "$tmp_dir/barplot_$appendix.html") {
	$extension++;
	$appendix = "$step.$extension";
    }
    my $results_page = "$tmp_dir/barplot_$appendix.html";
    open (RES, ">$results_page")
	or die "Can't write results page ($results_page): $!";

    # generate R-code:
    my $R_commands_file = $output_file.'.R';
    my $R_out = $output_file.".$appendix.ps";
    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";
    print RH "postscript(\"$R_out\",horizontal=FALSE, width = ".($IMAGE_HEIGHT/72).", height = ".($IMAGE_WIDTH/72).", paper = \"special\")\n";

    # read in data
    print RH "fm <- read.table(\"$output_file\", sep = \"\\t\")\n";

    # draw barplot
    if ($method eq 'Barplot') {
	print RH "barplot(fm\$V2, names.arg = fm\$V2)\n";
#	print RH "barplot(table(fm\$V1))\n";
    } else {
	my ($min_x,$max_x) = (sort { $a <=> $b } @elements)[0,$#elements];
	if ($hist_breaks =~ /\d/) {
	    $hist_breaks = ", breaks=$hist_breaks";
	}

#	my $max_x = $elements[-1];
	print RH "hist(fm\$V1, col=\"gray\", labels = TRUE, xlab = \"$header\", main = \"\", xlim=c($min_x,$max_x)$hist_breaks)\n";
#	warn "# hist(fm\$V1, col=\"gray\", labels = TRUE, xlab = \"$header\", main = \"\", xlim=c($min_x,$max_x)$hist_breaks)\n";
    }
    print RH "title(main = \"$method of $header\", line=-1)\n";
    print RH "dev.off()\n";
    close RH;

    my $image = &get_R_output($R_commands_file, $R_out, '', '', '', 'landscape');
    
    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	unlink $output_file;
    }

    # print results page:
    print RES &html_head("$progname ($method)");
    print RES "<DIV><H1>$progname ($method)</H1>\n";
    my $prev_page = '';
    if (defined $cgi->param('prev_page')) {
	$prev_page = $cgi->param('prev_page');
    } else {
	$prev_page = $referral_address;
    }
	    
    print RES "<A HREF=\"$referral_address\">[back to spreadsheet]</A>, <A HREF=\"$prev_page\">[back to ArrayPipe results page]</A><BR>\n";
    print RES "\n<H2><BR><A NAME=\"info_section\">Info</A></H2>\n";
    print RES &pointer."Overview of the results:\n<BR><OL>\n";
    print RES "<LI><A HREF=\"#barplot\">$method</A> created with R<BR>\n";
    if ($method ne 'Barplot') {
	print RES "<LI><A HREF=\"#listing1\">Listing 1</A> - elements sorted in alphabetical/numerical order<BR>\n" if ($show_lines > 0);
	print RES "<LI><A HREF=\"#listing2\">Listing 2</A> - elements sorted by occurrences<BR>\n" if ($show_lines > 0);
    } else {
	print RES "<LI><A HREF=\"#listing2\">Listing</A> - elements sorted by occurrences<BR>\n" if ($show_lines > 0);
    }
    print RES "</OL>\n";
    
    print RES "\n<H2><A NAME=\"barplot\">$method</A></H2>\n";
    $skip_empty-- if ($skip_empty);
    if ($skip_empty) {	
	print RES "<B>$skip_empty</B> entries were skipped because they contained no value!<BR><BR>\n";
    }
    print RES "<IMG SRC=\"$image\" ALT=\"$method\">\n";
    if ($show_lines > 0) {
	if ($method ne 'Barplot') {
	    print RES "<H2><A NAME=\"listing1\">Listing 1</A></H2>\n(sorted in same order as in the graph)<BR><BR>\n";
	    print RES "<TABLE BORDER=2 CELLPADDING=4><TR><TH>#</TH><TH>Element</TH><TH>occurrences</TH><TH>accumulation</TH></TR>\n";
	    $counter = 0;
	    my $accum = 0;
	    if ($text_input) {
		foreach (sort keys %sums) {
		    $counter++;
		    last if ($counter > $show_lines);
		    $accum += $sums{$_};
		    print RES "<TR><TD ALIGN=RIGHT>$counter</TD><TD>$_</TD><TD ALIGN=RIGHT>$sums{$_}</TD><TD ALIGN=RIGHT>$accum</TD></TR>\n";
		}
	    } else {
		foreach (sort { $a <=> $b } keys %sums) {
		    $counter++;
		    last if ($counter > $show_lines);
		    $accum += $sums{$_};
		    print RES "<TR><TD ALIGN=RIGHT>$counter</TD><TD>$_</TD><TD ALIGN=RIGHT>$sums{$_}</TD><TD ALIGN=RIGHT>$accum</TD></TR>\n";
		}
	    }
	    print RES "</TABLE>";
	}
	print RES "<H2><A NAME=\"listing2\">Listing ".($method ne 'Barplot' ? '2' : '')."</A></H2>\n(sorted in decreasing order by occurrences)<BR><BR>\n";
	print RES "<TABLE BORDER=2 CELLPADDING=4><TR><TH>#</TH><TH>Element</TH><TH>occurrences</TH></TR>\n";
	print RES $table_listing;
	print RES "</TABLE>";
	print RES "<BR><A HREF=\"$referral_address\">[back to spreadsheet]</A>, <A HREF=\"$prev_page\">[back to ArrayPipe results page]</A><BR>\n";
    }
    print RES "</DIV>\n".&ending('start over', 'home', 'tutorial', 'show_processing_time');
    close RES;

    print STDOUT $cgi->redirect("$tmp_dir_html/barplot_$appendix.html");

    return '';
}


sub input_section {

    my $out = '';
    my $example_info = '';

    # provides section that allows upload of files
    # or selection of predefined example files

    my $found_examples = 0;
    if ($example_dir
	and
	-d $example_dir) {
	# read all files from example directory
	opendir(DIR, $example_dir) || die "can't opendir $example_dir: $!";
	my @example_files = grep { -f "$example_dir/$_" } readdir(DIR);
	closedir DIR;

	# add them to example_files hash
	foreach (@example_files) {
	    if ($_ eq 'info.txt') {
		$example_info = `cat $example_dir/$_`;
		if ($example_info =~ /example_files/s) {
		    $found_examples = 1;
		}
		next;
	    } elsif ($_ =~ /~$/) {
		# backup file - skip
		next;
	    }
	    $example_files{"$example_dir/$_"} = $_;
	}
    }
    
    my $additional_text = '';
    if ($submit eq 'Load Files!') {
	$additional_text = 'additional ';
    }
    $out .= &pointer."\nSpecify ${additional_text}files and lists to be uploaded";
    if (%example_files
	or
	$found_examples) {
	$out .= " and/or pick from the <A HREF=\"#test_files\">example files</A> listed below";
#	$out .= " and/or pick from the <B>example files</B> listed below";
    }
#    $out .= ".&nbsp;$back_to_info<BR>\n";
    $out .= ".<BR>\n";
    $out .= "<H3>Upload:</H3>";
    $out .= "number of fields ".( $cgi ->  textfield({name => 'max_files', default => $max_files, size => 3}) )." &rarr; ".( $cgi -> submit(-name=>'submit',-value=>'Refresh!') ).&help('upload_fields')."<BR><BR>\n";
    
    $out .= "<TABLE CELLPADDING=2>
<TR>
<TH>#</TH>
<TD><CENTER><B>Data Files</B>\&nbsp;".&help('input_files')."</CENTER></TD>\n";

    if ($allow_web_file_load) {
	$out .= "<TD><CENTER><B>Web Files (experimental!)</B>\&nbsp;".&help('web_files')."</CENTER></TD>\n";
    }

    $out .= "<TD><CENTER><B>Reporter Lists</B>\&nbsp;".&help('input_lists')."</CENTER></TD>\n</TR>\n";

    for my $i (1..$max_files) {
	$out .= "<TR>
<TD>$i</TD>
<TD>".$cgi -> filefield({name => "file_name$i", size => 30, override => 1, default => ''})."</TD>\n";
	if ($allow_web_file_load) {
	    $out .= "<TD>".$cgi -> textfield({name => "web_file_name$i", size => 30, override => 1, default => ''})."</TD>\n";
	}
	
	$out .= "<TD>".$cgi -> filefield({name => "reporter_list$i", size => 30, override => 1, default => ''})."</TD>\n</TR>\n";
    }
    $out .= "</TABLE>\n";
	
    $out .= &pointer."Use file compression to reduce upload times!".&help('file_compression')."<BR><BR>\n"; 

    # if we have a MIAMExpress database link, provide info about new and saved experiments and files
    my $db_out = '';
    if ($data_db_setup
	and
	$username ne $public_username) {

	# easier to make it a loop instead of long nested if/else construct
	until ($db_out) {
	    $db_out = "<BR><BR>\n<BIG><B>MIAMExpress files:<BR></B></BIG>\n";
	    # connect to permissions table:
	    require DBI;
	    unless ($data_db_user
		    and
		    $data_db_host
		    and
		    $data_db_type) {
		$db_out .= "Information missing!\n";
		last;
	    }

	    my $dbh;
	    eval { $dbh = DBI->connect("DBI:$data_db_type:database=$data_db_setup;host=$data_db_host", $data_db_user, $data_db_passwd,{ RaiseError => 1, AutoCommit => 0 }); };
	    if ($@
		or
		!defined $dbh) {
		warn "Can't connect to 'DBI:$data_db_type:database=$data_db_setup;host=$data_db_host;user=$data_db_user;passwd=$data_db_passwd':  $@!\n";
		$db_out .= "Can't connect: $@!\n";		
		last;
	    }
		
	    # check if this user is defined			
	    my $statement = "select grp from users where usr = '$username'";
	    my $sth = $dbh->prepare($statement);
	    if ($DBI::errstr) {
		$db_out .= "Can't prepare '$statement': $DBI::errstr";
	    }
	
	    my $rv = $sth->execute();
	    unless ($rv >= 1) {
		$db_out .= "No permissions for user '$username' defined!<BR><BR>";
		$sth->finish;
		$dbh->disconnect;
		last;
	    }

	    %{$data_db{'GROUPS'}} = ();
	    while (my ($grp) = $sth->fetchrow_array) {
		$data_db{'GROUPS'}{$grp}++;
	    }

	    # get uerbergroups for each group
	    $statement = "select grp from groups where sub = ?";
	    $sth = $dbh->prepare($statement);
	    foreach my $grp (keys %{$data_db{'GROUPS'}}) {
		$rv = $sth->execute($grp);
		if ($rv >= 1) {
		    while (my ($ugrp) = $sth->fetchrow_array) {
			$data_db{'GROUPS'}{$ugrp}++;
		    }
		}
	    }

	    # go through the experiment table and pick the ones that were made readable to user
	    $statement = "select mx_exp_id from exp_user_read where name = $username";
	    $sth = $dbh->prepare($statement);
	    $rv = $sth->execute();
	    if ($rv >= 1) {
		while (my ($exp_id) = $sth->fetchrow_array) {
		    $data_db{'EXP_R'}{$exp_id}++;
		    $data_db{'EXP_RU'}{$exp_id}++;
		}
	    }

	    # go through the experiment table and pick the ones that were made readable to one of the groups
	    $statement = "select mx_exp_id from exp_group_read where name = ?";
	    $sth = $dbh->prepare($statement);
	    foreach my $grp (keys %{$data_db{'GROUPS'}}) {
		$rv = $sth->execute($grp);
		if ($rv >= 1) {
		    while (my ($exp_id) = $sth->fetchrow_array) {
			$data_db{'EXP_R'}{$exp_id}++;
			$data_db{'EXP_RG'}{$exp_id}++;
		    }
		}
	    }

	    # get the experiment names and owners:
	    $statement = "select owner, name from exp_owner where mx_exp_id = ?";
	    $sth = $dbh->prepare($statement);
	    foreach my $exp_id (keys %{$data_db{'EXP_R'}}) {
		$rv = $sth->execute($exp_id);
		if ($rv >= 1) {
		    while (my ($name,$owner) = $sth->fetchrow_array) {
			$data_db{'EXP_NAME'}{$exp_id} = $name;
			$data_db{'EXP_OWNER'}{$exp_id} = $owner;
		    }
		}
	    }

	    # check if files for this experiment are available
	    my $prev_files = 0;
	    $statement = "select col from exp_data_type where mx_exp_id = ?";
	    $sth = $dbh->prepare($statement);
	    
	    foreach my $exp_id (keys %{$data_db{'EXP_R'}}) {
		$rv = $sth->execute($exp_id);
		if ($rv >= 1) {
		    # get all the columns
		    while (my ($type) = $sth->fetchrow_array) {
			# get the files for each column			
			my $statement2 = "select $type from FPMI_data.exp_$exp_id where col NOT NULL";
			my $sth2 = $dbh->prepare($statement2);
			while (my ($path) = $sth->fetchrow_array) {
			    my $file = basename($path);
			    $data_db{'EXP_FILES'}{$exp_id}{$file}{$type} = $path;
			    $prev_files++;
			}
		    }
		}
	    }
	    
	    # all previously saved experiments that are allowed to be read by user or the group(s) it belongs to are now retrieved
	    # check for new files in our own MIAMExpress account:
	    # first get the details on where to find the information
	    $statement = "select db,path,mx_user from mx_setup where ap_user = $username";
	    $sth = $dbh->prepare($statement);
	    $rv = $sth->execute();
	    if ($rv >= 1) {
		while (my ($db,$path,$mx_user) = $sth->fetchrow_array) {
		    $data_db{'DB'}{$db}{$mx_user}{$path}++;
		}
	    }

	    if (%{$data_db{'DB'}}) {
		# check if new files are available
		my $new_files = 0;
		foreach my $mx_db (keys %{$data_db{'DB'}}) {
		    my $dbh2;
		    eval { $dbh2 = DBI->connect("DBI:$data_db_type:database=$mx_db;host=$data_db_host", $data_db_user, $data_db_passwd,{ RaiseError => 1, AutoCommit => 0 }); };
		    if ($@
			or
			!defined $dbh2) {
			warn "Can't connect to 'DBI:$data_db_type:database=$mx_db;host=$data_db_host;user=$data_db_user;passwd=$data_db_passwd':  $@!\n";
			$db_out .= "Can't connect to MX_DB $mx_db: $@!\n";		
			next;
		    }

		    # get the submission id
		    $statement = "select TSUBMIS_SYSUID, TSUBMIS_SUB_DESCR, TEXPRMNT_DESCR FROM TSUBMIS,TEXPRMNT WHERE TSUBMIS_LOGIN = ? AND TSUBMIS_DEL_STATUS = 'U' AND TEXPRMNT_SUBID = TSUBMIS_SYSUID AND TEXPRMNT_DEL_STATUS = 'U'";		    
		    $sth = $dbh2->prepare($statement);
		    foreach my $mx_user (keys %{$data_db{'DB'}{$mx_db}}) {
			$rv = $sth->execute($mx_user);
			if ($rv >= 1) {
			    while (my ($id,$name,$descr) = $sth->fetchrow_array) {
				$data_db{'EXP_MX'}{$id}{'NAME'} = $name;
				$data_db{'EXP_MX'}{$id}{'DESCR'} = $descr;
				$data_db{'EXP_MX'}{$id}{'OWNER'} = $mx_user;
			    }
			}
		    }

		    
		    # get all hybridization id's per experiment
		    $statement = "select TLABHYB_HYBRIDID FROM TLABHYB WHERE TLABHYB_SUBID = ? AND TLABHYB_DEL_STATUS = 'U'";		    
		    $sth = $dbh2->prepare($statement);
		    # with this data we could also get the labelling and annotation information
		    # TODO!

		    foreach my $exp (keys %{$data_db{'EXP_MX'}}) {
			$rv = $sth->execute($exp);
			if ($rv >= 1) {
			    while (my ($id) = $sth->fetchrow_array) {
				$data_db{'EXP_MX'}{$exp}{'HYBID'}{$id}++;
			    }
			}
		    }

			
		    # get all files for each experiment's hybridization
		    $statement = "select TFILE_PATH FROM TFILE WHERE TFILE_HYBRIDID = ? AND TFILE_DEL_STATUS = 'U'";		    
		    $sth = $dbh2->prepare($statement);

		    foreach my $exp (keys %{$data_db{'EXP_MX'}}) {
			foreach my $hybid (keys %{$data_db{'EXP_MX'}{$exp}{'HYBID'}}) {
			    $rv = $sth->execute($hybid);
			    if ($rv >= 1) {
				while (my ($file) = $sth->fetchrow_array) {
				    # check if file exists
				    my $owner = $data_db{'EXP_MX'}{$exp}{'OWNER'};
				    my $path = $data_db{'DB'}{$mx_db}{$owner};
				    $path .= "/$owner/submission$exp/hybrid$hybid/$file";
				    unless (-e $path) {
					warn "# KH: file does not exist: $path";
					next;
				    }
				    unless (-s $path) {
					warn "# KH: file is empty: $path";
					next;
				    }
				    unless (-r $path) {
					warn "# KH: can't read file: $path";
					next;
				    }
				    # if this file has been saved already, we don't need to do anything, otherwise offer it for processing
				    if (defined $data_db{'EXP_FILES'}{$exp}{$file}) {
					next;
				    }
				    $data_db{'EXP_MX_NEW'}{$exp}{$file} = $path;
				    $new_files++;
				}
			    }
			}
		    }
		    $dbh2->disconnect;
		}

		if ($new_files) {
		    $db_out .= "<BR>New MIAMExpress files:<BR>\n<OL>\n";
		    foreach my $exp (sort keys %{$data_db{'EXP_MX_NEW'}}) {
			$db_out .= "<LI><input type=\"checkbox\" name=\"new_mx_exp\" value=\"$exp\">$data_db{'EXP_MX'}{$exp}{'NAME'}:\n<OL>\n";
			foreach my $file (sort keys %{$data_db{'EXP_MX_NEW'}{$exp}}) {
			    $db_out .= "<LI><input type=\"checkbox\" name=\"new_mx_exp_files\" value=\"$data_db{'EXP_MX_NEW'}{$exp}{$file}\">$file\n";
			}
			$db_out .= "</OL>\n";
		    }
		    $db_out .= "</OL>\n";
		}		    
	    }

	    $dbh->disconnect;
	
	    # report previously saved files:
	    if ($prev_files) {
		$db_out .= "<BR>Previously saved MIAMExpress experiments:<BR>\n<OL>\n";
		foreach my $exp (sort keys %{$data_db{'EXP_FILES'}}) {
		    $db_out .= "<LI><input type=\"checkbox\" name=\"prev_mx_exp\" value=\"$exp\">$data_db{'EXP_NAME'}{$exp} ($data_db{'EXP_OWNER'}{$exp}):\n<OL>\n";
		    foreach my $file (sort keys %{$data_db{'EXP_FILES'}{$exp}}) {
			my @types = sort keys %{$data_db{'EXP_FILES'}{$exp}{$file}};
			
			$db_out .= "<LI>$file\n";
			foreach my $type (@types) {
			    $db_out .= "<INPUT TYPE=RADIO name=\"${exp}_files\" value=\"$data_db{'EXP_FILES'}{$exp}{$file}{$type}\">$type\n";
			}
		    }
		    $db_out .= "</OL>\n";
		}
		$db_out .= "</OL>\n";
	    }		    
	}
	$out .= $db_out
    }	    
    
#    $out .= "<BR>\n<BIG><B><A NAME=\"example_files\">Example files:</A></B></BIG>&nbsp;".&help('example_files')."<BR>\n$example_info\n";
    $out .= "<BR>\n<BIG><B><A NAME=\"test_files\">Example files:</A></B></BIG>&nbsp;".&help('example_files')."<BR>\n$example_info\n";

    # list example files (if specified/found):
    if (%example_files) {
	
#	$out .= 'Tick the boxes of those files you would like to work on.<BR>';
	my @example_files = ();
	my %example_files_label = ();
	unless ($example_dir_html) {
	    %example_files_label = %example_files;
	}
	my @defaults = ();
	my @tmp = sort keys %example_files;
	# filter out those that don't have the correct format
	foreach my $file (@tmp) {
	    my $format_check = 'ok';
	    ($format_check) = &check_file_format($file) if ($check_example_files);
	    unless ($format_check eq 'ok') {
		$wrong_format_files{$file} = $format_check;
	    } else {
		push @example_files, $file;
		if ($example_dir_html) {
		    $example_files_label{$file} = " <A HREF=\"$example_dir_html/".(basename $file)."\">[download]</A>";
		}
		# set hidden parameter
		my $size = (stat $file)[7];
		$out .= "<input type=\"hidden\" name=\"example_$file\" value=\"$size\">\n";	
		$out .= "<input type=\"checkbox\" name=\"example_files\" value=\"$file\"> ".(basename $file);
		if (defined $example_files_label{$file}) {
		    $out .= $example_files_label{$file};
		}
		$out .= "<BR>\n";
	    }
	}
#	$out .= $cgi -> checkbox_group(-name=>'example_files',
#				     -values=>[@example_files],
#				     -linebreak=>'true',
#				     -default=>[@defaults],
#				     -labels=>\%example_files);
    } else {
	$out .= "Sorry - no examples available. Ask your operator (<A HREF=\"mailto:$operator_email\">$operator_email</A>) for it!\n" unless ($found_examples);
    }


    $out .= '<BR><TABLE><TR><TD><TABLE CELLPADDING=5><TR><TD BGCOLOR=BLUE><A NAME="load_files_button">'.
	$cgi -> submit(-name=>'submit', 
		       -value=>'Load Files!')
	.'</A>';
    $out .= '</TD></TR></TABLE>';
    if ($cgi->param('work_files')) {
	$out .= '</TD><TD>'.$cgi -> submit(-name=>'submit', 
					-value=>'Load Files and Do it!')
	    . &help('load_files_and_do_it');
    }
    $out .= '</TD></TR></TABLE>';

    if (%wrong_format_files) {
	$out .= "<BR>
               <H3><FONT COLOR=RED>Invalid files:</FONT></H3>
              ";
	my @invalid_files = sort keys %wrong_format_files;
	$out .= $cgi -> checkbox_group(-name=>'invalidfiles',
				     -values=>[@invalid_files],
				     -linebreak=>'true',
				     -labels=>\%example_files);
	$out .= '<BR><BR>'.$cgi -> submit(-name=>'submit', 
					-value=>'Remove files!');
	
    }
    %wrong_format_files = ();			     

    if ($allow_debug) {
	$out .= "\n<BR><BR><TABLE BORDER=0>
<TR>
<TD COLSPAN=2>
<BR>For debugging only:
</TD>
</TR>
<TR>
<TD>&nbsp;&nbsp;Limit to</TD><TD>".($cgi ->  textfield(-name => 'spot_limit', 
						       -default => $spot_limit, 
						       -size => 6))."values</TD>
</TR>
<TR>
<TD>&nbsp;&nbsp;Starting from</TD><TD>".($cgi ->  textfield(-name => 'spot_limit_start', 
							    -default => $spot_limit_start, 
							    -size => 6))."th line</TD>
</TR>
<TR>
<TD ALIGN=RIGHT>or Subgrid</TD><TD>".($cgi ->  textfield(-name => 'subgrid_start', 
							 -default => $subgrid_start, 
							 -size => 6))."</TD>
</TR>
<TR>
<TD ALIGN=RIGHT>to</TD><TD>".($cgi ->  textfield(-name => 'subgrid_end', 
						 -default => $subgrid_end, 
						 -size => 6))."</TD>
</TR>
<TR>
<TD ALIGN=RIGHT>or Spot ID</TD><TD>".($cgi ->  textfield(-name => 'spotid_filter', 
							 -default => $spotid_filter, 
							 -size => 15))."</TD>
</TR>
</TABLE>
<BR>
";
    }

    return $out;
}


sub simple_background_correction_setup {
    my $num = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my @files = ();
    my @files_select = &get_user_settings($var_def_default{'fg_min_bg_files'},'fg_min_bg_files',$force_global_opt,$num);

    foreach (@files_select) {
	push @files, $_;
	if ($_ eq 'all') {
	    @files = ();
	    foreach (sort keys %data) {
		if (/\w/) {
		    push @files, $_;
		}
	    }
	    last;
	}
    }

    # carry out action on each file ...
    foreach my $file (@files) {
	
	# ... and each label
	foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	    
	    # skip columns that are not label specific ('spot labels')
	    next if ($label eq '__NO_LABEL__');
	    
	    my ($info);
	    my $bg_channel = $BG_SIGNAL_TERM;
	    if ($refine_parameters) {
		$info = 'Parameter Refinement!<BR>';
	    } else {
		if (defined $data_column_available{$file}{$label}{$BG_SIGNAL_TERM.'_M'}) {
		    $bg_channel = $BG_SIGNAL_TERM.'_M';
		}
		$info = &simple_background_correction($num,$file,$label,$bg_channel,$FG_COR_SIGNAL_TERM, $action);
	    }

	    # format output (include image and choice for parameters)
	    print &format_output( #'fg_min_bg_',$num,$file,$label,$action,'',$info);
				  {
				      'prefix' => 'fg_min_bg_',
				      'num' => $num,
				      'file' => $file,
				      'label' => $label,
				      'action' => $action,
				      'info' => $info
				      }
				  );	}
    }
}


sub simple_background_correction {
    my $action_num = shift;
    my $file = shift;
    my $label = shift;
    my $header2 = shift;
    my $out_header = shift;
    my $action = shift;
    my ($max_val,$min_val);

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $header1 = '';
    foreach my $column (@{$item_data_column{$action}}) {
	if ($data_column_available{$file}{$label}{$column}) {
	    $header1 = $column;
	    last;
	}
    }
    # check for specific settings:
    $header1 = &get_user_settings($header1,'fg_min_bg_data_column',$force_global_opt,$action_num,$file,$label);
    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'fg_min_bg_skip_flags'},'fg_min_bg_skip_flags',$force_global_opt,$action_num,$file,$label);
    %skip_flags = &get_skip_flags(@flags);
    my %skipped = ();

    unless ($header1) {
	return "Error: No data column defined to work on for $action!\n";
    }
 
    if ($header1 =~ /_M/) {
	$out_header = $out_header.'_M';
    }
    $last_column_used = $out_header;

    &read_data($file,$header1,$header2);

    # check for specific settings:
    my $operation = &get_user_settings($var_def_default{'fg_min_bg_operator'},'fg_min_bg_operator',$force_global_opt,$action_num,$file,$label);

#    warn "received operator '$operation'";

    my ($mrow,$mcol,$row,$col,$data1,$data2,$new_data);
    my $min_value = 9999999;
    my $low_values = 0;

    # transform data
    foreach $mrow (keys %{$data{$file}{'__DATA__'}{$label}{$header1}}) {
	foreach $mcol (keys %{$data{$file}{'__DATA__'}{$label}{$header1}{$mrow}}) {
	    foreach $row (keys %{$data{$file}{'__DATA__'}{$label}{$header1}{$mrow}{$mcol}}) {
		foreach $col (keys %{$data{$file}{'__DATA__'}{$label}{$header1}{$mrow}{$mcol}{$row}}) {

		    # skip spots without value
		    unless (defined $data{$file}{'__DATA__'}{$label}{$header1}{$mrow}{$mcol}{$row}{$col}
			    and
			    defined $data{$file}{'__DATA__'}{$label}{$header2}{$mrow}{$mcol}{$row}{$col}) {
			$skipped{'value not numerical'}++;
			next;
		    }

		    # skip flagged spots
		    my $skip = 0;
		    if (%skip_flags
			and
			(defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}
			or
			defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col})) {
			my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
			if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			    push @flags, &get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
			}
			foreach (@flags) {
			    if (defined $skip_flags{$_}) {
				$skip = 1;
				last;
			    }
			}
		    }			
		    
		    if ($skip) {
			$skipped{'spot was flagged'}++;
			next;
		    }

		    $data1 = $data{$file}{'__DATA__'}{$label}{$header1}{$mrow}{$mcol}{$row}{$col};
		    $data2 = $data{$file}{'__DATA__'}{$label}{$header2}{$mrow}{$mcol}{$row}{$col};
		    if ($operation eq '-') {
			$new_data = $data1 - $data2;
		    } elsif ($operation eq '/') {
			if ($data2 == 0) {
			    $new_data = '';
			} else {
			    $new_data = $data1 / $data2;
			}
		    }
		    if ($new_data < $min_value) {
			$min_value = $new_data;
		    }
		    if ($new_data <= 0) {
			$low_values++;
			my $skip_flagging = 0;
			# check if floor flag is set already
			if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}
			    and
			    $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} >= $flag_values{'floor'}) {
			    my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
			    foreach (@flags) {
				if ($_ == $flag_values{'floor'}) {
				    $skip_flagging++;
				    last;
				}
			    }
			}
			$data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'floor'} unless ($skip_flagging);
		    }
		    $data{$file}{'__DATA__'}{$label}{$out_header}{$mrow}{$mcol}{$row}{$col} = $new_data;
		}
	    }
	}
    }
    # take note of available columns for later

    unless (defined $data_column_available{$file}{$label}{$out_header}) {
	$data_column_available{$file}{$label}{$out_header} = 1;
	push @{$data_column_available{$file}{$label}{'__ORDER__'}}, $out_header;
    }
    $data_done{$file}{$label}{$out_header}++;

    my $out = "Background correction carried out as <B>'$header1 $operation $header2'</B>, values <= 0: <B>$low_values</B>, lowest corrected value = <B>$min_value</B>.";
    
    if (%skipped) {
	$out .= "<BR>The following data points were skipped:<BR><UL>\n";
	foreach (sort keys %skipped) {
	    $out .= "<LI>$_ -> $skipped{$_} entries</LI><BR>\n";
	}
	$out .= "</UL>\n";
    }
	
    return $out;
}


sub merge_dups {
    my $action_num = shift;
    my $action = shift;
    my $file = shift;
    my $spot_label = shift;
    my $flag_header = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);

    my $header = '';
  OUTCOL1:foreach my $column (@{$item_data_column{$action}}) {
      # sort labels reversed lexically 
      # (so '__NO_LABEL__' gets checked first. This way we prioritize ratios)
      foreach my $label (sort {$b cmp $a} keys %{$data{$file}{'__DATA__'}}) {
	  if ($data_column_available{$file}{$label}{$column}) {
	      $header = $column;
	      last OUTCOL1;
	  }
      }
  }
    # check for specific settings:
    my @header = &get_user_settings($header,'merge_dups_data_column',$force_global_opt,$action_num,$file,'__ALL__');
    $header = $header[0];

    my $use_label = '';

    unless ($header) {
	return "Error: No data column defined to work on for $action!\n";
    }
    
    my $spread_type = 'ratio';
    my $other_header = $header;

    unless ($header =~ /_R/) {
	$spread_type = 'intensities';
    } else {
	$other_header =~ s/_R//;
    } 
    my $other_header_out = $other_header.'_M';

    if ($header =~ /_L/) {
	$spread_type = 'log-'.$spread_type;
    }

    # get the labels to work on:
    my %labels = ();
    foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	if ($data_column_available{$file}{$label}{$header}) {
	    $labels{$label}++;
	}
    }	

    my @labels = keys (%labels);

    my $out_header = $header.'_M';

    $last_column_used = $out_header;

    if ($refine_parameters) {
	foreach my $label (@labels) {
	    $data_done{$file}{$label}{$out_header}++;
	    $data_column_available{$file}{$label}{$out_header}++;
	    print &format_output( # 'merge_dups_',$action_num,$file,$label,$action,'-','Parameter Refinement!<BR>'
				 {
				      'prefix' => 'merge_dups_',
				      'num' => $action_num,
				      'file' => $file,
				      'label' => $label,
				      'action' => $action,
				      'image' => '-',
				      'info' => 'Parameter Refinement!<BR>'
				      }
				  );
	}
	return;
    }

    # read in data if necessary
    my $rc = &read_data($file,$header);
    unless ($rc eq 'ok') {
	return "<BR><B><FONT COLOR=RED>WARNING: $rc</FONT></B><BR>\n";
    }

    # check for user specific settings
    my $target = &get_user_settings($var_def_default{'merge_dups_target'},'merge_dups_target',$force_global_opt,$action_num,$file,'__ALL__');

    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'merge_skip_flags'},'merge_skip_flags',$force_global_opt,$action_num,$file,'__ALL__');
    %skip_flags = &get_skip_flags(@flags);
    
    my ($mrow,$mcol,$row,$col);

    my %prev = ();

    my %lines = ();

    # collect values for each spot label
    foreach my $label (sort @labels) {

	my $bg_header = '';
	my $out_bg_header = '';
	if (defined $data_column_available{$file}{$label}{$BG_SIGNAL_TERM}) {
	    $bg_header = $BG_SIGNAL_TERM;
	    $out_bg_header = $bg_header.'_M';
	}

	my $counter = 0;

	my %values = ();
	my %other_values = ();
	my %bg_values = ();

	foreach $mrow (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}}) {
	    foreach $mcol (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}}) {
		foreach $row (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}}) {
		    foreach $col (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}}) {
		    
			# initialize output field:
			$data{$file}{'__DATA__'}{$label}{$out_header}{$mrow}{$mcol}{$row}{$col} = '';
			$data{$file}{'__DATA__'}{$label}{$out_header.'_i'}{$mrow}{$mcol}{$row}{$col} = '';

			my $spot = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$spot_label}{$mrow}{$mcol}{$row}{$col};
			my $value = '';
			
			# skip spots without value
			if ($data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}	eq ''
			    or
			    !defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}) {
			    next;
			}
			
			# skip flagged spots
			my $skip = 0;
			if (%skip_flags
			    and
			    (defined $data{$file}{'__DATA__'}{$label}{$flag_header}{$mrow}{$mcol}{$row}{$col}
			    or
			    defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$flag_header}{$mrow}{$mcol}{$row}{$col})) {
			    my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$flag_header}{$mrow}{$mcol}{$row}{$col});
			    if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$flag_header}{$mrow}{$mcol}{$row}{$col}) {
				push @flags, &get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$flag_header}{$mrow}{$mcol}{$row}{$col});
			    }
			    foreach (@flags) {
				if (defined $skip_flags{$_}) {
				    $skip = 1;
				    last;
				}
			    }
			}			

			if ($skip) {			
			    next;
			} else {
			    $value = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
			}

			my $bg_value = '';
			if ($bg_header
			    and
			    defined $data{$file}{'__DATA__'}{$label}{$bg_header}{$mrow}{$mcol}{$row}{$col}) {
			    $bg_value = $data{$file}{'__DATA__'}{$label}{$bg_header}{$mrow}{$mcol}{$row}{$col};
			}
			
			# special settings for Arrayvision:
			# delete numerator placed in brackets behind each gene name
			$spot =~ s/ \(\d+\)$// if (defined $file_specs{$file}{'FORMAT'} 
						   and
						   $file_specs{$file}{'FORMAT'} eq 'ArrayVision');
			
			if ($target eq 'whole slide') {
			    ($spot) = split /\^/, $spot, 2;
			}
			
			# create a single key for the same reporter in two consecutive lines
			# calculate line number
			my $line = ($row - 1 + ($mrow - 1) * $ROW) * $MCOL*$COL + ($mcol - 1) * $COL + $col;
			$lines{$line} = "$mrow,$mcol,$row,$col";
			
			my ($key,$prev);
			
			if (defined $prev{$spot}
			    and
			    $prev{$spot} >= 0) {
			    $prev = $prev{$spot};
			    if ($target eq 'whole slide') {
				$key = $prev;
			    } else {
				if ($prev == $line - 1) {
				    # we found the neighbouring duplicate!
				    $key = $prev;
				    # reset %prev entry for this spot
				    $prev{$spot} = -1;
				} else {
				    # the previous was somewhere else
				    $key = $line;
				    $prev{$spot} = $line;
				}
			    }
			} else {
			    # this is the first of a pair
			    $key = $line;
			    $prev{$spot} = $line;
			}
			
			unless ($value =~ /\d/) {
			    next;
			}

			# form an array of values for each spot under each label
			push @{$values{$key}{$spot}}, $value;
			push @{$bg_values{$key}{$spot}}, $bg_value;

			if ($label eq '__NO_LABEL__') {
			    # collect data from other channels as well
			    foreach my $ch ('CH1', 'CH2') {
				push @{$other_values{$key}{$spot}{$ch}}, $data{$file}{'__DATA__'}{$ch}{$other_header}{$mrow}{$mcol}{$row}{$col};
			    }
			}

			$counter++;
		    }
		}
	    }
	}

	# TODO: unless specificly requested, always merge on ratios!
	# e.g.: if one intensity is zero, this would not yield a ratio
	# but it can be merged with another intensity into a new value, 
	# which seems wrong!
	# Also, the two intensities for one spot are closer related to each
	# other than the intensities between two spots, therefore it would
	# be more natural to first calculate the ratio and then merge these
	# instead of merging the intensities and then calculate the ratios!

	my $counter2 = 0;
	my $warnings = '';
      LINE1:foreach my $line (keys %values) {
	  my $merge_val = '';
	  my $ch1_merge_val = '';
	  my $ch2_merge_val = '';
	  my $spread_val = '';
	  my @tmp = keys %{$values{$line}};
	  my $num = 0;
	  my ($mrow,$mcol,$row,$col) = split /,/, $lines{$line};
	  if (@tmp > 1) {
	      #error!
	      warn "#arraypipe: ".(scalar @tmp)." values for $line -> @tmp!";
	      $warnings .= "<B>".(scalar @tmp)." values for $line -> @tmp!</B><BR>\n";
	      $merge_val = '';
	  } else {
	      my $spot = $tmp[0];
	      my @tmp2 = ();
	      foreach (@{$values{$line}{$spot}}) {
		  if (/\d/) {
		      $num++;
		      push @tmp2, $_;
		  }
	      }
		  
	      if ($num < 1) {
		  # all values must have been flagged
		  $merge_val = '';
	      } elsif ($num == 1) {
		  $merge_val = $tmp2[0];
		  $spread_val = "0|$spread_type|$merge_val";
	      } else {
		  $merge_val = &median(@tmp2);
		  my @tmp3 = sort { $a <=> $b } @tmp2;
#		  $spread_val = abs($tmp3[-1] - $tmp3[0]);
		  $spread_val = &variance(@tmp2);
		  $spread_val = "$spread_val|$spread_type|".(join '|', @tmp2);
	      }	   

	      my $ch1_merge_value = '';
	      if (%other_values
		  and
		  defined $other_values{$line}) {

		  my @other_tmp = ();
		  my $other_num = 0;

		  foreach (@{$other_values{$line}{$spot}{'CH1'}}) {
		      if (/\d/) {
			  $other_num++;
			  push @other_tmp, $_;
		      }
		  }
		  if ($other_num < 1) {
		      # all values must have been flagged
		      $ch1_merge_val = '';
		  } elsif ($other_num == 1) {
		      $ch1_merge_val = $other_tmp[0];
		  } else {
		      $ch1_merge_val = &median(@other_tmp);
		  }	   
	      
		  if ($header =~ /_L/) {
		      $ch2_merge_val = $ch1_merge_val + $merge_val
		  } else {
		      $ch2_merge_val = $merge_val * $ch1_merge_val;
		  }	      
		  $data{$file}{'__DATA__'}{'CH1'}{$other_header_out}{$mrow}{$mcol}{$row}{$col} = $ch1_merge_val;
		  $data{$file}{'__DATA__'}{'CH2'}{$other_header_out}{$mrow}{$mcol}{$row}{$col} = $ch2_merge_val;
	      }   
	  }
    
	  # insert merged data
	  $data{$file}{'__DATA__'}{$label}{$out_header}{$mrow}{$mcol}{$row}{$col} = $merge_val;
	  $data{$file}{'__DATA__'}{$label}{'SPREAD'}{$mrow}{$mcol}{$row}{$col} = $spread_val;
	  $spread_done{$file}{$label}++;
	  $data{$file}{'__DATA__'}{$label}{$out_header.'_i'}{$mrow}{$mcol}{$row}{$col} = $num;
	  $counter2++ if ($num > 0);

	  if ($bg_header) {
	      # same for background values
	      @tmp = keys %{$bg_values{$line}};
	      $num = 0;
	      if (@tmp > 1) {
		  #error!
		  warn "#arraypipe: ".(scalar @tmp)." values for $line -> @tmp!";
		  $warnings .= "<B>".(scalar @tmp)." values for $line -> @tmp!</B><BR>\n";
		  $merge_val = '';
	      } else {
		  my $spot = $tmp[0];
		  my @tmp2 = ();
		  foreach (@{$bg_values{$line}{$spot}}) {
		      if (/\d/) {
			  $num++;
			  push @tmp2, $_;
		      }
		  }
		  if ($num < 1) {
		      # all values must have been flagged
		      $merge_val = '';
		  } elsif ($num == 1) {
		      $merge_val = $tmp2[0];
		  } else {
		      $merge_val = &median(@tmp2);
		  }
		  
	      }
	      # insert merged data
	      ($mrow,$mcol,$row,$col) = split /,/, $lines{$line};
	      $data{$file}{'__DATA__'}{$label}{$out_bg_header}{$mrow}{$mcol}{$row}{$col} = $merge_val;
	      $data{$file}{'__DATA__'}{$label}{$out_bg_header.'_i'}{$mrow}{$mcol}{$row}{$col} = $num;
	  }
      }
	
	# take note of available columns for later
	if ($bg_header) {
	    unless (defined $data_column_available{$file}{$label}{$out_bg_header}) {
		$data_column_available{$file}{$label}{$out_bg_header} = 1;
		push @{$data_column_available{$file}{$label}{'__ORDER__'}}, $out_bg_header;
	    }
	    $data_done{$file}{$label}{$out_bg_header}++;

	    unless (defined $data_column_available{$file}{$label}{$out_bg_header.'_i'}) {
		$data_column_available{$file}{$label}{$out_bg_header.'_i'} = 1;
		push @{$data_column_available{$file}{$label}{'__ORDER__'}}, $out_bg_header.'_i';
	    }
	    
	    $data_done{$file}{$label}{$out_bg_header.'_i'}++;
	}

	if (%other_values) {
	    foreach my $ch ('CH1', 'CH2') {
		unless (defined $data_column_available{$file}{$ch}{$other_header_out}) {
		    $data_column_available{$file}{$ch}{$other_header_out} = 1;
		    push @{$data_column_available{$file}{$ch}{'__ORDER__'}}, $other_header_out;
		}
		
		$data_done{$file}{$ch}{$other_header_out}++;	
	    }
	}

	unless (defined $data_column_available{$file}{$label}{$out_header}) {
	    $data_column_available{$file}{$label}{$out_header} = 1;
	    push @{$data_column_available{$file}{$label}{'__ORDER__'}}, $out_header;
	}

	$data_done{$file}{$label}{$out_header}++;
	
	unless (defined $data_column_available{$file}{$label}{$out_header.'_i'}) {
	    $data_column_available{$file}{$label}{$out_header.'_i'} = 1;
	    push @{$data_column_available{$file}{$label}{'__ORDER__'}}, $out_header.'_i';
	}

	$data_done{$file}{$label}{$out_header.'_i'}++;
	
#	unless (defined $data_column_available{$file}{$label}{'SPREAD'}) {
#	    $data_column_available{$file}{$label}{'SPREAD'} = 1;
#	    push @{$data_column_available{$file}{$label}{'__ORDER__'}}, 'SPREAD';
#	}
#
#	$data_done{$file}{$label}{'SPREAD'}++;
	
	my $out = "$warnings<BR>Merged <B>$counter</B> duplicated values from '$header'<BR>into <B>$counter2</B> single ones in '$out_header'"
	    . ($bg_header ? " and '$out_bg_header'" : '')
	    . " for label '$label'.";
	print &format_output( # 'merge_dups_',$action_num,$file,'-',$action,'',$out
			     {
				 'prefix' => 'merge_dups_',
				 'num' => $action_num,
				 'file' => $file,
				 'label' => '-',
				 'action' => $action,
				 'info' => $out
				 }
			     );
    }
    return '';
}

    
sub median {
    my @order = sort { $a <=> $b } @_;
    my $md = '';

    if ($#order % 2 == 0){
        $md = $order[$#order / 2];
    } else {
	my $index1 = int($#order / 2);
	my $index2 = int($#order / 2) + 1;
	my $low = $order[$index1];
	my $high = $order[$index2];
	$md = $low + ($high - $low)/2;
    }
    return $md;
}

    
sub compare_lists_setup {
    my $action_num = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my @files = &get_user_settings($var_def_default{'compare_lists_files'},'compare_lists_files',$force_global_opt,$action_num);

    foreach (@files) {
	if ($_ eq 'separately') {
	    @files = ();
	    last;
	}
	if ($_ eq 'all') {
	    @files = (sort keys %data);
	    last;
	}	
    }
    
    my @files_tmp = ();
    foreach (@files) {
	next if (defined $work_files_skip{$_});
	push @files_tmp, $_;
    }
    @files = @files_tmp;

    # carry out action on file(s)
    unless (@files) {
	foreach my $file (sort keys %data) {
	    next if (defined $work_files_skip{$file});
	    my $out = &compare_lists($action_num,$action,$file);
	    print &format_output( # 'compare_lists_',$action_num,$file,'__ALL__',$action,'',$out
				 {
				      'prefix' => 'compare_lists_',
				      'num' => $action_num,
				      'file' => $file,
				      'label' => '__ALL__',
				      'action' => $action,
				      'info' => $out
				      }
				  );
	}
    } else {
	my $out = &compare_lists($action_num,$action,@files);
	print &format_output( # 'compare_lists_',$action_num,$files[0],'__ALL__',$action,'',$out
			     {
				 'prefix' => 'compare_lists_',
				 'num' => $action_num,
				 'file' => $files[0],
				 'label' => '__ALL__',
				 'action' => $action,
				 'info' => $out
				 }
			     );
    }

}


sub compare_lists {
    my $action_num = shift;
    my $action = shift;
    my @files = @_;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }
    
    my $make_new_list = &get_user_settings($var_def_default{'compare_lists_make_new'},'compare_lists_make_new',$force_global_opt,$action_num,$files[0]);
    unless ($make_new_list eq 'yes') {
	$make_new_list = '';
    }

    my @selected_flags = &get_user_settings($var_def_default{'compare_lists_flags'},'compare_lists_flags',$force_global_opt,$action_num,$files[0]);
    my %flag_select = ();
    my %used_flags = ();
    foreach (@selected_flags) {
	if ($_ eq 'all') {
	    @selected_flags = ();
#	    foreach my $item (@{$var_def_values{'compare_lists_flags'}}) {
            foreach my $item (@all_flags) {
		next if ($item =~ /^all/);
		push @selected_flags, $item;
	    }
	} elsif ($_ eq 'all lists') {
	    @selected_flags = ();
#	    foreach my $item (@{$var_def_values{'compare_lists_flags'}}) {
            foreach my $item (@all_flags) {
		next unless ($item =~ /^list/);
		push @selected_flags, $item;
	    }
	}
    }

    # remove duplicates but preserve order
    @selected_flags = &remove_dups_preserve_order(@selected_flags);

    foreach (@selected_flags) {
	$flag_select{$_}++;
    }

    my ($mrow,$mcol,$row,$col);
    my %codes = ();
    my %spot_nums = ();

    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@files);
    foreach $mrow (1..$MROW) {
	foreach $mcol (1..$MCOL) {
	    foreach $row (1..$ROW) {
		foreach $col (1..$COL) {
		    # skip this unless at least spot is defined
		    my $skip = 1;
		    my %flags = ();
		    foreach my $file (@files) {
			# collect flags for each file
			my @flags = ();
			foreach my $label (keys %{$data{$file}{'__DATA__'}}) {
			    if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
				push @flags, (&get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
				$skip = 0;
			    }
			}
			unless (@flags) {
			    $flags{"$file|no flag"}++;
			    $spot_nums{$file}{"$mrow|$mcol|$row|$col"}++;
			} else {
			    foreach (@flags) {
				$flags{"$file|$flag_values_num{$_}"}++;
				$spot_nums{$file.$flag_values_num{$_}}{"$mrow|$mcol|$row|$col"}++;
			    }
			}
		    }
		    
		    # no flag defined here, skip this spot
		    next if ($skip);

		    # set up the code for this spot
		    my $code = '';
		    foreach my $file (@files) {
			foreach my $flag (@selected_flags) {
			    if (defined $flags{"$file|$flag"}) {
				$code .= '1';
				$used_flags{"$file|$flag"}++;
			    } else {
				$code .= '0';
			    }
			}
		    }
		    $codes{$code}{"$mrow|$mcol|$row|$col"}++;
		}
	    }
	}
    }

    # compile output:
   
    my $out = '';

	# list flags and lists
	foreach my $flag (@selected_flags) {
		foreach my $file (@{$list_flag_match{'FLAG'}{$flag}}) {
			my $num = $list_flag_match{'ENTRIES'}{$file};
			$out .= "$flag = ".(basename $file)." ($num entries)<BR>\n";
		}
	}

    # file short cuts
    my %shorts = ();
    if (@files > 1) {	
	$out .= "\nFile short cuts:\n<UL>\n";
	my $counter = 0;
	foreach my $file (@files)  {
	    $counter++;
	    $shorts{$file} = "F$counter";
	    my $file_out = (basename $file)." ";
	    $out .= "<LI>F$counter = $file_out<BR></LI>\n";
	}
	$out .= "</UL>\n";
    }

    my %flag_pos = ();
    my $max_list = 0;

    $out .= "<TABLE BORDER=1>\n<TR>\n<TH>#</TH>\n";
    my $pos_counter = 0;
    foreach my $file (@files)  {
	my $file_out = '';
	if (@files > 1) {
	    $file_out = $shorts{$file}. ' ';
	}
	foreach my $flag (@selected_flags) {
	    next unless ($used_flags{"$file|$flag"});
	    $out .= "<TH>$file_out$flag</TH>\n";
	    $pos_counter++;
	    $flag_pos{$pos_counter} = $flag;
	    if ($flag =~ /^list(\d+)/) {
		my $num = $1;
		if ($num > $max_list) {
		    $max_list = $num;
		}
	    }
	}
    }
    $out .= "<TH>elements</TH><TH>flags</TH></TR>\n";

    # rebuild codes (skipping flags that weren't used)
    my %new_codes = ();
    foreach my $code (sort keys %codes) {
	my @values  = (split / */, $code);
	my $new_code = '';
	unless ($code =~ /1/) {
	    # all flags failed
	    $new_code = '0' x (scalar(keys(%used_flags)));
	} else {
	    # there are scores for at least some flags:
	    foreach my $file (@files) {
		foreach my $flag (@selected_flags) {
		    my $value = shift @values;
		    next unless ($used_flags{"$file|$flag"});
		    $new_code .= $value;
		}
	    }
	}
	$new_codes{$new_code} = $codes{$code};
    }

    # print out codes and values
    my $code_counter = 0;
    my $found_full_match = '';

    my @add_out_headers = ();
#    foreach my $code (sort keys %new_codes) {
    foreach my $code (sort by_x_and_alphabetically keys %new_codes) {
	$code_counter++;
	$out .= "<TR ALIGN=RIGHT><TD>$code_counter</TD>\n";

	# write list to file
	my $list_file = "compare_list$action_step.$code_counter.txt";
	my $out_file = "$tmp_dir/$list_file";
	if (-e $out_file) {
	    unlink $out_file or return "Can't remove previous list '$out_file': $!\n";
	}
	unless (open (LIST, ">$out_file")) {
	    return "Can't write to list '$out_file': $!\n";
	}
	# add data to list
	my @add_header = ();
	my %add_header = ();       
	foreach my $header (@output) {
	    next if ($header eq $SPOT_ID_TERM);
	    foreach my $ch ('CH1', 'CH2', '__NO_LABEL__') {
		if (defined $data_column_available{$files[0]}{$ch}{$header}) {
		    push @add_header, $header;
		    push @{$add_header{$header}}, $ch;
		}
	    }
	}
#	my %done = ($SPOT_ID_TERM, 1);
	my %done = ();
	print LIST "COLUMNS:MROW\tMCOL\tROW\tCOL\t$SPOT_ID_TERM";
	foreach my $header (@add_header) {
	    my $ch = shift @{$add_header{$header}};
	    push @{$add_header{$header}}, $ch;
	    print LIST "\t$ch $header";
	}
	print LIST "\n";

#	print LIST "COLUMNS:MROW\tMCOL\tROW\tCOL";
	foreach my $position (sort keys %{$new_codes{$code}}) {
	    my ($mrow,$mcol,$row,$col) = split /\|/, $position;
	    my $spot_id = $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col} || 'NA';
	    print LIST "$mrow\t$mcol\t$row\t$col\t$spot_id";
	    foreach my $header (@add_header) {
		my $ch = shift @{$add_header{$header}};
		push @{$add_header{$header}}, $ch;
		my $val = 'NA';
		if (defined $data{$files[0]}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col}) {
		    $val = $data{$files[0]}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col};
		}
		print LIST "\t$val";
	    }
	    print LIST "\n";
	}
	close LIST;

	my $pos_counter = 0;
	my @flag_list = ();
	foreach my $value (split / */, $code) {
	    $pos_counter++;
	    my $out_val = 'x';
	    my $colour = ' BGCOLOR="LIGHTGREY"';
	    if ($value == 0) {
		$out_val = '-';
		$colour = '';
	    } else {
		# collect flags
		push @flag_list, $flag_pos{$pos_counter};
	    }
	    $out .= "<TD ALIGN=CENTER$colour>$out_val</TD>\n";
	    
	}
	$out .= "<TD><A HREF=\"$list_file\" TARGET=\"_lists\">".(scalar keys %{$new_codes{$code}})."</A></TD>\n";

	# add list of flags
	@flag_list =  &remove_dups_preserve_order(@flag_list);
	my $flag_list = 'no flags';
	if (@flag_list) {
	    $flag_list = join ',', @flag_list;
	}
	$out .= "<TD ALIGN=LEFT>$flag_list</TD>\n</TR>\n";

	# store name of list that has matches in each column:
	if ($make_new_list) {
	    unless ($code =~ /0/) {
		$found_full_match = 1;
		$make_new_list = $out_file;
	    }
	}
    }
	
    # last line - accumulated numbers for each flag
    $out .= "<TR><TD>sum</TD>\n";
    foreach my $file (@files)  {
	my $file_out = (basename $file)." ";
	if (@files == 1) {
	    $file_out = '';
	}
	foreach my $flag (@selected_flags) {
	    next unless ($used_flags{"$file|$flag"});

	    my $nums = scalar keys %{$spot_nums{"$file$flag"}};

	    # write list to file
	    my $list_file = "compare_list$action_step.$flag.txt";
	    if ($nums > 0) {
		my $out_file = "$tmp_dir/$list_file";
		if (-e $out_file) {
		    unlink $out_file or return "Can't remove previous list '$out_file': $!\n";
		}
		unless (open (LIST, ">$out_file")) {
		    return "Can't write to list '$out_file': $!\n";
		}
		print LIST "COLUMNS:MROW\tMCOL\tROW\tCOL\t$SPOT_ID_TERM\n";
		foreach my $position (sort keys %{$spot_nums{"$file$flag"}}) {
		    my ($mrow,$mcol,$row,$col) = split /\|/, $position;
		    my $spot_id = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col} || 'NA';
		    print LIST "$mrow\t$mcol\t$row\t$col\t$spot_id\n";
		}
		close LIST;
	    }

	    if ($nums > 0) {
		$out .= "<TH><A HREF=\"$list_file\" TARGET=\"_lists\">$nums</A></TH>\n";
	    } else {
		$out .= "<TH>$nums</TH>\n";
	    }
	}
    }
    $out .= "<TD>\&nbsp;</TD>\n";	
    $out .= "</TR></TABLE><BR>\n";

    if ($make_new_list
	and
	$found_full_match) {
	$max_list++;
	my $new_flag = "list$max_list";
	my $new_flag_num = 0;
	if (defined $flag_values{$new_flag}) {
	    $new_flag_num = $flag_values{$new_flag};
	} else {
	    $new_flag_num = (sort { $b <=> $a } keys %flag_values_num)[0];
	    $new_flag_num *= 2;
	}
	my $added_flags = &flag_files($make_new_list, $new_flag_num, (keys %data));
	my $add_info = $added_flags;
	unless ($added_flags =~ /^ERROR/) {
	    $add_info = "<BR>Added flag <B>$new_flag</B> ($new_flag_num) to <B>$added_flags</B> spots in all files.<BR>\n";
	}
	$out .= $add_info;
    }
    return $out;
}


sub by_x_and_alphabetically {

    my $len1 = 0;
    my $len2 = 0;
    my $leading1 = 0;
    my $leading2 = 0;
    my $leading_end1 = 0;
    my $leading_end2 = 0;

    my $counter = 0;
    foreach (split / */, $a) {
	$counter++;
	if ($_ eq '0') {
	    unless ($leading1) {
		$leading1 = $counter;
	    }
	    unless ($len1) {
		$leading_end1++;
	    }
	    next;
	} else {
	    $len1++;
	}
    }

    $counter = 0;
    foreach (split / */, $b) {
	$counter++;
	if ($_ eq '0') {
	    unless ($leading2) {
		$leading2 = $counter;
	    }
	    unless ($len1) {
		$leading_end1++;
	    }
	    next;
	} else {
	    $len2++;
	}
    }
    if ($len1 > $len2) {
	return 1;
    } elsif ($len1 < $len2) {
	return -1;
    } else {
	if ($leading1 > $leading2) {
	    return -1;
	} elsif ($leading1 < $leading2) {
	    return 1;
	} else {
	    if ($leading_end1 > $leading_end2) {
		return -1;
	    } elsif ($leading_end1 < $leading_end2) {
		return 1;
	    } else {
		return 0;
	    }
	}
    }
}

sub flag_files {
    my $file = shift;
    my $flag = shift;
    my @files = @_;

    my $counts = 0;

    unless (open (IN, $file)) {
	warn "Can't open file '$file': $!";
	return "ERROR: Couldn't add flag to positions in $file: $!\n";
    }
    
    while (<IN>) {
	my ($mrow,$mcol,$row,$col,$spot_id) = split /\t/, $_;
	next if ($mrow =~ /\D/);
	
	$counts++;
	foreach my $file (@files) {
	    # flag position
	    foreach my $ch (keys %{$data{$file}{'__DATA__'}}) {
		$data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} = $flag;
	    }
	}
    }
    return $counts;
}


sub remove_dups_preserve_order {
    my @list = @_;
    my %tmp = ();
    my @tmp = ();
    foreach (@list) {
	next if (defined $tmp{$_});
	$tmp{$_}++;
	push @tmp, $_;
    }

    return @tmp;
}

sub annotation_setup {
    my $action_num = shift;
    my $action = shift;

    if ($refine_parameters) {
	print &format_output( 
			      {
				  'prefix' => 'annotate_',
				  'num' => $action_num,
				  'file' => '_ALL__',
				  'label' => '__ALL__',
				  'action' => $action,
				  'info' => 'Parameter Refinement!<BR>'
				  }
			      );
	return;
    }

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    # db details
    require DBI;
    my $DBD = 'mysql';
    my $database = 'hostdb3';
    my $db_host = '192.168.10.3';
    my $user = 'pipeline';
    my $password = '9aleDL4k';
    my $table = 'oligo_position';
    
    my $dbh = DBI->connect("DBI:$DBD:database=$database;host=$db_host", $user, $password,{ RaiseError => 1, AutoCommit => 0 });
    unless (defined $dbh) {
	die "Can't connect to database: $DBI::errstr!\n";
    }

    my @files = &get_user_settings($var_def_default{'annotate_files'},'annotate_files',$force_global_opt,$action_num);

    foreach (@files) {
	if ($_ eq 'all') {
	    @files = (sort keys %data);
	    last;
	}	
    }

    # read in additional SwissProt information
    my $swissprot_file = '/www/koch.pathogenomics.ca/data/arraypipe/internal/annotation/swissprot';
    if (-s $swissprot_file
	and
	open (IN, $swissprot_file)
	) {
	print STDERR "\nReading file '$swissprot_file' at ".(localtime)."\n";
	
	my $line_counter = 0;
	
	while (<IN>) {
	    
	    $line_counter++;
	    
	    if ($line_counter % 1000 == 0) {
		print STDERR "$line_counter ";
	    }		    
	    
	    # skip comments
	    next if (/^\#/);
	    
	    chomp;
	    
	    my @tmp = split /\t/, $_;
	    my @accessions = split /; /, $tmp[0];
	    my @names = split /; /, $tmp[1];
	    foreach my $id (@accessions, @names) {
		if (defined $tmp[3]) {
		    $swissprot{$id}{'SWISSPROT gene name'} = $tmp[3];
		} else {
		    $swissprot{$id}{'SWISSPROT gene name'} = '';
		}
		if (defined $tmp[4]) {
		    $swissprot{$id}{'SWISSPROT protein name'} = $tmp[4];
		} else {
		    $swissprot{$id}{'SWISSPROT protein name'} = '';
		}
		if (defined $tmp[5]) {
		    $swissprot{$id}{'SWISSPROT function'} = $tmp[5];
		} else {
		    $swissprot{$id}{'SWISSPROT function'} = '';
		}
	    }
	}
	close IN;
    }
    

    $web_link_annotation_file = &get_user_settings($var_def_default{'annotate_web_file'},'annotate_web_file',$force_global_opt,$action_num);
    if ($web_link_annotation_file) {
	# loads files into $tmp_dir
	# and returns full path to file(s)
	my $upload_file = &download_1_file($cgi, $web_link_annotation_file, $tmp_dir,'');

	# unpack archives or uncompress file if necessary
	my @upload_files = &unpack_archive($upload_file);	    
	my ($format_check, $format_type) = &check_file_format($upload_files[0]);
	if ($format_check eq 'ok'
	    and
	    $format_type eq 'probelynx_annotation') {
	    $web_link_annotation_file = $upload_files[0];
	} else {
	    $web_link_annotation_file = '';
	    warn "# Annotation file '$upload_files[0]' not valid ($format_check, $format_type)";
	}
    }	
	    
    # carry out action on each file for each format
    foreach my $file (@files) {
	my $out = &annotate_data($action_num,$action,$file, $dbh);

	# add a column that contains one of multiple choices for a gene name and description
	my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);
	foreach my $mrow (1..$MROW) {
	    foreach my $mcol (1..$MCOL) {
		foreach my $row (1..$ROW) {
		    foreach my $col (1..$COL) {
			my $probe = $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col} if (defined $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col});
			next unless ($probe);
			my @vals = ();
			my $type = '';
			foreach my $annotation_item ('gene symbol', 'HUGO accession','SWISSPROT gene name','SWISSPROT accession', 'SPTREMBL accession', 'EMBL accession', 'gene', 'Ensembl accession') {
			    if (defined $annotation{$annotation_item}{$probe}
				and
				@{$annotation{$annotation_item}{$probe}} > 0
				and
				(grep /\w/, @{$annotation{$annotation_item}{$probe}})
				) {
				@vals = @{$annotation{$annotation_item}{$probe}};
				$type = $annotation_item;
				last;
			    }
			}
			@{$annotation{$COMMON_GENE_NAME}{$probe}} = @vals;		    
			$annotation{$COMMON_GENE_NAME_SOURCE}{$probe} = $type;
		    
			@vals = ();
			$type = '';
			foreach my $annotation_item ('SWISSPROT function','SWISSPROT protein name','interpro description', 'gene description', 'proteinfamily family description','domain description', 'feature prediction method') {
#			    warn "# $annotation_item: $annotation{$annotation_item}, $probe: $annotation{$annotation_item}{$probe}";
			    if (defined $annotation{$annotation_item}) {
#				warn "# probe: '$probe', '$annotation_item'";
				if (defined $annotation{$annotation_item}{$probe}) {
				    if (defined @{$annotation{$annotation_item}{$probe}}) {
					if (@{$annotation{$annotation_item}{$probe}} > 0) {					    
					    if (grep /\w/, @{$annotation{$annotation_item}{$probe}}) {						
						@vals = @{$annotation{$annotation_item}{$probe}};
						$type = $annotation_item;
						last;
					    }
					}
				    }
				}
			    }
			}
			@{$annotation{$COMMON_GENE_DESCRIPTION}{$probe}} = @vals;		    
			$annotation{$COMMON_GENE_DESCRIPTION_SOURCE}{$probe} = $type;		    
		    }
		}
	    }
	}
			    
	print &format_output( # 'annotate_',$action_num,$file,'__ALL__',$action,'',$out
			      {
				      'prefix' => 'annotate_',
				      'num' => $action_num,
				      'file' => $file,
				      'label' => '__ALL__',
				      'action' => $action,
				      'info' => $out
				      }
				  );
    }

    # disconnect from database
    $dbh->disconnect;

}


sub annotate_data {

    # annotation is based on the probe found at each spot
    # it is stored in a construct like this:
    # @{$annotation{$annotation_item}{$probe}} = (list of annotations)
    # Lateron, in &output_file a column is created for each annotation item
    # into which the collected values will be written

    my $action_num = shift;
    my $action = shift;
    my $file = shift;
    my $dbh = shift;
    
    my %listed = ();

    my $out = "<BR>The following table lists the numbers of probe or spot IDs for which annotation of a certain category was found.<BR>Categories without matches are not listed!\n<BR><TABLE BORDER=1>
<TR>
<TH>annotation</TH><TH>".
&help(
      {
	  'topic' => 'reporter_numbers',
	  'label' => 'reporters (probes)',
	  'no_small' => 1
	  }
      )
."</TH><TH>".
&help(
      {
	  'topic' => 'reporter_missing',
	  'label' => 'missing',
	  'no_small' => 1
	  }
      )
."</TH>
</TR>
";
    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    $last_column_used = $SPOT_ID_TERM;

    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);

    unless (defined $probe_ids{"$MROW,$MCOL,$ROW,$COL"}) {
	&get_probe_ids($file,$MROW,$MCOL,$ROW,$COL);
    }

    # now we have either specially mapped probe names for each position
    # or simply the reporter name


    # ProbeLynx section:
    # get required annotation fields:
    my @annotations_probelynx = &get_user_settings($var_def_default{'annotate_items_probelynx'},'annotate_items_probelynx',$force_global_opt,$action_num,$file);
    foreach (@annotations_probelynx) {
	if ($_ eq 'all') {
	    @annotations_probelynx = ();
	    foreach my $item (@{$var_def_values{'annotate_items_probelynx'}}) {
#		next if ($item eq 'all');
		push @annotations_probelynx, $item;
	    }
	}
    }
    foreach (@annotations_probelynx) {
	$annotation_select_probelynx{$_}++;
    }

    # determine which probelynx file to use:
    my $probe_lynx_file = '';
    if (-e "/www/koch.pathogenomics.ca/data/arraypipe/internal/annotation/ProbeLynx_${MROW}_${MCOL}_${ROW}_${COL}.tab"
	and
	-r "/www/koch.pathogenomics.ca/data/arraypipe/internal/annotation/ProbeLynx_${MROW}_${MCOL}_${ROW}_${COL}.tab") {
	$probe_lynx_file =  "/www/koch.pathogenomics.ca/data/arraypipe/internal/annotation/ProbeLynx_${MROW}_${MCOL}_${ROW}_${COL}.tab";
    }

    # an annotation file might have been uploaded:
    if (defined $cgi->param('annotation_upload_file')
	and
	$cgi->param('annotation_upload_file') =~ /\w/) {
	$probe_lynx_file = &upload_1_file($cgi, $cgi->param('annotation_upload_file'), $tmp_dir,'');
    } elsif (defined $web_link_annotation_file
	     and
	     -r $web_link_annotation_file) {
	$probe_lynx_file = $web_link_annotation_file;
    }

    # read in file if necessary
    unless (defined $annotation_file_read{$probe_lynx_file}) {
	$annotation_file_read{$probe_lynx_file}++;
	my ($rv,$doc) = &read_annotation($probe_lynx_file, \%annotation_select_probelynx, \@annotations_probelynx, $dbh);
	unless ($rv eq 'ok') {
	    return "Problem with annotation: '$rv'";
	}
	$out .= $doc;
    }

    # That's the annotation done for the uploaded/stored ProbeLynx file!


    # Let's now proceed with annotation from FuncDB:

    my @annotations_funcdb = &get_user_settings($var_def_default{'annotate_items_funcdb'},'annotate_items_funcdb',$force_global_opt,$action_num,$file);
    foreach (@annotations_funcdb) {
	if ($_ eq 'all') {
	    @annotations_funcdb = ();
	    foreach my $item (@{$var_def_values{'annotate_items_funcdb'}}) {
		next if ($item eq 'all');
		push @annotations_funcdb, $item;
	    }
	}
    }

    my @tmp_list = ();
    foreach (@annotations_funcdb) {
	# avoid repeating annotation that was done in ProbeLynx already:
	next if (defined $annotation{$_});
	next if (defined $annotation_item_match{$_}
		 and
		 defined $annotation{$annotation_item_match{$_}});
	push @tmp_list, $_;
	$annotation_select_funcdb{$_}++;
    }

    @annotations_funcdb = @tmp_list;

    my $funcdb_out = '';
    if (@annotations_funcdb) {
	$funcdb_out .= "<TR>
<TD ALIGN=CENTER COLSPAN=4 BGCOLOR=LIGHTGREY>--&nbsp;<B>FuncDB&nbsp;</B>--</TD>
</TR>
";
    }

    # to save time, check first if probes need to be annotated
    # (they might have all been done for a previous file already)

    my ($mrow,$mcol,$row,$col);
    my $header = $SPOT_ID_TERM;


    # prepare hostdb_id statement:
    my $statement_hostdb_id = "SELECT hostdb_id FROM hostdb_main WHERE accession = ?";
    my $sth_hostdb_id = $dbh->prepare($statement_hostdb_id);
    die "Can't prepare '$statement_hostdb_id': $DBI::errstr" if ($DBI::errstr);

    # prepare annotation statements
    # prepare HUGO statement:
    my $statement_hugo_id = "SELECT accession FROM hostdb_main WHERE hostdb_id = ? and db_name = ?";
    my $sth_hugo_id = $dbh->prepare($statement_hugo_id);
    die "Can't prepare '$statement_hugo_id': $DBI::errstr" if ($DBI::errstr);
    # prepare GO statement:
    my $statement_GO= "SELECT GO_term FROM ensembl_GO WHERE hostdb_id = ?";
    my $sth_GO = $dbh->prepare($statement_GO);
    die "Can't prepare '$statement_GO': $DBI::errstr" if ($DBI::errstr);
    # prepare BIND statement:		    
    my $statement_bind = "SELECT bind_interaction_id FROM bind_interactions WHERE hostdb_id = ?";
    my $sth_bind = $dbh->prepare($statement_bind);
    die "Can't prepare '$statement_bind': $DBI::errstr" if ($DBI::errstr);   
    # prepare BIND interaction partners statement:		    
    my $statement_bind_interact_a = "SELECT bind_interaction_id,molecule FROM bind_interactions WHERE hostdb_id = ?";
    my $sth_bind_interact_a = $dbh->prepare($statement_bind_interact_a);
    die "Can't prepare '$statement_bind_interact_a': $DBI::errstr" if ($DBI::errstr);   
    my $statement_bind_interact_b = "SELECT hostdb_id FROM bind_interactions WHERE bind_interaction_id = ? AND molecule = ?";
    my $sth_bind_interact_b = $dbh->prepare($statement_bind_interact_b);
    die "Can't prepare '$statement_bind_interact_b': $DBI::errstr" if ($DBI::errstr);   
    # prepare BioCarta statement:		    
    my $statement_biocarta= "SELECT pathway_id FROM biocarta_pathway_link WHERE hostdb_id = ?";
    my $sth_biocarta = $dbh->prepare($statement_biocarta);
    die "Can't prepare '$statement_biocarta': $DBI::errstr" if ($DBI::errstr);   
    # prepare DIP statement:		    
    my $statement_dip= "SELECT dip_node_id FROM dip_interactions WHERE hostdb_id = ?";
    my $sth_dip = $dbh->prepare($statement_dip);
    die "Can't prepare '$statement_dip': $DBI::errstr" if ($DBI::errstr);   
    # prepare DIP interaction partners statement:		    
    my $statement_dip_interact_a = "SELECT dip_edge_id,molecule FROM dip_interactions WHERE hostdb_id = ?";
    my $sth_dip_interact_a = $dbh->prepare($statement_dip_interact_a);
    die "Can't prepare '$statement_dip_interact_a': $DBI::errstr" if ($DBI::errstr);   
    my $statement_dip_interact_b = "SELECT hostdb_id FROM dip_interactions WHERE dip_edge_id = ? AND molecule = ?";
    my $sth_dip_interact_b = $dbh->prepare($statement_dip_interact_b);
    die "Can't prepare '$statement_dip_interact_b': $DBI::errstr" if ($DBI::errstr);   
    # prepare Ensembl Protein Family description statement:		    
    my $statement_protein_family_desc= "SELECT description FROM ensembl_description WHERE hostdb_id = ?";
    my $sth_protein_family_desc = $dbh->prepare($statement_protein_family_desc);
    die "Can't prepare '$statement_protein_family_desc': $DBI::errstr" if ($DBI::errstr);
    # prepare Ensembl domain prediction statement:		    
    my $statement_domain_method = "SELECT prediction_method FROM ensembl_domain WHERE hostdb_id = ?";
    my $sth_domain_method = $dbh->prepare($statement_domain_method);
    die "Can't prepare '$statement_domain_method': $DBI::errstr" if ($DBI::errstr);
    # prepare Ensembl domain accession statement:		    
    my $statement_domain_accession = "SELECT accession_number FROM ensembl_domain WHERE hostdb_id = ?";
    my $sth_domain_accession = $dbh->prepare($statement_domain_accession);
    die "Can't prepare '$statement_domain_accession': $DBI::errstr" if ($DBI::errstr);
    # prepare Ensembl domain description statement:		    
    my $statement_domain_description = "SELECT description FROM ensembl_domain WHERE hostdb_id = ?";
    my $sth_domain_description = $dbh->prepare($statement_domain_description);
    die "Can't prepare '$statement_domain_description': $DBI::errstr" if ($DBI::errstr);
    # prepare Ensembl domain start statement:		    
    my $statement_domain_start = "SELECT start FROM ensembl_domain WHERE hostdb_id = ?";
    my $sth_domain_start = $dbh->prepare($statement_domain_start);
    die "Can't prepare '$statement_domain_start': $DBI::errstr" if ($DBI::errstr);
    # prepare Ensembl domain end statement:		    
    my $statement_domain_end = "SELECT end FROM ensembl_domain WHERE hostdb_id = ?";
    my $sth_domain_end = $dbh->prepare($statement_domain_end);
    die "Can't prepare '$statement_domain_end': $DBI::errstr" if ($DBI::errstr);
    # prepare Ensembl feature prediction statement:		    
    my $statement_feature_method = "SELECT prediction_method FROM ensembl_feature WHERE hostdb_id = ?";
    my $sth_feature_method = $dbh->prepare($statement_feature_method);
    die "Can't prepare '$statement_feature_method': $DBI::errstr" if ($DBI::errstr);
    # prepare Ensembl feature start statement:		    
    my $statement_feature_start = "SELECT start FROM ensembl_feature WHERE hostdb_id = ?";
    my $sth_feature_start = $dbh->prepare($statement_feature_start);
    die "Can't prepare '$statement_feature_start': $DBI::errstr" if ($DBI::errstr);
    # prepare Ensembl feature end statement:		    
    my $statement_feature_end = "SELECT end FROM ensembl_feature WHERE hostdb_id = ?";
    my $sth_feature_end = $dbh->prepare($statement_feature_end);
    die "Can't prepare '$statement_feature_end': $DBI::errstr" if ($DBI::errstr);
    # prepare Ensembl Protein Family id statement:		    
    my $statement_protein_family_id= "SELECT family_id FROM ensembl_proteinfamily WHERE hostdb_id = ?";
    my $sth_protein_family_id = $dbh->prepare($statement_protein_family_id);
    die "Can't prepare '$statement_protein_family_id': $DBI::errstr" if ($DBI::errstr);
    # prepare Ensembl Protein Family description statement:		    
    my $statement_protein_family_description= "SELECT family_description FROM ensembl_proteinfamily WHERE hostdb_id = ?";
    my $sth_protein_family_description = $dbh->prepare($statement_protein_family_description);
    die "Can't prepare '$statement_protein_family_description': $DBI::errstr" if ($DBI::errstr);
    # prepare InterPro statement:		        
    my $statement_IP= "SELECT interpro_id,description FROM ensembl_interpro WHERE hostdb_id = ?";
    my $sth_IP = $dbh->prepare($statement_IP);
    die "Can't prepare '$statement_IP': $DBI::errstr" if ($DBI::errstr);    
    # prepare TRANSFAC_TYPE statement:		    
    my $statement_transfac_type_gene = "SELECT gene_id FROM transfac_gene_link WHERE hostdb_id = ?";
    my $sth_transfac_type_gene = $dbh->prepare($statement_transfac_type_gene);
    die "Can't prepare '$statement_transfac_type_gene': $DBI::errstr" if ($DBI::errstr);   
    my $statement_transfac_type_factor = "SELECT factor_id FROM transfac_factor_link WHERE hostdb_id = ?";
    my $sth_transfac_type_factor = $dbh->prepare($statement_transfac_type_factor);
    die "Can't prepare '$statement_transfac_type_factor': $DBI::errstr" if ($DBI::errstr);   
    # prepare TRANSFAC_FACTOR statement:		    
    my $statement_transfac_factor_a = "SELECT transfac_factor_link.factor_id, 
                                              transfac_factor_desc.factor_name 
                                         FROM transfac_factor_link,
                                              transfac_factor_desc
                                        WHERE transfac_factor_link.hostdb_id = ?
                                              AND
                                              transfac_factor_link.factor_id = transfac_factor_desc.factor_id";
    my $sth_transfac_factor_a = $dbh->prepare($statement_transfac_factor_a);
    die "Can't prepare '$statement_transfac_factor_a': $DBI::errstr" if ($DBI::errstr);   
    my $statement_transfac_factor_b = "SELECT transfac_factor_site.factor_id, 
                                              transfac_factor_desc.factor_name 
                                         FROM transfac_gene_link,
                                              transfac_factor_site,
                                              transfac_factor_desc
                                        WHERE transfac_gene_link.hostdb_id = ?
                                              AND
                                              transfac_gene_link.gene_id = transfac_factor_site.gene_id
                                              AND
                                              transfac_factor_site.factor_id = transfac_factor_desc.factor_id";
    my $sth_transfac_factor_b = $dbh->prepare($statement_transfac_factor_b);
    die "Can't prepare '$statement_transfac_factor_b': $DBI::errstr" if ($DBI::errstr);   
    # prepare TRANSFAC_GENE statement:		    
    my $statement_transfac_gene_a = "SELECT transfac_factor_site.gene_id, 
                                            transfac_gene_desc.short_desc 
                                       FROM transfac_factor_site,
                                            transfac_gene_desc,
                                            transfac_factor_link
                                      WHERE transfac_factor_link.hostdb_id = ?
                                            AND
                                            transfac_factor_link.factor_id = transfac_factor_site.factor_id
                                            AND
                                            transfac_factor_site.gene_id = transfac_gene_desc.gene_id";
    my $sth_transfac_gene_a = $dbh->prepare($statement_transfac_gene_a);
    die "Can't prepare '$statement_transfac_gene_a': $DBI::errstr" if ($DBI::errstr);   
    my $statement_transfac_gene_b = "SELECT transfac_gene_link.gene_id, 
                                            transfac_gene_desc.short_desc 
                                       FROM transfac_gene_link,
                                            transfac_gene_desc
                                      WHERE transfac_gene_link.hostdb_id = ?
                                            AND
                                            transfac_gene_link.gene_id = transfac_gene_desc.gene_id";
    my $sth_transfac_gene_b = $dbh->prepare($statement_transfac_gene_b);
    die "Can't prepare '$statement_transfac_gene_b': $DBI::errstr" if ($DBI::errstr);   
    # prepare TRANSPATH statement:		    
    my $statement_transpath = "SELECT node_id FROM transpath_reactions_link WHERE hostdb_id = ?";
    my $sth_transpath = $dbh->prepare($statement_transpath);
    die "Can't prepare '$statement_transpath': $DBI::errstr" if ($DBI::errstr);   
    my $statement_transpath_molecule = "SELECT name FROM transpath_molecule_desc WHERE molecule_id = ?";
    my $sth_transpath_molecule = $dbh->prepare($statement_transpath_molecule);
    die "Can't prepare '$statement_transpath_molecule': $DBI::errstr" if ($DBI::errstr);   
    my $statement_transpath_gene = "SELECT name FROM transpath_gene_desc WHERE gene_id = ?";
    my $sth_transpath_gene = $dbh->prepare($statement_transpath_gene);
    die "Can't prepare '$statement_transpath_gene': $DBI::errstr" if ($DBI::errstr);   
    
    my %no_probes = ();
    my %no_transcripts = ();

    # determine which probes need annotation:
    ANNOTATION:foreach my $annotation_item (@annotations_funcdb) {

	next if ($annotation_item =~ /--.*--/);

	my $annotation_item_out = $annotation_item;
	print STDERR "\n$annotation_item: \n";
	print STDERR "start: ".(localtime)."\n";
	my $found_annotation = 0;
	my $prev_annotation = 0;
	my $no_annotation = 0;

	my %found_annotation = ();
	my %prev_annotation = ();
	my %no_annotation = ();
	# go through each field
	my %probes = ();
	foreach $mrow (keys %{$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$header}}) {
	    foreach $mcol (keys %{$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$header}{$mrow}}) {
		foreach $row (keys %{$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$header}{$mrow}{$mcol}}) {
		    foreach $col (keys %{$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$header}{$mrow}{$mcol}{$row}}) {
			if (defined $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col}) {
			    my $probe = $probe_ids{"$MROW,$MCOL,$ROW,$COL"}{$mrow}{$mcol}{$row}{$col};
			    unless (defined @{$annotation{$annotation_item}{$probe}}) {
				# this probe needs to be annotated
				$probes{$probe}++;
			    } else {
				# annotation has been attempted previously
				if (@{$annotation{$annotation_item}{$probe}} > 0) {
				    # take note if something was found
				    $prev_annotation++;
				    $prev_annotation{$probe}++;
				} else {
				    $no_annotation++;
				    $no_annotation{$probe}++;
				}
			    }
			} else {
			    # no probe defined for this spot
			    # therefore no annotation possible
			    $no_probes{"$mrow,$mcol,$row,$col"}++;
			    next;
			}
		    }
		}
	    }
	}
	
	unless (%probes) {
	    if ($annotation_item =~ /--.*--/) {
		$funcdb_out .= "<TR>
<TD ALIGN=CENTER COLSPAN=4 BGCOLOR=LIGHTGREY>$annotation_item</TD>
</TR>
";
	    } else {
		my $found = scalar (keys %found_annotation) + scalar (keys %prev_annotation);
		if ($found > 0) {
		    unless ($listed{$annotation_item}) {
			$funcdb_out .= "<TR>
<TD>$annotation_item_out</TD>
<TD ALIGN=RIGHT>$found</TD>
<TD ALIGN=RIGHT>".(scalar keys %no_annotation)."</TD>
</TR>
";
			push @annotation_headers_all, $annotation_item;

			$listed{$annotation_item}++;
		    }
		}
	    }
	    next;
	}

	# get annotation

	# for speed-up reasons we might have annotation stored on file:
	if ($FPMI_specific) {

	    my $annotation_file = "/www/koch.pathogenomics.ca/data/arraypipe/internal/annotation/annotation_${MROW}_${MCOL}_${ROW}_${COL}.tab";
	    if (-s $annotation_file
		and
		open (IN, $annotation_file)
		) {

		print STDERR "\nReading file '$annotation_file' at ".(localtime)."\n";

		my %done = ();
		my %count = ();
		my %missing = ();
		my @order = ();

		my $line_counter = 0;

		while (<IN>) {

		    $line_counter++;
		    
		    if ($line_counter % 1000 == 0) {
			print STDERR "$line_counter ";
		    }		    

		    chomp;
		    my @tmp = split /\t/, $_;
		    my $annotation_item = shift @tmp;
		    unless ($done{$annotation_item}) {
			push @order, $annotation_item;
			$done{$annotation_item}++;
			$missing{$annotation_item} = scalar keys %no_probes;
			$count{$annotation_item} = 0;
		    }
		    my $probe = shift @tmp;
		    @{$annotation{$annotation_item}{$probe}} = @tmp;
		    $annotation_done{$annotation_item}++ if (@tmp);
		    if (@tmp) {
			$count{$annotation_item}++;
		    } else {
			$missing{$annotation_item}++;
			if ($annotation_item eq 'TRANSCRIPT') {
			    $no_transcripts{$probe}++;
			}
		    }

		    # add SwissProt info
		    if ($annotation_item eq 'SWISSPROT accession'
			and
			@tmp) {
			foreach my $id (@tmp) {
			    if (defined $swissprot{$id}) {
				foreach my $item ('SWISSPROT gene name','SWISSPROT protein name','SWISSPROT function') {
				    unless ($done{$item}) {
					push @order, $item;
					$done{$item}++;
					$missing{$item} = scalar keys %no_probes;
					$count{$item} = 0;
				    }
				    push @{$annotation{$item}{$probe}}, $swissprot{$id}{$item};
				}
			    }
			}
			foreach my $item ('SWISSPROT gene name','SWISSPROT protein name','SWISSPROT function') {
			    $count{$item}++;
			}
		    }
		}
		close IN;
  
		print STDERR "Finished reading file at ".(localtime)."\n";

		foreach my $annotation_item (@annotations_funcdb) {
		    my $annotation_item_out = $annotation_item;
		    if ($annotation_item =~ /--.*--/) {
			$funcdb_out .= "<TR>
<TD ALIGN=CENTER COLSPAN=4 BGCOLOR=LIGHTGREY>$annotation_item</TD>
</TR>
";
		    } else {
			unless (defined $count{$annotation_item}) {
			    if (defined $annotation{$annotation_item}) {
				$count{$annotation_item} = scalar keys %{$annotation{$annotation_item}};
			    }
			    unless (defined $missing{$annotation_item}) {
				if (defined $count{$annotation_item}
				    and
				    defined $count{'TRANSCRIPT HITS'}) {
				    $missing{$annotation_item} = $count{'TRANSCRIPT HITS'} - $count{$annotation_item};
				}
			    }
			}
			if (defined $count{$annotation_item}
			    and
			    $count{$annotation_item} > 0) {
			    unless ($listed{$annotation_item}) {
				$funcdb_out .= "<TR>
<TD>$annotation_item_out</TD>
<TD ALIGN=RIGHT>$count{$annotation_item}</TD>
<TD ALIGN=RIGHT>".((scalar keys %probe_names) - $count{$annotation_item})."</TD>
</TR>
";
				push @annotation_headers_all, $annotation_item;
				$listed{$annotation_item}++;
			    }
			}
		    }
		}

		if (%no_probes) {
		    $funcdb_out .= "<BR>For ".(scalar keys %no_probes)." spots no probe/reporter sequences were found!\n";
		}
		
#		if (%no_transcripts) {
#		    $funcdb_out .= "<BR>For ".(scalar keys %no_transcripts)." probes/reporters no transcripts were found, i.e. no TC matched or a match outside a gene border!\n";
#		}

		# if we had orthologue hits, add them to orthologues output:
		my $orthologue_out = 0;
		foreach my $item ('THC ENSEMBL HITS', 'THC ENSEMBL GENE', 'gene description', 'THC ENSEMBL TRANSCRIPT', 'THC ENSEMBL COVERAGE', 'THC ENSEMBL IDENTITY', 'THC ENSEMBL GAPS', 'THC ENSEMBL STATUS') {
		    if (defined $annotation{'THC ENSEMBL TRANSCRIPT'}) {
			unless ($orthologue_out) {
			    $orthologue_out = 1;
			    $out .= "<TR>
		<TD ALIGN=CENTER COLSPAN=3 BGCOLOR=LIGHTGREY><B>-- Putative Orthologues --</B></TD>
		</TR>
		";
			}
			$out .= "<TR>
		<TD>$item</TD>
		<TD ALIGN=RIGHT>".(scalar keys %{$annotation{$item}})."</TD>
		<TD ALIGN=RIGHT>".((scalar keys %probe_names) - (scalar keys %{$annotation{$item}}))."</TD>
		</TR>
		";
		    }
		}

		$funcdb_out .= "</TABLE>\n";

		return $out.$funcdb_out;
	    }
	}

	# end of annotation read from file

	# if we get here we must read annotation from FuncDB:

	my $probe_counter = 0;

	PROBE:foreach my $probe (keys %probes) {

	    $probe_counter++;

	    if ($probe_counter % 100 == 0) {
		print STDERR "$probe_counter ";
	    }

	    unless (defined $annotation{'number of transcript hits'}{$probe}
		    and
		    @{$annotation{'number of transcript hits'}{$probe}} > 0
		    ) {
	    
		# try annotation via orthologue:
		if (defined $orthologues{$probe}) {
		    my %orthologues_genes = ();
#		    foreach my $thc (sort keys %{$orthologues{$probe}}) {
		    foreach my $orthologue (@{$orthologues{$probe}{'TRANSCRIPT'}}) {
			    my $gene = shift @{$orthologues{$probe}{'GENE'}};
#			    my $cover = shift @{$orthologues{$probe}{'COVER'}};
#			    my $ident = shift @{$orthologues{$probe}{'IDENT'}};
#			    my $gaps = shift @{$orthologues{$probe}{'GAPS'}};
#			    my $status = shift @{$orthologues{$probe}{'STATUS'}};

#			    if ($cover eq '') {
#				$cover = 'acc';
#			    }
#			    if ($ident eq '') {
#				$ident = 'acc';
#			    }
#			    if ($gaps eq '') {
#				$gaps = 'acc';
#			    }
#			    if ($status eq '') {
#				$status = 'acc';
#			    }

			    $orthologues_genes{$gene}++;
			    push @{$annotation{'transcript'}{$probe}}, $orthologue;

#			    push @{$annotation{'THC ENSEMBL GENE'}{$probe}}, $gene;
#			    $annotation_done{'THC ENSEMBL GENE'}++;
#			    push @{$annotation{'THC ENSEMBL TRANSCRIPT'}{$probe}}, $orthologue;
#			    $annotation_done{'THC ENSEMBL TRANSCRIPT'}++;
#			    push @{$annotation{'THC ENSEMBL COVERAGE'}{$probe}}, $cover;
#			    $annotation_done{'THC ENSEMBL COVERAGE'}++;
#			    push @{$annotation{'THC ENSEMBL IDENTITY'}{$probe}}, $ident;
#			    $annotation_done{'THC ENSEMBL IDENTITY'}++;
#			    push @{$annotation{'THC ENSEMBL GAPS'}{$probe}}, $gaps;
#			    $annotation_done{'THC ENSEMBL GAPS'}++;
#			    push @{$annotation{'THC ENSEMBL STATUS'}{$probe}}, $status;
#			    $annotation_done{'THC ENSEMBL STATUS'}++;

#			}

			@{$annotation{'number of transcript hits'}{$probe}} = ('orthologues');
			$annotation_done{'number of transcript hits'}++;
#			@{$annotation{'THC ENSEMBL HITS'}{$probe}}[0]++;
#			$annotation_done{'THC ENSEMBL HITS'}++;
		}
		    @{$annotation{'gene hits'}{$probe}} = (scalar keys %orthologues_genes);
		    $annotation_done{'gene hits'}++;
		}
	    }

	    unless (defined @{$annotation{'number of transcript hits'}{$probe}}
		    and
		    (
		     @{$annotation{'number of transcript hits'}{$probe}}[0] eq 'orthologues'
		     or
		     @{$annotation{'number of transcript hits'}{$probe}}[0] > 0
		     )
		    ) {
		@{$annotation{$annotation_item}{$probe}} = ();
		$no_annotation++;		
		$no_annotation{$probe}++;
		$no_transcripts{$probe}++;
		next PROBE;		    
	    }
	    
	    # get annotation item for the transcripts of this probe
	    # initialize entry
	    @{$annotation{$annotation_item}{$probe}} = ();

	    foreach my $transcript (@{$annotation{'transcript'}{$probe}}) {

		next if ($transcript eq 'medium probability intergenic');

		# get hostdb_id if necessary
		unless (defined @{$annotation_hostdb_id{$transcript}}) {
		    # get hostdb_id from hostdb_main
		    @{$annotation_hostdb_id{$transcript}} = ();
		    my $rv = $sth_hostdb_id->execute($transcript);
		    if ($rv >= 1) {
			while (my ($id) = $sth_hostdb_id->fetchrow_array) {
			    push @{$annotation_hostdb_id{$transcript}}, $id;
			}
		    }
		}
			
		if ($annotation_item eq 'HUGO accession'
		    or
		    $annotation_item eq 'EMBL accession'
		    or
		    $annotation_item eq 'Ensembl accession'
		    or
		    $annotation_item eq 'LocusLink accession'
		    or
		    $annotation_item eq 'OMIM accession'
		    or
		    $annotation_item eq 'PDB accession'
		    or
		    $annotation_item eq 'RefSeq accession'
		    or
		    $annotation_item eq 'SPTREMBL accession'
		    or
		    $annotation_item eq 'SWISSPROT accession'
		    ) {
		    # get HUGO id and similar database items
		    # prepare HUGO statement:
		    # (see above)
		    my $db = $annotation_item;
		    $db =~ s/ accession$//;
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get HUGO id from hostdb_main
			my $rv = $sth_hugo_id->execute($hostdb_id, $db);
			if ($rv >= 1) {
			    while (my ($id) = $sth_hugo_id->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $id;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'GO') {
		    # get GO id 
		    # prepare GO statement:
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get GO id from hostdb_main
			my $rv = $sth_GO->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($id) = $sth_GO->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $id;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'BIND') {
		    # get BIND id 
		    # prepare BIND statement:
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get BIND id from hostdb_main
			my $rv = $sth_bind->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($id) = $sth_bind->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $id;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'BIND interaction partners') {
		    # get BIND id 
		    # prepare BIND statement:
		    # (see above)
		    my %hostdb_ids = ();
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get BIND id from hostdb_main
			my $rv = $sth_bind_interact_a->execute($hostdb_id);
			if ($rv >= 1) {
			    my %molecules = ();
			    while (my ($id,$molecule) = $sth_bind_interact_a->fetchrow_array) {
				$molecules{$id} = $molecule;
			    }
			    foreach my $id (sort keys %molecules) {
				my $molecule = $molecules{$id};
				# look for the opposite molecule:
				$molecule = ($molecule eq 'A' ? 'B' : 'A');
				my $rv = $sth_bind_interact_b->execute($id,$molecule);
				if ($rv >= 1) {
				    while (my ($bind_id) = $sth_bind_interact_b->fetchrow_array) {
					if (defined $bind_id) {
#					    warn "# undefined for '$id, $molecule, $rv': SELECT hostdb_id FROM bind_interactions WHERE bind_interaction_id = '$id' AND molecule = '$molecule';";
					    $hostdb_ids{$bind_id}++;
					}
				    }
				}
			    }
			}
		    }
		    # get HUGO IDs for the collected hostdb_ids
		    foreach my $hostdb_id (sort keys %hostdb_ids) {
			my $rv = $sth_hugo_id->execute($hostdb_id, 'HUGO');
			if ($rv >= 1) {
			    while (my ($acc) = $sth_hugo_id->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $acc;
				$annotation_done{$annotation_item}++;
			    }
			} else {
			    # try SwissProt
			    my $rv = $sth_hugo_id->execute($hostdb_id, 'SWISSPROT');
			    if ($rv >= 1) {
				while (my ($acc) = $sth_hugo_id->fetchrow_array) {
				    push @{$annotation{$annotation_item}{$probe}}, $acc;
				    $annotation_done{$annotation_item}++;
				}
			    } else {
				# try EMBL
				my $rv = $sth_hugo_id->execute($hostdb_id, 'EMBL');
				if ($rv >= 1) {
				    while (my ($acc) = $sth_hugo_id->fetchrow_array) {
					push @{$annotation{$annotation_item}{$probe}}, $acc;
					$annotation_done{$annotation_item}++;
				    }
				} else {
				    # try EnsEMBL
				    my $rv = $sth_hugo_id->execute($hostdb_id, 'Ensembl');
				    if ($rv >= 1) {
					while (my ($acc) = $sth_hugo_id->fetchrow_array) {
					    push @{$annotation{$annotation_item}{$probe}}, $acc;
					    $annotation_done{$annotation_item}++;
					}
				    }  else {
					# use hostdb_id:
					push @{$annotation{$annotation_item}{$probe}}, "NA ($hostdb_id)";
					$annotation_done{$annotation_item}++;
				    }
				}
			    }
			}
		    }
		} elsif ($annotation_item eq 'DIP') {
		    # get DIP id 
		    # prepare DIP statement:
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get DIP id from hostdb_main
			my $rv = $sth_dip->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($id) = $sth_dip->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $id;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'DIP interaction partners') {
		    # get DIP id 
		    # prepare DIP statement:
		    # (see above)
		    my %hostdb_ids = ();
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get DIP id from hostdb_main
			my $rv = $sth_dip_interact_a->execute($hostdb_id);
			if ($rv >= 1) {
			    my %molecules = ();
			    while (my ($id,$molecule) = $sth_dip_interact_a->fetchrow_array) {
				$molecules{$id} = $molecule;
			    }
			    foreach my $id (sort keys %molecules) {
				my $molecule = $molecules{$id};
				# look for the opposite molecule:
				$molecule = ($molecule eq 'A' ? 'B' : 'A');
				my $rv = $sth_dip_interact_b->execute($id,$molecule);
				if ($rv >= 1) {
				    while (my ($id) = $sth_dip_interact_b->fetchrow_array) {
					$hostdb_ids{$id}++ if (defined $id);
				    }
				}
			    }
			}
		    }
		    # get HUGO IDs for the collected hostdb_ids
		    foreach my $hostdb_id (sort keys %hostdb_ids) {
			my $rv = $sth_hugo_id->execute($hostdb_id, 'HUGO');
			if ($rv >= 1) {
			    while (my ($acc) = $sth_hugo_id->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $acc;
				$annotation_done{$annotation_item}++;
			    }
			} else {
			    # try SwissProt
			    my $rv = $sth_hugo_id->execute($hostdb_id, 'SWISSPROT');
			    if ($rv >= 1) {
				while (my ($acc) = $sth_hugo_id->fetchrow_array) {
				    push @{$annotation{$annotation_item}{$probe}}, $acc;
				    $annotation_done{$annotation_item}++;
				}
			    } else {
				# try EMBL
				my $rv = $sth_hugo_id->execute($hostdb_id, 'EMBL');
				if ($rv >= 1) {
				    while (my ($acc) = $sth_hugo_id->fetchrow_array) {
					push @{$annotation{$annotation_item}{$probe}}, $acc;
					$annotation_done{$annotation_item}++;
				    }
				}
			    }
			}
		    }
		} elsif ($annotation_item eq 'BioCarta') {
		    # get BioCarta id 
		    # prepare BioCarta statement:		    
		    # (see above)
		    # we need HUGO terms for this		    
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get BioCarta id from hostdb_main
			my $rv = $sth_biocarta->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($id) = $sth_biocarta->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $id;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'proteinfamily family description') {
		    # get Ensembl Description
		    # prepare Ensembl Description statement:		    
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get Ensembl Description from hostdb_main
			my $rv = $sth_protein_family_desc->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($id) = $sth_protein_family_desc->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $id;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'proteinfamily family id') {
		    # get Ensembl Protein Family
		    # prepare Ensembl Protein Family statement:		    
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get Ensembl Protein Family from hostdb_main
			my $rv = $sth_protein_family_id->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($id) = $sth_protein_family_id->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $id;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'proteinfamily family description') {
		    # get Ensembl Protein Family
		    # prepare Ensembl Protein Family statement:		    
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get Ensembl Protein Family from hostdb_main
			my $rv = $sth_protein_family_description->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($description) = $sth_protein_family_description->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $description;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'interpro interpro id') {
		    # get InterPro id 
		    # prepare InterPro statement:		        
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get InterPro Description from hostdb_main
			my $rv = $sth_IP->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($IP_id,$description) = $sth_IP->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $IP_id;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'interpro description') {
		    # get InterPro id 
		    # prepare InterPro statement:		        
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get InterPro Description from hostdb_main
			my $rv = $sth_IP->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($IP_id,$description) = $sth_IP->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $description;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'domain prediction method') {
		    # get domain method 
		    # prepare domain method statement:		        
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get domain method from ensembl_domain
			my $rv = $sth_domain_method->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($method) = $sth_domain_method->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $method;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'domain accession number') {
		    # get domain accession 
		    # prepare domain accession statement:		        
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get domain accession from ensembl_domain
			my $rv = $sth_domain_accession->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($accession) = $sth_domain_accession->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $accession;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'domain description') {
		    # get domain description 
		    # prepare domain description statement:		        
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get domain description from ensembl_domain
			my $rv = $sth_domain_description->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($description) = $sth_domain_description->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $description;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'domain start') {
		    # get domain start 
		    # prepare domain start statement:		        
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get domain start from ensembl_domain
			my $rv = $sth_domain_start->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($start) = $sth_domain_start->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $start;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'domain end') {
		    # get domain end 
		    # prepare domain end statement:		        
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get domain end from ensembl_domain
			my $rv = $sth_domain_end->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($end) = $sth_domain_end->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $end;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'feature prediction method') {
		    # get feature method 
		    # prepare feature method statement:		        
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get feature method from ensembl_feature
			my $rv = $sth_feature_method->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($method) = $sth_feature_method->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $method;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'feature start') {
		    # get feature start 
		    # prepare feature start statement:		        
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get feature start from ensembl_feature
			my $rv = $sth_feature_start->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($start) = $sth_feature_start->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $start;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'feature end') {
		    # get feature end 
		    # prepare feature end statement:		        
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get feature end from ensembl_feature
			my $rv = $sth_feature_end->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($end) = $sth_feature_end->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, $end;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'TransFac Factor Num') {
		    # get TRANSFAC TYPE 
		    # prepare TRANSFAC_TYPE statement:
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# check for factor related to probe
			my $rv = $sth_transfac_type_factor->execute($hostdb_id);
			if ($rv >= 1) {
			    push @{$annotation{$annotation_item}{$probe}}, $rv;
			    $annotation_done{$annotation_item}++;
			}
		    }
		} elsif ($annotation_item eq 'TransFac Gene Num') {
		    # get TRANSFAC TYPE 
		    # prepare TRANSFAC_TYPE statement:
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# check for gene related to probe
			my $rv = $sth_transfac_type_gene->execute($hostdb_id);
			if ($rv >= 1) {
			    push @{$annotation{$annotation_item}{$probe}}, $rv;
			    $annotation_done{$annotation_item}++;
			} 			
		    }
		} elsif ($annotation_item eq 'TransFac Factor') {
		    # get TRANSFAC_FACTOR id 
		    # prepare TRANSFAC_FACTOR statement:
		    # (see above)

		    # get transfac factors based on hostdb_id (
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			my $rv = $sth_transfac_factor_a->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($id, $description) = $sth_transfac_factor_a->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, "$description ($id)";
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		    # get transfac factors based on gene match to hostdb_id (
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get TRANSFAC_FACTOR id from hostdb_main
			my $rv = $sth_transfac_factor_b->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($id, $description) = $sth_transfac_factor_b->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, "$description ($id)";
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'TransFac Gene') {
		    # get TRANSFAC_GENE id 
		    # prepare TRANSFAC_GENE statement:
		    # (see above)

		    # get transfac genes based on factor match to hostdb_id (
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			my $rv = $sth_transfac_gene_a->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($id, $description) = $sth_transfac_gene_a->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, "$description ($id)";
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		    # get transfac genes based on hostdb_id (
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get TRANSFAC_GENE id from hostdb_main
			my $rv = $sth_transfac_gene_b->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($id, $description) = $sth_transfac_gene_b->fetchrow_array) {
				push @{$annotation{$annotation_item}{$probe}}, "$description ($id)";
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		} elsif ($annotation_item eq 'TransPath') {
		    # get TRANSPATH id 
		    # prepare TRANSPATH statement:
		    # (see above)
		    foreach my $hostdb_id (@{$annotation_hostdb_id{$transcript}}) {
			# get TRANSPATH id from hostdb_main
			my $rv = $sth_transpath->execute($hostdb_id);
			if ($rv >= 1) {
			    while (my ($id) = $sth_transpath->fetchrow_array) {
				my %names = ();
#				if ($id =~ /^M/) {
				# get the molecule names
				my $rv = $sth_transpath_molecule->execute($id);
				if ($rv >= 1) {
				    while (my ($molecule) = $sth_transpath_molecule->fetchrow_array) {
					$names{$molecule}++ if (defined $molecule);
				    }
				}
				$rv = $sth_transpath_gene->execute($id);
				if ($rv >= 1) {
				    while (my ($molecule) = $sth_transpath_gene->fetchrow_array) {
					$names{$molecule}++ if (defined $molecule);
				    }
				}
				my $molecule = $id;
				if (%names) {
				    $molecule = join ',', sort keys %names;
				    $molecule .= " ($id)";
				}
				
				push @{$annotation{$annotation_item}{$probe}}, $molecule;
				$annotation_done{$annotation_item}++;
			    }
			}
		    }
		}		
	    }
	    if (@{$annotation{$annotation_item}{$probe}} > 0) {
		$found_annotation++;
		$found_annotation{$probe}++;
	    } else {
		$no_annotation{$probe}++;
		$no_annotation++;
	    }

#	    if ($probe eq 'H09_H19k_13.1_151101') {		
#		my @tmp = @{$annotation{'HUGO'}{$probe}};
#		warn "#KH: HUGOS ($annotation_item): '@tmp'";
#	    }
#	    last ANNOTATION if ($probe eq 'H09_H19k_13.1_151101');

	} # end of PROBE:foreach my $probe (keys %probes) {

	if ($annotation_item =~ /--.*--/) {
	    $funcdb_out .= "<TR>
<TD ALIGN=CENTER COLSPAN=4 BGCOLOR=LIGHTGREY>$annotation_item</TD>
</TR>
";
	} else {
	    my $found = scalar (keys %found_annotation) + scalar (keys %prev_annotation);
	    if ($found > 0) {
		unless ($listed{$annotation_item}) {
		    $funcdb_out .= "<TR>
<TD>$annotation_item_out</TD>
<TD ALIGN=RIGHT>$found</TD>
<TD ALIGN=RIGHT>".(scalar keys %no_annotation)."</TD>
</TR>
";
		    push @annotation_headers_all, $annotation_item;
		    $listed{$annotation_item}++;
		}	    
	    }
	}

    } # end of ANNOTATION:foreach my $annotation_item (@annotations_funcdb) {

    # End of FuncDB annotation!

    # if we had orthologue hits, add them to orthologues output:
    my $orthologue_out = 0;
    foreach my $item ('THC ENSEMBL HITS', 'THC ENSEMBL GENE', 'THC ENSEMBL TRANSCRIPT', 'THC ENSEMBL COVERAGE', 'THC ENSEMBL IDENTITY', 'THC ENSEMBL GAPS', 'THC ENSEMBL STATUS') {
	if (defined $annotation{'THC ENSEMBL TRANSCRIPT'}) {
	    unless ($orthologue_out) {
		$orthologue_out = 1;
		$out .= "<TR>
    <TD ALIGN=CENTER COLSPAN=4 BGCOLOR=LIGHTGREY><B>-- Putative Orthologues --</B></TD>
    </TR>
    ";
	    }
	    $out .= "<TR>
    <TD>$item</TD>
    <TD ALIGN=RIGHT>".(scalar keys %{$annotation{$item}})."</TD>
    <TD ALIGN=RIGHT>".((scalar keys %probe_names) - (scalar keys %{$annotation{$item}}))."</TD>
    </TR>
    ";
	}
    }


    # add funcdb overview to previous output:
    $out .= $funcdb_out;
    
    $out .= "</TABLE>\n";
    if (%no_probes) {
	$out .= "<BR>For ".(scalar keys %no_probes)." spots no probe/reporter sequences were found!\n";
#	my @tmp = sort keys %no_probes;
#	warn "# No sequences for '".(join ',', @tmp)."'";
    }
#    if (%no_transcripts) {
#	$out .= "<BR>For ".(scalar keys %no_transcripts)." probes/reporters no transcripts were found, i.e. no TC matched or a match outside a gene border!\n";
#	my @tmp = sort keys %no_transcripts;
#	warn "# No hits for '".(join ',', @tmp)."'";
#    }

    # write annotation to file (for speed-up)
    my $write_annotation_file = 0;
    if ($write_annotation_file) {

	if (open (OUT, ">/tmp/annotation.$$")) {
#	    foreach my $annotation_item ( (sort @annotations_funcdb), ('THC ENSEMBL HITS', 'THC ENSEMBL GENE', 'THC ENSEMBL TRANSCRIPT', 'THC ENSEMBL COVERAGE', 'THC ENSEMBL IDENTITY', 'THC ENSEMBL GAPS', 'THC ENSEMBL STATUS')) {
	    foreach my $annotation_item ( (sort @annotations_funcdb)) {
		next if ($annotation_item eq 'all');

		foreach my $probe (sort keys %{$annotation{$annotation_item}}) {
		    next unless (@{$annotation{$annotation_item}{$probe}});
		    print OUT "$annotation_item\t$probe";
		    foreach (@{$annotation{$annotation_item}{$probe}}) {
			next unless (defined $_);
			print OUT "\t$_";
		    }
		    print OUT "\n";
		}
	    }
	    close OUT;
	    system "chmod 777 /tmp/annotation.$$ 2> /dev/null";
	}
    }

    return $out;
}


sub read_annotation {
    my $file = shift;
    my $ref = shift;
    my %annotation_select = %$ref;
    $ref = shift;
    my @annotation_order = @$ref;
    my $dbh = shift;

    my $out = '';

    # prepare orthology statement
    my ($statement_orthologue,$sth_orthologue);
    $statement_orthologue = "SELECT Ensembl_transcript, Ensembl_gene, cover, ident, gaps FROM probelynxdb.bovine_human_060504 WHERE THC = ?";
    $sth_orthologue = $dbh->prepare($statement_orthologue);

    # open annotation file provided by ProbeLynx
    unless ($file) {
	return "No annotation file found - please use <A href=\"http://www.pathogenomics.ca/probelynx\">ProbeLynx</A> and provide a link to the output file in the settings section for the annotation module!";
    }
    open (IN, $file)
	or return "Can't open file $file: $!\n";

    print STDERR "\nReading file '$file' at ".(localtime)."\n";

    $out .= "<TR>
<TD ALIGN=CENTER COLSPAN=4 BGCOLOR=LIGHTGREY>--&nbsp;<B>".(basename $file)."</B>&nbsp;--</TD>
</TR>
";

    my $header = <IN>;
    chomp $header;
    my @headers = ();

    # replace underscore with blank
    # so that headers can be split across multiple lines
    # in webbased output
    foreach (split /\t/, $header) {
	s/_/ /g;
	push @headers, $_;
    }
    
    # make a hash out of headers, so we know where to find certain elements
    my %position = ();
    my %position2 = ();
    for (my $i = 0; $i <= $#headers; $i++) {
	$position{$headers[$i]} = $i;
	$position2{$i} = $headers[$i];
    }

    # unless a header 'probe id' is provided
    # use the first column for probe specification:
    unless (defined $position{'probe id'}) {
	$position{'probe id'} = 0;
    }
    
    my @annotation_items = ();
    my %annotation_items_hash = ();

    my $line_counter = 0;
    my %extra_annotation = ();

    while (<IN>) {
	chomp;

	$line_counter++;
	
	if ($line_counter % 1000 == 0) {
	    print STDERR "$line_counter ";
	}

	my @tmp = split /\t/, $_;
	my $probe = $tmp[$position{'probe id'}];
	next unless (defined $probe_names{$probe});
	for (my $i = 0; $i <= $#tmp; $i++) {
	    my $annotation_item = $position2{$i};
	    next if ($annotation_item eq 'probe id');

	    # check if this annotation_item is wanted
	    next unless (defined $annotation_select{'all'}
			 or
			 defined $annotation_select{$annotation_item});
	    
	    unless ($annotation_items_hash{$annotation_item}) {
		$annotation_items_hash{$annotation_item}++;
		push @annotation_items, $annotation_item;
	    }
	    
	    # get the annotation:
	    my $annotation = $tmp[$i];
	    next if ($annotation eq 'NULL'
		     or
		     $annotation eq '');
	    
	    my @items = ($annotation);
	    if ($annotation =~ /\;/) {
		@items = (split /;/, $annotation);
	    }

	    # add SwissProt info
	    if ($annotation_item eq 'SWISSPROT accession'
		and
		@items) {

		my $match = 0;
		foreach my $id (@items) {
		    if (defined $swissprot{$id}) {
			foreach my $item ('SWISSPROT gene name','SWISSPROT protein name','SWISSPROT function') {
			    push @{$annotation{$item}{$probe}}, $swissprot{$id}{$item};
			    $annotation_done{$item}++ unless ($_ eq '');
			    $match = 1;
			}
		    }
		}
		if ($match) {
		    foreach my $item ('SWISSPROT gene name','SWISSPROT protein name','SWISSPROT function') {
			unless ($annotation_items_hash{$item}) {
			    $annotation_items_hash{$item}++;
			    push @annotation_items, $item;
			}
		    }
		}		
	    }
	    

	    foreach (@items) {
		
		# TC description contains sub-items:
		if ($annotation_item eq 'TC description') {
		    # extract database id's from the description
		    my $desc = $_;
		    if (defined $annotation_select{'all'}
			or
			defined $annotation_select{'SP'}) {
			while ($desc =~ /\bSP\|(.*?) /) {
			    push @{$annotation{'SP'}{$probe}}, (split /\|/, $1);
			    $annotation_done{'SP'}++;
			    $extra_annotation{'SP'}++;
			    $desc =~ s/\bSP\|.*? //;
			}
		    }

		    while ($desc =~ /\bPIR\|(.*?) /) {
			my @tmp = (split /\|/, $1);
			if (defined $annotation_select{'all'}
			    or
			    defined $annotation_select{'PIR'}) {
			    push @{$annotation{'PIR'}{$probe}}, $tmp[1];
			    $annotation_done{'PIR'}++;
			    $extra_annotation{'PIR'}++;
			}
			if (defined $annotation_select{'all'}
			    or
			    defined $annotation_select{'PIR'}) {
			    push @{$annotation{'GP'}{$probe}}, $tmp[1];
			    $annotation_done{'GP'}++;
			    $extra_annotation{'GP'}++;
			}
			$desc =~ s/\bPIR\|.*? //;
		    }

		    if (defined $annotation_select{'all'}
			or
			defined $annotation_select{'GP'}) {
			while ($desc =~ /\bGP\|(.*?) /) {
			    my @tmp = split /\|/, $1;
			    my $gp = $tmp[0];
			    push @{$annotation{'GP'}{$probe}}, $gp;
			    $annotation_done{'GP'}++;
			    $extra_annotation{'GP'}++;
			
			    my $db = $tmp[1];
			    my @ids = ();
			    if ($db eq 'pdb') {
				@ids = ($tmp[2]);
			    } else {
				my $prot = $tmp[2];
				@ids = ($tmp[4]) if (defined $tmp[4]);
				if (defined $tmp[3] 
				    and
				    $tmp[3] ne '') {
				    @ids = ($tmp[3]);
				    push @ids, $tmp[4] if (defined $tmp[4]);
				}
				
				if (defined $annotation_select{'all'}
				    or
				    defined $annotation_select{'protein'}) {				
				    push @{$annotation{'protein'}{$probe}}, $prot;
				    $annotation_done{'protein'}++;
				    $extra_annotation{'protein'}++;
				}
			    }
			    
			    if (defined $annotation_select{'all'}
				or
				defined $annotation_select{$db}) {				
				push @{$annotation{$db}{$probe}}, @ids;
				$annotation_done{$db}++;
				$extra_annotation{$db}++;
			    }
			    $desc =~ s/\bGP\|(.*?) //;
			}
		    }
		}
		
		push @{$annotation{$annotation_item}{$probe}}, $_;
		$annotation_done{$annotation_item}++ unless ($_ eq '');
		
		# new orthology information:
		if ($annotation_item eq 'putative ortholog Ensembl gene'
		    or
		    $annotation_item eq 'Ensembl homolog gene') {
#		    unless (defined $orthologues{$probe}{'GENE'}) {
			push @{$orthologues{$probe}{'GENE'}}, $_;
#		    }
		}
		if ($annotation_item eq 'putative ortholog Ensembl transcript'
		    or
		    $annotation_item eq 'Ensembl homolog transcript') {
#		    unless (defined $orthologues{$probe}{'TRANSCRIPT'}) {
			push @{$orthologues{$probe}{'TRANSCRIPT'}}, $_;
#		    }
		}
		if ($annotation_item eq 'putative ortholog Ensembl cover'
		    or
		    $annotation_item eq 'homolog cover') {
#		    unless (defined $orthologues{$probe}{'COVER'}) {
			push @{$orthologues{$probe}{'COVER'}}, $_;
#		    }
		}
		if ($annotation_item eq 'putative ortholog Ensembl ident'
		    or
		    $annotation_item eq 'homolog ident') {
#		    unless (defined $orthologues{$probe}{'IDENT'}) {
			push @{$orthologues{$probe}{'IDENT'}}, $_;
#		    }
		}
		if ($annotation_item eq 'putative ortholog Ensembl gaps'
		    or
		    $annotation_item eq 'homolog gaps') {
#		    unless (defined $orthologues{$probe}{'GAPS'}) {
			push @{$orthologues{$probe}{'GAPS'}}, $_;
#		    }
		}
		if ($annotation_item eq 'putative ortholog Ensembl status'
		    or
		    $annotation_item eq 'homology relationship') {
#		    unless (defined $orthologues{$probe}{'STATUS'}) {
			push @{$orthologues{$probe}{'STATUS'}}, $_;
#		    }
		}
	    
		# added '2' to practically disable this
		if ($annotation_item eq 'putative ortholog hit identifier2') {
		    unless (defined $orthologues{$probe}{$_}) {
			my $rv = $sth_orthologue->execute($_);
			if ($rv >= 1) {
			    while (my ($orthologue, $gene, $cover, $ident, $gaps, $method) = $sth_orthologue->fetchrow_array) {
				unless (defined $cover) {
				    $cover = '';
				}
				unless (defined $ident) {
				    $ident = '';
				}
				unless (defined $gaps) {
				    $gaps = '';
				}
				
				$orthologues{$probe}{$_}{$orthologue}{'GENE'} = $gene;
				$orthologues{$probe}{$_}{$orthologue}{'COVER'} = $cover;
				$orthologues{$probe}{$_}{$orthologue}{'IDENT'} = $ident;
				$orthologues{$probe}{$_}{$orthologue}{'GAPS'} = $gaps;
			    }
			}
		    }
		}
	    }
	}
    }
    close IN;

    print STDERR "Finished reading file at ".(localtime)."\n";

    # document number of annotated probes:
    foreach my $annotation_item (@annotation_items) {
	if (defined $annotation{$annotation_item}
	    and
	    (scalar keys %{$annotation{$annotation_item}} > 0)) {
	    my $found = scalar keys %{$annotation{$annotation_item}};
	    $out .= "<TR>
<TD>$annotation_item</TD>
<TD ALIGN=RIGHT>$found</TD>
<TD ALIGN=RIGHT>".((scalar keys %probe_names) - $found)."</TD>
</TR>
";
	    push @annotation_headers_all, $annotation_item;
	}
    }

#    @annotation_headers_all = @annotation_order;
    # didn't work because it includes headers such as 'all' and 'THC ENSEMBL'
    # which are not necessary in all cases

    if (%extra_annotation) {
	foreach my $annotation_item (sort keys %extra_annotation) {
	    push @annotation_headers_all, $annotation_item;
	    my $found = scalar keys %{$annotation{$annotation_item}};
	    $out .= "<TR>
<TD>$annotation_item</TD>
<TD ALIGN=RIGHT>$found</TD>
<TD ALIGN=RIGHT>".((scalar keys %probe_names) - $found)."</TD>
</TR>
";
	}
    }

    if (%orthologues) {
#	push @annotation_headers_all, ('THC ENSEMBL GENE', 'THC ENSEMBL TRANSCRIPT', 'THC ENSEMBL COVERAGE', 'THC ENSEMBL IDENTITY', 'THC ENSEMBL GAPS', 'THC ENSEMBL STATUS');
    }

    return 'ok', $out;
}


sub output_file_setup {
    my $action_num = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }
#    my $output_data = &get_user_settings($var_def_default{'output_files'},'output_files',$force_global_opt,$action_num);
    my $output_data = 'separately';
    my @output_format = ('default');

    if ($output_data eq 'separately') {
	@files = ();
#	@output_format = &get_user_settings($var_def_default{'output_format'},'output_format',$force_global_opt,$action_num);
    } else {
	@files = (sort keys %data);
#	@output_format = &get_user_settings($var_def_default{'output_format'},'output_format',$force_global_opt,$action_num,$files[0]);
    }


    # carry out action on each file for each format
    my $output_counter = 0;

    # print spreadsheet section:
    my $spreadsheet = "<!-- Generate spreadsheet table from data --> "
	. "<input type=\"hidden\" name=\"output_type\" value=\"$output_data\"><TABLE>\n<TR><TD>\n"
	. &pointer
	. "</TD><TD COLSPAN=3>Generate interactive spreadsheet table with ".&help( 
										       {
											   'topic' => 'simple_view',
											   'label' => 'pre-selected columns',
											   'no_small' => 1
										       }
										       )
	.": <TABLE CELLPADDING=5><TR><TD BGCOLOR=LIGHTGREEN><A NAME=\"full_html_button\"><input type=\"submit\" name=\"submit\" value=\"Simple Spreadsheet View\"></TD></TR></TABLE></TD></TR>\n<TR VALIGN=TOP><TD>&nbsp;</TD></TR><TR><TD>\n"
	. &pointer
	. "</TD><TD COLSPAN=3>Generate your own interactive spreadsheet table view with the data fields <B>selected below</B>: <TABLE CELLPADDING=5><TR><TD BGCOLOR=GREEN><A NAME=\"full_html_button\"><input type=\"submit\" name=\"submit\" value=\"Extended Spreadsheet View\"></TD></TR></TABLE></TD></TR>\n<TR VALIGN=TOP><TD>&nbsp;</TD>\n";

    # first column: selection of files to be included
    my $file_selection = "<CENTER>".&help( 
										       {
											   'topic' => 'spreadsheet_files',
											   'label' => 'file(s)',
											   'no_small' => 1
										       }
										       )
	.":</CENTER>\n<SELECT name=\"spreadsheet_files\" size=6 multiple>\n";
    my $file_selected = '';
    my %merged_files = ();
    foreach (keys %data) {
	my $file = basename($_);
	if ($file =~ /^$MERGED_FILE_NAME/) {
	    $merged_files{$_}++;
	}
    }

    if (scalar(keys %data) > 1) {
	if (%merged_files) {
	    $file_selection .= "<OPTION>all</OPTION>\n";
	} else {
	    $file_selection .= "<OPTION SELECTED>all</OPTION>\n";
	}   
    }
 
    foreach my $file (sort keys %data) {
	my $select = '';
	if (scalar(keys %data) == 1) {
	    $select = ' SELECTED';
	}
	if (defined $merged_files{$file}) {
	    $select = ' SELECTED';
	}
	$file_selection .= "<OPTION VALUE=\"$file\"$select>".(basename $file)."</OPTION>\n";
    }
    $file_selection .= "</SELECT>\n";
    $spreadsheet .= "<TD>$file_selection</TD>";

    # second column: selection of content
    my $content_selection = "<CENTER>".&help( 
										       {
											   'topic' => 'spreadsheet_content',
											   'label' => 'visible content',
											   'no_small' => 1
										       }
										       )
	.":</CENTER>\n<SELECT name=\"spreadsheet_content\" size=6 multiple>\n";
    my %defaults = ();
    my %defaults_hidden = ();
    if ($FPMI_specific) {
	if ($remote_address =~ /^128.233/
	    or
	    $remote_address =~ /^142.165/) {
	    $defaults{'ratios'}++;
	} else {
	    $defaults{'fold changes'}++;
	}
    } else {
	$defaults{'fold changes'}++;
    }
    
#    $defaults_hidden{'flags'}++;

    # if you add fields here, make sure they are deciphered for both content and hidden selections (see sub combine_files)!!!
    my @spreadsheet_content_hidden_choices = ('everything', 'all values', 'probe and spot ids', 'spot coordinates', 'fold changes', 'ratios');
    my @spreadsheet_content_choices = ('everything', 'all values', 'spot coordinates', 'probe and spot ids', 'fold changes', 'ratios');
    if (defined $annotation{$COMMON_GENE_NAME}) {
	push @spreadsheet_content_choices, ('gene info', 'all annotations');
	push @spreadsheet_content_hidden_choices, ('gene info', 'all annotations');
	$defaults{'gene info'}++;
	$defaults_hidden{'all annotations'}++;
	$defaults_hidden{'probe and spot ids'}++;
    } else {
	$defaults{'probe and spot ids'}++;
    }
    if (grep /Z-/i, @actions
	or
	grep /dep_z/i, @actions
	or
	grep /t-test/i, @actions
	or
	grep /ttest/i, @actions
	or
	grep /permutation_p/i, @actions
	or
	grep /p-value/i, @actions
	or
	grep /correlation/i, @actions
	or
	grep /anova/i, @actions) {
	push @spreadsheet_content_choices, ('significance values');
	push @spreadsheet_content_hidden_choices, ('significance values');
	$defaults{'significance values'}++;
    }
    if (grep /norm/i, @actions) {
	push @spreadsheet_content_choices, ('normalized values');
	push @spreadsheet_content_hidden_choices, ('normalized values');
    }
    if (grep /merge/i, @actions) {
	push @spreadsheet_content_choices, ('merged values');
	push @spreadsheet_content_hidden_choices, ('merged values');
	$defaults_hidden{'merged values'}++;
    }
    if (grep /correction/i, @actions
	or
	grep /background/i, @actions) {
	push @spreadsheet_content_choices, ('bkgd corrected values');
	push @spreadsheet_content_hidden_choices, ('bkgd corrected values');
    }
    push @spreadsheet_content_choices, ('raw values','flags');
    push @spreadsheet_content_hidden_choices, ('raw values','flags');

    foreach my $schema (@spreadsheet_content_choices) {	
	my $select = '';
	if (defined $defaults{$schema}) {
	    $select = ' SELECTED';
	}
	$content_selection .= "<OPTION$select>$schema</OPTION>\n";
    }
    $content_selection .= "</SELECT>\n";
    $spreadsheet .= "<TD>$content_selection</TD>\n";

    # third column: selection of hidden columns
    $content_selection = "<CENTER>".&help( 
										       {
											   'topic' => 'spreadsheet_hidden',
											   'label' => 'included but hidden',
											   'no_small' => 1
										       }
										       )
	.":</CENTER>\n<SELECT name=\"spreadsheet_hidden_content\" size=6 multiple>\n";
    foreach my $schema (@spreadsheet_content_hidden_choices) {	
	my $select = '';
	if (defined $defaults_hidden{$schema}) {
	    $select = ' SELECTED';
	}
	$content_selection .= "<OPTION$select>$schema</OPTION>\n";
    }
    $content_selection .= "</SELECT>\n";
    $spreadsheet .= "<TD>$content_selection</TD>";

    # end of spreadsheet section
    $spreadsheet .= "</TR></TABLE><BR><BR>".(&pointer)."Each file can be downloaded separately in tab-delimited format:<BR><BR>\n";

    print $spreadsheet;

    foreach my $output_format (@output_format) {
	if ($output_format eq 'TIGR MEV') {
	    # files need to be printed separately!		
	    unless (@files) {
		@files = sort keys %data;
	    }
	    my @mev_files = ();
	    foreach my $file (@files) {
		$output_counter++;
		print &output_file($action_num,$action,$output_format,$output_counter,\@output,$file,@files);
		push @mev_files, (basename $file)."_step$action_step.$action_num.mev";
	    }

	    if (grep /_annotate/, @actions) {
		# print annotation file
		$output_counter++;
		print &output_file($action_num,$action,$output_format,$output_counter,'annotation','annotation');
		push @mev_files, "annotation_step$action_step.$action_num.mev";
	    }

	    # create a gzip'ed tar file:
#	    my $dir_now = `pwd`;
#	    chomp $dir_now;
#	    chdir $tmp_dir;
	    my $out_file = "$tmp_dir/mev_files_$progname.$$.tgz";
	    my $job = "tar -C $tmp_dir -zcvf $out_file ".(join ' ', @mev_files);
#	    warn "job: $job, dir: $dir_now";
	    system $job;
	    print "<BR>".(&pointer)." Download <B>all</B> MEV files bundled into one compressed <a href=\"".(basename $out_file)."\">tar archive</a> (".(&file_size($out_file)).")<BR>\n";
#	    chdir $dir_now;	   
	} elsif ($output_format eq 'BASE'){
	    # files need to printed separately!		
	    unless (@files) {
		@files = sort keys %data;
	    }
	    foreach my $file (@files) {
		$output_counter++;
		print &output_file($action_num,$action,$output_format,$output_counter,\@output,$file);
	    }
	} else {
	    unless (@files) {
		foreach my $file (sort keys %data) {
		    $output_counter++;
		    print &output_file($action_num,$action,$output_format,$output_counter,\@output,$file);
		} 
	    } else {
		$output_counter++;
		print &output_file($action_num,$action,$output_format,$output_counter,\@output,@files);
	    }
	}
    }
}

sub boxplot_pt_setup {
    my $action_num = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my @files = &get_user_settings($var_def_default{'boxplot_pt_files'},'boxplot_pt_files',$force_global_opt,$action_num);
    foreach (@files) {
	if ($_ eq 'all') {
	    @files = (sort keys %data);
	    last;
	}	
    }
    
    if ($refine_parameters) {
	print &format_output( # 'boxplot_pt_',$action_num,'','__ALL__',$action,'','Parameter Refinement!<BR>'
			     {
				 'prefix' => 'boxplot_pt_',
				 'num' => $action_num,
				 'label' => '__ALL__',
				 'action' => $action,
				 'info' => 'Parameter Refinement!<BR>'
				 }
			     );
	return;
    }

    foreach my $file (sort @files) {	
	# get header:
	my $header = '';
	my $data_channel = '';
      BOXPLOT_PT_COL:foreach my $column (@{$item_data_column{$action}}) {
	  
	  my %labels = ();
	  my @labels = ();
	  foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	      unless ($label eq '__NO_LABEL__') {
		  push @labels, $label;
	      }
	      if (defined $data_column_available{$file}{$label}{$column}) {
		  $labels{$label}++;;
	      }
	  }

	  # priority to data that is written to '__NO_LABEL__'
	  if (defined $labels{'__NO_LABEL__'}) {
	      $header = $column;
	      $data_channel = '__NO_LABEL__';
	      last BOXPLOT_PT_COL;
	  } else {
	      foreach (@labels) {
		  unless (defined $labels{$_}) {
		      next BOXPLOT_PT_COL;
		  }
	      }
	      # if we arrived here, than all of the labels (except '__NO_LABEL__')
	      # have data defined
	      $data_channel = join ',', @labels;
	      $header = $column;
	      last BOXPLOT_PT_COL;
	  }	      
      }
	
	# check for specific settings:
	$header = &get_user_settings($header,'boxplot_pt_data_column',$force_global_opt,$action_num,'__ALL__');


	# get the labels
	my @work_labels = &get_user_settings($var_def_default{'boxplot_pt_labels'},'boxplot_pt_labels',$force_global_opt,$action_num,'__ALL__');
	foreach (@work_labels) {
	    if (/auto/) {
		if ($header =~ /^Ratio_/
		    or
		    $header =~ /_R$/
		    or
		    $header =~ /_R_M$/) {
		    @work_labels = ('__NO_LABEL__');
		} else {
		    @work_labels = ('CH1,CH2');
		}
	    }
	}
	my $combined = '';
	my $ratio = '';
	foreach (@work_labels) {
	    if (/ratio/) {
		@work_labels = ('__NO_LABEL__');
		$ratio = 1;
		last;
	    } elsif (/combined/) {
		@work_labels = ('CH1,CH2');
		$combined = ', combined';
		last;
	    } elsif (/together/) {
		@work_labels = ('CH1,CH2');
		last;
	    } elsif (/separate/) {
		@work_labels = ('CH1', 'CH2');
		last;
	    }
	}

	# we should check here if the labels are in 
	# accordance with the data_labels that were determined earlier

	foreach my $label (@work_labels) {
	    my ($image,$info) = &signal_boxplotting_printtip($action_num,$action,$header,$label,$combined,$ratio,$file);
	    print &format_output( # 'boxplot_pt_',$action_num,'','__ALL__',$action,$image,$info
				  {
				      'prefix' => 'boxplot_pt_',
				      'num' => $action_num,
				      'file' => $file,
				      'label' => $label,
				      'action' => $action,
				      'image' => $image,
				      'info' => $info
				      }
				  );
	}
    }
}


sub limma_loess_setup {
    my $action_num = shift;
    my $action = shift;
    my $scale = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my @files = &get_user_settings($var_def_default{"limma_loess_${scale}_files"},"limma_loess_${scale}_files",$force_global_opt,$action_num);
    foreach (@files) {
	if ($_ eq 'all') {
	    @files = (sort keys %data);
	    last;
	}	
    }
    
    if ($refine_parameters) {
	print &format_output( 
			     {
				 'prefix' => "limma_loess_${scale}_",
				 'num' => $action_num,
				 'label' => '__ALL__',
				 'action' => $action,
				 'info' => 'Parameter Refinement!<BR>'
				 }
			     );
	return;
    }

    foreach my $file (sort @files) {	

	# get header:

	my $header = '';
	my $data_channel = '';
	my @channels = ();

      LIMMA_LOESS_COL:foreach my $column (@{$item_data_column{$action}}) {
	  
	  my %channels = ();
	  @channels = ();

	  foreach my $ch (sort keys %{$data{$file}{'__DATA__'}}) {
	      unless ($ch eq '__NO_LABEL__') {
		  push @channels, $ch;
	      }
	      if (defined $data_column_available{$file}{$ch}{$column}) {
		  $channels{$ch}++;;
	      }
	  }

	  foreach (@channels) {
	      unless (defined $channels{$_}) {
		  next LIMMA_LOESS_COL;
	      }
	  }
	  
	  # if we arrived here, than all of the channels (except '__NO_LABEL__')
	  # have data defined
	  $data_channel = join ',', @channels;
	  $header = $column;
	  last LIMMA_LOESS_COL;
      }	      
	
	# check for specific settings:
	$header = &get_user_settings($header,"limma_loess_${scale}_data_column",$force_global_opt,$action_num,$file);

	# other parameters:
	my $span = &get_user_settings($var_def_default{"limma_loess_${scale}_span"},"limma_loess_${scale}_span",$force_global_opt,$action_num,$file);
	if ($span =~ /\d/) {
	    unless ($span >= 0
		    and
		    $span <= 1) {
		# TO DO: show error message!
		$span = '';
	    } else {
		$span = ", span=$span";
	    }
	} else {
	    $span = '';
	}

	my $bin = &get_user_settings($var_def_default{"limma_loess_${scale}_bin"},"limma_loess_${scale}_bin",$force_global_opt,$action_num,$file);
	if ($bin =~ /\d/) {
	    unless ($bin >= 0
		    and
		    $bin <= 1) {
		# TO DO: show error message!
		$bin = '';
	    } else {
		$bin = ", bin=$bin";
	    }
	} else {
	    $bin = '';
	}

	my $iter = &get_user_settings($var_def_default{"limma_loess_${scale}_iter"},"limma_loess_${scale}_iter",$force_global_opt,$action_num,$file);
	if ($iter =~ /\d/) {
	    $iter = ", iteration=$iter";
	}


	# determine which flags to skip:
	my %skip_flags = ();
	my @flags = &get_user_settings($var_def_default{"limma_loess_${scale}_skip_flags"},"limma_loess_${scale}_skip_flags",$force_global_opt,$action_num,$file);
	%skip_flags = &get_skip_flags(@flags);
	
	my ($image,$info) = &limma_loess($action_num,$action,$scale,$file,$header,$span,$bin,$iter,\%skip_flags, \@channels);
	print &format_output( 
			      {
				  'prefix' => "limma_loess_${scale}_",
				  'num' => $action_num,
				  'file' => $file,
				  'action' => $action,
				  'image' => $image,
				  'info' => $info
				  }
			      );
    }
}


sub limma_loess {
    my $action_num = shift;
    my $action = shift;
    my $scale = shift;
    my $file = shift;
    my $header = shift;
    my $span = shift;
    my $bin = shift;
    my $iter = shift;
    my $ref = shift;
    my %skip_flags = %$ref;
    $ref = shift;
    my @channels = @$ref;

    my %ma_data = ();
    my %order = ();

    my $info = '';

    $last_column_used = $header;
    
    my $fit_header = 'loessFIT';
    my $out_headerM = 'Norm_L_R';
    if ($header =~ /_M/) {
	$out_headerM = 'Norm_M_L_R';
    }
    
    my $out_headerA = 'Norm_L_I';
    if ($header =~ /_M/) {
	$out_headerA = 'Norm_M_L_I';
    }
    
    my $out_header = 'Norm_L';
    if ($header =~ /_M/) {
	$out_header = 'Norm_M_L';
    }
    
    my $take_log = 1;
    if ($header =~ /_L/) {
	$take_log = 0;
    }


    my %skipped = ();
    my @warnings = ();

    # get maximum array specifications:
    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);

    # go through all spots and store the ones to be used in a container (subgrid or global)
    
    my $grid_key = 'all';
    foreach my $mrow (1..$MROW) {
	foreach my $mcol (1..$MCOL) {

	    if ($scale eq 'printtip') {
		$grid_key = "$mrow,$mcol";
	    }

	    foreach my $row (1..$ROW) {
		foreach my $col (1..$COL) {

		    if (%skip_flags) {
			
			# check for flags
			my @flags = ();
			foreach my $ch (@channels) {
			    
			    # get flags for individual channel 
			    if (defined $data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
				push @flags, (&get_flags($data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
			    }
			}
			
			# and also flag from both channels
			if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			    push @flags, (&get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
			}
			
			# skip flagged spots
			my $skip = 0;
			foreach (@flags) {
			    if (defined $skip_flags{$_}) {
				$skip = 1;
				last;
			    }
			}						
			
			if ($skip) {
			    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$fit_header}{$mrow}{$mcol}{$row}{$col} = '';
			    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_headerM}{$mrow}{$mcol}{$row}{$col} = '';
			    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_headerA}{$mrow}{$mcol}{$row}{$col} = '';
			    $data{$file}{'__DATA__'}{$channels[0]}{$out_header}{$mrow}{$mcol}{$row}{$col} = '';
			    $data{$file}{'__DATA__'}{$channels[1]}{$out_header}{$mrow}{$mcol}{$row}{$col} = '';
			    $skipped{'of flagging'}++;
			    next;
			}		
		    }

		    # collect values

		    my %vals = ();
		    
		    my $skip_reason = '';
		    foreach my $ch (@channels) {
			my $val = '';
			if (defined $data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col}) {
			    $val = $data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col};
			}
			unless ($val =~ /\d/) {
			    $skip_reason = 'at least one of the values is undefined';
			    last;
			}

			if ($take_log) {
			    # only positive values allowed
			    unless ($val > 0) {
				$skip_reason = 'at least one of the values is undefined';
				last;
			    }
			}
			$vals{$ch} = $val;
		    }
		    
		    # initialize output fields:
		    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$fit_header}{$mrow}{$mcol}{$row}{$col} = '';
		    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_headerM}{$mrow}{$mcol}{$row}{$col} = '';
		    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_headerA}{$mrow}{$mcol}{$row}{$col} = '';
		    $data{$file}{'__DATA__'}{$channels[0]}{$out_header}{$mrow}{$mcol}{$row}{$col} = '';
		    $data{$file}{'__DATA__'}{$channels[1]}{$out_header}{$mrow}{$mcol}{$row}{$col} = '';

		    unless (keys %vals == 2) {
			$skipped{$skip_reason}++;
			next;
		    }
		    
		    my $ratio = '';
		    if ($take_log) {
			$ratio = &log_base(2,$vals{$channels[1]}) - &log_base(2,$vals{$channels[0]});
		    } else {
			$ratio = $vals{$channels[1]} - $vals{$channels[0]};
		    }
								       
		    my $product = '';
		    if ($take_log) {
			$product = &log_base(2,$vals{$channels[1]}) + &log_base(2,$vals{$channels[0]});
		    } else {
			$product = $vals{$channels[1]} + $vals{$channels[0]};
		    }
		    $product /= 2;

		    my $pos_key = "$mrow|$mcol|$row|$col";

		    $ma_data{$grid_key}{$pos_key}{'OUT'} = "$ratio\t$product";
		    push @{$order{$grid_key}}, $pos_key;
		}
	    }
	}
    } 
	
    # all data stored in %ma_data
	
    # get loess fit

    my $entries = 0;

    my %grid_data = ();
    
    foreach my $grid (sort keys %ma_data) {
	
	# write data to file for each grid:

	my $data_file = "R_in.$action.$action_num.$action_step.".(basename($file)).".$grid.$PID";
	$data_file =~ s/\//_/g;
	$data_file =~ s/ /_/g;
	$data_file = "$tmp_dir/$data_file";

	# assign an output file for each grid:

	my $R_out = "R_out.$action.$action_num.$action_step.".(basename($file)).".$grid.$PID";
	$R_out =~ s/\//_/g;
	$R_out =~ s/ /_/g;
	$R_out = "$tmp_dir/$R_out";
	
	my $data_counter = 0;
	open (DATAIN, ">$data_file") 
	    or die "Can't write to $data_file: $!";

	foreach my $key (@{$order{$grid}}) {
	    $data_counter++;
	    $entries++;
	    # prints "$ratio\t$product";
	    print DATAIN "$ma_data{$grid}{$key}{'OUT'}\n";
	}
	close DATAIN;

	# keep track of output file and number of elements:
	$grid_data{$grid}{'DATA'} = $data_file;
	$grid_data{$grid}{'RESULTS'} = $R_out;
	$grid_data{$grid}{'COUNTER'} = $data_counter;
    }

    
    if ($entries > 0) {

	# specify R commands for loessFit if we have at least one entry

	my $R_commands_file = "$tmp_dir/R_cmd.$action.$action_num.$action_step.".(basename($file)).".$PID";
	
	open (RH, ">$R_commands_file")
	    or die "Can't write to $R_commands_file: $!";
	
	print RH "library(limma)\n";
	
	# load data and do loess for each grid file:

	my @R_out = ();

	foreach my $grid (sort keys %grid_data) {

	    my $data_file = $grid_data{$grid}{'DATA'};
	    my $data_counter = $grid_data{$grid}{'COUNTER'};
	    next unless ($data_counter > 0);

	    my $R_out = $grid_data{$grid}{'RESULTS'};
	    push @R_out, $R_out;

	    # read in data
	    print RH "data <- read.table(\"$data_file\")\n";
	    
	    print RH "fit = loessFit(data\$V1,data\$V2$span$bin$iter)\$fitted\n";
	    
	    print RH "## write data to file\n";
	    print RH "write.table(fit, \"$R_out\", sep=\"\\t\", quote=F, col.names=F, row.names=F)\n";
	    
	}

	# this calculation might take a while, so we write out a file as a signal
	# that the process has finished
	my $signal_file = $R_commands_file.'.finished';
	if (-e $signal_file) {
	    unlink $signal_file
		or return "Can't delete signal file '$signal_file': $!\n";
	}
	
	print RH "out <- cbind(1,2,3,4,5,6,7,8,9)\n";
	print RH "write.table(out, \"$signal_file\")\n";
	
	close RH;
	
	# carry out loessFit:
	my @images = &get_R_output($R_commands_file, \@R_out, $action_num, $file, $signal_file, '', 1);
	
	foreach my $image (@images) {
	    if ($image =~ /Output from R/) {
		die "$image";
	    }
	}
	
	# read in results for each grid:


	foreach my $grid (sort keys %grid_data) {

	    my $data_file = $grid_data{$grid}{'DATA'};
	    my $data_counter = $grid_data{$grid}{'COUNTER'};
	    unless ($data_counter > 0) {
		push @warnings, "subgrid $grid contains no valid data!";
	    }

	    my $R_out = $grid_data{$grid}{'RESULTS'};
	    
	    open (IN, $R_out)
		or die "Can't read $R_out: $!";
	    
	    my $fit_counter = 0;
	    my @tmp = @{$order{$grid}};
	    while (my $fit = <IN>) {
		chomp $fit;
		$fit_counter++;
		my $pos_key = shift @tmp;
		$ma_data{$grid_key}{$pos_key}{'FIT'} = $fit;
	    }
	    
	    close IN;
	    
	    unless ($fit_counter eq $data_counter) {
		die "# ERROR: only $fit_counter instead of $data_counter values returned from $action!";
	    }
	    
	    foreach my $pos_key (@{$order{$grid}}) {
		# store new values:
		my ($mrow,$mcol,$row,$col) = split /\|/, $pos_key;
		my $fit = $ma_data{$grid_key}{$pos_key}{'FIT'};
		my ($valueM, $valueA) = split /\t/, $ma_data{$grid}{$pos_key}{'OUT'};
		$valueM -= $fit;

		# calculation of channel intensities from ratios and intensity products:
		#
		# M = log(R/G)
		#   = log(R) - log(G)
		# log(R) = M + log(G)  (i)
		#
		# A = 1/2 * log(R*G)
		#   = 1/2 * (log(R) + log(G))
		#
		# insert (i):
		# A = 1/2 * (M + log(G) + log(G))
		# log(G) = A - M/2
		# log(R) = M + log(G)

		my ($numerator, $denominator);
		if ($valueA =~ /\d/
		    and
		    $valueM =~ /\d/) {
		    $denominator = $valueA - $valueM / 2;
		    $numerator = $valueM + $denominator;
		} else {
		    $numerator = '';
		    $denominator = '';
		}

		# store new values:
		
		$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$fit_header}{$mrow}{$mcol}{$row}{$col} = $fit;
		$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_headerM}{$mrow}{$mcol}{$row}{$col} = $valueM;
		$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_headerA}{$mrow}{$mcol}{$row}{$col} = $valueA;
		$data{$file}{'__DATA__'}{$channels[0]}{$out_header}{$mrow}{$mcol}{$row}{$col} = $denominator;
		$data{$file}{'__DATA__'}{$channels[1]}{$out_header}{$mrow}{$mcol}{$row}{$col} = $numerator;
	    }	    	
	
	    if ($clean_up_files) {
		unlink $R_commands_file;
		unlink $R_out;
		unlink $data_file;
	    }
	}
    } else {
	push @warnings, "No valid data found for file $file!";
    }
    
    # take note of available columns for later
    
    my $channel = '__NO_LABEL__';
    
    foreach my $out_header ($out_headerM,$out_headerA,$fit_header) {
	unless (defined $data_column_available{$file}{$channel}{$out_header}) {
	    $data_column_available{$file}{$channel}{$out_header} = 1;
	    push @{$data_column_available{$file}{$channel}{'__ORDER__'}}, $out_header;
	    unless (grep /^$out_header$/, @output) {
		push @output, $out_header;
	    }
	}
	
	$data_done{$file}{$channel}{$out_header}++;
    }
    
    foreach my $channel (@channels) {
	unless (defined $data_column_available{$file}{$channel}{$out_header}) {
	    $data_column_available{$file}{$channel}{$out_header} = 1;
	    push @{$data_column_available{$file}{$channel}{'__ORDER__'}}, $out_header;
	    unless (grep /^$out_header$/, @output) {
		push @output, $out_header;
	    }
	    
	    $data_done{$file}{$channel}{$out_header}++;
	}
    }
    
    
    $info .= "<B>$entries</B> ";
    $info .= "entries processed<BR>\n";

    foreach (@warnings) {
	$info.= "<B>WARNING: $_</B><BR><BR>\n";
    }
	    
    foreach (keys %skipped) {
	$info .= "<B>$skipped{$_}</B> spots were skipped because $_!<BR>\n";
    }
    
    $info .= "Normalized ratios can be found in column <B>$out_headerM</B><BR>\n";

    return '', $info;

}


sub ttest_one_group_setup {
    my $action_num = shift;
    my $action = shift;
    my $ttest_method = shift || 'wilcox';

    if ($refine_parameters) {
	print &format_output( 
			     {
				 'prefix' => $ttest_method.'_ttest_1_',
				 'num' => $action_num,
				 'label' => '__ALL__',
				 'action' => $action,
				 'info' => 'Parameter Refinement!<BR>'
				 }
			     );
	return;
    }

    my @files = sort keys %data;

    my %channel_assign_combined = &channel_assign(@files, '__combine__');  
    my %channel_assign = &channel_assign(@files);  
    
    my @channel_keys = sort keys %channel_assign;

    my $key_num = scalar @channel_keys;

    unless ($key_num > 0) {
	print &format_output( 
			     {
				 'prefix' => $ttest_method.'_ttest_1_',
				 'num' => $action_num,
				 'action' => $action,
				 'info' => "No channel assignments found!"
				 }
			     );
	next;
    }
    
    my %keys_done = ();

    # list the assignments for each file
    my %file_assign = ();
    foreach my $assign (keys %channel_assign) {
	foreach my $file (keys %{$channel_assign{$assign}}) {
	    my $channel = $channel_assign{$assign}{$file}{'CH'};
	    $file_assign{$file}{$channel}{$assign}++;
	}
    }

    my $combined_channel = 1;
    if ($combined_channel) {
	# new merging based on keys from both channels:
	foreach my $assign (sort keys %channel_assign_combined) {
	    my %data1 = ();
	    my %data2 = ();
	    my @files = sort keys %{$channel_assign_combined{$assign}};
	    my ($assign1,$assign2) = split /\|/, $assign;
	    
	    unless (@files > 1) {
		print &format_output( 
				      {
					  'prefix' => $ttest_method.'_ttest_1_',
					  'num' => $action_num,
					  'label' => '__ALL__',
					  'action' => $action,
					  'info' => "Only one file for '$assign1' vs '$assign2' found - no test possible!"
					  }
				      );
		next;
	    }

	    if (@files == 2) {
		# unless usage of duplicate neighbours is selected this is useless for ratios!
	    }
		
	    foreach my $file (@files) {
		if (defined $file_assign{$file}{'CH1'}{$assign1}) {
		    $data1{$file} = 'CH1';
		} else {
		    $data1{$file} = 'CH2';
		}
		if (defined $file_assign{$file}{'CH1'}{$assign2}) {
		    $data2{$file} = 'CH1';
		} else {
		    $data2{$file} = 'CH2';
		}
	    }
	

	    # carry out action on specified data
#	warn "plot $file1, $ch1 against $file2, $ch2";
	    
	    my ($image,$info) = &ttest_one_group($action_num,$ttest_method,$assign1,$assign2,\%data1,\%data2,$action);
	    
	    print &format_output( 
				  {
				      'prefix' => $ttest_method.'_ttest_1_',
				      'num' => $action_num,
				      'label' => '__ALL__',
				      'image' => $image,
				      'action' => $action,
				      'info' => $info
				      }
				  );
	}
    } else {

	# do a merge based on individual channels:

	while (my $assign1 = shift @channel_keys) {
	    
	    next if ($keys_done{$assign1});
	    $keys_done{$assign1}++;
	    
	    my @files = sort keys %{$channel_assign{$assign1}};
	    unless (@files) {
		print &format_output( 
				      {
					  'prefix' => $ttest_method.'_ttest_1_',
					  'num' => $action_num,
					  'action' => $action,
					  'image' => '-',
					  'info' => "No files for channel assignment '$assign1' found!"
					  }
				      );
		next;
	    }
	    
	    # determine channels for each file
	    my %channels = ();
	    foreach my $file (@files) {
		my $channel = $channel_assign{$assign1}{$file}{'CH'};
		$channels{$file} = $channel;
	    }
	    
	    my $assign2 = '';
	    
	    my @left_keys = ();
	    foreach my $assign (@channel_keys) {
		unless (defined $keys_done{$assign}) {
		    push @left_keys, $assign;
		}
	    }
	    
	    unless (@left_keys) {
		print &format_output( 
				      {
					  'prefix' => $ttest_method.'_ttest_1_',
					  'num' => $action_num,
					  'action' => $action,
					  'image' => '-',
					  'info' => "No matching channel assignment for '$assign1' found!"
					  }
				      );
		next;
	    }
	    
	    # we might only have two keys in which case the combination is obvious:
	    if (@left_keys == 1) {
		$assign2 = shift @left_keys;
		$keys_done{$assign2}++;
	    }
	    
	    # otherwise, go through all the files for the first key 
	    # and check which other keys are predominant for them...
	    my %second_key = ();
	    foreach my $file (@files) {
		foreach my $channel (keys %{$file_assign{$file}}) {
		    next if ($channel eq $channels{$file});
		    foreach ( sort { $file_assign{$file}{$channel}{$b} <=> $file_assign{$file}{$channel}{$a} } keys %{$file_assign{$file}{$channel}} ) {
			$second_key{$_}++;
			# only get the assignment with the highest occurences
			last;
		    }
		}
	    }
	    
	    foreach (sort { $second_key{$b} <=> $second_key{$a} } keys %second_key) {
		$assign2 = $_;
		$keys_done{$assign2}++;
		last;
	    }
	    
	    # now that we have assign1 and assign2 let's assemble arrays with info
	    # regarding which channels to use for control and treatment:
	    my %data1 = ();
	    my %data2 = ();
	    foreach my $file (sort keys %{$channel_assign{$assign1}}) {
		my $channel = $channel_assign{$assign1}{$file}{'CH'};
		$data1{$file} = $channel;
	    }
	    foreach my $file (sort keys %{$channel_assign{$assign2}}) {
		my $channel = $channel_assign{$assign2}{$file}{'CH'};
		$data2{$file} = $channel;
	    }
	    
	    # carry out action on specified data
#	warn "plot $file1, $ch1 against $file2, $ch2";
	    
	    my ($image,$info) = &ttest_one_group($action_num,$ttest_method,$assign1,$assign2,\%data1,\%data2,$action);
	    
	    print &format_output( 
				  {
				      'prefix' => $ttest_method.'_ttest_1_',
				      'num' => $action_num,
				      'label' => '__ALL__',
				      'image' => $image,
				      'action' => $action,
				      'info' => $info
				      }
				  );
	}
    }
}


sub ttest_one_group {
    my $action_num = shift;
    my $ttest_method = shift;
    my $assign1 = shift;
    my $assign2 = shift;
    my $data1_ref = shift;
    my $data2_ref = shift;
    my $action = shift;

    my %data1 = %$data1_ref;
    my %data2 = %$data2_ref;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $data1_cols = scalar keys %data1;
    my $data2_cols = scalar keys %data2;

    # get a header that is available for all files and channels
    my $required_occ = $data1_cols + $data2_cols;
    my $header = '';
    foreach my $column (@{$item_data_column{$action}}) {

	my $occ = 0;
	
	foreach my $file (keys %data1) {
	    my $ch = $data1{$file};
	    if ($data_column_available{$file}{$ch}{$column}) {
		$header = $column;
		$occ++;
	    }
	}

	foreach my $file (keys %data2) {
	    my $ch = $data2{$file};
	    if ($data_column_available{$file}{$ch}{$column}) {
		$header = $column;
		$occ++;
	    }
	}

	if ($occ == $required_occ) {
	    last;
	} else {
	    $header = '';
	}
    }
    
    # check for specific settings:
    ($header) = &get_user_settings($header,$ttest_method.'_ttest_1_data_column',$force_global_opt,$action_num,'','__ALL__');

    my $log_transfered = 0;
    my $mu = 1;
    if ($header =~ /_L/) {
	$log_transfered = 1;
	$mu = 0;
    }

    $last_column_used = $header;

    my $input_type = '';
    if ($ttest_method eq 'paired') {
	# a paired t-test requires intensities as input:
	$input_type = 'intensities';
    } else {
	$input_type = &get_user_settings($var_def_default{$ttest_method.'_ttest_1_input'},$ttest_method.'_ttest_1_input',$force_global_opt,$action_num,'','__ALL__');
    }

    my $out_header_tmp = &get_user_settings($var_def_default{$ttest_method.'_ttest_1_outheader'},$ttest_method.'_ttest_1_outheader',$force_global_opt,$action_num,'','__ALL__');
    my @out_headers = ();
    if ($out_header_tmp eq 'both') {
	@out_headers = ('p-value', 'p-'.$ttest_method);
    } else {
	@out_headers = ($out_header_tmp);
    }

    my $out_channel = &get_user_settings($var_def_default{$ttest_method.'_ttest_1_outchannel'},$ttest_method.'_ttest_1_outchannel',$force_global_opt,$action_num,'','__ALL__');

    my $dup_usage = &get_user_settings($var_def_default{$ttest_method.'_ttest_1_dups'},$ttest_method.'_ttest_1_dups',$force_global_opt,$action_num,'','__ALL__');
    if ($dup_usage eq 'yes') {
	$dup_usage = 1;
    } else {
	$dup_usage = 0;
    }
    
    # determine which flags to skip:
    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{$ttest_method.'_ttest_1_skip_flags'},$ttest_method.'_ttest_1_skip_flags',$force_global_opt,$action_num,'','__ALL__');
    %skip_flags = &get_skip_flags(@flags);

    my $skip_flags_all = 0;
    my %skipped = ();
    my $no_flag = 0;
    my $entries = 0;
    my $flag_threshold = 0;

    # define data file:
    my $data_file = "R_in.$assign1.$assign2.$action.$action_num.$action_step.$PID";
    $data_file =~ s/\//_/g;
    $data_file =~ s/ /_/g;
    $data_file = "$tmp_dir/$data_file";
    open (DATAIN, ">$data_file") 
	or die "Can't write to $data_file: $!";
	
    # write out data:
    my @files1 = (sort keys %data1);
    my @files2 = (sort keys %data2);
    my @files = (@files1, @files2);
    my %valid_entries = ();
     
    # get maximum array specifications:
    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@files);

    # go through each field
    my @coords = ();
    foreach my $mrow (1..$MROW) {
	foreach my $mcol (1..$MCOL) {
	    foreach my $row (1..$ROW) {

		my %done_col = ();
	      COL:foreach my $col (1..$COL) {
		  
		  $entries++;
		  
		  # initialize output field:
		  for (my $i = 0; $i <= $#files; $i++) {
		      my $file = $files[$i];
		      my $ch = '';
		      if ($i < $data1_cols) {
			  $ch = $data1{$file};
		      } else {
			  $ch = $data2{$file};
		      }

		      if ($out_channel eq 'file') {
			  $ch = '__NO_LABEL__';
		      }

		      foreach my $out_header (@out_headers) {
			  $data{$file}{'__DATA__'}{$ch}{$out_header}{$mrow}{$mcol}{$row}{$col} = '';
		      }
		  }

		  next COL if ($done_col{$col});

		  my $range = 0;
		    
		  if ($dup_usage) {
		      # we do this only for the first file
		      
		      my $file = $files1[0];
		      my $spot = '';
		      my $next_spot = '';

		      # get current spot label:
		      if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col}) {
			  $spot = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col};
			  
			  # special settings for Arrayvision:
			  # delete numerator placed in brackets behind each gene name
			  $spot =~ s/ \(\d+\)$// if (defined $file_specs{$file}{'FORMAT'} 
						     and
						     $file_specs{$file}{'FORMAT'} eq 'ArrayVision');
		      }
		      
		      # get next spot label:
		      if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col+1}) {
			  $next_spot = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col+1};
			  
			  # special settings for Arrayvision:
			  # delete numerator placed in brackets behind each gene name
			  $next_spot =~ s/ \(\d+\)$// if (defined $file_specs{$file}{'FORMAT'} 
							  and
							  $file_specs{$file}{'FORMAT'} eq 'ArrayVision');
		      }
		      
		      unless ($spot eq $next_spot) {
			  # mismatch of spot labels,
			  # skip to next;
			  next COL;
		      } else {
			  # mark next column as done
			  $done_col{$col+1}++;
			  $range = 1;
		      }
		  }

		  my @val = ();
		  my %ratios = ();
		  my $skip_flags = 0;
		  my $valid_spots = 0;

		  # retrieve values from all channels
		  for (my $i = 0; $i <= $#files; $i++) {
		      my $file = $files[$i];
		      my $ch = '';
		      if ($i < $data1_cols) {
			  $ch = $data1{$file};
		      } else {
			  $ch = $data2{$file};
		      }

		      for (my $j = 0; $j <= $range; $j++) {

			  my $val = '';
			  
			  if (defined $data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col+$j}) {
			      $val = $data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col+$j};
			  }

			  my @flags = ();
			  # get flags for this channel as well
			  if (defined $data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}) {
			      push @flags, (&get_flags($data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}));
			  }
			  
			  # and also flag from both channels
			  if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}) {
			      push @flags, (&get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}));
			  }
			  
			  # skip flagged spots
			  my $skip = 0;
			  if (%skip_flags) {
			      
			      foreach (@flags) {
				  if (defined $skip_flags{$_}) {
				      $skip = 1;
				      last;
				  }
			      }			
			  }
			  
			  if ($skip) {
			      $skip_flags++;
			      $val = 'NA';
			  }

			  # skip unless value is numeric
			  unless ($val =~ /\d/) {
			      $val = 'NA';
			  } else {
			      $valid_spots++;
			  }
			  
			  push @val, $val;
			  $ratios{$file}{$col+$j}{$ch} = $val;
		      }
		  }

#		  unless ($valid_spots > 1) {
#		      $skipped{'not enough valid intensities amongst all replicates'}++;
#		      next COL;
#		  }

#		  if ($skip_flags > $flag_threshold) {
#		      $skip_flags_all++;
#		      next COL;
#		  } else {
#		      $no_flag++;
#		  }
	      
		  my @ratios = ();
		  my $valid_ratios = 0;
		  foreach my $file (sort keys %ratios) {
		      foreach my $pos (sort { $a <=> $b } keys %{$ratios{$file}}) {
			  unless (scalar(keys %{$ratios{$file}{$pos}}) == 2) {
			      next;
			  }
			  my $val1 = $ratios{$file}{$pos}{'CH1'};
			  my $val2 = $ratios{$file}{$pos}{'CH2'};

			  my $ratio = 'NA';

			  if ($val1 =~ /\d/
			      and
			      $val2 =~ /\d/) {
			      # ratio calculation depends on type of data
			      if ($log_transfered) {
				  $ratio = $val2 - $val1;
				  $valid_ratios++;
			      } else {
				  unless ($val1 == 0) {
				      $ratio = $val2 / $val1;
				      $valid_ratios++;
				  }
			      }		
			  }      
			  push @ratios, $ratio;
		      }
		  }		  

		  # add information about the number of values that go into the calculation:
		  foreach my $file (@files) {
		      for (my $j = 0; $j <= $range; $j++) {
			  $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$VALID_ENTRIES_TERM}{$mrow}{$mcol}{$row}{$col+$j} = $valid_ratios;
		      }
		  }

		  unless ($valid_ratios > 1) {
		      $skipped{'there were less than two valid ratios amongst all replicates'}++;
		      next COL;
		  }

		  $no_flag++;

		  # store values in specified destination
		  if ($input_type eq 'intensities') {
		      print DATAIN (join "\t", @val)."\n";
		  } else {
		      print DATAIN (join "\t", @ratios)."\n";
		  }

		  my $coord = join '|', ($mrow,$mcol,$row,$col);
		  push @coords, $coord;

		  
	      }
	    }
	}
    }
    
    close DATAIN;

    my $ratio_operator = '/';
    if ($log_transfered) {
	$ratio_operator = '-';
    }

    my $info = "Keys <B>$assign1</B> vs <B>$assign2</B>:<BR><TABLE><TR><TD><TABLE>";
    if ($input_type eq 'intensities') {
	foreach my $file (@files1) {
	    my $ch = $data1{$file};
	    
	    # get channel label name
	    my $ch_out = '';
	    if ($ch eq 'CH1') {
		if (defined $file_specs{$file}{'CH1_LABEL'}) {
		    $ch_out = " ($file_specs{$file}{'CH1_LABEL'})";
		}
	    } elsif ($ch eq 'CH2') {
		if (defined $file_specs{$file}{'CH2_LABEL'}) {
		    $ch_out = " ($file_specs{$file}{'CH2_LABEL'})";
		}
	    }
	    
	    $info .= "<TR><TD>".(basename $file).", $ch$ch_out</TD></TR>\n";
	}
	$info .= "</TABLE>\n</TD><TD>vs</TD><TD>\n<TABLE>\n";
	foreach my $file (@files2) {
	    my $ch = $data2{$file};
	    
	    # get channel label name
	    my $ch_out = '';
	    if ($ch eq 'CH1') {
		if (defined $file_specs{$file}{'CH1_LABEL'}) {
		    $ch_out = " ($file_specs{$file}{'CH1_LABEL'})";
		}
	    } elsif ($ch eq 'CH2') {
		if (defined $file_specs{$file}{'CH2_LABEL'}) {
		    $ch_out = " ($file_specs{$file}{'CH2_LABEL'})";
		}
	    }
	    
	    $info .= "<TR><TD>".(basename $file).", $ch$ch_out</TD></TR>\n";
	}
    } else {

	foreach my $file1 (@files1) {
	    
	    # get channel label name
	    my $ch1 = $data1{$file1};
	    my $ch1_out = '';
	    if ($ch1 eq 'CH1') {
		if (defined $file_specs{$file1}{'CH1_LABEL'}) {
		    $ch1_out = " ($file_specs{$file1}{'CH1_LABEL'})";
		}
	    } elsif ($ch1 eq 'CH2') {
		if (defined $file_specs{$file1}{'CH2_LABEL'}) {
		    $ch1_out = " ($file_specs{$file1}{'CH2_LABEL'})";
		}
	    }

	    my $file2 = shift @files2 || 'undefined';
	    
	    # get channel label name
	    my $ch2 = $data2{$file2};
	    my $ch2_out = '';
	    if ($ch2 eq 'CH1') {
		if (defined $file_specs{$file2}{'CH1_LABEL'}) {
		    $ch2_out = " ($file_specs{$file2}{'CH1_LABEL'})";
		}
	    } elsif ($ch2 eq 'CH2') {
		if (defined $file_specs{$file2}{'CH2_LABEL'}) {
		    $ch2_out = " ($file_specs{$file2}{'CH2_LABEL'})";
		}
	    }

	    $info .= "<TR><TD>".(basename $file2).", $ch2$ch2_out&nbsp;$ratio_operator";
	    $info .= "&nbsp;".(basename $file1).", $ch1$ch1_out\n";
	    $info .= "</TD></TR>\n";
	}
	$info .= "</TABLE>\n</TD><TD>vs</TD><TD>\n<TABLE>\n";
	$info .= "<TR><TD>$mu</TD></TR>";
    }

    $info .= "</TABLE>\n</TD></TR>\n</TABLE>\n<BR>";
	

    unless ($no_flag) {
	$info .= "<B>No input generated!</B>\n";
	return ('', $info);
    }

    # carry out t-test
    my $R_commands_file = "$tmp_dir/R_cmd.$assign1.$assign2.$action.$action_num.$action_step.$PID";
    my $R_out = "$tmp_dir/R_out.$assign1.$assign2.$action.$action_num.$action_step.$PID";
    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";
    
    print RH "library(ctest)\n";
    
    # read in data
    print RH "dataf <- read.table(\"$data_file\")\n";
    
    # define function
    if ($input_type eq 'intensities') {
	print RH "get.pval.$ttest_method.ttest <- function(dataf,index1,index2,datafilter=as.numeric){\n";	
	print RH "f <- function(i) {\n";
	if ($ttest_method eq 'welch') {
	    print RH "return(t.test(datafilter(dataf[i,index1]), datafilter(dataf[i,index2]))\$p.value)\n";	
	} elsif ($ttest_method eq 'student') {
	    print RH "return(t.test(datafilter(dataf[i,index1]), datafilter(dataf[i,index2]), var.equal = TRUE )\$p.value)\n";	
	} elsif ($ttest_method eq 'paired') {
	    print RH "return(t.test(datafilter(dataf[i,index1]), datafilter(dataf[i,index2]), paired = TRUE )\$p.value)\n";	
	} elsif ($ttest_method eq 'wilcox') {
	    print RH "return(wilcox.test(datafilter(dataf[i,index1]), datafilter(dataf[i,index2]))\$p.value)\n"; 
	}
	print RH "}\n";	
	print RH "return(sapply(1:length(dataf[,1]),f))\n";	
	print RH "}\n";
	
	# call function
	my $index1 = $data1_cols * (1 + $dup_usage);
	my $index2 = $index1 + 1;
	my $index3 = $index1 * 2;
	print RH "pVal.$ttest_method.ttest <- get.pval.$ttest_method.ttest(dataf,1:$index1,$index2:$index3)\n";
    } else {
	print RH "get.pval.$ttest_method.ttest <- function(dataf,datafilter=as.numeric){\n";	
	print RH "f <- function(i) {\n";	
	if ($ttest_method eq 'welch') {
	    print RH "return(t.test(datafilter(dataf[i,]), mu=$mu)\$p.value)\n";	
	} elsif ($ttest_method eq 'student') {
	    print RH "return(t.test(datafilter(dataf[i,]), mu=$mu, var.equal = TRUE )\$p.value)\n";	
	} elsif ($ttest_method eq 'wilcox') {
	    print RH "return(wilcox.test(datafilter(dataf[i,]), mu=$mu)\$p.value)\n";
	}
	print RH "}\n";	
	print RH "return(sapply(1:length(dataf[,1]),f))\n";	
	print RH "}\n";
	
	# call function
	print RH "pVal.$ttest_method.ttest <- get.pval.$ttest_method.ttest(dataf)\n";
    }

    # write out data
    print RH "write.table(pVal.$ttest_method.ttest, \"$R_out\", sep=\"\\t\", quote=F, col.names=F, row.names=F)\n";
 
    # this calculation might take a while, so we write out a file as a signal
    # that the process has finished
    my $signal_file = $R_out.'.finished';
    if (-e $signal_file) {
	unlink $signal_file
	    or return "Can't delete signal file '$signal_file': $!\n";
    }

    print RH "out <- cbind(1,2,3,4,5,6,7,8,9)\n";
    print RH "write.table(out, \"$signal_file\")\n";
    
    close RH;

    my $image = &get_R_output($R_commands_file, $R_out, $action_num, 'all_files', $signal_file, '', 1);
    
    unless (-e $R_out) {
	die "$image";
    }

    open (IN, $R_out)
	or die "Can't read $R_out: $!";

    my $nas = 0;
    $counter = 0;
    while (my $val = <IN>) {
	chomp $val;

	if ($val =~ /na/i) {
	    $nas++;
	    $val = '';
	}

	$counter++;

	# get coordinates and background corrected value
	my ($mrow,$mcol,$row,$col) = split /\|/, shift @coords;

	
	# write back values for each spot label
	foreach my $file (@files1) {
	    my $ch = $data1{$file};

	    if ($out_channel eq 'file') {
		$ch = '__NO_LABEL__';
	    }

	    foreach my $out_header (@out_headers) {
		$data{$file}{'__DATA__'}{$ch}{$out_header}{$mrow}{$mcol}{$row}{$col} = $val;
		if ($dup_usage) {
		    $data{$file}{'__DATA__'}{$ch}{$out_header}{$mrow}{$mcol}{$row}{$col+1} = $val;
		}
	    }
	}

	foreach my $file (@files2) {
	    my $ch = $data2{$file};

	    if ($out_channel eq 'file') {
		$ch = '__NO_LABEL__';
	    }

	    foreach my $out_header (@out_headers) {
		$data{$file}{'__DATA__'}{$ch}{$out_header}{$mrow}{$mcol}{$row}{$col} = $val;
		if ($dup_usage) {
		    $data{$file}{'__DATA__'}{$ch}{$out_header}{$mrow}{$mcol}{$row}{$col+1} = $val;
		}
	    }
	}
    }
    close IN;

    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	unlink $data_file;
    }

    # take note of available columns for later
    for (my $i = 0; $i <= $#files; $i++) {
	my $file = $files[$i];
	my $channel = '';
	if ($i < $data1_cols) {
	    $channel = $data1{$file};
	} else {
	    $channel = $data2{$file};
	}

	if ($out_channel eq 'file') {
	    $channel = '__NO_LABEL__';
	}

	foreach my $out_header (@out_headers) {
	    unless (defined $data_column_available{$file}{$channel}{$out_header}) {
		$data_column_available{$file}{$channel}{$out_header} = 1;
		push @{$data_column_available{$file}{$channel}{'__ORDER__'}}, $out_header;
	    }
	    
	    $data_done{$file}{$channel}{$out_header}++;
	}
	unless (defined $data_column_available{$file}{'__NO_LABEL__'}{$VALID_ENTRIES_TERM}) {
	    $data_column_available{$file}{'__NO_LABEL__'}{$VALID_ENTRIES_TERM} = 1;
	    push @{$data_column_available{$file}{'__NO_LABEL__'}{'__ORDER__'}}, $VALID_ENTRIES_TERM;
	    $data_done{$file}{'__NO_LABEL__'}{$VALID_ENTRIES_TERM}++;
	}
    }

    $info .= "<B>$no_flag</B> ";
    if ($dup_usage) {
	$info .= 'duplicated ';
    }
    $info .= "entries processed<BR>\n";
    if ($skip_flags_all) {
	$info .= "<B>$skip_flags_all</B> spots skipped due to flagging.<BR>\n";
    }
    if ($nas) {
	$info .=  "<B>$nas</B> undefined spots<BR>\n";
    }
    $info .=  "<BR>\n";

    unless ($counter == $no_flag) {
	$info .= "<B>WARNING: only $counter rows (instead of $no_flag) returned from $ttest_method t-test (file $R_out)!</B><BR><BR>\n";
    }
	    
    foreach (keys %skipped) {
	$info .= "<B>$skipped{$_}</B> spots were skipped because<BR>$_!<BR>\n";
    }
    
    return $image, $info;
} 


sub correlation_coefficient_setup {
    my $action_num = shift;
    my $action = shift;
    my %file_setup = ();

    if ($refine_parameters) {
	print &format_output( # 'correlation_coefficient_ ',$action_num,'','__ALL__',$action,'','Parameter Refinement!<BR>'
			     {
				 'prefix' => 'correlation_coefficient_',
				 'num' => $action_num,
#				 'label' => '__ALL__',
				 'label' => '-',
				 'action' => $action,
				 'info' => 'Parameter Refinement!<BR>'
				 }
			     );
	return;
    }

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my @files = &get_user_settings($var_def_default{'correlation_coefficient_files'},'correlation_coefficient_files',$force_global_opt,$action_num);

    my ($rounding) = &get_user_settings($var_def_default{'correlation_coefficient_rounding'},'correlation_coefficient_rounding',$force_global_opt,$action_num);
    my ($use_ratio) = &get_user_settings($var_def_default{'correlation_coefficient_ratio'},'correlation_coefficient_ratio',$force_global_opt,$action_num);
    my @data_type = ($use_ratio);
    if ($use_ratio eq 'both') {
	@data_type = ('intensities', 'ratios');
    }

    my ($method) = &get_user_settings($var_def_default{'correlation_coefficient_method'},'correlation_coefficient_method',$force_global_opt,$action_num);
    my @methods = ($method);
    if ($method eq 'both') {
	@methods = ('Spearman', 'Pearson');
    }	

    my ($take_log) = &get_user_settings($var_def_default{'correlation_coefficient_log'},'correlation_coefficient_log',$force_global_opt,$action_num);
    if ($take_log eq 'yes') {
	$take_log = 1;
    } else {
	$take_log = 0;
    }

    if (grep /^all$/, @files) {

	@files = sort keys %data;
	@{$file_setup{'01 all vs all'}} = @files;

    } elsif (grep /^within treatment groups$/, @files) {

	@files = sort keys %data;

	my $counter = 0;
	my $counter_out = 0;
	my %counter = ();
	foreach my $file (@files) {
	    
	    my $full_file = $file;
	    
	    my ($ch1_assign,$ch2_assign);
	    
	    if (defined $cgi->param("CH1_ANNOTATE_$full_file")) {
		$ch1_assign = $cgi->param("CH1_ANNOTATE_$full_file");
	    }
	    if (defined $cgi->param("CH2_ANNOTATE_$full_file")) {
		$ch2_assign = $cgi->param("CH2_ANNOTATE_$full_file");
	    }
	    
	    if ($counter{"$ch1_assign vs $ch2_assign"}) {
		$counter_out = $counter{"$ch1_assign vs $ch2_assign"};
	    } else {
		$counter++;
		$counter_out = $counter;
		until ((length $counter_out) >= 3) {
		    $counter_out = '0'.$counter_out;
		}
		$counter{"$ch1_assign vs $ch2_assign"} = $counter_out;
	    }
	    push @{$file_setup{"$counter_out $ch1_assign vs $ch2_assign"}}, $file;
	}
    } elsif (grep /^between technical replicates$/, @files) {

        @files = sort keys %data;

        my $counter = 0;
        my $counter_out = 0;
        my %counter = ();
        foreach my $file (@files) {

            my $full_file = $file;

            my ($ch1_assign,$ch2_assign, $bio_rep);

            if (defined $cgi->param("CH1_ANNOTATE_$full_file")) {
                $ch1_assign = $cgi->param("CH1_ANNOTATE_$full_file");
            }
            if (defined $cgi->param("CH2_ANNOTATE_$full_file")) {
                $ch2_assign = $cgi->param("CH2_ANNOTATE_$full_file");
            }
            if (defined $cgi->param("BIO_REP_$full_file")) {
                $bio_rep = $cgi->param("BIO_REP_$full_file");
            }
	    if ($bio_rep eq '') {
		$bio_rep = 'bio_rep1';
	    } else {
		if ($bio_rep =~ /^\d/) {
		    $bio_rep = 'B'.$bio_rep;
		}
	    }

            if ($counter{"$ch1_assign vs $ch2_assign, $bio_rep"}) {
                $counter_out = $counter{"$ch1_assign vs $ch2_assign, $bio_rep"};
            } else {
                $counter++;
                $counter_out = $counter;
                until ((length $counter_out) >= 3) {
                    $counter_out = '0'.$counter_out;
                }
                $counter{"$ch1_assign vs $ch2_assign, $bio_rep"} = $counter_out;
            }
            push @{$file_setup{"$counter_out $ch1_assign vs $ch2_assign, $bio_rep"}}, $file;
        }
    } else {
	@{$file_setup{'01 selection vs selection'}} = @files;
    }
	

    foreach (sort keys %file_setup) {
	@files = @{$file_setup{$_}};
	my ($num,$title) = split / /, $_, 2;
    
	foreach my $method (@methods) {

	    foreach my $use_ratio (@data_type) {

		my $info = '';
		if (@files < 2) {
		    $info = "Not enough files for $title comparison!\n";
		} else {
		    ($info) = &correlation_coefficient($action_num, $action, $title, $use_ratio, $method, $take_log, $rounding, @files);
		}
		
		print &format_output( # 'correlation_coefficient_',$action_num,'','__ALL__',$action,$image,$info
				      {
					  'prefix' => 'correlation_coefficient_',
					  'num' => $action_num,
					  'label' => '-',
					  'action' => $action,
					  'info' => $info
					  }
				      );
	    }
	}
    }
}


sub correlation_coefficient {
    my $action_num = shift;
    my $action = shift;
    my $title = shift;
    my $use_ratio = shift;
    my $method = shift;
    my $take_log = shift;
    my $rounding = shift;
    my @files = @_;

    if ($use_ratio eq 'ratios') {
	$use_ratio = 1;
    } else {
	$use_ratio = 0;
    }

    my %results = ();

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $file_cols = scalar @files;

    # get a header that is available for all files and channels
    my $required_occ = $file_cols;
    my $header = '';
    COL:foreach my $column (@{$item_data_column{$action}}) {

	my $occ = 0;

	my $check_column = $column;
	unless ($use_ratio) {
	    if ($column =~ /(.*)_R/) {
		$check_column = $1;
	    }
	}
	
	my @channels = ('CH1', 'CH2');
	foreach my $file (@files) {
	    if ($column =~ /_R/
		or
		$column =~ /$MERGED_RATIOS_TERM/) {
		@channels = ('__NO_LABEL__');
	    }
	    my $unavailable = 0;
	    foreach my $channel (@channels) {
		unless ($data_column_available{$file}{$channel}{$check_column}) {
		    $unavailable = 1;
		}
	    }
	    if ($unavailable) {
		next COL;
	    } else {
		$header = $check_column;
		$occ++;
	    }
	}


	if ($occ == $required_occ) {
	    last;
	} else {
	    $header = '';
	}
    }
    
    # check for specific settings:
    my $header_before = $header;
    ($header) = &get_user_settings($header,'correlation_coefficient_data_column',$force_global_opt,$action_num);

    my @channels = ();
    my $take_ratio = 0;
    my $header_no_ratio = $header;

    if ($take_log
	and
	$header =~ /_L/) {
	$take_log = 0;
    }
    
    if ($header =~ /_R/
	or
	$header =~ /$MERGED_RATIOS_TERM/) {
	# force ratios
	$use_ratio = 1;
	@channels = ('__NO_LABEL__');
    } else {
	if ($header_before eq $header) {
	    # the header was automatically determined
	    # take the ratio if desired
	    $header_no_ratio = $header;
	    
	    if ($use_ratio == 1) {

		$take_ratio = 1;
		
		$header .= '_R';
		
		# make this channel known to the rest of the program
		foreach my $file (@files) {
		    unless (defined $data_column_available{$file}{'__NO_LABEL__'}{$header}) {
			$data_column_available{$file}{'__NO_LABEL__'}{$header} = 1;
			push @{$data_column_available{$file}{'__NO_LABEL__'}{'__ORDER__'}}, $header;
			$data_done{$file}{'__NO_LABEL__'}{$header}++;
		    }
		}
		
		@channels = ('__NO_LABEL__');
	    } else {
		@channels = ('CH1', 'CH2');
	    }
	} else {
	    # a header for individual channels was explicitly selected
	    # don't use ratios
	    $use_ratio = 0;
	    @channels = ('CH1', 'CH2');
	}
    }

    
    # make this channel known to the rest of the program
    if ($take_log) {
#	foreach my $file (@files) {
#	    foreach my $ch (@channels) {
#		unless (defined $data_column_available{$file}{$ch}{$header.'_L'}) {
#		    $data_column_available{$file}{$ch}{$header.'_L'} = 1;
#		    push @{$data_column_available{$file}{$ch}{'__ORDER__'}}, $header.'_L';
#		    $data_done{$file}{$ch}{$header.'_L'}++;
#		}
#	    }
#	}
	$take_log = '_L';
    } else {
	$take_log = '';
    }
	    
    $last_column_used = $header.$take_log;
    
    # determine which flags to skip:
    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'correlation_coefficient_skip_flags'},'correlation_coefficient_skip_flags',$force_global_opt,$action_num);
    %skip_flags = &get_skip_flags(@flags);

    for (my $i = 0; $i < $#files; $i++) {
	for (my $j = $i+1; $j <= $#files; $j++) {

	    # get the file pair
	    my @compare = ($files[$i], $files[$j]);		

	    # get maximum array specifications:
	    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@compare);
	    
	    foreach my $ch (@channels) {
		
		my %skipped = ();
		my $entries = 0;
		my $valid = 0;

		# define data files:
		my $data_file1 = "corr_${title}_".(basename $files[$i])."_${ch}_${action_num}_$action_step.txt";
		my $data_file2 = "corr_${title}_".(basename $files[$j])."_${ch}_${action_num}_$action_step.txt";
		
		$data_file1 =~ s/\//_/g;
		$data_file1 =~ s/ /_/g;
		$data_file1 = "$tmp_dir/$data_file1";
		open (OUT1, ">$data_file1") 
		    or die "Can't write to $data_file1: $!";
		
		$data_file2 =~ s/\//_/g;
		$data_file2 =~ s/ /_/g;
		$data_file2 = "$tmp_dir/$data_file2";
		open (OUT2, ">$data_file2") 
		    or die "Can't write to $data_file2: $!";
		
		# go through each field
		foreach my $mrow (1..$MROW) {
		    foreach my $mcol (1..$MCOL) {
			foreach my $row (1..$ROW) {
			    foreach my $col (1..$COL) {
				
				$entries++;
				
				my @val1 = ();
				my @val2 = ();
				my $skip_flags = 0;
				my %flag1 = ();
				my %flag2 = ();
				my %names = ();
				
				    
				my $val1 = '';
				my $val2 = '';

				if ($take_ratio) {

				    # calculate ratios first
				    my ($ratio);

				    if (defined $data{$compare[0]}{'__DATA__'}{'CH1'}{$header_no_ratio}{$mrow}{$mcol}{$row}{$col}) {
					$val1 = $data{$compare[0]}{'__DATA__'}{'CH1'}{$header_no_ratio}{$mrow}{$mcol}{$row}{$col};
				    }
				    if (defined $data{$compare[0]}{'__DATA__'}{'CH2'}{$header_no_ratio}{$mrow}{$mcol}{$row}{$col}) {
					$val2 = $data{$compare[0]}{'__DATA__'}{'CH2'}{$header_no_ratio}{$mrow}{$mcol}{$row}{$col};
				    }
				    
				    if ($val1 =~ /\d/
					and
					$val2 =~ /\d/) {
					if ($header =~ /_L/) {
					    $ratio = $val2 - $val1;
					} else {
					    if ($val1 == 0) {
						$ratio = '';
					    } else {
						$ratio = $val2 / $val1;
					    }
					}
				    }
				    $data{$compare[0]}{'__DATA__'}{'__NO_LABEL__'}{$header}{$mrow}{$mcol}{$row}{$col} = $ratio;

				    # same for second file:
				    $ratio = '';

				    if (defined $data{$compare[1]}{'__DATA__'}{'CH1'}{$header_no_ratio}{$mrow}{$mcol}{$row}{$col}) {
					$val1 = $data{$compare[1]}{'__DATA__'}{'CH1'}{$header_no_ratio}{$mrow}{$mcol}{$row}{$col};
				    }
				    if (defined $data{$compare[1]}{'__DATA__'}{'CH2'}{$header_no_ratio}{$mrow}{$mcol}{$row}{$col}) {
					$val2 = $data{$compare[1]}{'__DATA__'}{'CH2'}{$header_no_ratio}{$mrow}{$mcol}{$row}{$col};
				    }
				    
				    if ($val1 =~ /\d/
					and
					$val2 =~ /\d/) {
					if ($header =~ /_L/) {
					    $ratio = $val2 - $val1;
					} else {
					    if ($val1 == 0) {
						$ratio = '';
					    } else {
						$ratio = $val2 / $val1;
					    }
					}
				    }
				    $data{$compare[1]}{'__DATA__'}{'__NO_LABEL__'}{$header}{$mrow}{$mcol}{$row}{$col} = $ratio;
				}
				    
				if (defined $data{$compare[0]}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col}) {
				    $val1 = $data{$compare[0]}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col};
				}
				if (defined $data{$compare[1]}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col}) {
				    $val2 = $data{$compare[1]}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col};
				}

				if ($take_log) {
				    # initialize new field
				    $data{$compare[0]}{'__DATA__'}{$ch}{$header.'_L'}{$mrow}{$mcol}{$row}{$col} = '';
				    $data{$compare[1]}{'__DATA__'}{$ch}{$header.'_L'}{$mrow}{$mcol}{$row}{$col} = '';
				}

				unless ($val1 =~ /\d/
					and
					$val2 =~ /\d/) {
				    # no values at this position, skip
				    $skipped{'no valid content'}++;
				    next;
				}
			    
				my @flags = ();

				# get flags for this channel
				if (defined $data{$compare[0]}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
				    push @flags, (&get_flags($data{$compare[0]}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
				}
				
				# and also flag from both channels
				unless ($ch eq '__NO_LABEL__') {
				    if (defined $data{$compare[0]}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}) {
					push @flags, (&get_flags($data{$compare[0]}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}));
				    }				    
				}

				# get flags for this channel
				if (defined $data{$compare[1]}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
				    push @flags, (&get_flags($data{$compare[1]}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
				}
				
				# and also flag from both channels
				unless ($ch eq '__NO_LABEL__') {
				    if (defined $data{$compare[1]}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}) {
					push @flags, (&get_flags($data{$compare[1]}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}));
				    }				    
				}

				# skip flagged spots
				my $skip = 0;
				if (%skip_flags) {
				    
				    foreach (@flags) {
					if (defined $skip_flags{$_}) {
					    $skip = 1;
					    last;
					}
				    }			
				}
				
				if ($skip) {
				    $skipped{'flagged'}++;
				    next;
				}

				if ($take_log) {
				    # initialize new field
				    unless ($val1 > 0
					    and
					    $val2 > 0) {
					$skipped{'no log possible'}++;
				    } else {
					$val1 = &log_base(2,$val1);
					$val2 = &log_base(2,$val2);
					$data{$compare[0]}{'__DATA__'}{$ch}{$header.'_L'}{$mrow}{$mcol}{$row}{$col} = $val1;
					$data{$compare[1]}{'__DATA__'}{$ch}{$header.'_L'}{$mrow}{$mcol}{$row}{$col} = $val2;
				    }
				}

				# print values into file
				$valid++;
				print OUT1 "$val1\n";
				print OUT2 "$val2\n";
			    }
			}
		    }
		}
		
		close OUT1;
		close OUT2;

		# get the coefficient

		# create R script
		my $R_commands_file = "$tmp_dir/R_cmd.$action.$action_num.$action_step";
		my $R_out = "$tmp_dir/R_out.$action.$action_num.$action_step";
		open (RH, ">$R_commands_file")
		    or die "Can't write to $R_commands_file: $!";
		
		print RH "data1 = read.table(\"$data_file1\", header=F, sep=\"\\t\")\n";
		print RH "data2 = read.table(\"$data_file2\", header=F, sep=\"\\t\")\n";

		if ($method eq 'Spearman') {
		    # sort vectors
		    print RH "data1 = rank(data1)\n";
		    print RH "data2 = rank(data2)\n";
		}

#		print RH "coeff = cor(data1, data2, method = \"pearson\")\n";
		print RH "coeff = cor(data1, data2)\n";
		print RH "write.table(coeff, \"$R_out\")\n";
		close RH;
 
		my $image = &get_R_output($R_commands_file, $R_out, $action_num, 'all_files', '', '', 'no_image');
    
		open (IN, $R_out)
		    or die "Can't read $R_out: $!";
		
		my $coefficient = 0;
		while (<IN>) {
		    chomp;
		    s/\"//g;
		    if (/^V?1 (.*)/) {
			$coefficient = $1;
			last;
		    }
		}
		close IN;

		if ($clean_up_files) {
		    unlink $R_commands_file;
		    unlink $R_out;
		    unlink $data_file1;
		    unlink $data_file2;
		}

		# round if necessary
		if ($coefficient =~ /\.(\d+)/) {
		    if (length($1) > $rounding) {
			$coefficient = sprintf "%.${rounding}f", $coefficient;
		    }
		}

		$results{$i}{$j}{$ch} = $coefficient;
	    }
	}
    }


    my $info = "Correlation Coefficient $title ($method, ".($use_ratio ? 'ratios' : 'intensities')."):<BR>\n<TABLE>\n";
    foreach my $i (sort { $a <=> $b } keys %results) {
	foreach my $j (sort { $a <=> $b } keys %{$results{$i}}) {
	    foreach my $ch (sort keys %{$results{$i}{$j}}) {
		my $coefficient = $results{$i}{$j}{$ch};
		my $ch_out = "($ch) ";
		if ($ch eq '__NO_LABEL__') {
		    $ch_out = "(ratio) ";
		}
		$info .= "\n<TR><TD>\n".(basename $files[$i])." - ".(basename $files[$j])." $ch_out: $coefficient\n</TD></TR>\n";
	    }
	}
    }

    $info .= "\n</TABLE>\n";

    return $info;
}


sub merge_replicates_setup {
    my $action_num = shift;
    my $action = shift;
    my $target = shift || 'all';

    if ($refine_parameters) {
	print &format_output( 
			     {
				 'prefix' => 'merge_'.$target.'_reps_',
				 'num' => $action_num,
				 'label' => '-',
				 'action' => $action,
				 'info' => 'Parameter Refinement!<BR>'
				 }
			     );
	return;
    }

    my @files = sort keys %data;

    # retrieve all channel assignments and replicate labels for all files
    my %channel_assign_combined = ();
    %channel_assign_combined = &channel_assign(@files, '__combine__');  
    my %channel_assign = ();
    %channel_assign = &channel_assign(@files);  

    my @channel_keys = sort keys %channel_assign;

    my $key_num = scalar @channel_keys;

    # return if none were found
    unless ($key_num > 0) {
	print &format_output( 
			     {
				 'prefix' => 'merge_'.$target.'_reps_',
				 'num' => $action_num,
				 'label' => '-',
				 'action' => $action,
				 'info' => "No channel assignments found!"
				 }
			     );
	return;
    }
    
    my %keys_done = ();

    # list the assignments for each file
    my %file_assign = ();
    foreach my $assign (keys %channel_assign) {
	foreach my $file (keys %{$channel_assign{$assign}}) {
	    my $channel = $channel_assign{$assign}{$file}{'CH'};
	    $file_assign{$file}{$channel}{$assign}++;
	}
    }
    
    # merging based on keys from both channels:
    foreach my $assign (sort keys %channel_assign_combined) {
	my @files = sort keys %{$channel_assign_combined{$assign}};
	    
	my ($assign1,$assign2) = split /\|/, $assign;
	# careful: assign1 and 2 will be switched now
	# due to the workings of &channel_assign
	
	if (@files == 1) {
	    # simply copy the data from normal to merged channel
	    my ($header, $out_header) = &single_file_merge($files[0],$action,$action_num);
	    
	    print &format_output( 
				  {
				      'prefix' => 'merge_'.$target.'_reps_',
				      'num' => $action_num,
				      'action' => $action,
				      'image' => '-',
				      'info' => "Only one file for '$assign1' vs '$assign2' found - copied from $header to $out_header!"
				      }
				  );
	    next;
	} elsif (@files < 1) {
	    
	    print &format_output( 
				  {
				      'prefix' => 'merge_'.$target.'_reps_',
				      'num' => $action_num,
				      'action' => $action,
				      'image' => '-',
				      'info' => "No file for '$assign1' vs '$assign2' found - no merging possible!"
				      }
				  );
	    next;
	}
	    
	# carry out action on specified data

	my @input = ();
	my @reps = ();
	
	my %tech_reps = ();
	my %bio_reps = ();

	if ($target eq 'tech') {
	    # get annotation for biological and technical replicates
	    my ($ref1,$ref2) = &replicate_assign(@files); 
	    %bio_reps = %$ref1;
	    %tech_reps = %$ref2;
	    my %input = ();

	    foreach my $bio_rep (sort keys %{$bio_reps{'REP'}}) {

		# get all files that belong to the same biological replicate
		my @tmp = @{$bio_reps{'REP'}{$bio_rep}};

		# add them to the output array
		push @input, [ @tmp ];

		# get the labelling for the technical replicates
		# and define a new label for the merged file:
		@tmp = sort keys %{$tech_reps{$bio_rep}};
		if (@tmp > 1) {
		    if (@tmp > 2) {
			push @reps, "B$bio_rep --- T$tmp[0]-$tmp[-1]";
		    } else {
			push @reps, "B$bio_rep --- T$tmp[0],$tmp[1]";
		    }
		} else {
		    push @reps, "B$bio_rep --- T$tmp[0]";
		}
	    }
	} elsif ($target eq 'bio') {
	    # not defined yet, but should eventually work just like 'all'
	    @input = ( [ @files ] );
	} else {
	    @input = ( [ @files ] );
	}

	# input contains one or more sets of files that need to be merged:
	foreach my $input (@input) {
	    my @files = @{$input};
	    
	    my %data1 = ();
	    my %data2 = ();
	    
	    foreach my $file (@files) {
		# determine which channels to use:
		if (defined $file_assign{$file}{'CH1'}{$assign1}) {
		    $data1{$file} = 'CH1';
		} else {
		    $data1{$file} = 'CH2';
		}
		if (defined $file_assign{$file}{'CH1'}{$assign2}) {
		    $data2{$file} = 'CH1';
		} else {
		    $data2{$file} = 'CH2';
		}
	    }
	    
	    my $bio_rep = '';
	    my $tech_rep = '';
	    
	    if (@reps) {
		$_ = shift @reps;
		# remove duplicate labelling with b (biological)
		# and t (technical):		s/tt/t/i;
		($bio_rep, $tech_rep) = split / --- /, $_;
	    }

	    my ($info) = &merge_replicates($action_num,$assign1,$assign2,$bio_rep,$tech_rep,\%data1,\%data2,$action,$target);
	    
	    print &format_output( 
				  {
				      'prefix' => 'merge_'.$target.'_reps_',
				      'num' => $action_num,
				      'label' => '-',
				      'action' => $action,
				      'info' => $info
				      }
				  );
	}
    }
}


sub single_file_merge {
    my $file = shift;
    my $action = shift;
    my $num = shift;
    my $header = '';
    my $out_header = '';
    foreach my $column (@{$item_data_column{$action}}) {
	foreach my $ch (keys %{$data{$file}{'__DATA__'}}) {
	    if ($data_column_available{$file}{$ch}{$column}) {
		$header = $column;
		last;
	    }
	}
	$last_column_used = $header;
	
    }
    return $header, $out_header;
}



sub merge_replicates {
    my $action_num = shift;
    my $assign1 = shift;
    my $assign2 = shift;
    my $bio_rep = shift;
    my $tech_rep = shift;
    my $data1_ref = shift;
    my $data2_ref = shift;
    my $action = shift;
    my $target = shift;

    my %data1 = %$data1_ref;
    my %data2 = %$data2_ref;

    my %work_files_skip_tmp = ();

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $data1_cols = scalar keys %data1;
    my $data2_cols = scalar keys %data2;

    # get listing of all files:
#    my @files = ();
#    my %files = ();
#    foreach (keys %data1) {
#	$files{$_}++;
#    }
#    foreach (keys %data2) {
#	$files{$_}++;
#    }
#    foreach (sort keys %files) {
#	push @files, $_;
#    }

    # get a header for each file and channels
    my %headers = ();
    my %channels = ();

    foreach my $file (keys %data1) {
	my $header = '';
	my $ch = $data1{$file};
	foreach my $column (@{$item_data_column{$action}}) {
	    if ($data_column_available{$file}{$ch}{$column}) {
		$header = $column;
		last;
	    }
	}
	# check for specific settings:
	$header = &get_user_settings($header,'merge_'.$target.'_reps_data_column',$force_global_opt,$action_num,$file);
	$headers{'HEADER1'}{$file} = $header;	

	my $out_header = $header;
	unless ($out_header =~ /_M/) {
	    $out_header .= '_M';
	}	
	$headers{'OUTHEADER1'}{$file} = $out_header;	    
    }

    foreach my $file (keys %data2) {
	my $header = '';
	my $ch = $data2{$file};
	foreach my $column (@{$item_data_column{$action}}) {
	    if ($data_column_available{$file}{$ch}{$column}) {
		$header = $column;
		last;
	    }
	}    
	# check for specific settings:
	$header = &get_user_settings($header,'merge_'.$target.'_reps_data_column',$force_global_opt,$action_num,$file);
	$headers{'HEADER2'}{$file} = $header;	    
	$last_column_used = $header;

	my $out_header = $header;
	unless ($out_header =~ /_M/) {
	    $out_header .= '_M';
	}	
	$headers{'OUTHEADER2'}{$file} = $out_header;	    
    }

    # TODO: if different headers were used, pick the most suitable!
    # 
    my $out_header = $last_column_used;
    unless ($out_header =~ /_M/) {
	$out_header .= '_M';
    }

    # columns that are added from the other files
    my @add_headers = ();
    my @interesting_headers = ('F-value','p-value', 'p-anova', 'p-paired', 'p-welch', 'p-wilcox', 'p-permute', 'p-student');
    my %interesting_headers = ();
    foreach (@interesting_headers) {
	$interesting_headers{$_}++;
    }
    foreach (@output) {
	# add other p-value columns 
	# (from pairwise t-tests between multiple groups)
	next if (defined $interesting_headers{$_});
	if (/p-/) {
	    push @interesting_headers, $_;
	    $interesting_headers{$_}++;
	}
    }
    HEAD:foreach my $column (@interesting_headers) {
	foreach my $file (keys %data) {
	    if (defined $data_column_available{$file}{'__NO_LABEL__'}{$column}) {
		push @add_headers, $column;
		next HEAD;
	    }
	}
    }
        
    my $dup_usage = &get_user_settings($var_def_default{'merge_'.$target.'_reps_dups'},'merge_'.$target.'_reps_dups',$force_global_opt,$action_num);
    if ($dup_usage eq 'yes') {
	$dup_usage = 1;
    } else {
	$dup_usage = 0;
    }
    
    # determine which flags to skip:
    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'merge_'.$target.'_reps_skip_flags'},'merge_'.$target.'_reps_skip_flags',$force_global_opt,$action_num);
    %skip_flags = &get_skip_flags(@flags);

    my $skip_flags_all = 0;
    my %skipped = ();
    my $no_flag = 0;
    my $entries = 0;
    my $flag_threshold = 0;

    # define data files:
    my $out_file = "$MERGED_FILE_NAME${assign1}_${assign2}";
    if ($bio_rep ne 'B') {
	$out_file .= "_$bio_rep";
    }
    if ($tech_rep ne 'T') {
	$out_file .= "_$tech_rep";
    }
    $out_file .= "_${action_step}_$action_num.tab";
    $out_file =~ s/\//_/g;
    $out_file =~ s/ /_/g;
    $out_file =~ s/__/_/g;
    $out_file = "$tmp_dir/$out_file";
    open (OUT, ">$out_file") 
	or die "Can't write to $out_file: $!";

    # get the channel labels
    my %label1 = ();
    my %label2 = ();
    foreach my $file (keys %data1) {
	$label1{$data1{$file}}++;
    }
    my $label1 = join ',', sort keys %label1;
    foreach my $file (keys %data2) {
	$label2{$data2{$file}}++;
    }
    my $label2 = join ',', sort keys %label2;

    # define some more headers:
    my $out_header_list = $out_header . '_list';
    my $out_header_orig = $FG_SIGNAL_TERM . '_M_orig';
    if ($out_header eq $FG_SIGNAL_TERM.'_M') {
	# no need to put the original values in 
	# because they are listed under '_list' already
	$out_header_orig = $out_header_list;
    }

    my $sd_header = $out_header.'_SD';
    my $i_header = $out_header .'_i';

    my $merge_ratio_header = $MERGED_RATIOS_TERM;
    my $merge_ratio_input = $out_header.'_R';
    
    # write header:
    my $raw_header = '';
    my $add_headers = join "\t", @add_headers;
    $add_headers .= "\t";
    $merge_ratio_header .= '_L';
    $raw_header = "${FG_SIGNAL_TERM}_M";
    print OUT "MROW\tMCOL\tROW\tCOL\t$SPOT_ID_TERM\t$label1 $raw_header\t$label2 $raw_header\t$label1 $out_header\t$label2 $out_header\t$label1 $sd_header\t$label2 $sd_header\t$label1 $i_header\t$label2 $i_header\t$merge_ratio_header$add_headers$label1\t$FLAG_CHANNEL_TERM\t$label2 $FLAG_CHANNEL_TERM\n";

    # write out data:
    my @files1 = (sort keys %data1);
    my @files2 = (sort keys %data2);
    my @files = (@files1, @files2);

    my $average_method = &get_user_settings($var_def_default{'merge_'.$target.'_reps_method'},'merge_'.$target.'_reps_method',$force_global_opt,$action_num);
    
    if ($average_method eq 'weighted mean') {
	# if no spreads are available, we have to change this to 'median'
	foreach my $file (@files) {
	    unless (defined $spread_done{$file}) {
		$average_method = 'median';
		last;
	    }
	}
    }

    my $merge_continuation = &get_user_settings($var_def_default{'merge_'.$target.'_reps_continuation'},'merge_'.$target.'_reps_continuation',$force_global_opt,$action_num);
    if ($merge_continuation eq 'yes') {
	$merge_continuation = 1;
	foreach (@files) {
	    $work_files_skip{$_}++;
	    $work_files_skip_tmp{$_}++;
	}
    } else {
	$merge_continuation = 0;
    }
    
    # check if the ratio column for the input files are available already
    # CHANGED: ALWAYS calculate ratios
    # TODO: make sure that whenever new values are set the individual channels are up to date!
    my $initialize_ratio_input = 1;
#    foreach my $file (@files) {
#	unless (defined $data_column_available{$file}{'__NO_LABEL__'}{$merge_ratio_input}) {
#	    # if one or more are missing, initialize and calculate them all
#	    $initialize_ratio_input = 1;
#	}
#    }

    my $input_size1 = scalar @files1;
    my $input_size2 = scalar @files2;
    if ($dup_usage) {
	$input_size1 *= 2;
	$input_size2 *= 2;
    }

    my %data_out = ();
    my %coord = ();

    # get maximum array specifications:
    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@files);

    # go through each field
    foreach my $mrow (1..$MROW) {
	foreach my $row (1..$ROW) {
	    my $all_col = 0;
	    foreach my $mcol (1..$MCOL) {
	      COL:foreach my $col (1..$COL) {

		  $all_col++;
		  
		  $entries++;

		  # get current spot label:
		  my $spot = '';
		  if (defined $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col}) {
		      $spot = $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col};
		      
		      # special settings for Arrayvision:
		      # delete numerator placed in brackets behind each gene name
		      $spot =~ s/ \(\d+\)$// if (defined $file_specs{$files[0]}{'FORMAT'} 
						 and
						 $file_specs{$files[0]}{'FORMAT'} eq 'ArrayVision');
		  }
		  unless ($spot =~ /\w/) {
		      # TODO: uncomment warning
		      warn "No label for spot at $mrow,$mcol,$row,$col!";
		      next;
		  }
		  # write spot name to new file:
		  $data{$out_file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col} = $spot;

		  my $key = "$mrow|$mcol|$row|$col";
		  if ($dup_usage) {
		      $key = $spot;
		  }
		  push @{$coord{$key}}, "$mrow|$mcol|$row|$col";		  

		  my %done = ();
		  my @val1 = ();
		  my @val2 = ();
		  my @val1_orig = ();
		  my @val2_orig = ();
		  my @spread1 = ();
		  my @spread2 = ();
		  my @flag1 = ();
		  my @flag2 = ();
		  for (my $i = 0; $i <= $#files; $i++) {
		      my $file = $files[$i];
		      my $ch = '';
		      if ($i < $data1_cols) {
			  $ch = $data1{$file};
		      } else {
			  $ch = $data2{$file};
		      }	
	      
		      if (defined $done{$file.$ch}) {
			  next;
		      } else {
			  $done{$file.$ch}++;
		      }


		      # check the data is flagged
		      my @flags = ();
		      # get flags for this channel as well
		      if (defined $data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			  push @flags, (&get_flags($data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
		      }
		      
		      # and also flag from both channels
		      if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			  push @flags, (&get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
		      }
		      
		      # skip flagged spots
		      my $skip = 0;
		      my $flag = 0;
		      my %flags = ();		      
		      if (%skip_flags) {
			  
			  foreach (@flags) {
			      $flags{$_}++;
			      if (defined $skip_flags{$_}) {
				  $skip = 1;
			      }
			  }			
		      }
		      
		      foreach (keys %flags) {
			  next if ($_ eq '');
			  $flag += $_;
		      }

		      my $type = 'HEADER2';
		      if ($i < $data1_cols) {
			  $type = 'HEADER1';
		      }

		      # collect data
		      my $val = '';
		      my $spread = '';
		      my $val_orig = '';
		      if (defined $data{$file}{'__DATA__'}{$ch}{$headers{$type}{$file}}{$mrow}{$mcol}{$row}{$col}) {
			  $val = $data{$file}{'__DATA__'}{$ch}{$headers{$type}{$file}}{$mrow}{$mcol}{$row}{$col};
			  $val_orig = $val;
		      }

		      # if a spread for the ratios exist, take that first
		      if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{'SPREAD'}{$mrow}{$mcol}{$row}{$col}) {
			  $spread = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{'SPREAD'}{$mrow}{$mcol}{$row}{$col};
			  # otherwise look for a spread for each individual channel
		      } elsif (defined $data{$file}{'__DATA__'}{$ch}{'SPREAD'}{$mrow}{$mcol}{$row}{$col}) {
			  $spread = $data{$file}{'__DATA__'}{$ch}{'SPREAD'}{$mrow}{$mcol}{$row}{$col};
		      }

		      if ($skip) {
			  $val = '';
		      }

		      if ($i < $data1_cols) {
			  push @flag1, $flag;
			  push @spread1, $spread;
			  push @val1, $val;
			  push @val1_orig, $val_orig;
		      } else {
			  push @flag2, $flag;
			  push @spread2, $spread;
			  push @val2, $val;
			  push @val2_orig, $val_orig;
		      }

		      # initialize output
		      foreach my $header ($out_header,$sd_header,$i_header,$FLAG_CHANNEL_TERM) {
			  $data{$out_file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col} = '';
			  if ($header eq $FLAG_CHANNEL_TERM) {
			      $data{$out_file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col} = 0;
			  }
			  $data{$out_file}{'__DATA__'}{'__NO_LABEL__'}{$merge_ratio_header}{$mrow}{$mcol}{$row}{$col} = '';
			  if ($initialize_ratio_input) {
			      $data{$out_file}{'__DATA__'}{'__NO_LABEL__'}{$merge_ratio_input}{$mrow}{$mcol}{$row}{$col} = '';
			  }
		      }
		  }

		  # store data
		  push @{$data_out{$key}{'CH1'}{'VAL'}}, @val1;
		  push @{$data_out{$key}{'CH1'}{'VAL_ORIG'}}, @val1_orig;
		  push @{$data_out{$key}{'CH1'}{'SPREAD'}}, @spread1;
		  push @{$data_out{$key}{'CH1'}{'FLAG'}}, @flag1;
		  push @{$data_out{$key}{'CH2'}{'VAL'}}, @val2;
		  push @{$data_out{$key}{'CH2'}{'VAL_ORIG'}}, @val2_orig;
		  push @{$data_out{$key}{'CH2'}{'SPREAD'}}, @spread2;
		  push @{$data_out{$key}{'CH2'}{'FLAG'}}, @flag2;
	      }
	    }
	}
    }

    # all data to be merged is now stored in data_out 

    # go through each element
    foreach my $key (keys %data_out) {
	my @val1 = @{$data_out{$key}{'CH1'}{'VAL'}};
	my @val1_orig = @{$data_out{$key}{'CH1'}{'VAL_ORIG'}};
	my @spread1 = @{$data_out{$key}{'CH1'}{'SPREAD'}};
	my @flag1 = @{$data_out{$key}{'CH1'}{'FLAG'}};

	my @val2 = @{$data_out{$key}{'CH2'}{'VAL'}};
	my @val2_orig = @{$data_out{$key}{'CH2'}{'VAL_ORIG'}};
	my @spread2 = @{$data_out{$key}{'CH2'}{'SPREAD'}};
	my @flag2 = @{$data_out{$key}{'CH2'}{'FLAG'}};

	# merge the two spreads
	# (if they are based on ratios, then they are the same already
	# if not, calculate log-ratios and new spread
	my @all_spread = ();

	while (@spread1) {
	    my $spread1 = shift @spread1;
	    my $spread2 = shift @spread2;
	    my $spread = '';

	    if ($spread1 eq ''
		or
		$spread2 eq '') {
		push @all_spread, '';
		next;
	    }
	    
	    my ($spread1_num,$type1,@spread1_nums) = split /\|/, $spread1;
	    my ($spread2_num,$type2,@spread2_nums) = split /\|/, $spread2;

	    my @new_spread = ();

	    if ($type1 =~ /intensities/
		and
		$type2 =~ /intensities/) {
		# calculate spread from ratios
		while (@spread1_nums) {
		    my $spread1 = shift @spread1_nums;
		    my $spread2 = shift @spread2_nums;
		    my $new_spread = '';
		    if ($spread1 =~ /\d/
			and
			$spread2 =~ /\d/) {
			unless ($type1 =~ /log/) {
			    if ($spread1 > 0) {
				$spread1 = &log_base(2,$spread1);
			    } else {
				$spread1 = '';
			    }
			}
			unless ($type2 =~ /log/) {
			    if ($spread2 > 0) {
				$spread2 = &log_base(2,$spread2);
			    } else {
				$spread2 = '';
			    }
			}
			if ($spread1 =~ /\d/
			    and
			    $spread2 =~ /\d/) {
			    $new_spread = $spread2 - $spread1;		    
			    push @new_spread, $new_spread;		
			}	
		    }
		}
	    } elsif ($type1 =~ /ratio/
		     and
		     $type2 =~ /ratio/) {
		if ($type1 =~ /log/) {
		    @new_spread = @spread1_nums;
		} else {
		    foreach my $new_spread (@spread1_nums) {
			if ($new_spread > 0) {
			    $new_spread = &log_base(2,$new_spread);
			}
			push @new_spread, $new_spread;
		    }
		}
	    } else {
		die "Mix of ratios and intensities in spread!";
	    }
	    	    
	    my $new_spread_num = 0;
	    if ($spread1_num != 0
		and
		$spread2_num != 0) {
		# if both previous spreads are different from zero, 
		# calculate new spread
		my @new_spread2 = sort { $a <=> $b } @new_spread;
#		$new_spread_num = abs($new_spread2[-1] - $new_spread2[0]);
		$new_spread_num = &variance(@new_spread2);
	    }
	    push @all_spread, "$new_spread_num|".(join '|', @new_spread);
	}
    
	my ($mrow, $mcol, $row, $col) = split /\|/, @{$coord{$key}}[0];
	my $name = $data{$out_file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col};

	# calculate the ratios:
	my @all_ratios = ();
	my @ratios = ();
	my $stdev_r = '-';
	my $average_r = '-';
	my @val1_used = ();
	my @val2_used = ();
	if ($initialize_ratio_input) {
	    for (my $i = 0; $i <= $#val1; $i++) {

		my $val1 = $val1[$i];
		my $val2 = $val2[$i];

		my $ratio = '';

		if ($val1 =~ /\d/
		    and
		    $val2 =~ /\d/) {

		    # log-transfer if necessary:
		    my $header1 = '';
		    my $header2 = '';
		    my $j = int($i/2) % @files1;
		    unless ($headers{'HEADER1'}{$files1[$j]} =~ /_L/) {
			if ($val1 <= 0) {
			    $val1 = '';
			} else {
			    $val1 = &log_base(2,$val1);
			}
		    }
		    
		    unless ($headers{'HEADER2'}{$files2[$j]} =~ /_L/) {
			if ($val2 <= 0) {
			    $val2 = '';
			} else {
			    $val2 = &log_base(2,$val2);
			}
		    }

		    if ($val1 =~ /\d/
			and
			$val2 =~ /\d/) {

			$ratio = $val2 - $val1;
			
			push @val1_used, $val1[$i];
			push @val2_used, $val2[$i];
		    }
		}
		# collect all ratios
		push @all_ratios, $ratio;
	    }
	} else {
	    #TODO: retrieve pre-calculated ratios!
	}

	unless (@all_ratios == @all_spread) {
	    die "# ERROR: discrepancy between ratios (".(join ',', @all_ratios).") and spread (".(join ',', @all_spread).") for key $key!";
	}

	my @spread = ();
	my @flag1a = ();
	my @flag2a = ();
	foreach my $ratio (@all_ratios) {
	    my $spread = shift @all_spread;
	    my $flag1 = shift @flag1;
	    my $flag2 = shift @flag2;
	    # sort out those that are not numbers:
	    if ($ratio =~ /\d/) {
		push @ratios, $ratio;
		push @spread, $spread;
		push @flag1a, $flag1;
		push @flag2a, $flag2;
	    }
	}

	@flag1 = @flag1a;
	@flag2 = @flag2a;

	# calculate median and SD:
	if (@ratios > 1) {
	    $stdev_r = &stdev(@ratios);
	}

	my $weights = '';
	if (@ratios) {
	    if ($average_method eq 'median') {
		$average_r = &median(@ratios);
	    } elsif ($average_method eq 'weighted mean') {
		($average_r, $weights) = &weighted_mean(\@ratios,\@spread);
	    } else {
		$average_r = &mean(@ratios);
	    }
	}
	
	my $i_r = scalar(@ratios);
	
	$data{$out_file}{'__DATA__'}{'__NO_LABEL__'}{$merge_ratio_header}{$mrow}{$mcol}{$row}{$col} = $average_r;
	$data{$out_file}{'__DATA__'}{'__NO_LABEL__'}{$merge_ratio_header.'_list'}{$mrow}{$mcol}{$row}{$col} = join '; ', @ratios;
	$data{$out_file}{'__DATA__'}{'__NO_LABEL__'}{$merge_ratio_header.'_SD'}{$mrow}{$mcol}{$row}{$col} = $stdev_r;
	$data{$out_file}{'__DATA__'}{'__NO_LABEL__'}{$merge_ratio_header.'_i'}{$mrow}{$mcol}{$row}{$col} = $i_r;
	if ($average_method eq 'weighted mean') {
	    $data{$out_file}{'__DATA__'}{'__NO_LABEL__'}{$WEIGHTS_TERM}{$mrow}{$mcol}{$row}{$col} = $weights;
	}
	

	# separate merged ratios into merged intensities
	# merge the first of the individual values and calculate the second from that and the merged ratio
	my $average1 = '';
	my $average2 = '';

	my $average1_spread = '';
	my $average2_spread = '';

	my @val_num1 = ();
	my @val_num2 = ();

	if ($average_r =~ /\d/) {
	    my @spread = @{$data_out{$key}{'CH1'}{'SPREAD'}};
	    my @spread_num = ();
	    foreach (@val1) {
		my $spread = shift @spread;
		unless (/\d/) {
		    next;
		}
		push @val_num1, $_;
		push @spread_num, $spread;
	    }
	    if ($average_method eq 'median') {
		$average1 = &median(@val_num1);
	    } elsif ($average_method eq 'weighted mean') {
		# warning: if the spread contains a zero and it
		# has been calculated from (log-)ratios, then 
		# the average1 will be the median of the ratios!
		my $done = 0;
		foreach (@spread_num) {
		    if (/^0\|/
			and
			/ratio/) {
			# calculate the median from the intensities
			# we can't use spread because one of them us null
			# normally, &weighted_mean takes care of this, but
			# it can't use the spread_vals because they are ratios
			# and not intensities
			$average1 = &median(@val_num1);
			$done = 1;
			last;
		    }
		}
		unless ($done) {
		    my $weights = '';
		    ($average1, $weights) = &weighted_mean(\@val_num1,\@spread_num);
		}
	    } else {
		$average1 = &mean(@val_num1);
	    }

	    # TODO: What if there are both raw and log-transferred intensities in the SPREAD?
	    if ($average_method eq 'weighted mean') {
		unless (grep /log/, @{$data_out{$key}{'CH1'}{'SPREAD'}}) {
		    $average1 = &log_base(2, $average1);
		}
	    } 		

	    $average2 = $average_r + $average1;

	    foreach (@val2) {
		unless (/\d/) {
		    next;
		}
		push @val_num2, $_;
	    }

	    my $spread_type = 'log-intensities';
	    if (@val_num1 == 1) {
		$average1_spread = "0|$spread_type|$val_num1[0]";
	    } else {
		my @tmp = sort { $a <=> $b } @val_num1;
#		my $spread_val = abs($tmp[-1] - $tmp[0]);
		my $spread_val = &variance(@tmp);
		$average1_spread = "$spread_val|$spread_type|".(join '|', @val_num1);
	    }
	    if (@val_num2 == 1) {
		$average2_spread = "0|$spread_type|$val_num2[0]";
	    } else {
		my @tmp = sort { $a <=> $b } @val_num2;
#		my $spread_val = abs($tmp[-1] - $tmp[0]);
		my $spread_val = &variance(@tmp);
		$average2_spread = "$spread_val|$spread_type|".(join '|', @val_num2);
	    }
	}
	    
	# calculate median and SD:
	my $stdev1 = '-';
	if (@val1_used > 1) {
	    $stdev1 = &stdev(@val1_used);
	}

	my $stdev2 = '-';
	if (@val2_used > 1) {
	    $stdev2 = &stdev(@val2_used);
	}
    
	my $i1 = @val1_used;
	my $i2 = @val2_used;
	
	# TODO: don't take over negative flags unless the data is invalid
	my $flag1 = 0;
	if ($average1 eq '-'
	    or
	    $average1 eq '') {
	    # no value found for this spot
	    $flag1 = $flag_values{'undefined'};
	} else {
	    # check for flags
	    my @flags = ();
	    my %flags = ();
	    foreach my $flag_num (@flag1) {
		# extract all individual flags from merged value
		push @flags, (&get_flags($flag_num));
	    }
	    foreach (@flags) {
		# remove duplicates
		$flags{$_}++;
	    }
	    foreach (keys %flags) {
		# sum up again
		$flag1 += $_;
	    }
	}
	
	# same for flag of channel2:
	my $flag2 = 0;
	if ($average2 eq '-'
	    or
	    $average2 eq '') {
	    # no value found for this spot
	    $flag2 = $flag_values{'undefined'};
	} else {
	    # check for flags
	    # check for flags
	    my @flags = ();
	    my %flags = ();
	    foreach my $flag_num (@flag2) {
		# extract all individual flags from merged value
		push @flags, (&get_flags($flag_num));
	    }
	    foreach (@flags) {
		# remove duplicates
		$flags{$_}++;
	    }
	    foreach (keys %flags) {
		# sum up again
		$flag2 += $_;
	    }
	} 
	
#	my $name = join ',', sort keys %names;
	
	# store values
	$data{$out_file}{'__DATA__'}{$label1}{$out_header}{$mrow}{$mcol}{$row}{$col} = $average1;
	$data{$out_file}{'__DATA__'}{$label1}{'SPREAD'}{$mrow}{$mcol}{$row}{$col} = $average1_spread;
	$data{$out_file}{'__DATA__'}{$label1}{$out_header_list}{$mrow}{$mcol}{$row}{$col} = join '; ', @val1_used;
	$data{$out_file}{'__DATA__'}{$label1}{$out_header_orig}{$mrow}{$mcol}{$row}{$col} = join '; ', @val1_orig if (@val1_orig);
	$data{$out_file}{'__DATA__'}{$label1}{$sd_header}{$mrow}{$mcol}{$row}{$col} = $stdev1;
	$data{$out_file}{'__DATA__'}{$label1}{$i_header}{$mrow}{$mcol}{$row}{$col} = $i1;
	$data{$out_file}{'__DATA__'}{$label1}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} = $flag1;
	$data{$out_file}{'__DATA__'}{$label2}{$out_header}{$mrow}{$mcol}{$row}{$col} = $average2;
	$data{$out_file}{'__DATA__'}{$label2}{'SPREAD'}{$mrow}{$mcol}{$row}{$col} = $average2_spread;
	$data{$out_file}{'__DATA__'}{$label2}{$out_header_list}{$mrow}{$mcol}{$row}{$col} = join '; ', @val2_used;
	$data{$out_file}{'__DATA__'}{$label2}{$out_header_orig}{$mrow}{$mcol}{$row}{$col} = join '; ', @val2_orig if (@val2_orig);
	$data{$out_file}{'__DATA__'}{$label2}{$sd_header}{$mrow}{$mcol}{$row}{$col} = $stdev2;
	$data{$out_file}{'__DATA__'}{$label2}{$i_header}{$mrow}{$mcol}{$row}{$col} = $i2;
	$data{$out_file}{'__DATA__'}{$label2}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} = $flag2;
#	$data{$out_file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col} = $name;
	$no_flag++;
	
	$spread_done{$out_file}{$label1}++;
	$spread_done{$out_file}{$label2}++;

	my @add_values = ();
	foreach my $add_header (@add_headers) {
	    # these values are the same for each file - pick the first best
	    my $add = '';
	    if (defined $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$add_header}{$mrow}{$mcol}{$row}{$col}) {
		$add = $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$add_header}{$mrow}{$mcol}{$row}{$col};
	    }
	    $data{$out_file}{'__DATA__'}{'__NO_LABEL__'}{$add_header}{$mrow}{$mcol}{$row}{$col} = $add;
	    push @add_values, $add;
	}
	
	my $add_values = join "\t", @add_values;
	$add_values .= "\t";
	my $raw1 = '';
	$raw1 = 2**$average1 if ($average1 =~ /\d/);
	my $raw2 = '';
	$raw2 = 2**$average2 if ($average2 =~ /\d/);
	$data{$out_file}{'__DATA__'}{$label1}{$raw_header}{$mrow}{$mcol}{$row}{$col} = $raw1;
	$data{$out_file}{'__DATA__'}{$label2}{$raw_header}{$mrow}{$mcol}{$row}{$col} = $raw2;
	print OUT "$mrow\t$mcol\t$row\t$col\t$name\t$raw1\t$raw2\t$average1\t$average2\t$stdev1\t$stdev2\t$i1\t$i2$add_values$flag1\t$flag2\n";
    }

    close OUT;
    
    # write file_specs
    $file_specs{$out_file}{'MROW'} = $MROW;
    $file_specs{$out_file}{'MCOL'} = $MCOL;
    $file_specs{$out_file}{'ROW'} = $ROW;
    $file_specs{$out_file}{'COL'} = $COL;
    $file_specs{$out_file}{'CH1_ANNOTATE'} = $assign1;
    $file_specs{$out_file}{'CH2_ANNOTATE'} = $assign2;
    $file_specs{$out_file}{'CH1_LABEL'} = $label1;
    $file_specs{$out_file}{'CH2_LABEL'} = $label2;
    $file_specs{$out_file}{'CH1FG'} = "$label1 $raw_header";
    $file_specs{$out_file}{'CH2FG'} = "$label2 $raw_header";

    if ($bio_rep ne '') {
	$file_specs{$out_file}{'BIO_REP'} = $bio_rep;
    }

    foreach my $header ($out_header,$out_header_list,$out_header_orig,$sd_header,$i_header,$FLAG_CHANNEL_TERM) {
	foreach my $label ($label1,$label2) {
	    unless (defined $data_column_available{$out_file}{$label}{$header}) {
		$data_column_available{$out_file}{$label}{$header} = 1;
		push @{$data_column_available{$out_file}{$label}{'__ORDER__'}}, $header;
		$data_done{$out_file}{$label}{$header}++;
		if ($header =~ /_L/) {
		    # add information about raw data channels
		    $data_column_available{$out_file}{$label}{$raw_header} = 1;
		    push @{$data_column_available{$out_file}{$label}{'__ORDER__'}}, $raw_header;
		    $data_done{$out_file}{$label}{$raw_header}++;
		}
	    }
	}
    }
    foreach my $header ($merge_ratio_header,$merge_ratio_header.'_list',$merge_ratio_header.'_SD',$merge_ratio_header.'_i',$merge_ratio_input,@add_headers) {
	unless (defined $data_column_available{$out_file}{'__NO_LABEL__'}{$header}) {
	    $data_column_available{$out_file}{'__NO_LABEL__'}{$header} = 1;
	    push @{$data_column_available{$out_file}{'__NO_LABEL__'}{'__ORDER__'}}, $header;
	    $data_done{$out_file}{'__NO_LABEL__'}{$header}++;
	}
    }

    if ($average_method eq 'weighted mean') {
	unless (defined $data_column_available{$out_file}{'__NO_LABEL__'}{$WEIGHTS_TERM}) {
	    $data_column_available{$out_file}{'__NO_LABEL__'}{$WEIGHTS_TERM} = 1;
	    push @{$data_column_available{$out_file}{'__NO_LABEL__'}{'__ORDER__'}}, $WEIGHTS_TERM;
	    $data_done{$out_file}{'__NO_LABEL__'}{$WEIGHTS_TERM}++;
	}
    }
    
    my $info = "Merged ratios were calculated from <B>$merge_ratio_input</B><BR>";
    $info .= "Merger (<B>$average_method</B>) for channel annotations <B>$assign1</B> vs <B>$assign2</B>";
    if ($target eq 'tech') {
	$info .= ", sample <B>$bio_rep</B>";
    }

    $info .= ":<BR><TABLE><TR><TD><TABLE><TR><TH>new CH1</TH></TR>";
    foreach my $file (@files1) {
	my $ch = $data1{$file};

	# get channel label name
	my $ch_out = '';
	if ($ch eq 'CH1') {
	    if (defined $file_specs{$file}{'CH1_LABEL'}) {
		$ch_out = " ($file_specs{$file}{'CH1_LABEL'})";
	    }
	} elsif ($ch eq 'CH2') {
	    if (defined $file_specs{$file}{'CH2_LABEL'}) {
		$ch_out = " ($file_specs{$file}{'CH2_LABEL'})";
	    }
	}

	$info .= "<TR><TD>".(basename $file).", $ch$ch_out</TD></TR>\n";
    }
    $info .= "</TABLE>\n</TD><TD>&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</TD><TD>\n<TABLE><TR><TH>new CH2</TH></TR>\n";
    foreach my $file (@files2) {
	my $ch = $data2{$file};

	# get channel label name
	my $ch_out = '';
	if ($ch eq 'CH1') {
	    if (defined $file_specs{$file}{'CH1_LABEL'}) {
		$ch_out = " ($file_specs{$file}{'CH1_LABEL'})";
	    }
	} elsif ($ch eq 'CH2') {
	    if (defined $file_specs{$file}{'CH2_LABEL'}) {
		$ch_out = " ($file_specs{$file}{'CH2_LABEL'})";
	    }
	}

	$info .= "<TR><TD>".(basename $file).", $ch$ch_out</TD></TR>\n";
    }
    $info .= "</TABLE>\n</TD></TR>\n</TABLE>\n<BR>";
    
    unless ($no_flag) {
	# TODO: check why we have no value in $no_flag
#	$info .= "<B>No output generated!</B>\n";
    } else {
	my $base = basename $out_file;
	$info .= "Output file: <A HREF=\"$base\">$base</A> (".(&file_size($out_file)).")<BR>\n";
    }

    if ($merge_continuation) {
	$info .= "<BR><B>Note:</B>The following files are now removed from subsequent analysis steps:<BR><UL>";
	foreach (sort keys %work_files_skip_tmp) {
	    $info .= "<LI>".(basename $_)."</LI>\n";
	}
	$info .= "</UL>\n";
	foreach ((keys %data1), (keys %data2)) {
	    # delete files from data hash
	    if (defined $data{$_}) {
		delete $data{$_};
	    }
	}
    }
    return $info;
} 


sub permutation_p_value_setup {
    my $action_num = shift;
    my $action = shift;

    if ($refine_parameters) {
	print &format_output( # 'permutation_p_value_',$action_num,'','__ALL__',$action,'','Parameter Refinement!<BR>'
			     {
				 'prefix' => 'permutation_p_value_',
				 'num' => $action_num,
				 'label' => '__ALL__',
				 'action' => $action,
				 'info' => 'Parameter Refinement!<BR>'
				 }
			     );
	return;
    }

    my @files = sort keys %data;

    my %channel_assign_combined = &channel_assign(@files, '__combine__');  
    my %channel_assign = &channel_assign(@files);  
    
    my @channel_keys = sort keys %channel_assign;

    my $key_num = scalar @channel_keys;

    unless ($key_num > 0) {
	print &format_output( # 'permutation_p_value_',$action_num,'','',$action,'-',"No channel assignments found!"
			     {
				 'prefix' => 'permutation_p_value_',
				 'num' => $action_num,
				 'action' => $action,
				 'info' => "No channel assignments found!"
				 }
			     );
	next;
    }
    
    my %keys_done = ();

    # list the assignments for each file
    my %file_assign = ();
    foreach my $assign (keys %channel_assign) {
	foreach my $file (keys %{$channel_assign{$assign}}) {
	    my $channel = $channel_assign{$assign}{$file}{'CH'};
	    $file_assign{$file}{$channel}{$assign}++;
	}
    }

    my $combined_channel = 1;
    if ($combined_channel) {
	# new merging based on keys from both channels:
	foreach my $assign (sort keys %channel_assign_combined) {
	    my %data1 = ();
	    my %data2 = ();
	    my @files = sort keys %{$channel_assign_combined{$assign}};

	    my ($assign1,$assign2) = split /\|/, $assign;
	    
	    unless (@files > 1) {
		print &format_output( # 'permutation_p_value_',$action_num,'','__ALL__',$action,$image,$info
				      {
					  'prefix' => 'permutation_p_value_',
					  'num' => $action_num,
					  'label' => '__ALL__',
					  'action' => $action,
					  'info' => "Only one file for '$assign1' vs '$assign2' found - no testing possible!"
					  }
				      );
		next;
	    }
		
	    foreach my $file (@files) {
		if (defined $file_assign{$file}{'CH1'}{$assign1}) {
		    $data1{$file} = 'CH1';
		} else {
		    $data1{$file} = 'CH2';
		}
		if (defined $file_assign{$file}{'CH1'}{$assign2}) {
		    $data2{$file} = 'CH1';
		} else {
		    $data2{$file} = 'CH2';
		}
	    }
	
	    # carry out action on specified data
#	warn "plot $file1, $ch1 against $file2, $ch2";
	    
	    my ($info) = &permutation_p_value($action_num,$assign1,$assign2,\%data1,\%data2,$action);
	    
	    print &format_output( # 'permutation_p_value_',$action_num,'','__ALL__',$action,$image,$info
				  {
				      'prefix' => 'permutation_p_value_',
				      'num' => $action_num,
				      'label' => '__ALL__',
				      'action' => $action,
				      'info' => $info
				      }
				  );
	}

    } else {

	# do a merge based on individual channels:

	while (my $assign1 = shift @channel_keys) {
	    
	    next if ($keys_done{$assign1});
	    $keys_done{$assign1}++;
	    
	    my @files = sort keys %{$channel_assign{$assign1}};
	    unless (@files) {
		print &format_output( # 'permutation_p_value_',$action_num,'','',$action,'-',"No files for channel assignment '$assign1' found!"
				      {
					  'prefix' => 'permutation_p_value_',
					  'num' => $action_num,
					  'action' => $action,
					  'image' => '-',
					  'info' => "No files for channel assignment '$assign1' found!"
					  }
				      );
		next;
	    }
	    
	    # determine channels for each file
	    my %channels = ();
	    foreach my $file (@files) {
		my $channel = $channel_assign{$assign1}{$file}{'CH'};
		$channels{$file} = $channel;
	    }
	    
	    my $assign2 = '';
	    
	    my @left_keys = ();
	    foreach my $assign (@channel_keys) {
		unless (defined $keys_done{$assign}) {
		    push @left_keys, $assign;
		}
	    }
	    
	    unless (@left_keys) {
		print &format_output( # 'permutation_p_value_',$action_num,'','',$action,'-',"No matching channel assignment for '$assign1' found!"
				      {
					  'prefix' => 'permutation_p_value_',
					  'num' => $action_num,
					  'action' => $action,
					  'image' => '-',
					  'info' => "No matching channel assignment for '$assign1' found!"
				 }
				      );
		next;
	    }
	    
	    # we might only have two keys in which case the combination is obvious:
	    if (@left_keys == 1) {
		$assign2 = shift @left_keys;
		$keys_done{$assign2}++;
	    }
	    
	    # otherwise, go through all the files for the first key 
	    # and check which other keys are predominant for them...
	    my %second_key = ();
	    foreach my $file (@files) {
		foreach my $channel (keys %{$file_assign{$file}}) {
		    next if ($channel eq $channels{$file});
		    foreach ( sort { $file_assign{$file}{$channel}{$b} <=> $file_assign{$file}{$channel}{$a} } keys %{$file_assign{$file}{$channel}} ) {
			$second_key{$_}++;
			# only get the assignment with the highest occurences
			last;
		    }
		}
	    }
	    
	    foreach (sort { $second_key{$b} <=> $second_key{$a} } keys %second_key) {
		$assign2 = $_;
	    $keys_done{$assign2}++;
		last;
	    }
	    
	    # now that we have assign1 and assign2 let's assemble arrays with info
	    # regarding which channels to use for control and treatment:
	    my %data1 = ();
	    my %data2 = ();
	    foreach my $file (sort keys %{$channel_assign{$assign1}}) {
		my $channel = $channel_assign{$assign1}{$file}{'CH'};
		$data1{$file} = $channel;
	    }
	    foreach my $file (sort keys %{$channel_assign{$assign2}}) {
		my $channel = $channel_assign{$assign2}{$file}{'CH'};
		$data2{$file} = $channel;
	    }
	    
	    # carry out action on specified data
#	warn "plot $file1, $ch1 against $file2, $ch2";
	    
	    my ($info) = &permutation_p_value($action_num,$assign1,$assign2,\%data1,\%data2,$action);
	    
	    print &format_output( # 'permutation_p_value_',$action_num,'','__ALL__',$action,$image,$info
				  {
				      'prefix' => 'permutation_p_value_',
				      'num' => $action_num,
				      'label' => '__ALL__',
				      'action' => $action,
				      'info' => $info
				      }
				  );
	}
    }
}


sub permutation_p_value {
    my $action_num = shift;
    my $assign1 = shift;
    my $assign2 = shift;
    my $data1_ref = shift;
    my $data2_ref = shift;
    my $action = shift;

    my %data1 = %$data1_ref;
    my %data2 = %$data2_ref;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $data1_cols = scalar keys %data1;
    my $data2_cols = scalar keys %data2;

    # get a header that is available for all files and channels
    my $required_occ = $data1_cols + $data2_cols;
    my $header = '';
    foreach my $column (@{$item_data_column{$action}}) {

	my $occ = 0;
	
	foreach my $file (keys %data1) {
	    my $ch = $data1{$file};
	    if ($data_column_available{$file}{$ch}{$column}) {
		$header = $column;
		$occ++;
	    }
	}

	foreach my $file (keys %data2) {
	    my $ch = $data2{$file};
	    if ($data_column_available{$file}{$ch}{$column}) {
		$header = $column;
		$occ++;
	    }
	}

	if ($occ == $required_occ) {
	    last;
	} else {
	    $header = '';
	}
    }
    
    # check for specific settings:
    ($header) = &get_user_settings($header,'permutation_p_value_data_column',$force_global_opt,$action_num,'','__ALL__');

    $last_column_used = $header;

    my $out_header_tmp = &get_user_settings($var_def_default{'permutation_p_value_outheader'},'permutation_p_value_outheader',$force_global_opt,$action_num,'','__ALL__');
    my @out_headers = ();
    if ($out_header_tmp eq 'both') {
	@out_headers = ('p-value', 'p-permute');
    } else {
	@out_headers = ($out_header_tmp);
    }

    my $out_channel = &get_user_settings($var_def_default{'permutation_p_value_outchannel'},'permutation_p_value_outchannel',$force_global_opt,$action_num,'','__ALL__');

    my $dup_usage = &get_user_settings($var_def_default{'permutation_p_value_dups'},'permutation_p_value_dups',$force_global_opt,$action_num,'','__ALL__');
    if ($dup_usage eq 'yes') {
	$dup_usage = 1;
    } else {
	$dup_usage = 0;
    }
    
    my $iterations = &get_user_settings($var_def_default{'permutation_p_value_iterations'},'permutation_p_value_iterations',$force_global_opt,$action_num,'','__ALL__');
    unless ($iterations > 0) {
	die "Number of iterations must be greater than zero!";
    }
    if ($iterations =~ /\D/) {
	die "Non-digit found in value for 'iterations'!";
    }
	
    # determine which flags to skip:
    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'permutation_p_value_skip_flags'},'permutation_p_value_skip_flags',$force_global_opt,$action_num,'','__ALL__');
    %skip_flags = &get_skip_flags(@flags);

    my $skip_flags_all = 0;
    my %skipped = ();
    my $no_flag = 0;
    my $entries = 0;
    my $flag_threshold = 0;

    my @remove_files = ();

    # define data files:
    my $data_file1 = "perm_in1.$assign1.$assign2.$action.$action_num.$action_step.$PID";
    $data_file1 =~ s/\//_/g;
    $data_file1 =~ s/ /_/g;
    $data_file1 = "$tmp_dir/$data_file1";
    open (DATA1, ">$data_file1") 
	or die "Can't write to $data_file1: $!";
    push @remove_files, $data_file1;

    my $data_file2 = "perm_in2.$assign1.$assign2.$action.$action_num.$action_step.$PID";
    $data_file2 =~ s/\//_/g;
    $data_file2 =~ s/ /_/g;
    $data_file2 = "$tmp_dir/$data_file2";
    open (DATA2, ">$data_file2") 
	or die "Can't write to $data_file2: $!";
    push @remove_files, $data_file2;
	
    my $coordinates_file = "coordinates.$assign1.$assign2.$action.$action_num.$action_step.$PID";
    $coordinates_file =~ s/\//_/g;
    $coordinates_file =~ s/ /_/g;
    $coordinates_file = "$tmp_dir/$coordinates_file";
    open (COORDS, ">$coordinates_file") 
	or die "Can't write to $coordinates_file: $!";
    push @remove_files, $coordinates_file;
	
    # write out data:
    my @files1 = (sort keys %data1);
    my @files2 = (sort keys %data2);
    my @files = (@files1, @files2);

    my $input_size1 = scalar @files1;
    my $input_size2 = scalar @files2;
    if ($dup_usage) {
	$input_size1 *= 2;
	$input_size2 *= 2;
    }
     

    # get maximum array specifications:
    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@files);

    # go through each field
    foreach my $mrow (1..$MROW) {
	foreach my $mcol (1..$MCOL) {
	    foreach my $row (1..$ROW) {

		my %done_col = ();
	      COL:foreach my $col (1..$COL) {
		  
		  $entries++;
		  
		  # initialize output field:
		  for (my $i = 0; $i <= $#files; $i++) {
		      my $file = $files[$i];
		      my $ch = '';
		      if ($i < $data1_cols) {
			  $ch = $data1{$file};
		      } else {
			  $ch = $data2{$file};
		      }

		      if ($out_channel eq 'file') {
			  $ch = '__NO_LABEL__';
		      }

		      foreach my $out_header (@out_headers) {
			  $data{$file}{'__DATA__'}{$ch}{$out_header}{$mrow}{$mcol}{$row}{$col} = '';
		      }
		  }

		  next COL if ($done_col{$col});

		  my $range = 0;
		    
		  if ($dup_usage) {
		      # we do this only for the first file
		      
		      my $file = $files1[0];
		      my $spot = '';
		      my $next_spot = '';

		      # get current spot label:
		      if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col}) {
			  $spot = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col};
			  
			  # special settings for Arrayvision:
			  # delete numerator placed in brackets behind each gene name
			  $spot =~ s/ \(\d+\)$// if (defined $file_specs{$file}{'FORMAT'} 
						     and
						     $file_specs{$file}{'FORMAT'} eq 'ArrayVision');
		      }
		      
		      # get next spot label:
		      if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col+1}) {
			  $next_spot = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col+1};
			  
			  # special settings for Arrayvision:
			  # delete numerator placed in brackets behind each gene name
			  $next_spot =~ s/ \(\d+\)$// if (defined $file_specs{$file}{'FORMAT'} 
							  and
							  $file_specs{$file}{'FORMAT'} eq 'ArrayVision');
		      }
		      
		      unless ($spot eq $next_spot) {
			  # mismatch of spot labels,
			  # skip to next;
			  next COL;
		      } else {
			  # mark next column as done
			  $done_col{$col+1}++;
			  $range = 1;
		      }
		  }

		  my @val1 = ();
		  my @val2 = ();
		  my $skip_flags = 0;
		  
		  # retrieve values from all channels
		  for (my $i = 0; $i <= $#files; $i++) {
		      my $file = $files[$i];
		      my $ch = '';
		      if ($i < $data1_cols) {
			  $ch = $data1{$file};
		      } else {
			  $ch = $data2{$file};
		      }

		      for (my $j = 0; $j <= $range; $j++) {

			  my $val = '';
			  
			  if (defined $data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col+$j}) {
			      $val = $data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col+$j};
			  }
			  if ($i < $data1_cols) {
			      push @val1, $val;
			  } else {
			      push @val2, $val;
			  }

			  # skip unless value is numeric
			  unless ($val =~ /\d/) {
			      $skipped{'at least one of the values is undefined'}++;
			      next COL;
			  }
			  
			  my @flags = ();
			  # get flags for this channel as well
			  if (defined $data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}) {
			      push @flags, (&get_flags($data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}));
			  }
			  
			  # and also flag from both channels
			  if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}) {
			      push @flags, (&get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}));
			  }
			  
			  # skip flagged spots
			  my $skip = 0;
			  if (%skip_flags) {
			      
			      foreach (@flags) {
				  if (defined $skip_flags{$_}) {
				      $skip = 1;
				      last;
				  }
			      }			
			  }
			  
			  if ($skip) {
			      $skip_flags++;
			  }
		      }
		  }

		  if ($skip_flags > $flag_threshold) {
		      $skip_flags_all++;
		      next;
		  } else {
		      $no_flag++;
		  }

		  # store values in specified destination
		  print DATA1 (join "\t", @val1)."\n";
		  print DATA2 (join "\t", @val2)."\n";
		  print COORDS (join '|', ($mrow,$mcol,$row,$col))."\n";
	      }
	    }
	}
    }
    
    close DATA1;
    close DATA2;
    close COORDS;

    my $info = "Keys <B>$assign1</B> vs <B>$assign2</B>:<BR><TABLE><TR><TD><TABLE>";
    foreach my $file (@files1) {
	my $ch = $data1{$file};

	# get channel label name
	my $ch_out = '';
	if ($ch eq 'CH1') {
	    if (defined $file_specs{$file}{'CH1_LABEL'}) {
		$ch_out = " ($file_specs{$file}{'CH1_LABEL'})";
	    }
	} elsif ($ch eq 'CH2') {
	    if (defined $file_specs{$file}{'CH2_LABEL'}) {
		$ch_out = " ($file_specs{$file}{'CH2_LABEL'})";
	    }
	}

	$info .= "<TR><TD>".(basename $file).", $ch$ch_out</TD></TR>\n";
    }
    $info .= "</TABLE>\n</TD><TD>vs</TD><TD>\n<TABLE>\n";
    foreach my $file (@files2) {
	my $ch = $data2{$file};

	# get channel label name
	my $ch_out = '';
	if ($ch eq 'CH1') {
	    if (defined $file_specs{$file}{'CH1_LABEL'}) {
		$ch_out = " ($file_specs{$file}{'CH1_LABEL'})";
	    }
	} elsif ($ch eq 'CH2') {
	    if (defined $file_specs{$file}{'CH2_LABEL'}) {
		$ch_out = " ($file_specs{$file}{'CH2_LABEL'})";
	    }
	}

	$info .= "<TR><TD>".(basename $file).", $ch$ch_out</TD></TR>\n";
    }
    $info .= "</TABLE>\n</TD></TR>\n</TABLE>\n<BR>";
    
    unless ($no_flag) {
	$info .= "<B>No input generated!</B>\n";
	return $info;
    }

    # carry out permutation
    my $permutation_program = '/home/khokamp/FPMI/arraypipe/permutationTest';
    my $data_out = "$tmp_dir/permutation_out.$assign1.$assign2.$action.$action_num.$action_step.$PID";
    my $job = "$permutation_program $coordinates_file $data_file1 $input_size1 $data_file2 $input_size2 $data_out $iterations";
#    warn "$job";
    system "$job";

    open (IN, $data_out)
	or die "Can't read $data_out: $!";
    push @remove_files, $data_out;

    $counter = 0;
    while (my $line = <IN>) {
	chomp $line;
	$counter++;

	my ($coords,$val) = split /\t/, $line;

	# get coordinates and background corrected value
	my ($mrow,$mcol,$row,$col) = split /\|/, $coords;

	
	# write back values for each spot label
	foreach my $file (@files1) {
	    my $ch = $data1{$file};

	    if ($out_channel eq 'file') {
		$ch = '__NO_LABEL__';
	    }

	    foreach my $out_header (@out_headers) {
		$data{$file}{'__DATA__'}{$ch}{$out_header}{$mrow}{$mcol}{$row}{$col} = $val;
		if ($dup_usage) {
		    $data{$file}{'__DATA__'}{$ch}{$out_header}{$mrow}{$mcol}{$row}{$col+1} = $val;
		}
	    }
	}

	foreach my $file (@files2) {
	    my $ch = $data2{$file};

	    if ($out_channel eq 'file') {
		$ch = '__NO_LABEL__';
	    }

	    foreach my $out_header (@out_headers) {
		$data{$file}{'__DATA__'}{$ch}{$out_header}{$mrow}{$mcol}{$row}{$col} = $val;
		if ($dup_usage) {
		    $data{$file}{'__DATA__'}{$ch}{$out_header}{$mrow}{$mcol}{$row}{$col+1} = $val;
		}
	    }
	}
    }
    close IN;

    if ($clean_up_files) {
	foreach my $data_file (@remove_files) {
	    unlink $data_file;
	}
    }

    # take note of available columns for later
    for (my $i = 0; $i <= $#files; $i++) {
	my $file = $files[$i];
	my $channel = '';
	if ($i < $data1_cols) {
	    $channel = $data1{$file};
	} else {
	    $channel = $data2{$file};
	}

	if ($out_channel eq 'file') {
	    $channel = '__NO_LABEL__';
	}

	foreach my $out_header (@out_headers) {
	    unless (defined $data_column_available{$file}{$channel}{$out_header}) {
		$data_column_available{$file}{$channel}{$out_header} = 1;
		push @{$data_column_available{$file}{$channel}{'__ORDER__'}}, $out_header;
	    }
	    
	    $data_done{$file}{$channel}{$out_header}++;
	}
    }

    $info .= "<B>$no_flag</B> ";
    if ($dup_usage) {
	$info .= 'duplicated ';
    }
    $info .= "entries processed<BR>\n";
    if ($skip_flags_all) {
	$info .= "<B>$skip_flags_all</B> spots skipped due to flagging.<BR>\n";
    }
    $info .=  "<BR>\n";

    unless ($counter == $no_flag) {
	$info.= "<B>WARNING: only $counter rows (instead of $no_flag) returned from Permutation module (file $data_out)!</B><BR><BR>\n";
    }
	    
    foreach (keys %skipped) {
	$info .= "<B>$skipped{$_}</B> spots were skipped because<BR>$_!<BR>\n";
    }
    
    return $info;
} 


sub ttest_between_groups_setup {
    my $action_num = shift;
    my $action = shift;
    my $ttest_method = shift || 'wilcox';

    if ($refine_parameters) {
	print &format_output( 
			     {
				 'prefix' => $ttest_method.'_ttest_2_',
				 'num' => $action_num,
				 'label' => '__ALL__',
				 'action' => $action,
				 'info' => 'Parameter Refinement!<BR>'
				 }
			     );
	return;
    }

    my @files = sort keys %data;

    my %channel_assign = &channel_assign(@files, '__combine__');  
    
    my @channel_keys = sort keys %channel_assign;

    my $key_num = scalar @channel_keys;

    unless ($key_num > 1) {
	print &format_output( 
			     {
				 'prefix' => $ttest_method.'_ttest_2_',
				 'num' => $action_num,
				 'action' => $action,
				 'info' => "<B>At least two data groups necessary - please add files and/or adjust channel annotation!</B>"
				 }
			     );
	next;
    }
    
    my %keys_done = ();

    # list the assignments for each file
    my %file_assign = ();
    foreach my $assign (keys %channel_assign) {
	foreach my $file (keys %{$channel_assign{$assign}}) {
	    $file_assign{$file}{$assign}++;
	}
    }

    foreach my $assign (@channel_keys) {

	my @files = sort keys %{$channel_assign{$assign}};
	unless (@files) {
	    print &format_output( 
			     {
				 'prefix' => $ttest_method.'_ttest_2_',
				 'num' => $action_num,
				 'action' => $action,
				 'image' => '-',
				 'info' => "<B>No files for channel assignment '$assign' found!</B>"
				 }
			     );
	    next;
	}
    }

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $dup_usage = &get_user_settings($var_def_default{$ttest_method.'_ttest_2_dups'},$ttest_method.'_ttest_2_dups',$force_global_opt,$action_num,'','__ALL__');
    if ($dup_usage eq 'yes') {
	$dup_usage = 1;
    } else {
	$dup_usage = 0;
    }
    
	
    unless ( (keys %file_assign) > (keys %channel_assign) ) {
	unless ($dup_usage) {
	    print &format_output( 
			     {
				 'prefix' => $ttest_method.'_ttest_2_',
				 'num' => $action_num,
				 'action' => $action,
				 'image' => '-',
				 'info' => "<B>Not enough input files! Please add more files or activate usage of duplicate neighbours (if available).</B>"
				 }
			     );
	    return;
	}
    }
	
	    
    # carry out action on specified data
    my @channel_assigns = sort keys %channel_assign;
    for (my $i = 0; $i < $#channel_assigns; $i++) {
	for (my $j = $i + 1; $j <= $#channel_assigns; $j++) {
	    my $channel_assign1 = $channel_assigns[$i];
	    my $channel_assign2 = $channel_assigns[$j];
	    my @files1 = (sort keys %{$channel_assign{$channel_assign1}});
	    my @files2 = (sort keys %{$channel_assign{$channel_assign2}});
	    my ($image,$info) = &ttest_between_groups($action_num,$ttest_method,$channel_assign1,$channel_assign2,\@files1,\@files2,$action,$dup_usage);
    
	    print &format_output( 
				  {
				      'prefix' => $ttest_method.'_ttest_2_',
				      'num' => $action_num,
				      'label' => '__ALL__',
				      'action' => $action,
				      'info' => $info
				      }
				  );
	}
    }
    
}


sub ttest_between_groups {
    my $action_num = shift;
    my $ttest_method = shift;
    my $assign1 = shift;
    my $assign2 = shift;
    my $data_ref1 = shift;
    my $data_ref2 = shift;
    my $action = shift;
    my $dup_usage = shift;

    my @files1 = @$data_ref1;
    my @files2 = @$data_ref2;

    my %files = ();
    @{$files{$assign1}} = @files1;
    @{$files{$assign2}} = @files2;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    # get a header that is available for all files and channels
    my @channels = ('CH1', 'CH2');
    my $required_occ = (@files1 + @files2) * @channels;
    my $header = '';
    foreach my $column (@{$item_data_column{$action}}) {

	my $occ = 0;
	
	foreach my $file (@files1,@files2) {
	    foreach my $ch (@channels) {
		if ($data_column_available{$file}{$ch}{$column}) {
		    $header = $column;
		    $occ++;
		}
	    }
	}

	if ($occ == $required_occ) {
	    last;
	} else {
	    $header = '';
	}
    }
    
    # check for specific settings:
    ($header) = &get_user_settings($header,$ttest_method.'_ttest_2_data_column',$force_global_opt,$action_num,'','__ALL__');

    unless ($header =~ /\w/) {
	die "Couldn't find header!";
    }

    $last_column_used = $header;
    my $log_transfered = 0;
    if ($header =~ /_L/) {
	$log_transfered = 1;
    }
    my $header_ratio = $header . '_R';

    my $new_label = "$assign1 vs $assign2";    
    my $ctrl1 = 'ctrl1';
    if ($assign1 =~ /(\w+)\|/) {
	$ctrl1 = $1;
    }
    my $ctrl2 = 'ctrl2';
    if ($assign2 =~ /(\w+)\|/) {
	$ctrl2 = $1;
    }
    $new_label =~ s/\|/,/g;
    if ($ctrl1 eq $ctrl2) {
	$new_label =~ s/$ctrl1,//g;
    }
    my @out_header_tmp = &get_user_settings($var_def_default{$ttest_method.'_ttest_2_outheader'},$ttest_method.'_ttest_2_outheader',$force_global_opt,$action_num,'','__ALL__');
    my @out_headers = ($header_ratio);
    foreach my $out_header_tmp (@out_header_tmp) {
	if ($out_header_tmp eq 'label p-value') {
	    push @out_headers, "$new_label p-value";
	} elsif ($out_header_tmp eq "label p-$ttest_method") {
	    push @out_headers, "$new_label p-$ttest_method";
	} else {
	    push @out_headers, $out_header_tmp;
	}
    }

    my $out_channel = &get_user_settings($var_def_default{$ttest_method.'_ttest_2_outchannel'},$ttest_method.'_ttest_2_outchannel',$force_global_opt,$action_num,'','__ALL__');

    # determine which flags to skip:
    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{$ttest_method.'_ttest_2_skip_flags'},$ttest_method.'_ttest_2_skip_flags',$force_global_opt,$action_num,'','__ALL__');
    %skip_flags = &get_skip_flags(@flags);

    my $skip_flags_all = 0;
    my %skipped = ();
    my $no_flag = 0;
    my $entries = 0;
    my $flag_threshold = 0;

    # define data file:
    my $data_file = "R_in.$assign1.$assign2.$action.$action_num.$action_step.$PID";
    $data_file =~ s/\//_/g;
    $data_file =~ s/ /_/g;
    $data_file =~ s/\|//g;
    $data_file = "$tmp_dir/$data_file";
    open (DATAIN, ">$data_file") 
	or die "Can't write to $data_file: $!";

    # get maximum array specifications:
    my @files = (@files1, @files2);
    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@files);

    # go through each field
    my @coords = ();
    foreach my $mrow (1..$MROW) {
	foreach my $mcol (1..$MCOL) {
	    foreach my $row (1..$ROW) {

		my %done_col = ();
	      COL:foreach my $col (1..$COL) {
		  
		  $entries++;
		  
		  # initialize output field:
		  for (my $i = 0; $i <= $#files; $i++) {
		      my $file = $files[$i];

		      my @channels2 = ('CH1', 'CH2');

		      if ($out_channel eq 'file') {
			  @channels2 = ('__NO_LABEL__');
		      }

		      foreach my $out_header (@out_headers) {
			  foreach my $ch (@channels2) {
			      $data{$file}{'__DATA__'}{$ch}{$out_header}{$mrow}{$mcol}{$row}{$col} = '';
			  }
		      }
		  }

		  next COL if ($done_col{$col});

		  my $range = 0;
		    
		  if ($dup_usage) {
		      # we do this only for the first file
		      
		      my $file = $files1[0];
		      my $spot = '';
		      my $next_spot = '';

		      # get current spot label:
		      if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col}) {
			  $spot = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col};
			  
			  # special settings for Arrayvision:
			  # delete numerator placed in brackets behind each gene name
			  $spot =~ s/ \(\d+\)$// if (defined $file_specs{$file}{'FORMAT'} 
						     and
						     $file_specs{$file}{'FORMAT'} eq 'ArrayVision');
		      }
		      
		      # get next spot label:
		      if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col+1}) {
			  $next_spot = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col+1};
			  
			  # special settings for Arrayvision:
			  # delete numerator placed in brackets behind each gene name
			  $next_spot =~ s/ \(\d+\)$// if (defined $file_specs{$file}{'FORMAT'} 
							  and
							  $file_specs{$file}{'FORMAT'} eq 'ArrayVision');
		      }
		      
		      unless ($spot eq $next_spot) {
			  # mismatch of spot labels,
			  # skip to next;
			  next COL;
		      } else {
			  # mark next column as done
			  $done_col{$col+1}++;
			  $range = 1;
		      }
		  }

		  # DATA COLLECTION
		  # go through each group
		  my %ratios = ();
		  foreach my $group ($assign1, $assign2) {

		      my $found_ratio = 0;

		      # retrieve values from all files of this group
		      foreach my $file (@{$files{$group}}) {

			  # deal with duplicate neighbours
			  for (my $j = 0; $j <= $range; $j++) {

			      my $ratio = 'NA';
			      $ratios{$group}{$file}{$j} = $ratio;

			      my %vals = ();

			      # check for flags
			      my @flags = ();
			      foreach my $ch (@channels) {
				  
				  # get flags for individual channel 
				  if (defined $data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}) {
				      push @flags, (&get_flags($data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}));
				  }
			      }
			  
			      # and also flag from both channels
			      if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}) {
				  push @flags, (&get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}));
			      }
			  
			      # skip flagged spots
			      my $skip = 0;
			      if (%skip_flags) {
				  
				  foreach (@flags) {
				      if (defined $skip_flags{$_}) {
					  $skip = 1;
					  last;
				      }
				  }			
			      }
			      
			      if ($skip) {
				  next;
			      }			  		      
		      
			      # get the value
			      
			      foreach my $ch (@channels) {
				  my $val = '';
				  if (defined $data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col+$j}) {
				      $val = $data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col+$j};
				  }
				  unless ($val =~ /\d/) {
				      next;
				  }
				  $vals{$ch} = $val;
			      }
			  
			      unless (keys %vals == 2) {
				  next;
			      }
			      
			      if ($header =~ /_L/) {
				  $ratio = $vals{'CH2'} - $vals{'CH1'};
			      } else {
				  if ($vals{'CH1'} == 0) {
				      next;
				  }
				  $ratio = $vals{'CH2'} / $vals{'CH1'};
			      }
			      # store ratio:
			      $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$header_ratio}{$mrow}{$mcol}{$row}{$col+$j} = $ratio;

			      # collect ratio for anova
			      $ratios{$group}{$file}{$j} = $ratio;
			      $found_ratio++;
			  }
		      }
		      
		      # make sure we have at least two values per group
		      unless ($found_ratio > 1) {
			  next COL;
		      }
		  }

		  # all values are collected in %ratios (categorized by groups)


		  # write values to file:
		  my @vals = ();
		  foreach my $group ($assign1, $assign2) {
		      foreach my $file (@{$files{$group}}) {
			  for (my $j = 0; $j <= $range; $j++) {
			      push @vals, $ratios{$group}{$file}{$j};
			  }
		      }
		  }
		  $no_flag++;
		  print DATAIN join("\t", @vals)."\n";
		  push @coords, (join '|', ($mrow,$mcol,$row,$col));
	      }
	    }
	}
    }

    close DATAIN;

    my $info = "Keys <B>$assign1</B> vs <B>$assign2</B>:<BR><TABLE><TR><TD><TABLE>";
    foreach my $file1 (@files1) {
	
	# get channel label name
	my $ch1 = 'CH1';
	my $ch1_out = '';
	if ($ch1 eq 'CH1') {
	    if (defined $file_specs{$file1}{'CH1_LABEL'}) {
		$ch1_out = " ($file_specs{$file1}{'CH1_LABEL'})";
	    }
	} elsif ($ch1 eq 'CH2') {
	    if (defined $file_specs{$file1}{'CH2_LABEL'}) {
		$ch1_out = " ($file_specs{$file1}{'CH2_LABEL'})";
	    }
	}

	my $ch2 = 'CH2';
	my $ch2_out = '';
	if ($ch2 eq 'CH1') {
	    if (defined $file_specs{$file1}{'CH1_LABEL'}) {
		$ch2_out = " ($file_specs{$file1}{'CH1_LABEL'})";
	    }
	} elsif ($ch2 eq 'CH2') {
	    if (defined $file_specs{$file1}{'CH2_LABEL'}) {
		$ch2_out = " ($file_specs{$file1}{'CH2_LABEL'})";
	    }
	}

	$info .= "<TR><TD>".(basename $file1).", $ch2$ch2_out&nbsp;/&nbsp;$ch1$ch1_out\n";
	$info .= "</TD></TR>\n";
    }
    
    $info .= "</TABLE>\n</TD><TD>vs</TD><TD>\n<TABLE>\n";
    
    foreach my $file2 (@files2) {
	    
	# get channel label name
	my $ch1 = 'CH1';
	my $ch1_out = '';
	if ($ch1 eq 'CH1') {
	    if (defined $file_specs{$file2}{'CH1_LABEL'}) {
		$ch1_out = " ($file_specs{$file2}{'CH1_LABEL'})";
	    }
	} elsif ($ch1 eq 'CH2') {
	    if (defined $file_specs{$file2}{'CH2_LABEL'}) {
		$ch1_out = " ($file_specs{$file2}{'CH2_LABEL'})";
	    }
	}

	my $ch2 = 'CH2';
	my $ch2_out = '';
	if ($ch2 eq 'CH1') {
	    if (defined $file_specs{$file2}{'CH1_LABEL'}) {
		$ch2_out = " ($file_specs{$file2}{'CH1_LABEL'})";
	    }
	} elsif ($ch2 eq 'CH2') {
	    if (defined $file_specs{$file2}{'CH2_LABEL'}) {
		$ch2_out = " ($file_specs{$file2}{'CH2_LABEL'})";
	    }
	}
	
	$info .= "<TR><TD>".(basename $file2).", $ch2$ch2_out&nbsp;/&nbsp;$ch1$ch1_out\n";
	$info .= "</TD></TR>\n";
    }
    
    $info .= "</TABLE>\n</TD></TR>\n</TABLE>\n<BR>";

    unless ($no_flag) {
	$info .= "<B>No input generated!</B>\n";
	return ('', $info);
    }

    # carry out t-test
    my $R_commands_file = "$tmp_dir/R_cmd.$assign1.$assign2.$action.$action_num.$action_step.$PID";
    $R_commands_file =~ s/\|//g;
    my $R_out = "$tmp_dir/R_out.$assign1.$assign2.$action.$action_num.$action_step.$PID";
    $R_out =~ s/\|//g;
    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";
    
    print RH "library(ctest)\n";
    
    # read in data
    print RH "dataf <- read.table(\"$data_file\")\n";
    
    # define function
    print RH "get.pval.$ttest_method.ttest <- function(dataf,index1,index2,datafilter=as.numeric){\n";	
    print RH "f <- function(i) {\n";
    if ($ttest_method eq 'welch') {
	print RH "return(t.test(datafilter(dataf[i,index1]), datafilter(dataf[i,index2]))\$p.value)\n";	
    } elsif ($ttest_method eq 'student') {
	print RH "return(t.test(datafilter(dataf[i,index1]), datafilter(dataf[i,index2], var.equal = TRUE))\$p.value)\n";	
    } elsif ($ttest_method eq 'wilcox') {
	print RH "return(wilcox.test(datafilter(dataf[i,index1]), datafilter(dataf[i,index2]))\$p.value)\n"; 
    }
    print RH "}\n";	
    print RH "return(sapply(1:length(dataf[,1]),f))\n";	
    print RH "}\n";
    
    # call function
    my $index1 = (scalar @files1) * (1 + $dup_usage);
    my $index2 = $index1 + 1;
    my $index3 = ((scalar @files1) + (scalar @files2)) * (1 + $dup_usage);
    print RH "pVal.$ttest_method.ttest <- get.pval.$ttest_method.ttest(dataf,1:$index1,$index2:$index3)\n";
    
    # write out data
    print RH "write.table(pVal.$ttest_method.ttest, \"$R_out\", sep=\"\\t\", quote=F, col.names=F, row.names=F)\n";
 
    # this calculation might take a while, so we write out a file as a signal
    # that the process has finished
    my $signal_file = $R_out.'.finished';
    if (-e $signal_file) {
	unlink $signal_file
	    or return "Can't delete signal file '$signal_file': $!\n";
    }

    print RH "out <- cbind(1,2,3,4,5,6,7,8,9)\n";
    print RH "write.table(out, \"$signal_file\")\n";
    
    close RH;

    my $image = &get_R_output($R_commands_file, $R_out, $action_num, 'all_files', $signal_file, '', 1);
    
    unless (-e $R_out) {
	die "$image";
    }

    open (IN, $R_out)
	or die "Can't read $R_out: $!";

    my $nas = 0;
    $counter = 0;
    while (my $val = <IN>) {
	chomp $val;

	if ($val =~ /na/i) {
	    $nas++;
	    $val = '';
	}

	$counter++;

	# get coordinates and background corrected value
	my ($mrow,$mcol,$row,$col) = split /\|/, shift @coords;

	
	# write back values for each spot label
	foreach my $file (@files1) {

	    my @channels2 = ('CH1', 'CH2');
	    
	    if ($out_channel eq 'file') {
		@channels2 = ('__NO_LABEL__');
	    }
	    
	    foreach my $out_header (@out_headers) {
		next if ($header_ratio eq $out_header);
		foreach my $ch (@channels2) {
		    $data{$file}{'__DATA__'}{$ch}{$out_header}{$mrow}{$mcol}{$row}{$col} = $val;
		    if ($dup_usage) {
			$data{$file}{'__DATA__'}{$ch}{$out_header}{$mrow}{$mcol}{$row}{$col+1} = $val;
		    }
		}
	    }
	}

	foreach my $file (@files2) {

	    my @channels2 = ('CH1', 'CH2');
	    
	    if ($out_channel eq 'file') {
		@channels2 = ('__NO_LABEL__');
	    }
	    
	    foreach my $out_header (@out_headers) {
		next if ($header_ratio eq $out_header);
		foreach my $ch (@channels2) {
		    $data{$file}{'__DATA__'}{$ch}{$out_header}{$mrow}{$mcol}{$row}{$col} = $val;
		    if ($dup_usage) {
			$data{$file}{'__DATA__'}{$ch}{$out_header}{$mrow}{$mcol}{$row}{$col+1} = $val;
		    }
		}
	    }
	}
    }
    close IN;

    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	unlink $data_file;
    }

    # take note of available columns for later
    for (my $i = 0; $i <= $#files; $i++) {
	my $file = $files[$i];

	my $channel = '__NO_LABEL__';

	foreach my $out_header (@out_headers) {
	    unless (defined $data_column_available{$file}{$channel}{$out_header}) {
		$data_column_available{$file}{$channel}{$out_header} = 1;
		push @{$data_column_available{$file}{$channel}{'__ORDER__'}}, $out_header;
		unless (grep /^$out_header$/, @output) {
		    push @output, $out_header;
		}
	    }
	    
	    $data_done{$file}{$channel}{$out_header}++;
	}
    }

    $info .= "<B>$no_flag</B> ";
    if ($dup_usage) {
	$info .= 'duplicated ';
    }
    $info .= "entries processed<BR>\n";
    if ($skip_flags_all) {
	$info .= "<B>$skip_flags_all</B> spots skipped due to flagging.<BR>\n";
    }
    if ($nas) {
	$info .=  "<B>$nas</B> undefined spots<BR>\n";
    }
    $info .=  "<BR>\n";

    unless ($counter == $no_flag) {
	$info.= "<B>WARNING: only $counter rows (instead of $no_flag) returned from $ttest_method t-test (file $R_out)!</B><BR><BR>\n";
    }
	    
    foreach (keys %skipped) {
	$info .= "<B>$skipped{$_}</B> spots were skipped because<BR>$_!<BR>\n";
    }
    
    $info .= "Results can be found in columns <B>".(join ', ', @out_headers)."</B><BR>\n";

    return $image, $info;
} 


sub anova_setup {
    my $action_num = shift;
    my $action = shift;

    if ($refine_parameters) {
	print &format_output( # 'anova_',$action_num,'','__ALL__',$action,'','Parameter Refinement!<BR>'
			     {
				 'prefix' => 'anova_',
				 'num' => $action_num,
				 'label' => '__ALL__',
				 'action' => $action,
				 'info' => 'Parameter Refinement!<BR>'
				 }
			     );
	return;
    }

    my @files = sort keys %data;

    my %channel_assign = &channel_assign(@files, '__combine__');  
    
    my @channel_keys = sort keys %channel_assign;

    my $key_num = scalar @channel_keys;

    unless ($key_num > 1) {
	print &format_output( # 'anova_',$action_num,'','',$action,'-',"No channel assignments found!"
			     {
				 'prefix' => 'anova_',
				 'num' => $action_num,
				 'action' => $action,
				 'info' => "<B>At least two data groups necessary - please add files and/or adjust channel annotation!</B>"
				 }
			     );
	next;
    }
    
    my %keys_done = ();

    # list the assignments for each file
    my %file_assign = ();
    foreach my $assign (keys %channel_assign) {
	foreach my $file (keys %{$channel_assign{$assign}}) {
	    $file_assign{$file}{$assign}++;
	}
    }

    foreach my $assign (@channel_keys) {

	my @files = sort keys %{$channel_assign{$assign}};
	unless (@files) {
	    print &format_output( # 'anova_',$action_num,'','',$action,'-',"No files for channel assignment '$assign1' found!"
			     {
				 'prefix' => 'anova_',
				 'num' => $action_num,
				 'action' => $action,
				 'image' => '-',
				 'info' => "<B>No files for channel assignment '$assign' found!</B>"
				 }
			     );
	    next;
	}
    }

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $dup_usage = &get_user_settings($var_def_default{'anova_dups'},'anova_dups',$force_global_opt,$action_num,'','__ALL__');
    if ($dup_usage eq 'yes') {
	$dup_usage = 1;
    } else {
	$dup_usage = 0;
    }
    
	
    unless ( (keys %file_assign) > (keys %channel_assign) ) {
	unless ($dup_usage) {
	    print &format_output( # 'anova_',$action_num,'','',$action,'-',"No files for channel assignment '$assign1' found!"
			     {
				 'prefix' => 'anova_',
				 'num' => $action_num,
				 'action' => $action,
				 'image' => '-',
				 'info' => "<B>Not enough input files! Please add more files or activate usage of duplicate neighbours (if available).</B>"
				 }
			     );
	    next;
	}
    }
	
	    
    # carry out action on specified data
#	warn "plot $file1, $ch1 against $file2, $ch2";
    
    my ($info) = &anova($action_num,\%channel_assign,\%file_assign,$action,$dup_usage);
    
    print &format_output( # 'anova_',$action_num,'','__ALL__',$action,$image,$info
			  {
			      'prefix' => 'anova_',
			      'num' => $action_num,
			      'label' => '__ALL__',
			      'action' => $action,
			      'info' => $info
			      }
			  );
}



sub anova {
    my $action_num = shift;
    my $data1_ref = shift;
    my $data2_ref = shift;
    my $action = shift;
    my $dup_usage = shift;

    my %channel_assign = %$data1_ref;
    my %file_assign = %$data2_ref;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my @files = sort keys %file_assign;

    # get a header that is available for all files and channels
    my @channels = ('CH1', 'CH2');
    my $required_occ = @files * @channels;
    my $header = '';
    foreach my $column (@{$item_data_column{$action}}) {

	my $occ = 0;
	
	foreach my $file (@files) {
	    foreach my $ch (@channels) {
		if ($data_column_available{$file}{$ch}{$column}) {
		    $header = $column;
		    $occ++;
		}
	    }
	}

	if ($occ == $required_occ) {
	    last;
	} else {
	    $header = '';
	}
    }
    
    # check for specific settings:
    ($header) = &get_user_settings($header,'anova_data_column',$force_global_opt,$action_num,'','__ALL__');

    $last_column_used = $header;
    my $header_ratio = $header . '_R';

    my $out_header_tmp = &get_user_settings($var_def_default{'anova_outheader'},'anova_outheader',$force_global_opt,$action_num,'','__ALL__');
    my @out_headers = ();
    if ($out_header_tmp eq 'both') {
	@out_headers = ('p-value', 'p-anova');
    } else {
	@out_headers = ($out_header_tmp);
    }
    @out_headers = ($header_ratio, 'F-value', @out_headers);

    my $out_channel = &get_user_settings($var_def_default{'anova_outchannel'},'anova_outchannel',$force_global_opt,$action_num,'','__ALL__');

    # determine which flags to skip:
    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'anova_skip_flags'},'anova_skip_flags',$force_global_opt,$action_num,'','__ALL__');
    %skip_flags = &get_skip_flags(@flags);

    my $skip_flags_all = 0;
    my %skipped = ();
    my $no_flag = 0;
    my $entries = 0;
    my $flag_threshold = 0;


    # get maximum array specifications:
    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@files);

    # go through each field
    foreach my $mrow (1..$MROW) {
	foreach my $mcol (1..$MCOL) {
	    foreach my $row (1..$ROW) {

		my %done_col = ();
	      COL:foreach my $col (1..$COL) {
		  
		  $entries++;
		  
		  next COL if ($done_col{$col});

		  my $range = 0;
		    
		  if ($dup_usage) {
		      # we do this only for the first file
		      
		      my $file = $files[0];
		      my $spot = '';
		      my $next_spot = '';

		      # get current spot label:
		      if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col}) {
			  $spot = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col};
			  
			  # special settings for Arrayvision:
			  # delete numerator placed in brackets behind each gene name
			  $spot =~ s/ \(\d+\)$// if (defined $file_specs{$file}{'FORMAT'} 
						     and
						     $file_specs{$file}{'FORMAT'} eq 'ArrayVision');
		      }
		      
		      # get next spot label:
		      if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col+1}) {
			  $next_spot = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col+1};
			  
			  # special settings for Arrayvision:
			  # delete numerator placed in brackets behind each gene name
			  $next_spot =~ s/ \(\d+\)$// if (defined $file_specs{$file}{'FORMAT'} 
							  and
							  $file_specs{$file}{'FORMAT'} eq 'ArrayVision');
		      }
		      
		      unless ($spot eq $next_spot) {
			  # mismatch of spot labels,
			  # skip to next;
			  next COL;
		      } else {
			  # mark next column as done
			  $done_col{$col+1}++;
			  $range = 1;
		      }
		  }

		  my %ratios = ();
		  my %skip = ();

		  # INITIALIZATION
		  # go through each group
		  foreach my $group (keys %channel_assign) {

		      # retrieve values from all files of this group
		      foreach my $file (sort keys %{$channel_assign{$group}}) {

			  # deal with duplicate neighbours
			  for (my $j = 0; $j <= $range; $j++) {

			      # initialize output field:
			      for (my $i = 0; $i <= $#files; $i++) {
				  my $file = $files[$i];
				  foreach my $out_header (@out_headers) {
				      $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header}{$mrow}{$mcol}{$row}{$col+$j} = '';
				  }
			      }
			  }
		      }
		  }

		  # DATA COLLECTION
		  # go through each group
		  my $values = 0;
		  foreach my $group (keys %channel_assign) {

		      # retrieve values from all files of this group
		      foreach my $file (sort keys %{$channel_assign{$group}}) {

			  # deal with duplicate neighbours
			  for (my $j = 0; $j <= $range; $j++) {

			      my $ratio = '';
			      my @vals = ();

			      # check for flags
			      my @flags = ();
			      foreach my $ch (@channels) {
				  
				  # get flags for individual channel 
				  if (defined $data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}) {
				      push @flags, (&get_flags($data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}));
				  }
			      }
			  
			      # and also flag from both channels
			      if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}) {
				  push @flags, (&get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col+$j}));
			      }
			  
			      # skip flagged spots
			      my $skip = 0;
			      if (%skip_flags) {
				  
				  foreach (@flags) {
				      if (defined $skip_flags{$_}) {
					  $skip = 1;
					  last;
				      }
				  }			
			      }
			      
			      if ($skip) {
				  $skip{$file}{$j}++;
				  next;
			      }			  		      
		      
			      # get the value
			      
			      foreach my $ch (@channels) {
				  my $val = '';
				  if (defined $data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col+$j}) {
				      $val = $data{$file}{'__DATA__'}{$ch}{$header}{$mrow}{$mcol}{$row}{$col+$j};
				  }
				  unless ($val =~ /\d/) {
				      $skip{$file}{$j}++;
				      next;
				  }
				  push @vals, $val;
			      }
			  
			      unless (@vals == 2) {
				  $skip{$file}{$j}++;
				  next;
			      }
			      
			      if ($header =~ /_L/) {
				  $ratio = $vals[1] - $vals[0];
			      } else {
				  if ($vals[0] == 0) {
				      $skip{$file}{$j}++;
				      next;
				  }
				  $ratio = $vals[1] / $vals[0];
			      }
			      # store ratio:
			      $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$header_ratio}{$mrow}{$mcol}{$row}{$col+$j} = $ratio;

			      # collect ratio for anova
			      push @{$ratios{$group}}, $ratio;
			  }
		      }
		      
		      # make sure we have at least one value per group
		      unless (defined @{$ratios{$group}}) {
			  next COL;
		      } else {
			  $values += scalar @{$ratios{$group}};
		      }
		  }

		  # all values are collected in %ratios (categorized by groups)

		  # make sure we have sufficient values:
		  # otherwise $df_error is 0
		  unless ($values > (scalar keys %ratios) ) {
		      next COL;
		  }

		  # carry out ANOVA
		  my ($F_value,$p_value) = &getANOVA(\%ratios);

		  # write back values
		  # go through each group
		  foreach my $group (keys %channel_assign) {

		      # retrieve values from all files of this group
		      foreach my $file (sort keys %{$channel_assign{$group}}) {

			  # deal with duplicate neighbours
			  for (my $j = 0; $j <= $range; $j++) {
#			      next if (defined $skip{$file}{$j});
			      $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_headers[1]}{$mrow}{$mcol}{$row}{$col+$j} = $F_value;
			      for (my $i = 2; $i <= $#out_headers; $i++) {
				  $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_headers[$i]}{$mrow}{$mcol}{$row}{$col+$j} = $p_value;
			      }
			  }
		      }
		  }

	      }
	    }
	}
    }
    


    # take note of available columns for later
    for (my $i = 0; $i <= $#files; $i++) {
	my $file = $files[$i];
	my $channel = '__NO_LABEL__';
	foreach my $out_header (@out_headers) {
	    unless (defined $data_column_available{$file}{$channel}{$out_header}) {
		$data_column_available{$file}{$channel}{$out_header} = 1;
		push @{$data_column_available{$file}{$channel}{'__ORDER__'}}, $out_header;
	    }
	    
	    $data_done{$file}{$channel}{$out_header}++;
	}
    }

    
    return 'ANOVA finished!';
} 


sub getANOVA {
    my $ref = shift;
    my %groups = %$ref;

    my %SS = ();
    my @val_g = ();


    # 1) input consists of two or more groups of values, i.e. ratios from
    # replicates in experimment T1 and ratios from replicates in experiment T2

    foreach my $group (keys %groups) {
        my @val = @{$groups{$group}};

        # 2) calculate the mean and sums of squares (SS) for each group

        my $mean = &mean(@val);
        my $ss = &getSS($mean,@val);
        $SS{$group} = $ss;
        push @val_g, @val;
    }

    # 3) calculate the mean and SS of the combined input
    my $mean_g = &mean(@val_g);
    my $SS_g = &getSS($mean_g,@val_g);

    my $SS_error = 0;

    # 4) sum up SS from all the groups to SS_error
    foreach my $group (keys %groups) {
        $SS_error += $SS{$group};
    }

    # 5) calculate SS_effect as overall SS - SS_error
    my $SS_effect = $SS_g - $SS_error;

    # 6) assign two different degrees of freedom
    #    a) df_effect = number of groups - 1
    my $df_effect = scalar(keys %groups) - 1;

    #    b) df_error = number of all input elements - number of groups
    my $df_error = scalar(@val_g) - scalar(keys %groups);

    if ($df_error <= 0) {
        die "Not enough input values!";
    }

    # 7) calculate the mean squares (MS) for effect and error
    #    by difiding the SS by their according df
    my $MS_effect = $SS_effect / $df_effect;
    my $MS_error = $SS_error / $df_error;

    if ($MS_error == 0) {
        return ('undefined','');
    }

    # 8) get the F-value through dividing MS_effect by MS_error
    my $F = $MS_effect / $MS_error;

    # 9) get p-value from Statistics::Distribution module:
    my $p = 'undefined';

    # check if the Statistics::Distributions module is available
    eval { require Statistics::Distributions; };
#    warn "error: $@";
    if ($@) {
	# use the internal copy
#	warn "using internal fprob";
	$p = &Statistics_Distributions_fprob($df_effect,$df_error, $F);
    } else {
#	warn "using installed module";
	$p = Statistics::Distributions::fprob($df_effect,$df_error, $F);
    }

    return $F,$p;
    
}

sub getSS {
    my $mean = shift;
    my @values = @_;
    my $ss = 0;
    foreach (@values) {
	my $add = $mean - $_;
	$add = $add * $add;
	$ss += $add;
    }
    return $ss;
}


###################################################
#### The following subroutines were taken from the 
#### Statistics::Distributions Perl module 
#### provided by Michael Kospach:
###################################################

sub Statistics_Distributions_fprob { # Upper probability   F(x,n1,n2)
	my ($n, $m, $x) = @_;
	if (($n<=0) || ((abs($n)-(abs(int($n))))!=0)) {
		die "Invalid n: $n\n"; # first degree of freedom
	}
	if (($m<=0) || ((abs($m)-(abs(int($m))))!=0)) {
		die "Invalid m: $m\n"; # second degree of freedom
	} 
	return precision_string(_subfprob($n, $m, $x));
}

sub _subfprob {
	my ($n, $m, $x) = @_;
	my $p;

	if ($x<=0) {
		$p=1;
	} elsif ($m % 2 == 0) {
		my $z = $m / ($m + $n * $x);
		my $a = 1;
		for (my $i = $m - 2; $i >= 2; $i -= 2) {
			$a = 1 + ($n + $i - 2) / $i * $z * $a;
		}
		$p = 1 - ((1 - $z) ** ($n / 2) * $a);
	} elsif ($n % 2 == 0) {
		my $z = $n * $x / ($m + $n * $x);
		my $a = 1;
		for (my $i = $n - 2; $i >= 2; $i -= 2) {
			$a = 1 + ($m + $i - 2) / $i * $z * $a;
		}
		$p = (1 - $z) ** ($m / 2) * $a;
	} else {
		my $y = atan2(sqrt($n * $x / $m), 1);
		my $z = sin($y) ** 2;
		my $a = ($n == 1) ? 0 : 1;
		for (my $i = $n - 2; $i >= 3; $i -= 2) {
			$a = 1 + ($m + $i - 2) / $i * $z * $a;
		} 
		my $b = PI;
		for (my $i = 2; $i <= $m - 1; $i += 2) {
			$b *= ($i - 1) / $i;
		}
		my $p1 = 2 / $b * sin($y) * cos($y) ** $m * $a;

		$z = cos($y) ** 2;
		$a = ($m == 1) ? 0 : 1;
		for (my $i = $m-2; $i >= 3; $i -= 2) {
			$a = 1 + ($i - 1) / $i * $z * $a;
		}
		$p = max(0, $p1 + 1 - 2 * $y / PI
			- 2 / PI * sin($y) * cos($y) * $a);
	}
	return $p;
}

sub precision {
	my ($x) = @_;
	return abs int(log10(abs $x) - SIGNIFICANT);
}

sub precision_string {
	my ($x) = @_;
	if ($x) {
		return sprintf "%." . precision($x) . "f", $x;
	} else {
		return "0";
	}
}

sub max {
        my $max = shift;
        my $next;
        while (@_) {
                $next = shift;
                $max = $next if ($next > $max);
        }
        return $max;
}

sub log10 {
	my $n = shift;
	return log($n) / log(10);
}

####################################################
### end of subroutines from Statistics::Distribution
####################################################



sub channel_assign {

    my @files = @_;
    my $combine = '';
    if ($files[-1] eq '__combine__') {
	pop @files;
	$combine = 1;
    }
    my %channel_assign = ();

    foreach my $file (@files) {
	
	my $full_file = $file;
	my ($ch1_assign,$ch2_assign);

	# get info from file specs hash:
	if (defined $file_specs{$file}{'CH1_ANNOTATE'}) {
	    $ch1_assign = $file_specs{$file}{'CH1_ANNOTATE'};
	}
	if (defined $file_specs{$file}{'CH2_ANNOTATE'}) {
	    $ch2_assign = $file_specs{$file}{'CH2_ANNOTATE'};
	}
	
	# overwrite with any CGI parameters
	if (defined $cgi->param("CH1_ANNOTATE_$full_file")) {
	    $ch1_assign = $cgi->param("CH1_ANNOTATE_$full_file");
	}
	if (defined $cgi->param("CH2_ANNOTATE_$full_file")) {
	    $ch2_assign = $cgi->param("CH2_ANNOTATE_$full_file");
	}
	

	if ($combine) {
	    if ($ch1_assign =~ /\w/
		and
		$ch2_assign =~ /\w/) {		
		my ($first,$second) = ($ch1_assign,$ch2_assign);
#		($first,$second) = sort ($ch1_assign,$ch2_assign);
		if (defined $channel_assign{"$second|$first"}) {
		    $channel_assign{"$second|$first"}{$file}++;
		} else {
		    $channel_assign{"$first|$second"}{$file}++;
		}
	    }
	} else {
	    if ($ch1_assign =~ /\w/) {
		$channel_assign{$ch1_assign}{$file}{'CH'} = 'CH1';
	    }
	    if ($ch2_assign =~ /\w/) {
		$channel_assign{$ch2_assign}{$file}{'CH'} = 'CH2';
	    }
	}
    }

    return %channel_assign;
}


sub replicate_assign {

    my %bio_rep = ();
    my %tech_rep = ();

    my @files = @_;

    foreach my $file (@files) {
	
	my $bio_rep = '';
	my $tech_rep = '';
	
	my $full_file = $file;

	if (defined $cgi->param("BIO_REP_$full_file")) {
	    $bio_rep = $cgi->param("BIO_REP_$full_file");
	}
	if (defined $cgi->param("TECH_REP_$full_file")) {
	    $tech_rep = $cgi->param("TECH_REP_$full_file");
	}
	push @{$bio_rep{'REP'}{$bio_rep}}, $file;
	$bio_rep{'FILE'}{$file} = $bio_rep;
	$tech_rep{$bio_rep}{$tech_rep}++;
    }
    
    return (\%bio_rep, \%tech_rep);
}


sub local_int_dep_z_setup {
    my $action_num = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my @files = &get_user_settings($var_def_default{'local_int_dep_z_files'},'local_int_dep_z_files',$force_global_opt,$action_num);
    foreach (@files) {
	if ($_ eq 'all') {
	    @files = (sort keys %data);
	    last;
	}	
    }
    
    if ($refine_parameters) {
	print &format_output( # 'local_int_dep_z_',$action_num,'','__ALL__',$action,'','Parameter Refinement!<BR>'
			     {
				 'prefix' => 'local_int_dep_z_',
				 'num' => $action_num,
				 'label' => '__ALL__',
				 'action' => $action,
				 'info' => 'Parameter Refinement!<BR>'
				 }
			     );
	return;
    }

    foreach my $file (@files) {
	# carry out action on each file ...
	# but skip those that are banned:
	next if (defined $work_files_skip{$file});
	my ($image,$info) = &local_int_dep_z($action_num,$action,$file);
	print &format_output( # 'local_int_dep_z_',$action_num,$file,'__ALL__',$action,$image,$info
			      {
				  'prefix' => 'local_int_dep_z_',
				  'num' => $action_num,
				  'file' => $file,
				  'label' => '__ALL__',
				  'action' => $action,
				  'image' => $image,
				  'info' => $info
				  }
			      );
    }

}

sub local_int_dep_z {
    my $action_num = shift;
    my $action = shift;
    my @files = @_;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $header = '';
    LOCAL_INT_DEP_Z_COL:foreach my $column (@{$item_data_column{$action}}) {

	my $occ = 0;
	foreach my $label (sort {$b cmp $a} keys %{$data{$files[0]}{'__DATA__'}}) {
	    my $not_found = 0;
	    foreach my $file (@files) {
		unless (defined $data_column_available{$file}{$label}{$column}) {
		    $not_found = 1;
		}
	    }
	    
	    if ($not_found == 0) {
		$header = $column;
		last LOCAL_INT_DEP_Z_COL;
	    }
	}
    }

    # check for specific settings:
    $header = &get_user_settings($header,'local_int_dep_z_data_column',$force_global_opt,$action_num,$files[0],'__ALL__');

    $last_column_used = $header;

    # get labels for channel 1 (
    my $ch_up = $var_def_default{'local_int_dep_z_ch1'};
    
    # check for user specific settings
    $ch_up = &get_user_settings($ch_up,'local_int_dep_z_ch1',$force_global_opt,$action_num,$files[0],'__ALL__');

    # collect labels
    my %labels = ();

    foreach my $label (sort keys %{$data{$files[0]}{'__DATA__'}}) {
	
	# skip columns that are not label specific ('spot labels')
	next if ($label eq '__NO_LABEL__');
	$labels{$label}++;
    }

    unless ($labels{$ch_up}) {
	warn "# ERROR: Label '$ch_up' is not defined for ".(basename $files[0])."!";
	die "# ERROR: Label '$ch_up' is not defined for ".(basename $files[0])."!\n";
    }

    my $ch_down = '';
    foreach (keys %labels) {
	next unless (/\w/);
	unless ($ch_up eq $_) {
	    $ch_down = $_;
	    last;
	}
    }
    unless ($ch_down) {
	die "# ERROR: No channel 2 defined for ".(basename $files[0])." (Ch1 = '$ch_up', labels available: '".(join "','", keys %labels)."')!\n";
    }

    
    my ($ch_up_out, $ch_down_out) = ($ch_up,$ch_down);
    if ($ch_up eq 'CH1') {
	if (defined $file_specs{$files[0]}{'CH1_LABEL'}) {
	    $ch_up_out = $file_specs{$files[0]}{'CH1_LABEL'};
	}
	if (defined $file_specs{$files[0]}{'CH2_LABEL'}) {
	    $ch_down_out = $file_specs{$files[0]}{'CH2_LABEL'};
	}
    } else {
	if (defined $file_specs{$files[0]}{'CH2_LABEL'}) {
	    $ch_up_out = $file_specs{$files[0]}{'CH2_LABEL'};
	}
	if (defined $file_specs{$files[0]}{'CH1_LABEL'}) {
	    $ch_down_out = $file_specs{$files[0]}{'CH1_LABEL'};
	}
    }

    my @flags = &get_user_settings($var_def_default{'local_int_dep_z_skip_flags'},'local_int_dep_z_skip_flags',$force_global_opt,$action_num,$files[0],'__ALL__');
    my %skip_flags = ();
    %skip_flags = &get_skip_flags(@flags);

    my $out_header = "Z-score";

    my @work_labels = sort keys %labels;

    my $flag_no_ratio = &get_user_settings($var_def_default{'local_int_dep_z_no_ratio'},'local_int_dep_z_no_ratio',$force_global_opt,$action_num,$files[0],'__ALL__');

    # order all elements by intensity
    my %intensities = ();
    my %ratio = ();
    my %product = ();
    my ($mrow,$mcol,$row,$col);
    my %flags = ();
    my %skipped = ();
    my $no_flag = 0;
    my $dots = 0;

    my $loged = 0;
    my $ratio_header = $header.'_R';
    if ($header =~ /_L/) {
	$loged = 1;
    } else {
	$ratio_header = $header.'_L_R';
    }
    my $product_header = $header.'_I';
    unless ($header =~ /_L/) {
	$product_header = $header.'_L_I';
    }

    # go through each field
    foreach my $file (sort @files) {
	foreach $mrow (keys %{$data{$file}{'__DATA__'}{$ch_up}{$header}}) {
	    foreach $mcol (keys %{$data{$file}{'__DATA__'}{$ch_up}{$header}{$mrow}}) {
		foreach $row (keys %{$data{$file}{'__DATA__'}{$ch_up}{$header}{$mrow}{$mcol}}) {
		    foreach $col (keys %{$data{$file}{'__DATA__'}{$ch_up}{$header}{$mrow}{$mcol}{$row}}) {
			
			# retrieve values from both channels
			my ($val_up,$val_down,$ratio,$product);
			$val_up = '';
			$val_down = '';
			
			# initialize output field:
			$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header}{$mrow}{$mcol}{$row}{$col} = '';
			$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$ratio_header}{$mrow}{$mcol}{$row}{$col} = '';


			if (defined $data{$file}{'__DATA__'}{$ch_up}{$header}{$mrow}{$mcol}{$row}{$col}) {
			    $val_up = $data{$file}{'__DATA__'}{$ch_up}{$header}{$mrow}{$mcol}{$row}{$col};
			}
			if (defined $data{$file}{'__DATA__'}{$ch_down}{$header}{$mrow}{$mcol}{$row}{$col}) {
			    $val_down = $data{$file}{'__DATA__'}{$ch_down}{$header}{$mrow}{$mcol}{$row}{$col};
			}

			# skip empty or undefined values
			unless ($val_up =~ /\d/
				and
				$val_down =~ /\d/) {
			    $skipped{'at least one of the values is undefined'}++;
			    next;
			}
			
			if ($loged == 0) {
			    # take log of data
			    if ($val_up <= 0
				or
				$val_down <= 0) {
				if ($flag_no_ratio) {
				    $data{$file}{'__DATA__'}{$ch_up}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'no_ratio'};
				    $data{$file}{'__DATA__'}{$ch_down}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'no_ratio'};
				}
				$skipped{'at least one of the values is <= 0'}++;
				next;
			    }
			    my $val_up_orig = $val_up;
			    my $val_down_orig = $val_down;
			    $val_up = &log_base(2,$val_up);
			    $val_down = &log_base(2,$val_down);
			    $ratio = $val_up - $val_down;
			    $product = $val_up + $val_down;
			} else {
			    # calculate ratio and product:
			    $ratio = $val_up - $val_down;
			    $product = $val_up + $val_down;			    
			}

			$product *= 0.5;


			# get flags for first channel:
			my @flags = ();
			if (defined $data{$file}{'__DATA__'}{$ch_up}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			    @flags = &get_flags($data{$file}{'__DATA__'}{$ch_up}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
			}
			
			# get flags for second channel as well
			if (defined $data{$file}{'__DATA__'}{$ch_down}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			    push @flags, (&get_flags($data{$file}{'__DATA__'}{$ch_down}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
			}
			
			# and also flag from both channels
			if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			    push @flags, (&get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
			}
		    
			# skip flagged spots
			my $skip = 0;
			if (%skip_flags) {			    
			    foreach (@flags) {
				if (defined $skip_flags{$_}) {
				    $skip = 1;
				    last;
				}
			    }			
			}
			
			if ($skip) {
			    $skipped{'at least one of the values is flagged'}++;
			    next;
			}
			
			# if we have flags, put the data into a special bin
			if (@flags) {
			    # take the lowest-numbered flag
			    my $flag = (sort { $a <=> $b } @flags)[0];
			    push @{$flags{$flag}}, "$product,$ratio";
			} else {
			    $no_flag++;
			}
			
			# store values
			$ratio{"$file|$mrow|$mcol|$row|$col"} = $ratio;
			$product{"$file|$mrow|$mcol|$row|$col"} = 0.5*$product;
			push @{$intensities{$product}}, "$file|$mrow|$mcol|$row|$col";

			# count total number of dots to be plotted
			$dots++;
		    }
		}
	    }
	}
    }

    my $add_orientation = &get_user_settings($var_def_default{'local_int_dep_z_orient'},'local_int_dep_z_orient',$force_global_opt,$action_num,$files[0],'__ALL__');
    if ($add_orientation eq 'yes') {
	$add_orientation = 1;
    } else {
	$add_orientation = 0;
    }

    my $window = &get_user_settings($var_def_default{'local_int_dep_z_window'},'local_int_dep_z_window',$force_global_opt,$action_num,$files[0],'__ALL__');
    my @keys = (sort { $a <=> $b } keys %intensities);
    my @positions = ();
    my @values = ();

    # fill up values array
    for (my $index = 0; $index <= $#keys; $index++) {
	my $intensity = $keys[$index];
	foreach my $pos (@{$intensities{$intensity}}) {
	    unless (defined $ratio{$pos}) {
		warn "No ratio for position $pos!";
		next;
	    }
	    push @values, $ratio{$pos};
	    push @positions, $pos;
	}
    }

    my ($start,$stop,$mean,$stdev);
    my $prev_start = -1;
    my @window = ();
    my %scores = ();

    my $mean_file = "$tmp_dir/$action_num.$action.$action_step".(basename $files[0]).".$header.mean.$PID.tmp";
    open (MEAN , ">$mean_file")
	or return '',"ERROR: Can't write to file '$mean_file': $!\n";
    
    for (my $index = 0; $index <= $#values; $index++) {
	
	# fill window with values
	if ($index < $window/2) {
	    $start = 0;
	    $stop = $window-1;
	} elsif ( $index > ($#values - $window/2) ) {
	    $start = $#values - $window;
	    $stop = $#values;
	} else {
	    $start = $index - $window/2;
	    $stop = $start + $window-1;
	}

	unless ($prev_start == $start) {
	    @window = ();
	    foreach (my $i = $start; $i <= $stop; $i++) {
		last if ($i > $#values);
		push @window, $values[$i];
	    }
	    $prev_start = $start;
	    # get mean and SD for window
	    ($mean,$stdev) = &window_values(@window);
	}       
    
	# get z value for spot
	my $pos = $positions[$index];
	my $ratio = $values[$index];
	my $z;
	if ($stdev == 0) {
	    $z = 'N/A';
	} else {
	    $z = abs($mean - $ratio) / $stdev;

	    if ($add_orientation
		and
		$ratio < $mean) {
		$z *= -1;
	    }
	}

	my $prod = $product{$pos};

	# store z-value
	my ($file,$mrow,$mcol,$row,$col) = split /\|/, $pos;
	$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_header}{$mrow}{$mcol}{$row}{$col} = $z;
	$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$ratio_header}{$mrow}{$mcol}{$row}{$col} = $ratio;
	$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$product_header}{$mrow}{$mcol}{$row}{$col} = $prod;

	print MEAN "$prod\t$mean\n";

	# store values for plotting
	if (abs($z) > 1.96) {
	    push @{$scores{1.96}}, "$prod,$ratio";
	} elsif (abs($z) >= 1) {
	    push @{$scores{1}}, "$prod,$ratio";
	} else {
	    push @{$scores{0}}, "$prod,$ratio";
	}
    }

    close MEAN;

    # take note of available columns for later
    
    foreach my $file (@files) {
	unless (defined $data_column_available{$file}{'__NO_LABEL__'}{$out_header}) {
	    $data_column_available{$file}{'__NO_LABEL__'}{$out_header} = 1;
	    push @{$data_column_available{$file}{'__NO_LABEL__'}{'__ORDER__'}}, $out_header;
	    $data_done{$file}{'__NO_LABEL__'}{$out_header}++;
	}
	unless (defined $data_column_available{$file}{'__NO_LABEL__'}{$ratio_header}) {
	    $data_column_available{$file}{'__NO_LABEL__'}{$ratio_header} = 1;
	    push @{$data_column_available{$file}{'__NO_LABEL__'}{'__ORDER__'}}, $ratio_header;
	    $data_done{$file}{'__NO_LABEL__'}{$ratio_header}++;
	}
	unless (defined $data_column_available{$file}{'__NO_LABEL__'}{$product_header}) {
	    $data_column_available{$file}{'__NO_LABEL__'}{$product_header} = 1;
	    push @{$data_column_available{$file}{'__NO_LABEL__'}{'__ORDER__'}}, $product_header;
	    $data_done{$file}{'__NO_LABEL__'}{$product_header}++;
	}
    }


    # set user defined colours for score
    my %score2col_tmp = %score2col;
    foreach my $score (keys %score2col_tmp) {
 
	my $score_colour = $score2col_tmp{$score};
	# check for user specific settings
	$score_colour = &get_user_settings($score_colour,'local_int_dep_z_score_colour'.$score,$force_global_opt,$action_num);

	# skip if colour is the same
	next if ($score_colour eq $score2col_tmp{$score});
	
	# otherwise change settings
	$score2col_tmp{$score} = $score_colour;
    }
    
    # set user defined colours for flag
    my %flag2col_tmp = ();
    foreach my $flag (0, (sort { $a <=> $b } keys %flag_values_num)) {
 
	my $default = '';
	if (defined $pref_flag2col{$flag}) {
	    $default = $pref_flag2col{$flag};
	}
	# check for user specific settings
	my $flag_colour = &get_user_settings($default,'local_int_dep_z_flag_colour'.$flag,$force_global_opt,$action_num);
	$flag2col_tmp{$flag} = $flag_colour if ($flag_colour =~ /\w/);	    
    }
    
    # plot data
    my %data_files = ();
    my %flag_files = ();
    my $min_x = 99999;
    my $min_y = 99999;
    my $max_x = -99999;
    my $max_y = -99999;
    my @remove_files = ();
    foreach (keys %scores) {
	my $out_file = "$tmp_dir/$action_num.$action.$action_step".(basename $files[0]).".$header.data.$_.$PID.tmp";
	open (OUT , ">$out_file")
	    or return "ERROR: Can't write to file '$out_file': $!\n";
	foreach (@{$scores{$_}}) {
	    my ($x,$y) = split ',', $_;
	    if ($x > $max_x) {
		$max_x = $x;
	    }
	    if ($x < $min_x) {
		$min_x = $x;
	    }
	    if ($y > $max_y) {
		$max_y = $y;
	    }
	    if ($y < $min_y) {
		$min_y = $y;
	    }
	    print OUT "$x\t$y\n";
	}
	close OUT;
	push @remove_files, $out_file;
	$data_files{$_} = $out_file;
    }

    foreach (keys %flags) {
	my $out_file = "$tmp_dir/$action_num.$action.$action_step".(basename $files[0]).".$header.flags.$_.$PID.tmp";
	open (OUT , ">$out_file")
	    or return "ERROR: Can't write to file '$out_file': $!\n";
	foreach (@{$flags{$_}}) {
	    my ($x,$y) = split ',', $_;
	    if ($x > $max_x) {
		$max_x = $x;
	    }
	    if ($x < $min_x) {
		$min_x = $x;
	    }
	    if ($y > $max_y) {
		$max_y = $y;
	    }
	    if ($y < $min_y) {
		$min_y = $y;
	    }
	    print OUT "$x\t$y\n";
	}
	close OUT;
	push @remove_files, $out_file;
	$flag_files{$_} = $out_file;
    }

    # create R script
    my $R_commands_file = "$tmp_dir/R_cmd.$action.$action_num.$action_step.$PID";
    my $file_tmp = basename $files[0];
    $file_tmp =~ s/\//_/g;
    $file_tmp =~ s/ /_/g;
    my $R_out = "$tmp_dir/R_out.$file_tmp.$action.$action_num.$action_step.$PID";
    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";
    print RH "postscript(\"$R_out\",horizontal=FALSE, width = ".($IMAGE_HEIGHT/72).", height = ".($IMAGE_WIDTH/72).", paper = \"special\")\n";
    my $plotted = 0;

    # print scores in different colours
    foreach my $score ( sort { $a <=> $b } keys %data_files) {
	my $colour = 'black';
	if (defined $score2col{$score}) {
	    $colour = $score2col{$score};
	}
	my $data_file = $data_files{$score};
	print RH "data = read.table(\"$data_file\", header=F, sep=\"\\t\")\n";
	if ($plotted == 0) {
#	    print RH "plot(data\$V1,data\$V2, xlab = \"log10($ch_up*$ch_down)\", ylab = \"log2($ch_up/$ch_down)\", xlim=range($min_x:$max_x), ylim=range($min_y:$max_y), pch='.', col='$colour')\n";
	    print RH "plot(data\$V1,data\$V2, xlab = \"log2($ch_down_out*$ch_up_out)\", ylab = \"log2($ch_up_out/$ch_down_out)\", xlim=range($min_x:$max_x), ylim=range($min_y:$max_y), pch='.', col='$colour')\n";
	    $plotted++;
	} else {
	    print RH "points(data\$V1,data\$V2, pch='.', col='$colour')\n";
	}
    }
		 
    # print flags
    foreach my $flag ( sort { $a <=> $b } keys %flag_files) {
	my $colour = 'black';
	# get flag colours
	if (defined $flag2col_tmp{$flag}) {
	    unless (defined $pref_flag2col{$flag}) {
		$pref_flag2col{$flag} = $flag2col_tmp{$flag};
		$pref_flag2col{$flag2col_tmp{$flag}} = $flag;
	    }
	} else {
	    if (defined $pref_flag2col{$flag}) {
		$flag2col_tmp{$flag} = $pref_flag2col{$flag};
	    } else {
		foreach my $colour (@pref_flag_colours) {
		    unless (defined $pref_flag2col{$colour}) {
			$flag2col_tmp{$flag} = $colour;
			$pref_flag2col{$colour} = $flag;
			$pref_flag2col{$flag} = $colour;
			last;
		    }
		}
	    }
	}
	if (defined $flag2col_tmp{$flag}) {
	    $colour = $flag2col_tmp{$flag};
	}
	my $data_file = $flag_files{$flag};
	print RH "data = read.table(\"$data_file\", header=F, sep=\"\\t\")\n";
	if ($plotted == 0) {
#	    print RH "plot(data\$V1,data\$V2, xlab = \"log10($ch_up*$ch_down)\", ylab = \"log2($ch_up/$ch_down)\", xlim=range($min_x:$max_x), ylim=range($min_y:$max_y), pch='.', col='$colour')\n";
	    print RH "plot(data\$V1,data\$V2, xlab = \"log2($ch_up_out*$ch_down_out)\", ylab = \"log2($ch_up_out/$ch_down_out)\", xlim=range($min_x:$max_x), ylim=range($min_y:$max_y), pch='*', col='$colour')\n";
	    $plotted++;
	} else {
	    print RH "points(data\$V1,data\$V2, pch='*', col='$colour')\n";
	}
    }

    # print means
    print RH "data = read.table(\"$mean_file\", header=F, sep=\"\\t\")\n";
    my $colour = 'yellow';
    if ($plotted == 0) {
#	    print RH "plot(data\$V1,data\$V2, xlab = \"log10($ch_up*$ch_down)\", ylab = \"log2($ch_up/$ch_down)\", xlim=range($min_x:$max_x), ylim=range($min_y:$max_y), pch='.', col='$colour')\n";
	print RH "plot(data\$V1,data\$V2, xlab = \"0.5*log2($ch_up*$ch_down)\", ylab = \"log2($ch_up/$ch_down)\", xlim=range($min_x:$max_x), ylim=range($min_y:$max_y), pch='.', col='$colour')\n";
	$plotted++;
    } else {
	print RH "points(data\$V1,data\$V2, pch='.', col='$colour')\n";
    }
    
    print RH "abline(0,0, col = \"red\")\n";
    print RH "abline(1,0, col = \"grey\")\n";
    print RH "abline(-1,0, col = \"grey\")\n";
    
    # add legend
    print RH "legend($min_x, $max_y, 'mean ratio', lty = 1, col='$colour')\n";
    print RH "title(main = \"MA plot with z-scores for ".(basename $files[0])."\", line=-1)\n";

    print RH "dev.off()\n";

    close RH;

    my $image = &get_R_output($R_commands_file, $R_out, $action_num, 'all_files','','landscape');

    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	unlink $mean_file;
	foreach my $data_file (@remove_files) {
	    unlink $data_file;
	}
    }

    my $info = "<B>$dots</B> dots plotted<BR>\n";
    if (%scores) {
	$info .=  "<BR>
     <TABLE>
      <TR>
       <TH>Z-scores</TH><TH>#</TH><TH>colour</TH>
      </TR>\n";

	foreach my $score (sort { $a <=> $b } keys %scores) {
	    # make field to choose colour:
	    my $colour_field = '';
	    $colour_field = "<SELECT name=\"action${action_num}local_int_dep_z_score_colour$score\" size=\"1\">";
	    foreach my $col (sort { $a <=> $b } keys %score2col_tmp) {
		$colour_field .= "<OPTION".($score == $col ? ' SELECTED' : '').">$score2col_tmp{$col}</OPTION>\n";
	    }
	    $colour_field .= "\n</select>\n";
	    
	    my $range = '1&nbsp;<=&nbsp;|Z|&nbsp;<=&nbsp;1.96:&nbsp;';
	    if ($score < 1) {
		$range = '|Z|&nbsp;<&nbsp;1:&nbsp;';
	    } elsif ($score > 1) {
		$range = '|Z|&nbsp;>&nbsp;1.96:&nbsp;';
	    }
	    $info.=  "
      <TR>
       <TD ALIGN=RIGHT>$range</TD><TD ALIGN=RIGHT><B>".(scalar @{$scores{$score}})."</B></TD><TD ALIGN=RIGHT>$colour_field</TD>
      </TR>";
	}
	$info .=  "
     </TABLE>
";
    }
	    
    if (%flags) {
	$info .=  "<BR>
     <TABLE>
      <TR>
       <TH>flag</TH><TH>#</TH><TH>colour</TH>
      </TR>\n";

	my $colour_field = '';
	$colour_field = "<SELECT name=\"action${action_num}local_int_dep_z_flag_colour0\" size=\"1\">";
	foreach my $col ('black', @pref_flag_colours) {
	    $colour_field .= "<OPTION".($col eq 'black' ? ' SELECTED' : '').">$col</OPTION>\n";
	}
	$colour_field .= "\n</select>\n";

	$info .= "<TR><TD>unflagged</TD><TD><B>$no_flag</B></TD><TD ALIGN=RIGHT>$colour_field</TD></TR>\n";
	foreach my $flag (sort { $a <=> $b } keys %flags) {
	    # make field to choose colour:
	    my $colour_field = '';
	    $colour_field = "<SELECT name=\"action${action_num}local_int_dep_z_flag_colour$flag\" size=\"1\">";
	    foreach my $col (@pref_flag_colours) {
		my $selected = '';
		if (defined $flag2col_tmp{$flag}
		    and
		    $flag2col_tmp{$flag} eq $col) {
		    $selected = ' SELECTED';
		}
		$colour_field .= "<OPTION$selected>$col</OPTION>\n";
	    }
	    $colour_field .= "\n</select>\n";
	    
	    $info.=  "
      <TR>
       <TD ALIGN=RIGHT>$flag: $flag_values_num{$flag}</TD><TD ALIGN=RIGHT><B>".(scalar @{$flags{$flag}})."</B></TD><TD ALIGN=RIGHT>$colour_field</TD>
      </TR>";
	}
	$info .=  "
     </TABLE>
";
    }
	    
    foreach (keys %skipped) {
	$info .= "<B>$skipped{$_}</B> spots were skipped because<BR>$_!<BR>\n";
    }
    
    return $image, $info;
}


sub stdev {
    my @values = @_;

    my $sum = 0;
    foreach (@values) {
	$sum += $_;
    }

    my $sumsq = 0;
    foreach (@values) {
	$sumsq += $_**2;
    }

    my $size = scalar (@values);

    my $value = ($size * $sumsq - $sum**2) / $size / ($size - 1);

    # due to rounding errors we might end up with a tiny negative value here:
    if ($value < 0
	and
	$value >= -1e-10) {
	$value = 0;
    }

    my $stdev = sqrt($value);

    return $stdev;

}

sub variance {

    # unbiased estimator of the variance

    my @values = @_;

    # simplest form of calculating the spread:
#    @values = sort { $a <=> $b } @values;
#    my $min = $values[0];
#    my $max = $values[-1];
#    my $spread = abs($max - $min);
#    return $spread;

    my $sum = 0;
    foreach (@values) {
	$sum += $_;
    }

    my $sumsq = 0;
    foreach (@values) {
	$sumsq += $_**2;
    }

    my $size = scalar (@values);

    my $value = ($size * $sumsq - $sum**2) / $size / ($size - 1);

    # due to rounding errors we might end up with a tiny negative value here:
#    if ($value < 0
#	and
#	$value >= -1e-10) {
#	$value = 0;
#    }

    return $value;

}

sub window_values {

    my @values = @_;

#    my $median = &median(@values);
    my $mean = &mean(@values);

    my $sd = &stdev(@values);

    return ($mean,$sd);
}


sub weighted_mean {


# warning: apply only to ratios!
# e.g., consider the following example:
#
#  CH1 | CH2 |span|ratio
#------+-----+----+-----
# 15491|15599| 108|~1   
# 15491|15599| 108|~1
#      |     |    |
# 34221|34227|   6|~1
# 37563|42074|4511|~1.1
#
# after duplicate merging:
#  CH1 | CH2 |ratio
#------+-----+-----
# 15545|15545| 1   
# 34224|39818| 1.16
#
# after weighted merging across channels, this will result in the following:
#  CH1 | CH2 |ratio
#------+-----+-----
# 16113|33241|2.1 (!)


    my $ref = shift;
    my @values = @$ref;
    $ref = shift;
    my @spread = @$ref;
    my $average = '';

    unless (@spread) {
	$average = &median(@values);
	return $average;
    }

    my @values2 = ();
    my @spread2 = ();
    my $take_median = 0;

    foreach (@spread) {

	# separate spread from original values:
	my ($spread,@vals) = split /\|/, $_;
	if (defined $vals[0]
	    and
	    (
	     $vals[0] =~ /intensi/
	     or
	     $vals[0] =~ /ratio/
	     )
	    ) {
	    shift @vals;
	}

	# if at least one of the spreads is '0' we have to revert to median
	if ($spread == 0) {
	    $take_median = 1;
	}

	# collect spread and original values
	push @spread2, $spread;
	push @values2, @vals;
    }
    
    if ($take_median) {
	# take the median of all the original values:
	$average = &median(@values2);
	return $average;
    }

    # calculate weights according to spread:
    # (Thanks to Dr. Jenny Bryan, Statistics, UBC for this!)
    my @weights = ();
    my $scale = 0;
    foreach (@spread2) {
	# reverse spread to calculate index
	my $index = 1/$_;
	# sum up all indices for the scaling
	$scale += $index;
	push @weights, $index;
    }
    
    for (my $i = 0; $i <= $#weights; $i++) {
	# divide each index by the scale
	$weights[$i] /= $scale;
    }

    unless (@weights == @values) {
	die "#ERROR different values in ".(join ',', @weights)." and ".(join ',', @values);
    }

    my $test_sum = 0;
    foreach (@weights) {
	$test_sum += $_;
    }

    unless ( (sprintf "%.6f", $test_sum) == 1) {
	die "#ERROR: sum of weights is not 1: $test_sum";
    }

    $average = 0;
    
    my $weights_out = join '; ', @weights;

    foreach my $val (@values) {
	my $weight = shift @weights;
	$val *= $weight;
	$average += $val;
    }

    return ($average, $weights_out);
}


sub mean {
    my @values = @_;
    my $sum = 0;
    foreach (@values) {
	$sum += $_;
    }
    return $sum / @values;
}


sub flag_change_setup {
    my $action_num = shift;
    my $action = shift;
    my $scale = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my @files = &get_user_settings($var_def_default{'flag_change_files'},'flag_change_files',$force_global_opt,$action_num);
    foreach (@files) {
	if ($_ eq 'all') {
	    @files = (sort keys %data);
	    last;
	}	
    }
    
    if ($refine_parameters) {
	print &format_output( 
			     {
				 'prefix' => "limma_loess_${scale}_",
				 'num' => $action_num,
				 'label' => '__ALL__',
				 'action' => $action,
				 'info' => 'Parameter Refinement!<BR>'
				 }
			     );
	return;
    }

    foreach my $file (sort @files) {	

	my $test = $cgi->param('flag_change_in');
	warn "#KH: flags: $test";
	my @flag_change_in = &get_user_settings($var_def_default{'flag_change_in'},'flag_change_in',$force_global_opt,$action_num,$file);
	warn "#KH: flags: @flag_change_in";
	my $flag_change_out = &get_user_settings($var_def_default{'flag_change_out'},'flag_change_out',$force_global_opt,$action_num,$file);
	
	my ($image,$info) = &flag_change($action_num,$action,$file,$flag_change_out,\@flag_change_in);
	print &format_output( 
			      {
				  'prefix' => 'flag_change_',
				  'num' => $action_num,
				  'file' => $file,
				  'action' => $action,
				  'image' => $image,
				  'info' => $info
				  }
			      );
    }
}


sub flag_change {
    my $action_num = shift;
    my $action = shift;
    my $file = shift;
    my $to_flag = shift;
    my $ref = shift;
    my @from_flags = @$ref;
    
    # get numerical value for flags
    my %from_flags_tmp = ();
    my $highest_list = 0;
    my $exclude_highest_list = 0;

    foreach (@from_flags) {

	if ($_ =~ /all lists/) {

	    if ($_ eq 'all lists but last') {
		$exclude_highest_list = 1;
	    }

	    $from_flags_tmp{'list1'}++;

	    foreach my $fl (%used_flags) {
		if ($fl =~ /^list(\d+)/) {
		    if ($1 > $highest_list) {
			$highest_list = $1;;
		    }
		    $from_flags_tmp{$fl}++;
		}
	    }
	} else {
	    $from_flags_tmp{$_}++;
	}
    }

    if ($exclude_highest_list) {
	delete $from_flags_tmp{"list$highest_list"};
    }
    
    @from_flags = sort keys %from_flags_tmp;

    my %from_flags = ();
    foreach my $flag_num (@from_flags) {
	if ($flag_num =~ /\D/) {
	    $flag_num = $flag_values{$flag_num}
	}
	$from_flags{$flag_num}++;
    }

    my $to_flag_num = $to_flag;
    if ($to_flag eq 'none') {
	$to_flag = 0;
    }
    if ($to_flag_num =~ /\D/) {
	$to_flag_num = $flag_values{$to_flag_num}
    }

    # TODO: here we should register the flag with %used_flags an $highest_... if it is a list flag 

    my $info = '';
    $last_column_used = $FLAG_CHANNEL_TERM;

    # get maximum array specifications:
    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);

    # go through all spots and change flags if necessary
    my @channels = keys %{$data{$file}{'__DATA__'}};
    
    my $all_flag_change = 0;
    my %all_spot_flag_change = ();

    foreach my $mrow (1..$MROW) {
	foreach my $mcol (1..$MCOL) {
	    foreach my $row (1..$ROW) {
		foreach my $col (1..$COL) {

		    my $spot_flag_change = 0;
		    my %spot_flag_change = ();
		    foreach my $ch (@channels) {
			if (defined $data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			    my @flags = &get_flags($data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});			
			    my %new_flags = ();
			    my $flag_changed = 0;
			    foreach (@flags) {
				if (defined $from_flags{$_}) {
				    $new_flags{$to_flag_num}++;
				    $spot_flag_change{$_}++;
				    $flag_changed++;
				} else {
				    $new_flags{$_}++;
				}
			    }
			    
			    if ($flag_changed) {
				# set new flags
				my $new_flag = 0;
				foreach (keys %new_flags) {
				    $new_flag += $_;
				}
				$data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} = $new_flag;
				$spot_flag_change++;
			    }
			}
		    }
			
		    if ($spot_flag_change) {
			$all_flag_change++;
			foreach (keys %spot_flag_change) {
			    $all_spot_flag_change{$_}++;
			}		    
		    }
		}
	    }
	}
    }

    if ($all_flag_change) {	
	$info = "In $all_flag_change spots the flag(s) ";
	foreach my $num (@from_flags) {
	    $info .= "$flag_values_num{$num} ($all_spot_flag_change{$num}), ";
	}
	chop $info;
	chop $info;
	$info .= " have been changed to $to_flag.";
    } else {
	$info = "No flags were changed.";
    }

    return '', $info;

}


sub set_cutoffs_setup {
    my $action_num = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my @files = &get_user_settings($var_def_default{'cutoff_files'},'cutoff_files',$force_global_opt,$action_num);
    foreach (@files) {
	if ($_ eq 'separately') {
	    @files = ();
	    last;
	}
	if ($_ eq 'all') {
	    @files = (sort keys %data);
	    last;
	}	
    }
    
    if ($refine_parameters) {
	print &format_output( # 'cutoff_',$action_num,'','-',$action,'','Parameter Refinement!<BR>'
			     {
				 'prefix' => 'cutoff_',
				 'num' => $action_num,
				 'label' => '-',
				 'action' => $action,
				 'info' => 'Parameter Refinement!<BR>'
				 }
			     );
	return;
    }

    unless (@files) {
	# carry out action on each file ...
	foreach my $file (sort keys %data) {	
	    my $out = &adjust_to_cutoffs($action_num,$action,$file);       
	    print &format_output( # 'cutoffs_',$action_num,$file,'-',$action,'',$out
			     {
				 'prefix' => 'cutoff_',
				 'num' => $action_num,
				 'file' => $file,
				 'label' => '-',
				 'action' => $action,
				 'info' => $out
				 }
			     );
	}
    } else {
	my $out = &adjust_to_cutoffs($action_num,$action,@files);       
	print &format_output( # 'cutoffs_',$action_num,'','-',$action,'',$out
			     {
				 'prefix' => 'cutoff_',
				 'num' => $action_num,
				 'label' => '-',
				 'action' => $action,
				 'info' => $out
				 }
			     );
    }
}


sub filter_by_value_setup {
    my $action_num = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my @files = &get_user_settings($var_def_default{'by_value_files'},'by_value_files',$force_global_opt,$action_num);
    foreach (@files) {
	if ($_ eq 'separately') {
	    @files = ();
	    last;
	}
	if ($_ eq 'all') {
	    @files = (sort keys %data);
	    last;
	}	
    }
    my @files_tmp = ();
    foreach (@files) {
	next if (defined $work_files_skip{$_});
	push @files_tmp, $_;
    }
    @files = @files_tmp;
    
    if ($refine_parameters) {
	print &format_output( # 'by_value_',$action_num,'','-',$action,'','Parameter Refinement!<BR>'
			     {
				 'prefix' => 'by_value_',
				 'num' => $action_num,
				 'label' => '-',
				 'action' => $action,
				 'info' => 'Parameter Refinement!<BR>'
				 }
			     );
	return;
    }

    unless (@files) {
	foreach my $file (sort keys %data) {
	    next if ($file eq '');
	    next if (defined $work_files_skip{$file});
	    my $out = &filter_by_value($action_num,$action,$file);
	    print &format_output( # 'by_value_',$action_num,$file,'-',$action,'',$out
				 {
				     'prefix' => 'by_value_',
				     'num' => $action_num,
				     'file' => $file,
				     'label' => '-',
				     'action' => $action,
				     'info' => $out
				     }
				 );
	}
    } else {
	my $out = &filter_by_value($action_num,$action,@files);
	print &format_output( # 'by_value_',$action_num,'','-',$action,'',$out
			     {
				 'prefix' => 'by_value_',
				 'num' => $action_num,
				 'label' => '-',
				 'action' => $action,
				 'info' => $out
				 }
			     );
    }
}

sub filter_by_value {
    my $action_num = shift;
    my $action = shift;
    my @files = @_;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $header = &get_user_settings($var_def_default{'by_value_max_field'},'by_value_max_field',$force_global_opt,$action_num,$files[0]);
    if ($header eq 'auto') {
      FILTER_COL:foreach my $column (@{$item_data_column{$action}}) {
	  foreach my $file (@files) {
	      foreach my $label (keys %{$data{$file}{'__DATA__'}}) {
		  if ($data_column_available{$file}{$label}{$column}) {
		      # check that it exists for each other file/label:
		      my $not_found = 0;
		      foreach my $file2 (@files) {
			  unless (defined $data_column_available{$file2}{$label}{$column}) {
			      $not_found = 1;
			  }
		      }
		      
		      if ($not_found == 0) {
			  $header = $column;
			  last FILTER_COL;
		      }
		  }
	      }
	  }
      }
    }

    unless ($header) {
	return "Error: No data column defined to work on for $action!\n";
    }

    my $header2 = &get_user_settings($var_def_default{'by_value_min_field'},'by_value_min_field',$force_global_opt,$action_num,$files[0]);
    if ($header2 eq 'auto') {
	$header2 = $header;
    }

    $last_column_used = $header;

    my %out_lines = ();
    my %lines = ();

    my $list_flag = $flag_values{'list1'};
    my $list_file = '';
    my $sticky_list = &get_user_settings($var_def_default{'by_value_sticky_list'},'by_value_sticky_list',$force_global_opt,$action_num,$files[0]);   

    my $copy_list = &get_user_settings($var_def_default{'by_value_copy_list'},'by_value_copy_list',$force_global_opt,$action_num,$files[0]);
    unless ($copy_list eq 'yes') {
	$copy_list = '';
    }

    my $make_list = &get_user_settings($var_def_default{'by_value_make_list'},'by_value_make_list',$force_global_opt,$action_num,$files[0]);
#    warn "make_list: $make_list";
    if ($make_list eq '-') {
	$make_list = '';
    }
    if ($make_list ne ''
	and
	$make_list =~ /\W/) {
	return "Invalid character in listname! Please specify a name with alpha-numerical characters only under which to save the list.\n";
    }
    
    if ($make_list) {
	if ($make_list =~ /^auto$/i) {
	    my $ending = 1;
	    while (-e "$tmp_dir/list_file$ending") {
		$ending++;
	    }
	    $list_file = "$tmp_dir/list_file$ending";
	} else {
	    $list_file = "$tmp_dir/$make_list";
	    if (scalar (keys %data) > 1) {
		# append name of data file to avoid overwriting:
		$list_file = "$tmp_dir/$make_list.".(basename $files[0]);
	    }
	}

	if (-e $list_file) {
	    unlink $list_file or return "Can't remove previous list '$list_file': $!\n";
	}
	unless (open (LIST, ">$list_file")) {
	    return "Can't write to list '$list_file': $!\n";
	}
	# add list to working lists if sticky:
	print "<input type=\"hidden\" name=\"work_lists\" value=\"$list_file\">\n" if ($sticky_list);
	
	# try to find the next highest flag
	if ($make_list =~ /^auto$/i) {
	    my $count = 0;
	    if (%work_lists) {
		$count = scalar keys %work_lists;
	    }
	    $count++;
	    unless (defined $flag_values{"list$count"}) {
		# create a new list
		$highest_flag_num *= 2;
		$flag_values{"list$count"} = $highest_flag_num;
		$flag_values_num{$highest_flag_num} = "list$count";
		push @all_flags, "list$count";
		unshift @all_flags_num, $highest_flag_num;

		# keep track of list flags:
		if ("list$count" =~ /^list\d+/) {
		    $used_flags{"list$count"}++;
		}	    
	    }
	    $list_flag = $flag_values{"list$count"};
       
	} else {
	    # check if this flag exists,
	    # otherwise create a new number for it
	    unless (defined $flag_values{$make_list}) {
		# create a new list
		$highest_flag_num *= 2;
		$flag_values{$make_list} = $highest_flag_num;
		$flag_values_num{$highest_flag_num} = $make_list;
		push @all_flags, $make_list;
		unshift @all_flags_num, $highest_flag_num;

		# keep track of list flags:
		if ($make_list =~ /^list\d+/) {
		    $used_flags{$make_list}++;
		}	    
		
	    }
	    $list_flag = $flag_values{$make_list};
	}
	
	$work_lists{"$make_list$list_flag"}++;
    }

    my $mark = &get_user_settings($var_def_default{'by_value_flag'},'by_value_flag',$force_global_opt,$action_num,$files[0]);
    if ($mark eq 'auto') {
	$mark = 'list1';
	if (%work_lists) {
	    my $count = scalar keys %work_lists;
	    $count++;
	    unless (defined $flag_values{"list$count"}) {
		# create a new list
		$highest_flag_num *= 2;
		$flag_values{"list$count"} = $highest_flag_num;
		$flag_values_num{$highest_flag_num} = "list$count";
		push @all_flags, "list$count";
		unshift @all_flags_num, $highest_flag_num;

		# keep track of list flags:
		if ("list$count" =~ /^list\d+/) {
		    $used_flags{"list$count"}++;
		}	    
		
	    }
	    $mark = "list$count";
	}
	if ($header eq $FG_COR_SIGNAL_TERM) {
	    $mark = 'floor';
	}
    }
	
    unless (defined $flag_values{$mark}) {
	die "## $progname ERROR: Flag '$mark' not known!\n";
    }
    my $mark_num = $flag_values{$mark};

    my $threshold1 = &get_user_settings($var_def_default{'by_value_max'},'by_value_max',$force_global_opt,$action_num,$files[0]);

    if ($threshold1 =~ /auto/) {
	if ($header =~ /Z-score/) {
	    $threshold1 = 1.96;
	} elsif ($header =~ /\bp-\w/) {
	    $threshold1 = 0.05;
	} elsif ($header =~ /_R/
		 or
		 $header =~ /$MERGED_RATIOS_TERM/) {
	    if ($header =~ /_L/) {
		# twofold on the log2-scale
		$threshold1 = 1;
	    } else {
		# twofold change might be the standard thing to look for:
		$threshold1 = 2;
	    }
	} elsif ($header =~ /^${FG_COR_SIGNAL_TERM}_?/) {
	    # threshold on foreground corrected data
	    $threshold1 = -50;
	} elsif ($header =~ /_L/) {
	    $threshold1 = 10;
	} else {
	    # threshold on raw data:
	    $threshold1 = 1000;
	}
    }

    my $op1 = &get_user_settings($var_def_default{'by_value_max_op'},'by_value_max_op',$force_global_opt,$action_num,$files[0]);
    if ($op1 =~ /auto/) {
	if ($header =~ /p-/) {
	    $op1 = '<';
	} else {
	    # threshold on raw data:
	    $op1 = '>';
	}
    }
    my $connector = &get_user_settings($var_def_default{'by_value_connector'},'by_value_connector',$force_global_opt,$action_num,$files[0]);
    my $ignore_sign1 = &get_user_settings($var_def_default{'by_value_ignore_sign1'},'by_value_ignore_sign1',$force_global_opt,$action_num,$files[0]);
    my $ignore_sign2 = &get_user_settings($var_def_default{'by_value_ignore_sign2'},'by_value_ignore_sign2',$force_global_opt,$action_num,$files[0]);
    my $threshold2 = &get_user_settings($var_def_default{'by_value_min'},'by_value_min',$force_global_opt,$action_num,$files[0]);
    if ($threshold2 =~ /auto/) {
	if ($header =~ /Z-score/) {
	    $threshold2 = 1.96;
	} elsif ($header =~ /^p-\w/) {
	    $threshold2 = 0.05;
	} elsif ($header =~ /_R/
		 or
		 $header =~ /$MERGED_RATIOS_TERM/) {
	    if ($header =~ /_L/) {
		# twofold on the log2-scale
		$threshold2 = -1;
	    } else {
		# twofold change might be the standard thing to look for:
		$threshold2 = 0.5;
	    }
	} elsif ($header =~ /^${FG_COR_SIGNAL_TERM}_?/) {
	    # threshold on foreground corrected data
	    $threshold2 = -50;
	} elsif ($header =~ /_L/) {
	    $threshold2 = 10;
	} else {
	    # threshold on raw data:
	    $threshold2 = 1000;
	}
    }

    my $op2 = &get_user_settings($var_def_default{'by_value_min_op'},'by_value_min_op',$force_global_opt,$action_num,$files[0]);
    if ($op2 =~ /auto/) {
	if ($header =~ /p-/
	    or
	    $header =~ /^${FG_COR_SIGNAL_TERM}_?/
	    or
	    $header =~ /Z-score/
	    ) {
	    $op2 = '>';
	} else {
	    # threshold on raw data:
	    $op2 = '<';
	}
    }

    my @work_labels = &get_user_settings($var_def_default{'by_value_labels'},'by_value_labels',$force_global_opt,$action_num,$files[0]);
    foreach (@work_labels) {
	if (/auto/) {
	    if ($header =~ /^Ratio_/
		or
		$header =~ /_R$/
		or
		$header =~ /_I$/
		or
		$header =~ /Z-score/
		or
		$header =~ /_R_M$/
		or
		$header =~ /_R_L_M$/
		or
		$header =~ /_SD$/
		or
		$header =~ /$MERGED_RATIOS_TERM/
		) {
		@work_labels = ('__NO_LABEL__');
	    } elsif ($header =~ /^p-/) {
		# this is mostly in '__NO_LABEL__' but could also be per channel
		# so we have to check
		if (defined $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$header}) {
		    @work_labels = ('__NO_LABEL__');
		} else {
		    @work_labels = ('CH1 and CH2');
		}
	    } else {
		@work_labels = ('CH1 and CH2');
	    }
	    last;
	}
    }
    my $log_ratios1 = 0;
    foreach (@work_labels) {
	if (/ratio/) {
	    @work_labels = ('__NO_LABEL__');
	    if (/log/) {
		$log_ratios1 = 1;
	    }
	    last;
	}	
    }

    my @work_labels_out = @work_labels;
    if ($work_labels[0] =~/ and /
	or
	$work_labels[0] =~/ or /) {
	@work_labels_out = ('CH1', 'CH2');
    }
    if ($make_list) {
	print LIST "COLUMNS:MROW\tMCOL\tROW\tCOL\t$SPOT_ID_TERM";
	foreach (@work_labels_out) {
	    my $out_label = '';
	    unless ($_ eq '__NO_LABEL__') {
		$out_label = "$_ ";
	    }
	    print LIST "\t$out_label$header";
	}
	print LIST "\n";
    }

    my @work_labels2 = &get_user_settings($var_def_default{'by_value_labels2'},'by_value_labels2',$force_global_opt,$action_num,$files[0]);
    foreach (@work_labels2) {
	if (/auto/) {
	    if ($header2 =~ /^Ratio_/
		or
		$header2 =~ /_R$/
		or
		$header2 =~ /_I$/
		or
		$header2 =~ /Z-score/
		or
		$header2 =~ /_R_M$/
		or
		$header2 =~ /_R_L_M$/
		or
		$header2 =~ /_SD$/
		or
		$header2 =~ /$MERGED_RATIOS_TERM/
		) {
		@work_labels2 = ('__NO_LABEL__');
	    } elsif ($header2 =~ /^p-/) {
		# this is mostly in '__NO_LABEL__' but could also be per channel
		# so we have to check
		if (defined $data{$files[0]}{'__DATA__'}{'__NO_LABEL__'}{$header2}) {
		    @work_labels2 = ('__NO_LABEL__');
		} else {
		    @work_labels2 = ('CH1 and CH2');
		}
	    } else {
		@work_labels2 = ('CH1', 'CH2');	    
		if (@work_labels == 1) {
		    if ($work_labels[0] eq 'CH1') {
			@work_labels2 = ('CH2');
		    } elsif ($work_labels[0] eq 'CH2') {
			@work_labels2 = ('CH1');
		    }
		}
	    }
	    last;
	}
    }
    my $log_ratios2 = 0;
    foreach (@work_labels2) {
	if (/ratio/) {
	    @work_labels2 = ('__NO_LABEL__');
	    if (/log/) {
		$log_ratios2 = 1;
	    }
	    last;
	}	
    }

    my $strict_skip = &get_user_settings($var_def_default{'by_value_strict_skip'},'by_value_strict_skip',$force_global_opt,$action_num,$files[0]);
    

    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'by_value_skip_flags'},'by_value_skip_flags',$force_global_opt,$action_num,$files[0]);
    %skip_flags = &get_skip_flags(@flags);
    
    my %skip_lines = ();
    my %skipped = ('filter', 0);
    
    # filter 1:
    my ($mrow,$mcol,$row,$col);
    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@files);
    foreach $mrow (1..$MROW) {
	foreach $mcol (1..$MCOL) {
	    foreach $row (1..$ROW) {
		COL1:foreach $col (1..$COL) {
		    my $valid = 0;
		    my $reason = '';
		    foreach my $file (@files) {
			
			my ($valid1,$reason1) = &check_data($file,\@work_labels,$log_ratios1,$threshold1,$op1,$header,\%skip_flags,$ignore_sign1,$mrow,$mcol,$row,$col);
			if ($connector eq '--') {
			    if ($valid1 == 0) {
				if ($strict_skip eq 'yes'
				    or
				    @files == 1) {
				    $skip_lines{$mrow}{$mcol}{$row}{$col}++;
				    $skipped{$reason1}++;
				    next COL1;
				} else {
				    $reason = $reason1;
				    next;
				}
			    } else {
				$valid += $valid1;
				next;
			    }
			}

			# filter2:
			my $file2 = &get_user_settings($var_def_default{'by_value_files2'},'by_value_files2',$force_global_opt,$action_num,$files[0]);
			if ($file2 eq 'separately') {
			    $file2 = $file;
			}

			my ($valid2,$reason2) = &check_data($file,\@work_labels2,$log_ratios2,$threshold2,$op2,$header2,\%skip_flags,$ignore_sign2,$mrow,$mcol,$row,$col);
			my $valid_tmp = 0;
			if ($connector eq 'AND') {
			    $valid_tmp = $valid1 && $valid2;
			} elsif ($connector eq 'OR') {
			    $valid_tmp = $valid1 || $valid2;
			} else {
			    if ($valid1
				and
				$valid2 == 0) {
				$valid_tmp = 1;
			    } else {
				$valid_tmp = 0;
			    }
			}
			if ($valid_tmp == 0) {
			    $reason = $reason2;
			    if ($strict_skip eq 'yes') {
				$skip_lines{$mrow}{$mcol}{$row}{$col}++;
				$skipped{$reason2}++;
				next COL1;
			    }
			} else {
			    $valid += $valid_tmp;
			}
		    }
		    # if we have valid entries in each file
		    # then flag this as a valid entry
		    if ($valid == 0) {
			$skip_lines{$mrow}{$mcol}{$row}{$col}++;
			$skipped{$reason}++;
		    } else {
			$out_lines{$mrow}{$mcol}{$row}{$col}++;
		    }
		}
	    }
	}
    }

    my $out_label1 = join ' and ', @work_labels;
    if ($work_labels[0] =~ /(.*) (and|or) (.*)/) {
	@work_labels = ($1,$3);
    }
    my $out_label2 = $work_labels2[0];
    if ($work_labels2[0] =~ /(.*) (and|or) (.*)/) {
	@work_labels2 = ($1,$3);
    }
    my %tmp = ();
    foreach (@work_labels,@work_labels2) {
	$tmp{$_}++;
    }
    @work_labels = keys %tmp;

    if (grep '__NO_LABEL__', @work_labels) {
	# instead of flagging the common channel
	# I'd prefer to set a flag in each signal's channel
	@work_labels = ('CH1', 'CH2');
    }
    
    my @copy_files = ();
    if ($copy_list) {
	@copy_files = (keys %data);
    }

    my %hits = ();
    my %flags = ();
    foreach $mrow (1..$MROW) {
	foreach $mcol (1..$MCOL) {
	    foreach $row (1..$ROW) {
		foreach $col (1..$COL) {
		    foreach my $file (@files) {

			my @apply_flag_files = ($file);

			if ($copy_list) {
			    @apply_flag_files = @copy_files;
			}

			foreach my $label (@work_labels) {
			    if (defined $skip_lines{$mrow}{$mcol}{$row}{$col}) {
				unless ($make_list) {
				    foreach my $file (@apply_flag_files) {
					unless (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
					    $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} = 0;
					}
					if ($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} eq '') {
					    $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} = 0;
					}
					my $skip = 0;
					# make sure flag doesn't exist already
					if ($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} >= $mark_num) {
					    my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
					    foreach (@flags) {
						if ($_ == $mark_num) {
						    $skip++;
						    last;
						}
					    }
					}
					next if ($skip);
					$data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $mark_num;
				    }
				    $flags{"$mrow,$mcol,$row,$col"}++;
				}
			    } else {
				$hits{"$mrow,$mcol,$row,$col"}++;
				if ($make_list) {
				    foreach my $file (@apply_flag_files) {
					my $skip = 0;
					# make sure flag doesn't exist already
					if ($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} >= $list_flag) {
					    my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
					    foreach (@flags) {
						if ($_ == $list_flag) {
						    $skip++;
						    last;
						}
					    }
					}
					next if ($skip);
					$data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $list_flag;
				    }
				    my $spot_id = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col} || 'NA';
				    unless (defined $flags{"$mrow,$mcol,$row,$col"}) {
					# todo: skip this if we are in debugging mode and don't have values for each spot!
					if ($make_list) {
					    
					    print LIST "$mrow\t$mcol\t$row\t$col\t$spot_id";
					    foreach (@work_labels_out) {
						print LIST "\t$data{$file}{'__DATA__'}{$_}{$header}{$mrow}{$mcol}{$row}{$col}";
					    }
					    print LIST "\n";
					}
				    }
				    $flags{"$mrow,$mcol,$row,$col"}++;
				} else {
				    unless (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
					$data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} = 0;
				    }
				}
			    }
			}
			# set flag in 'no_label' if this column exists already:
			my $label = '__NO_LABEL__';
			if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}) {
			    if (defined $skip_lines{$mrow}{$mcol}{$row}{$col}) {
				unless ($make_list) {
				    foreach my $file (@apply_flag_files) {
					unless (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
					    $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} = 0;
					}
					if ($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} eq '') {
					    $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} = 0;
					}				           
					my $skip = 0;
					# make sure flag doesn't exist already
					if ($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} >= $mark_num) {
					    my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
					    foreach (@flags) {
						if ($_ == $mark_num) {
						    $skip++;
						    last;
						}
					    }
					}
					next if ($skip);
					$data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $mark_num;
				    }
				}
			    } else {
				if ($make_list) {
				    foreach my $file (@apply_flag_files) {
					my $skip = 0;
					# make sure flag doesn't exist already
					if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}
					    and
					    $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} >= $list_flag) {
					    my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
					    foreach (@flags) {
						if ($_ == $list_flag) {
						    $skip++;
						    last;
						}
					    }
					}
					next if ($skip);
					$data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $list_flag;
				    }
				    my $spot_id = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$SPOT_ID_TERM}{$mrow}{$mcol}{$row}{$col} || 'NA';
				    unless (defined $flags{"$mrow,$mcol,$row,$col"}) {
					if ($make_list) {
					    print LIST "$mrow\t$mcol\t$row\t$col\t$spot_id";
					    foreach (@work_labels_out) {
						print LIST "\t$data{$file}{'__DATA__'}{$_}{$header}{$mrow}{$mcol}{$row}{$col}";
					    }
					    print LIST "\n";
					}
				    }
				} else {
				    unless (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
					$data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} = 0;
				    }
				}
			    }
			}			    
		    }
		}
	    }
	}
    }
	
    if ($make_list) {
	close LIST;
    }

    # take note of available columns for later
    
    foreach my $file (@files) {
	foreach my $label (@work_labels) {
	    unless (defined $data_column_available{$file}{$label}{$FLAG_CHANNEL_TERM}) {
		$data_column_available{$file}{$label}{$FLAG_CHANNEL_TERM} = 1;
		push @{$data_column_available{$file}{$label}{'__ORDER__'}}, $FLAG_CHANNEL_TERM;
		$data_done{$file}{$label}{$FLAG_CHANNEL_TERM}++;
	    }
	}
    }

    my $num1 = scalar keys %hits;
    my $num2 = scalar keys %flags;
    my $header_info = $header;
    if ($ignore_sign1 eq 'yes') {
	$header_info = "|$header|";
    }
    my $header_info2 = $header2;
    if ($ignore_sign2 eq 'yes') {
	$header_info2 = "|$header2|";
    }
    my $info = "Filter criteria: <B>$header_info $op1 $threshold1";
    unless ($out_label1 eq '__NO_LABEL__') {
	$info .= " in $out_label1";
    }
    unless ($connector eq '--') { 
	$info .= " $connector $header_info2 $op2 $threshold2";
    }
    unless ($out_label2 eq '__NO_LABEL__') {
	$info .= " in $out_label2";
    }
    $info .= '</B><BR>';

    foreach (keys %skipped) {
	$info .= "<B>$skipped{$_}</B> entries were skipped due to $_!<BR>\n";
    }
    $info .= "<B>$num1</B> values passed through";
    if ($make_list) {
	my $list_file_html = basename($list_file);
	$info .= " (saved as list <B><A HREF=\"$list_file_html\" TARGET=\"_list\">".(basename $list_file)."</A></B>, flagged with <B>$list_flag</B> = <B>$flag_values_num{$list_flag}</B>)";
    }
    $info .= "!<BR>\n";
    if ($copy_list) {
	$info .= "<BR>This flag has been applied to the same positions in all other files as well.<BR>\n";
    }
    return $info;
}


sub check_data {
    my $file = shift;
    my $ref = shift;
    my @labels = @$ref;
    my $log_ratios = shift;
    my $threshold = shift;
    my $op = shift;
    my $header = shift;
    $ref = shift;
    my %skip_flags = %$ref;
    
    my $ignore_sign = shift;
    if ($ignore_sign eq 'yes') {
	$ignore_sign = 1;
    } else {
	$ignore_sign = 0;
    }

    my ($mrow,$mcol,$row,$col) = @_;

    my $filter_reason = 'filter';

    my $label_no_strict = 0;
    if ($labels[0] =~/ and /) {
	$label_no_strict = 0;
	@labels = ('CH1', 'CH2');
    } elsif ($labels[0] =~/ or /) {
	$label_no_strict = 1;
	@labels = ('CH1', 'CH2');
    }
	

    my $first_label_val = '';
    foreach my $label (@labels) {
	my $val = '';
	unless (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}) {
	    my $found = 0;
	    if ($label eq '__NO_LABEL__') {
		# we might have to calculate the ratio first:
		if (defined $data{$file}{'__DATA__'}{'CH1'}{$header}{$mrow}{$mcol}{$row}{$col}
		    and
		    defined $data{$file}{'__DATA__'}{'CH2'}{$header}{$mrow}{$mcol}{$row}{$col}) {
		    my $val1 = $data{$file}{'__DATA__'}{'CH1'}{$header}{$mrow}{$mcol}{$row}{$col};
		    my $val2 = $data{$file}{'__DATA__'}{'CH2'}{$header}{$mrow}{$mcol}{$row}{$col};
		    if ($val1 =~ /\d/
			and
			$val2 =~ /\d/) {
			if ($header =~ /_L/) {
			    $val = $val2 - $val1;
			    $found = 1;
			} else {
			    unless ($val1 == 0) {
				$val = $val2 / $val1;
				if ($log_ratios) {
				    $val = &log_base(2,$val);
				}
				$found = 1;
			    }
			}
		    }
		}
	    }
	    unless ($found) {
		if ($label_no_strict) {
		    if ($first_label_val eq '') {
			$first_label_val = -1;
			# give the next label a chance
			next;
		    } else {
			# first label didn't pass
		    }
		}
		return 0, 'missing data';
	    }
	} else {
	    $val = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
	}

	unless ($val =~ /\d/) {
	    if ($label_no_strict) {
		if ($first_label_val eq '') {
		    $first_label_val = -1;
		    # give the next label a chance
		    next;
		} else {
		    # first label didn't pass
		}
	    }
	    return 0, 'missing data';
	}

	if ($ignore_sign
	    and
	    $val < 0) {
	    $val *= -1;
	}

	if ($op eq '<') {
	    if ($val >= $threshold) {
		if ($label_no_strict) {
		    if ($first_label_val eq '') {
			$first_label_val = -1;
			# give the next label a chance
			next;
		    } else {
			# first label didn't pass
		    }
		}
		return 0, $filter_reason;
	    }
	} elsif ($op eq '<=') {
	    if ($val > $threshold) {
		if ($label_no_strict) {
		    if ($first_label_val eq '') {
			$first_label_val = -1;
			# give the next label a chance
			next;
		    } else {
			# first label didn't pass
		    }
		}
		return 0, $filter_reason;
	    }
	} elsif ($op eq '>') {
	    if ($val <= $threshold) {
		if ($label_no_strict) {
		    if ($first_label_val eq '') {
			$first_label_val = -1;
			# give the next label a chance
			next;
		    } else {
			# first label didn't pass
		    }
		}
		return 0, $filter_reason;
	    } 
	} elsif ($op eq '>=') {
	    if ($val < $threshold) {
		if ($label_no_strict) {
		    if ($first_label_val eq '') {
			$first_label_val = -1;
			# give the next label a chance
			next;
		    } else {
			# first label didn't pass
		    }
		}
		return 0, $filter_reason;
	    } 
	} else {
	    unless ($val == $threshold) {
		if ($label_no_strict) {
		    if ($first_label_val eq '') {
			$first_label_val = -1;
			# give the next label a chance
			next;
		    } else {
			# first label didn't pass
		    }
		}
		return 0, $filter_reason;
	    }
	}

	my $flag = 0;
	if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
	    $flag = $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col};
	    if ($flag eq '') {
		$flag = 0;
	    }
	}
	if ($flag >= 1) {
	    if ($skip_flags{$flag}) {
		$filter_reason = 'flagging';
		if ($label_no_strict) {
		    if ($first_label_val eq '') {
			$first_label_val = -1;
			# give the next label a chance
			next;
		    } else {
			# first label didn't pass
		    }
		}
		return 0, $filter_reason;
	    } else {
		foreach my $num (@all_flags_num) {
		    if ($num > $flag) {
			next;
		    }
		    $flag = $flag - $num;
		    
		    if ($skip_flags{$flag}
			or
			$skip_flags{$num}) {
			$filter_reason = 'flagging';
			if ($label_no_strict) {
			    if ($first_label_val eq '') {
				$first_label_val = -1;
				# give the next label a chance
				next;
			    } else {
				# first label didn't pass
			    }
			}
			return 0, $filter_reason;
		    }
		}
	    }
	}
    }
    return 1, $filter_reason;
}

sub check_data_simple {
    my $val = shift;
    unless ($val =~ /\w/) {
	return 0;
    }
    my $absolute = shift;
    if ($absolute) {
	$val = abs($val);
    }
    my $op = shift;
    my $threshold = shift;
    
    if ($op eq '<') {
	if ($val >= $threshold) {
	    return 0;
	}
    } elsif ($op eq '<=') {
	if ($val > $threshold) {
	    return 0;
	}
    } elsif ($op eq '>') {
	if ($val <= $threshold) {
	    return 0;
	} 
    } elsif ($op eq '>=') {
	if ($val < $threshold) {
	    return 0;
	} 
    } elsif ($op eq '=') {
	if ($val =~ /([a-z]| )/i) {
	    unless ($val eq $threshold) {
		return 0;
	    } 
	} else {
	    unless ($val == $threshold) {
		return 0;
	    } 	    
	} 
    } else {
	if ($val =~ /$threshold/is) {	    
	    if ($op eq '!//') {
		return 0;
	    }
	} else {
	    if ($op eq '//') {
		return 0;
	    }
	}
    }
    return 1;
}

sub pathway_plot_job {
    my $action_num = shift;
    my $action = shift;
    my $file = shift;
    my $job = shift;
    my $dir = shift;
    my $file_out = shift;
    my $out_img_dir = shift;
    my $out = '';

    system "$job";
    unless ($? == 0) {
	return "# ERROR: Couldn't execute job '$job': '$?'!\n";
    }

    # check that results have been written:
    unless (-s "$dir/$file_out") {
	return "#ERROR: No output produced!\n";
    } else {
	return $file_out;
    }
}


    

sub marrayNorm_normalization_setup {
    my $action_num = shift;
    my $action = shift;
    my $method = shift || '';
    my $method_var = $method;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $global = &get_user_settings($var_def_default{"marrayNorm${method}_global"},"marrayNorm${method}_global",$force_global_opt,$action_num);
    my $span = &get_user_settings($var_def_default{"marrayNorm${method}_span"},"marrayNorm${method}_span",$force_global_opt,$action_num);
    my @flags = &get_user_settings($var_def_default{"marrayNorm${method}_skip_flags"},"marrayNorm${method}_skip_flags",$force_global_opt,$action_num);
    unless ($method) {
	$method = &get_user_settings($var_def_default{"marrayNorm${method}_method"},"marrayNorm${method}_method",$force_global_opt,$action_num);
    }

    # check if multiple files are submitted
    # and the global normalization is requested:

    my @files = ();

    my @files_select = &get_user_settings($var_def_default{"marrayNorm${method_var}_files"},"marrayNorm${method_var}_files",$force_global_opt,$action_num);

    foreach (@files_select) {
	push @files, $_;
	if ($_ eq 'all') {
	    @files = ();
	    foreach (sort keys %data) {
		if (/\w/) {
		    push @files, $_;
		}
	    }
	    last;
	}
    }

#    print &marrayNorm_normalization($action_num,$action,\@files,$header,$FLAG_CHANNEL_TERM,$out_headerM,$out_headerA,$out_header2,$span,$global,$method,@flags);
    my %layout = ();
    foreach my $file (sort @files) {
	# group them by layout:
	my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);
	push @{$layout{"$MROW,$MCOL,$ROW,$COL"}}, $file;
    }
    
    foreach my $layout (sort keys %layout) {
	my @files = @{$layout{$layout}};

	my %headers = ();


	foreach my $file (@files) {
	    my $header = '';
	  MARRAYNORM_COL:foreach my $column (@{$item_data_column{$action}}) {
	      foreach my $label (keys %{$data{$file}{'__DATA__'}}) {
		  if ($data_column_available{$file}{$label}{$column}) {
		      $header = $column;
		      last MARRAYNORM_COL;
		  }
	      }
	  }
	    
	    
	    # check for specific settings:
	    $header = &get_user_settings($header,"marrayNorm${method_var}_data_column",$force_global_opt,$action_num,$file);
	    
	    unless ($header) {
		print "Error: No data column defined to work on for $action!<BR>\n";
	    }
	    
	    $headers{'HEADER'}{$file} = $header;
	    
	    $last_column_used = $header;
	    
	    my $out_headerM = 'Norm_L_R';
	    if ($header =~ /_M/) {
		$out_headerM = 'Norm_M_L_R';
	    }
	    
	    $headers{'out_headerM'}{$file} = $out_headerM;
	    
	    my $out_headerA = 'Norm_L_I';
	    if ($header =~ /_M/) {
		$out_headerA = 'Norm_M_L_I';
	    }
	    
	    $headers{'out_headerA'}{$file} = $out_headerA;
	    
	    my $out_header2 = 'Norm_L';
	    if ($header =~ /_M/) {
		$out_header2 = 'Norm_M_L';
	    }
	    
	    $headers{'out_header2'}{$file} = $out_header2;	

	    if ($refine_parameters) {
		print &format_output(
				     {
					 'prefix' => "marrayNorm${method_var}_",
					 'num' => $action_num,
					 'label' => '-',
					 'action' => $action,
					 'info' => 'Parameter Refinement!<BR>',
					 'file' => $file
					 }
				     );
		$data_done{$file}{'__NO_LABEL__'}{$out_headerM}++;
		$data_column_available{$file}{'__NO_LABEL__'}{$out_headerM}++;
		$data_done{$file}{'__NO_LABEL__'}{$out_headerA}++;
		$data_column_available{$file}{'__NO_LABEL__'}{$out_headerA}++;
		foreach my $label (keys %{$data{$file}{'__DATA__'}}) {
		    next if ($label eq '__NO_LABEL__');
		    $data_done{$file}{$label}{$out_header2}++;
		    $data_column_available{$file}{$label}{$out_header2}++;
		}
		next;
	    }
	}

#	my @tmp = ($file);
#	# TODO: files with the same array layout should be combined here, 
#	# because the MAD option might require all files to be looked at once!
#	print &marrayNorm_normalization($action_num,$action,\@tmp,$header,$FLAG_CHANNEL_TERM,$out_headerM,$out_headerA,$out_header2,$span,$global,$method,@flags);

	# carry out normalization on all files that have the same layout:
	print &marrayNorm_normalization($action_num,$action,\@files,\%headers,$FLAG_CHANNEL_TERM,$span,$global,$method,@flags);
	
    }
}


sub marrayNorm_normalization {
    my $action_num = shift;
    my $action = shift;
    my $ref = shift;
    my @files = @$ref;
    $ref = shift;
    my %headers = %$ref;
    my $flag_header = shift;
    my $span = shift;
    my $global = shift;
    my $method = shift;
    my @flags = (@_);

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my %skip_flags = ();
    %skip_flags = &get_skip_flags(@flags);

    # read in data if necessary
    foreach my $file (@files) {
	my $rc = &read_data($file,$headers{'HEADER'}{$file},$flag_header);
	unless ($rc eq 'ok') {
	    return "<BR><B><FONT COLOR=RED>WARNING: $rc</FONT></B><BR>\n";
	}
    }

    # write data to file
    my $file_tmp = basename $files[0];
    $file_tmp =~ s/\//_/g;
    $file_tmp =~ s/ /_/g;
    my $normM_file = "$tmp_dir/normM_$action_num.$action_step.$file_tmp.$headers{'HEADER'}{$files[0]}.$PID.tmp";
    my $normA_file = "$tmp_dir/normA_$action_num.$action_step.$file_tmp.$headers{'HEADER'}{$files[0]}.$PID.tmp";
    my $params_file = "$tmp_dir/params_$action_num.$action_step.$file_tmp.$headers{'HEADER'}{$files[0]}.$PID.tmp";         
     
    if (-e $normM_file) {
	unlink $normM_file 
	    or return "Can't delete normalized file '$normM_file': $!\n";
    }

    if (-e $normA_file) {
	unlink $normA_file 
	    or return "Can't delete normalized file '$normA_file': $!\n";
    }

    if (-e $params_file) {
	unlink $params_file 
	    or return "Can't delete params file '$params_file': $!\n";
    }

    # collect labels
    my $label = '';
    my %labels = ();
    foreach $label (sort keys %{$data{$files[0]}{'__DATA__'}}) {
	next if ($label eq '__NO_LABEL__');
	
	# collect labels for later
	$labels{$label}++;
    }

    my @labels = sort keys %labels;
    $label = $labels[0];
	
    # open output files
    my %out_files = ();
    my %fhs = ();
    my @data_files = ();
    my %data_files = ();
    $counter = 0;
    no strict 'refs';
    foreach my $file (sort @files) {

	next if ($file eq '');

	my $file_tmp = basename $file;
	$file_tmp =~ s/\//_/g;
	$file_tmp =~ s/ /_/g;
	my $header_tmp = $headers{'HEADER'}{$file};
	$header_tmp =~ s/ /_/g;

	my $data_file = "$tmp_dir/data_$action_num.$action_step.$file_tmp.$header_tmp.$PID.tmp";

	if (-e $data_file) {
	    unlink $data_file 
		or return "Can't delete data file '$data_file': $!\n";
	}

	$counter++;
	my $fh = \*$file;
	
	open ($fh, ">$data_file")
	    or die "Can't write to $data_file: $!";       
	print $file (join "\t", @labels)."\n";

	$out_files{$file} = $data_file;
	$fhs{$file} = $fh;

	push @data_files, $data_file;
	$data_files{$data_file} = basename $file;

    }


    my ($mrow,$mcol,$row,$col);
    my %lines = ();
    my @lines = ();

    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@files);
    # collect values for each spot label
  OUTER2_TMP:foreach my $mrow (1..$MROW) {
	foreach my $mcol (1..$MCOL) {
	    foreach my $row (1..$ROW) {
		foreach my $col (1..$COL) {
		  
		  # calculate line number
		  my $line = ($row - 1 + ($mrow - 1) * $ROW) * $MCOL*$COL + ($mcol - 1) * $COL + $col;
		  # keep track of conversion
		  $lines{$line} = "$mrow,$mcol,$row,$col";
		  push @lines, $line;

		  # make sure there are values for each label and each file
		  FILE1:foreach my $file (@files) {

		      my $skip = 0;
		  
		      my @out = ();

		      my $header = $headers{'HEADER'}{$file};
		      my $out_headerM = $headers{'out_headerM'}{$file};
		      my $out_headerA = $headers{'out_headerA'}{$file};
		      my $out_header2 = $headers{'out_header2'}{$file};
		      
		      # initialize fields
		      $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_headerM}{$mrow}{$mcol}{$row}{$col} = '';
		      $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_headerA}{$mrow}{$mcol}{$row}{$col} = '';

		      foreach my $label (@labels) {
			  
			  # initialize fields
			  # unless we are using normalized values (e.g. in a second run of the module)
			  unless ($out_header2 eq $header) {
			      $data{$file}{'__DATA__'}{$label}{$out_header2}{$mrow}{$mcol}{$row}{$col} = '';
			  }

			  # skip spots without value
			  unless (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}
				  and
				  $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col} =~ /\d/) {
			      $skip++;
			  }
			  
			  # skip flagged values if requested
			  if (%skip_flags
			      and
			      (defined $data{$file}{'__DATA__'}{$label}{$flag_header}{$mrow}{$mcol}{$row}{$col}
			      or
			      defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$flag_header}{$mrow}{$mcol}{$row}{$col})) {
			      my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$flag_header}{$mrow}{$mcol}{$row}{$col});
			      if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$flag_header}{$mrow}{$mcol}{$row}{$col}) {
				  push @flags, &get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$flag_header}{$mrow}{$mcol}{$row}{$col});
			      }
			      foreach (@flags) {
				  if (defined $skip_flags{$_}) {
				      $skip++;
				  }
			      }			
			  } 
			  if ($skip) {
			      push @out, 'NA';
			  } else {
			      my $val = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
			      if ($header =~ /_L/) {
				  # de-normalize data first
				  $val = 2**$val;
			      }
			      push @out, $val;
			  }
		      }

		      # if we arrived here we can print values fore each label into the file
		      print $file (join "\t", @out)."\n";
		      
		  }
	      }
	  }
      }
  }

    foreach (@files) {
	close $fhs{$_};
    }
    
    # create R commands
    my $R_commands_file = "$tmp_dir/R_cmd.$action.$action_num.$action_step.$PID";
    my $R_out = "$tmp_dir/R_out.$action.$action_num.$file_tmp.$action_step.$PID";
    my @R_out = ();
    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";
    print RH "library(marrayInput)\n";
    print RH "library(marrayNorm)\n";
    print RH "library(marrayPlots)\n";

    # define array layout
    print RH "layout <- read.marrayLayout(ngr=$MROW, ngc=$MCOL, nsr=$ROW, nsc=$COL)\n";

    # read in raw data:
    my @tmp = ();
    foreach (@data_files) {
	push @tmp, basename($_);
    }
    print RH "fnames <- c(\"".(join '","', @tmp)."\")\n";
    print RH "raw <- read.marrayRaw(fnames, path = \"$tmp_dir\", name.Gf = \"$labels[0]\", name.Rf = \"$labels[1]\", layout = layout)\n";
    
    # span factor
    if ($span =~ /\d/) {
	# this can only be set by functions using loess
	$span = ",span=$span";
    } else {
	$span = '';
    }

    if ($global eq 'no') {
	# this can only be set with the scale functions
	# i.e. globalMAD and printTipMAD
	$global = ', geo=FALSE';
    } else {
	$global = ', geo=TRUE';
    }

    # Write marrayNorm call and process table.
    my $call = "norm <- maNorm(raw, norm = \"$method\"$span, echo=TRUE)";
    if ($method eq 'globalMAD'
	or
	$method eq 'printTipMAD') {
	$call = "norm <- maNormScale(raw, norm = \"$method\"$global, echo=TRUE)";
    }
    print RH "$call\n";

    # Write the write-to-file statement in R to save data in "$normM_file".
    print RH "## write data to file\n";
    print RH "write.table(norm\@maM, \"$normM_file\", sep=\"\\t\", quote=F, col.names=F, row.names=F)\n";

    # Write the write-to-file statement in R to save data in "$normA_file".
    print RH "## write data to file\n";
    print RH "write.table(norm\@maA, \"$normA_file\", sep=\"\\t\", quote=F, col.names=F, row.names=F)\n";

    # print graphics
    print RH "## print graphics\n";
    for (my $i = 0; $i < (scalar @data_files); $i++) {
	print RH "postscript(\"$R_out$i\",horizontal=FALSE)\n";
	print RH "par(mfrow = c(2,1))\n";
	print RH "maBoxplot(raw[,".($i+1)."], x = \"maPrintTip\", y = \"maM\", main = \"$data_files{$data_files[$i]}: pre--normalization\")\n";
	print RH "maBoxplot(norm[,".($i+1)."], x = \"maPrintTip\", y = \"maM\", main = \"$data_files{$data_files[$i]}: post--$method$span\")\n";
	print RH "dev.off()\n";
	push @R_out, "$R_out$i";
    }

    # this calculation might take a while, so we write out a file as a signal
    # that the process has finished
    my $signal_file = $normM_file.'.finished';
    if (-e $signal_file) {
	unlink $signal_file
	    or return "Can't delete signal file '$signal_file': $!\n";
    }

    print RH "out <- cbind(1,2,3,4,5,6,7,8,9)\n";
    print RH "write.table(out, \"$signal_file\")\n";
    
    close RH;

    my @images = &get_R_output($R_commands_file, \@R_out, $action_num, 'all_files', $signal_file);
    
    # read normalized data    

    my %out = ();

    unless (-s $normM_file) {

	# the error messages are located in $images[0]:
	return "<B>WARNING: no output from normalization! Error message: </B><br><pre>$images[0]</pre><BR>\n";
    }

    unless (-s $normA_file) {
	return "<B>WARNING: no output from normalization! Error message: </B><br><pre>$images[0]</pre><BR>\n";
    }

    open (INM, $normM_file)
	or die "Can't read $normM_file: $!";

    open (INA, $normA_file)
	or die "Can't read $normA_file: $!";

    my $input_rows = scalar(@lines);

    my %nas = ();
    $counter = 0;
    while (<INM>) {
	chomp;
	my @tmpM = split /\t/, $_;

	$_ = <INA>;
	chomp;
	my @tmpA = split /\t/, $_;

	# get coordinates
	my $line = shift @lines;
	($mrow,$mcol,$row,$col) = split /,/, $lines{$line};
	
	foreach my $file (@files) {
	    
	    my $header = $headers{'HEADER'}{$file};
	    my $out_headerM = $headers{'out_headerM'}{$file};
	    my $out_headerA = $headers{'out_headerA'}{$file};
	    my $out_header2 = $headers{'out_header2'}{$file};

	    my $valueM = '';
	    if (@tmpM) {
		$valueM = shift @tmpM;
	    }
	    
	    # write back values for each spot label
	    if ($valueM =~ /na/i) {
		$nas{$file}++;
		$valueM = '';
	    }
	    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_headerM}{$mrow}{$mcol}{$row}{$col} = $valueM;

	    # get A value to calculate logs for each channel
	    my $valueA = '';
	    if (@tmpA) {
		$valueA = shift @tmpA;
	    }
	    if ($valueA =~ /na/i) {
		$valueA = '';
	    }
	    $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$out_headerA}{$mrow}{$mcol}{$row}{$col} = $valueA;

	    # calculation of channel intensities from ratios and intensity products:
	    #
	    # M = log(R/G)
            #   = log(R) - log(G)
	    # log(R) = M + log(G)  (i)
	    #
	    # A = 1/2 * log(R*G)
            #   = 1/2 * (log(R) + log(G))
	    #
	    # insert (i):
	    # A = 1/2 * (M + log(G) + log(G))
            # log(G) = A - M/2
	    # log(R) = M + log(G)
	    
	    my ($chG, $chR);
	    if ($valueA =~ /\d/
		and
		$valueM =~ /\d/) {
		$chG = $valueA - $valueM / 2;
		$chR = $valueM + $chG;
	    } else {
		$chG = '';
		$chR = '';
	    }

	    # in our case labels[0] was used for green channel:

	    $data{$file}{'__DATA__'}{$labels[0]}{$out_header2}{$mrow}{$mcol}{$row}{$col} = $chG;
	    $data{$file}{'__DATA__'}{$labels[1]}{$out_header2}{$mrow}{$mcol}{$row}{$col} = $chR;
	    
	}
	$counter++;
    }
    close INM;
    close INA;
		
    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	foreach my $data_file (@data_files) {
	    unlink $data_file;
	}
	foreach my $R_out_file (@R_out) {
	    unlink $R_out_file;
	}
	unlink $normA_file;
	unlink $normM_file;
    }

    # take note of available columns for later

    foreach my $file (@files) {
	
	my $header = $headers{'HEADER'}{$file};
	my $out_headerM = $headers{'out_headerM'}{$file};
	my $out_headerA = $headers{'out_headerA'}{$file};
	my $out_header2 = $headers{'out_header2'}{$file};
	
	unless (defined $data_column_available{$file}{'__NO_LABEL__'}{$out_headerM}) {
	    $data_column_available{$file}{'__NO_LABEL__'}{$out_headerM} = 1;
	    push @{$data_column_available{$file}{'__NO_LABEL__'}{'__ORDER__'}}, $out_headerM;
	}
	unless (defined $data_column_available{$file}{'__NO_LABEL__'}{$out_headerA}) {
	    $data_column_available{$file}{'__NO_LABEL__'}{$out_headerA} = 1;
	    push @{$data_column_available{$file}{'__NO_LABEL__'}{'__ORDER__'}}, $out_headerA;
	}

	$data_done{$file}{'__NO_LABEL__'}{$out_headerM}++;
	$data_done{$file}{'__NO_LABEL__'}{$out_headerA}++;

	foreach my $label (@labels) {
	    unless (defined $data_column_available{$file}{$label}{$out_header2}) {
		$data_column_available{$file}{$label}{$out_header2} = 1;
		push @{$data_column_available{$file}{$label}{'__ORDER__'}}, $out_header2;
	    }
	    $data_done{$file}{$label}{$out_header2}++;
	}
    }

    my $info = "<B>$counter</B> values processed<BR>\n";
    if (%nas) {
	$info .=  "<BR>
     <TABLE>
      <TR>
       <TH>file</TH><TH>NA's</TH>
      </TR>\n";
	foreach (@files) {
	    my $nas = 0;
	    if (defined $nas{$_}) {
		$nas = $nas{$_};
	    }
	    $info .= "<TR>
       <TD>".(basename $_)."</TD><TD><B>$nas</B></TD>
      </TR>";
	}
	$info .=  "
     </TABLE>\n";
    }

    # format output (include image and choice for parameters)
    my $file_lnk = '';
    if (@files == 1) {
	$file_lnk = $files[0];
    }
    my $image_lnk = &format_output( # 'marrayNorm_',$action_num,'','-',$action,\@images,$info
				   {
				       'prefix' => "marrayNorm${method}_",
				       'num' => $action_num,
				       'label' => '-',
				       'action' => $action,
				       'image' => \@images,
				       'info' => $info,
				       'file' => $file_lnk
				       }
				   );
    
    unless ($counter == $input_rows) {
	return "<B>WARNING: only $counter rows (instead of $input_rows)  returned from normalization!</B><BR><BR>$image_lnk\n";
    }

    return $image_lnk;
}


sub vsn_normalization_setup {
    my $action_num = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $global = &get_user_settings($var_def_default{'vsn_global'},'vsn_global',$force_global_opt,$action_num);
    my $iterations = &get_user_settings($var_def_default{'vsn_iter'},'vsn_iter',$force_global_opt,$action_num);
    my @flags = &get_user_settings($var_def_default{'vsn_skip_flags'},'vsn_skip_flags',$force_global_opt,$action_num);
    my $dehuberize = &get_user_settings($var_def_default{'vsn_dehuberize'},'vsn_dehuberize',$force_global_opt,$action_num);
    if ($dehuberize eq 'yes') {
	$dehuberize = 1;
    } else {
	$dehuberize = 0;
    }
    my $header = '';
    VSN_COL:foreach my $column (@{$item_data_column{$action}}) {
	foreach my $file (sort keys %data) {
	    foreach my $label (keys %{$data{$file}{'__DATA__'}}) {
		if ($data_column_available{$file}{$label}{$column}) {
		    $header = $column;
		    last VSN_COL;
		}
	    }
	}
    }

    # check for specific settings:
    $header = &get_user_settings($header,'vsn_data_column',$force_global_opt,$action_num);

    unless ($header) {
	print "Error: No data column defined to work on for $action!<BR>\n";
    }

    $last_column_used = $header;

    my $out_header = 'Norm_L';
    if ($header =~ /_M/) {
	$out_header = 'Norm_M_L';
    }
    my $dehub_header = 'Dehub';
    if ($header =~ /_M/) {
	$dehub_header = 'Dehub_M';
    }

    if ($refine_parameters) {
	print &format_output( # 'vsn_',$action_num,'','-',$action,'','Parameter Refinement!<BR>'
			     {
				 'prefix' => 'vsn_',
				 'num' => $action_num,
				 'label' => '-',
				 'action' => $action,
				 'info' => 'Parameter Refinement!<BR>'
				 }
			     );
	foreach my $file (sort keys %data) {
	    foreach my $label (keys %{$data{$file}{'__DATA__'}}) {
		$data_done{$file}{$label}{$out_header}++;
		$data_column_available{$file}{$label}{$out_header}++;
		$data_done{$file}{$label}{$dehub_header}++;
		$data_column_available{$file}{$label}{$dehub_header}++;
	    }
	}
	return;
    }

    # check if multiple files are submitted
    # and the global normalization is requested:

    my @files = ();
    foreach (sort keys %data) {
	if (/\w/) {
	    push @files, $_;
	}
    }

    if ($global eq 'global') {
	
	# carry out normalization on all of them at once:
	print &vsn_normalization($action_num,$action,\@files,$header,$FLAG_CHANNEL_TERM,$out_header,$dehuberize,$dehub_header,$iterations,@flags);

    } else {

	# carry out action on each file separately
	foreach my $file (@files) {
	    my @tmp = ($file);
	    print &vsn_normalization($action_num,$action,\@tmp,$header,$FLAG_CHANNEL_TERM,$out_header,$dehuberize,$dehub_header,$iterations,@flags);
	}
    }
}


sub vsn_normalization {
    my $action_num = shift;
    my $action = shift;
    my $ref = shift;
    my @files = @$ref;
    my $header = shift;
    my $flag_header = shift;
    my $norm_header = shift;
    my $dehuberize = shift;
    my $dehub_header = shift;
    my $iterations = shift;
    my @flags = (@_);


    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my %skip_flags = ();
    %skip_flags = &get_skip_flags(@flags);

    # read in data if necessary
    foreach my $file (@files) {
	my $rc = &read_data($file,$header,$flag_header);
	unless ($rc eq 'ok') {
	    return "<BR><B><FONT COLOR=RED>WARNING: $rc</FONT></B><BR>\n";
	}
    }

    # write data to file
    my $file_tmp = basename $files[0];
    $file_tmp =~ s/\//_/g;
    $file_tmp =~ s/ /_/g;
    my $data_file = "$tmp_dir/data_$action_num.$action_step.$file_tmp.$header.$PID.tmp";
    my $norm_file = "$tmp_dir/norm_$action_num.$action_step.$file_tmp.$header.$PID.tmp";
    my $dehub_file = "$tmp_dir/dehub_$action_num.$action_step.$file_tmp.$header.$PID.tmp";
    my $params_file = "$tmp_dir/params_$action_num.$action_step.$file_tmp.$header.$PID.tmp";         
     
    if (-e $data_file) {
	unlink $data_file 
	    or return "Can't delete data file '$data_file': $!\n";
    }

    if (-e $norm_file) {
	unlink $norm_file 
	    or return "Can't delete normalized file '$norm_file': $!\n";
    }

    if (-e $dehub_file) {
	unlink $dehub_file 
	    or return "Can't delete dehub file '$dehub_file': $!\n";
    }

    if (-e $params_file) {
	unlink $params_file 
	    or return "Can't delete params file '$dehub_file': $!\n";
    }

    open (OUT, ">$data_file")
	  or die "Can't write to $data_file: $!";


    my ($mrow,$mcol,$row,$col);
    my %tmp = ();
    my %labels = ();
    my @labels = ();
    my %lines = ();
    my $skipped = 0;
    my %rows = ();
    my %mrows = ();
    my %cols = ();
    my %mcols = ();
    my %skip_row = ();
    my %skip_lines = ();

    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs(@files);
    # collect values for each spot label
    foreach my $file (@files) {

	next if ($file eq '');

	foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	    next if ($label eq '__NO_LABEL__');
	    
	    # collect labels for later
	    $labels{$label}++;
	    
	    OUTER2_TMP:foreach $mrow (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}}) {
		$mrows{$mrow}++;
		foreach $mcol (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}}) {
		    $mcols{$mcol}++;
		    foreach $row (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}}) {
			$rows{$row}++;
			foreach $col (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}}) {
			    $cols{$col}++;
			    
			    my $value;

			    # skip spots without value
			    unless (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}) {
				$value = '';
			    }
			    
			    # get value
			    $value = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};

			    # skip flagged values if requested
			    my $skip = 0;
			    if (%skip_flags
				and
				(defined $data{$file}{'__DATA__'}{$label}{$flag_header}{$mrow}{$mcol}{$row}{$col}
				or
				defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$flag_header}{$mrow}{$mcol}{$row}{$col})) {
				my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$flag_header}{$mrow}{$mcol}{$row}{$col});
				if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$flag_header}{$mrow}{$mcol}{$row}{$col}) {
				    push @flags, &get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$flag_header}{$mrow}{$mcol}{$row}{$col});
				}

				foreach (@flags) {
				    if (defined $skip_flags{$_}) {
					$skip = 1;
					last;
				    }
				}			
			    } 
			    
			    
			    # calculate line number
			    my $line = ($row - 1 + ($mrow - 1) * $ROW) * $MCOL*$COL + ($mcol - 1) * $COL + $col;

			    # keep track of conversion
			    $lines{$line} = "$mrow,$mcol,$row,$col";

			    if ($skip) {
				$skip_lines{$line}++;
				next;
			    }

			    unless ($value =~ /\d/) {
				$skip_lines{$line}++;
				next;
			    }

			    if ($header =~ /_L/) {
				# reverse log if we get data from previous normalization
				# presume that log2 was used
				$value = 2**$value;
			    }

			    push @{$tmp{$line}}, $value;
			}
		    }
		}
	    }
	}
    }

    @labels = sort keys %labels;
	
    my $input_rows = 0;
    my @tmp2 = ();
    my $columns = @files * @labels;

    # second check - make sure all rows have the right amount of data points:
    foreach my $line (sort { $a <=> $b } keys %tmp) {

	# skipped already
	next if (defined $skip_lines{$line});

	@tmp2 = @{$tmp{$line}};
	unless (@tmp2 == $columns) {
	    # one or more values missing (due to flagging)
#	    warn "#WARNING: only ".(scalar @tmp2)." values for $line!\n";
	    $skip_lines{$line}++;
	    next;
	}

	# print those values connected by TAB
	print OUT (join "\t", $line,@tmp2)."\n";
	$input_rows++;
    }
    close OUT;

    foreach my $file (@files) {
	foreach my $line (keys %skip_lines) {
	    ($mrow,$mcol,$row,$col) = split /,/, $lines{$line};
	    foreach my $label (@labels) {
		# initialize field that was skipped
		$data{$file}{'__DATA__'}{$label}{$norm_header}{$mrow}{$mcol}{$row}{$col} = '';
		$data{$file}{'__DATA__'}{$label}{$dehub_header}{$mrow}{$mcol}{$row}{$col} = '';
	    }
	}
    }
    
    # add line number to columns
    $columns++;

    # create R commands
    my $R_commands_file = "$tmp_dir/R_cmd.$action.$action_num.$action_step.$PID";
    my $R_out = "$tmp_dir/R_out.$action.$action_num.".(basename $file_tmp).".$action_step.$PID";
    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";
    print RH "library(vsn)\n";
    print RH "postscript(\"$R_out\",horizontal=FALSE)\n";

    # read in data
    print RH "data = read.table(\"$data_file\", header=F, sep=\"\\t\")\n";
    print RH "pool <- cbind(";
    for (my $i = 2; $i < $columns; $i++) {
	print RH "data[,$i],";
    }
    print RH "data[,$columns])\n";

    # Write vsn call and process table.
    print RH "## load vsn library\n";
    print RH "## feed data to vsn\n";
    print RH "n = vsn(pool, niter=$iterations)\n";

    # Write the write-to-file statement in R to save data in "$norm_file".
    print RH "## write data to file\n";
    print RH "out <- cbind(data[,1]";
    for (my $i = 1; $i < $columns; $i++) {
	print RH ",exprs(n)[,$i]";
    }
    print RH ")\nwrite.table(out, \"$norm_file\", sep=\"\\t\", quote=F, col.names=F, row.names=F)\n";

    # get dehuberized values:
    # big thanks to Dana Aeschliman for help with this!
    if ($dehuberize > 0) {
	print RH "params = preproc(description(n))\$vsnParams\n";
	print RH "offs=matrix(params[1:ncol(pool)],nrow=nrow(pool),ncol=ncol(pool),byrow=TRUE)\n";
	print RH "facs=matrix(params[(ncol(pool)+1):(2*ncol(pool))],nrow=nrow(pool),ncol=ncol(pool),byrow=TRUE)\n";
	print RH "a = offs[1,1]\n";
	print RH "b = facs[1,1] \n";
	print RH "dehub=sinh(exprs(n) + log(2*b))\n";
	print RH "dehub=(dehub-a)/b\n";
	print RH "out <- cbind(data[,1]";
	for (my $i = 1; $i < $columns; $i++) {
	    print RH ",dehub[,$i]";
	}
	print RH ")\nwrite.table(out, \"$dehub_file\", sep=\"\\t\", quote=F, col.names=F, row.names=F)\n";
    }

#    print RH "params <- preproc(description(n))\$vsnParams\n";
#    print RH "write.table(params, \"$params_file\", sep=\"\\t\", quote=F, col.names=F, row.names=F)\n";

    # define 2x2 output panel
    print RH "par(mfrow = c(2,2))\n";

    # plot raw and normalized data
#    print RH "plot(log.na(pool[,1],2),log.na(pool[,2],2), pch = \".\", main = \"raw logs\")\n";
#    print RH "abline(0,1)\n";
#    print RH "plot(exprs(n), pch = \".\", main = \"vsn normalized\")\n";
#    print RH "abline(0,1)\n";

    # plot control (should produce a horizontal line)
    print RH "meanSdPlot(n, ranks = TRUE, main = \"mean Stdev\")\n";

    # calculate ratios
    if ($columns <= 3) {
	# only data from one file were inserted
	print RH "M <- exprs(n)[,2] - exprs(n)[,1]\n";
    } else {
	for (my $i = 1; $i < $columns-1; $i +=2) {
	    print RH "M$i <- exprs(n)[,".($i+1)."] - exprs(n)[,$i]\n";
	}
	print RH "M <- c(";
	for (my $i = 1; $i < $columns-1; $i +=2) {
	    print RH "M$i";
	    print RH ',' unless ($i + 2 >= $columns - 1);
	}
	print RH ")\n";
    }

    # draw histogram    
    print RH "hist(M, breaks = 50, col=\"#d95f0e\", labels = FALSE, xlab = \"normalized $header ratios\", main = \"g-log ratios (Ch2/Ch1) after vsn\")\n";
#    print RH "title(main = \"glog-ratios for ".(basename $file)." after vns\", line=-1)\n";

    # plot extimated calibration and variance staibilization parameters
    print RH "vsnPlotPar(n, \"offsets\")\n";
    print RH "vsnPlotPar(n, \"factors\")\n";
    
    print RH "mtext(\"all_files\",side=3,line=-1.5,font=2,outer=TRUE)\n";

    print RH "dev.off()\n";

    # this calculation might take a while, so we write out a file as a signal
    # that the process has finished
    my $signal_file = $norm_file.'.finished';
    if (-e $signal_file) {
	unlink $signal_file
	    or return "Can't delete signal file '$signal_file': $!\n";
    }

    print RH "out <- cbind(1,2,3,4,5,6,7,8,9)\n";
    print RH "write.table(out, \"$signal_file\")\n";
    
    close RH;

    my $image = &get_R_output($R_commands_file, $R_out, $action_num, 'all_files', $signal_file);
    
#    my $image_lnk = &get_R_output($R_commands_file, $R_out, $action_num, 'all_files');

    # read in all parameters
#    unless (-s $params_file) {
#	return "<B>WARNING: no output from normalization!\n";
#    }
#    open (IN, $params_file)
#	or die "Can't read $params_file: $!";
#    
#    my @fac = ();
#    my @offs = ();
#    my $counter = 1;
#    my $entries = @files * @labels;
#    while (<IN>) {
#	chomp;
#	next unless (/\d/);
#	if ($counter > $entries) {
#	    push @fac, $_;
#	} else {	    
#	    push @offs, $_;
#	}
#	$counter++;
#    }
#    close IN;
#    

    # read normalized and dehuberized data    

    my %out = ();
    my %dehub = ();

    unless (-s $norm_file) {
	return "<B>WARNING: no output from normalization!</B><BR>\n";
    }

    open (IN, $norm_file)
	or die "Can't read $norm_file: $!";

    while (<IN>) {
	chomp;
	my ($line,@tmp2) = split /\t/, $_;
	# get coordinates
	($mrow,$mcol,$row,$col) = split /,/, $lines{$line};
	@{$out{$mrow}{$mcol}{$row}{$col}} = @tmp2;
    }
    close IN;

    if ($dehuberize) {
	unless (-s $dehub_file) {
	    return "<B>WARNING: no output from dehuberization!</B>\n";
	}
	
	open (IN2, $dehub_file)
	    or die "Can't read $dehub_file: $!";
	
	while (<IN2>) {
	    chomp;
	    my ($line,@tmp2) = split /\t/, $_;
	    # get coordinates
	    ($mrow,$mcol,$row,$col) = split /,/, $lines{$line};
	    @{$dehub{$mrow}{$mcol}{$row}{$col}} = @tmp2;
	}
	close IN2;
    }

    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	unlink $data_file;
	unlink $norm_file;
	unlink $dehub_file if ($dehuberize);
    }

    $counter = 0;
    %tmp = ();
    foreach $mrow (1..$MROW) {
	foreach $mcol (1..$MCOL) {
	    foreach $row (1..$ROW) {
		foreach $col (1..$COL) {

		    $counter++ if (defined @{$out{$mrow}{$mcol}{$row}{$col}});

		    my $params = 0;
		    foreach my $file (@files) {
			
			foreach my $label (@labels) {
			    
			    my $value = '';
			    my $dehub = '';
#			    my ($offset,$factor) = ($offs[$params],$fac[$params]);
			    if (defined @{$out{$mrow}{$mcol}{$row}{$col}}) {
				$value = shift @{$out{$mrow}{$mcol}{$row}{$col}};
#				$dehub = &sinh($value);
#				my $raw = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
#				$dehub = $offset + $raw * $factor;
			    }

			    next if ($value eq '');
			    
			    # write back values for each spot label
			    # (round if necessary)
#			    if ($value =~ /\.\d{5,}/) {
#				$value = sprintf "%.5f", $value;
#			    }
			    $data{$file}{'__DATA__'}{$label}{$norm_header}{$mrow}{$mcol}{$row}{$col} = $value;
			    $params++;


			    if (defined @{$dehub{$mrow}{$mcol}{$row}{$col}}) {
				$dehub = shift @{$dehub{$mrow}{$mcol}{$row}{$col}};
			    }

			    next if ($dehub eq '');
			    
			    $data{$file}{'__DATA__'}{$label}{$dehub_header}{$mrow}{$mcol}{$row}{$col} = $dehub;
			}
		    }
		}
	    }
	}
    }

    close IN;
		
    # take note of available columns for later

    foreach my $file (@files) {
	foreach my $label (keys %{$data{$file}{'__DATA__'}}) {
	    next if ($label eq '__NO_LABEL__');

	    unless (defined $data_column_available{$file}{$label}{$norm_header}) {
		$data_column_available{$file}{$label}{$norm_header} = 1;
		push @{$data_column_available{$file}{$label}{'__ORDER__'}}, $norm_header;
		$data_column_available{$file}{$label}{$dehub_header} = 1 if ($dehuberize);
		push @{$data_column_available{$file}{$label}{'__ORDER__'}}, $dehub_header if ($dehuberize);
	    }

	    $data_done{$file}{$label}{$norm_header}++;
	    $data_done{$file}{$label}{$dehub_header}++ if ($dehuberize);
	}
    }


    # format output (include image and choice for parameters)
    my $image_lnk = &format_output( # 'vsn_',$action_num,'','-',$action,$image,''
				   {
				       'prefix' => 'vsn_',
				       'num' => $action_num,
				       'label' => '-',
				       'action' => $action,
				       'image' => $image,
				       }
				   );
    
    unless ($counter == $input_rows) {
	return "<B>WARNING: only $counter rows (instead of $input_rows)  returned from normalization!</B>\n";
    }

    return $image_lnk;
}


sub get_skip_flags {
    my @flags = @_;
    my %skip_flags = ();
    my %no_lists = ();
    foreach (@flag_names) {
	next if (/list/);
	$no_lists{$_}++;
    }
    foreach (@flags) {
	if ($_ eq 'none') {
	    %skip_flags = ();
	    last;
	}
	if ($_ eq 'all but lists') {
	    foreach (keys %flag_values_num) {
		if ($_ > 0) {
#		    $skip_flags{$_}++ unless ($flag_values_num{$_} =~ /^list/);
		    $skip_flags{$_}++ if (defined $no_lists{$flag_values_num{$_}});
		}
	    }
	}
	if ($_ eq 'all but lists and floor') {
	    foreach (keys %flag_values_num) {
		if ($_ > 0) {
		    next if ($flag_values_num{$_} eq 'floor');
#		    $skip_flags{$_}++ unless ($flag_values_num{$_} =~ /^list/);
		    $skip_flags{$_}++ if (defined $no_lists{$flag_values_num{$_}});
		}
	    }
	}
	if ($_ eq 'all') {
	    foreach (keys %flag_values_num) {
		if ($_ > 0) {
		    $skip_flags{$_}++;
		}
	    }
	    last;
	}
	if (defined $flag_values{$_}) {
	    $skip_flags{$flag_values{$_}}++;
	} elsif ($_ eq 'unflagged'
		 or
		 $_ eq 'no flag') {
	    $skip_flags{0}++;
	} else {
	    warn "# $progname: no flag '$_' known!" unless ($_ eq ''
							    or
							    $_ =~ /all but /
							    );
	}
    }

    return %skip_flags;
}


sub sinh {
    my($x) = exp shift;
    return ($x - 1 / $x) / 2;
}

sub flag_markers {
    my $action_num = shift;
    my $file = shift;
    my $action = shift;
    my $spot_label = shift;
    my $flag_header = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    $last_column_used = $spot_label;

    # check for user specific settings for marker flag
     my $marker_flag = &get_user_settings($var_def_default{'marker_flag'},'marker_flag',$force_global_opt,$action_num,$file);
    $marker_flag = $flag_values{$marker_flag} if ($marker_flag =~ /\D/);

    my @markers = &get_user_settings($var_def_default{'markers'},'markers',$force_global_opt,$action_num,$file);
    unless (@markers) {
	return "ERROR: No markers specified!<BR>\n";
    }
    if (grep /all/, @markers) {
	@markers = ();
	foreach my $marker (@{$var_def_values{'markers'}}) {
	    next if ($marker eq 'all');
	    push @markers, $marker;
	}
    }

    my %markers = ();
    foreach (@markers) {
	$markers{$_}++;
    }

    my %flagged = ();

    # read in data if necessary
    my $rc = &read_data($file,$spot_label);
    unless ($rc eq 'ok') {
	return "<BR><B><FONT COLOR=RED>WARNING: $rc</FONT></B><BR>\n";
    }

    my ($mrow,$mcol,$row,$col);

    foreach $mrow (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$spot_label}}) {
	foreach $mcol (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$spot_label}{$mrow}}) {
	    foreach $row (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$spot_label}{$mrow}{$mcol}}) {
		foreach $col (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{'__NO_LABEL__'}{$spot_label}{$mrow}{$mcol}{$row}}) {
		    
		    my $spot = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$spot_label}{$mrow}{$mcol}{$row}{$col};
#		    if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$spot_label}{$mrow}{$mcol}{$row}{$col}) {
#			$spot = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$spot_label}{$mrow}{$mcol}{$row}{$col};
#		    }
		    
		    # special settings for Arrayvision:
		    # delete numerator placed in brackets behind each gene name
		    $spot =~ s/ \(\d+\)$// if (defined $file_specs{$file}{'FORMAT'} 
					       and
					       $file_specs{$file}{'FORMAT'} eq 'ArrayVision');

		    # remove trailing white space
		    $spot =~ s/\s+$//;

		    # special settings for Arrayvision:
		    # remove description from id
		    $spot =~ s/^\^+// if (defined $file_specs{$file}{'FORMAT'} 
					  and
					  $file_specs{$file}{'FORMAT'} eq 'ArrayVision');
		    ($spot) = split /\^/, $spot, 2 if ($FPMI_specific
						       and 
						       $spot =~ /\^/);
		    unless ($spot =~ /\w/) {
			$spot = 'no label';
		    }
		    
		    if ($markers{$spot}) {
			foreach my $label (keys %{$data{$file}{'__DATA__'}}) {
			    next if ($label eq '__NO_LABEL__');
			    $data{$file}{'__DATA__'}{$label}{$flag_header}{$mrow}{$mcol}{$row}{$col} += $marker_flag;
			}
			$flagged{$spot}++;
		    }
		}
	    }
	}
    }
    # print report
    my $total = 0;
    my $out = "
     <TABLE BORDER=1 CELLPADDING=4>
      <TR>
       <TH>marker</TH><TH>#</TH>
      </TR>\n";
    foreach (sort { $flagged{$b} <=> $flagged{$a} } keys %flagged) {
	$out .= "
      <TR>
       <TD>$_</TD><TD ALIGN=RIGHT><B>$flagged{$_}</B></TD>
      </TR>\n";
	$total += $flagged{$_};
    }
    $out .= "
      <TR>
       <TD>total:</TD><TD ALIGN=RIGHT><B>$total</B></TD>
      </TR>
      </TABLE>\n";

    unless ($total) {
	$out = 'No markers were detected!';
    }
    
    return $out;
}


sub dup_spot_flagging {
    my $action_num = shift;
    my $file = shift;
    my $action = shift;
    my $spot_label = shift;
    my $flag_header = shift;

    if ($refine_parameters) {
	return 'Paramter Refinement!';
    }

    $last_column_used = $spot_label;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $header = '';
    foreach my $column (@{$item_data_column{$action}}) {
	if ($data_column_available{$file}{'__ALL__'}{$column}) {
	    $header = $column;
	    last;
	}
    }
    # check for specific settings:
    $header = &get_user_settings($header,'flag_dups_data_column',$force_global_opt,$action_num,$file,'__ALL__');

    unless ($header) {
	return "Error: No data column defined to work on for $action!\n";
    }

    # default dup unit to use:
    my $dup_unit = $var_def_default{'dup_unit'};
    $dup_unit = &get_user_settings($dup_unit,'dup_unit',$force_global_opt,$action_num,$file,'__ALL__');
    
    # default flag to use:
    my $dup_flag = $var_def_default{'dup_flag'};

    # check for user specific settings
    $dup_flag = &get_user_settings($dup_flag,'dup_flag',$force_global_opt,$action_num,$file,'__ALL__');
    $dup_flag = $flag_values{$dup_flag} if ($dup_flag =~ /\D/
					    and
					    defined $flag_values{$dup_flag});

    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'dup_skip_flags'},'dup_skip_flags',$force_global_opt,$action_num,$file,'__ALL__');
    %skip_flags = &get_skip_flags(@flags);

   # default dup_cutoff to use:
    my $dup_cutoff = $var_def_default{'dup_cutoff'};

    # check for user specific settings
    $dup_cutoff = &get_user_settings($dup_cutoff,'dup_cutoff',$force_global_opt,$action_num,$file,'__ALL__');

    # read in data if necessary
    my $rc = &read_data($file,$spot_label,$header);
    unless ($rc eq 'ok') {
	return "<BR><B><FONT COLOR=RED>WARNING: $rc</FONT></B><BR>\n";
    }

    my @labels = ();
    foreach my $label (keys %{$data{$file}{'__DATA__'}}) {
	next if ($label eq '__NO_LABEL__');
	push @labels, $label;
    }

    my ($mrow,$mcol,$row,$col);
    my $flag_pos = 0;
    my %flag = ();
    my %ch_flag = ();
    my %ratios = ();
    my $found_dups = 0;
    my %dup_positions = ();

   # collect values for each spot label
    my @differences = ();
    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);
    my $CELLS = $ROW*$COL;
    foreach my $mrow (1..$MROW) {
	foreach my $mcol (1..$MCOL) {
	    # the next is necessary to catch those duplicated neighbours that swap over into the next row
	    COL:foreach my $col_all (1..$CELLS) {
		my $row = int(($col_all - 1) / $COL) + 1;
		my $col = $col_all - ($row - 1) * $COL;
			    
		my %done = ();

		foreach my $label (@labels) {
		    # initialize all data output for DUP_DIFF
		    unless (defined $data{$file}{'__DATA__'}{$label}{'DUP_DIFF'}{$mrow}{$mcol}{$row}{$col}) {			
			$data{$file}{'__DATA__'}{$label}{'DUP_DIFF'}{$mrow}{$mcol}{$row}{$col} = '';   
		    }
		}
		
		next if ($done{$mcol}{$col});
		my %values = ();
		my ($flag,$skip,$spot,$next_spot,$val1,$val2);
		
		# get current spot label:
		if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$spot_label}{$mrow}{$mcol}{$row}{$col}) {
		    $spot = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$spot_label}{$mrow}{$mcol}{$row}{$col};
		    
		    # special settings for Arrayvision:
		    # delete numerator placed in brackets behind each gene name
		    $spot =~ s/ \(\d+\)$// if (defined $file_specs{$file}{'FORMAT'} 
					       and
					       $file_specs{$file}{'FORMAT'} eq 'ArrayVision');
		} else {
		    $spot = '';
		}

		if ($row == $ROW
		    and 
		    $col == $COL) {
		    next;
		}
		    
		my $col_next = $col + 1;
		my $row_next = $row;
		if ($col_next > $COL) {
		    $row_next++;
		    $col_next = 1;
		}
		
		# get next spot label:
		if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$spot_label}{$mrow}{$mcol}{$row_next}{$col_next}) {
		    $next_spot = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$spot_label}{$mrow}{$mcol}{$row_next}{$col_next};
		    
		    # special settings for Arrayvision:
		    # delete numerator placed in brackets behind each gene name
		    $next_spot =~ s/ \(\d+\)$// if (defined $file_specs{$file}{'FORMAT'} 
						    and
						    $file_specs{$file}{'FORMAT'} eq 'ArrayVision');
		} else {
		    $next_spot = '';
		}
		
		unless ($spot eq $next_spot) {
		    # mismatch of spot labels,
		    # skip to next;
		    next;
		} else {
		    # mark next column as done
		    $done{$row_next}{$col_next}++;
		}
		
		$flag = 0;
		
		# get values for each label on current column:
		
		foreach my $label (@labels) {
		    
		    # skip spots where a value is missing
		    unless (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}) {
			next;
		    }
		    
		    # check for flags that might have to be skipped:
		    # get flags 
		    my @flags = ();
		    if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			@flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
		    }
		    if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			push @flags, (&get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}));
		    }
		    
		    # skip flagged spots
		    foreach (@flags) {
			if (defined $skip_flags{$_}) {
			    $flag++;
			    last;
			}
		    }
		    
		    if ($flag) {
			next COL;
		    }
		    
		    
		    my $value = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
		    
		    push @{$values{$label}}, $value;
		}
		
		$flag = 0;
		
		# get values for each label on next column:
		
		foreach my $label (@labels) {
		    
		    # skip spots where a value is missing
		    unless (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row_next}{$col_next}) {
			next;
		    }
		    
		    # check for flags that might have to be skipped:
		    # get flags 
		    my @flags = ();
		    if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row_next}{$col_next}) {
			@flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row_next}{$col_next});
		    }
		    if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row_next}{$col_next}) {
			push @flags, (&get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row_next}{$col_next}));
		    }
		    
		    # skip flagged spots
		    foreach (@flags) {
			if (defined $skip_flags{$_}) {
			    $flag++;
			    last;
			}
		    }
		    
		    if ($flag) {
			next COL;
		    }
		    
			
		    my $value = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row_next}{$col_next};
		    
		    push @{$values{$label}}, $value;
		}
		
		# check for high ratio in either channel
		my $high_ratio = 0;
		foreach my $label (@labels) {
		    unless (defined @{$values{$label}}) {
			next;
		    }
		    if (@{$values{$label}} > 1) {
			# get the min and max
			my (@tmp) = sort { $a <=> $b } @{$values{$label}};
			my ($low,$high) = ($tmp[0],$tmp[-1]);
			
			if ($low < 0
			    and 
			    $high > 0) {
			    # special dealing with numbers that are on both sides of zero
			    if (abs($low) > abs($high)) {
				my $low_tmp = $high;
				$high = abs($low) + $high;
				$low = $low_tmp;
			    } else {
				$low = abs($low);
				$high += $low;
			    }
			} elsif ($low < 0
				 and
				 $high < 0) {
			    ($low,$high) = ($high,$low);
			}
			
			# calculate ratio
			my $dup_difference;
			if ($dup_unit =~ /abs/) {
			    $dup_difference = abs($high - $low);
			} else {
			    if ($low == 0
				or
				$high == 0) {
				$dup_difference = 0;
				warn "#WARNING: can't calculate ratio for $spot: ($low,$high)!";
			    } else {
				$dup_difference = $high / $low;
			    }
			}
			
			# store dup_difference in data file
			$data{$file}{'__DATA__'}{$label}{'DUP_DIFF'}{$mrow}{$mcol}{$row}{$col} = $dup_difference;
			$data{$file}{'__DATA__'}{$label}{'DUP_DIFF'}{$mrow}{$mcol}{$row_next}{$col_next} = $dup_difference;
			
			$found_dups++;
			$dup_positions{$mrow}{$mcol}{$row}{$col} = "$mrow|$mcol|$row_next|$col_next";
			push @differences, $dup_difference;
			
			unless ($dup_unit =~ /stddev/i) {
			    # we don't need to know the standard deviation,
			    # just flag spots here:
			    
			    if ($dup_difference > $dup_cutoff
				or
				$dup_difference == 0) {
				$high_ratio++;
				$ch_flag{$label}++;
			    }
			    
			    # keep track of (rounded) dup_differences
#			    $dup_difference = sprintf "%.1f", $dup_difference;
			    # changed to simple cut-off, so that, for example,
			    # 1.49 does not appear in 1.5 bin
			    $dup_difference =~ s/\.(\d).*/\.$1/;
			    $ratios{$dup_difference}++;
			}
		    }
		}
		unless ($dup_unit =~ /stddev/i) {
		    if ($high_ratio > 0) {
			
			if (defined $flag{"$mrow|$mcol|$row|$col"}) {
			    warn "col '$col' is flagged already!";
			}
			
			$flag{"$mrow|$mcol|$row|$col"}++;
			
			$flag_pos += 2;
			# flag all spots:			
			foreach my $label (keys %{$data{$file}{'__DATA__'}}) {
			    next if ($label eq '__NO_LABEL__');
			    $data{$file}{'__DATA__'}{$label}{$flag_header}{$mrow}{$mcol}{$row}{$col} += $dup_flag;
			    $data{$file}{'__DATA__'}{$label}{$flag_header}{$mrow}{$mcol}{$row_next}{$col_next} += $dup_flag;
			}
		    }
		}
	    }
	}
    }

    # we have now collected all differences and can calculate std deviations
    my $stdev = '?';
    if (@differences > 1) {
	$stdev = &stdev(@differences);
    }
    my $mean = '?';
    if (@differences) {
	$mean = &mean(@differences);
    }
    my $median = '?';
    if (@differences) {
	$median = &median(@differences);
    }
		       
    if ($dup_unit =~ /stddev/i) {
	foreach my $mrow (1..$MROW) {
	    foreach my $mcol (1..$MCOL) {
		foreach my $row (1..$ROW) {		    
		    foreach my $col (1..$COL) {

			next unless ($dup_positions{$mrow}{$mcol}{$row}{$col});
			
			my $dup_difference = $data{$file}{'__DATA__'}{'CH1'}{'DUP_DIFF'}{$mrow}{$mcol}{$row}{$col};
			
			# check for high ratio in either channel
			my $high_ratio = 0;
			foreach my $label (@labels) {
			    my $dup_difference = $data{$file}{'__DATA__'}{$label}{'DUP_DIFF'}{$mrow}{$mcol}{$row}{$col};
			    $dup_difference = abs($median - $dup_difference);
			    $dup_difference /= $stdev;
			    if ($dup_difference >= $dup_cutoff) {
				$high_ratio++;
				$ch_flag{$label}++;
			    }
			    
			    # keep track of (rounded) dup_differences
#			    $dup_difference = sprintf "%.1f", $dup_difference;
			    # changed to simple cut-off, so that, for example,
			    # 1.49 does not appear in 1.5 bin
			    $dup_difference =~ s/\.(\d).*/\.$1/;
			    $ratios{$dup_difference}++;
			}
			
			
			if ($high_ratio > 0) {
			    
			    if (defined $flag{"$mrow|$mcol|$row|$col"}) {
				warn "col '$col' is flagged already!";
			    }
			    
			    $flag{"$mrow|$mcol|$row|$col"}++;
			    
			    $flag_pos += 2;
			    # flag all spots:			
			    foreach my $label (keys %{$data{$file}{'__DATA__'}}) {
				next if ($label eq '__NO_LABEL__');
				$data{$file}{'__DATA__'}{$label}{$flag_header}{$mrow}{$mcol}{$row}{$col} += $dup_flag;
				my ($mrow_plus,$mcol_plus,$row_plus,$col_plus) = split /\|/, $dup_positions{$mrow}{$mcol}{$row}{$col};
				$data{$file}{'__DATA__'}{$label}{$flag_header}{$mrow_plus}{$mcol_plus}{$row_plus}{$col_plus} += $dup_flag;
			    }
			}
		    }
		}
	    }
	}
    }

    # indicate available data fields
    foreach my $label (@labels) {
	$data_done{$file}{$label}{'DUP_DIFF'}++;
	$data_column_available{$file}{$label}{'DUP_DIFF'}++;
    }

   # print report
    my $out = '';
    
    my $round = 1;
    if ($mean =~ /\.(\d+)/) {
	if (length($1) > $round) {
	    $mean = sprintf "%.${round}f", $mean;
	}
    }
    if ($median =~ /\.(\d+)/) {
	if (length($1) > $round) {
	    $median = sprintf "%.${round}f", $median;
	}
    }
    if ($stdev =~ /\.(\d+)/) {
	if (length($1) > $round) {
	    $stdev = sprintf "%.${round}f", $stdev;
	}
    }

    my $effective_cutoff = $dup_cutoff;
    if ($dup_unit =~ /stddev/i) {
	$effective_cutoff = $median + $dup_cutoff * $stdev;
    }
    $found_dups = int($found_dups/2);
    $out .= "<B>$found_dups</B> pairs of neighbouring duplicate spots found.<BR>\n"
	. "mean: <B>$mean</B><BR>median: <B>$median</B><BR>StdDev: <B>$stdev</B><BR>\n"
	. "With ".($dup_unit =~ /abs/ ? 'absolute difference' : 'fold difference')
	. " > <B>$dup_cutoff</B>"
	. ($dup_unit =~ /stddev/i ? " standard deviations from the median (> <B>$effective_cutoff</B>)" : '')
	. "<BR><B>".(scalar keys %flag)."</B> reporters ($flag_pos spots) were flagged in total";
    my $ch_sum = 0;
    foreach my $ch (sort keys %ch_flag) {
	my $ch_out = '';
	if ($ch eq 'CH1') {
	    if (defined $file_specs{$file}{'CH1_LABEL'}) {
		$ch_out = " ($file_specs{$file}{'CH1_LABEL'})";
	    }
	} elsif ($ch eq 'CH2') {
	    if (defined $file_specs{$file}{'CH2_LABEL'}) {
		$ch_out = " ($file_specs{$file}{'CH2_LABEL'})";
	    }
	}
	
	$out .= ", <B>$ch_flag{$ch}</B> in $ch$ch_out";
	$ch_sum += $ch_flag{$ch};
    }
    if ($ch_sum > (scalar keys %flag)) {
	$out .= " (some spots were flagged due to large differences in both channels)";
    }
    $out .= ".<BR><BR><B>Top fold differences (for both channels):</B><BR>\n<TABLE BORDER=1>\n<TR ALIGN=CENTER>\n<TH>"
	. ($dup_unit =~ /stddev/i ? ' StdDev ' : '')
	. "difference</TH>\n";
    my $counter = 0;
    my $ratio2 = sprintf "%.1f", $dup_cutoff;
    my @num = ();
    my @acc = ();
    my $acc = 0;
    my @ratios = (sort { $b <=> $a } keys %ratios);
    my $coloured = 0;
    my $counter2 = $#ratios;
   for (my $i = 0; $i <= $#ratios; $i++) {
	my $ratio = $ratios[$i];
#	$out .= "<TR>\n<TD ALIGN=RIGHT>$counter</TD><TD>$ratios{$ratio}</TD><TD>$ratio</TD></TR>\n";
	if ($coloured == 0
	    and
	    defined $ratios[$i+1]
	    and 
	    $ratios[$i+1] < $ratio2) {
	    $counter2 = $counter;
	    $out .= "<TD BGCOLOR=RED>$ratio</TD>\n";
	    $coloured = 1;
	} else {
	    $out .= "<TD>$ratio</TD>\n";
	}
	push @num, $ratios{$ratio};
	$acc += $ratios{$ratio};
	push @acc, $acc;
	$counter++;
    }

    $out .= "</TR>\n<TR ALIGN=RIGHT>\n<TH>occurrence</TH>\n";    
    $counter = 0;
    foreach (@num) {
	if ($counter == $counter2) {
	    $out .= "<TD BGCOLOR=RED>$_</TD>\n";
	} else {
	    $out .= "<TD>$_</TD>\n";
	}
	$counter++;
    }

    $out .= "</TR>\n<TR ALIGN=RIGHT>\n<TH>accumulated</TH>\n";    
    $counter = 0;
    foreach (@acc) {
	if ($counter == $counter2) {
	    $out .= "<TD BGCOLOR=RED>$_</TD>\n";
	} else {
	    $out .= "<TD>$_</TD>\n";
	}
	$counter++;
    }
    $out .= "</TR>\n</TABLE><BR><BR>\n";

    # add histogram of dup_differences
    my $file_tmp = basename $file;
    $file_tmp =~ s/\//_/g;
    $file_tmp =~ s/ /_/g;
    my $output_file = "$tmp_dir/data.$file_tmp.$action.$action_num.$action_step.$PID";
    my $R_commands_file = "$tmp_dir/R_cmd.$file_tmp.$action.$action_num.$action_step.$PID";

    if (open OUT, ">$output_file"
	and
	open RH, ">$R_commands_file"
	) {
		
	# write dup_differences to file:
	foreach (@differences) {
	    print OUT "$_\n";
	}
	close OUT;

	# write R commands:
	my $R_out = $output_file.".ps";
	open (RH, ">$R_commands_file")
	    or die "Can't write to $R_commands_file: $!";
	print RH "postscript(\"$R_out\",horizontal=FALSE, width = ".($IMAGE_HEIGHT/72).", height = ".($IMAGE_WIDTH/72).", paper = \"special\")\n";
	
	# read in data
	print RH "fm <- read.table(\"$output_file\", sep = \"\\t\")\n";
	
	print RH "hist(fm\$V1, col=\"gray\", labels = TRUE, xlab = \"DUP_DIFF (".($dup_unit =~ /abs/ ? 'absolute' : 'fold').")\", main = \"\", breaks=100)\n";

	print RH "title(main = \"differences in duplicated neighbours for $file_tmp\", line=-1)\n";
	print RH "dev.off()\n";
	close RH;
	
	my $image = &get_R_output($R_commands_file, $R_out, '', '', '', 'landscape');
    
	if ($clean_up_files) {
	    unlink $R_commands_file;
	    unlink $R_out;
	    unlink $output_file;
	}

	$image = basename($image);
	$out .= "<IMG SRC=\"$image\">\n<BR><BR>\n";	
    } else {
	$out .= "Histogram couldn't not be drawn: $!\n<BR><BR>\n";
    }
    
    unless ($found_dups) {
	$out = "According to entries in $spot_label there were no neighbouring duplicate spots!<BR><BR>";
    }
    return $out;
}


sub rep_flawed_flagging_setup {
    my $num = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
        and
        $cgi->param("${action}_force") == 1) {
        $force_global_opt = 1;
    }

    if ($refine_parameters) {
        print &format_output( 
                             {
                                 'prefix' => 'rep_flawed_',
                                 'num' => $num,
                                 'label' => '-',
                                 'action' => $action,
                                 'info' => 'Parameter Refinement!<BR>'
                                 }
                             );
        return;
    }

    # default dup unit to use:
    my $dup_unit = $var_def_default{'rep_flawed_unit'};
    $dup_unit = &get_user_settings($dup_unit,'rep_flawed_unit',$force_global_opt,$num);
    
    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'rep_flawed_skip_flags'},'rep_flawed_skip_flags',$force_global_opt,$num);
    %skip_flags = &get_skip_flags(@flags);

    my @labels = &get_user_settings($var_def_default{'rep_flawed_label'},'rep_flawed_label',$force_global_opt,$num);

    foreach (@labels) {
	if ($_ eq 'all') {
	    @labels = ('CH1', 'CH2', 'ratio');
	    last;
	}
    }

    # default dup_cutoff to use:
    my $dup_cutoff = $var_def_default{'rep_flawed_cutoff'};

    # check for user specific settings
    $dup_cutoff = &get_user_settings($dup_cutoff,'rep_flawed_cutoff',$force_global_opt,$num);


    # default dup_percent to use:
    my $dup_percent = $var_def_default{'rep_flawed_percent'};

    # check for user specific settings
    $dup_percent = &get_user_settings($dup_percent,'rep_flawed_percent',$force_global_opt,$num);

    # default dup_extend to use:
    my $dup_extend = $var_def_default{'rep_flawed_extend'};

    # check for user specific settings
    $dup_extend = &get_user_settings($dup_extend,'rep_flawed_extend',$force_global_opt,$num);
    
    # default dup_within_slide to use:
    my $dup_within_slide = $var_def_default{'rep_flawed_within_slide'};

    # check for user specific settings
    $dup_within_slide = &get_user_settings($dup_within_slide,'rep_flawed_within_slide',$force_global_opt,$num);
    if ($dup_within_slide eq 'yes') {
	$dup_within_slide = 1;
    } else {
	$dup_within_slide = 0;
    }

    # default dup_flag_single to use:
    my $dup_flag_single = $var_def_default{'rep_flawed_flag_single'};

    # check for user specific settings
    $dup_flag_single = &get_user_settings($dup_flag_single,'rep_flawed_flag_single',$force_global_opt,$num);
    if ($dup_flag_single eq 'yes') {
	$dup_flag_single = 1;
    } else {
	$dup_flag_single = 0;
    }

    my @files = &get_user_settings($var_def_default{'rep_flawed_files'},'rep_flawed_files',$force_global_opt,$num);

    my %file_setup = ();

    if (grep /^all$/, @files) {

	@files = sort keys %data;
	@{$file_setup{'01 all vs all'}} = @files;

    } elsif (grep /^within treatment groups$/, @files) {

	@files = sort keys %data;

	my $counter = 0;
	my $counter_out = 0;
	my %counter = ();
	foreach my $file (@files) {
	    
	    my $full_file = $file;
	    
	    my ($ch1_assign,$ch2_assign);
	    
	    if (defined $cgi->param("CH1_ANNOTATE_$full_file")) {
		$ch1_assign = $cgi->param("CH1_ANNOTATE_$full_file");
	    }
	    if (defined $cgi->param("CH2_ANNOTATE_$full_file")) {
		$ch2_assign = $cgi->param("CH2_ANNOTATE_$full_file");
	    }
	    
	    if ($counter{"$ch1_assign vs $ch2_assign"}) {
		$counter_out = $counter{"$ch1_assign vs $ch2_assign"};
	    } else {
		$counter++;
		$counter_out = $counter;
		until ((length $counter_out) >= 3) {
		    $counter_out = '0'.$counter_out;
		}
		$counter{"$ch1_assign vs $ch2_assign"} = $counter_out;
	    }
	    push @{$file_setup{"$counter_out $ch1_assign vs $ch2_assign"}}, $file;
	}
    } elsif (grep /^between technical replicates$/, @files) {

        @files = sort keys %data;

        my $counter = 0;
        my $counter_out = 0;
        my %counter = ();
        foreach my $file (@files) {

            my $full_file = $file;

            my ($ch1_assign,$ch2_assign, $bio_rep);

            if (defined $cgi->param("CH1_ANNOTATE_$full_file")) {
                $ch1_assign = $cgi->param("CH1_ANNOTATE_$full_file");
            }
            if (defined $cgi->param("CH2_ANNOTATE_$full_file")) {
                $ch2_assign = $cgi->param("CH2_ANNOTATE_$full_file");
            }
            if (defined $cgi->param("BIO_REP_$full_file")) {
                $bio_rep = $cgi->param("BIO_REP_$full_file");
            }
	    unless (defined $bio_rep) {
		$bio_rep = '';
	    }
	    if ($bio_rep eq '') {
		$bio_rep = 'bio_rep1';
	    } else {
		if ($bio_rep =~ /^\d/) {
		    $bio_rep = 'B'.$bio_rep;
		}
	    }

            if ($counter{"$ch1_assign vs $ch2_assign, $bio_rep"}) {
                $counter_out = $counter{"$ch1_assign vs $ch2_assign, $bio_rep"};
            } else {
                $counter++;
                $counter_out = $counter;
                until ((length $counter_out) >= 3) {
                    $counter_out = '0'.$counter_out;
                }
                $counter{"$ch1_assign vs $ch2_assign, $bio_rep"} = $counter_out;
            }
            push @{$file_setup{"$counter_out $ch1_assign vs $ch2_assign, $bio_rep"}}, $file;
        }
    } else {
	@{$file_setup{'01 selection vs selection'}} = @files;
    }
	

    foreach my $label (@labels) {
	foreach my $title (sort keys %file_setup) {
	    @files = @{$file_setup{$title}};
	    my ($num,$title) = split / /, $title, 2;
	    
	    my $info = '';
	    if (@files < 2) {
		$info = "Not enough files for analysis!\n";
	    } else {
		($info) = &rep_flawed_flagging(
					       $num, 
					       $action, 
					       $dup_unit,
					       $dup_cutoff,
					       \%skip_flags,
					       $dup_percent,
					       $dup_extend,
					       $dup_within_slide,
					       $dup_flag_single,
					       $title,
					       $label,
					       @files
					       );
	    }
	    
	    my $label_out = $label;
	    if ($label eq 'ratio') {
		$label_out = '-';
	    }
	    print &format_output( 
				  {
				      'prefix' => 'rep_flawed_',
				      'num' => $num,
				      'label' => $label_out,
				      'action' => $action,
				      'info' => $info
				      }
				  );
	}
    }
}


sub rep_flawed_flagging {
    my $action_num = shift;
    my $action = shift;
    my $dup_unit = shift;
    my $dup_cutoff = shift;
    my $ref = shift;
    my %skip_flags = %$ref;
    my $dup_percent = shift;
    my $dup_extend = shift;
    my $dup_within_slide = shift;
    my $dup_flag_single = shift;
    my $title = shift;
    my $label = shift;
    my @files = @_;

    my $use_ratio = 0;
    my $channel = $label;
    if ($label eq 'ratio') {
	$use_ratio = 1;
	$channel = '__NO_LABEL__';
    }

    my $spot_label = $SPOT_ID_TERM;
    my $flag_header = $FLAG_CHANNEL_TERM;

    if ($refine_parameters) {
	return 'Paramter Refinement!';
    }

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    # default flag to use:
    my $dup_flag = $var_def_default{'rep_flawed_flag'};

    # check for user specific settings
    $dup_flag = &get_user_settings($dup_flag,'rep_flawed_flag',$force_global_opt,$action_num);
#    warn "#kh: flag: $dup_flag";
    if ($dup_flag =~ /\D/) {
	if ($dup_flag eq 'list') {
	    # find the next free list
	    my $counter = 1;
	    until ($dup_flag =~ /\d/) {
		if (defined $used_flags{$dup_flag.$counter}) {
		    $counter++;
		} else {
		    $dup_flag .= $counter;
		    $used_flags{$dup_flag}++;
		}
	    }
	}
	if (defined $flag_values{$dup_flag}) {
	    $dup_flag = $flag_values{$dup_flag};
	} else {
	    $highest_flag_num *= 2;
#	    warn "# KH: new value for $dup_flag: $highest_flag_num";
	    $flag_values{$dup_flag} = $highest_flag_num;
	    $flag_values_num{$highest_flag_num} = $dup_flag;
	    push @all_flags, $dup_flag;
            unshift @all_flags_num, $highest_flag_num;
	    $dup_flag = $highest_flag_num;
	    # keep track of list flags:
	    if ($dup_flag =~ /^list\d+/) {
		$used_flags{$dup_flag}++;
	    }	    

	}
    }

    # define best header to use for each file
    my %headers = ();

    my @check_ch = ($label); 
    if ($use_ratio) {
	@check_ch = ('CH1', 'CH2', '__NO_LABEL__');
    }
	
    foreach my $file (@files) {

	# determine header to use for each file:
	my $header = '';

	HEADER:foreach my $column (@{$item_data_column{$action}}) {

	    unless ($use_ratio) {
		next if ($column =~ /_R/);
	    }

	    foreach my $ch (@check_ch) {
		if ($data_column_available{$file}{$ch}{$column}) {
		    $header = $column;
		    last HEADER;
		}
	    }
	}


	# check for specific settings:
	$header = &get_user_settings($header,'flag_dups_data_column',$force_global_opt,$action_num,$file,$channel);
	
	unless ($header) {
	    return "Error: No data column defined to work on for $action, file $file, label $channel!\n";
	}
	$headers{$file} = $header;

	# read in data if necessary
	my $rc = &read_data($file,$spot_label,$header);
	unless ($rc eq 'ok') {
	    return "<BR><B><FONT COLOR=RED>WARNING: $rc</FONT></B><BR>\n";
	}
    }

    my %header_tmp = ();
    foreach (keys %headers) {
	$header_tmp{$headers{$_}}++;
    }
    $last_column_used = join ', ', sort keys %header_tmp;

    my ($mrow,$mcol,$row,$col);
    my %values = ();
    my %done = ();
    my %skipped = ();
    my %positions = ();
    my %flagged_already = ();

    # collect values for each element in each file

    my $file_counter = -1;
    foreach my $file (@files) {

	$file_counter++;

	my @channels = ();
	if ($use_ratio) {
	    @channels = sort keys %{$data{$file}{'__DATA__'}};
	} else {
	    @channels = ($label);
	}

	my $header = $headers{$file};
	
	my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);
	
	foreach my $mrow (1..$MROW) {
	    foreach my $mcol (1..$MCOL) {
		foreach my $row (1..$ROW) {
		    COL:foreach my $col (1..$COL) {

			# initialize all data output for DUP_DIFF
			unless (defined $data{$file}{'__DATA__'}{$channel}{'DUP_DIFF'}{$mrow}{$mcol}{$row}{$col}) {
			    $data{$file}{'__DATA__'}{$channel}{'DUP_DIFF'}{$mrow}{$mcol}{$row}{$col} = '';
			}
  
			# if we have skip flags, check for these first
			# and skip spot if necessary:
			if (%skip_flags) {
			    # get flags for this spot:
			    my @flags = ();
			    foreach my $ch (@channels) {
				if (defined $data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
				    push @flags, &get_flags($data{$file}{'__DATA__'}{$ch}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
				}
			    }
			    my $flag = 0;
			    # skip flagged spots
			    foreach (@flags) {
				if (defined $skip_flags{$_}) {
				    $flag++;
				    if ($_ eq $dup_flag) {
					$flagged_already{$file}{"$mrow|$mcol|$row|$col"}++;
				    }
				}
			    }
			    
			    if ($flag) {
				$skipped{'individually flagged'}++;
				next COL;
			    }
			}
		    
			# determine identifier for this spot
			my $id = '';
			if ($dup_within_slide) {
			    # get the name as an id
			    if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$spot_label}{$mrow}{$mcol}{$row}{$col}) {
				$id = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$spot_label}{$mrow}{$mcol}{$row}{$col};
				
				# special settings for Arrayvision:
				# delete numerator placed in brackets behind each gene name
				$id =~ s/ \(\d+\)$// if (defined $file_specs{$file}{'FORMAT'} 
							   and
							   $file_specs{$file}{'FORMAT'} eq 'ArrayVision');
				push @{$positions{$id}}, "$mrow|$mcol|$row|$col";
			    } else {
				$skipped{'no spot label'}++;
				next COL;
			    }
			} else {
			    # otherwise use the coordinates as id:
			    $id = "$mrow|$mcol|$row|$col";
			}
			    

			# get the value for this spot
			my $value = '';
			if ($use_ratio) {
			    unless ($header =~ /_R/) {
				# we have to collect both intensities first
				my $val1 = '';
				if (defined $data{$file}{'__DATA__'}{'CH1'}{$header}{$mrow}{$mcol}{$row}{$col}) {
				    $val1 = $data{$file}{'__DATA__'}{'CH1'}{$header}{$mrow}{$mcol}{$row}{$col};
				}
				my $val2 = '';
				if (defined $data{$file}{'__DATA__'}{'CH2'}{$header}{$mrow}{$mcol}{$row}{$col}) {
				    $val2 = $data{$file}{'__DATA__'}{'CH2'}{$header}{$mrow}{$mcol}{$row}{$col};
				}
				if ($val1 eq ''
				    or
				    $val2 eq '') {
				    $skipped{'individual value missing'}++;
				    next COL;
				}

				if ($header =~ /_L/) {
				    $value = $val2 - $val1;
				} else {
				    if ($val1 == 0) {
					$skipped{'invalid value'}++;
					next COL;
				    } else {
					$value = $val2 / $val1;
				    }
				}
			    } else {
				if (defined $data{$file}{'__DATA__'}{$channel}{$header}{$mrow}{$mcol}{$row}{$col}) {
				    $value = $data{$file}{'__DATA__'}{$channel}{$header}{$mrow}{$mcol}{$row}{$col};
				}
			    }
			} else {
			    if (defined $data{$file}{'__DATA__'}{$channel}{$header}{$mrow}{$mcol}{$row}{$col}) {
				$value = $data{$file}{'__DATA__'}{$channel}{$header}{$mrow}{$mcol}{$row}{$col};
			    }
			}
			
			if ($value eq '') {
			    $skipped{'no value defined'}++;
			    next COL;
			}

			push @{$done{$id}}, $file_counter;

			# reverse logs:
			if ($header =~ /_L/) {
			    $value = 2**$value;
			}
			push @{$values{$id}}, $value;
		    }
		  }
	    }
	}
    }
		
    # all useable values are now collected in %values

    # calculate mean, median and standard deviation:
    my @all_differences = ();
    my %differences = ();
    my %apply_flag = ();
    my $found_dups = 0;
    my $found_values = 0;
    my %out_diff = ();

    foreach my $id (keys %values) {
	
	my @tmp = @{$values{$id}};
	
	if (@tmp == 1) {
	    # can't work with only one value -skip!
	    delete $done{$id};
	    # should a flag be applied for single values?
	    if ($dup_flag_single) {
		$apply_flag{$id}++;
		$out_diff{$id} = '';
	    } else {
		$skipped{'only one value'}++;
	    }
	    next;
	}

	$found_dups++;
	$found_values += scalar @tmp;

	my @differences = ();

	# get the difference between each pairing
	for (my $i = 0; $i < $#tmp; $i++) {
	    for (my $j = $i+1; $j <= $#tmp; $j++) {
		my @vals = sort { $a <=> $b } ($tmp[$i], $tmp[$j]);
		my $diff = '';
		if ($dup_unit =~ /abs/) {
		    # this can be tricky if we have ratios between 0 and 1:
		    if ($vals[0] >= 1) {
			# both at least one - simply subtract:
			$diff = $vals[1] - $vals[0];
		    } elsif ($vals[1] < 1) {
			# both between 0 and 1, inverse and subtract:
			my $val0_tmp = 1/$vals[0];
			my $val1_tmp = 1/$vals[1];
			$diff = $val0_tmp - $val1_tmp;
		    } else {
			# one between 0 and 1, inverse that one and subtract:
			my $val0_tmp = 1/$vals[0] - 1;
			my $val1_tmp = $vals[1] - 1;
			$diff = $val1_tmp + $val0_tmp;
		    }
		} else {
		    $diff = $vals[1] / $vals[0];
		}
		push @differences, $diff;
		push @all_differences, $diff;
	    }
	}
	$differences{$id} = join '|', @differences; # to save RAM
    }
    
    # we have now collected all differences and can calculate std deviations
    my $stdev = '?';
    if (@all_differences > 1) {
	$stdev = &stdev(@all_differences);
    }
    my $mean = '?';
    if (@all_differences) {
	$mean = &mean(@all_differences);
    }
    my $median = '?';
    if (@all_differences) {
	$median = &median(@all_differences);
    }

    # check each set of replicates for need of flagging
    foreach my $id (keys %done) {

	my @differences = sort { $b <=> $a } (split /\|/, $differences{$id});
	$out_diff{$id} = $differences[0];


	if ($dup_unit =~ /StdDev/i) {
	    my $median = &median(@differences);
	    foreach my $diff (@differences) {
		$diff = abs($median - $diff);
		$diff /= $stdev;
		if ($diff > $dup_cutoff) {
		    $apply_flag{$id}++;
		    last;
		}
	    }
	} else {
	    foreach my $diff (@differences) {
		if ($diff > $dup_cutoff) {
		    $apply_flag{$id}++;
		    last;
		}
	    }
	}
    }

    # insert max difference into the DUP_DIFF field
    
    foreach my $id (keys %out_diff) {
	my @pos = ($id);
	if ($dup_within_slide) {
	    @pos = @{$positions{$id}};
	}
	
	foreach my $pos (@pos) {
	    my ($mrow, $mcol, $row, $col) = split /\|/, $pos;

	    foreach my $file (@files) {
		$data{$file}{'__DATA__'}{$channel}{'DUP_DIFF'}{$mrow}{$mcol}{$row}{$col} = $out_diff{$id};
	    }

	    # carry out flagging
	    if (defined $apply_flag{$id}) {
		foreach my $file (@files) {
		    next if (defined $flagged_already{$file}{"$mrow|$mcol|$row|$col"});
		    foreach my $ch (@check_ch) {
			# make sure the same flag is only applied once:
			$data{$file}{'__DATA__'}{$ch}{$flag_header}{$mrow}{$mcol}{$row}{$col} += $dup_flag;
			$flagged_already{$file}{"$mrow|$mcol|$row|$col"}++;
		    }
		}
	    }
	}
    }

    # indicate available data fields
    foreach my $file (@files) {
        $data_done{$file}{$channel}{'DUP_DIFF'}++;
        $data_column_available{$file}{$channel}{'DUP_DIFF'}++;
    }
    
    # print report
    my $out = "Results for $title:\n";
    
    my $round = 3;
    if ($mean =~ /\.(\d+)/) {
	if (length($1) > $round) {
	    $mean = sprintf "%.${round}f", $mean;
	}
    }
    if ($median =~ /\.(\d+)/) {
	if (length($1) > $round) {
	    $median = sprintf "%.${round}f", $median;
	}
    }
    if ($stdev =~ /\.(\d+)/) {
	if (length($1) > $round) {
	    $stdev = sprintf "%.${round}f", $stdev;
	}
    }

    my $effective_cutoff = $dup_cutoff;
    if ($dup_unit =~ /stddev/i) {
	$effective_cutoff = $median + $dup_cutoff * $stdev;
    }

    $out .= "<BR><B>$found_dups</B> sets of replicate spots with a total of $found_values values found.<BR>\nSummary for "
	. ($dup_unit =~ /abs/ ? 'absolute difference' : 'relative difference')
	. " on $label:<BR>\n"
	. "&nbsp;&nbsp;mean: <B>$mean</B><BR>&nbsp;&nbsp;median: <B>$median</B><BR>&nbsp;&nbsp;StdDev: <B>$stdev</B><BR>\n"
	. "With ".($dup_unit =~ /abs/ ? 'absolute difference' : 'relative difference')
	. " > <B>$dup_cutoff</B>"
	. ($dup_unit =~ /stddev/i ? " standard deviations from the median (> <B>$effective_cutoff</B>)" : '')
	. "<BR><B>".(scalar keys %apply_flag)."</B> sets of replicates were flagged with $flag_values_num{$dup_flag} ($dup_flag) in total.";
    
    unless ($found_dups) {
	$out = "Results for $title:\n<BR><B>No useable set of replicate spots found!<B><BR><BR>";
    }

    if (%skipped) {
	$out .= "<BR>The following number(s) of values were skipped:\n<UL>\n";
	foreach ( sort { $skipped{$b} <=> $skipped{$a} } keys %skipped) {
	    $out .= "<LI>$_: $skipped{$_}\n";
	}
	$out .= "</UL><BR>\n";
    }

    return $out;
}


sub bg_correction_percentile_setup {
    my $num = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my @files = ();
    my @files_select = &get_user_settings($var_def_default{'fg_min_percentile_files'},'fg_min_percentile_files',$force_global_opt,$num);

    foreach (@files_select) {
	push @files, $_;
	if ($_ eq 'all') {
	    @files = ();
	    foreach (sort keys %data) {
		if (/\w/) {
		    push @files, $_;
		}
	    }
	    last;
	}
    }

    # carry out action on each file ...
    foreach my $file (sort @files) {
	
	# ... and each label
	foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	    
	    # skip columns that are not label specific ('spot labels')
	    next if ($label eq '__NO_LABEL__');
	    
	    print &bg_correction_percentile($num,$file,$label,$action,$FG_COR_SIGNAL_TERM);
	}
    }
}


sub bg_correction_percentile {
    my $action_num = shift;
    my $file = shift;
    my $label = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $header = '';
    foreach my $column (@{$item_data_column{$action}}) {
	if ($data_column_available{$file}{$label}{$column}) {
	    $header = $column;
	    last;
	}
    }
    # check for specific settings:
    $header = &get_user_settings($header,'bg_data_column',$force_global_opt,$action_num,$file,$label);

    unless ($header) {
	return "Error: No data column defined to work on for $action!\n";
    }

    $last_column_used = $header;

    my $out_header = shift || $header;

    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'fg_min_percentile_skip_flags'},'fg_min_percentile_skip_flags',$force_global_opt,$action_num,$file,$label);
    %skip_flags = &get_skip_flags(@flags);
    my %skipped = ();

    if ($refine_parameters) {
	# format output (include image and choice for parameters)
	print &format_output( # 'bg_',$action_num,$file,$label,$action,'','Parameter Refinement!<BR>','',''
			     {
				 'prefix' => 'bg_',
				 'num' => $action_num,
				 'file' => $file,
				 'label' => $label,
				 'action' => $action,
				 'info' => 'Parameter Refinement!<BR>'
				 }
			     );
	$data_done{$file}{$label}{$out_header}++;
	$data_column_available{$file}{$label}{$out_header}++;
	return '';
    }

    # check for user specific settings
    my $percentile = &get_user_settings($var_def_default{'percentile'},'percentile',$force_global_opt,$action_num,$file,$label);

    # check for user specific settings
    my $percentile_action = &get_user_settings($var_def_default{'percentile_action'},'percentile_action',$force_global_opt,$action_num,$file,$label);

    my $rc = &read_data($file,$header);
    unless ($rc eq 'ok') {
	return "<BR><B><FONT COLOR=RED>WARNING: $rc</FONT></B><BR>\n";
    }

    my ($mrow,$mcol,$row,$col,$data1,$new_data);

    my $flagged = 0;
    my $min_value = 9999999;
    my $low_values = 0;

    # transform data
    foreach $mrow (keys %{$data{$file}{'__DATA__'}{$label}{$header}}) {
	foreach $mcol (keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}}) {

	    # first collect the signals for each subgrid
	    my @signals = ();	    
	    foreach $row (keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}}) {
		foreach $col (keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}}) {

		    # skip spots without value
		    next unless (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col});

		    # skip flagged spots
		    my $skip = 0;
		    if (%skip_flags
			and
			(defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}
			or
			defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col})) {
			my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
			if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			    push @flags, &get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
			}
			foreach (@flags) {
			    if (defined $skip_flags{$_}) {
				$skip = 1;
				last;
			    }
			}
		    }			
		    
		    if ($skip) {
			$skipped{'at least one of the values is flagged'}++;
			next;
		    }

		    push @signals, $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
		}
	    }
	    
	    # calculate the lower x% percentile
#	    my $quant = (&quantile($percentile, \@signals))[1];

	    my @percentile_signals = ();

	    my @signals_ordered = sort { $a <=> $b } @signals;
	    # calculate the number of elements which go into median calculation
	    my $elements = sprintf "%.0f", ($percentile * @signals_ordered);
	    for (my $i = 0; $i <= $elements; $i++) {
		push @percentile_signals, $signals_ordered[$i];
	    }
	    my $percentile_bg = &median(@percentile_signals);

	    # update data for this subgrid:
	    foreach $row (keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}}) {
		foreach $col (keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}}) {

		    # initialize field
		    $data{$file}{'__DATA__'}{$label}{'BG_CALC'}{$mrow}{$mcol}{$row}{$col} = '';

		    # skip spots without value
		    next unless (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col});

		    # skip flagged spots
		    my $skip = 0;
		    if (%skip_flags
			and
			(defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}
			or
			defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col})) {
			my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
			if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			    push @flags, &get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
			}
			foreach (@flags) {
			    if (defined $skip_flags{$_}) {
				$skip = 1;
				last;
			    }
			}
		    }			
		    
		    if ($skip) {
			next;
		    }

		    $data1 = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};

		    # subtract quant from each signal in this subgrid:
		    if ($data1 =~ /[a-z]/i
			or
			$percentile_bg =~ /[a-z]/i) {
			die "ERROR: invalid data: file: $file, header: $header, label: $label, data: '$data1', quant: '$percentile_bg'\n";
		    }
		    $new_data = $data1 - $percentile_bg;

		    if ($new_data < $min_value) {
			$min_value = $new_data;
		    }
		    if ($new_data <= 0) {
			$low_values++;
		    }

		    # store new data and quant
		    $data{$file}{'__DATA__'}{$label}{'BG_CALC'}{$mrow}{$mcol}{$row}{$col} = $percentile_bg;
		    if ($percentile_action eq 'adjust') {
			$data{$file}{'__DATA__'}{$label}{$out_header}{$mrow}{$mcol}{$row}{$col} = $new_data;
		    } else {
			if ($new_data <= 0) {
			    my $skip_flagging = 0;
			    # check if floor flag is set already
			    if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}
				and
				$data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} >= $flag_values{'floor'}) {
				my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
				foreach (@flags) {
				    if ($_ == $flag_values{'floor'}) {
					$skip_flagging++;
					last;
				    }
				}
			    }
			    $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'floor'} unless ($skip_flagging);
			    $flagged++;
			}
			if ($percentile_action eq 'both') {
			    $data{$file}{'__DATA__'}{$label}{$out_header}{$mrow}{$mcol}{$row}{$col} = $new_data;
			}
		    }
		}
	    }
	}
    }

    # take note of available columns for later

    if ($percentile_action eq 'adjust'
	or
	$percentile_action eq 'both') {
	unless (defined $data_column_available{$file}{$label}{$out_header}) {
	    $data_column_available{$file}{$label}{$out_header} = 1;
	    push @{$data_column_available{$file}{$label}{'__ORDER__'}}, $out_header;
	}
	$data_done{$file}{$label}{$out_header}++;
    }

    unless (defined $data_column_available{$file}{$label}{'BG_CALC'}) {
	$data_column_available{$file}{$label}{'BG_CALC'} = 1;
	push @{$data_column_available{$file}{$label}{'__ORDER__'}}, 'BG_CALC';
    }
    $data_done{$file}{$label}{'BG_CALC'}++;

     # draw chip with values from quant column
#    warn "#KH: doing draw_chip($action_num,$file,$label,'BG_CALC',$action)";
    my ($image,$info,$min_val,$max_val) = &draw_chip( # 'bg_',$action_num,$file,$label,'BG_CALC',$action.'_draw'
							   {
							       'prefix' => 'bg_',
							       'num' => $action_num,
							       'file' => $file,
							       'label' => $label,
							       'header' => 'BG_CALC',
							       'action' => $action.'_draw'
							       }
							   );
    
    if ($image =~ /Error\:/i) {
	print "<BR><B>$image<B><BR>\n";
	next;
    }

    $info =~ s/<TD COLSPAN=2>/<TD COLSPAN=2>Lowest corrected value = <B>$min_value<\/B><BR><B>$low_values<\/B> values <= 0<BR>\n/is;

    if ($percentile_action eq 'flag') {
	$info =~ s/<TD COLSPAN=2>/<TD COLSPAN=2><B>$flagged<\/B> spots were flagged as 'floor'!<BR><BR>\n<table>/is;
    }

    # format output (include image and choice for parameters)
    print &format_output( # 'bg_',$action_num,$file,$label,$action,$image,$info,$min_val,$max_val
			 {
			     'prefix' => 'bg_',
			     'num' => $action_num,
			     'file' => $file,
			     'label' => $label,
			     'action' => $action,
			     'image' => $image,
			     'info' => $info,
			     'min' => $min_val,
			     'max' => $max_val
			     }
			 );
    
    return '';
}


sub bg_correction_loess_setup {
    my $num = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
        and
        $cgi->param("${action}_force") == 1) {
        $force_global_opt = 1;
    }

    my @files = ();
    my @files_select = &get_user_settings($var_def_default{'bg_loess_files'},'bg_loess_files',$force_global_opt,$num);

    foreach (@files_select) {
        push @files, $_;
        if ($_ eq 'all') {
            @files = ();
            foreach (sort keys %data) {
                if (/\w/) {
                    push @files, $_;
                }
            }
            last;
        }
    }

    # carry out action on each file ...
    foreach my $file (@files) {
	
	# ... and each label
	foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	    
	    # skip columns that are not label specific ('spot labels')
	    next if ($label eq '__NO_LABEL__');
	    
	    print &bg_correction_loess($num,$file,$label,$action,$FG_COR_SIGNAL_TERM.'_L');
	}
    }
}


sub bg_correction_loess {
    my $action_num = shift;
    my $file = shift;
    my $label = shift;
    my $action = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $header = '';
    foreach my $column (@{$item_data_column{$action}}) {
	if ($data_column_available{$file}{$label}{$column}) {
	    $header = $column;
	    last;
	}
    }
    # check for specific settings:
    $header = &get_user_settings($header,'bg_loess_data_column',$force_global_opt,$action_num,$file,$label);
    my $span = &get_user_settings($var_def_default{'bg_loess_span'},'bg_loess_span',$force_global_opt,$action_num,$file,$label);
    unless ($span > 0
	    and
	    $span <= 1) {
	warn "# $progname: invalid span: $span, setting to default ($var_def_default{'bg_loess_span'})";
	$span = $var_def_default{'bg_loess_span'};
    }

    unless ($header) {
	return "Error: No data column defined to work on for $action!\n";
    }

    $last_column_used = $header;

    my $out_header = shift || $header;
    $out_header =~ s/_L//;
    
    my %skip_flags = ();
    my @flags = &get_user_settings($var_def_default{'bg_loess_skip_flags'},'bg_loess_skip_flags',$force_global_opt,$action_num,$file,$label);
    %skip_flags = &get_skip_flags(@flags);
    my %skipped = ();

    if ($refine_parameters) {
	# format output (include image and choice for parameters)
	print &format_output( # 'bg_',$action_num,$file,$label,$action,'','Parameter Refinement!<BR>','',''
			     {
				 'prefix' => 'bg_',
				 'num' => $action_num,
				 'file' => $file,
				 'label' => $label,
				 'action' => $action,
				 'info' => 'Parameter Refinement!<BR>'
				 }
			     );
	$data_done{$file}{$label}{$out_header}++;
	$data_column_available{$file}{$label}{$out_header}++;
	return '';
    }

    my $rc = &read_data($file,$header);
    unless ($rc eq 'ok') {
	return "<BR><B><FONT COLOR=RED>WARNING: $rc</FONT></B><BR>\n";
    }

    my ($mrow,$mcol,$row,$col,$data1,$new_data);

    my $flagged = 0;
    my $min_value = 9999999;

    # write output file with foreground signals
    
    my $data_file = "data_action.$action.$action_num.$action_step.file.".(basename $file).".label$label.PID$PID";
    $data_file =~ s/ /_/g;
    $data_file =~ s/-/_/g;
    my $R_commands_file = "$tmp_dir/$data_file.R";
    my $R_out = "$tmp_dir/$data_file.out";
    $data_file = "$tmp_dir/$data_file";

    my $input_rows = 0;

    open (OUT, ">$data_file")
	  or die "Can't write to $data_file: $!";

    my ($dim_mr,$dim_mc,$dim_r,$dim_c) = (0,0,0,0);
    my @signals = ();	    
    foreach $mrow (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}}) {
	if ($mrow > $dim_mr) {
	    $dim_mr = $mrow;
	}
	foreach $mcol (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}}) {
	    if ($mcol > $dim_mc) {
		$dim_mc = $mcol;
	    }
	    foreach $row (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}}) {
		if ($row > $dim_r) {
		    $dim_r = $row;
		}
		foreach $col (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}}) {
		    if ($col > $dim_c) {
			$dim_c = $col;
		    }

		    # skip spots without value
		    my $value = '';
		    if (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}) {
			$value = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
		    }

		    # skip flagged spots
		    my $skip = 0;
		    if (%skip_flags
			and
			(defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}
			or
			defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col})) {
			my @flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
			if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			    push @flags, &get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
			}
			foreach (@flags) {
			    if (defined $skip_flags{$_}) {
				$skip = 1;
				last;
			    }
			}
		    }			
		    
		    if ($skip) {
			$skipped{'at least one of the values is flagged'}++;
		    }

		    if ($value !~ /\d/
			or
			$skip) {
#			$value = 'NA';
			# initialize output fields to empty
                        $data{$file}{'__DATA__'}{$label}{$out_header}{$mrow}{$mcol}{$row}{$col} = '';
                        $data{$file}{'__DATA__'}{$label}{'BG_CALC'}{$mrow}{$mcol}{$row}{$col} = '';
			# set flag for this label/position to 'undefined'
                        $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} = $flag_values{'undefined'};

                        next;
		    }
		    print OUT "$mrow\t$mcol\t$row\t$col\t$value\n";
		    $input_rows++;
		}
	    }
	}
    }

    close OUT;

    open (RH, ">$R_commands_file")
	or die "Can't write to $R_commands_file: $!";

    # read in data
    print RH "all.int <- read.table(\"$data_file\")\n";
    print RH "attach(all.int)\n";

    # cols of all.int are MetaRow, MetaColumn, Row, Column, then
    # 1 col of signal intensity data from 1 channel of an array
    # dims of each subgrid are dim_r x dim_c
    print RH "ypos = all.int[,3] + $dim_r * (all.int[,1] - 1)\n";
    print RH "xpos = all.int[,4] + $dim_c * (all.int[,2] - 1)\n";
                                                                                                                             
    print RH "BG.crd.all.int = all.int\n";
#    print RH "BG.crd.all.int[,5] = predict(loess(all.int[,5]~xpos+ypos, span = $span))\n";
    print RH "BG.crd.all.int[,5] = predict(loess(all.int[,5]~xpos*ypos, span = $span, family=\"symmetric\"))\n";
    print RH "BG.crd.all.int[,6] = all.int[,5] - BG.crd.all.int[,5]\n";

    # print out data
    print RH "write.table(BG.crd.all.int, \"$R_out\", sep=\"\\t\", quote=F, col.names=F, row.names=F)\n";

    # this calculation might take a while, so we write out a file as a signal
    # that the process has finished
    my $signal_file = $R_out.'.finished';
    if (-e $signal_file) {
	unlink $signal_file
	    or return "Can't delete signal file '$signal_file': $!\n";
    }

    print RH "out <- cbind(1,2,3,4,5,6,7,8,9)\n";
    print RH "write.table(out, \"$signal_file\")\n";
    
    close RH;

    my @images = &get_R_output($R_commands_file, $R_out, $action_num, 'all_files', $signal_file, '', 1);
    
    open (IN, $R_out)
	or die "Can't read $R_out: $!";

    my $nas = 0;
    $counter = 0;
    while (<IN>) {
	chomp;

	# get coordinates and background corrected value
	my ($mrow,$mcol,$row,$col,$bg_calc,$val) = split /\t/, $_;
	
	# write back values for each spot label
	if ($val =~ /na/i) {
	    $nas++;
	    $val = '';
	} else {
	    # keep track of lowest value:
	    if ($val < $min_value) {
		$min_value = $val;
	    }
	}
	
	$data{$file}{'__DATA__'}{$label}{$out_header}{$mrow}{$mcol}{$row}{$col} = $val;
	$data{$file}{'__DATA__'}{$label}{'BG_CALC'}{$mrow}{$mcol}{$row}{$col} = $bg_calc;
	
	$counter++;
    }
    close IN;

    if ($clean_up_files) {
	unlink $R_commands_file;
	unlink $R_out;
	unlink $data_file;
    }
		

    # take note of available columns for later
    unless (defined $data_column_available{$file}{$label}{$out_header}) {
	$data_column_available{$file}{$label}{$out_header} = 1;
	push @{$data_column_available{$file}{$label}{'__ORDER__'}}, $out_header;
    }

    $data_done{$file}{$label}{$out_header}++;

    unless (defined $data_column_available{$file}{$label}{'BG_CALC'}) {
	$data_column_available{$file}{$label}{'BG_CALC'} = 1;
	push @{$data_column_available{$file}{$label}{'__ORDER__'}}, 'BG_CALC';
    }

    $data_done{$file}{$label}{'BG_CALC'}++;

    my $info = "Label $label:<BR><B>$counter</B> values processed<BR>\n";
    if ($nas) {
	$info .=  "<B>$nas</B> undefined spots<BR>\n";
    }
    $info .=  "<BR>\n";

    unless ($counter == $input_rows) {
	print "<B>WARNING: only $counter rows (instead of $input_rows)  returned from loess background correction!</B><BR><BR>\n";
    }

     # draw chip with values from quant column
#    warn "#KH: doing draw_chip($action_num,$file,$label,'BG_CALC',$action)";
    my ($image,$info2,$min_val,$max_val) = &draw_chip( # 'bg_',$action_num,$file,$label,'BG_CALC_L',$action.'_draw'
						      {
							  'prefix' => 'bg_',
							  'num' => $action_num,
							  'file' => $file,
							  'label' => $label,
							  'header' => 'BG_CALC',
							  'action' => $action.'_draw'
							  }
						      );
    
    if ($image =~ /Error\:/i) {
	print "<BR><B>$image<B><BR>\n";
	next;
    }

    $info =~ s/<TD COLSPAN=2>/<TD COLSPAN=2>Lowest corrected value = <B>$min_value<\/B>.<BR>\n/is;

    # format output (include image and choice for parameters)
    print &format_output( # 'bg_loess_',$action_num,$file,$label,$action,$image,$info2.$info
			 {
			     'prefix' => 'bg_loess',
			     'num' => $action_num,
			     'file' => $file,
			     'label' => $label,
			     'action' => $action,
			     'image' => $image,
			     'info' => $info2.$info
			     }
			 );
    
    return '';
}

sub quantile_average {

    my $ref = shift;
    my @percentiles = @$ref;
    $ref = shift;
    my @tmp = @$ref;

    # filter out those items that are not numeric
    my @data = ();
    foreach (@tmp) {
	unless (/\d/) {
	    next;
	} else {
	    push @data, $_;
	}
    }

    my $len = scalar @data;
    my $window = sprintf "%.0f", $len / 10;
    unless ($window > 0) {
	$window = 0;
    }

    my %percentiles = ();
    my %positions = ();
    foreach my $percentile (@percentiles) {
	my $pos = sprintf "%.0f", $len * $percentile;
	$percentiles{$percentile} = $pos;
	$positions{$pos}++;
    }

    my @average = @data[0..$window-1];
    for (my $i = 0; $i <= $#data; $i++) {

	if ($i >= $window) {
	    push @average, $data[$i];
	}

	if ( (scalar @average) > $window) {
	    shift @average;
	}
	if (defined $positions{$i}) {
	    my $average = &median(@average);
	    $positions{$i} = $average;
	}
    }
	
    my @out = ();

    foreach my $percentile (@percentiles) {
	my $pos = $percentiles{$percentile};
	my $val = $positions{$pos};
	push @out, $val;
    }
    return @out;
}


sub quantile {
    
    # details from the R help(quantile) text:
    # 
    #   `quantile(x,p)' as a function of `p' linearly interpolates the
    #   points ( (i-1)/(n-1), ox[i] ), where `ox <- sort(x)' and `n <-
    #   length(x)'.
    #  
    #   This gives `quantile(x, p) == (1-f)*ox[i] + f*ox[i+1]', where `r
    #   <- 1 + (n-1)*p', `i <- floor(r)', `f <- r - i' and `ox[n+1] :=
    #   ox[n]'.

    my $percentile = shift;
    my $ref = shift;
    my @tmp = @$ref;

    # filter out those items that are not numeric
    my @data = ();
    foreach (@tmp) {
	unless (/\d/) {
	    next;
	} else {
	    push @data, $_;
	}
    }
    my @order = (sort { $a <=> $b } @data);
    my $n = scalar @data;
    my $rank = 1 + ($n - 1) * $percentile;
    my $i = &floor($rank);
    my $num1 = $order[$i-1];
    my $num2 = ($i > $#order ? $order[-1] : $order[$i]);
    my $fraction = $rank - $i;
    my $quant = (1 - $fraction) * $num1 + $fraction * $num2;

    # get the inverse quant as well
    $percentile = 1 - $percentile;
    $rank = 1 + ($n - 1) * $percentile;
    $i = &floor($rank);
    $num1 = $order[$i-1];
    $num2 = ($i > $#order ? $order[-1] : $order[$i]);
    $fraction = $rank - $i;
    my $quant2 = (1 - $fraction) * $num1 + $fraction * $num2;
    
    # get the median:    
    my $md = &median(@order);

    # return min, first quant, median, second quant, max:
    return $order[0],$quant,$md,$quant2,$order[-1];
}

sub floor {

    # get the floor of a number 
    # (greatest integer that is lower or equal to number)

    my $num = shift;
    my $floor;
    if ($num >= 0) {
	$floor = int($num);
    } else {
	$floor = int($num);
	unless ($floor == $num) {
	    $floor -= 1;
	}
    }
    return $floor;
}


sub write_data_to_file {

    # write data from specified headers into a
    # temporary file, using some predefined format;
    # possible formats so far:
    # - oneline_tab
    # - oneline_space
    # - tab
    # - space

    my $mrows = shift;
    my $mcols = shift;
    my $file = shift;
    my $label = shift;
    my $format = shift;
    my $id = shift;
    my @header = @_;

    my $header = join ',', @header;

    my $tmp_file = "data_mrows$mrows.mcols$mcols.format$format.header$header.label$label.id$id.PID$PID.tmp";
    $tmp_file =~ s/ /_/g;
    $tmp_file =~ s/-/_/g;
    $tmp_file = "$tmp_dir/$tmp_file";

    open (OUT, ">$tmp_file")
	  or die "Can't write to $tmp_file: $!";

    my ($mrow,$mcol,$row,$col);
    my @tmp = ();

    my @flags = ();
    # check if flags are defined
    if (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{'__FLAGS__'}
	and
	keys %{$data{$file}{'__DATA__'}{'__NO_LABEL__'}{'__FLAGS__'}}) {
	@flags = sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{'__NO_LABEL__'}{'__FLAGS__'}};
    }

    # pick first header to cycle through each spot
    $header = $header[0];    

    my @mrows = sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}};
    my @mcols = ();
    if (@mrows) {
	@mcols = sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrows[0]}};
    }

    if ($mrows =~ /\d/) {
	@mrows = ($mrows);
    }
    if ($mcols =~ /\d/) {
	@mcols = ($mcols);
    }

    foreach $mrow (@mrows) {
	foreach $mcol (@mcols) {
	    foreach $row (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}}) {
		COL:foreach $col (sort { $a <=> $b } keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}}) {
		    
		    # collect all data in @tmp if it is to be written
		    # in one line, otherwise clear print out each row
		    # separately
		    @tmp = () unless ($format =~ /oneline/);
		    
		    # collect data for all headers into @tmp
		    foreach (@header) {			
			
			my $val;
			# skip spots without value
			if (!defined $data{$file}{'__DATA__'}{$label}{$_}{$mrow}{$mcol}{$row}{$col}
			    or
			    $data{$file}{'__DATA__'}{$label}{$_}{$mrow}{$mcol}{$row}{$col} eq '') {
			    next COL;
			} else {
			    $val = $data{$file}{'__DATA__'}{$label}{$_}{$mrow}{$mcol}{$row}{$col};
			}
			push @tmp, $val;
		    }
		    
		    # add columns for each flag/header
		    foreach my $flag (@flags) {
			foreach my $header (@header) {			
			    my $val;
			    # skip spots without value
			    unless (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{'__FLAGS__'}{$flag}{$header}{$mrow}{$mcol}{$row}{$col}) {
				$val = 0;
			    } else {
				$val = $data{$file}{'__DATA__'}{'__NO_LABEL__'}{'__FLAGS__'}{$flag}{$header}{$mrow}{$mcol}{$row}{$col};
			    }
			    push @tmp, $val;
			}
		    }

		    # print row combined with specified separator
		    if ($format eq 'tab') {
			print OUT (join "\t", @tmp)."\n";
		    } elsif ($format eq 'space') {
			print OUT (join " ", @tmp)."\n";
		    }
		}
	    }
	}
    }

    # write all data as one long line
    if ($format =~ /oneline/) {

	# ... combined with specified separator
	if ($format =~ /space/) {
	    print OUT (join ' ', @tmp)."\n";
	} elsif ($format eq 'space') {
	    print OUT (join " ", @tmp)."\n";
	}
    }

    close OUT;

    return $tmp_file;
}
    
    
sub draw_chip {
    my $ref = shift;
    my %args = %$ref;

    my $prefix = $args{'prefix'} || '';
    my $action_num = '';
    if (defined $args{'num'}) {
	$action_num = $args{'num'};
    }
    my $file = $args{'file'} || '';
    my $label = $args{'label'} || '';
    my $header = $args{'header'} || '';
    my $action = $args{'action'} || '';

    my ($max_val,$min_val,$md,$lq,$uq);
    my %flag_stats = ();
    my %flag_stats_all = ();
    my @images = ();

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    # size of each square can be overriden with user specified value:
    my $orientation = 'landscape';

    # check for user specific settings
    $orientation = &get_user_settings($orientation,$prefix.'orientation',$force_global_opt,$action_num,$file,$label);

    my $flip_image = ($orientation eq 'landscape' ? 1 : 0);

    # general variables for graphical output:
    my ($MROW,$MCOL,$ROW,$COL) = &get_array_specs($file);
    my $pix_x = $MCOL * $COL;
    my $pix_y = $MROW * $ROW;
    
    if ($flip_image) {
	$pix_x = $MROW * $ROW;
	$pix_y = $MCOL * $COL;
    }
    
    my $space_x = 15;
    my $space_y = 25;
    my $legend_y = 0;
    my $legend_font = gdMediumBoldFont;
    my $legend_text_width = 6*$legend_font->width;
    my $legend_x = 25;
    if ($flip_image) {
	($legend_y,$legend_x) = ($legend_x,$legend_y);
	$legend_text_width = 0;
    }

    my $square_size = $var_def_default{$prefix.'square_size'} || 2;
    # check for user specific settings
    $square_size = &get_user_settings($square_size,$prefix.'square_size',$force_global_opt,$action_num,$file,$label);

 #   my $draw_flags = $var_def_default{$prefix.'draw_flags'};
 #   # check for user specific settings
 #   $draw_flags = &get_user_settings($draw_flags,$prefix.'draw_flags',$force_global_opt,$action_num,$file,$label);
 #   $draw_flags = 0 if ($draw_flags eq 'off');

    my @flags = &get_user_settings($var_def_default{$prefix.'skip_flags'},$prefix.'skip_flags',$force_global_opt,$action_num,$file,$label);
    my %skip_flags = ();
    %skip_flags = &get_skip_flags(@flags);

   # make sure we have all data
    &read_data($file,$header);

    # get the size of the picture
    my $all_x = $square_size * $pix_x + 2*$space_x + $legend_x + $legend_text_width;
    my $all_y = $square_size * $pix_y + 2*$space_y + $legend_y;

    my $im = new GD::Image($all_x,$all_y);

    my ($intensity,$mcol,$mrow,$col,$row,$x1,$x2,$y1,$y2,$xt,$yt);

    # get min and max vals
    my @vals = ();
    my $log_transfer = 0;
    foreach $mrow (keys %{$data{$file}{'__DATA__'}{$label}{$header}}) {
	foreach $mcol (keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}}) {
	    foreach $row (keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}}) {
		foreach $col (keys %{$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}}) {
		    if (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}) {
			my $val = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
			if ($val eq 'error') {
			    $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'error'};
			    next;
			}
#			if ($val =~ /[^\d\.\-\+e]/) {
#			    $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'error'};
#			    return "ERROR: Invalid value for $header in $file, $label at $mrow,$mcol,$row,$col: '$val'\n";
#			} 
			if ($log_transfer) {
			    $val = &log_base(2,$val);
			}
			push @vals, $val;			    
		    }
		}
	    }
	}
    }

    # get box plot values
    my ($max_val_orig,$min_val_orig);
    ($min_val_orig,$lq,$md,$uq,$max_val_orig) = &quantile(0.25,\@vals);

    # avoid divisions by zero when min and max are the same
    if ($min_val_orig == $max_val_orig) {
	$max_val_orig += 1;
    }

    my $min_border_set = 0;
    my $max_border_set = 0;

    # absolute or relative application of borders:
    my $percentile_borders = $var_def_default{$prefix.'border_setting'};
    $percentile_borders = &get_user_settings($percentile_borders,$prefix.'border_setting',$force_global_opt,$action_num,$file,$label);
    $percentile_borders = ( $percentile_borders eq 'abs' ? 0 : 1 );

    # check for user specific settings
    my $min_val_input = &get_user_settings($var_def_default{$prefix.'min'},$prefix.'min',$force_global_opt,$action_num,$file,$label);
    $min_val = $min_val_input;
    if ($min_val =~ /def/
	or
	$min_val eq '') {
	if ($percentile_borders) {
	    $min_val = $lq;
	} else {
	    $min_val = $min_val_orig;
	}
    } else {
	if ($percentile_borders) {
	    $min_val =~ s/\s//g;
	    $min_val =~ s/\%//g;	    
	    if ($min_val < 0
		or
		$min_val >= 100) {
		return "ERROR: please specify a minimum border between 0 and 100!\n";
	    } else {
		if ($min_val > 1) {
		    $min_val /= 100;
		}
		if ($min_val == 0) {
		    $min_val = $min_val_orig;
		} elsif ($min_val == 1) {
		    $min_val = $max_val_orig;
		} else {
		    my ($min_val_tmp,$lq_tmp,$md_tmp,$uq_tmp,$max_val_tmp) = &quantile($min_val,\@vals);
		    $min_val = $lq_tmp;
		}
	    }
	} else {
	    if ($min_val < $min_val_orig) {
		$min_val = $min_val_orig;
	    }
	}
    }

    $min_border_set = 1 unless ($min_val == $min_val_orig);

    my $max_val_input = &get_user_settings($var_def_default{$prefix.'max'},$prefix.'max',$force_global_opt,$action_num,$file,$label);
    $max_val = $max_val_input;
    if ($max_val =~ /def/
	or
	$max_val eq '') {
	if ($percentile_borders) {
	    $max_val = $uq;
	} else {
	    $max_val = $max_val_orig;
	}
    } else {
	if ($percentile_borders) {
	    $max_val =~ s/\s//g;
	    $max_val =~ s/\%//g;	    
	    if ($max_val < 0
		or
		$max_val > 100) {
		return "ERROR: please specify a maximum border between 0 and 100!\n";
	    } else {
		if ($max_val > 1) {
		    $max_val /= 100;
		}
		if ($max_val == 0) {
		    $max_val = $min_val_orig;
		} elsif ($max_val == 1) {
		    $max_val = $max_val_orig;
		} else {
		    my ($min_val_tmp,$lq_tmp,$md_tmp,$uq_tmp,$max_val_tmp) = &quantile(1-$max_val,\@vals);
		    $max_val = $uq_tmp;
		}
	    }
	} else {
	    if ($max_val > $max_val_orig) {
		$max_val = $max_val_orig;
	    }
	}
    }

    $max_border_set = 1 unless ($max_val == $max_val_orig);

#    print "MAX: $max_val, MIN: $min_val, ".(scalar @vals)." elements<BR>\n";
    

    # allocate some colors
    my $white = $im->colorAllocate(255,255,255);
    my $black = $im->colorAllocate(0,0,0);
    my $grey = $im->colorAllocate(150,150,150);
    my $red = $im->colorAllocate(255,0,0);
    my $blue = $im->colorAllocate(0,0,255);
    my $orange = $im->colorAllocate(255,165,0);
    my $green = $im->colorAllocate(0,255,0);
    my $yellow = $im->colorAllocate(255,255,0);
    my $cyan = $im->colorAllocate(0,255,255);
    my $magenta = $im->colorAllocate(255,0,255);
    my $brown = $im->colorAllocate(103,67,0);
    my $sienna = $im->colorAllocate(142,107,35);
    my $pink = $im->colorAllocate(255,174,185);
    my $thistle = $im->colorAllocate(146,62,112);
    my $coral = $im->colorAllocate(248,137,117);
    my $olive = $im->colorAllocate(59,94,43);
    my $teal = $im->colorAllocate(0,128,128);
    my $purple = $im->colorAllocate(128,0,128);
    my $plum = $im->colorAllocate(76,46,87);
    my $orchid = $im->colorAllocate(218,107,212);
    my $turquoise = $im->colorAllocate(64,224,208);
    my $springgreen = $im->colorAllocate(0,255,127);
    my $darkcyan = $im->colorAllocate(0,139,139);
    my $maroon = $im->colorAllocate(176,48,96);
    my $firebrick = $im->colorAllocate(178,34,34);
    my $darkorange = $im->colorAllocate(205,102,0);
    my $darkmagenta = $im->colorAllocate(139,0,139);
    my $darkgreen = $im->colorAllocate(0,100,0);
    my %colour = ('red',$red,
		  'blue',$blue,
		  'green',$green,
		  'orange',$orange,
		  'yellow', $yellow,
		  'magenta', $magenta,
		  'cyan', $cyan,
		  'brown', $brown,
		  'sienna', $sienna,
		  'pink', $pink,
		  'thistle', $thistle,
		  'coral', $coral,
		  'olive', $olive,
		  'teal', $teal,
		  'purple', $purple,
		  'darkgreen', $darkgreen,
		  'plum', $plum,
		  'orchid', $orchid,
		  'grey', $grey,
		  'turquoise', $turquoise,
		  'springgreen', $springgreen,
		  'darkcyan', $darkcyan,
		  'maroon', $maroon,
		  'firebrick', $firebrick,
		  'darkorange', $darkorange,
		  'darkmagenta', $darkmagenta,
		  'darkgreen', $darkgreen,
		  );

    # set user defined colours for flag
    my %flag2col_tmp = ();
    foreach my $flag (0, (sort { $a <=> $b } keys %flag_values_num)) {
 
	my $default = '';
	if (defined $pref_flag2col{$flag}) {
	    $default = $pref_flag2col{$flag};
	}
	# check for user specific settings
	my $flag_colour = &get_user_settings($default,$prefix.'flag_colour'.$flag,$force_global_opt,$action_num,$file,$label);
	$flag2col_tmp{$flag} = $flag_colour if ($flag_colour =~ /\w/);
    }
    

    # image names:
    my $file_tmp = basename $file;
    $file_tmp =~ s/\//_/g;
    $file_tmp =~ s/ /_/g;
    my $image_name = "$image_stem.$PID..$action_step.$action_num.$file_tmp.$label.png";
    my $image_out = "$tmp_dir/$image_name";
    
    # frame:
    $im->rectangle($space_x-1,$space_y-1,$all_x-($space_x-1)-$legend_x-$legend_text_width,$all_y-($space_y-1)-$legend_y,$black);

    my $startx = $space_x;
    my $starty = $all_y - $space_y - $legend_y;

    # print legend

    if ($flip_image) {
	$x1 = $startx;
	$y1 = $starty + 5;
	$y2 = $y1 + $legend_y;
    } else {
	$x1 = $all_x - $legend_x - $legend_text_width - 5;
	$x2 = $x1 + $legend_x;
	$y1 = $space_y;
    }

    my @borders = ();
    my %borders = ();
    my $gradient_steps = 10;

    # make sure min and max are not the same
    if ($min_val_orig == $max_val_orig) {
	$max_val_orig = $min_val_orig + 1;
    }

    # make sure min and max are not the same
    if ($min_val == $max_val) {
	$max_val = $min_val + 1;
    }

    # if requested, set the borders within 
    # certain percentiles of the data
    my $main_spread_borders = 0;
    
    # modify lower and upper border 
    # unless they are the min and max values, respectively:
    if ($min_val != $min_val_orig
	and
	$max_val != $max_val_orig) {
	my $min_val_tmp = $min_val;
	$min_val -= ($max_val - $min_val) / ($gradient_steps - 2);
	$max_val += ($max_val - $min_val_tmp) / ($gradient_steps - 2);
    } elsif ($min_val != $min_val_orig) {
	$min_val -= ($max_val - $min_val) / ($gradient_steps - 1);
    } elsif ($max_val != $max_val_orig) {
	$max_val += ($max_val - $min_val) / ($gradient_steps - 1);
    }

    for my $i (0..$gradient_steps-1) {
	# define bin borders
	$intensity = $min_val + $i*(($max_val - $min_val) / $gradient_steps);
	if ($intensity == 0
	    or
	    $intensity == -0) {
	    $intensity = 0;
	}
	push @borders, $intensity;
    }
    push @borders, $max_val;

    foreach my $i (0..$#borders-1) {
	# get the colours for each bin
	$intensity = $borders[$i+1];
	unless (defined $colour{$intensity}) {
 	    my ($red,$green,$blue) = &gradient_grey($intensity,$min_val + ($max_val - $min_val) / $gradient_steps,$max_val);
	    $red = 255 if ($red > 255);
	    $green = 255 if ($green > 255);
	    $blue = 255 if ($blue > 255);

#		$colour{$intensity} = $white; # otherwise I get a black rectangle
	    $colour{$intensity} = $im->colorAllocate($red,$green,$blue);
	}

	# draw rectangles
	if ($flip_image) {
	    $x2 = $x1 + 40;
	} else {
	    $y2 = $y1 + 40;
	}

	$im->filledRectangle($x1,$y1,$x2,$y2,$colour{$intensity});
	$im->rectangle($x1,$y1,$x2,$y2,$black);
	my $text = sprintf "%.1f", $borders[$i];
	if (int($borders[0]) == int($borders[-1])) {
	    $text = sprintf "%.2f", $borders[$i];
	}
	$text =~ s/\.0+$//;
	if (($main_spread_borders
	     or
	     $min_border_set)
	    and
	    $i == 0) {
	    $text = (sprintf "%.1f", $borders[1]);
	    $text =~ s/\.0+$//;
	    $text = "<".$text;
	}
	
	if ($flip_image) {
	    $xt = sprintf "%.0f", $x1 - (length($text) * $legend_font->width / 2);
	    $yt = $y2 + 5;
	} else {
	    $yt = sprintf "%.0f", $y1 - $legend_font->height / 2;
	    $xt = $x2 + 5;
	}

	$im->string($legend_font,$xt,$yt,$text,$black);

	if ($flip_image) {
	    $x1 = $x2;
	} else {
	    $y1 = $y2;
	}
    }

    # last label:
    if ($flip_image) {
	$xt = sprintf "%.0f",$x1 - (length(sprintf "%.0f",$borders[-1]) * $legend_font->width / 2);
	$yt = $y2+5;
    } else {
	$yt = sprintf "%.0f", $y2 - $legend_font->height / 2;
	$xt = $x2 + 5;
    }

    my $text = sprintf "%.1f", $borders[-1];
    $text =~ s/\.0+$//;
    if ($main_spread_borders
	or
	$max_border_set) {
	$text = (sprintf "%.1f", $borders[-2]);
	$text =~ s/\.0+$//;
	$text = ">".$text;
    }
    $im->string($legend_font,$xt,$yt,$text,$black);

    # end of legend


    # draw squares
#    print "all_x: $all_x, all_y: $all_y, space_x: $space_x,space_y: $space_y<BR>\n";
    $counter = 0;
    
    # collect list of real labels
    my @labels = ();
    foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	
	# skip columns that are not label specific ('spot labels')
	next if ($label eq '__NO_LABEL__');

	push @labels, $label;
    }

    my %done = ();

    # cycle through all spots
    foreach my $mrow (1..$MROW) {
	foreach my $mcol (1..$MCOL) {
	    foreach my $row (1..$ROW) {
		foreach my $col (1..$COL) {
		    $done{"$mrow,$mcol,$row,$col"}++;
		    $counter++;
		    if (defined $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col}
			and
			$data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col} =~ /\d/) {
			$intensity = $data{$file}{'__DATA__'}{$label}{$header}{$mrow}{$mcol}{$row}{$col};
			if ($log_transfer) {
			    $intensity = &log_base(2,$intensity);
			}
			

			# place intensity into a bin
			my $border_found = 0;
			foreach my $i (1..$#borders) {
			    if ($intensity <= $borders[$i]) {
				$borders{$borders[$i]}++;
				$border_found = 1;
				$intensity = $borders[$i];
				last;
			    }
			}
			unless ($border_found) {
			    $borders{$borders[-1]}++;
			    $intensity = $borders[-1];
			}
		    } else {
			$intensity = 'undef';
		    }

		    # calculate coordinates for square

		    if ($flip_image) {
			$x1 = $startx + (($mrow-1)*$ROW + $row - 1)*$square_size;
			$y1 = $starty - (($mcol-1)*$COL + $col - 1)*$square_size;
			$y2 = $y1 - $square_size;
		    } else {
			$x1 = $startx + (($mcol-1)*$COL + $col - 1)*$square_size;
			$y1 = $space_y + $square_size + (($mrow-1)*$ROW + $row - 1)*$square_size;
			$y2 = $y1 - $square_size;
		    }

		    $x2 = $x1 + $square_size;

		    # get colour
		    if ($intensity eq 'undef') {
			$colour{'undef'} = $red;
		    } else {
			unless (defined $colour{$intensity}) {
			    my ($red,$green,$blue) = &gradient_grey($intensity,$min_val + ($max_val - $min_val) / $gradient_steps,$max_val);
			    $colour{$intensity} = $im->colorAllocate($red,$green,$blue);
			}
		    }

		    # draw square
		    $im->filledRectangle($x1,$y2,$x2,$y1,$colour{$intensity});

		    # draw flag
		    my $flag = 0;
		    my @flags = ();
		    if (defined $data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			# label specific flags
			@flags = &get_flags($data{$file}{'__DATA__'}{$label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
		    } 
		    if ($label eq '__NO_LABEL__') {
			foreach my $flag_label ('CH1', 'CH2') {
			    push @flags, &get_flags($data{$file}{'__DATA__'}{$flag_label}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
			}
		    } elsif (defined $data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col}) {
			# general flags
			push @flags, &get_flags($data{$file}{'__DATA__'}{'__NO_LABEL__'}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col});
		    }
		    
		    my %tmp = ();
		    foreach (@flags) {
			$tmp{$_}++;
		    }
		    
		    if ($intensity eq 'undef') {
			# unless this spot has been flagged with 'error' already 
			# (and set to '', see sub read_data)
			# we should flag it with 'undefined'
			unless (defined $tmp{$flag_values{'error'}}) {
			    $tmp{$flag_values{'undefined'}}++;
			}
		    }

		    # draw the flag with the lowest value
		    # that is not skipped
		    foreach my $flag_tmp ( sort { $b <=> $a } keys %tmp) {
			next if (defined $skip_flags{$flag_tmp});
			$flag_stats_all{$flag_tmp}++;
			$flag = $flag_tmp;
		    }
		    $flag_stats{$flag}++;

		    if ($flag > 0) {
			if (defined $flag2col_tmp{$flag}) {
			    unless (defined $pref_flag2col{$flag}) {
				$pref_flag2col{$flag2col_tmp{$flag}} = $flag;
				$pref_flag2col{$flag} = $flag2col_tmp{$flag};
			    } 
			} else {
			    if (defined $pref_flag2col{$flag}) {
				$flag2col_tmp{$flag} = $pref_flag2col{$flag};
			    } else {
				foreach my $colour (@pref_flag_colours) {
				    unless (defined $pref_flag2col{$colour}) {
					$flag2col_tmp{$flag} = $colour;
					$pref_flag2col{$colour} = $flag;
					$pref_flag2col{$flag} = $colour;
					last;
				    }
				}
			    }
			}
			
			$intensity = $flag2col_tmp{$flag};
			$im->filledRectangle($x1,$y2,$x2,$y1,$colour{$intensity});
#			    $im->rectangle($x1,$y2,$x2,$y1,$colour{$intensity});
#			    if ($square_size > 3) {
#				$im->rectangle($x1+1,$y2+1,$x2-1,$y1-1,$colour{$intensity});
#			    }
#			} else {
#			    warn "# WARNING: No colour defined for flag '$flag'!";
#			}
		    }
		}
	    }
	}
    }

    # draw grid for subgrids
    my $grid = 'on'; # (draw by default)

    # check for user specific settings
    $grid = &get_user_settings($grid,$prefix.'grid',$force_global_opt,$action_num,$file,$label);

    my $draw_grid = ($grid eq 'off' ? 0 : 1);

    if ($draw_grid == 1) {

	my ($end,$param) = ($MCOL,$COL);
	if ($flip_image) {
	    ($end,$param) = ($MROW,$ROW);
	}

	$y1 = $starty - $square_size * $pix_y - 4;
	$y2 = $starty + 4;    

	for (my $i = 1; $i <= $end; $i++) {
	    $x1 = $startx + $i * $param * $square_size;
	    $x2 = $x1;
	    $im->line($x1,$y1,$x2,$y2,$grey) unless ($i == $end);
	    my $xt = $x1;
	    my $yt = $y1;
	    my $text = $i;
	    $xt -= $param * $square_size/2;
	    $yt -= gdTinyFont->height - 1;
	    if ($flip_image) {
	    }
	    $im->string(gdTinyFont,$xt,$yt,$text,$black);
	}

	($end,$param) = ($MROW,$ROW);
	if ($flip_image) {
	    ($end,$param) = ($MCOL,$COL);
	}

	$x1 = $startx - 4;
	$x2 = $startx + $square_size * $pix_x + 4;

	for (my $k = 1; $k <= $end; $k++) {
	    $y1 = $starty - $k * $param * $square_size;
	    $y2 = $y1;
	    $im->line($x1,$y1,$x2,$y2,$grey) unless ($k == $end);
	    my $xt = $x1;
	    my $yt = $y1;
	    my $text = $k;
	    unless ($flip_image) {
		$text = $end - $k + 1;
	    }
	    $xt -= length ($text) * gdTinyFont->width - 1;
	    $yt += $param * $square_size/2;
	    $im->string(gdTinyFont,$xt,$yt,$text,$black);
	}
    }

    # draw label
    my $elements = $data{$file}{'__SPOTS__'};
#    $im->string($legend_font, 0,0,(basename $file)."($elements spots), $label, $action:",$black);
    $im->string($legend_font, 0,0,(basename $file).", $label, $action:",$black);

    open (IMG, ">$image_out")
        or die "Can't write image $image_out: $!";
    # make sure we are writing to a binary stream
    binmode IMG;
    
    # Convert the image to PNG and print it to filehandle
    print IMG $im->png;

    close IMG;
    
    # write up border info as a HTML table:
    my $border_info = "<TABLE BORDER=0 CELLPADDING=2>\n<TR><TD COLSPAN=2><B>Borders:</B></TD></TR>\n";
    my $border_out = $min_val_orig;
    if ($border_out =~ /\.\d{3,}/) {
	$border_out = sprintf "%.2f", $border_out;
    }
    $border_info .= "<TR><TD>min: </TD><TD>$border_out</TD></TR>\n";

    $border_out = $max_val_orig;
    if ($border_out =~ /\.\d{3,}/) {
	$border_out = sprintf "%.2f", $border_out;
    }
    $border_info .= "<TR><TD>max: </TD><TD>$border_out</TD></TR>\n";

    $border_out = $min_val;
    if ($border_out =~ /\.\d{3,}/) {
	$border_out = sprintf "%.2f", $border_out;
    }
    $border_info .= "<TR><TD>'$min_val_input'</TD><TD>-> $border_out</TD></TR>\n";

    $border_out = $max_val;
    if ($border_out =~ /\.\d{3,}/) {
	$border_out = sprintf "%.2f", $border_out;
    }
    $border_info .= "<TR><TD>'$max_val_input'</TD><TD>-> $border_out</TD></TR>\n";

    $border_info .= "</TABLE>\n";

    
    # write up bin info as a HTML table:
    my $range_info .= "<TABLE BORDER=1 CELLPADDING=2>\n<TR><TH>range</TH><TH>#</TH></TR>\n";
    my $sum = 0;
    my $decimal = 1;
    if (int($borders[0]) == int($borders[-1])) {
	$decimal = 2;
    }
    foreach my $i (0..$#borders-1) {
	my ($lower,$upper) = ((sprintf "%.${decimal}f",$borders[$i]), (sprintf "%.${decimal}f",$borders[$i+1]));
	$lower =~ s/\.0+$//;
	$upper =~ s/\.0+$//;

	if ($i == 0
	    and
	    ($main_spread_borders
	     or
	     $min_border_set)
	    ) {
	    $range_info .= "  <TR><TD>&lt;$upper</TD><TD>$borders{$borders[$i+1]}</TD></TR>\n";
	} elsif ($i == $#borders-1
		 and
		 ($main_spread_borders
		  or
		  $max_border_set)) {
	    $range_info .= "  <TR><TD>&gt;$lower</TD><TD>$borders{$borders[$i+1]}</TD></TR>\n";
	} else {
	    $range_info .= "  <TR><TD>$lower\&nbsp;-\&nbsp;$upper</TD><TD>".
		(defined $borders{$borders[$i+1]} ? $borders{$borders[$i+1]} : '').
		"</TD></TR>\n";
	}
	$sum += $borders{$borders[$i+1]} if (defined $borders{$borders[$i+1]});
    }
    $range_info .= "<TR><TD>total:</TD><TD>$sum</TD></TR>\n";
    $range_info .= " </TABLE>\n";
    if ($min_border_set
	or
	$main_spread_borders) {
	$min_val = $borders[1];
    }
    if ($max_border_set
	or
	$main_spread_borders) {
	$max_val = $borders[-2];
    }
    
    # chop off trailing zero's
    if ($min_val =~ /\./) {
	$min_val =~ s/0+$//g;
	$min_val =~ s/\.$//;
    }
    if ($max_val =~ /\./) {
	$max_val =~ s/0+$//g;
	$max_val =~ s/\.$//;
    }


    # add stats for flags if necessary
    my $unflagged = $flag_stats{0};

    my $flag_info = '';

    if (%flag_stats) {
	$flag_info .=  "\n<BR>
     <TABLE>
      <TR>
       <TH>flag</TH><TH>#</TH><TH>colour</TH>
      </TR>\n";

	if ($unflagged) {
	    $flag_info .= "<TR ALIGN=RIGHT><TD>0\&nbsp;=\&nbsp;no\&nbsp;flag:</TD><TD>$unflagged</TD><TD ALIGN=CENTER>gradient</TD></TR>\n";
	}

	foreach my $flag (sort { $a <=> $b } keys %flag_stats_all) {
	    
	    # make field to choose colour:
	    my $colour_field = '';
	    if ($flag > 0) {
		$colour_field = "<SELECT name=\"action$action_num$file$label${prefix}flag_colour$flag\" size=\"1\">";
		foreach my $col (@pref_flag_colours) {
		    my $selected = '';
		    if (defined $flag2col_tmp{$flag}
			and
			$flag2col_tmp{$flag} eq $col) {
			$selected = ' SELECTED';
		    }
		    $colour_field .= "<OPTION$selected>$col</OPTION>\n";
		}
		$colour_field .= "\n</select>\n";
	    }
	    
	    $flag_info .=  "
      <TR>
       <TD ALIGN=RIGHT>$flag\&nbsp;=\&nbsp;$flag_values_num{$flag}:</TD><TD ALIGN=RIGHT>"
       .(defined $flag_stats{$flag} ? $flag_stats{$flag} : '0')
       .(defined $flag_stats_all{$flag} ? "\&nbsp;($flag_stats_all{$flag})" : '')
       ."</TD><TD ALIGN=RIGHT>$colour_field</TD>
      </TR>";
	}
	$flag_info .=  "
     </TABLE>
";
    }

    my $info_out = $border_info.$range_info.$flag_info;

    if ($orientation eq 'landscape') {
	$info_out = '<TABLE><TR><TD>'.$range_info.'</TD><TD>'.$border_info.$flag_info.'</TD></TR></TABLE>';
#	$info_out = $range_info.$flag_info.$border_info;
    }

    return $image_name, $info_out, $min_val, $max_val;
}


sub ratio_calc {
    # calculate ratios from specified column header
    # subtract values if normalized already
    # otherwise divide one channel by the other
    # write ratios into specified column
    my $action_num = shift;
    my $file = shift;
    my $action = shift;
    my $out_label = shift;

    my $force_global_opt = 0;
    if (defined $cgi->param("${action}_force")
	and
	$cgi->param("${action}_force") == 1) {
	$force_global_opt = 1;
    }

    my $header = '';
    foreach my $column (@{$item_data_column{$action}}) {

	my $occ = 0;
	foreach my $label (keys %{$data{$file}{'__DATA__'}}) {
	    next if ($label eq '__NO_LABEL__');
	    if ($data_column_available{$file}{$label}{$column}) {
		$header = $column;
		$occ++;
	    }
	}
	if ($occ == 2) {
	    last;
	} else {
	    $header = '';
	}
    }
    # check for specific settings:
    $header = &get_user_settings($header,'ratio_data_column',$force_global_opt,$action_num,$file);

    my $logtransferred = 0;
    if ($header =~ /_L/) {
	$logtransferred = 1;
    }

    $last_column_used = $header;

    my $out_header = $header.'_R';

    # collect labels
    my %labels = ();

    foreach my $label (sort keys %{$data{$file}{'__DATA__'}}) {
	
	# skip columns that are not label specific ('spot labels')
	next if ($label eq '__NO_LABEL__');
	$labels{$label}++;
    }

    if (keys %labels > 2) {
	die "# ERROR: More than 2 labels defined for ".(basename $file)." - can't deal with that, yet!\n";
    }

    my $flag_no_ratio = &get_user_settings($var_def_default{'ratio_flag_no_ratio'},'ratio_flag_no_ratio',$force_global_opt,$action_num,$file,'__ALL__');;
    $flag_no_ratio = 1;

    # get labels for channel 1 (
    my $ch1 = $var_def_default{'ratio_ch1'};
    
    # check for user specific settings
    $ch1 = &get_user_settings($ch1,'ratio_ch1',$force_global_opt,$action_num,$file);

    unless ($labels{$ch1}) {
	warn "# ERROR: Label '$ch1' is not defined for ".(basename $file)."!";
	die "# ERROR: Label '$ch1' is not defined for ".(basename $file)."!\n";
    }

    my $ch2 = '';
    foreach (keys %labels) {
	unless ($ch1 eq $_) {
	    $ch2 = $_;
	    last;
	}
    }
    unless ($ch2) {
	die "# ERROR: No channel 2 defined for ".(basename $file)."!\n";
    }

    my ($mrow,$mcol,$row,$col);
    my $counter = 0;

    # if specified, raw data will be log-transferred
    my $log_trans_ratios = $var_def_default{'log_trans_ratios'};

    # check for user specific settings
    $log_trans_ratios = &get_user_settings($log_trans_ratios,'log_trans_ratios',$force_global_opt,$action_num,$file);
    $log_trans_ratios = ($log_trans_ratios eq 'no' ? 0 : 1);
    if ($log_trans_ratios) {
	$out_header =~ s/_R$/_L_R/;
    }

    my $skipped = 0;

    # calculate ratio and store in array
    foreach $mrow (keys %{$data{$file}{'__DATA__'}{$ch1}{$header}}) {
	foreach $mcol (keys %{$data{$file}{'__DATA__'}{$ch1}{$header}{$mrow}}) {
	    foreach $row (keys %{$data{$file}{'__DATA__'}{$ch1}{$header}{$mrow}{$mcol}}) {
		foreach $col (keys %{$data{$file}{'__DATA__'}{$ch1}{$header}{$mrow}{$mcol}{$row}}) {

		    # retrieve values from both channels
		    my ($val1,$val2,$ratio);

		    if (defined $data{$file}{'__DATA__'}{$ch1}{$header}{$mrow}{$mcol}{$row}{$col}) {
			$val1 = $data{$file}{'__DATA__'}{$ch1}{$header}{$mrow}{$mcol}{$row}{$col};
		    }
		    if (defined $data{$file}{'__DATA__'}{$ch2}{$header}{$mrow}{$mcol}{$row}{$col}) {
			$val2 = $data{$file}{'__DATA__'}{$ch2}{$header}{$mrow}{$mcol}{$row}{$col};
		    }

		    # skip unless both values set
		    # (in case of normalization on unflagged values only)
		    unless ($val1 =~ /\d/
			    and
			    $val2 =~ /\d/) {
			$skipped++;
			next;
		    }

		    if ($logtransferred) {
			$ratio = $val1 - $val2;
		    } else {
			if ($val2 == 0) {
			    $ratio = '';
			    $skipped++;
			    # set flag
			    if ($flag_no_ratio) {
				$data{$file}{'__DATA__'}{$ch1}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'no_ratio'};
				$data{$file}{'__DATA__'}{$ch2}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'no_ratio'};
			    }
			} else {
			    $ratio = $val1 / $val2;
			}
			if ($log_trans_ratios
			    and
			    $ratio =~ /\d/) {
			     if ($ratio <= 0) {
				 $ratio = '';
				 $skipped++;
				 if ($flag_no_ratio) {
				     $data{$file}{'__DATA__'}{$ch1}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'no_ratio'};
				     $data{$file}{'__DATA__'}{$ch2}{$FLAG_CHANNEL_TERM}{$mrow}{$mcol}{$row}{$col} += $flag_values{'no_ratio'};
				 }
			     } else {
				 $ratio = &log_base(2,$ratio);
			     }
			}
		    }
		    
		    # store ratio in specified destination
		    $data{$file}{'__DATA__'}{$out_label}{$out_header}{$mrow}{$mcol}{$row}{$col} = $ratio if ($ratio =~ /\d/);
		}
	    }
	}
    }

    # take note of available columns for later

    unless (defined $data_column_available{$file}{$out_label}{$out_header}) {
	$data_column_available{$file}{$out_label}{$out_header} = 1;
	push @{$data_column_available{$file}{$out_label}{'__ORDER__'}}, $out_header;
    }
    $data_done{$file}{$out_label}{$out_header}++;
    
    my $info = '';
    if ($skipped) {
	$info = "$skipped ratios could not be calculated,<BR>due to invalid data ranges!<BR>\n";
    }
    return $out_header,$info;
}


sub log_base {
    my $base = shift;
    my $n = shift;

#    if ($n == 0) {
#	return 'Error: no log of 0 possible!\n';
#    }

    return log($n)/log($base);
}


sub get_user_settings {
    # checks for user specific settings for a variable
    # on different levels (global, action, file, label)

    my $default = shift || '';

    # special treatment of default for working files:
    unless (%added_parameters) {
	foreach (sort keys %work_files) {
	    push @{$added_parameters{'work_files'}}, $_;
	}
	foreach (keys %work_files) {
	    push @{$added_parameters{'work_files_base'}}, ($_, basename $_);
	}
    }


    if ($default =~ /^_X__/) {
	my @tmp = split /_X__/, $default;
	my @tmp2 = ();
	foreach (@tmp) {
	    next unless (/\w/);
	    if (/^\@(.*)/) {
		push @tmp2, @{$added_parameters{$1}};
	    } else {
		push @tmp2, $_;
	    }
	}
	$default = \@tmp2;
    }

    my @new_val = ($default);
    my @prev_val = ();
    if (ref($default) eq 'ARRAY') {
	@new_val = @$default;
    }
    my $var = shift;
    my $force_global_opt = shift;

    my ($action_num,$file,$label);

    # special action for single-check boxes that are switched on by default
#    unless (defined $var_def_default{$var}) {
#	warn "# WARNING: No default for $var!\n";
#    }
    if (defined $var
	and
	defined $var_def_type{$var}
	and
	$var_def_type{$var} eq 'check_single'
	and
	$var_def_default{$var} == 1) {
	unless (defined $cgi->param($var)) {
	    $cgi->param(-name=>$var, -value=>0);
	}
    }

    # check for global setting
    if (defined $cgi->param($var)
	and
	($cgi->param($var) =~ /\w/
	 or
	 $cgi->param($var) =~ /[-|<|>|=|\/]/)
	and
	($cgi->param($var) ne $default
	 or
	 $force_global_opt)
	) {
	@new_val = $cgi->param($var);
    }
    
    
    if ($force_global_opt) {
	if (@new_val == 1) {
	    return $new_val[0];
	}
	return @new_val;
    }
    
    unless (@_) {
	if (@new_val == 1) {
	    return $new_val[0];
	}
	return @new_val;
    } else {
	@prev_val = @new_val;
    }

    # this can be overwritten with a value set per action step
    $action_num = shift;
    if (defined $cgi->param("action$action_num$var")
	and
	($cgi->param("action$action_num$var") =~ /\w/
	 or
	 $cgi->param("action$action_num$var") =~ /[-|<|>|=|\/]/)
#	and
#	$cgi->param("action$action_num$var") ne $default
	) {
	@new_val = $cgi->param("action$action_num$var");

	# during parameter refinement the value 'default' can be chosen for data column
	# in which case the value from the previous selection section is to be used:
	if ($new_val[0] eq 'default') {
	    $new_val[0] = $prev_val[0];
	}
    }

    unless (@_) {
	if (@new_val == 1) {
	    return $new_val[0];
	}
	return @new_val;
    } else {
	@prev_val = @new_val if (@new_val);
    }

    # this can be overwritten with a user defined value (set per file)
    $file = shift;
    if (defined $cgi->param("action$action_num$file$var")
	and
	($cgi->param("action$action_num$file$var") =~ /\w/
	 or
	 $cgi->param("action$action_num$file$var") =~ /[-|<|>|=|\/]/)
#	and
#	$cgi->param("action$action_num$file$var") ne $default
	) {
	@new_val = $cgi->param("action$action_num$file$var");
 
	# during parameter refinement the value 'default' can be chosen for data column
	# in which case the value from the previous selection section is to be used:
	if ($new_val[0] eq 'default') {
	    $new_val[0] = $prev_val[0];
	}
    }
    
    unless (@_) {
	if (@new_val == 1) {
	    return $new_val[0];
	}
	return @new_val;
    } else {
	@prev_val = @new_val if (@new_val);
    }

#    if ($_[0] eq '') {
	unless ($file =~ /\w/) {
	    if (@new_val == 1) {
		return $new_val[0];
	    }
	    return @new_val;
	} else {
	    @prev_val = @new_val if (@new_val);
	}
#    }
    
    # this can be overwritten with a user defined value (set per label)
    $label = shift;
    my @labels = ($label);
    if ($label eq '__ALL__') {
	foreach my $label_tmp (keys %{$data{$file}{'__DATA__'}}) {
	    push @labels, $label_tmp;
	}
    }
    foreach my $label (@labels) {
	if (defined $cgi->param("action$action_num$file$label$var")
	    and
	    ($cgi->param("action$action_num$file$label$var") =~ /\w/
	     or
	     $cgi->param("action$action_num$file$label$var") =~ /[-|<|>|=|\/]/)
	    and
	    $cgi->param("action$action_num$file$label$var") ne $default
	    ) {
	    @new_val = $cgi->param("action$action_num$file$label$var");

	    # during parameter refinement the value 'default' can be chosen for data column
	    # in which case the value from the previous selection section is to be used:
	    if ($new_val[0] eq 'default') {
		$new_val[0] = $prev_val[0];
	    }

	    last;
	}
    }

    if (@new_val == 1) {
	return $new_val[0];
    }
    return @new_val;
}

    
sub boxplot_values {
    # calculates median, upper and lower quartiles
    # from an array of values
    # also returns min and max value
    my $ref = shift;
    my @tmp = @$ref;
    my $md = '0';
    my $uq = '0';
    my $lq = '0';

    @tmp = sort { $a <=> $b } @tmp;
    if ($#tmp % 2 == 0){
        $md = $tmp[$#tmp / 2];
    } else {
	my $index1 = int($#tmp / 2);
	my $index2 = int($#tmp / 2) + 1;
        $md = (abs($tmp[$index1]) + abs($tmp[$index2])) / 2;
	if ($tmp[$index1] < 0
            and
            $tmp[$index2] > 0
            and
            abs($tmp[$index1]) > $tmp[$index2]) {
            $md *= -1;
	}
    }
#    print "Median: $md\n";

    if ($#tmp % 4 == 0){
        $uq = $tmp[$#tmp * 3 / 4];
        $lq = $tmp[$#tmp / 4];
    }
    else{
	my $index1 = int($#tmp / 4);
	my $index2 = int($#tmp / 4) + 1;
        $lq = (abs($tmp[$index1]) + abs($tmp[$index2])) / 2;
	if ($tmp[$index1] < 0
            and
            $tmp[$index2] > 0
            and
            abs($tmp[$index1]) > $tmp[$index2]) {
            $lq *= -1;
	}
	$index1 = int($#tmp * 3 / 4);
	$index2 = int($#tmp * 3 / 4) + 1;
        $uq = ($tmp[$index1] + $tmp[$index2]) / 2;
	if ($tmp[$index1] < 0
            and
            $tmp[$index2] > 0
            and
            abs($tmp[$index1]) > $tmp[$index2]) {
            $uq *= -1;
	}
    }
#    print "lq: $lq\nuq: $uq\n";

    return ($tmp[-1],$md,$lq,$uq,$tmp[0]);
}

sub gradient {

    # returns a colour gradient
    # according to a value and the upper and lower boundaries

    my $rat = shift;
    my $inlow = shift || 0;
    my $high = shift || 1;
    
    # shift everything to start with 0:
    $high -= $inlow;
    $high = 1 if ($high == 0);

    $rat -= $inlow;
    my $low = 0;
    if ($rat < 0) {
        $rat = 0;
    } elsif ($rat > $high) {
        $rat = $high;
    }

    # scale everyting down to 1
    $rat /= $high;
    $high /= $high;

    my $half = $high/2;
    my $low_col = 256;
    my $high_col = 256;
    if ($rat < $half) {
	$low_col = 256;
        $high_col = int (256 - (256 * 2 * ($half - $rat)));
    } elsif ($rat > $half) {
	$high_col = 256;
        $low_col = int (256 - (256 * 2 * ($rat - $half)));
    }

    $low_col-- unless ($low_col == 0);
    $high_col-- unless ($high_col == 0);

    # change position of '00' to vary between red,green,blue
    return ('00',$low_col,$high_col);
}


sub gradient_grey {

    # returns a grey-colour gradient
    # according to a value and the upper and lower boundaries

    my $rat = shift;
    my $inlow = shift || 0;
    my $high = shift || 1;

#    my $warn = "#KH DEBUG: $rat, $inlow - $high => ";
    
    # shift everything to start with 0:
    $high -= $inlow;
    $high = 1 if ($high == 0);

    $rat -= $inlow;
    my $low = 0;
    if ($rat < 0) {
        $rat = 0;
    } elsif ($rat > $high) {
        $rat = $high;
    }

    # scale everyting down to 1
    $rat /= $high;
    $high /= $high;

    my $col = int (256 * $rat);

#    $warn .= "($col,$col,$col)\n";
#    warn $warn;

    return ($col,$col,$col);
}

sub ending {
    my @options = @_;
    my $out = '';

    $time2 = time;
    $time_diff = $time2 - $time1;
    my $overall_time_diff = $time_diff;
    my $hostname = `hostname`;
    chomp $hostname;
    my $time_out = '';

    # calculate processing time:
    if (@options
	and
	$options[$#options] eq 'show_processing_time') {
	$time_out = "\nOverall processing time:";
	my ($hours,$min);
	if ($time_diff >= 3600) {
	    $hours = int($time_diff / 3600);
	    $time_diff = $time_diff % 3600;
	    $time_out .= " $hours hours";
	}
	if ($time_diff > 60) {
	    $min = int($time_diff / 60);
	    $time_out .= " $min minutes";
	    $time_diff = $time_diff % 60;
	}
	if ($time_diff >= 0) {
	    if ($time_diff < 1
		and
		$overall_time_diff < 60) {
		$time_out .=  " < 1 second";
	    } else {
		my $s = 's';
		if ($time_diff == 1) {
		    $s = '';
		}
		$time_out .=  " $time_diff second$s";
	    }
	}
	
	$time_out .= " on $hostname";
    }

	
    $out .= "<BR>\n";

    my ($url,$cmd_config_file2) = &get_cgi_params;
    $url =~ s/ /\+/g;
    my $url_out = $url;
#    $url_out =~ s/(\&.*?=)/<FONT COLOR=RED>$1<\/FONT>/g;
    # (makes output too long)
    $url_out = 'right click and copy';
    $url_out = 0;
    if ($url_out) {
	$out .= "<BR>URL (for debugging): <A HREF=\"$url\">$url_out</A><BR>\n";
    }

    $out .= "<BR>
        $time_out\n<BR>
        <CENTER>
        ";

    my %link = ('home', "<A HREF=\"$home_link\">home</A>",
		'tutorial', "<A HREF=\"$tutorial_link\">tutorial</A>",
		'start over', "<A HREF=\"$cgi_link\">start over</A>",
		);
    if ($FPMI_specific) {
	if ($progname =~ /FPMI/i) {
	    $link{'start over'} =~ s/\.pl/_FPMI\.pl/;
	}
    }

    foreach (@options) {
	next unless (defined $link{$_});
        $out .= "[$link{$_}]&nbsp;&nbsp;\n";
    }
    $out .= "</CENTER>
        <HR>
        <IMG SRC=\"$project_logo\" ALIGN=\"RIGHT\">
        <address>automatically generated at ".(localtime)."
        <BR>for questions or remarks e-mail <A HREF=\"MAILTO:$operator_email\">$operator_email</A>.
        </address>
        </DIV>
        </BODY>
        </HTML>
";
    return $out;
}

sub html_head_refresh {

    # creates HTML head for returned document
    # parameter title is mandatory
    # optional parameters:
    # - expire time
    # - cookie1 and cookie2

    my $title = shift;
    my $refresh = shift;
    unless ($refresh =~ /\d/) {
	$refresh = 2;
    }
    my $expires = shift || '-1';
    my $node = shift || '';

    my $out = '';

#    my $css_style = '<style type="text/css">

    $out .= $cgi->start_html(-title=>"$title",
			   -author=>"$operator_email",
			   -BGCOLOR=>'white',
		           -style=>{-code=>$css_style},
			   -script=>$help_javascript,
                           -head=>[meta({-http_equiv => 'refresh',
                                           -content  => $refresh}),
                                   meta({-http_equiv => 'Pragma',
                                           -content  =>'no-cache'}),
                                   meta({-http_equiv => 'Expires',
                                           -content  =>'Tue, 04 Dec 1993 21:29:02 GMT'})],
  );

    my $hostname = `hostname`;
    chomp $hostname;
    unless ($node) {
	$node = $hostname;
    }
# doesn't get a value yet:
# Total waiting time approx. $waiting_time seconds.
    $refresh =~ s/^(\d+).*/$1/;
#    $out .= "<BR><B><FONT COLOR=RED>Data processing in progress (on node $node) - page will refresh automatically after $refresh seconds!<BR>(Partial results might be displayed already.)<BR>Please click 'Reload/Refresh' if page comes up blank!</FONT>".&help('refresh')."</B><BR><BR>\n";
    $node = (split /\./, $node)[0];
    $out .= "\n<H1>ArrayPipe Progress Status</H1>\n<TABLE CELLSPACING=10 BORDER=2 bgcolor=\"#FFFFCC\" bordercolor=\"#006699\" WIDTH=750>\n<TR>\n<TD><BIG>\n<B><FONT COLOR=BLACK>Data processing in progress (on node $node) - page will refresh automatically after ".(&time_formated($refresh))."!<P>Final results will become available at <BR><BR><A HREF=\"$results_file_html\">$results_file_html</A>\n<P>To see partial results as they come in you can click on that link.</FONT>"
#."<BR><FONT COLOR=red>If an error page appears, please reload!</FONT>"
.&help('refresh')."</B>\n</BIG></TD>\n</TR>\n</TABLE>\n<BR><BR>\n";

    return $out;
}


sub html_head {

    # creates HTML head for returned document
    # parameter title is mandatory
    # optional parameters:
    # - expire time
    # - cookie1 and cookie2

    my $title = shift;
    my $expires = shift || '';
    my $out = '';

    $out .= $cgi->start_html(-title=>"$title",
			     -author=>"$operator_email",
			     -script=>$help_javascript,
			     -BGCOLOR=>'white',			     
			     -style=>{-code=>$css_style}
                           );
    return $out;
}


__END__


- changes:
  * 

- bug fixes:
  * 

- started:
  *

